package using_serviceloader;
public class DemoImpls {
    public static class Impl0 implements DemoService {public int type() {return 0;} public void run() {System.out.println(0);}}
    public static class Impl1 implements DemoService {public int type() {return 1;} public void run() {System.out.println(1);}}
    public static class Impl2 implements DemoService {public int type() {return 2;} public void run() {System.out.println(2);}}
    public static class Impl3 implements DemoService {public int type() {return 3;} public void run() {System.out.println(3);}}
    public static class Impl4 implements DemoService {public int type() {return 4;} public void run() {System.out.println(4);}}
    public static class Impl5 implements DemoService {public int type() {return 5;} public void run() {System.out.println(5);}}
    public static class Impl6 implements DemoService {public int type() {return 6;} public void run() {System.out.println(6);}}
    public static class Impl7 implements DemoService {public int type() {return 7;} public void run() {System.out.println(7);}}
    public static class Impl8 implements DemoService {public int type() {return 8;} public void run() {System.out.println(8);}}
    public static class Impl9 implements DemoService {public int type() {return 9;} public void run() {System.out.println(9);}}
    public static class Impl10 implements DemoService {public int type() {return 10;} public void run() {System.out.println(10);}}
    public static class Impl11 implements DemoService {public int type() {return 11;} public void run() {System.out.println(11);}}
    public static class Impl12 implements DemoService {public int type() {return 12;} public void run() {System.out.println(12);}}
    public static class Impl13 implements DemoService {public int type() {return 13;} public void run() {System.out.println(13);}}
    public static class Impl14 implements DemoService {public int type() {return 14;} public void run() {System.out.println(14);}}
    public static class Impl15 implements DemoService {public int type() {return 15;} public void run() {System.out.println(15);}}
    public static class Impl16 implements DemoService {public int type() {return 16;} public void run() {System.out.println(16);}}
    public static class Impl17 implements DemoService {public int type() {return 17;} public void run() {System.out.println(17);}}
    public static class Impl18 implements DemoService {public int type() {return 18;} public void run() {System.out.println(18);}}
    public static class Impl19 implements DemoService {public int type() {return 19;} public void run() {System.out.println(19);}}
    public static class Impl20 implements DemoService {public int type() {return 20;} public void run() {System.out.println(20);}}
    public static class Impl21 implements DemoService {public int type() {return 21;} public void run() {System.out.println(21);}}
    public static class Impl22 implements DemoService {public int type() {return 22;} public void run() {System.out.println(22);}}
    public static class Impl23 implements DemoService {public int type() {return 23;} public void run() {System.out.println(23);}}
    public static class Impl24 implements DemoService {public int type() {return 24;} public void run() {System.out.println(24);}}
    public static class Impl25 implements DemoService {public int type() {return 25;} public void run() {System.out.println(25);}}
    public static class Impl26 implements DemoService {public int type() {return 26;} public void run() {System.out.println(26);}}
    public static class Impl27 implements DemoService {public int type() {return 27;} public void run() {System.out.println(27);}}
    public static class Impl28 implements DemoService {public int type() {return 28;} public void run() {System.out.println(28);}}
    public static class Impl29 implements DemoService {public int type() {return 29;} public void run() {System.out.println(29);}}
    public static class Impl30 implements DemoService {public int type() {return 30;} public void run() {System.out.println(30);}}
    public static class Impl31 implements DemoService {public int type() {return 31;} public void run() {System.out.println(31);}}
    public static class Impl32 implements DemoService {public int type() {return 32;} public void run() {System.out.println(32);}}
    public static class Impl33 implements DemoService {public int type() {return 33;} public void run() {System.out.println(33);}}
    public static class Impl34 implements DemoService {public int type() {return 34;} public void run() {System.out.println(34);}}
    public static class Impl35 implements DemoService {public int type() {return 35;} public void run() {System.out.println(35);}}
    public static class Impl36 implements DemoService {public int type() {return 36;} public void run() {System.out.println(36);}}
    public static class Impl37 implements DemoService {public int type() {return 37;} public void run() {System.out.println(37);}}
    public static class Impl38 implements DemoService {public int type() {return 38;} public void run() {System.out.println(38);}}
    public static class Impl39 implements DemoService {public int type() {return 39;} public void run() {System.out.println(39);}}
    public static class Impl40 implements DemoService {public int type() {return 40;} public void run() {System.out.println(40);}}
    public static class Impl41 implements DemoService {public int type() {return 41;} public void run() {System.out.println(41);}}
    public static class Impl42 implements DemoService {public int type() {return 42;} public void run() {System.out.println(42);}}
    public static class Impl43 implements DemoService {public int type() {return 43;} public void run() {System.out.println(43);}}
    public static class Impl44 implements DemoService {public int type() {return 44;} public void run() {System.out.println(44);}}
    public static class Impl45 implements DemoService {public int type() {return 45;} public void run() {System.out.println(45);}}
    public static class Impl46 implements DemoService {public int type() {return 46;} public void run() {System.out.println(46);}}
    public static class Impl47 implements DemoService {public int type() {return 47;} public void run() {System.out.println(47);}}
    public static class Impl48 implements DemoService {public int type() {return 48;} public void run() {System.out.println(48);}}
    public static class Impl49 implements DemoService {public int type() {return 49;} public void run() {System.out.println(49);}}
    public static class Impl50 implements DemoService {public int type() {return 50;} public void run() {System.out.println(50);}}
    public static class Impl51 implements DemoService {public int type() {return 51;} public void run() {System.out.println(51);}}
    public static class Impl52 implements DemoService {public int type() {return 52;} public void run() {System.out.println(52);}}
    public static class Impl53 implements DemoService {public int type() {return 53;} public void run() {System.out.println(53);}}
    public static class Impl54 implements DemoService {public int type() {return 54;} public void run() {System.out.println(54);}}
    public static class Impl55 implements DemoService {public int type() {return 55;} public void run() {System.out.println(55);}}
    public static class Impl56 implements DemoService {public int type() {return 56;} public void run() {System.out.println(56);}}
    public static class Impl57 implements DemoService {public int type() {return 57;} public void run() {System.out.println(57);}}
    public static class Impl58 implements DemoService {public int type() {return 58;} public void run() {System.out.println(58);}}
    public static class Impl59 implements DemoService {public int type() {return 59;} public void run() {System.out.println(59);}}
    public static class Impl60 implements DemoService {public int type() {return 60;} public void run() {System.out.println(60);}}
    public static class Impl61 implements DemoService {public int type() {return 61;} public void run() {System.out.println(61);}}
    public static class Impl62 implements DemoService {public int type() {return 62;} public void run() {System.out.println(62);}}
    public static class Impl63 implements DemoService {public int type() {return 63;} public void run() {System.out.println(63);}}
    public static class Impl64 implements DemoService {public int type() {return 64;} public void run() {System.out.println(64);}}
    public static class Impl65 implements DemoService {public int type() {return 65;} public void run() {System.out.println(65);}}
    public static class Impl66 implements DemoService {public int type() {return 66;} public void run() {System.out.println(66);}}
    public static class Impl67 implements DemoService {public int type() {return 67;} public void run() {System.out.println(67);}}
    public static class Impl68 implements DemoService {public int type() {return 68;} public void run() {System.out.println(68);}}
    public static class Impl69 implements DemoService {public int type() {return 69;} public void run() {System.out.println(69);}}
    public static class Impl70 implements DemoService {public int type() {return 70;} public void run() {System.out.println(70);}}
    public static class Impl71 implements DemoService {public int type() {return 71;} public void run() {System.out.println(71);}}
    public static class Impl72 implements DemoService {public int type() {return 72;} public void run() {System.out.println(72);}}
    public static class Impl73 implements DemoService {public int type() {return 73;} public void run() {System.out.println(73);}}
    public static class Impl74 implements DemoService {public int type() {return 74;} public void run() {System.out.println(74);}}
    public static class Impl75 implements DemoService {public int type() {return 75;} public void run() {System.out.println(75);}}
    public static class Impl76 implements DemoService {public int type() {return 76;} public void run() {System.out.println(76);}}
    public static class Impl77 implements DemoService {public int type() {return 77;} public void run() {System.out.println(77);}}
    public static class Impl78 implements DemoService {public int type() {return 78;} public void run() {System.out.println(78);}}
    public static class Impl79 implements DemoService {public int type() {return 79;} public void run() {System.out.println(79);}}
    public static class Impl80 implements DemoService {public int type() {return 80;} public void run() {System.out.println(80);}}
    public static class Impl81 implements DemoService {public int type() {return 81;} public void run() {System.out.println(81);}}
    public static class Impl82 implements DemoService {public int type() {return 82;} public void run() {System.out.println(82);}}
    public static class Impl83 implements DemoService {public int type() {return 83;} public void run() {System.out.println(83);}}
    public static class Impl84 implements DemoService {public int type() {return 84;} public void run() {System.out.println(84);}}
    public static class Impl85 implements DemoService {public int type() {return 85;} public void run() {System.out.println(85);}}
    public static class Impl86 implements DemoService {public int type() {return 86;} public void run() {System.out.println(86);}}
    public static class Impl87 implements DemoService {public int type() {return 87;} public void run() {System.out.println(87);}}
    public static class Impl88 implements DemoService {public int type() {return 88;} public void run() {System.out.println(88);}}
    public static class Impl89 implements DemoService {public int type() {return 89;} public void run() {System.out.println(89);}}
    public static class Impl90 implements DemoService {public int type() {return 90;} public void run() {System.out.println(90);}}
    public static class Impl91 implements DemoService {public int type() {return 91;} public void run() {System.out.println(91);}}
    public static class Impl92 implements DemoService {public int type() {return 92;} public void run() {System.out.println(92);}}
    public static class Impl93 implements DemoService {public int type() {return 93;} public void run() {System.out.println(93);}}
    public static class Impl94 implements DemoService {public int type() {return 94;} public void run() {System.out.println(94);}}
    public static class Impl95 implements DemoService {public int type() {return 95;} public void run() {System.out.println(95);}}
    public static class Impl96 implements DemoService {public int type() {return 96;} public void run() {System.out.println(96);}}
    public static class Impl97 implements DemoService {public int type() {return 97;} public void run() {System.out.println(97);}}
    public static class Impl98 implements DemoService {public int type() {return 98;} public void run() {System.out.println(98);}}
    public static class Impl99 implements DemoService {public int type() {return 99;} public void run() {System.out.println(99);}}
    public static class Impl100 implements DemoService {public int type() {return 100;} public void run() {System.out.println(100);}}
    public static class Impl101 implements DemoService {public int type() {return 101;} public void run() {System.out.println(101);}}
    public static class Impl102 implements DemoService {public int type() {return 102;} public void run() {System.out.println(102);}}
    public static class Impl103 implements DemoService {public int type() {return 103;} public void run() {System.out.println(103);}}
    public static class Impl104 implements DemoService {public int type() {return 104;} public void run() {System.out.println(104);}}
    public static class Impl105 implements DemoService {public int type() {return 105;} public void run() {System.out.println(105);}}
    public static class Impl106 implements DemoService {public int type() {return 106;} public void run() {System.out.println(106);}}
    public static class Impl107 implements DemoService {public int type() {return 107;} public void run() {System.out.println(107);}}
    public static class Impl108 implements DemoService {public int type() {return 108;} public void run() {System.out.println(108);}}
    public static class Impl109 implements DemoService {public int type() {return 109;} public void run() {System.out.println(109);}}
    public static class Impl110 implements DemoService {public int type() {return 110;} public void run() {System.out.println(110);}}
    public static class Impl111 implements DemoService {public int type() {return 111;} public void run() {System.out.println(111);}}
    public static class Impl112 implements DemoService {public int type() {return 112;} public void run() {System.out.println(112);}}
    public static class Impl113 implements DemoService {public int type() {return 113;} public void run() {System.out.println(113);}}
    public static class Impl114 implements DemoService {public int type() {return 114;} public void run() {System.out.println(114);}}
    public static class Impl115 implements DemoService {public int type() {return 115;} public void run() {System.out.println(115);}}
    public static class Impl116 implements DemoService {public int type() {return 116;} public void run() {System.out.println(116);}}
    public static class Impl117 implements DemoService {public int type() {return 117;} public void run() {System.out.println(117);}}
    public static class Impl118 implements DemoService {public int type() {return 118;} public void run() {System.out.println(118);}}
    public static class Impl119 implements DemoService {public int type() {return 119;} public void run() {System.out.println(119);}}
    public static class Impl120 implements DemoService {public int type() {return 120;} public void run() {System.out.println(120);}}
    public static class Impl121 implements DemoService {public int type() {return 121;} public void run() {System.out.println(121);}}
    public static class Impl122 implements DemoService {public int type() {return 122;} public void run() {System.out.println(122);}}
    public static class Impl123 implements DemoService {public int type() {return 123;} public void run() {System.out.println(123);}}
    public static class Impl124 implements DemoService {public int type() {return 124;} public void run() {System.out.println(124);}}
    public static class Impl125 implements DemoService {public int type() {return 125;} public void run() {System.out.println(125);}}
    public static class Impl126 implements DemoService {public int type() {return 126;} public void run() {System.out.println(126);}}
    public static class Impl127 implements DemoService {public int type() {return 127;} public void run() {System.out.println(127);}}
    public static class Impl128 implements DemoService {public int type() {return 128;} public void run() {System.out.println(128);}}
    public static class Impl129 implements DemoService {public int type() {return 129;} public void run() {System.out.println(129);}}
    public static class Impl130 implements DemoService {public int type() {return 130;} public void run() {System.out.println(130);}}
    public static class Impl131 implements DemoService {public int type() {return 131;} public void run() {System.out.println(131);}}
    public static class Impl132 implements DemoService {public int type() {return 132;} public void run() {System.out.println(132);}}
    public static class Impl133 implements DemoService {public int type() {return 133;} public void run() {System.out.println(133);}}
    public static class Impl134 implements DemoService {public int type() {return 134;} public void run() {System.out.println(134);}}
    public static class Impl135 implements DemoService {public int type() {return 135;} public void run() {System.out.println(135);}}
    public static class Impl136 implements DemoService {public int type() {return 136;} public void run() {System.out.println(136);}}
    public static class Impl137 implements DemoService {public int type() {return 137;} public void run() {System.out.println(137);}}
    public static class Impl138 implements DemoService {public int type() {return 138;} public void run() {System.out.println(138);}}
    public static class Impl139 implements DemoService {public int type() {return 139;} public void run() {System.out.println(139);}}
    public static class Impl140 implements DemoService {public int type() {return 140;} public void run() {System.out.println(140);}}
    public static class Impl141 implements DemoService {public int type() {return 141;} public void run() {System.out.println(141);}}
    public static class Impl142 implements DemoService {public int type() {return 142;} public void run() {System.out.println(142);}}
    public static class Impl143 implements DemoService {public int type() {return 143;} public void run() {System.out.println(143);}}
    public static class Impl144 implements DemoService {public int type() {return 144;} public void run() {System.out.println(144);}}
    public static class Impl145 implements DemoService {public int type() {return 145;} public void run() {System.out.println(145);}}
    public static class Impl146 implements DemoService {public int type() {return 146;} public void run() {System.out.println(146);}}
    public static class Impl147 implements DemoService {public int type() {return 147;} public void run() {System.out.println(147);}}
    public static class Impl148 implements DemoService {public int type() {return 148;} public void run() {System.out.println(148);}}
    public static class Impl149 implements DemoService {public int type() {return 149;} public void run() {System.out.println(149);}}
    public static class Impl150 implements DemoService {public int type() {return 150;} public void run() {System.out.println(150);}}
    public static class Impl151 implements DemoService {public int type() {return 151;} public void run() {System.out.println(151);}}
    public static class Impl152 implements DemoService {public int type() {return 152;} public void run() {System.out.println(152);}}
    public static class Impl153 implements DemoService {public int type() {return 153;} public void run() {System.out.println(153);}}
    public static class Impl154 implements DemoService {public int type() {return 154;} public void run() {System.out.println(154);}}
    public static class Impl155 implements DemoService {public int type() {return 155;} public void run() {System.out.println(155);}}
    public static class Impl156 implements DemoService {public int type() {return 156;} public void run() {System.out.println(156);}}
    public static class Impl157 implements DemoService {public int type() {return 157;} public void run() {System.out.println(157);}}
    public static class Impl158 implements DemoService {public int type() {return 158;} public void run() {System.out.println(158);}}
    public static class Impl159 implements DemoService {public int type() {return 159;} public void run() {System.out.println(159);}}
    public static class Impl160 implements DemoService {public int type() {return 160;} public void run() {System.out.println(160);}}
    public static class Impl161 implements DemoService {public int type() {return 161;} public void run() {System.out.println(161);}}
    public static class Impl162 implements DemoService {public int type() {return 162;} public void run() {System.out.println(162);}}
    public static class Impl163 implements DemoService {public int type() {return 163;} public void run() {System.out.println(163);}}
    public static class Impl164 implements DemoService {public int type() {return 164;} public void run() {System.out.println(164);}}
    public static class Impl165 implements DemoService {public int type() {return 165;} public void run() {System.out.println(165);}}
    public static class Impl166 implements DemoService {public int type() {return 166;} public void run() {System.out.println(166);}}
    public static class Impl167 implements DemoService {public int type() {return 167;} public void run() {System.out.println(167);}}
    public static class Impl168 implements DemoService {public int type() {return 168;} public void run() {System.out.println(168);}}
    public static class Impl169 implements DemoService {public int type() {return 169;} public void run() {System.out.println(169);}}
    public static class Impl170 implements DemoService {public int type() {return 170;} public void run() {System.out.println(170);}}
    public static class Impl171 implements DemoService {public int type() {return 171;} public void run() {System.out.println(171);}}
    public static class Impl172 implements DemoService {public int type() {return 172;} public void run() {System.out.println(172);}}
    public static class Impl173 implements DemoService {public int type() {return 173;} public void run() {System.out.println(173);}}
    public static class Impl174 implements DemoService {public int type() {return 174;} public void run() {System.out.println(174);}}
    public static class Impl175 implements DemoService {public int type() {return 175;} public void run() {System.out.println(175);}}
    public static class Impl176 implements DemoService {public int type() {return 176;} public void run() {System.out.println(176);}}
    public static class Impl177 implements DemoService {public int type() {return 177;} public void run() {System.out.println(177);}}
    public static class Impl178 implements DemoService {public int type() {return 178;} public void run() {System.out.println(178);}}
    public static class Impl179 implements DemoService {public int type() {return 179;} public void run() {System.out.println(179);}}
    public static class Impl180 implements DemoService {public int type() {return 180;} public void run() {System.out.println(180);}}
    public static class Impl181 implements DemoService {public int type() {return 181;} public void run() {System.out.println(181);}}
    public static class Impl182 implements DemoService {public int type() {return 182;} public void run() {System.out.println(182);}}
    public static class Impl183 implements DemoService {public int type() {return 183;} public void run() {System.out.println(183);}}
    public static class Impl184 implements DemoService {public int type() {return 184;} public void run() {System.out.println(184);}}
    public static class Impl185 implements DemoService {public int type() {return 185;} public void run() {System.out.println(185);}}
    public static class Impl186 implements DemoService {public int type() {return 186;} public void run() {System.out.println(186);}}
    public static class Impl187 implements DemoService {public int type() {return 187;} public void run() {System.out.println(187);}}
    public static class Impl188 implements DemoService {public int type() {return 188;} public void run() {System.out.println(188);}}
    public static class Impl189 implements DemoService {public int type() {return 189;} public void run() {System.out.println(189);}}
    public static class Impl190 implements DemoService {public int type() {return 190;} public void run() {System.out.println(190);}}
    public static class Impl191 implements DemoService {public int type() {return 191;} public void run() {System.out.println(191);}}
    public static class Impl192 implements DemoService {public int type() {return 192;} public void run() {System.out.println(192);}}
    public static class Impl193 implements DemoService {public int type() {return 193;} public void run() {System.out.println(193);}}
    public static class Impl194 implements DemoService {public int type() {return 194;} public void run() {System.out.println(194);}}
    public static class Impl195 implements DemoService {public int type() {return 195;} public void run() {System.out.println(195);}}
    public static class Impl196 implements DemoService {public int type() {return 196;} public void run() {System.out.println(196);}}
    public static class Impl197 implements DemoService {public int type() {return 197;} public void run() {System.out.println(197);}}
    public static class Impl198 implements DemoService {public int type() {return 198;} public void run() {System.out.println(198);}}
    public static class Impl199 implements DemoService {public int type() {return 199;} public void run() {System.out.println(199);}}
    public static class Impl200 implements DemoService {public int type() {return 200;} public void run() {System.out.println(200);}}
    public static class Impl201 implements DemoService {public int type() {return 201;} public void run() {System.out.println(201);}}
    public static class Impl202 implements DemoService {public int type() {return 202;} public void run() {System.out.println(202);}}
    public static class Impl203 implements DemoService {public int type() {return 203;} public void run() {System.out.println(203);}}
    public static class Impl204 implements DemoService {public int type() {return 204;} public void run() {System.out.println(204);}}
    public static class Impl205 implements DemoService {public int type() {return 205;} public void run() {System.out.println(205);}}
    public static class Impl206 implements DemoService {public int type() {return 206;} public void run() {System.out.println(206);}}
    public static class Impl207 implements DemoService {public int type() {return 207;} public void run() {System.out.println(207);}}
    public static class Impl208 implements DemoService {public int type() {return 208;} public void run() {System.out.println(208);}}
    public static class Impl209 implements DemoService {public int type() {return 209;} public void run() {System.out.println(209);}}
    public static class Impl210 implements DemoService {public int type() {return 210;} public void run() {System.out.println(210);}}
    public static class Impl211 implements DemoService {public int type() {return 211;} public void run() {System.out.println(211);}}
    public static class Impl212 implements DemoService {public int type() {return 212;} public void run() {System.out.println(212);}}
    public static class Impl213 implements DemoService {public int type() {return 213;} public void run() {System.out.println(213);}}
    public static class Impl214 implements DemoService {public int type() {return 214;} public void run() {System.out.println(214);}}
    public static class Impl215 implements DemoService {public int type() {return 215;} public void run() {System.out.println(215);}}
    public static class Impl216 implements DemoService {public int type() {return 216;} public void run() {System.out.println(216);}}
    public static class Impl217 implements DemoService {public int type() {return 217;} public void run() {System.out.println(217);}}
    public static class Impl218 implements DemoService {public int type() {return 218;} public void run() {System.out.println(218);}}
    public static class Impl219 implements DemoService {public int type() {return 219;} public void run() {System.out.println(219);}}
    public static class Impl220 implements DemoService {public int type() {return 220;} public void run() {System.out.println(220);}}
    public static class Impl221 implements DemoService {public int type() {return 221;} public void run() {System.out.println(221);}}
    public static class Impl222 implements DemoService {public int type() {return 222;} public void run() {System.out.println(222);}}
    public static class Impl223 implements DemoService {public int type() {return 223;} public void run() {System.out.println(223);}}
    public static class Impl224 implements DemoService {public int type() {return 224;} public void run() {System.out.println(224);}}
    public static class Impl225 implements DemoService {public int type() {return 225;} public void run() {System.out.println(225);}}
    public static class Impl226 implements DemoService {public int type() {return 226;} public void run() {System.out.println(226);}}
    public static class Impl227 implements DemoService {public int type() {return 227;} public void run() {System.out.println(227);}}
    public static class Impl228 implements DemoService {public int type() {return 228;} public void run() {System.out.println(228);}}
    public static class Impl229 implements DemoService {public int type() {return 229;} public void run() {System.out.println(229);}}
    public static class Impl230 implements DemoService {public int type() {return 230;} public void run() {System.out.println(230);}}
    public static class Impl231 implements DemoService {public int type() {return 231;} public void run() {System.out.println(231);}}
    public static class Impl232 implements DemoService {public int type() {return 232;} public void run() {System.out.println(232);}}
    public static class Impl233 implements DemoService {public int type() {return 233;} public void run() {System.out.println(233);}}
    public static class Impl234 implements DemoService {public int type() {return 234;} public void run() {System.out.println(234);}}
    public static class Impl235 implements DemoService {public int type() {return 235;} public void run() {System.out.println(235);}}
    public static class Impl236 implements DemoService {public int type() {return 236;} public void run() {System.out.println(236);}}
    public static class Impl237 implements DemoService {public int type() {return 237;} public void run() {System.out.println(237);}}
    public static class Impl238 implements DemoService {public int type() {return 238;} public void run() {System.out.println(238);}}
    public static class Impl239 implements DemoService {public int type() {return 239;} public void run() {System.out.println(239);}}
    public static class Impl240 implements DemoService {public int type() {return 240;} public void run() {System.out.println(240);}}
    public static class Impl241 implements DemoService {public int type() {return 241;} public void run() {System.out.println(241);}}
    public static class Impl242 implements DemoService {public int type() {return 242;} public void run() {System.out.println(242);}}
    public static class Impl243 implements DemoService {public int type() {return 243;} public void run() {System.out.println(243);}}
    public static class Impl244 implements DemoService {public int type() {return 244;} public void run() {System.out.println(244);}}
    public static class Impl245 implements DemoService {public int type() {return 245;} public void run() {System.out.println(245);}}
    public static class Impl246 implements DemoService {public int type() {return 246;} public void run() {System.out.println(246);}}
    public static class Impl247 implements DemoService {public int type() {return 247;} public void run() {System.out.println(247);}}
    public static class Impl248 implements DemoService {public int type() {return 248;} public void run() {System.out.println(248);}}
    public static class Impl249 implements DemoService {public int type() {return 249;} public void run() {System.out.println(249);}}
    public static class Impl250 implements DemoService {public int type() {return 250;} public void run() {System.out.println(250);}}
    public static class Impl251 implements DemoService {public int type() {return 251;} public void run() {System.out.println(251);}}
    public static class Impl252 implements DemoService {public int type() {return 252;} public void run() {System.out.println(252);}}
    public static class Impl253 implements DemoService {public int type() {return 253;} public void run() {System.out.println(253);}}
    public static class Impl254 implements DemoService {public int type() {return 254;} public void run() {System.out.println(254);}}
    public static class Impl255 implements DemoService {public int type() {return 255;} public void run() {System.out.println(255);}}
    public static class Impl256 implements DemoService {public int type() {return 256;} public void run() {System.out.println(256);}}
    public static class Impl257 implements DemoService {public int type() {return 257;} public void run() {System.out.println(257);}}
    public static class Impl258 implements DemoService {public int type() {return 258;} public void run() {System.out.println(258);}}
    public static class Impl259 implements DemoService {public int type() {return 259;} public void run() {System.out.println(259);}}
    public static class Impl260 implements DemoService {public int type() {return 260;} public void run() {System.out.println(260);}}
    public static class Impl261 implements DemoService {public int type() {return 261;} public void run() {System.out.println(261);}}
    public static class Impl262 implements DemoService {public int type() {return 262;} public void run() {System.out.println(262);}}
    public static class Impl263 implements DemoService {public int type() {return 263;} public void run() {System.out.println(263);}}
    public static class Impl264 implements DemoService {public int type() {return 264;} public void run() {System.out.println(264);}}
    public static class Impl265 implements DemoService {public int type() {return 265;} public void run() {System.out.println(265);}}
    public static class Impl266 implements DemoService {public int type() {return 266;} public void run() {System.out.println(266);}}
    public static class Impl267 implements DemoService {public int type() {return 267;} public void run() {System.out.println(267);}}
    public static class Impl268 implements DemoService {public int type() {return 268;} public void run() {System.out.println(268);}}
    public static class Impl269 implements DemoService {public int type() {return 269;} public void run() {System.out.println(269);}}
    public static class Impl270 implements DemoService {public int type() {return 270;} public void run() {System.out.println(270);}}
    public static class Impl271 implements DemoService {public int type() {return 271;} public void run() {System.out.println(271);}}
    public static class Impl272 implements DemoService {public int type() {return 272;} public void run() {System.out.println(272);}}
    public static class Impl273 implements DemoService {public int type() {return 273;} public void run() {System.out.println(273);}}
    public static class Impl274 implements DemoService {public int type() {return 274;} public void run() {System.out.println(274);}}
    public static class Impl275 implements DemoService {public int type() {return 275;} public void run() {System.out.println(275);}}
    public static class Impl276 implements DemoService {public int type() {return 276;} public void run() {System.out.println(276);}}
    public static class Impl277 implements DemoService {public int type() {return 277;} public void run() {System.out.println(277);}}
    public static class Impl278 implements DemoService {public int type() {return 278;} public void run() {System.out.println(278);}}
    public static class Impl279 implements DemoService {public int type() {return 279;} public void run() {System.out.println(279);}}
    public static class Impl280 implements DemoService {public int type() {return 280;} public void run() {System.out.println(280);}}
    public static class Impl281 implements DemoService {public int type() {return 281;} public void run() {System.out.println(281);}}
    public static class Impl282 implements DemoService {public int type() {return 282;} public void run() {System.out.println(282);}}
    public static class Impl283 implements DemoService {public int type() {return 283;} public void run() {System.out.println(283);}}
    public static class Impl284 implements DemoService {public int type() {return 284;} public void run() {System.out.println(284);}}
    public static class Impl285 implements DemoService {public int type() {return 285;} public void run() {System.out.println(285);}}
    public static class Impl286 implements DemoService {public int type() {return 286;} public void run() {System.out.println(286);}}
    public static class Impl287 implements DemoService {public int type() {return 287;} public void run() {System.out.println(287);}}
    public static class Impl288 implements DemoService {public int type() {return 288;} public void run() {System.out.println(288);}}
    public static class Impl289 implements DemoService {public int type() {return 289;} public void run() {System.out.println(289);}}
    public static class Impl290 implements DemoService {public int type() {return 290;} public void run() {System.out.println(290);}}
    public static class Impl291 implements DemoService {public int type() {return 291;} public void run() {System.out.println(291);}}
    public static class Impl292 implements DemoService {public int type() {return 292;} public void run() {System.out.println(292);}}
    public static class Impl293 implements DemoService {public int type() {return 293;} public void run() {System.out.println(293);}}
    public static class Impl294 implements DemoService {public int type() {return 294;} public void run() {System.out.println(294);}}
    public static class Impl295 implements DemoService {public int type() {return 295;} public void run() {System.out.println(295);}}
    public static class Impl296 implements DemoService {public int type() {return 296;} public void run() {System.out.println(296);}}
    public static class Impl297 implements DemoService {public int type() {return 297;} public void run() {System.out.println(297);}}
    public static class Impl298 implements DemoService {public int type() {return 298;} public void run() {System.out.println(298);}}
    public static class Impl299 implements DemoService {public int type() {return 299;} public void run() {System.out.println(299);}}
    public static class Impl300 implements DemoService {public int type() {return 300;} public void run() {System.out.println(300);}}
    public static class Impl301 implements DemoService {public int type() {return 301;} public void run() {System.out.println(301);}}
    public static class Impl302 implements DemoService {public int type() {return 302;} public void run() {System.out.println(302);}}
    public static class Impl303 implements DemoService {public int type() {return 303;} public void run() {System.out.println(303);}}
    public static class Impl304 implements DemoService {public int type() {return 304;} public void run() {System.out.println(304);}}
    public static class Impl305 implements DemoService {public int type() {return 305;} public void run() {System.out.println(305);}}
    public static class Impl306 implements DemoService {public int type() {return 306;} public void run() {System.out.println(306);}}
    public static class Impl307 implements DemoService {public int type() {return 307;} public void run() {System.out.println(307);}}
    public static class Impl308 implements DemoService {public int type() {return 308;} public void run() {System.out.println(308);}}
    public static class Impl309 implements DemoService {public int type() {return 309;} public void run() {System.out.println(309);}}
    public static class Impl310 implements DemoService {public int type() {return 310;} public void run() {System.out.println(310);}}
    public static class Impl311 implements DemoService {public int type() {return 311;} public void run() {System.out.println(311);}}
    public static class Impl312 implements DemoService {public int type() {return 312;} public void run() {System.out.println(312);}}
    public static class Impl313 implements DemoService {public int type() {return 313;} public void run() {System.out.println(313);}}
    public static class Impl314 implements DemoService {public int type() {return 314;} public void run() {System.out.println(314);}}
    public static class Impl315 implements DemoService {public int type() {return 315;} public void run() {System.out.println(315);}}
    public static class Impl316 implements DemoService {public int type() {return 316;} public void run() {System.out.println(316);}}
    public static class Impl317 implements DemoService {public int type() {return 317;} public void run() {System.out.println(317);}}
    public static class Impl318 implements DemoService {public int type() {return 318;} public void run() {System.out.println(318);}}
    public static class Impl319 implements DemoService {public int type() {return 319;} public void run() {System.out.println(319);}}
    public static class Impl320 implements DemoService {public int type() {return 320;} public void run() {System.out.println(320);}}
    public static class Impl321 implements DemoService {public int type() {return 321;} public void run() {System.out.println(321);}}
    public static class Impl322 implements DemoService {public int type() {return 322;} public void run() {System.out.println(322);}}
    public static class Impl323 implements DemoService {public int type() {return 323;} public void run() {System.out.println(323);}}
    public static class Impl324 implements DemoService {public int type() {return 324;} public void run() {System.out.println(324);}}
    public static class Impl325 implements DemoService {public int type() {return 325;} public void run() {System.out.println(325);}}
    public static class Impl326 implements DemoService {public int type() {return 326;} public void run() {System.out.println(326);}}
    public static class Impl327 implements DemoService {public int type() {return 327;} public void run() {System.out.println(327);}}
    public static class Impl328 implements DemoService {public int type() {return 328;} public void run() {System.out.println(328);}}
    public static class Impl329 implements DemoService {public int type() {return 329;} public void run() {System.out.println(329);}}
    public static class Impl330 implements DemoService {public int type() {return 330;} public void run() {System.out.println(330);}}
    public static class Impl331 implements DemoService {public int type() {return 331;} public void run() {System.out.println(331);}}
    public static class Impl332 implements DemoService {public int type() {return 332;} public void run() {System.out.println(332);}}
    public static class Impl333 implements DemoService {public int type() {return 333;} public void run() {System.out.println(333);}}
    public static class Impl334 implements DemoService {public int type() {return 334;} public void run() {System.out.println(334);}}
    public static class Impl335 implements DemoService {public int type() {return 335;} public void run() {System.out.println(335);}}
    public static class Impl336 implements DemoService {public int type() {return 336;} public void run() {System.out.println(336);}}
    public static class Impl337 implements DemoService {public int type() {return 337;} public void run() {System.out.println(337);}}
    public static class Impl338 implements DemoService {public int type() {return 338;} public void run() {System.out.println(338);}}
    public static class Impl339 implements DemoService {public int type() {return 339;} public void run() {System.out.println(339);}}
    public static class Impl340 implements DemoService {public int type() {return 340;} public void run() {System.out.println(340);}}
    public static class Impl341 implements DemoService {public int type() {return 341;} public void run() {System.out.println(341);}}
    public static class Impl342 implements DemoService {public int type() {return 342;} public void run() {System.out.println(342);}}
    public static class Impl343 implements DemoService {public int type() {return 343;} public void run() {System.out.println(343);}}
    public static class Impl344 implements DemoService {public int type() {return 344;} public void run() {System.out.println(344);}}
    public static class Impl345 implements DemoService {public int type() {return 345;} public void run() {System.out.println(345);}}
    public static class Impl346 implements DemoService {public int type() {return 346;} public void run() {System.out.println(346);}}
    public static class Impl347 implements DemoService {public int type() {return 347;} public void run() {System.out.println(347);}}
    public static class Impl348 implements DemoService {public int type() {return 348;} public void run() {System.out.println(348);}}
    public static class Impl349 implements DemoService {public int type() {return 349;} public void run() {System.out.println(349);}}
    public static class Impl350 implements DemoService {public int type() {return 350;} public void run() {System.out.println(350);}}
    public static class Impl351 implements DemoService {public int type() {return 351;} public void run() {System.out.println(351);}}
    public static class Impl352 implements DemoService {public int type() {return 352;} public void run() {System.out.println(352);}}
    public static class Impl353 implements DemoService {public int type() {return 353;} public void run() {System.out.println(353);}}
    public static class Impl354 implements DemoService {public int type() {return 354;} public void run() {System.out.println(354);}}
    public static class Impl355 implements DemoService {public int type() {return 355;} public void run() {System.out.println(355);}}
    public static class Impl356 implements DemoService {public int type() {return 356;} public void run() {System.out.println(356);}}
    public static class Impl357 implements DemoService {public int type() {return 357;} public void run() {System.out.println(357);}}
    public static class Impl358 implements DemoService {public int type() {return 358;} public void run() {System.out.println(358);}}
    public static class Impl359 implements DemoService {public int type() {return 359;} public void run() {System.out.println(359);}}
    public static class Impl360 implements DemoService {public int type() {return 360;} public void run() {System.out.println(360);}}
    public static class Impl361 implements DemoService {public int type() {return 361;} public void run() {System.out.println(361);}}
    public static class Impl362 implements DemoService {public int type() {return 362;} public void run() {System.out.println(362);}}
    public static class Impl363 implements DemoService {public int type() {return 363;} public void run() {System.out.println(363);}}
    public static class Impl364 implements DemoService {public int type() {return 364;} public void run() {System.out.println(364);}}
    public static class Impl365 implements DemoService {public int type() {return 365;} public void run() {System.out.println(365);}}
    public static class Impl366 implements DemoService {public int type() {return 366;} public void run() {System.out.println(366);}}
    public static class Impl367 implements DemoService {public int type() {return 367;} public void run() {System.out.println(367);}}
    public static class Impl368 implements DemoService {public int type() {return 368;} public void run() {System.out.println(368);}}
    public static class Impl369 implements DemoService {public int type() {return 369;} public void run() {System.out.println(369);}}
    public static class Impl370 implements DemoService {public int type() {return 370;} public void run() {System.out.println(370);}}
    public static class Impl371 implements DemoService {public int type() {return 371;} public void run() {System.out.println(371);}}
    public static class Impl372 implements DemoService {public int type() {return 372;} public void run() {System.out.println(372);}}
    public static class Impl373 implements DemoService {public int type() {return 373;} public void run() {System.out.println(373);}}
    public static class Impl374 implements DemoService {public int type() {return 374;} public void run() {System.out.println(374);}}
    public static class Impl375 implements DemoService {public int type() {return 375;} public void run() {System.out.println(375);}}
    public static class Impl376 implements DemoService {public int type() {return 376;} public void run() {System.out.println(376);}}
    public static class Impl377 implements DemoService {public int type() {return 377;} public void run() {System.out.println(377);}}
    public static class Impl378 implements DemoService {public int type() {return 378;} public void run() {System.out.println(378);}}
    public static class Impl379 implements DemoService {public int type() {return 379;} public void run() {System.out.println(379);}}
    public static class Impl380 implements DemoService {public int type() {return 380;} public void run() {System.out.println(380);}}
    public static class Impl381 implements DemoService {public int type() {return 381;} public void run() {System.out.println(381);}}
    public static class Impl382 implements DemoService {public int type() {return 382;} public void run() {System.out.println(382);}}
    public static class Impl383 implements DemoService {public int type() {return 383;} public void run() {System.out.println(383);}}
    public static class Impl384 implements DemoService {public int type() {return 384;} public void run() {System.out.println(384);}}
    public static class Impl385 implements DemoService {public int type() {return 385;} public void run() {System.out.println(385);}}
    public static class Impl386 implements DemoService {public int type() {return 386;} public void run() {System.out.println(386);}}
    public static class Impl387 implements DemoService {public int type() {return 387;} public void run() {System.out.println(387);}}
    public static class Impl388 implements DemoService {public int type() {return 388;} public void run() {System.out.println(388);}}
    public static class Impl389 implements DemoService {public int type() {return 389;} public void run() {System.out.println(389);}}
    public static class Impl390 implements DemoService {public int type() {return 390;} public void run() {System.out.println(390);}}
    public static class Impl391 implements DemoService {public int type() {return 391;} public void run() {System.out.println(391);}}
    public static class Impl392 implements DemoService {public int type() {return 392;} public void run() {System.out.println(392);}}
    public static class Impl393 implements DemoService {public int type() {return 393;} public void run() {System.out.println(393);}}
    public static class Impl394 implements DemoService {public int type() {return 394;} public void run() {System.out.println(394);}}
    public static class Impl395 implements DemoService {public int type() {return 395;} public void run() {System.out.println(395);}}
    public static class Impl396 implements DemoService {public int type() {return 396;} public void run() {System.out.println(396);}}
    public static class Impl397 implements DemoService {public int type() {return 397;} public void run() {System.out.println(397);}}
    public static class Impl398 implements DemoService {public int type() {return 398;} public void run() {System.out.println(398);}}
    public static class Impl399 implements DemoService {public int type() {return 399;} public void run() {System.out.println(399);}}
    public static class Impl400 implements DemoService {public int type() {return 400;} public void run() {System.out.println(400);}}
    public static class Impl401 implements DemoService {public int type() {return 401;} public void run() {System.out.println(401);}}
    public static class Impl402 implements DemoService {public int type() {return 402;} public void run() {System.out.println(402);}}
    public static class Impl403 implements DemoService {public int type() {return 403;} public void run() {System.out.println(403);}}
    public static class Impl404 implements DemoService {public int type() {return 404;} public void run() {System.out.println(404);}}
    public static class Impl405 implements DemoService {public int type() {return 405;} public void run() {System.out.println(405);}}
    public static class Impl406 implements DemoService {public int type() {return 406;} public void run() {System.out.println(406);}}
    public static class Impl407 implements DemoService {public int type() {return 407;} public void run() {System.out.println(407);}}
    public static class Impl408 implements DemoService {public int type() {return 408;} public void run() {System.out.println(408);}}
    public static class Impl409 implements DemoService {public int type() {return 409;} public void run() {System.out.println(409);}}
    public static class Impl410 implements DemoService {public int type() {return 410;} public void run() {System.out.println(410);}}
    public static class Impl411 implements DemoService {public int type() {return 411;} public void run() {System.out.println(411);}}
    public static class Impl412 implements DemoService {public int type() {return 412;} public void run() {System.out.println(412);}}
    public static class Impl413 implements DemoService {public int type() {return 413;} public void run() {System.out.println(413);}}
    public static class Impl414 implements DemoService {public int type() {return 414;} public void run() {System.out.println(414);}}
    public static class Impl415 implements DemoService {public int type() {return 415;} public void run() {System.out.println(415);}}
    public static class Impl416 implements DemoService {public int type() {return 416;} public void run() {System.out.println(416);}}
    public static class Impl417 implements DemoService {public int type() {return 417;} public void run() {System.out.println(417);}}
    public static class Impl418 implements DemoService {public int type() {return 418;} public void run() {System.out.println(418);}}
    public static class Impl419 implements DemoService {public int type() {return 419;} public void run() {System.out.println(419);}}
    public static class Impl420 implements DemoService {public int type() {return 420;} public void run() {System.out.println(420);}}
    public static class Impl421 implements DemoService {public int type() {return 421;} public void run() {System.out.println(421);}}
    public static class Impl422 implements DemoService {public int type() {return 422;} public void run() {System.out.println(422);}}
    public static class Impl423 implements DemoService {public int type() {return 423;} public void run() {System.out.println(423);}}
    public static class Impl424 implements DemoService {public int type() {return 424;} public void run() {System.out.println(424);}}
    public static class Impl425 implements DemoService {public int type() {return 425;} public void run() {System.out.println(425);}}
    public static class Impl426 implements DemoService {public int type() {return 426;} public void run() {System.out.println(426);}}
    public static class Impl427 implements DemoService {public int type() {return 427;} public void run() {System.out.println(427);}}
    public static class Impl428 implements DemoService {public int type() {return 428;} public void run() {System.out.println(428);}}
    public static class Impl429 implements DemoService {public int type() {return 429;} public void run() {System.out.println(429);}}
    public static class Impl430 implements DemoService {public int type() {return 430;} public void run() {System.out.println(430);}}
    public static class Impl431 implements DemoService {public int type() {return 431;} public void run() {System.out.println(431);}}
    public static class Impl432 implements DemoService {public int type() {return 432;} public void run() {System.out.println(432);}}
    public static class Impl433 implements DemoService {public int type() {return 433;} public void run() {System.out.println(433);}}
    public static class Impl434 implements DemoService {public int type() {return 434;} public void run() {System.out.println(434);}}
    public static class Impl435 implements DemoService {public int type() {return 435;} public void run() {System.out.println(435);}}
    public static class Impl436 implements DemoService {public int type() {return 436;} public void run() {System.out.println(436);}}
    public static class Impl437 implements DemoService {public int type() {return 437;} public void run() {System.out.println(437);}}
    public static class Impl438 implements DemoService {public int type() {return 438;} public void run() {System.out.println(438);}}
    public static class Impl439 implements DemoService {public int type() {return 439;} public void run() {System.out.println(439);}}
    public static class Impl440 implements DemoService {public int type() {return 440;} public void run() {System.out.println(440);}}
    public static class Impl441 implements DemoService {public int type() {return 441;} public void run() {System.out.println(441);}}
    public static class Impl442 implements DemoService {public int type() {return 442;} public void run() {System.out.println(442);}}
    public static class Impl443 implements DemoService {public int type() {return 443;} public void run() {System.out.println(443);}}
    public static class Impl444 implements DemoService {public int type() {return 444;} public void run() {System.out.println(444);}}
    public static class Impl445 implements DemoService {public int type() {return 445;} public void run() {System.out.println(445);}}
    public static class Impl446 implements DemoService {public int type() {return 446;} public void run() {System.out.println(446);}}
    public static class Impl447 implements DemoService {public int type() {return 447;} public void run() {System.out.println(447);}}
    public static class Impl448 implements DemoService {public int type() {return 448;} public void run() {System.out.println(448);}}
    public static class Impl449 implements DemoService {public int type() {return 449;} public void run() {System.out.println(449);}}
    public static class Impl450 implements DemoService {public int type() {return 450;} public void run() {System.out.println(450);}}
    public static class Impl451 implements DemoService {public int type() {return 451;} public void run() {System.out.println(451);}}
    public static class Impl452 implements DemoService {public int type() {return 452;} public void run() {System.out.println(452);}}
    public static class Impl453 implements DemoService {public int type() {return 453;} public void run() {System.out.println(453);}}
    public static class Impl454 implements DemoService {public int type() {return 454;} public void run() {System.out.println(454);}}
    public static class Impl455 implements DemoService {public int type() {return 455;} public void run() {System.out.println(455);}}
    public static class Impl456 implements DemoService {public int type() {return 456;} public void run() {System.out.println(456);}}
    public static class Impl457 implements DemoService {public int type() {return 457;} public void run() {System.out.println(457);}}
    public static class Impl458 implements DemoService {public int type() {return 458;} public void run() {System.out.println(458);}}
    public static class Impl459 implements DemoService {public int type() {return 459;} public void run() {System.out.println(459);}}
    public static class Impl460 implements DemoService {public int type() {return 460;} public void run() {System.out.println(460);}}
    public static class Impl461 implements DemoService {public int type() {return 461;} public void run() {System.out.println(461);}}
    public static class Impl462 implements DemoService {public int type() {return 462;} public void run() {System.out.println(462);}}
    public static class Impl463 implements DemoService {public int type() {return 463;} public void run() {System.out.println(463);}}
    public static class Impl464 implements DemoService {public int type() {return 464;} public void run() {System.out.println(464);}}
    public static class Impl465 implements DemoService {public int type() {return 465;} public void run() {System.out.println(465);}}
    public static class Impl466 implements DemoService {public int type() {return 466;} public void run() {System.out.println(466);}}
    public static class Impl467 implements DemoService {public int type() {return 467;} public void run() {System.out.println(467);}}
    public static class Impl468 implements DemoService {public int type() {return 468;} public void run() {System.out.println(468);}}
    public static class Impl469 implements DemoService {public int type() {return 469;} public void run() {System.out.println(469);}}
    public static class Impl470 implements DemoService {public int type() {return 470;} public void run() {System.out.println(470);}}
    public static class Impl471 implements DemoService {public int type() {return 471;} public void run() {System.out.println(471);}}
    public static class Impl472 implements DemoService {public int type() {return 472;} public void run() {System.out.println(472);}}
    public static class Impl473 implements DemoService {public int type() {return 473;} public void run() {System.out.println(473);}}
    public static class Impl474 implements DemoService {public int type() {return 474;} public void run() {System.out.println(474);}}
    public static class Impl475 implements DemoService {public int type() {return 475;} public void run() {System.out.println(475);}}
    public static class Impl476 implements DemoService {public int type() {return 476;} public void run() {System.out.println(476);}}
    public static class Impl477 implements DemoService {public int type() {return 477;} public void run() {System.out.println(477);}}
    public static class Impl478 implements DemoService {public int type() {return 478;} public void run() {System.out.println(478);}}
    public static class Impl479 implements DemoService {public int type() {return 479;} public void run() {System.out.println(479);}}
    public static class Impl480 implements DemoService {public int type() {return 480;} public void run() {System.out.println(480);}}
    public static class Impl481 implements DemoService {public int type() {return 481;} public void run() {System.out.println(481);}}
    public static class Impl482 implements DemoService {public int type() {return 482;} public void run() {System.out.println(482);}}
    public static class Impl483 implements DemoService {public int type() {return 483;} public void run() {System.out.println(483);}}
    public static class Impl484 implements DemoService {public int type() {return 484;} public void run() {System.out.println(484);}}
    public static class Impl485 implements DemoService {public int type() {return 485;} public void run() {System.out.println(485);}}
    public static class Impl486 implements DemoService {public int type() {return 486;} public void run() {System.out.println(486);}}
    public static class Impl487 implements DemoService {public int type() {return 487;} public void run() {System.out.println(487);}}
    public static class Impl488 implements DemoService {public int type() {return 488;} public void run() {System.out.println(488);}}
    public static class Impl489 implements DemoService {public int type() {return 489;} public void run() {System.out.println(489);}}
    public static class Impl490 implements DemoService {public int type() {return 490;} public void run() {System.out.println(490);}}
    public static class Impl491 implements DemoService {public int type() {return 491;} public void run() {System.out.println(491);}}
    public static class Impl492 implements DemoService {public int type() {return 492;} public void run() {System.out.println(492);}}
    public static class Impl493 implements DemoService {public int type() {return 493;} public void run() {System.out.println(493);}}
    public static class Impl494 implements DemoService {public int type() {return 494;} public void run() {System.out.println(494);}}
    public static class Impl495 implements DemoService {public int type() {return 495;} public void run() {System.out.println(495);}}
    public static class Impl496 implements DemoService {public int type() {return 496;} public void run() {System.out.println(496);}}
    public static class Impl497 implements DemoService {public int type() {return 497;} public void run() {System.out.println(497);}}
    public static class Impl498 implements DemoService {public int type() {return 498;} public void run() {System.out.println(498);}}
    public static class Impl499 implements DemoService {public int type() {return 499;} public void run() {System.out.println(499);}}
    public static class Impl500 implements DemoService {public int type() {return 500;} public void run() {System.out.println(500);}}
    public static class Impl501 implements DemoService {public int type() {return 501;} public void run() {System.out.println(501);}}
    public static class Impl502 implements DemoService {public int type() {return 502;} public void run() {System.out.println(502);}}
    public static class Impl503 implements DemoService {public int type() {return 503;} public void run() {System.out.println(503);}}
    public static class Impl504 implements DemoService {public int type() {return 504;} public void run() {System.out.println(504);}}
    public static class Impl505 implements DemoService {public int type() {return 505;} public void run() {System.out.println(505);}}
    public static class Impl506 implements DemoService {public int type() {return 506;} public void run() {System.out.println(506);}}
    public static class Impl507 implements DemoService {public int type() {return 507;} public void run() {System.out.println(507);}}
    public static class Impl508 implements DemoService {public int type() {return 508;} public void run() {System.out.println(508);}}
    public static class Impl509 implements DemoService {public int type() {return 509;} public void run() {System.out.println(509);}}
    public static class Impl510 implements DemoService {public int type() {return 510;} public void run() {System.out.println(510);}}
    public static class Impl511 implements DemoService {public int type() {return 511;} public void run() {System.out.println(511);}}
    public static class Impl512 implements DemoService {public int type() {return 512;} public void run() {System.out.println(512);}}
    public static class Impl513 implements DemoService {public int type() {return 513;} public void run() {System.out.println(513);}}
    public static class Impl514 implements DemoService {public int type() {return 514;} public void run() {System.out.println(514);}}
    public static class Impl515 implements DemoService {public int type() {return 515;} public void run() {System.out.println(515);}}
    public static class Impl516 implements DemoService {public int type() {return 516;} public void run() {System.out.println(516);}}
    public static class Impl517 implements DemoService {public int type() {return 517;} public void run() {System.out.println(517);}}
    public static class Impl518 implements DemoService {public int type() {return 518;} public void run() {System.out.println(518);}}
    public static class Impl519 implements DemoService {public int type() {return 519;} public void run() {System.out.println(519);}}
    public static class Impl520 implements DemoService {public int type() {return 520;} public void run() {System.out.println(520);}}
    public static class Impl521 implements DemoService {public int type() {return 521;} public void run() {System.out.println(521);}}
    public static class Impl522 implements DemoService {public int type() {return 522;} public void run() {System.out.println(522);}}
    public static class Impl523 implements DemoService {public int type() {return 523;} public void run() {System.out.println(523);}}
    public static class Impl524 implements DemoService {public int type() {return 524;} public void run() {System.out.println(524);}}
    public static class Impl525 implements DemoService {public int type() {return 525;} public void run() {System.out.println(525);}}
    public static class Impl526 implements DemoService {public int type() {return 526;} public void run() {System.out.println(526);}}
    public static class Impl527 implements DemoService {public int type() {return 527;} public void run() {System.out.println(527);}}
    public static class Impl528 implements DemoService {public int type() {return 528;} public void run() {System.out.println(528);}}
    public static class Impl529 implements DemoService {public int type() {return 529;} public void run() {System.out.println(529);}}
    public static class Impl530 implements DemoService {public int type() {return 530;} public void run() {System.out.println(530);}}
    public static class Impl531 implements DemoService {public int type() {return 531;} public void run() {System.out.println(531);}}
    public static class Impl532 implements DemoService {public int type() {return 532;} public void run() {System.out.println(532);}}
    public static class Impl533 implements DemoService {public int type() {return 533;} public void run() {System.out.println(533);}}
    public static class Impl534 implements DemoService {public int type() {return 534;} public void run() {System.out.println(534);}}
    public static class Impl535 implements DemoService {public int type() {return 535;} public void run() {System.out.println(535);}}
    public static class Impl536 implements DemoService {public int type() {return 536;} public void run() {System.out.println(536);}}
    public static class Impl537 implements DemoService {public int type() {return 537;} public void run() {System.out.println(537);}}
    public static class Impl538 implements DemoService {public int type() {return 538;} public void run() {System.out.println(538);}}
    public static class Impl539 implements DemoService {public int type() {return 539;} public void run() {System.out.println(539);}}
    public static class Impl540 implements DemoService {public int type() {return 540;} public void run() {System.out.println(540);}}
    public static class Impl541 implements DemoService {public int type() {return 541;} public void run() {System.out.println(541);}}
    public static class Impl542 implements DemoService {public int type() {return 542;} public void run() {System.out.println(542);}}
    public static class Impl543 implements DemoService {public int type() {return 543;} public void run() {System.out.println(543);}}
    public static class Impl544 implements DemoService {public int type() {return 544;} public void run() {System.out.println(544);}}
    public static class Impl545 implements DemoService {public int type() {return 545;} public void run() {System.out.println(545);}}
    public static class Impl546 implements DemoService {public int type() {return 546;} public void run() {System.out.println(546);}}
    public static class Impl547 implements DemoService {public int type() {return 547;} public void run() {System.out.println(547);}}
    public static class Impl548 implements DemoService {public int type() {return 548;} public void run() {System.out.println(548);}}
    public static class Impl549 implements DemoService {public int type() {return 549;} public void run() {System.out.println(549);}}
    public static class Impl550 implements DemoService {public int type() {return 550;} public void run() {System.out.println(550);}}
    public static class Impl551 implements DemoService {public int type() {return 551;} public void run() {System.out.println(551);}}
    public static class Impl552 implements DemoService {public int type() {return 552;} public void run() {System.out.println(552);}}
    public static class Impl553 implements DemoService {public int type() {return 553;} public void run() {System.out.println(553);}}
    public static class Impl554 implements DemoService {public int type() {return 554;} public void run() {System.out.println(554);}}
    public static class Impl555 implements DemoService {public int type() {return 555;} public void run() {System.out.println(555);}}
    public static class Impl556 implements DemoService {public int type() {return 556;} public void run() {System.out.println(556);}}
    public static class Impl557 implements DemoService {public int type() {return 557;} public void run() {System.out.println(557);}}
    public static class Impl558 implements DemoService {public int type() {return 558;} public void run() {System.out.println(558);}}
    public static class Impl559 implements DemoService {public int type() {return 559;} public void run() {System.out.println(559);}}
    public static class Impl560 implements DemoService {public int type() {return 560;} public void run() {System.out.println(560);}}
    public static class Impl561 implements DemoService {public int type() {return 561;} public void run() {System.out.println(561);}}
    public static class Impl562 implements DemoService {public int type() {return 562;} public void run() {System.out.println(562);}}
    public static class Impl563 implements DemoService {public int type() {return 563;} public void run() {System.out.println(563);}}
    public static class Impl564 implements DemoService {public int type() {return 564;} public void run() {System.out.println(564);}}
    public static class Impl565 implements DemoService {public int type() {return 565;} public void run() {System.out.println(565);}}
    public static class Impl566 implements DemoService {public int type() {return 566;} public void run() {System.out.println(566);}}
    public static class Impl567 implements DemoService {public int type() {return 567;} public void run() {System.out.println(567);}}
    public static class Impl568 implements DemoService {public int type() {return 568;} public void run() {System.out.println(568);}}
    public static class Impl569 implements DemoService {public int type() {return 569;} public void run() {System.out.println(569);}}
    public static class Impl570 implements DemoService {public int type() {return 570;} public void run() {System.out.println(570);}}
    public static class Impl571 implements DemoService {public int type() {return 571;} public void run() {System.out.println(571);}}
    public static class Impl572 implements DemoService {public int type() {return 572;} public void run() {System.out.println(572);}}
    public static class Impl573 implements DemoService {public int type() {return 573;} public void run() {System.out.println(573);}}
    public static class Impl574 implements DemoService {public int type() {return 574;} public void run() {System.out.println(574);}}
    public static class Impl575 implements DemoService {public int type() {return 575;} public void run() {System.out.println(575);}}
    public static class Impl576 implements DemoService {public int type() {return 576;} public void run() {System.out.println(576);}}
    public static class Impl577 implements DemoService {public int type() {return 577;} public void run() {System.out.println(577);}}
    public static class Impl578 implements DemoService {public int type() {return 578;} public void run() {System.out.println(578);}}
    public static class Impl579 implements DemoService {public int type() {return 579;} public void run() {System.out.println(579);}}
    public static class Impl580 implements DemoService {public int type() {return 580;} public void run() {System.out.println(580);}}
    public static class Impl581 implements DemoService {public int type() {return 581;} public void run() {System.out.println(581);}}
    public static class Impl582 implements DemoService {public int type() {return 582;} public void run() {System.out.println(582);}}
    public static class Impl583 implements DemoService {public int type() {return 583;} public void run() {System.out.println(583);}}
    public static class Impl584 implements DemoService {public int type() {return 584;} public void run() {System.out.println(584);}}
    public static class Impl585 implements DemoService {public int type() {return 585;} public void run() {System.out.println(585);}}
    public static class Impl586 implements DemoService {public int type() {return 586;} public void run() {System.out.println(586);}}
    public static class Impl587 implements DemoService {public int type() {return 587;} public void run() {System.out.println(587);}}
    public static class Impl588 implements DemoService {public int type() {return 588;} public void run() {System.out.println(588);}}
    public static class Impl589 implements DemoService {public int type() {return 589;} public void run() {System.out.println(589);}}
    public static class Impl590 implements DemoService {public int type() {return 590;} public void run() {System.out.println(590);}}
    public static class Impl591 implements DemoService {public int type() {return 591;} public void run() {System.out.println(591);}}
    public static class Impl592 implements DemoService {public int type() {return 592;} public void run() {System.out.println(592);}}
    public static class Impl593 implements DemoService {public int type() {return 593;} public void run() {System.out.println(593);}}
    public static class Impl594 implements DemoService {public int type() {return 594;} public void run() {System.out.println(594);}}
    public static class Impl595 implements DemoService {public int type() {return 595;} public void run() {System.out.println(595);}}
    public static class Impl596 implements DemoService {public int type() {return 596;} public void run() {System.out.println(596);}}
    public static class Impl597 implements DemoService {public int type() {return 597;} public void run() {System.out.println(597);}}
    public static class Impl598 implements DemoService {public int type() {return 598;} public void run() {System.out.println(598);}}
    public static class Impl599 implements DemoService {public int type() {return 599;} public void run() {System.out.println(599);}}
    public static class Impl600 implements DemoService {public int type() {return 600;} public void run() {System.out.println(600);}}
    public static class Impl601 implements DemoService {public int type() {return 601;} public void run() {System.out.println(601);}}
    public static class Impl602 implements DemoService {public int type() {return 602;} public void run() {System.out.println(602);}}
    public static class Impl603 implements DemoService {public int type() {return 603;} public void run() {System.out.println(603);}}
    public static class Impl604 implements DemoService {public int type() {return 604;} public void run() {System.out.println(604);}}
    public static class Impl605 implements DemoService {public int type() {return 605;} public void run() {System.out.println(605);}}
    public static class Impl606 implements DemoService {public int type() {return 606;} public void run() {System.out.println(606);}}
    public static class Impl607 implements DemoService {public int type() {return 607;} public void run() {System.out.println(607);}}
    public static class Impl608 implements DemoService {public int type() {return 608;} public void run() {System.out.println(608);}}
    public static class Impl609 implements DemoService {public int type() {return 609;} public void run() {System.out.println(609);}}
    public static class Impl610 implements DemoService {public int type() {return 610;} public void run() {System.out.println(610);}}
    public static class Impl611 implements DemoService {public int type() {return 611;} public void run() {System.out.println(611);}}
    public static class Impl612 implements DemoService {public int type() {return 612;} public void run() {System.out.println(612);}}
    public static class Impl613 implements DemoService {public int type() {return 613;} public void run() {System.out.println(613);}}
    public static class Impl614 implements DemoService {public int type() {return 614;} public void run() {System.out.println(614);}}
    public static class Impl615 implements DemoService {public int type() {return 615;} public void run() {System.out.println(615);}}
    public static class Impl616 implements DemoService {public int type() {return 616;} public void run() {System.out.println(616);}}
    public static class Impl617 implements DemoService {public int type() {return 617;} public void run() {System.out.println(617);}}
    public static class Impl618 implements DemoService {public int type() {return 618;} public void run() {System.out.println(618);}}
    public static class Impl619 implements DemoService {public int type() {return 619;} public void run() {System.out.println(619);}}
    public static class Impl620 implements DemoService {public int type() {return 620;} public void run() {System.out.println(620);}}
    public static class Impl621 implements DemoService {public int type() {return 621;} public void run() {System.out.println(621);}}
    public static class Impl622 implements DemoService {public int type() {return 622;} public void run() {System.out.println(622);}}
    public static class Impl623 implements DemoService {public int type() {return 623;} public void run() {System.out.println(623);}}
    public static class Impl624 implements DemoService {public int type() {return 624;} public void run() {System.out.println(624);}}
    public static class Impl625 implements DemoService {public int type() {return 625;} public void run() {System.out.println(625);}}
    public static class Impl626 implements DemoService {public int type() {return 626;} public void run() {System.out.println(626);}}
    public static class Impl627 implements DemoService {public int type() {return 627;} public void run() {System.out.println(627);}}
    public static class Impl628 implements DemoService {public int type() {return 628;} public void run() {System.out.println(628);}}
    public static class Impl629 implements DemoService {public int type() {return 629;} public void run() {System.out.println(629);}}
    public static class Impl630 implements DemoService {public int type() {return 630;} public void run() {System.out.println(630);}}
    public static class Impl631 implements DemoService {public int type() {return 631;} public void run() {System.out.println(631);}}
    public static class Impl632 implements DemoService {public int type() {return 632;} public void run() {System.out.println(632);}}
    public static class Impl633 implements DemoService {public int type() {return 633;} public void run() {System.out.println(633);}}
    public static class Impl634 implements DemoService {public int type() {return 634;} public void run() {System.out.println(634);}}
    public static class Impl635 implements DemoService {public int type() {return 635;} public void run() {System.out.println(635);}}
    public static class Impl636 implements DemoService {public int type() {return 636;} public void run() {System.out.println(636);}}
    public static class Impl637 implements DemoService {public int type() {return 637;} public void run() {System.out.println(637);}}
    public static class Impl638 implements DemoService {public int type() {return 638;} public void run() {System.out.println(638);}}
    public static class Impl639 implements DemoService {public int type() {return 639;} public void run() {System.out.println(639);}}
    public static class Impl640 implements DemoService {public int type() {return 640;} public void run() {System.out.println(640);}}
    public static class Impl641 implements DemoService {public int type() {return 641;} public void run() {System.out.println(641);}}
    public static class Impl642 implements DemoService {public int type() {return 642;} public void run() {System.out.println(642);}}
    public static class Impl643 implements DemoService {public int type() {return 643;} public void run() {System.out.println(643);}}
    public static class Impl644 implements DemoService {public int type() {return 644;} public void run() {System.out.println(644);}}
    public static class Impl645 implements DemoService {public int type() {return 645;} public void run() {System.out.println(645);}}
    public static class Impl646 implements DemoService {public int type() {return 646;} public void run() {System.out.println(646);}}
    public static class Impl647 implements DemoService {public int type() {return 647;} public void run() {System.out.println(647);}}
    public static class Impl648 implements DemoService {public int type() {return 648;} public void run() {System.out.println(648);}}
    public static class Impl649 implements DemoService {public int type() {return 649;} public void run() {System.out.println(649);}}
    public static class Impl650 implements DemoService {public int type() {return 650;} public void run() {System.out.println(650);}}
    public static class Impl651 implements DemoService {public int type() {return 651;} public void run() {System.out.println(651);}}
    public static class Impl652 implements DemoService {public int type() {return 652;} public void run() {System.out.println(652);}}
    public static class Impl653 implements DemoService {public int type() {return 653;} public void run() {System.out.println(653);}}
    public static class Impl654 implements DemoService {public int type() {return 654;} public void run() {System.out.println(654);}}
    public static class Impl655 implements DemoService {public int type() {return 655;} public void run() {System.out.println(655);}}
    public static class Impl656 implements DemoService {public int type() {return 656;} public void run() {System.out.println(656);}}
    public static class Impl657 implements DemoService {public int type() {return 657;} public void run() {System.out.println(657);}}
    public static class Impl658 implements DemoService {public int type() {return 658;} public void run() {System.out.println(658);}}
    public static class Impl659 implements DemoService {public int type() {return 659;} public void run() {System.out.println(659);}}
    public static class Impl660 implements DemoService {public int type() {return 660;} public void run() {System.out.println(660);}}
    public static class Impl661 implements DemoService {public int type() {return 661;} public void run() {System.out.println(661);}}
    public static class Impl662 implements DemoService {public int type() {return 662;} public void run() {System.out.println(662);}}
    public static class Impl663 implements DemoService {public int type() {return 663;} public void run() {System.out.println(663);}}
    public static class Impl664 implements DemoService {public int type() {return 664;} public void run() {System.out.println(664);}}
    public static class Impl665 implements DemoService {public int type() {return 665;} public void run() {System.out.println(665);}}
    public static class Impl666 implements DemoService {public int type() {return 666;} public void run() {System.out.println(666);}}
    public static class Impl667 implements DemoService {public int type() {return 667;} public void run() {System.out.println(667);}}
    public static class Impl668 implements DemoService {public int type() {return 668;} public void run() {System.out.println(668);}}
    public static class Impl669 implements DemoService {public int type() {return 669;} public void run() {System.out.println(669);}}
    public static class Impl670 implements DemoService {public int type() {return 670;} public void run() {System.out.println(670);}}
    public static class Impl671 implements DemoService {public int type() {return 671;} public void run() {System.out.println(671);}}
    public static class Impl672 implements DemoService {public int type() {return 672;} public void run() {System.out.println(672);}}
    public static class Impl673 implements DemoService {public int type() {return 673;} public void run() {System.out.println(673);}}
    public static class Impl674 implements DemoService {public int type() {return 674;} public void run() {System.out.println(674);}}
    public static class Impl675 implements DemoService {public int type() {return 675;} public void run() {System.out.println(675);}}
    public static class Impl676 implements DemoService {public int type() {return 676;} public void run() {System.out.println(676);}}
    public static class Impl677 implements DemoService {public int type() {return 677;} public void run() {System.out.println(677);}}
    public static class Impl678 implements DemoService {public int type() {return 678;} public void run() {System.out.println(678);}}
    public static class Impl679 implements DemoService {public int type() {return 679;} public void run() {System.out.println(679);}}
    public static class Impl680 implements DemoService {public int type() {return 680;} public void run() {System.out.println(680);}}
    public static class Impl681 implements DemoService {public int type() {return 681;} public void run() {System.out.println(681);}}
    public static class Impl682 implements DemoService {public int type() {return 682;} public void run() {System.out.println(682);}}
    public static class Impl683 implements DemoService {public int type() {return 683;} public void run() {System.out.println(683);}}
    public static class Impl684 implements DemoService {public int type() {return 684;} public void run() {System.out.println(684);}}
    public static class Impl685 implements DemoService {public int type() {return 685;} public void run() {System.out.println(685);}}
    public static class Impl686 implements DemoService {public int type() {return 686;} public void run() {System.out.println(686);}}
    public static class Impl687 implements DemoService {public int type() {return 687;} public void run() {System.out.println(687);}}
    public static class Impl688 implements DemoService {public int type() {return 688;} public void run() {System.out.println(688);}}
    public static class Impl689 implements DemoService {public int type() {return 689;} public void run() {System.out.println(689);}}
    public static class Impl690 implements DemoService {public int type() {return 690;} public void run() {System.out.println(690);}}
    public static class Impl691 implements DemoService {public int type() {return 691;} public void run() {System.out.println(691);}}
    public static class Impl692 implements DemoService {public int type() {return 692;} public void run() {System.out.println(692);}}
    public static class Impl693 implements DemoService {public int type() {return 693;} public void run() {System.out.println(693);}}
    public static class Impl694 implements DemoService {public int type() {return 694;} public void run() {System.out.println(694);}}
    public static class Impl695 implements DemoService {public int type() {return 695;} public void run() {System.out.println(695);}}
    public static class Impl696 implements DemoService {public int type() {return 696;} public void run() {System.out.println(696);}}
    public static class Impl697 implements DemoService {public int type() {return 697;} public void run() {System.out.println(697);}}
    public static class Impl698 implements DemoService {public int type() {return 698;} public void run() {System.out.println(698);}}
    public static class Impl699 implements DemoService {public int type() {return 699;} public void run() {System.out.println(699);}}
    public static class Impl700 implements DemoService {public int type() {return 700;} public void run() {System.out.println(700);}}
    public static class Impl701 implements DemoService {public int type() {return 701;} public void run() {System.out.println(701);}}
    public static class Impl702 implements DemoService {public int type() {return 702;} public void run() {System.out.println(702);}}
    public static class Impl703 implements DemoService {public int type() {return 703;} public void run() {System.out.println(703);}}
    public static class Impl704 implements DemoService {public int type() {return 704;} public void run() {System.out.println(704);}}
    public static class Impl705 implements DemoService {public int type() {return 705;} public void run() {System.out.println(705);}}
    public static class Impl706 implements DemoService {public int type() {return 706;} public void run() {System.out.println(706);}}
    public static class Impl707 implements DemoService {public int type() {return 707;} public void run() {System.out.println(707);}}
    public static class Impl708 implements DemoService {public int type() {return 708;} public void run() {System.out.println(708);}}
    public static class Impl709 implements DemoService {public int type() {return 709;} public void run() {System.out.println(709);}}
    public static class Impl710 implements DemoService {public int type() {return 710;} public void run() {System.out.println(710);}}
    public static class Impl711 implements DemoService {public int type() {return 711;} public void run() {System.out.println(711);}}
    public static class Impl712 implements DemoService {public int type() {return 712;} public void run() {System.out.println(712);}}
    public static class Impl713 implements DemoService {public int type() {return 713;} public void run() {System.out.println(713);}}
    public static class Impl714 implements DemoService {public int type() {return 714;} public void run() {System.out.println(714);}}
    public static class Impl715 implements DemoService {public int type() {return 715;} public void run() {System.out.println(715);}}
    public static class Impl716 implements DemoService {public int type() {return 716;} public void run() {System.out.println(716);}}
    public static class Impl717 implements DemoService {public int type() {return 717;} public void run() {System.out.println(717);}}
    public static class Impl718 implements DemoService {public int type() {return 718;} public void run() {System.out.println(718);}}
    public static class Impl719 implements DemoService {public int type() {return 719;} public void run() {System.out.println(719);}}
    public static class Impl720 implements DemoService {public int type() {return 720;} public void run() {System.out.println(720);}}
    public static class Impl721 implements DemoService {public int type() {return 721;} public void run() {System.out.println(721);}}
    public static class Impl722 implements DemoService {public int type() {return 722;} public void run() {System.out.println(722);}}
    public static class Impl723 implements DemoService {public int type() {return 723;} public void run() {System.out.println(723);}}
    public static class Impl724 implements DemoService {public int type() {return 724;} public void run() {System.out.println(724);}}
    public static class Impl725 implements DemoService {public int type() {return 725;} public void run() {System.out.println(725);}}
    public static class Impl726 implements DemoService {public int type() {return 726;} public void run() {System.out.println(726);}}
    public static class Impl727 implements DemoService {public int type() {return 727;} public void run() {System.out.println(727);}}
    public static class Impl728 implements DemoService {public int type() {return 728;} public void run() {System.out.println(728);}}
    public static class Impl729 implements DemoService {public int type() {return 729;} public void run() {System.out.println(729);}}
    public static class Impl730 implements DemoService {public int type() {return 730;} public void run() {System.out.println(730);}}
    public static class Impl731 implements DemoService {public int type() {return 731;} public void run() {System.out.println(731);}}
    public static class Impl732 implements DemoService {public int type() {return 732;} public void run() {System.out.println(732);}}
    public static class Impl733 implements DemoService {public int type() {return 733;} public void run() {System.out.println(733);}}
    public static class Impl734 implements DemoService {public int type() {return 734;} public void run() {System.out.println(734);}}
    public static class Impl735 implements DemoService {public int type() {return 735;} public void run() {System.out.println(735);}}
    public static class Impl736 implements DemoService {public int type() {return 736;} public void run() {System.out.println(736);}}
    public static class Impl737 implements DemoService {public int type() {return 737;} public void run() {System.out.println(737);}}
    public static class Impl738 implements DemoService {public int type() {return 738;} public void run() {System.out.println(738);}}
    public static class Impl739 implements DemoService {public int type() {return 739;} public void run() {System.out.println(739);}}
    public static class Impl740 implements DemoService {public int type() {return 740;} public void run() {System.out.println(740);}}
    public static class Impl741 implements DemoService {public int type() {return 741;} public void run() {System.out.println(741);}}
    public static class Impl742 implements DemoService {public int type() {return 742;} public void run() {System.out.println(742);}}
    public static class Impl743 implements DemoService {public int type() {return 743;} public void run() {System.out.println(743);}}
    public static class Impl744 implements DemoService {public int type() {return 744;} public void run() {System.out.println(744);}}
    public static class Impl745 implements DemoService {public int type() {return 745;} public void run() {System.out.println(745);}}
    public static class Impl746 implements DemoService {public int type() {return 746;} public void run() {System.out.println(746);}}
    public static class Impl747 implements DemoService {public int type() {return 747;} public void run() {System.out.println(747);}}
    public static class Impl748 implements DemoService {public int type() {return 748;} public void run() {System.out.println(748);}}
    public static class Impl749 implements DemoService {public int type() {return 749;} public void run() {System.out.println(749);}}
    public static class Impl750 implements DemoService {public int type() {return 750;} public void run() {System.out.println(750);}}
    public static class Impl751 implements DemoService {public int type() {return 751;} public void run() {System.out.println(751);}}
    public static class Impl752 implements DemoService {public int type() {return 752;} public void run() {System.out.println(752);}}
    public static class Impl753 implements DemoService {public int type() {return 753;} public void run() {System.out.println(753);}}
    public static class Impl754 implements DemoService {public int type() {return 754;} public void run() {System.out.println(754);}}
    public static class Impl755 implements DemoService {public int type() {return 755;} public void run() {System.out.println(755);}}
    public static class Impl756 implements DemoService {public int type() {return 756;} public void run() {System.out.println(756);}}
    public static class Impl757 implements DemoService {public int type() {return 757;} public void run() {System.out.println(757);}}
    public static class Impl758 implements DemoService {public int type() {return 758;} public void run() {System.out.println(758);}}
    public static class Impl759 implements DemoService {public int type() {return 759;} public void run() {System.out.println(759);}}
    public static class Impl760 implements DemoService {public int type() {return 760;} public void run() {System.out.println(760);}}
    public static class Impl761 implements DemoService {public int type() {return 761;} public void run() {System.out.println(761);}}
    public static class Impl762 implements DemoService {public int type() {return 762;} public void run() {System.out.println(762);}}
    public static class Impl763 implements DemoService {public int type() {return 763;} public void run() {System.out.println(763);}}
    public static class Impl764 implements DemoService {public int type() {return 764;} public void run() {System.out.println(764);}}
    public static class Impl765 implements DemoService {public int type() {return 765;} public void run() {System.out.println(765);}}
    public static class Impl766 implements DemoService {public int type() {return 766;} public void run() {System.out.println(766);}}
    public static class Impl767 implements DemoService {public int type() {return 767;} public void run() {System.out.println(767);}}
    public static class Impl768 implements DemoService {public int type() {return 768;} public void run() {System.out.println(768);}}
    public static class Impl769 implements DemoService {public int type() {return 769;} public void run() {System.out.println(769);}}
    public static class Impl770 implements DemoService {public int type() {return 770;} public void run() {System.out.println(770);}}
    public static class Impl771 implements DemoService {public int type() {return 771;} public void run() {System.out.println(771);}}
    public static class Impl772 implements DemoService {public int type() {return 772;} public void run() {System.out.println(772);}}
    public static class Impl773 implements DemoService {public int type() {return 773;} public void run() {System.out.println(773);}}
    public static class Impl774 implements DemoService {public int type() {return 774;} public void run() {System.out.println(774);}}
    public static class Impl775 implements DemoService {public int type() {return 775;} public void run() {System.out.println(775);}}
    public static class Impl776 implements DemoService {public int type() {return 776;} public void run() {System.out.println(776);}}
    public static class Impl777 implements DemoService {public int type() {return 777;} public void run() {System.out.println(777);}}
    public static class Impl778 implements DemoService {public int type() {return 778;} public void run() {System.out.println(778);}}
    public static class Impl779 implements DemoService {public int type() {return 779;} public void run() {System.out.println(779);}}
    public static class Impl780 implements DemoService {public int type() {return 780;} public void run() {System.out.println(780);}}
    public static class Impl781 implements DemoService {public int type() {return 781;} public void run() {System.out.println(781);}}
    public static class Impl782 implements DemoService {public int type() {return 782;} public void run() {System.out.println(782);}}
    public static class Impl783 implements DemoService {public int type() {return 783;} public void run() {System.out.println(783);}}
    public static class Impl784 implements DemoService {public int type() {return 784;} public void run() {System.out.println(784);}}
    public static class Impl785 implements DemoService {public int type() {return 785;} public void run() {System.out.println(785);}}
    public static class Impl786 implements DemoService {public int type() {return 786;} public void run() {System.out.println(786);}}
    public static class Impl787 implements DemoService {public int type() {return 787;} public void run() {System.out.println(787);}}
    public static class Impl788 implements DemoService {public int type() {return 788;} public void run() {System.out.println(788);}}
    public static class Impl789 implements DemoService {public int type() {return 789;} public void run() {System.out.println(789);}}
    public static class Impl790 implements DemoService {public int type() {return 790;} public void run() {System.out.println(790);}}
    public static class Impl791 implements DemoService {public int type() {return 791;} public void run() {System.out.println(791);}}
    public static class Impl792 implements DemoService {public int type() {return 792;} public void run() {System.out.println(792);}}
    public static class Impl793 implements DemoService {public int type() {return 793;} public void run() {System.out.println(793);}}
    public static class Impl794 implements DemoService {public int type() {return 794;} public void run() {System.out.println(794);}}
    public static class Impl795 implements DemoService {public int type() {return 795;} public void run() {System.out.println(795);}}
    public static class Impl796 implements DemoService {public int type() {return 796;} public void run() {System.out.println(796);}}
    public static class Impl797 implements DemoService {public int type() {return 797;} public void run() {System.out.println(797);}}
    public static class Impl798 implements DemoService {public int type() {return 798;} public void run() {System.out.println(798);}}
    public static class Impl799 implements DemoService {public int type() {return 799;} public void run() {System.out.println(799);}}
    public static class Impl800 implements DemoService {public int type() {return 800;} public void run() {System.out.println(800);}}
    public static class Impl801 implements DemoService {public int type() {return 801;} public void run() {System.out.println(801);}}
    public static class Impl802 implements DemoService {public int type() {return 802;} public void run() {System.out.println(802);}}
    public static class Impl803 implements DemoService {public int type() {return 803;} public void run() {System.out.println(803);}}
    public static class Impl804 implements DemoService {public int type() {return 804;} public void run() {System.out.println(804);}}
    public static class Impl805 implements DemoService {public int type() {return 805;} public void run() {System.out.println(805);}}
    public static class Impl806 implements DemoService {public int type() {return 806;} public void run() {System.out.println(806);}}
    public static class Impl807 implements DemoService {public int type() {return 807;} public void run() {System.out.println(807);}}
    public static class Impl808 implements DemoService {public int type() {return 808;} public void run() {System.out.println(808);}}
    public static class Impl809 implements DemoService {public int type() {return 809;} public void run() {System.out.println(809);}}
    public static class Impl810 implements DemoService {public int type() {return 810;} public void run() {System.out.println(810);}}
    public static class Impl811 implements DemoService {public int type() {return 811;} public void run() {System.out.println(811);}}
    public static class Impl812 implements DemoService {public int type() {return 812;} public void run() {System.out.println(812);}}
    public static class Impl813 implements DemoService {public int type() {return 813;} public void run() {System.out.println(813);}}
    public static class Impl814 implements DemoService {public int type() {return 814;} public void run() {System.out.println(814);}}
    public static class Impl815 implements DemoService {public int type() {return 815;} public void run() {System.out.println(815);}}
    public static class Impl816 implements DemoService {public int type() {return 816;} public void run() {System.out.println(816);}}
    public static class Impl817 implements DemoService {public int type() {return 817;} public void run() {System.out.println(817);}}
    public static class Impl818 implements DemoService {public int type() {return 818;} public void run() {System.out.println(818);}}
    public static class Impl819 implements DemoService {public int type() {return 819;} public void run() {System.out.println(819);}}
    public static class Impl820 implements DemoService {public int type() {return 820;} public void run() {System.out.println(820);}}
    public static class Impl821 implements DemoService {public int type() {return 821;} public void run() {System.out.println(821);}}
    public static class Impl822 implements DemoService {public int type() {return 822;} public void run() {System.out.println(822);}}
    public static class Impl823 implements DemoService {public int type() {return 823;} public void run() {System.out.println(823);}}
    public static class Impl824 implements DemoService {public int type() {return 824;} public void run() {System.out.println(824);}}
    public static class Impl825 implements DemoService {public int type() {return 825;} public void run() {System.out.println(825);}}
    public static class Impl826 implements DemoService {public int type() {return 826;} public void run() {System.out.println(826);}}
    public static class Impl827 implements DemoService {public int type() {return 827;} public void run() {System.out.println(827);}}
    public static class Impl828 implements DemoService {public int type() {return 828;} public void run() {System.out.println(828);}}
    public static class Impl829 implements DemoService {public int type() {return 829;} public void run() {System.out.println(829);}}
    public static class Impl830 implements DemoService {public int type() {return 830;} public void run() {System.out.println(830);}}
    public static class Impl831 implements DemoService {public int type() {return 831;} public void run() {System.out.println(831);}}
    public static class Impl832 implements DemoService {public int type() {return 832;} public void run() {System.out.println(832);}}
    public static class Impl833 implements DemoService {public int type() {return 833;} public void run() {System.out.println(833);}}
    public static class Impl834 implements DemoService {public int type() {return 834;} public void run() {System.out.println(834);}}
    public static class Impl835 implements DemoService {public int type() {return 835;} public void run() {System.out.println(835);}}
    public static class Impl836 implements DemoService {public int type() {return 836;} public void run() {System.out.println(836);}}
    public static class Impl837 implements DemoService {public int type() {return 837;} public void run() {System.out.println(837);}}
    public static class Impl838 implements DemoService {public int type() {return 838;} public void run() {System.out.println(838);}}
    public static class Impl839 implements DemoService {public int type() {return 839;} public void run() {System.out.println(839);}}
    public static class Impl840 implements DemoService {public int type() {return 840;} public void run() {System.out.println(840);}}
    public static class Impl841 implements DemoService {public int type() {return 841;} public void run() {System.out.println(841);}}
    public static class Impl842 implements DemoService {public int type() {return 842;} public void run() {System.out.println(842);}}
    public static class Impl843 implements DemoService {public int type() {return 843;} public void run() {System.out.println(843);}}
    public static class Impl844 implements DemoService {public int type() {return 844;} public void run() {System.out.println(844);}}
    public static class Impl845 implements DemoService {public int type() {return 845;} public void run() {System.out.println(845);}}
    public static class Impl846 implements DemoService {public int type() {return 846;} public void run() {System.out.println(846);}}
    public static class Impl847 implements DemoService {public int type() {return 847;} public void run() {System.out.println(847);}}
    public static class Impl848 implements DemoService {public int type() {return 848;} public void run() {System.out.println(848);}}
    public static class Impl849 implements DemoService {public int type() {return 849;} public void run() {System.out.println(849);}}
    public static class Impl850 implements DemoService {public int type() {return 850;} public void run() {System.out.println(850);}}
    public static class Impl851 implements DemoService {public int type() {return 851;} public void run() {System.out.println(851);}}
    public static class Impl852 implements DemoService {public int type() {return 852;} public void run() {System.out.println(852);}}
    public static class Impl853 implements DemoService {public int type() {return 853;} public void run() {System.out.println(853);}}
    public static class Impl854 implements DemoService {public int type() {return 854;} public void run() {System.out.println(854);}}
    public static class Impl855 implements DemoService {public int type() {return 855;} public void run() {System.out.println(855);}}
    public static class Impl856 implements DemoService {public int type() {return 856;} public void run() {System.out.println(856);}}
    public static class Impl857 implements DemoService {public int type() {return 857;} public void run() {System.out.println(857);}}
    public static class Impl858 implements DemoService {public int type() {return 858;} public void run() {System.out.println(858);}}
    public static class Impl859 implements DemoService {public int type() {return 859;} public void run() {System.out.println(859);}}
    public static class Impl860 implements DemoService {public int type() {return 860;} public void run() {System.out.println(860);}}
    public static class Impl861 implements DemoService {public int type() {return 861;} public void run() {System.out.println(861);}}
    public static class Impl862 implements DemoService {public int type() {return 862;} public void run() {System.out.println(862);}}
    public static class Impl863 implements DemoService {public int type() {return 863;} public void run() {System.out.println(863);}}
    public static class Impl864 implements DemoService {public int type() {return 864;} public void run() {System.out.println(864);}}
    public static class Impl865 implements DemoService {public int type() {return 865;} public void run() {System.out.println(865);}}
    public static class Impl866 implements DemoService {public int type() {return 866;} public void run() {System.out.println(866);}}
    public static class Impl867 implements DemoService {public int type() {return 867;} public void run() {System.out.println(867);}}
    public static class Impl868 implements DemoService {public int type() {return 868;} public void run() {System.out.println(868);}}
    public static class Impl869 implements DemoService {public int type() {return 869;} public void run() {System.out.println(869);}}
    public static class Impl870 implements DemoService {public int type() {return 870;} public void run() {System.out.println(870);}}
    public static class Impl871 implements DemoService {public int type() {return 871;} public void run() {System.out.println(871);}}
    public static class Impl872 implements DemoService {public int type() {return 872;} public void run() {System.out.println(872);}}
    public static class Impl873 implements DemoService {public int type() {return 873;} public void run() {System.out.println(873);}}
    public static class Impl874 implements DemoService {public int type() {return 874;} public void run() {System.out.println(874);}}
    public static class Impl875 implements DemoService {public int type() {return 875;} public void run() {System.out.println(875);}}
    public static class Impl876 implements DemoService {public int type() {return 876;} public void run() {System.out.println(876);}}
    public static class Impl877 implements DemoService {public int type() {return 877;} public void run() {System.out.println(877);}}
    public static class Impl878 implements DemoService {public int type() {return 878;} public void run() {System.out.println(878);}}
    public static class Impl879 implements DemoService {public int type() {return 879;} public void run() {System.out.println(879);}}
    public static class Impl880 implements DemoService {public int type() {return 880;} public void run() {System.out.println(880);}}
    public static class Impl881 implements DemoService {public int type() {return 881;} public void run() {System.out.println(881);}}
    public static class Impl882 implements DemoService {public int type() {return 882;} public void run() {System.out.println(882);}}
    public static class Impl883 implements DemoService {public int type() {return 883;} public void run() {System.out.println(883);}}
    public static class Impl884 implements DemoService {public int type() {return 884;} public void run() {System.out.println(884);}}
    public static class Impl885 implements DemoService {public int type() {return 885;} public void run() {System.out.println(885);}}
    public static class Impl886 implements DemoService {public int type() {return 886;} public void run() {System.out.println(886);}}
    public static class Impl887 implements DemoService {public int type() {return 887;} public void run() {System.out.println(887);}}
    public static class Impl888 implements DemoService {public int type() {return 888;} public void run() {System.out.println(888);}}
    public static class Impl889 implements DemoService {public int type() {return 889;} public void run() {System.out.println(889);}}
    public static class Impl890 implements DemoService {public int type() {return 890;} public void run() {System.out.println(890);}}
    public static class Impl891 implements DemoService {public int type() {return 891;} public void run() {System.out.println(891);}}
    public static class Impl892 implements DemoService {public int type() {return 892;} public void run() {System.out.println(892);}}
    public static class Impl893 implements DemoService {public int type() {return 893;} public void run() {System.out.println(893);}}
    public static class Impl894 implements DemoService {public int type() {return 894;} public void run() {System.out.println(894);}}
    public static class Impl895 implements DemoService {public int type() {return 895;} public void run() {System.out.println(895);}}
    public static class Impl896 implements DemoService {public int type() {return 896;} public void run() {System.out.println(896);}}
    public static class Impl897 implements DemoService {public int type() {return 897;} public void run() {System.out.println(897);}}
    public static class Impl898 implements DemoService {public int type() {return 898;} public void run() {System.out.println(898);}}
    public static class Impl899 implements DemoService {public int type() {return 899;} public void run() {System.out.println(899);}}
    public static class Impl900 implements DemoService {public int type() {return 900;} public void run() {System.out.println(900);}}
    public static class Impl901 implements DemoService {public int type() {return 901;} public void run() {System.out.println(901);}}
    public static class Impl902 implements DemoService {public int type() {return 902;} public void run() {System.out.println(902);}}
    public static class Impl903 implements DemoService {public int type() {return 903;} public void run() {System.out.println(903);}}
    public static class Impl904 implements DemoService {public int type() {return 904;} public void run() {System.out.println(904);}}
    public static class Impl905 implements DemoService {public int type() {return 905;} public void run() {System.out.println(905);}}
    public static class Impl906 implements DemoService {public int type() {return 906;} public void run() {System.out.println(906);}}
    public static class Impl907 implements DemoService {public int type() {return 907;} public void run() {System.out.println(907);}}
    public static class Impl908 implements DemoService {public int type() {return 908;} public void run() {System.out.println(908);}}
    public static class Impl909 implements DemoService {public int type() {return 909;} public void run() {System.out.println(909);}}
    public static class Impl910 implements DemoService {public int type() {return 910;} public void run() {System.out.println(910);}}
    public static class Impl911 implements DemoService {public int type() {return 911;} public void run() {System.out.println(911);}}
    public static class Impl912 implements DemoService {public int type() {return 912;} public void run() {System.out.println(912);}}
    public static class Impl913 implements DemoService {public int type() {return 913;} public void run() {System.out.println(913);}}
    public static class Impl914 implements DemoService {public int type() {return 914;} public void run() {System.out.println(914);}}
    public static class Impl915 implements DemoService {public int type() {return 915;} public void run() {System.out.println(915);}}
    public static class Impl916 implements DemoService {public int type() {return 916;} public void run() {System.out.println(916);}}
    public static class Impl917 implements DemoService {public int type() {return 917;} public void run() {System.out.println(917);}}
    public static class Impl918 implements DemoService {public int type() {return 918;} public void run() {System.out.println(918);}}
    public static class Impl919 implements DemoService {public int type() {return 919;} public void run() {System.out.println(919);}}
    public static class Impl920 implements DemoService {public int type() {return 920;} public void run() {System.out.println(920);}}
    public static class Impl921 implements DemoService {public int type() {return 921;} public void run() {System.out.println(921);}}
    public static class Impl922 implements DemoService {public int type() {return 922;} public void run() {System.out.println(922);}}
    public static class Impl923 implements DemoService {public int type() {return 923;} public void run() {System.out.println(923);}}
    public static class Impl924 implements DemoService {public int type() {return 924;} public void run() {System.out.println(924);}}
    public static class Impl925 implements DemoService {public int type() {return 925;} public void run() {System.out.println(925);}}
    public static class Impl926 implements DemoService {public int type() {return 926;} public void run() {System.out.println(926);}}
    public static class Impl927 implements DemoService {public int type() {return 927;} public void run() {System.out.println(927);}}
    public static class Impl928 implements DemoService {public int type() {return 928;} public void run() {System.out.println(928);}}
    public static class Impl929 implements DemoService {public int type() {return 929;} public void run() {System.out.println(929);}}
    public static class Impl930 implements DemoService {public int type() {return 930;} public void run() {System.out.println(930);}}
    public static class Impl931 implements DemoService {public int type() {return 931;} public void run() {System.out.println(931);}}
    public static class Impl932 implements DemoService {public int type() {return 932;} public void run() {System.out.println(932);}}
    public static class Impl933 implements DemoService {public int type() {return 933;} public void run() {System.out.println(933);}}
    public static class Impl934 implements DemoService {public int type() {return 934;} public void run() {System.out.println(934);}}
    public static class Impl935 implements DemoService {public int type() {return 935;} public void run() {System.out.println(935);}}
    public static class Impl936 implements DemoService {public int type() {return 936;} public void run() {System.out.println(936);}}
    public static class Impl937 implements DemoService {public int type() {return 937;} public void run() {System.out.println(937);}}
    public static class Impl938 implements DemoService {public int type() {return 938;} public void run() {System.out.println(938);}}
    public static class Impl939 implements DemoService {public int type() {return 939;} public void run() {System.out.println(939);}}
    public static class Impl940 implements DemoService {public int type() {return 940;} public void run() {System.out.println(940);}}
    public static class Impl941 implements DemoService {public int type() {return 941;} public void run() {System.out.println(941);}}
    public static class Impl942 implements DemoService {public int type() {return 942;} public void run() {System.out.println(942);}}
    public static class Impl943 implements DemoService {public int type() {return 943;} public void run() {System.out.println(943);}}
    public static class Impl944 implements DemoService {public int type() {return 944;} public void run() {System.out.println(944);}}
    public static class Impl945 implements DemoService {public int type() {return 945;} public void run() {System.out.println(945);}}
    public static class Impl946 implements DemoService {public int type() {return 946;} public void run() {System.out.println(946);}}
    public static class Impl947 implements DemoService {public int type() {return 947;} public void run() {System.out.println(947);}}
    public static class Impl948 implements DemoService {public int type() {return 948;} public void run() {System.out.println(948);}}
    public static class Impl949 implements DemoService {public int type() {return 949;} public void run() {System.out.println(949);}}
    public static class Impl950 implements DemoService {public int type() {return 950;} public void run() {System.out.println(950);}}
    public static class Impl951 implements DemoService {public int type() {return 951;} public void run() {System.out.println(951);}}
    public static class Impl952 implements DemoService {public int type() {return 952;} public void run() {System.out.println(952);}}
    public static class Impl953 implements DemoService {public int type() {return 953;} public void run() {System.out.println(953);}}
    public static class Impl954 implements DemoService {public int type() {return 954;} public void run() {System.out.println(954);}}
    public static class Impl955 implements DemoService {public int type() {return 955;} public void run() {System.out.println(955);}}
    public static class Impl956 implements DemoService {public int type() {return 956;} public void run() {System.out.println(956);}}
    public static class Impl957 implements DemoService {public int type() {return 957;} public void run() {System.out.println(957);}}
    public static class Impl958 implements DemoService {public int type() {return 958;} public void run() {System.out.println(958);}}
    public static class Impl959 implements DemoService {public int type() {return 959;} public void run() {System.out.println(959);}}
    public static class Impl960 implements DemoService {public int type() {return 960;} public void run() {System.out.println(960);}}
    public static class Impl961 implements DemoService {public int type() {return 961;} public void run() {System.out.println(961);}}
    public static class Impl962 implements DemoService {public int type() {return 962;} public void run() {System.out.println(962);}}
    public static class Impl963 implements DemoService {public int type() {return 963;} public void run() {System.out.println(963);}}
    public static class Impl964 implements DemoService {public int type() {return 964;} public void run() {System.out.println(964);}}
    public static class Impl965 implements DemoService {public int type() {return 965;} public void run() {System.out.println(965);}}
    public static class Impl966 implements DemoService {public int type() {return 966;} public void run() {System.out.println(966);}}
    public static class Impl967 implements DemoService {public int type() {return 967;} public void run() {System.out.println(967);}}
    public static class Impl968 implements DemoService {public int type() {return 968;} public void run() {System.out.println(968);}}
    public static class Impl969 implements DemoService {public int type() {return 969;} public void run() {System.out.println(969);}}
    public static class Impl970 implements DemoService {public int type() {return 970;} public void run() {System.out.println(970);}}
    public static class Impl971 implements DemoService {public int type() {return 971;} public void run() {System.out.println(971);}}
    public static class Impl972 implements DemoService {public int type() {return 972;} public void run() {System.out.println(972);}}
    public static class Impl973 implements DemoService {public int type() {return 973;} public void run() {System.out.println(973);}}
    public static class Impl974 implements DemoService {public int type() {return 974;} public void run() {System.out.println(974);}}
    public static class Impl975 implements DemoService {public int type() {return 975;} public void run() {System.out.println(975);}}
    public static class Impl976 implements DemoService {public int type() {return 976;} public void run() {System.out.println(976);}}
    public static class Impl977 implements DemoService {public int type() {return 977;} public void run() {System.out.println(977);}}
    public static class Impl978 implements DemoService {public int type() {return 978;} public void run() {System.out.println(978);}}
    public static class Impl979 implements DemoService {public int type() {return 979;} public void run() {System.out.println(979);}}
    public static class Impl980 implements DemoService {public int type() {return 980;} public void run() {System.out.println(980);}}
    public static class Impl981 implements DemoService {public int type() {return 981;} public void run() {System.out.println(981);}}
    public static class Impl982 implements DemoService {public int type() {return 982;} public void run() {System.out.println(982);}}
    public static class Impl983 implements DemoService {public int type() {return 983;} public void run() {System.out.println(983);}}
    public static class Impl984 implements DemoService {public int type() {return 984;} public void run() {System.out.println(984);}}
    public static class Impl985 implements DemoService {public int type() {return 985;} public void run() {System.out.println(985);}}
    public static class Impl986 implements DemoService {public int type() {return 986;} public void run() {System.out.println(986);}}
    public static class Impl987 implements DemoService {public int type() {return 987;} public void run() {System.out.println(987);}}
    public static class Impl988 implements DemoService {public int type() {return 988;} public void run() {System.out.println(988);}}
    public static class Impl989 implements DemoService {public int type() {return 989;} public void run() {System.out.println(989);}}
    public static class Impl990 implements DemoService {public int type() {return 990;} public void run() {System.out.println(990);}}
    public static class Impl991 implements DemoService {public int type() {return 991;} public void run() {System.out.println(991);}}
    public static class Impl992 implements DemoService {public int type() {return 992;} public void run() {System.out.println(992);}}
    public static class Impl993 implements DemoService {public int type() {return 993;} public void run() {System.out.println(993);}}
    public static class Impl994 implements DemoService {public int type() {return 994;} public void run() {System.out.println(994);}}
    public static class Impl995 implements DemoService {public int type() {return 995;} public void run() {System.out.println(995);}}
    public static class Impl996 implements DemoService {public int type() {return 996;} public void run() {System.out.println(996);}}
    public static class Impl997 implements DemoService {public int type() {return 997;} public void run() {System.out.println(997);}}
    public static class Impl998 implements DemoService {public int type() {return 998;} public void run() {System.out.println(998);}}
    public static class Impl999 implements DemoService {public int type() {return 999;} public void run() {System.out.println(999);}}
    public static class Impl1000 implements DemoService {public int type() {return 1000;} public void run() {System.out.println(1000);}}
    public static class Impl1001 implements DemoService {public int type() {return 1001;} public void run() {System.out.println(1001);}}
    public static class Impl1002 implements DemoService {public int type() {return 1002;} public void run() {System.out.println(1002);}}
    public static class Impl1003 implements DemoService {public int type() {return 1003;} public void run() {System.out.println(1003);}}
    public static class Impl1004 implements DemoService {public int type() {return 1004;} public void run() {System.out.println(1004);}}
    public static class Impl1005 implements DemoService {public int type() {return 1005;} public void run() {System.out.println(1005);}}
    public static class Impl1006 implements DemoService {public int type() {return 1006;} public void run() {System.out.println(1006);}}
    public static class Impl1007 implements DemoService {public int type() {return 1007;} public void run() {System.out.println(1007);}}
    public static class Impl1008 implements DemoService {public int type() {return 1008;} public void run() {System.out.println(1008);}}
    public static class Impl1009 implements DemoService {public int type() {return 1009;} public void run() {System.out.println(1009);}}
    public static class Impl1010 implements DemoService {public int type() {return 1010;} public void run() {System.out.println(1010);}}
    public static class Impl1011 implements DemoService {public int type() {return 1011;} public void run() {System.out.println(1011);}}
    public static class Impl1012 implements DemoService {public int type() {return 1012;} public void run() {System.out.println(1012);}}
    public static class Impl1013 implements DemoService {public int type() {return 1013;} public void run() {System.out.println(1013);}}
    public static class Impl1014 implements DemoService {public int type() {return 1014;} public void run() {System.out.println(1014);}}
    public static class Impl1015 implements DemoService {public int type() {return 1015;} public void run() {System.out.println(1015);}}
    public static class Impl1016 implements DemoService {public int type() {return 1016;} public void run() {System.out.println(1016);}}
    public static class Impl1017 implements DemoService {public int type() {return 1017;} public void run() {System.out.println(1017);}}
    public static class Impl1018 implements DemoService {public int type() {return 1018;} public void run() {System.out.println(1018);}}
    public static class Impl1019 implements DemoService {public int type() {return 1019;} public void run() {System.out.println(1019);}}
    public static class Impl1020 implements DemoService {public int type() {return 1020;} public void run() {System.out.println(1020);}}
    public static class Impl1021 implements DemoService {public int type() {return 1021;} public void run() {System.out.println(1021);}}
    public static class Impl1022 implements DemoService {public int type() {return 1022;} public void run() {System.out.println(1022);}}
    public static class Impl1023 implements DemoService {public int type() {return 1023;} public void run() {System.out.println(1023);}}
    public static class Impl1024 implements DemoService {public int type() {return 1024;} public void run() {System.out.println(1024);}}
    public static class Impl1025 implements DemoService {public int type() {return 1025;} public void run() {System.out.println(1025);}}
    public static class Impl1026 implements DemoService {public int type() {return 1026;} public void run() {System.out.println(1026);}}
    public static class Impl1027 implements DemoService {public int type() {return 1027;} public void run() {System.out.println(1027);}}
    public static class Impl1028 implements DemoService {public int type() {return 1028;} public void run() {System.out.println(1028);}}
    public static class Impl1029 implements DemoService {public int type() {return 1029;} public void run() {System.out.println(1029);}}
    public static class Impl1030 implements DemoService {public int type() {return 1030;} public void run() {System.out.println(1030);}}
    public static class Impl1031 implements DemoService {public int type() {return 1031;} public void run() {System.out.println(1031);}}
    public static class Impl1032 implements DemoService {public int type() {return 1032;} public void run() {System.out.println(1032);}}
    public static class Impl1033 implements DemoService {public int type() {return 1033;} public void run() {System.out.println(1033);}}
    public static class Impl1034 implements DemoService {public int type() {return 1034;} public void run() {System.out.println(1034);}}
    public static class Impl1035 implements DemoService {public int type() {return 1035;} public void run() {System.out.println(1035);}}
    public static class Impl1036 implements DemoService {public int type() {return 1036;} public void run() {System.out.println(1036);}}
    public static class Impl1037 implements DemoService {public int type() {return 1037;} public void run() {System.out.println(1037);}}
    public static class Impl1038 implements DemoService {public int type() {return 1038;} public void run() {System.out.println(1038);}}
    public static class Impl1039 implements DemoService {public int type() {return 1039;} public void run() {System.out.println(1039);}}
    public static class Impl1040 implements DemoService {public int type() {return 1040;} public void run() {System.out.println(1040);}}
    public static class Impl1041 implements DemoService {public int type() {return 1041;} public void run() {System.out.println(1041);}}
    public static class Impl1042 implements DemoService {public int type() {return 1042;} public void run() {System.out.println(1042);}}
    public static class Impl1043 implements DemoService {public int type() {return 1043;} public void run() {System.out.println(1043);}}
    public static class Impl1044 implements DemoService {public int type() {return 1044;} public void run() {System.out.println(1044);}}
    public static class Impl1045 implements DemoService {public int type() {return 1045;} public void run() {System.out.println(1045);}}
    public static class Impl1046 implements DemoService {public int type() {return 1046;} public void run() {System.out.println(1046);}}
    public static class Impl1047 implements DemoService {public int type() {return 1047;} public void run() {System.out.println(1047);}}
    public static class Impl1048 implements DemoService {public int type() {return 1048;} public void run() {System.out.println(1048);}}
    public static class Impl1049 implements DemoService {public int type() {return 1049;} public void run() {System.out.println(1049);}}
    public static class Impl1050 implements DemoService {public int type() {return 1050;} public void run() {System.out.println(1050);}}
    public static class Impl1051 implements DemoService {public int type() {return 1051;} public void run() {System.out.println(1051);}}
    public static class Impl1052 implements DemoService {public int type() {return 1052;} public void run() {System.out.println(1052);}}
    public static class Impl1053 implements DemoService {public int type() {return 1053;} public void run() {System.out.println(1053);}}
    public static class Impl1054 implements DemoService {public int type() {return 1054;} public void run() {System.out.println(1054);}}
    public static class Impl1055 implements DemoService {public int type() {return 1055;} public void run() {System.out.println(1055);}}
    public static class Impl1056 implements DemoService {public int type() {return 1056;} public void run() {System.out.println(1056);}}
    public static class Impl1057 implements DemoService {public int type() {return 1057;} public void run() {System.out.println(1057);}}
    public static class Impl1058 implements DemoService {public int type() {return 1058;} public void run() {System.out.println(1058);}}
    public static class Impl1059 implements DemoService {public int type() {return 1059;} public void run() {System.out.println(1059);}}
    public static class Impl1060 implements DemoService {public int type() {return 1060;} public void run() {System.out.println(1060);}}
    public static class Impl1061 implements DemoService {public int type() {return 1061;} public void run() {System.out.println(1061);}}
    public static class Impl1062 implements DemoService {public int type() {return 1062;} public void run() {System.out.println(1062);}}
    public static class Impl1063 implements DemoService {public int type() {return 1063;} public void run() {System.out.println(1063);}}
    public static class Impl1064 implements DemoService {public int type() {return 1064;} public void run() {System.out.println(1064);}}
    public static class Impl1065 implements DemoService {public int type() {return 1065;} public void run() {System.out.println(1065);}}
    public static class Impl1066 implements DemoService {public int type() {return 1066;} public void run() {System.out.println(1066);}}
    public static class Impl1067 implements DemoService {public int type() {return 1067;} public void run() {System.out.println(1067);}}
    public static class Impl1068 implements DemoService {public int type() {return 1068;} public void run() {System.out.println(1068);}}
    public static class Impl1069 implements DemoService {public int type() {return 1069;} public void run() {System.out.println(1069);}}
    public static class Impl1070 implements DemoService {public int type() {return 1070;} public void run() {System.out.println(1070);}}
    public static class Impl1071 implements DemoService {public int type() {return 1071;} public void run() {System.out.println(1071);}}
    public static class Impl1072 implements DemoService {public int type() {return 1072;} public void run() {System.out.println(1072);}}
    public static class Impl1073 implements DemoService {public int type() {return 1073;} public void run() {System.out.println(1073);}}
    public static class Impl1074 implements DemoService {public int type() {return 1074;} public void run() {System.out.println(1074);}}
    public static class Impl1075 implements DemoService {public int type() {return 1075;} public void run() {System.out.println(1075);}}
    public static class Impl1076 implements DemoService {public int type() {return 1076;} public void run() {System.out.println(1076);}}
    public static class Impl1077 implements DemoService {public int type() {return 1077;} public void run() {System.out.println(1077);}}
    public static class Impl1078 implements DemoService {public int type() {return 1078;} public void run() {System.out.println(1078);}}
    public static class Impl1079 implements DemoService {public int type() {return 1079;} public void run() {System.out.println(1079);}}
    public static class Impl1080 implements DemoService {public int type() {return 1080;} public void run() {System.out.println(1080);}}
    public static class Impl1081 implements DemoService {public int type() {return 1081;} public void run() {System.out.println(1081);}}
    public static class Impl1082 implements DemoService {public int type() {return 1082;} public void run() {System.out.println(1082);}}
    public static class Impl1083 implements DemoService {public int type() {return 1083;} public void run() {System.out.println(1083);}}
    public static class Impl1084 implements DemoService {public int type() {return 1084;} public void run() {System.out.println(1084);}}
    public static class Impl1085 implements DemoService {public int type() {return 1085;} public void run() {System.out.println(1085);}}
    public static class Impl1086 implements DemoService {public int type() {return 1086;} public void run() {System.out.println(1086);}}
    public static class Impl1087 implements DemoService {public int type() {return 1087;} public void run() {System.out.println(1087);}}
    public static class Impl1088 implements DemoService {public int type() {return 1088;} public void run() {System.out.println(1088);}}
    public static class Impl1089 implements DemoService {public int type() {return 1089;} public void run() {System.out.println(1089);}}
    public static class Impl1090 implements DemoService {public int type() {return 1090;} public void run() {System.out.println(1090);}}
    public static class Impl1091 implements DemoService {public int type() {return 1091;} public void run() {System.out.println(1091);}}
    public static class Impl1092 implements DemoService {public int type() {return 1092;} public void run() {System.out.println(1092);}}
    public static class Impl1093 implements DemoService {public int type() {return 1093;} public void run() {System.out.println(1093);}}
    public static class Impl1094 implements DemoService {public int type() {return 1094;} public void run() {System.out.println(1094);}}
    public static class Impl1095 implements DemoService {public int type() {return 1095;} public void run() {System.out.println(1095);}}
    public static class Impl1096 implements DemoService {public int type() {return 1096;} public void run() {System.out.println(1096);}}
    public static class Impl1097 implements DemoService {public int type() {return 1097;} public void run() {System.out.println(1097);}}
    public static class Impl1098 implements DemoService {public int type() {return 1098;} public void run() {System.out.println(1098);}}
    public static class Impl1099 implements DemoService {public int type() {return 1099;} public void run() {System.out.println(1099);}}
    public static class Impl1100 implements DemoService {public int type() {return 1100;} public void run() {System.out.println(1100);}}
    public static class Impl1101 implements DemoService {public int type() {return 1101;} public void run() {System.out.println(1101);}}
    public static class Impl1102 implements DemoService {public int type() {return 1102;} public void run() {System.out.println(1102);}}
    public static class Impl1103 implements DemoService {public int type() {return 1103;} public void run() {System.out.println(1103);}}
    public static class Impl1104 implements DemoService {public int type() {return 1104;} public void run() {System.out.println(1104);}}
    public static class Impl1105 implements DemoService {public int type() {return 1105;} public void run() {System.out.println(1105);}}
    public static class Impl1106 implements DemoService {public int type() {return 1106;} public void run() {System.out.println(1106);}}
    public static class Impl1107 implements DemoService {public int type() {return 1107;} public void run() {System.out.println(1107);}}
    public static class Impl1108 implements DemoService {public int type() {return 1108;} public void run() {System.out.println(1108);}}
    public static class Impl1109 implements DemoService {public int type() {return 1109;} public void run() {System.out.println(1109);}}
    public static class Impl1110 implements DemoService {public int type() {return 1110;} public void run() {System.out.println(1110);}}
    public static class Impl1111 implements DemoService {public int type() {return 1111;} public void run() {System.out.println(1111);}}
    public static class Impl1112 implements DemoService {public int type() {return 1112;} public void run() {System.out.println(1112);}}
    public static class Impl1113 implements DemoService {public int type() {return 1113;} public void run() {System.out.println(1113);}}
    public static class Impl1114 implements DemoService {public int type() {return 1114;} public void run() {System.out.println(1114);}}
    public static class Impl1115 implements DemoService {public int type() {return 1115;} public void run() {System.out.println(1115);}}
    public static class Impl1116 implements DemoService {public int type() {return 1116;} public void run() {System.out.println(1116);}}
    public static class Impl1117 implements DemoService {public int type() {return 1117;} public void run() {System.out.println(1117);}}
    public static class Impl1118 implements DemoService {public int type() {return 1118;} public void run() {System.out.println(1118);}}
    public static class Impl1119 implements DemoService {public int type() {return 1119;} public void run() {System.out.println(1119);}}
    public static class Impl1120 implements DemoService {public int type() {return 1120;} public void run() {System.out.println(1120);}}
    public static class Impl1121 implements DemoService {public int type() {return 1121;} public void run() {System.out.println(1121);}}
    public static class Impl1122 implements DemoService {public int type() {return 1122;} public void run() {System.out.println(1122);}}
    public static class Impl1123 implements DemoService {public int type() {return 1123;} public void run() {System.out.println(1123);}}
    public static class Impl1124 implements DemoService {public int type() {return 1124;} public void run() {System.out.println(1124);}}
    public static class Impl1125 implements DemoService {public int type() {return 1125;} public void run() {System.out.println(1125);}}
    public static class Impl1126 implements DemoService {public int type() {return 1126;} public void run() {System.out.println(1126);}}
    public static class Impl1127 implements DemoService {public int type() {return 1127;} public void run() {System.out.println(1127);}}
    public static class Impl1128 implements DemoService {public int type() {return 1128;} public void run() {System.out.println(1128);}}
    public static class Impl1129 implements DemoService {public int type() {return 1129;} public void run() {System.out.println(1129);}}
    public static class Impl1130 implements DemoService {public int type() {return 1130;} public void run() {System.out.println(1130);}}
    public static class Impl1131 implements DemoService {public int type() {return 1131;} public void run() {System.out.println(1131);}}
    public static class Impl1132 implements DemoService {public int type() {return 1132;} public void run() {System.out.println(1132);}}
    public static class Impl1133 implements DemoService {public int type() {return 1133;} public void run() {System.out.println(1133);}}
    public static class Impl1134 implements DemoService {public int type() {return 1134;} public void run() {System.out.println(1134);}}
    public static class Impl1135 implements DemoService {public int type() {return 1135;} public void run() {System.out.println(1135);}}
    public static class Impl1136 implements DemoService {public int type() {return 1136;} public void run() {System.out.println(1136);}}
    public static class Impl1137 implements DemoService {public int type() {return 1137;} public void run() {System.out.println(1137);}}
    public static class Impl1138 implements DemoService {public int type() {return 1138;} public void run() {System.out.println(1138);}}
    public static class Impl1139 implements DemoService {public int type() {return 1139;} public void run() {System.out.println(1139);}}
    public static class Impl1140 implements DemoService {public int type() {return 1140;} public void run() {System.out.println(1140);}}
    public static class Impl1141 implements DemoService {public int type() {return 1141;} public void run() {System.out.println(1141);}}
    public static class Impl1142 implements DemoService {public int type() {return 1142;} public void run() {System.out.println(1142);}}
    public static class Impl1143 implements DemoService {public int type() {return 1143;} public void run() {System.out.println(1143);}}
    public static class Impl1144 implements DemoService {public int type() {return 1144;} public void run() {System.out.println(1144);}}
    public static class Impl1145 implements DemoService {public int type() {return 1145;} public void run() {System.out.println(1145);}}
    public static class Impl1146 implements DemoService {public int type() {return 1146;} public void run() {System.out.println(1146);}}
    public static class Impl1147 implements DemoService {public int type() {return 1147;} public void run() {System.out.println(1147);}}
    public static class Impl1148 implements DemoService {public int type() {return 1148;} public void run() {System.out.println(1148);}}
    public static class Impl1149 implements DemoService {public int type() {return 1149;} public void run() {System.out.println(1149);}}
    public static class Impl1150 implements DemoService {public int type() {return 1150;} public void run() {System.out.println(1150);}}
    public static class Impl1151 implements DemoService {public int type() {return 1151;} public void run() {System.out.println(1151);}}
    public static class Impl1152 implements DemoService {public int type() {return 1152;} public void run() {System.out.println(1152);}}
    public static class Impl1153 implements DemoService {public int type() {return 1153;} public void run() {System.out.println(1153);}}
    public static class Impl1154 implements DemoService {public int type() {return 1154;} public void run() {System.out.println(1154);}}
    public static class Impl1155 implements DemoService {public int type() {return 1155;} public void run() {System.out.println(1155);}}
    public static class Impl1156 implements DemoService {public int type() {return 1156;} public void run() {System.out.println(1156);}}
    public static class Impl1157 implements DemoService {public int type() {return 1157;} public void run() {System.out.println(1157);}}
    public static class Impl1158 implements DemoService {public int type() {return 1158;} public void run() {System.out.println(1158);}}
    public static class Impl1159 implements DemoService {public int type() {return 1159;} public void run() {System.out.println(1159);}}
    public static class Impl1160 implements DemoService {public int type() {return 1160;} public void run() {System.out.println(1160);}}
    public static class Impl1161 implements DemoService {public int type() {return 1161;} public void run() {System.out.println(1161);}}
    public static class Impl1162 implements DemoService {public int type() {return 1162;} public void run() {System.out.println(1162);}}
    public static class Impl1163 implements DemoService {public int type() {return 1163;} public void run() {System.out.println(1163);}}
    public static class Impl1164 implements DemoService {public int type() {return 1164;} public void run() {System.out.println(1164);}}
    public static class Impl1165 implements DemoService {public int type() {return 1165;} public void run() {System.out.println(1165);}}
    public static class Impl1166 implements DemoService {public int type() {return 1166;} public void run() {System.out.println(1166);}}
    public static class Impl1167 implements DemoService {public int type() {return 1167;} public void run() {System.out.println(1167);}}
    public static class Impl1168 implements DemoService {public int type() {return 1168;} public void run() {System.out.println(1168);}}
    public static class Impl1169 implements DemoService {public int type() {return 1169;} public void run() {System.out.println(1169);}}
    public static class Impl1170 implements DemoService {public int type() {return 1170;} public void run() {System.out.println(1170);}}
    public static class Impl1171 implements DemoService {public int type() {return 1171;} public void run() {System.out.println(1171);}}
    public static class Impl1172 implements DemoService {public int type() {return 1172;} public void run() {System.out.println(1172);}}
    public static class Impl1173 implements DemoService {public int type() {return 1173;} public void run() {System.out.println(1173);}}
    public static class Impl1174 implements DemoService {public int type() {return 1174;} public void run() {System.out.println(1174);}}
    public static class Impl1175 implements DemoService {public int type() {return 1175;} public void run() {System.out.println(1175);}}
    public static class Impl1176 implements DemoService {public int type() {return 1176;} public void run() {System.out.println(1176);}}
    public static class Impl1177 implements DemoService {public int type() {return 1177;} public void run() {System.out.println(1177);}}
    public static class Impl1178 implements DemoService {public int type() {return 1178;} public void run() {System.out.println(1178);}}
    public static class Impl1179 implements DemoService {public int type() {return 1179;} public void run() {System.out.println(1179);}}
    public static class Impl1180 implements DemoService {public int type() {return 1180;} public void run() {System.out.println(1180);}}
    public static class Impl1181 implements DemoService {public int type() {return 1181;} public void run() {System.out.println(1181);}}
    public static class Impl1182 implements DemoService {public int type() {return 1182;} public void run() {System.out.println(1182);}}
    public static class Impl1183 implements DemoService {public int type() {return 1183;} public void run() {System.out.println(1183);}}
    public static class Impl1184 implements DemoService {public int type() {return 1184;} public void run() {System.out.println(1184);}}
    public static class Impl1185 implements DemoService {public int type() {return 1185;} public void run() {System.out.println(1185);}}
    public static class Impl1186 implements DemoService {public int type() {return 1186;} public void run() {System.out.println(1186);}}
    public static class Impl1187 implements DemoService {public int type() {return 1187;} public void run() {System.out.println(1187);}}
    public static class Impl1188 implements DemoService {public int type() {return 1188;} public void run() {System.out.println(1188);}}
    public static class Impl1189 implements DemoService {public int type() {return 1189;} public void run() {System.out.println(1189);}}
    public static class Impl1190 implements DemoService {public int type() {return 1190;} public void run() {System.out.println(1190);}}
    public static class Impl1191 implements DemoService {public int type() {return 1191;} public void run() {System.out.println(1191);}}
    public static class Impl1192 implements DemoService {public int type() {return 1192;} public void run() {System.out.println(1192);}}
    public static class Impl1193 implements DemoService {public int type() {return 1193;} public void run() {System.out.println(1193);}}
    public static class Impl1194 implements DemoService {public int type() {return 1194;} public void run() {System.out.println(1194);}}
    public static class Impl1195 implements DemoService {public int type() {return 1195;} public void run() {System.out.println(1195);}}
    public static class Impl1196 implements DemoService {public int type() {return 1196;} public void run() {System.out.println(1196);}}
    public static class Impl1197 implements DemoService {public int type() {return 1197;} public void run() {System.out.println(1197);}}
    public static class Impl1198 implements DemoService {public int type() {return 1198;} public void run() {System.out.println(1198);}}
    public static class Impl1199 implements DemoService {public int type() {return 1199;} public void run() {System.out.println(1199);}}
    public static class Impl1200 implements DemoService {public int type() {return 1200;} public void run() {System.out.println(1200);}}
    public static class Impl1201 implements DemoService {public int type() {return 1201;} public void run() {System.out.println(1201);}}
    public static class Impl1202 implements DemoService {public int type() {return 1202;} public void run() {System.out.println(1202);}}
    public static class Impl1203 implements DemoService {public int type() {return 1203;} public void run() {System.out.println(1203);}}
    public static class Impl1204 implements DemoService {public int type() {return 1204;} public void run() {System.out.println(1204);}}
    public static class Impl1205 implements DemoService {public int type() {return 1205;} public void run() {System.out.println(1205);}}
    public static class Impl1206 implements DemoService {public int type() {return 1206;} public void run() {System.out.println(1206);}}
    public static class Impl1207 implements DemoService {public int type() {return 1207;} public void run() {System.out.println(1207);}}
    public static class Impl1208 implements DemoService {public int type() {return 1208;} public void run() {System.out.println(1208);}}
    public static class Impl1209 implements DemoService {public int type() {return 1209;} public void run() {System.out.println(1209);}}
    public static class Impl1210 implements DemoService {public int type() {return 1210;} public void run() {System.out.println(1210);}}
    public static class Impl1211 implements DemoService {public int type() {return 1211;} public void run() {System.out.println(1211);}}
    public static class Impl1212 implements DemoService {public int type() {return 1212;} public void run() {System.out.println(1212);}}
    public static class Impl1213 implements DemoService {public int type() {return 1213;} public void run() {System.out.println(1213);}}
    public static class Impl1214 implements DemoService {public int type() {return 1214;} public void run() {System.out.println(1214);}}
    public static class Impl1215 implements DemoService {public int type() {return 1215;} public void run() {System.out.println(1215);}}
    public static class Impl1216 implements DemoService {public int type() {return 1216;} public void run() {System.out.println(1216);}}
    public static class Impl1217 implements DemoService {public int type() {return 1217;} public void run() {System.out.println(1217);}}
    public static class Impl1218 implements DemoService {public int type() {return 1218;} public void run() {System.out.println(1218);}}
    public static class Impl1219 implements DemoService {public int type() {return 1219;} public void run() {System.out.println(1219);}}
    public static class Impl1220 implements DemoService {public int type() {return 1220;} public void run() {System.out.println(1220);}}
    public static class Impl1221 implements DemoService {public int type() {return 1221;} public void run() {System.out.println(1221);}}
    public static class Impl1222 implements DemoService {public int type() {return 1222;} public void run() {System.out.println(1222);}}
    public static class Impl1223 implements DemoService {public int type() {return 1223;} public void run() {System.out.println(1223);}}
    public static class Impl1224 implements DemoService {public int type() {return 1224;} public void run() {System.out.println(1224);}}
    public static class Impl1225 implements DemoService {public int type() {return 1225;} public void run() {System.out.println(1225);}}
    public static class Impl1226 implements DemoService {public int type() {return 1226;} public void run() {System.out.println(1226);}}
    public static class Impl1227 implements DemoService {public int type() {return 1227;} public void run() {System.out.println(1227);}}
    public static class Impl1228 implements DemoService {public int type() {return 1228;} public void run() {System.out.println(1228);}}
    public static class Impl1229 implements DemoService {public int type() {return 1229;} public void run() {System.out.println(1229);}}
    public static class Impl1230 implements DemoService {public int type() {return 1230;} public void run() {System.out.println(1230);}}
    public static class Impl1231 implements DemoService {public int type() {return 1231;} public void run() {System.out.println(1231);}}
    public static class Impl1232 implements DemoService {public int type() {return 1232;} public void run() {System.out.println(1232);}}
    public static class Impl1233 implements DemoService {public int type() {return 1233;} public void run() {System.out.println(1233);}}
    public static class Impl1234 implements DemoService {public int type() {return 1234;} public void run() {System.out.println(1234);}}
    public static class Impl1235 implements DemoService {public int type() {return 1235;} public void run() {System.out.println(1235);}}
    public static class Impl1236 implements DemoService {public int type() {return 1236;} public void run() {System.out.println(1236);}}
    public static class Impl1237 implements DemoService {public int type() {return 1237;} public void run() {System.out.println(1237);}}
    public static class Impl1238 implements DemoService {public int type() {return 1238;} public void run() {System.out.println(1238);}}
    public static class Impl1239 implements DemoService {public int type() {return 1239;} public void run() {System.out.println(1239);}}
    public static class Impl1240 implements DemoService {public int type() {return 1240;} public void run() {System.out.println(1240);}}
    public static class Impl1241 implements DemoService {public int type() {return 1241;} public void run() {System.out.println(1241);}}
    public static class Impl1242 implements DemoService {public int type() {return 1242;} public void run() {System.out.println(1242);}}
    public static class Impl1243 implements DemoService {public int type() {return 1243;} public void run() {System.out.println(1243);}}
    public static class Impl1244 implements DemoService {public int type() {return 1244;} public void run() {System.out.println(1244);}}
    public static class Impl1245 implements DemoService {public int type() {return 1245;} public void run() {System.out.println(1245);}}
    public static class Impl1246 implements DemoService {public int type() {return 1246;} public void run() {System.out.println(1246);}}
    public static class Impl1247 implements DemoService {public int type() {return 1247;} public void run() {System.out.println(1247);}}
    public static class Impl1248 implements DemoService {public int type() {return 1248;} public void run() {System.out.println(1248);}}
    public static class Impl1249 implements DemoService {public int type() {return 1249;} public void run() {System.out.println(1249);}}
    public static class Impl1250 implements DemoService {public int type() {return 1250;} public void run() {System.out.println(1250);}}
    public static class Impl1251 implements DemoService {public int type() {return 1251;} public void run() {System.out.println(1251);}}
    public static class Impl1252 implements DemoService {public int type() {return 1252;} public void run() {System.out.println(1252);}}
    public static class Impl1253 implements DemoService {public int type() {return 1253;} public void run() {System.out.println(1253);}}
    public static class Impl1254 implements DemoService {public int type() {return 1254;} public void run() {System.out.println(1254);}}
    public static class Impl1255 implements DemoService {public int type() {return 1255;} public void run() {System.out.println(1255);}}
    public static class Impl1256 implements DemoService {public int type() {return 1256;} public void run() {System.out.println(1256);}}
    public static class Impl1257 implements DemoService {public int type() {return 1257;} public void run() {System.out.println(1257);}}
    public static class Impl1258 implements DemoService {public int type() {return 1258;} public void run() {System.out.println(1258);}}
    public static class Impl1259 implements DemoService {public int type() {return 1259;} public void run() {System.out.println(1259);}}
    public static class Impl1260 implements DemoService {public int type() {return 1260;} public void run() {System.out.println(1260);}}
    public static class Impl1261 implements DemoService {public int type() {return 1261;} public void run() {System.out.println(1261);}}
    public static class Impl1262 implements DemoService {public int type() {return 1262;} public void run() {System.out.println(1262);}}
    public static class Impl1263 implements DemoService {public int type() {return 1263;} public void run() {System.out.println(1263);}}
    public static class Impl1264 implements DemoService {public int type() {return 1264;} public void run() {System.out.println(1264);}}
    public static class Impl1265 implements DemoService {public int type() {return 1265;} public void run() {System.out.println(1265);}}
    public static class Impl1266 implements DemoService {public int type() {return 1266;} public void run() {System.out.println(1266);}}
    public static class Impl1267 implements DemoService {public int type() {return 1267;} public void run() {System.out.println(1267);}}
    public static class Impl1268 implements DemoService {public int type() {return 1268;} public void run() {System.out.println(1268);}}
    public static class Impl1269 implements DemoService {public int type() {return 1269;} public void run() {System.out.println(1269);}}
    public static class Impl1270 implements DemoService {public int type() {return 1270;} public void run() {System.out.println(1270);}}
    public static class Impl1271 implements DemoService {public int type() {return 1271;} public void run() {System.out.println(1271);}}
    public static class Impl1272 implements DemoService {public int type() {return 1272;} public void run() {System.out.println(1272);}}
    public static class Impl1273 implements DemoService {public int type() {return 1273;} public void run() {System.out.println(1273);}}
    public static class Impl1274 implements DemoService {public int type() {return 1274;} public void run() {System.out.println(1274);}}
    public static class Impl1275 implements DemoService {public int type() {return 1275;} public void run() {System.out.println(1275);}}
    public static class Impl1276 implements DemoService {public int type() {return 1276;} public void run() {System.out.println(1276);}}
    public static class Impl1277 implements DemoService {public int type() {return 1277;} public void run() {System.out.println(1277);}}
    public static class Impl1278 implements DemoService {public int type() {return 1278;} public void run() {System.out.println(1278);}}
    public static class Impl1279 implements DemoService {public int type() {return 1279;} public void run() {System.out.println(1279);}}
    public static class Impl1280 implements DemoService {public int type() {return 1280;} public void run() {System.out.println(1280);}}
    public static class Impl1281 implements DemoService {public int type() {return 1281;} public void run() {System.out.println(1281);}}
    public static class Impl1282 implements DemoService {public int type() {return 1282;} public void run() {System.out.println(1282);}}
    public static class Impl1283 implements DemoService {public int type() {return 1283;} public void run() {System.out.println(1283);}}
    public static class Impl1284 implements DemoService {public int type() {return 1284;} public void run() {System.out.println(1284);}}
    public static class Impl1285 implements DemoService {public int type() {return 1285;} public void run() {System.out.println(1285);}}
    public static class Impl1286 implements DemoService {public int type() {return 1286;} public void run() {System.out.println(1286);}}
    public static class Impl1287 implements DemoService {public int type() {return 1287;} public void run() {System.out.println(1287);}}
    public static class Impl1288 implements DemoService {public int type() {return 1288;} public void run() {System.out.println(1288);}}
    public static class Impl1289 implements DemoService {public int type() {return 1289;} public void run() {System.out.println(1289);}}
    public static class Impl1290 implements DemoService {public int type() {return 1290;} public void run() {System.out.println(1290);}}
    public static class Impl1291 implements DemoService {public int type() {return 1291;} public void run() {System.out.println(1291);}}
    public static class Impl1292 implements DemoService {public int type() {return 1292;} public void run() {System.out.println(1292);}}
    public static class Impl1293 implements DemoService {public int type() {return 1293;} public void run() {System.out.println(1293);}}
    public static class Impl1294 implements DemoService {public int type() {return 1294;} public void run() {System.out.println(1294);}}
    public static class Impl1295 implements DemoService {public int type() {return 1295;} public void run() {System.out.println(1295);}}
    public static class Impl1296 implements DemoService {public int type() {return 1296;} public void run() {System.out.println(1296);}}
    public static class Impl1297 implements DemoService {public int type() {return 1297;} public void run() {System.out.println(1297);}}
    public static class Impl1298 implements DemoService {public int type() {return 1298;} public void run() {System.out.println(1298);}}
    public static class Impl1299 implements DemoService {public int type() {return 1299;} public void run() {System.out.println(1299);}}
    public static class Impl1300 implements DemoService {public int type() {return 1300;} public void run() {System.out.println(1300);}}
    public static class Impl1301 implements DemoService {public int type() {return 1301;} public void run() {System.out.println(1301);}}
    public static class Impl1302 implements DemoService {public int type() {return 1302;} public void run() {System.out.println(1302);}}
    public static class Impl1303 implements DemoService {public int type() {return 1303;} public void run() {System.out.println(1303);}}
    public static class Impl1304 implements DemoService {public int type() {return 1304;} public void run() {System.out.println(1304);}}
    public static class Impl1305 implements DemoService {public int type() {return 1305;} public void run() {System.out.println(1305);}}
    public static class Impl1306 implements DemoService {public int type() {return 1306;} public void run() {System.out.println(1306);}}
    public static class Impl1307 implements DemoService {public int type() {return 1307;} public void run() {System.out.println(1307);}}
    public static class Impl1308 implements DemoService {public int type() {return 1308;} public void run() {System.out.println(1308);}}
    public static class Impl1309 implements DemoService {public int type() {return 1309;} public void run() {System.out.println(1309);}}
    public static class Impl1310 implements DemoService {public int type() {return 1310;} public void run() {System.out.println(1310);}}
    public static class Impl1311 implements DemoService {public int type() {return 1311;} public void run() {System.out.println(1311);}}
    public static class Impl1312 implements DemoService {public int type() {return 1312;} public void run() {System.out.println(1312);}}
    public static class Impl1313 implements DemoService {public int type() {return 1313;} public void run() {System.out.println(1313);}}
    public static class Impl1314 implements DemoService {public int type() {return 1314;} public void run() {System.out.println(1314);}}
    public static class Impl1315 implements DemoService {public int type() {return 1315;} public void run() {System.out.println(1315);}}
    public static class Impl1316 implements DemoService {public int type() {return 1316;} public void run() {System.out.println(1316);}}
    public static class Impl1317 implements DemoService {public int type() {return 1317;} public void run() {System.out.println(1317);}}
    public static class Impl1318 implements DemoService {public int type() {return 1318;} public void run() {System.out.println(1318);}}
    public static class Impl1319 implements DemoService {public int type() {return 1319;} public void run() {System.out.println(1319);}}
    public static class Impl1320 implements DemoService {public int type() {return 1320;} public void run() {System.out.println(1320);}}
    public static class Impl1321 implements DemoService {public int type() {return 1321;} public void run() {System.out.println(1321);}}
    public static class Impl1322 implements DemoService {public int type() {return 1322;} public void run() {System.out.println(1322);}}
    public static class Impl1323 implements DemoService {public int type() {return 1323;} public void run() {System.out.println(1323);}}
    public static class Impl1324 implements DemoService {public int type() {return 1324;} public void run() {System.out.println(1324);}}
    public static class Impl1325 implements DemoService {public int type() {return 1325;} public void run() {System.out.println(1325);}}
    public static class Impl1326 implements DemoService {public int type() {return 1326;} public void run() {System.out.println(1326);}}
    public static class Impl1327 implements DemoService {public int type() {return 1327;} public void run() {System.out.println(1327);}}
    public static class Impl1328 implements DemoService {public int type() {return 1328;} public void run() {System.out.println(1328);}}
    public static class Impl1329 implements DemoService {public int type() {return 1329;} public void run() {System.out.println(1329);}}
    public static class Impl1330 implements DemoService {public int type() {return 1330;} public void run() {System.out.println(1330);}}
    public static class Impl1331 implements DemoService {public int type() {return 1331;} public void run() {System.out.println(1331);}}
    public static class Impl1332 implements DemoService {public int type() {return 1332;} public void run() {System.out.println(1332);}}
    public static class Impl1333 implements DemoService {public int type() {return 1333;} public void run() {System.out.println(1333);}}
    public static class Impl1334 implements DemoService {public int type() {return 1334;} public void run() {System.out.println(1334);}}
    public static class Impl1335 implements DemoService {public int type() {return 1335;} public void run() {System.out.println(1335);}}
    public static class Impl1336 implements DemoService {public int type() {return 1336;} public void run() {System.out.println(1336);}}
    public static class Impl1337 implements DemoService {public int type() {return 1337;} public void run() {System.out.println(1337);}}
    public static class Impl1338 implements DemoService {public int type() {return 1338;} public void run() {System.out.println(1338);}}
    public static class Impl1339 implements DemoService {public int type() {return 1339;} public void run() {System.out.println(1339);}}
    public static class Impl1340 implements DemoService {public int type() {return 1340;} public void run() {System.out.println(1340);}}
    public static class Impl1341 implements DemoService {public int type() {return 1341;} public void run() {System.out.println(1341);}}
    public static class Impl1342 implements DemoService {public int type() {return 1342;} public void run() {System.out.println(1342);}}
    public static class Impl1343 implements DemoService {public int type() {return 1343;} public void run() {System.out.println(1343);}}
    public static class Impl1344 implements DemoService {public int type() {return 1344;} public void run() {System.out.println(1344);}}
    public static class Impl1345 implements DemoService {public int type() {return 1345;} public void run() {System.out.println(1345);}}
    public static class Impl1346 implements DemoService {public int type() {return 1346;} public void run() {System.out.println(1346);}}
    public static class Impl1347 implements DemoService {public int type() {return 1347;} public void run() {System.out.println(1347);}}
    public static class Impl1348 implements DemoService {public int type() {return 1348;} public void run() {System.out.println(1348);}}
    public static class Impl1349 implements DemoService {public int type() {return 1349;} public void run() {System.out.println(1349);}}
    public static class Impl1350 implements DemoService {public int type() {return 1350;} public void run() {System.out.println(1350);}}
    public static class Impl1351 implements DemoService {public int type() {return 1351;} public void run() {System.out.println(1351);}}
    public static class Impl1352 implements DemoService {public int type() {return 1352;} public void run() {System.out.println(1352);}}
    public static class Impl1353 implements DemoService {public int type() {return 1353;} public void run() {System.out.println(1353);}}
    public static class Impl1354 implements DemoService {public int type() {return 1354;} public void run() {System.out.println(1354);}}
    public static class Impl1355 implements DemoService {public int type() {return 1355;} public void run() {System.out.println(1355);}}
    public static class Impl1356 implements DemoService {public int type() {return 1356;} public void run() {System.out.println(1356);}}
    public static class Impl1357 implements DemoService {public int type() {return 1357;} public void run() {System.out.println(1357);}}
    public static class Impl1358 implements DemoService {public int type() {return 1358;} public void run() {System.out.println(1358);}}
    public static class Impl1359 implements DemoService {public int type() {return 1359;} public void run() {System.out.println(1359);}}
    public static class Impl1360 implements DemoService {public int type() {return 1360;} public void run() {System.out.println(1360);}}
    public static class Impl1361 implements DemoService {public int type() {return 1361;} public void run() {System.out.println(1361);}}
    public static class Impl1362 implements DemoService {public int type() {return 1362;} public void run() {System.out.println(1362);}}
    public static class Impl1363 implements DemoService {public int type() {return 1363;} public void run() {System.out.println(1363);}}
    public static class Impl1364 implements DemoService {public int type() {return 1364;} public void run() {System.out.println(1364);}}
    public static class Impl1365 implements DemoService {public int type() {return 1365;} public void run() {System.out.println(1365);}}
    public static class Impl1366 implements DemoService {public int type() {return 1366;} public void run() {System.out.println(1366);}}
    public static class Impl1367 implements DemoService {public int type() {return 1367;} public void run() {System.out.println(1367);}}
    public static class Impl1368 implements DemoService {public int type() {return 1368;} public void run() {System.out.println(1368);}}
    public static class Impl1369 implements DemoService {public int type() {return 1369;} public void run() {System.out.println(1369);}}
    public static class Impl1370 implements DemoService {public int type() {return 1370;} public void run() {System.out.println(1370);}}
    public static class Impl1371 implements DemoService {public int type() {return 1371;} public void run() {System.out.println(1371);}}
    public static class Impl1372 implements DemoService {public int type() {return 1372;} public void run() {System.out.println(1372);}}
    public static class Impl1373 implements DemoService {public int type() {return 1373;} public void run() {System.out.println(1373);}}
    public static class Impl1374 implements DemoService {public int type() {return 1374;} public void run() {System.out.println(1374);}}
    public static class Impl1375 implements DemoService {public int type() {return 1375;} public void run() {System.out.println(1375);}}
    public static class Impl1376 implements DemoService {public int type() {return 1376;} public void run() {System.out.println(1376);}}
    public static class Impl1377 implements DemoService {public int type() {return 1377;} public void run() {System.out.println(1377);}}
    public static class Impl1378 implements DemoService {public int type() {return 1378;} public void run() {System.out.println(1378);}}
    public static class Impl1379 implements DemoService {public int type() {return 1379;} public void run() {System.out.println(1379);}}
    public static class Impl1380 implements DemoService {public int type() {return 1380;} public void run() {System.out.println(1380);}}
    public static class Impl1381 implements DemoService {public int type() {return 1381;} public void run() {System.out.println(1381);}}
    public static class Impl1382 implements DemoService {public int type() {return 1382;} public void run() {System.out.println(1382);}}
    public static class Impl1383 implements DemoService {public int type() {return 1383;} public void run() {System.out.println(1383);}}
    public static class Impl1384 implements DemoService {public int type() {return 1384;} public void run() {System.out.println(1384);}}
    public static class Impl1385 implements DemoService {public int type() {return 1385;} public void run() {System.out.println(1385);}}
    public static class Impl1386 implements DemoService {public int type() {return 1386;} public void run() {System.out.println(1386);}}
    public static class Impl1387 implements DemoService {public int type() {return 1387;} public void run() {System.out.println(1387);}}
    public static class Impl1388 implements DemoService {public int type() {return 1388;} public void run() {System.out.println(1388);}}
    public static class Impl1389 implements DemoService {public int type() {return 1389;} public void run() {System.out.println(1389);}}
    public static class Impl1390 implements DemoService {public int type() {return 1390;} public void run() {System.out.println(1390);}}
    public static class Impl1391 implements DemoService {public int type() {return 1391;} public void run() {System.out.println(1391);}}
    public static class Impl1392 implements DemoService {public int type() {return 1392;} public void run() {System.out.println(1392);}}
    public static class Impl1393 implements DemoService {public int type() {return 1393;} public void run() {System.out.println(1393);}}
    public static class Impl1394 implements DemoService {public int type() {return 1394;} public void run() {System.out.println(1394);}}
    public static class Impl1395 implements DemoService {public int type() {return 1395;} public void run() {System.out.println(1395);}}
    public static class Impl1396 implements DemoService {public int type() {return 1396;} public void run() {System.out.println(1396);}}
    public static class Impl1397 implements DemoService {public int type() {return 1397;} public void run() {System.out.println(1397);}}
    public static class Impl1398 implements DemoService {public int type() {return 1398;} public void run() {System.out.println(1398);}}
    public static class Impl1399 implements DemoService {public int type() {return 1399;} public void run() {System.out.println(1399);}}
    public static class Impl1400 implements DemoService {public int type() {return 1400;} public void run() {System.out.println(1400);}}
    public static class Impl1401 implements DemoService {public int type() {return 1401;} public void run() {System.out.println(1401);}}
    public static class Impl1402 implements DemoService {public int type() {return 1402;} public void run() {System.out.println(1402);}}
    public static class Impl1403 implements DemoService {public int type() {return 1403;} public void run() {System.out.println(1403);}}
    public static class Impl1404 implements DemoService {public int type() {return 1404;} public void run() {System.out.println(1404);}}
    public static class Impl1405 implements DemoService {public int type() {return 1405;} public void run() {System.out.println(1405);}}
    public static class Impl1406 implements DemoService {public int type() {return 1406;} public void run() {System.out.println(1406);}}
    public static class Impl1407 implements DemoService {public int type() {return 1407;} public void run() {System.out.println(1407);}}
    public static class Impl1408 implements DemoService {public int type() {return 1408;} public void run() {System.out.println(1408);}}
    public static class Impl1409 implements DemoService {public int type() {return 1409;} public void run() {System.out.println(1409);}}
    public static class Impl1410 implements DemoService {public int type() {return 1410;} public void run() {System.out.println(1410);}}
    public static class Impl1411 implements DemoService {public int type() {return 1411;} public void run() {System.out.println(1411);}}
    public static class Impl1412 implements DemoService {public int type() {return 1412;} public void run() {System.out.println(1412);}}
    public static class Impl1413 implements DemoService {public int type() {return 1413;} public void run() {System.out.println(1413);}}
    public static class Impl1414 implements DemoService {public int type() {return 1414;} public void run() {System.out.println(1414);}}
    public static class Impl1415 implements DemoService {public int type() {return 1415;} public void run() {System.out.println(1415);}}
    public static class Impl1416 implements DemoService {public int type() {return 1416;} public void run() {System.out.println(1416);}}
    public static class Impl1417 implements DemoService {public int type() {return 1417;} public void run() {System.out.println(1417);}}
    public static class Impl1418 implements DemoService {public int type() {return 1418;} public void run() {System.out.println(1418);}}
    public static class Impl1419 implements DemoService {public int type() {return 1419;} public void run() {System.out.println(1419);}}
    public static class Impl1420 implements DemoService {public int type() {return 1420;} public void run() {System.out.println(1420);}}
    public static class Impl1421 implements DemoService {public int type() {return 1421;} public void run() {System.out.println(1421);}}
    public static class Impl1422 implements DemoService {public int type() {return 1422;} public void run() {System.out.println(1422);}}
    public static class Impl1423 implements DemoService {public int type() {return 1423;} public void run() {System.out.println(1423);}}
    public static class Impl1424 implements DemoService {public int type() {return 1424;} public void run() {System.out.println(1424);}}
    public static class Impl1425 implements DemoService {public int type() {return 1425;} public void run() {System.out.println(1425);}}
    public static class Impl1426 implements DemoService {public int type() {return 1426;} public void run() {System.out.println(1426);}}
    public static class Impl1427 implements DemoService {public int type() {return 1427;} public void run() {System.out.println(1427);}}
    public static class Impl1428 implements DemoService {public int type() {return 1428;} public void run() {System.out.println(1428);}}
    public static class Impl1429 implements DemoService {public int type() {return 1429;} public void run() {System.out.println(1429);}}
    public static class Impl1430 implements DemoService {public int type() {return 1430;} public void run() {System.out.println(1430);}}
    public static class Impl1431 implements DemoService {public int type() {return 1431;} public void run() {System.out.println(1431);}}
    public static class Impl1432 implements DemoService {public int type() {return 1432;} public void run() {System.out.println(1432);}}
    public static class Impl1433 implements DemoService {public int type() {return 1433;} public void run() {System.out.println(1433);}}
    public static class Impl1434 implements DemoService {public int type() {return 1434;} public void run() {System.out.println(1434);}}
    public static class Impl1435 implements DemoService {public int type() {return 1435;} public void run() {System.out.println(1435);}}
    public static class Impl1436 implements DemoService {public int type() {return 1436;} public void run() {System.out.println(1436);}}
    public static class Impl1437 implements DemoService {public int type() {return 1437;} public void run() {System.out.println(1437);}}
    public static class Impl1438 implements DemoService {public int type() {return 1438;} public void run() {System.out.println(1438);}}
    public static class Impl1439 implements DemoService {public int type() {return 1439;} public void run() {System.out.println(1439);}}
    public static class Impl1440 implements DemoService {public int type() {return 1440;} public void run() {System.out.println(1440);}}
    public static class Impl1441 implements DemoService {public int type() {return 1441;} public void run() {System.out.println(1441);}}
    public static class Impl1442 implements DemoService {public int type() {return 1442;} public void run() {System.out.println(1442);}}
    public static class Impl1443 implements DemoService {public int type() {return 1443;} public void run() {System.out.println(1443);}}
    public static class Impl1444 implements DemoService {public int type() {return 1444;} public void run() {System.out.println(1444);}}
    public static class Impl1445 implements DemoService {public int type() {return 1445;} public void run() {System.out.println(1445);}}
    public static class Impl1446 implements DemoService {public int type() {return 1446;} public void run() {System.out.println(1446);}}
    public static class Impl1447 implements DemoService {public int type() {return 1447;} public void run() {System.out.println(1447);}}
    public static class Impl1448 implements DemoService {public int type() {return 1448;} public void run() {System.out.println(1448);}}
    public static class Impl1449 implements DemoService {public int type() {return 1449;} public void run() {System.out.println(1449);}}
    public static class Impl1450 implements DemoService {public int type() {return 1450;} public void run() {System.out.println(1450);}}
    public static class Impl1451 implements DemoService {public int type() {return 1451;} public void run() {System.out.println(1451);}}
    public static class Impl1452 implements DemoService {public int type() {return 1452;} public void run() {System.out.println(1452);}}
    public static class Impl1453 implements DemoService {public int type() {return 1453;} public void run() {System.out.println(1453);}}
    public static class Impl1454 implements DemoService {public int type() {return 1454;} public void run() {System.out.println(1454);}}
    public static class Impl1455 implements DemoService {public int type() {return 1455;} public void run() {System.out.println(1455);}}
    public static class Impl1456 implements DemoService {public int type() {return 1456;} public void run() {System.out.println(1456);}}
    public static class Impl1457 implements DemoService {public int type() {return 1457;} public void run() {System.out.println(1457);}}
    public static class Impl1458 implements DemoService {public int type() {return 1458;} public void run() {System.out.println(1458);}}
    public static class Impl1459 implements DemoService {public int type() {return 1459;} public void run() {System.out.println(1459);}}
    public static class Impl1460 implements DemoService {public int type() {return 1460;} public void run() {System.out.println(1460);}}
    public static class Impl1461 implements DemoService {public int type() {return 1461;} public void run() {System.out.println(1461);}}
    public static class Impl1462 implements DemoService {public int type() {return 1462;} public void run() {System.out.println(1462);}}
    public static class Impl1463 implements DemoService {public int type() {return 1463;} public void run() {System.out.println(1463);}}
    public static class Impl1464 implements DemoService {public int type() {return 1464;} public void run() {System.out.println(1464);}}
    public static class Impl1465 implements DemoService {public int type() {return 1465;} public void run() {System.out.println(1465);}}
    public static class Impl1466 implements DemoService {public int type() {return 1466;} public void run() {System.out.println(1466);}}
    public static class Impl1467 implements DemoService {public int type() {return 1467;} public void run() {System.out.println(1467);}}
    public static class Impl1468 implements DemoService {public int type() {return 1468;} public void run() {System.out.println(1468);}}
    public static class Impl1469 implements DemoService {public int type() {return 1469;} public void run() {System.out.println(1469);}}
    public static class Impl1470 implements DemoService {public int type() {return 1470;} public void run() {System.out.println(1470);}}
    public static class Impl1471 implements DemoService {public int type() {return 1471;} public void run() {System.out.println(1471);}}
    public static class Impl1472 implements DemoService {public int type() {return 1472;} public void run() {System.out.println(1472);}}
    public static class Impl1473 implements DemoService {public int type() {return 1473;} public void run() {System.out.println(1473);}}
    public static class Impl1474 implements DemoService {public int type() {return 1474;} public void run() {System.out.println(1474);}}
    public static class Impl1475 implements DemoService {public int type() {return 1475;} public void run() {System.out.println(1475);}}
    public static class Impl1476 implements DemoService {public int type() {return 1476;} public void run() {System.out.println(1476);}}
    public static class Impl1477 implements DemoService {public int type() {return 1477;} public void run() {System.out.println(1477);}}
    public static class Impl1478 implements DemoService {public int type() {return 1478;} public void run() {System.out.println(1478);}}
    public static class Impl1479 implements DemoService {public int type() {return 1479;} public void run() {System.out.println(1479);}}
    public static class Impl1480 implements DemoService {public int type() {return 1480;} public void run() {System.out.println(1480);}}
    public static class Impl1481 implements DemoService {public int type() {return 1481;} public void run() {System.out.println(1481);}}
    public static class Impl1482 implements DemoService {public int type() {return 1482;} public void run() {System.out.println(1482);}}
    public static class Impl1483 implements DemoService {public int type() {return 1483;} public void run() {System.out.println(1483);}}
    public static class Impl1484 implements DemoService {public int type() {return 1484;} public void run() {System.out.println(1484);}}
    public static class Impl1485 implements DemoService {public int type() {return 1485;} public void run() {System.out.println(1485);}}
    public static class Impl1486 implements DemoService {public int type() {return 1486;} public void run() {System.out.println(1486);}}
    public static class Impl1487 implements DemoService {public int type() {return 1487;} public void run() {System.out.println(1487);}}
    public static class Impl1488 implements DemoService {public int type() {return 1488;} public void run() {System.out.println(1488);}}
    public static class Impl1489 implements DemoService {public int type() {return 1489;} public void run() {System.out.println(1489);}}
    public static class Impl1490 implements DemoService {public int type() {return 1490;} public void run() {System.out.println(1490);}}
    public static class Impl1491 implements DemoService {public int type() {return 1491;} public void run() {System.out.println(1491);}}
    public static class Impl1492 implements DemoService {public int type() {return 1492;} public void run() {System.out.println(1492);}}
    public static class Impl1493 implements DemoService {public int type() {return 1493;} public void run() {System.out.println(1493);}}
    public static class Impl1494 implements DemoService {public int type() {return 1494;} public void run() {System.out.println(1494);}}
    public static class Impl1495 implements DemoService {public int type() {return 1495;} public void run() {System.out.println(1495);}}
    public static class Impl1496 implements DemoService {public int type() {return 1496;} public void run() {System.out.println(1496);}}
    public static class Impl1497 implements DemoService {public int type() {return 1497;} public void run() {System.out.println(1497);}}
    public static class Impl1498 implements DemoService {public int type() {return 1498;} public void run() {System.out.println(1498);}}
    public static class Impl1499 implements DemoService {public int type() {return 1499;} public void run() {System.out.println(1499);}}
    public static class Impl1500 implements DemoService {public int type() {return 1500;} public void run() {System.out.println(1500);}}
    public static class Impl1501 implements DemoService {public int type() {return 1501;} public void run() {System.out.println(1501);}}
    public static class Impl1502 implements DemoService {public int type() {return 1502;} public void run() {System.out.println(1502);}}
    public static class Impl1503 implements DemoService {public int type() {return 1503;} public void run() {System.out.println(1503);}}
    public static class Impl1504 implements DemoService {public int type() {return 1504;} public void run() {System.out.println(1504);}}
    public static class Impl1505 implements DemoService {public int type() {return 1505;} public void run() {System.out.println(1505);}}
    public static class Impl1506 implements DemoService {public int type() {return 1506;} public void run() {System.out.println(1506);}}
    public static class Impl1507 implements DemoService {public int type() {return 1507;} public void run() {System.out.println(1507);}}
    public static class Impl1508 implements DemoService {public int type() {return 1508;} public void run() {System.out.println(1508);}}
    public static class Impl1509 implements DemoService {public int type() {return 1509;} public void run() {System.out.println(1509);}}
    public static class Impl1510 implements DemoService {public int type() {return 1510;} public void run() {System.out.println(1510);}}
    public static class Impl1511 implements DemoService {public int type() {return 1511;} public void run() {System.out.println(1511);}}
    public static class Impl1512 implements DemoService {public int type() {return 1512;} public void run() {System.out.println(1512);}}
    public static class Impl1513 implements DemoService {public int type() {return 1513;} public void run() {System.out.println(1513);}}
    public static class Impl1514 implements DemoService {public int type() {return 1514;} public void run() {System.out.println(1514);}}
    public static class Impl1515 implements DemoService {public int type() {return 1515;} public void run() {System.out.println(1515);}}
    public static class Impl1516 implements DemoService {public int type() {return 1516;} public void run() {System.out.println(1516);}}
    public static class Impl1517 implements DemoService {public int type() {return 1517;} public void run() {System.out.println(1517);}}
    public static class Impl1518 implements DemoService {public int type() {return 1518;} public void run() {System.out.println(1518);}}
    public static class Impl1519 implements DemoService {public int type() {return 1519;} public void run() {System.out.println(1519);}}
    public static class Impl1520 implements DemoService {public int type() {return 1520;} public void run() {System.out.println(1520);}}
    public static class Impl1521 implements DemoService {public int type() {return 1521;} public void run() {System.out.println(1521);}}
    public static class Impl1522 implements DemoService {public int type() {return 1522;} public void run() {System.out.println(1522);}}
    public static class Impl1523 implements DemoService {public int type() {return 1523;} public void run() {System.out.println(1523);}}
    public static class Impl1524 implements DemoService {public int type() {return 1524;} public void run() {System.out.println(1524);}}
    public static class Impl1525 implements DemoService {public int type() {return 1525;} public void run() {System.out.println(1525);}}
    public static class Impl1526 implements DemoService {public int type() {return 1526;} public void run() {System.out.println(1526);}}
    public static class Impl1527 implements DemoService {public int type() {return 1527;} public void run() {System.out.println(1527);}}
    public static class Impl1528 implements DemoService {public int type() {return 1528;} public void run() {System.out.println(1528);}}
    public static class Impl1529 implements DemoService {public int type() {return 1529;} public void run() {System.out.println(1529);}}
    public static class Impl1530 implements DemoService {public int type() {return 1530;} public void run() {System.out.println(1530);}}
    public static class Impl1531 implements DemoService {public int type() {return 1531;} public void run() {System.out.println(1531);}}
    public static class Impl1532 implements DemoService {public int type() {return 1532;} public void run() {System.out.println(1532);}}
    public static class Impl1533 implements DemoService {public int type() {return 1533;} public void run() {System.out.println(1533);}}
    public static class Impl1534 implements DemoService {public int type() {return 1534;} public void run() {System.out.println(1534);}}
    public static class Impl1535 implements DemoService {public int type() {return 1535;} public void run() {System.out.println(1535);}}
    public static class Impl1536 implements DemoService {public int type() {return 1536;} public void run() {System.out.println(1536);}}
    public static class Impl1537 implements DemoService {public int type() {return 1537;} public void run() {System.out.println(1537);}}
    public static class Impl1538 implements DemoService {public int type() {return 1538;} public void run() {System.out.println(1538);}}
    public static class Impl1539 implements DemoService {public int type() {return 1539;} public void run() {System.out.println(1539);}}
    public static class Impl1540 implements DemoService {public int type() {return 1540;} public void run() {System.out.println(1540);}}
    public static class Impl1541 implements DemoService {public int type() {return 1541;} public void run() {System.out.println(1541);}}
    public static class Impl1542 implements DemoService {public int type() {return 1542;} public void run() {System.out.println(1542);}}
    public static class Impl1543 implements DemoService {public int type() {return 1543;} public void run() {System.out.println(1543);}}
    public static class Impl1544 implements DemoService {public int type() {return 1544;} public void run() {System.out.println(1544);}}
    public static class Impl1545 implements DemoService {public int type() {return 1545;} public void run() {System.out.println(1545);}}
    public static class Impl1546 implements DemoService {public int type() {return 1546;} public void run() {System.out.println(1546);}}
    public static class Impl1547 implements DemoService {public int type() {return 1547;} public void run() {System.out.println(1547);}}
    public static class Impl1548 implements DemoService {public int type() {return 1548;} public void run() {System.out.println(1548);}}
    public static class Impl1549 implements DemoService {public int type() {return 1549;} public void run() {System.out.println(1549);}}
    public static class Impl1550 implements DemoService {public int type() {return 1550;} public void run() {System.out.println(1550);}}
    public static class Impl1551 implements DemoService {public int type() {return 1551;} public void run() {System.out.println(1551);}}
    public static class Impl1552 implements DemoService {public int type() {return 1552;} public void run() {System.out.println(1552);}}
    public static class Impl1553 implements DemoService {public int type() {return 1553;} public void run() {System.out.println(1553);}}
    public static class Impl1554 implements DemoService {public int type() {return 1554;} public void run() {System.out.println(1554);}}
    public static class Impl1555 implements DemoService {public int type() {return 1555;} public void run() {System.out.println(1555);}}
    public static class Impl1556 implements DemoService {public int type() {return 1556;} public void run() {System.out.println(1556);}}
    public static class Impl1557 implements DemoService {public int type() {return 1557;} public void run() {System.out.println(1557);}}
    public static class Impl1558 implements DemoService {public int type() {return 1558;} public void run() {System.out.println(1558);}}
    public static class Impl1559 implements DemoService {public int type() {return 1559;} public void run() {System.out.println(1559);}}
    public static class Impl1560 implements DemoService {public int type() {return 1560;} public void run() {System.out.println(1560);}}
    public static class Impl1561 implements DemoService {public int type() {return 1561;} public void run() {System.out.println(1561);}}
    public static class Impl1562 implements DemoService {public int type() {return 1562;} public void run() {System.out.println(1562);}}
    public static class Impl1563 implements DemoService {public int type() {return 1563;} public void run() {System.out.println(1563);}}
    public static class Impl1564 implements DemoService {public int type() {return 1564;} public void run() {System.out.println(1564);}}
    public static class Impl1565 implements DemoService {public int type() {return 1565;} public void run() {System.out.println(1565);}}
    public static class Impl1566 implements DemoService {public int type() {return 1566;} public void run() {System.out.println(1566);}}
    public static class Impl1567 implements DemoService {public int type() {return 1567;} public void run() {System.out.println(1567);}}
    public static class Impl1568 implements DemoService {public int type() {return 1568;} public void run() {System.out.println(1568);}}
    public static class Impl1569 implements DemoService {public int type() {return 1569;} public void run() {System.out.println(1569);}}
    public static class Impl1570 implements DemoService {public int type() {return 1570;} public void run() {System.out.println(1570);}}
    public static class Impl1571 implements DemoService {public int type() {return 1571;} public void run() {System.out.println(1571);}}
    public static class Impl1572 implements DemoService {public int type() {return 1572;} public void run() {System.out.println(1572);}}
    public static class Impl1573 implements DemoService {public int type() {return 1573;} public void run() {System.out.println(1573);}}
    public static class Impl1574 implements DemoService {public int type() {return 1574;} public void run() {System.out.println(1574);}}
    public static class Impl1575 implements DemoService {public int type() {return 1575;} public void run() {System.out.println(1575);}}
    public static class Impl1576 implements DemoService {public int type() {return 1576;} public void run() {System.out.println(1576);}}
    public static class Impl1577 implements DemoService {public int type() {return 1577;} public void run() {System.out.println(1577);}}
    public static class Impl1578 implements DemoService {public int type() {return 1578;} public void run() {System.out.println(1578);}}
    public static class Impl1579 implements DemoService {public int type() {return 1579;} public void run() {System.out.println(1579);}}
    public static class Impl1580 implements DemoService {public int type() {return 1580;} public void run() {System.out.println(1580);}}
    public static class Impl1581 implements DemoService {public int type() {return 1581;} public void run() {System.out.println(1581);}}
    public static class Impl1582 implements DemoService {public int type() {return 1582;} public void run() {System.out.println(1582);}}
    public static class Impl1583 implements DemoService {public int type() {return 1583;} public void run() {System.out.println(1583);}}
    public static class Impl1584 implements DemoService {public int type() {return 1584;} public void run() {System.out.println(1584);}}
    public static class Impl1585 implements DemoService {public int type() {return 1585;} public void run() {System.out.println(1585);}}
    public static class Impl1586 implements DemoService {public int type() {return 1586;} public void run() {System.out.println(1586);}}
    public static class Impl1587 implements DemoService {public int type() {return 1587;} public void run() {System.out.println(1587);}}
    public static class Impl1588 implements DemoService {public int type() {return 1588;} public void run() {System.out.println(1588);}}
    public static class Impl1589 implements DemoService {public int type() {return 1589;} public void run() {System.out.println(1589);}}
    public static class Impl1590 implements DemoService {public int type() {return 1590;} public void run() {System.out.println(1590);}}
    public static class Impl1591 implements DemoService {public int type() {return 1591;} public void run() {System.out.println(1591);}}
    public static class Impl1592 implements DemoService {public int type() {return 1592;} public void run() {System.out.println(1592);}}
    public static class Impl1593 implements DemoService {public int type() {return 1593;} public void run() {System.out.println(1593);}}
    public static class Impl1594 implements DemoService {public int type() {return 1594;} public void run() {System.out.println(1594);}}
    public static class Impl1595 implements DemoService {public int type() {return 1595;} public void run() {System.out.println(1595);}}
    public static class Impl1596 implements DemoService {public int type() {return 1596;} public void run() {System.out.println(1596);}}
    public static class Impl1597 implements DemoService {public int type() {return 1597;} public void run() {System.out.println(1597);}}
    public static class Impl1598 implements DemoService {public int type() {return 1598;} public void run() {System.out.println(1598);}}
    public static class Impl1599 implements DemoService {public int type() {return 1599;} public void run() {System.out.println(1599);}}
    public static class Impl1600 implements DemoService {public int type() {return 1600;} public void run() {System.out.println(1600);}}
    public static class Impl1601 implements DemoService {public int type() {return 1601;} public void run() {System.out.println(1601);}}
    public static class Impl1602 implements DemoService {public int type() {return 1602;} public void run() {System.out.println(1602);}}
    public static class Impl1603 implements DemoService {public int type() {return 1603;} public void run() {System.out.println(1603);}}
    public static class Impl1604 implements DemoService {public int type() {return 1604;} public void run() {System.out.println(1604);}}
    public static class Impl1605 implements DemoService {public int type() {return 1605;} public void run() {System.out.println(1605);}}
    public static class Impl1606 implements DemoService {public int type() {return 1606;} public void run() {System.out.println(1606);}}
    public static class Impl1607 implements DemoService {public int type() {return 1607;} public void run() {System.out.println(1607);}}
    public static class Impl1608 implements DemoService {public int type() {return 1608;} public void run() {System.out.println(1608);}}
    public static class Impl1609 implements DemoService {public int type() {return 1609;} public void run() {System.out.println(1609);}}
    public static class Impl1610 implements DemoService {public int type() {return 1610;} public void run() {System.out.println(1610);}}
    public static class Impl1611 implements DemoService {public int type() {return 1611;} public void run() {System.out.println(1611);}}
    public static class Impl1612 implements DemoService {public int type() {return 1612;} public void run() {System.out.println(1612);}}
    public static class Impl1613 implements DemoService {public int type() {return 1613;} public void run() {System.out.println(1613);}}
    public static class Impl1614 implements DemoService {public int type() {return 1614;} public void run() {System.out.println(1614);}}
    public static class Impl1615 implements DemoService {public int type() {return 1615;} public void run() {System.out.println(1615);}}
    public static class Impl1616 implements DemoService {public int type() {return 1616;} public void run() {System.out.println(1616);}}
    public static class Impl1617 implements DemoService {public int type() {return 1617;} public void run() {System.out.println(1617);}}
    public static class Impl1618 implements DemoService {public int type() {return 1618;} public void run() {System.out.println(1618);}}
    public static class Impl1619 implements DemoService {public int type() {return 1619;} public void run() {System.out.println(1619);}}
    public static class Impl1620 implements DemoService {public int type() {return 1620;} public void run() {System.out.println(1620);}}
    public static class Impl1621 implements DemoService {public int type() {return 1621;} public void run() {System.out.println(1621);}}
    public static class Impl1622 implements DemoService {public int type() {return 1622;} public void run() {System.out.println(1622);}}
    public static class Impl1623 implements DemoService {public int type() {return 1623;} public void run() {System.out.println(1623);}}
    public static class Impl1624 implements DemoService {public int type() {return 1624;} public void run() {System.out.println(1624);}}
    public static class Impl1625 implements DemoService {public int type() {return 1625;} public void run() {System.out.println(1625);}}
    public static class Impl1626 implements DemoService {public int type() {return 1626;} public void run() {System.out.println(1626);}}
    public static class Impl1627 implements DemoService {public int type() {return 1627;} public void run() {System.out.println(1627);}}
    public static class Impl1628 implements DemoService {public int type() {return 1628;} public void run() {System.out.println(1628);}}
    public static class Impl1629 implements DemoService {public int type() {return 1629;} public void run() {System.out.println(1629);}}
    public static class Impl1630 implements DemoService {public int type() {return 1630;} public void run() {System.out.println(1630);}}
    public static class Impl1631 implements DemoService {public int type() {return 1631;} public void run() {System.out.println(1631);}}
    public static class Impl1632 implements DemoService {public int type() {return 1632;} public void run() {System.out.println(1632);}}
    public static class Impl1633 implements DemoService {public int type() {return 1633;} public void run() {System.out.println(1633);}}
    public static class Impl1634 implements DemoService {public int type() {return 1634;} public void run() {System.out.println(1634);}}
    public static class Impl1635 implements DemoService {public int type() {return 1635;} public void run() {System.out.println(1635);}}
    public static class Impl1636 implements DemoService {public int type() {return 1636;} public void run() {System.out.println(1636);}}
    public static class Impl1637 implements DemoService {public int type() {return 1637;} public void run() {System.out.println(1637);}}
    public static class Impl1638 implements DemoService {public int type() {return 1638;} public void run() {System.out.println(1638);}}
    public static class Impl1639 implements DemoService {public int type() {return 1639;} public void run() {System.out.println(1639);}}
    public static class Impl1640 implements DemoService {public int type() {return 1640;} public void run() {System.out.println(1640);}}
    public static class Impl1641 implements DemoService {public int type() {return 1641;} public void run() {System.out.println(1641);}}
    public static class Impl1642 implements DemoService {public int type() {return 1642;} public void run() {System.out.println(1642);}}
    public static class Impl1643 implements DemoService {public int type() {return 1643;} public void run() {System.out.println(1643);}}
    public static class Impl1644 implements DemoService {public int type() {return 1644;} public void run() {System.out.println(1644);}}
    public static class Impl1645 implements DemoService {public int type() {return 1645;} public void run() {System.out.println(1645);}}
    public static class Impl1646 implements DemoService {public int type() {return 1646;} public void run() {System.out.println(1646);}}
    public static class Impl1647 implements DemoService {public int type() {return 1647;} public void run() {System.out.println(1647);}}
    public static class Impl1648 implements DemoService {public int type() {return 1648;} public void run() {System.out.println(1648);}}
    public static class Impl1649 implements DemoService {public int type() {return 1649;} public void run() {System.out.println(1649);}}
    public static class Impl1650 implements DemoService {public int type() {return 1650;} public void run() {System.out.println(1650);}}
    public static class Impl1651 implements DemoService {public int type() {return 1651;} public void run() {System.out.println(1651);}}
    public static class Impl1652 implements DemoService {public int type() {return 1652;} public void run() {System.out.println(1652);}}
    public static class Impl1653 implements DemoService {public int type() {return 1653;} public void run() {System.out.println(1653);}}
    public static class Impl1654 implements DemoService {public int type() {return 1654;} public void run() {System.out.println(1654);}}
    public static class Impl1655 implements DemoService {public int type() {return 1655;} public void run() {System.out.println(1655);}}
    public static class Impl1656 implements DemoService {public int type() {return 1656;} public void run() {System.out.println(1656);}}
    public static class Impl1657 implements DemoService {public int type() {return 1657;} public void run() {System.out.println(1657);}}
    public static class Impl1658 implements DemoService {public int type() {return 1658;} public void run() {System.out.println(1658);}}
    public static class Impl1659 implements DemoService {public int type() {return 1659;} public void run() {System.out.println(1659);}}
    public static class Impl1660 implements DemoService {public int type() {return 1660;} public void run() {System.out.println(1660);}}
    public static class Impl1661 implements DemoService {public int type() {return 1661;} public void run() {System.out.println(1661);}}
    public static class Impl1662 implements DemoService {public int type() {return 1662;} public void run() {System.out.println(1662);}}
    public static class Impl1663 implements DemoService {public int type() {return 1663;} public void run() {System.out.println(1663);}}
    public static class Impl1664 implements DemoService {public int type() {return 1664;} public void run() {System.out.println(1664);}}
    public static class Impl1665 implements DemoService {public int type() {return 1665;} public void run() {System.out.println(1665);}}
    public static class Impl1666 implements DemoService {public int type() {return 1666;} public void run() {System.out.println(1666);}}
    public static class Impl1667 implements DemoService {public int type() {return 1667;} public void run() {System.out.println(1667);}}
    public static class Impl1668 implements DemoService {public int type() {return 1668;} public void run() {System.out.println(1668);}}
    public static class Impl1669 implements DemoService {public int type() {return 1669;} public void run() {System.out.println(1669);}}
    public static class Impl1670 implements DemoService {public int type() {return 1670;} public void run() {System.out.println(1670);}}
    public static class Impl1671 implements DemoService {public int type() {return 1671;} public void run() {System.out.println(1671);}}
    public static class Impl1672 implements DemoService {public int type() {return 1672;} public void run() {System.out.println(1672);}}
    public static class Impl1673 implements DemoService {public int type() {return 1673;} public void run() {System.out.println(1673);}}
    public static class Impl1674 implements DemoService {public int type() {return 1674;} public void run() {System.out.println(1674);}}
    public static class Impl1675 implements DemoService {public int type() {return 1675;} public void run() {System.out.println(1675);}}
    public static class Impl1676 implements DemoService {public int type() {return 1676;} public void run() {System.out.println(1676);}}
    public static class Impl1677 implements DemoService {public int type() {return 1677;} public void run() {System.out.println(1677);}}
    public static class Impl1678 implements DemoService {public int type() {return 1678;} public void run() {System.out.println(1678);}}
    public static class Impl1679 implements DemoService {public int type() {return 1679;} public void run() {System.out.println(1679);}}
    public static class Impl1680 implements DemoService {public int type() {return 1680;} public void run() {System.out.println(1680);}}
    public static class Impl1681 implements DemoService {public int type() {return 1681;} public void run() {System.out.println(1681);}}
    public static class Impl1682 implements DemoService {public int type() {return 1682;} public void run() {System.out.println(1682);}}
    public static class Impl1683 implements DemoService {public int type() {return 1683;} public void run() {System.out.println(1683);}}
    public static class Impl1684 implements DemoService {public int type() {return 1684;} public void run() {System.out.println(1684);}}
    public static class Impl1685 implements DemoService {public int type() {return 1685;} public void run() {System.out.println(1685);}}
    public static class Impl1686 implements DemoService {public int type() {return 1686;} public void run() {System.out.println(1686);}}
    public static class Impl1687 implements DemoService {public int type() {return 1687;} public void run() {System.out.println(1687);}}
    public static class Impl1688 implements DemoService {public int type() {return 1688;} public void run() {System.out.println(1688);}}
    public static class Impl1689 implements DemoService {public int type() {return 1689;} public void run() {System.out.println(1689);}}
    public static class Impl1690 implements DemoService {public int type() {return 1690;} public void run() {System.out.println(1690);}}
    public static class Impl1691 implements DemoService {public int type() {return 1691;} public void run() {System.out.println(1691);}}
    public static class Impl1692 implements DemoService {public int type() {return 1692;} public void run() {System.out.println(1692);}}
    public static class Impl1693 implements DemoService {public int type() {return 1693;} public void run() {System.out.println(1693);}}
    public static class Impl1694 implements DemoService {public int type() {return 1694;} public void run() {System.out.println(1694);}}
    public static class Impl1695 implements DemoService {public int type() {return 1695;} public void run() {System.out.println(1695);}}
    public static class Impl1696 implements DemoService {public int type() {return 1696;} public void run() {System.out.println(1696);}}
    public static class Impl1697 implements DemoService {public int type() {return 1697;} public void run() {System.out.println(1697);}}
    public static class Impl1698 implements DemoService {public int type() {return 1698;} public void run() {System.out.println(1698);}}
    public static class Impl1699 implements DemoService {public int type() {return 1699;} public void run() {System.out.println(1699);}}
    public static class Impl1700 implements DemoService {public int type() {return 1700;} public void run() {System.out.println(1700);}}
    public static class Impl1701 implements DemoService {public int type() {return 1701;} public void run() {System.out.println(1701);}}
    public static class Impl1702 implements DemoService {public int type() {return 1702;} public void run() {System.out.println(1702);}}
    public static class Impl1703 implements DemoService {public int type() {return 1703;} public void run() {System.out.println(1703);}}
    public static class Impl1704 implements DemoService {public int type() {return 1704;} public void run() {System.out.println(1704);}}
    public static class Impl1705 implements DemoService {public int type() {return 1705;} public void run() {System.out.println(1705);}}
    public static class Impl1706 implements DemoService {public int type() {return 1706;} public void run() {System.out.println(1706);}}
    public static class Impl1707 implements DemoService {public int type() {return 1707;} public void run() {System.out.println(1707);}}
    public static class Impl1708 implements DemoService {public int type() {return 1708;} public void run() {System.out.println(1708);}}
    public static class Impl1709 implements DemoService {public int type() {return 1709;} public void run() {System.out.println(1709);}}
    public static class Impl1710 implements DemoService {public int type() {return 1710;} public void run() {System.out.println(1710);}}
    public static class Impl1711 implements DemoService {public int type() {return 1711;} public void run() {System.out.println(1711);}}
    public static class Impl1712 implements DemoService {public int type() {return 1712;} public void run() {System.out.println(1712);}}
    public static class Impl1713 implements DemoService {public int type() {return 1713;} public void run() {System.out.println(1713);}}
    public static class Impl1714 implements DemoService {public int type() {return 1714;} public void run() {System.out.println(1714);}}
    public static class Impl1715 implements DemoService {public int type() {return 1715;} public void run() {System.out.println(1715);}}
    public static class Impl1716 implements DemoService {public int type() {return 1716;} public void run() {System.out.println(1716);}}
    public static class Impl1717 implements DemoService {public int type() {return 1717;} public void run() {System.out.println(1717);}}
    public static class Impl1718 implements DemoService {public int type() {return 1718;} public void run() {System.out.println(1718);}}
    public static class Impl1719 implements DemoService {public int type() {return 1719;} public void run() {System.out.println(1719);}}
    public static class Impl1720 implements DemoService {public int type() {return 1720;} public void run() {System.out.println(1720);}}
    public static class Impl1721 implements DemoService {public int type() {return 1721;} public void run() {System.out.println(1721);}}
    public static class Impl1722 implements DemoService {public int type() {return 1722;} public void run() {System.out.println(1722);}}
    public static class Impl1723 implements DemoService {public int type() {return 1723;} public void run() {System.out.println(1723);}}
    public static class Impl1724 implements DemoService {public int type() {return 1724;} public void run() {System.out.println(1724);}}
    public static class Impl1725 implements DemoService {public int type() {return 1725;} public void run() {System.out.println(1725);}}
    public static class Impl1726 implements DemoService {public int type() {return 1726;} public void run() {System.out.println(1726);}}
    public static class Impl1727 implements DemoService {public int type() {return 1727;} public void run() {System.out.println(1727);}}
    public static class Impl1728 implements DemoService {public int type() {return 1728;} public void run() {System.out.println(1728);}}
    public static class Impl1729 implements DemoService {public int type() {return 1729;} public void run() {System.out.println(1729);}}
    public static class Impl1730 implements DemoService {public int type() {return 1730;} public void run() {System.out.println(1730);}}
    public static class Impl1731 implements DemoService {public int type() {return 1731;} public void run() {System.out.println(1731);}}
    public static class Impl1732 implements DemoService {public int type() {return 1732;} public void run() {System.out.println(1732);}}
    public static class Impl1733 implements DemoService {public int type() {return 1733;} public void run() {System.out.println(1733);}}
    public static class Impl1734 implements DemoService {public int type() {return 1734;} public void run() {System.out.println(1734);}}
    public static class Impl1735 implements DemoService {public int type() {return 1735;} public void run() {System.out.println(1735);}}
    public static class Impl1736 implements DemoService {public int type() {return 1736;} public void run() {System.out.println(1736);}}
    public static class Impl1737 implements DemoService {public int type() {return 1737;} public void run() {System.out.println(1737);}}
    public static class Impl1738 implements DemoService {public int type() {return 1738;} public void run() {System.out.println(1738);}}
    public static class Impl1739 implements DemoService {public int type() {return 1739;} public void run() {System.out.println(1739);}}
    public static class Impl1740 implements DemoService {public int type() {return 1740;} public void run() {System.out.println(1740);}}
    public static class Impl1741 implements DemoService {public int type() {return 1741;} public void run() {System.out.println(1741);}}
    public static class Impl1742 implements DemoService {public int type() {return 1742;} public void run() {System.out.println(1742);}}
    public static class Impl1743 implements DemoService {public int type() {return 1743;} public void run() {System.out.println(1743);}}
    public static class Impl1744 implements DemoService {public int type() {return 1744;} public void run() {System.out.println(1744);}}
    public static class Impl1745 implements DemoService {public int type() {return 1745;} public void run() {System.out.println(1745);}}
    public static class Impl1746 implements DemoService {public int type() {return 1746;} public void run() {System.out.println(1746);}}
    public static class Impl1747 implements DemoService {public int type() {return 1747;} public void run() {System.out.println(1747);}}
    public static class Impl1748 implements DemoService {public int type() {return 1748;} public void run() {System.out.println(1748);}}
    public static class Impl1749 implements DemoService {public int type() {return 1749;} public void run() {System.out.println(1749);}}
    public static class Impl1750 implements DemoService {public int type() {return 1750;} public void run() {System.out.println(1750);}}
    public static class Impl1751 implements DemoService {public int type() {return 1751;} public void run() {System.out.println(1751);}}
    public static class Impl1752 implements DemoService {public int type() {return 1752;} public void run() {System.out.println(1752);}}
    public static class Impl1753 implements DemoService {public int type() {return 1753;} public void run() {System.out.println(1753);}}
    public static class Impl1754 implements DemoService {public int type() {return 1754;} public void run() {System.out.println(1754);}}
    public static class Impl1755 implements DemoService {public int type() {return 1755;} public void run() {System.out.println(1755);}}
    public static class Impl1756 implements DemoService {public int type() {return 1756;} public void run() {System.out.println(1756);}}
    public static class Impl1757 implements DemoService {public int type() {return 1757;} public void run() {System.out.println(1757);}}
    public static class Impl1758 implements DemoService {public int type() {return 1758;} public void run() {System.out.println(1758);}}
    public static class Impl1759 implements DemoService {public int type() {return 1759;} public void run() {System.out.println(1759);}}
    public static class Impl1760 implements DemoService {public int type() {return 1760;} public void run() {System.out.println(1760);}}
    public static class Impl1761 implements DemoService {public int type() {return 1761;} public void run() {System.out.println(1761);}}
    public static class Impl1762 implements DemoService {public int type() {return 1762;} public void run() {System.out.println(1762);}}
    public static class Impl1763 implements DemoService {public int type() {return 1763;} public void run() {System.out.println(1763);}}
    public static class Impl1764 implements DemoService {public int type() {return 1764;} public void run() {System.out.println(1764);}}
    public static class Impl1765 implements DemoService {public int type() {return 1765;} public void run() {System.out.println(1765);}}
    public static class Impl1766 implements DemoService {public int type() {return 1766;} public void run() {System.out.println(1766);}}
    public static class Impl1767 implements DemoService {public int type() {return 1767;} public void run() {System.out.println(1767);}}
    public static class Impl1768 implements DemoService {public int type() {return 1768;} public void run() {System.out.println(1768);}}
    public static class Impl1769 implements DemoService {public int type() {return 1769;} public void run() {System.out.println(1769);}}
    public static class Impl1770 implements DemoService {public int type() {return 1770;} public void run() {System.out.println(1770);}}
    public static class Impl1771 implements DemoService {public int type() {return 1771;} public void run() {System.out.println(1771);}}
    public static class Impl1772 implements DemoService {public int type() {return 1772;} public void run() {System.out.println(1772);}}
    public static class Impl1773 implements DemoService {public int type() {return 1773;} public void run() {System.out.println(1773);}}
    public static class Impl1774 implements DemoService {public int type() {return 1774;} public void run() {System.out.println(1774);}}
    public static class Impl1775 implements DemoService {public int type() {return 1775;} public void run() {System.out.println(1775);}}
    public static class Impl1776 implements DemoService {public int type() {return 1776;} public void run() {System.out.println(1776);}}
    public static class Impl1777 implements DemoService {public int type() {return 1777;} public void run() {System.out.println(1777);}}
    public static class Impl1778 implements DemoService {public int type() {return 1778;} public void run() {System.out.println(1778);}}
    public static class Impl1779 implements DemoService {public int type() {return 1779;} public void run() {System.out.println(1779);}}
    public static class Impl1780 implements DemoService {public int type() {return 1780;} public void run() {System.out.println(1780);}}
    public static class Impl1781 implements DemoService {public int type() {return 1781;} public void run() {System.out.println(1781);}}
    public static class Impl1782 implements DemoService {public int type() {return 1782;} public void run() {System.out.println(1782);}}
    public static class Impl1783 implements DemoService {public int type() {return 1783;} public void run() {System.out.println(1783);}}
    public static class Impl1784 implements DemoService {public int type() {return 1784;} public void run() {System.out.println(1784);}}
    public static class Impl1785 implements DemoService {public int type() {return 1785;} public void run() {System.out.println(1785);}}
    public static class Impl1786 implements DemoService {public int type() {return 1786;} public void run() {System.out.println(1786);}}
    public static class Impl1787 implements DemoService {public int type() {return 1787;} public void run() {System.out.println(1787);}}
    public static class Impl1788 implements DemoService {public int type() {return 1788;} public void run() {System.out.println(1788);}}
    public static class Impl1789 implements DemoService {public int type() {return 1789;} public void run() {System.out.println(1789);}}
    public static class Impl1790 implements DemoService {public int type() {return 1790;} public void run() {System.out.println(1790);}}
    public static class Impl1791 implements DemoService {public int type() {return 1791;} public void run() {System.out.println(1791);}}
    public static class Impl1792 implements DemoService {public int type() {return 1792;} public void run() {System.out.println(1792);}}
    public static class Impl1793 implements DemoService {public int type() {return 1793;} public void run() {System.out.println(1793);}}
    public static class Impl1794 implements DemoService {public int type() {return 1794;} public void run() {System.out.println(1794);}}
    public static class Impl1795 implements DemoService {public int type() {return 1795;} public void run() {System.out.println(1795);}}
    public static class Impl1796 implements DemoService {public int type() {return 1796;} public void run() {System.out.println(1796);}}
    public static class Impl1797 implements DemoService {public int type() {return 1797;} public void run() {System.out.println(1797);}}
    public static class Impl1798 implements DemoService {public int type() {return 1798;} public void run() {System.out.println(1798);}}
    public static class Impl1799 implements DemoService {public int type() {return 1799;} public void run() {System.out.println(1799);}}
    public static class Impl1800 implements DemoService {public int type() {return 1800;} public void run() {System.out.println(1800);}}
    public static class Impl1801 implements DemoService {public int type() {return 1801;} public void run() {System.out.println(1801);}}
    public static class Impl1802 implements DemoService {public int type() {return 1802;} public void run() {System.out.println(1802);}}
    public static class Impl1803 implements DemoService {public int type() {return 1803;} public void run() {System.out.println(1803);}}
    public static class Impl1804 implements DemoService {public int type() {return 1804;} public void run() {System.out.println(1804);}}
    public static class Impl1805 implements DemoService {public int type() {return 1805;} public void run() {System.out.println(1805);}}
    public static class Impl1806 implements DemoService {public int type() {return 1806;} public void run() {System.out.println(1806);}}
    public static class Impl1807 implements DemoService {public int type() {return 1807;} public void run() {System.out.println(1807);}}
    public static class Impl1808 implements DemoService {public int type() {return 1808;} public void run() {System.out.println(1808);}}
    public static class Impl1809 implements DemoService {public int type() {return 1809;} public void run() {System.out.println(1809);}}
    public static class Impl1810 implements DemoService {public int type() {return 1810;} public void run() {System.out.println(1810);}}
    public static class Impl1811 implements DemoService {public int type() {return 1811;} public void run() {System.out.println(1811);}}
    public static class Impl1812 implements DemoService {public int type() {return 1812;} public void run() {System.out.println(1812);}}
    public static class Impl1813 implements DemoService {public int type() {return 1813;} public void run() {System.out.println(1813);}}
    public static class Impl1814 implements DemoService {public int type() {return 1814;} public void run() {System.out.println(1814);}}
    public static class Impl1815 implements DemoService {public int type() {return 1815;} public void run() {System.out.println(1815);}}
    public static class Impl1816 implements DemoService {public int type() {return 1816;} public void run() {System.out.println(1816);}}
    public static class Impl1817 implements DemoService {public int type() {return 1817;} public void run() {System.out.println(1817);}}
    public static class Impl1818 implements DemoService {public int type() {return 1818;} public void run() {System.out.println(1818);}}
    public static class Impl1819 implements DemoService {public int type() {return 1819;} public void run() {System.out.println(1819);}}
    public static class Impl1820 implements DemoService {public int type() {return 1820;} public void run() {System.out.println(1820);}}
    public static class Impl1821 implements DemoService {public int type() {return 1821;} public void run() {System.out.println(1821);}}
    public static class Impl1822 implements DemoService {public int type() {return 1822;} public void run() {System.out.println(1822);}}
    public static class Impl1823 implements DemoService {public int type() {return 1823;} public void run() {System.out.println(1823);}}
    public static class Impl1824 implements DemoService {public int type() {return 1824;} public void run() {System.out.println(1824);}}
    public static class Impl1825 implements DemoService {public int type() {return 1825;} public void run() {System.out.println(1825);}}
    public static class Impl1826 implements DemoService {public int type() {return 1826;} public void run() {System.out.println(1826);}}
    public static class Impl1827 implements DemoService {public int type() {return 1827;} public void run() {System.out.println(1827);}}
    public static class Impl1828 implements DemoService {public int type() {return 1828;} public void run() {System.out.println(1828);}}
    public static class Impl1829 implements DemoService {public int type() {return 1829;} public void run() {System.out.println(1829);}}
    public static class Impl1830 implements DemoService {public int type() {return 1830;} public void run() {System.out.println(1830);}}
    public static class Impl1831 implements DemoService {public int type() {return 1831;} public void run() {System.out.println(1831);}}
    public static class Impl1832 implements DemoService {public int type() {return 1832;} public void run() {System.out.println(1832);}}
    public static class Impl1833 implements DemoService {public int type() {return 1833;} public void run() {System.out.println(1833);}}
    public static class Impl1834 implements DemoService {public int type() {return 1834;} public void run() {System.out.println(1834);}}
    public static class Impl1835 implements DemoService {public int type() {return 1835;} public void run() {System.out.println(1835);}}
    public static class Impl1836 implements DemoService {public int type() {return 1836;} public void run() {System.out.println(1836);}}
    public static class Impl1837 implements DemoService {public int type() {return 1837;} public void run() {System.out.println(1837);}}
    public static class Impl1838 implements DemoService {public int type() {return 1838;} public void run() {System.out.println(1838);}}
    public static class Impl1839 implements DemoService {public int type() {return 1839;} public void run() {System.out.println(1839);}}
    public static class Impl1840 implements DemoService {public int type() {return 1840;} public void run() {System.out.println(1840);}}
    public static class Impl1841 implements DemoService {public int type() {return 1841;} public void run() {System.out.println(1841);}}
    public static class Impl1842 implements DemoService {public int type() {return 1842;} public void run() {System.out.println(1842);}}
    public static class Impl1843 implements DemoService {public int type() {return 1843;} public void run() {System.out.println(1843);}}
    public static class Impl1844 implements DemoService {public int type() {return 1844;} public void run() {System.out.println(1844);}}
    public static class Impl1845 implements DemoService {public int type() {return 1845;} public void run() {System.out.println(1845);}}
    public static class Impl1846 implements DemoService {public int type() {return 1846;} public void run() {System.out.println(1846);}}
    public static class Impl1847 implements DemoService {public int type() {return 1847;} public void run() {System.out.println(1847);}}
    public static class Impl1848 implements DemoService {public int type() {return 1848;} public void run() {System.out.println(1848);}}
    public static class Impl1849 implements DemoService {public int type() {return 1849;} public void run() {System.out.println(1849);}}
    public static class Impl1850 implements DemoService {public int type() {return 1850;} public void run() {System.out.println(1850);}}
    public static class Impl1851 implements DemoService {public int type() {return 1851;} public void run() {System.out.println(1851);}}
    public static class Impl1852 implements DemoService {public int type() {return 1852;} public void run() {System.out.println(1852);}}
    public static class Impl1853 implements DemoService {public int type() {return 1853;} public void run() {System.out.println(1853);}}
    public static class Impl1854 implements DemoService {public int type() {return 1854;} public void run() {System.out.println(1854);}}
    public static class Impl1855 implements DemoService {public int type() {return 1855;} public void run() {System.out.println(1855);}}
    public static class Impl1856 implements DemoService {public int type() {return 1856;} public void run() {System.out.println(1856);}}
    public static class Impl1857 implements DemoService {public int type() {return 1857;} public void run() {System.out.println(1857);}}
    public static class Impl1858 implements DemoService {public int type() {return 1858;} public void run() {System.out.println(1858);}}
    public static class Impl1859 implements DemoService {public int type() {return 1859;} public void run() {System.out.println(1859);}}
    public static class Impl1860 implements DemoService {public int type() {return 1860;} public void run() {System.out.println(1860);}}
    public static class Impl1861 implements DemoService {public int type() {return 1861;} public void run() {System.out.println(1861);}}
    public static class Impl1862 implements DemoService {public int type() {return 1862;} public void run() {System.out.println(1862);}}
    public static class Impl1863 implements DemoService {public int type() {return 1863;} public void run() {System.out.println(1863);}}
    public static class Impl1864 implements DemoService {public int type() {return 1864;} public void run() {System.out.println(1864);}}
    public static class Impl1865 implements DemoService {public int type() {return 1865;} public void run() {System.out.println(1865);}}
    public static class Impl1866 implements DemoService {public int type() {return 1866;} public void run() {System.out.println(1866);}}
    public static class Impl1867 implements DemoService {public int type() {return 1867;} public void run() {System.out.println(1867);}}
    public static class Impl1868 implements DemoService {public int type() {return 1868;} public void run() {System.out.println(1868);}}
    public static class Impl1869 implements DemoService {public int type() {return 1869;} public void run() {System.out.println(1869);}}
    public static class Impl1870 implements DemoService {public int type() {return 1870;} public void run() {System.out.println(1870);}}
    public static class Impl1871 implements DemoService {public int type() {return 1871;} public void run() {System.out.println(1871);}}
    public static class Impl1872 implements DemoService {public int type() {return 1872;} public void run() {System.out.println(1872);}}
    public static class Impl1873 implements DemoService {public int type() {return 1873;} public void run() {System.out.println(1873);}}
    public static class Impl1874 implements DemoService {public int type() {return 1874;} public void run() {System.out.println(1874);}}
    public static class Impl1875 implements DemoService {public int type() {return 1875;} public void run() {System.out.println(1875);}}
    public static class Impl1876 implements DemoService {public int type() {return 1876;} public void run() {System.out.println(1876);}}
    public static class Impl1877 implements DemoService {public int type() {return 1877;} public void run() {System.out.println(1877);}}
    public static class Impl1878 implements DemoService {public int type() {return 1878;} public void run() {System.out.println(1878);}}
    public static class Impl1879 implements DemoService {public int type() {return 1879;} public void run() {System.out.println(1879);}}
    public static class Impl1880 implements DemoService {public int type() {return 1880;} public void run() {System.out.println(1880);}}
    public static class Impl1881 implements DemoService {public int type() {return 1881;} public void run() {System.out.println(1881);}}
    public static class Impl1882 implements DemoService {public int type() {return 1882;} public void run() {System.out.println(1882);}}
    public static class Impl1883 implements DemoService {public int type() {return 1883;} public void run() {System.out.println(1883);}}
    public static class Impl1884 implements DemoService {public int type() {return 1884;} public void run() {System.out.println(1884);}}
    public static class Impl1885 implements DemoService {public int type() {return 1885;} public void run() {System.out.println(1885);}}
    public static class Impl1886 implements DemoService {public int type() {return 1886;} public void run() {System.out.println(1886);}}
    public static class Impl1887 implements DemoService {public int type() {return 1887;} public void run() {System.out.println(1887);}}
    public static class Impl1888 implements DemoService {public int type() {return 1888;} public void run() {System.out.println(1888);}}
    public static class Impl1889 implements DemoService {public int type() {return 1889;} public void run() {System.out.println(1889);}}
    public static class Impl1890 implements DemoService {public int type() {return 1890;} public void run() {System.out.println(1890);}}
    public static class Impl1891 implements DemoService {public int type() {return 1891;} public void run() {System.out.println(1891);}}
    public static class Impl1892 implements DemoService {public int type() {return 1892;} public void run() {System.out.println(1892);}}
    public static class Impl1893 implements DemoService {public int type() {return 1893;} public void run() {System.out.println(1893);}}
    public static class Impl1894 implements DemoService {public int type() {return 1894;} public void run() {System.out.println(1894);}}
    public static class Impl1895 implements DemoService {public int type() {return 1895;} public void run() {System.out.println(1895);}}
    public static class Impl1896 implements DemoService {public int type() {return 1896;} public void run() {System.out.println(1896);}}
    public static class Impl1897 implements DemoService {public int type() {return 1897;} public void run() {System.out.println(1897);}}
    public static class Impl1898 implements DemoService {public int type() {return 1898;} public void run() {System.out.println(1898);}}
    public static class Impl1899 implements DemoService {public int type() {return 1899;} public void run() {System.out.println(1899);}}
    public static class Impl1900 implements DemoService {public int type() {return 1900;} public void run() {System.out.println(1900);}}
    public static class Impl1901 implements DemoService {public int type() {return 1901;} public void run() {System.out.println(1901);}}
    public static class Impl1902 implements DemoService {public int type() {return 1902;} public void run() {System.out.println(1902);}}
    public static class Impl1903 implements DemoService {public int type() {return 1903;} public void run() {System.out.println(1903);}}
    public static class Impl1904 implements DemoService {public int type() {return 1904;} public void run() {System.out.println(1904);}}
    public static class Impl1905 implements DemoService {public int type() {return 1905;} public void run() {System.out.println(1905);}}
    public static class Impl1906 implements DemoService {public int type() {return 1906;} public void run() {System.out.println(1906);}}
    public static class Impl1907 implements DemoService {public int type() {return 1907;} public void run() {System.out.println(1907);}}
    public static class Impl1908 implements DemoService {public int type() {return 1908;} public void run() {System.out.println(1908);}}
    public static class Impl1909 implements DemoService {public int type() {return 1909;} public void run() {System.out.println(1909);}}
    public static class Impl1910 implements DemoService {public int type() {return 1910;} public void run() {System.out.println(1910);}}
    public static class Impl1911 implements DemoService {public int type() {return 1911;} public void run() {System.out.println(1911);}}
    public static class Impl1912 implements DemoService {public int type() {return 1912;} public void run() {System.out.println(1912);}}
    public static class Impl1913 implements DemoService {public int type() {return 1913;} public void run() {System.out.println(1913);}}
    public static class Impl1914 implements DemoService {public int type() {return 1914;} public void run() {System.out.println(1914);}}
    public static class Impl1915 implements DemoService {public int type() {return 1915;} public void run() {System.out.println(1915);}}
    public static class Impl1916 implements DemoService {public int type() {return 1916;} public void run() {System.out.println(1916);}}
    public static class Impl1917 implements DemoService {public int type() {return 1917;} public void run() {System.out.println(1917);}}
    public static class Impl1918 implements DemoService {public int type() {return 1918;} public void run() {System.out.println(1918);}}
    public static class Impl1919 implements DemoService {public int type() {return 1919;} public void run() {System.out.println(1919);}}
    public static class Impl1920 implements DemoService {public int type() {return 1920;} public void run() {System.out.println(1920);}}
    public static class Impl1921 implements DemoService {public int type() {return 1921;} public void run() {System.out.println(1921);}}
    public static class Impl1922 implements DemoService {public int type() {return 1922;} public void run() {System.out.println(1922);}}
    public static class Impl1923 implements DemoService {public int type() {return 1923;} public void run() {System.out.println(1923);}}
    public static class Impl1924 implements DemoService {public int type() {return 1924;} public void run() {System.out.println(1924);}}
    public static class Impl1925 implements DemoService {public int type() {return 1925;} public void run() {System.out.println(1925);}}
    public static class Impl1926 implements DemoService {public int type() {return 1926;} public void run() {System.out.println(1926);}}
    public static class Impl1927 implements DemoService {public int type() {return 1927;} public void run() {System.out.println(1927);}}
    public static class Impl1928 implements DemoService {public int type() {return 1928;} public void run() {System.out.println(1928);}}
    public static class Impl1929 implements DemoService {public int type() {return 1929;} public void run() {System.out.println(1929);}}
    public static class Impl1930 implements DemoService {public int type() {return 1930;} public void run() {System.out.println(1930);}}
    public static class Impl1931 implements DemoService {public int type() {return 1931;} public void run() {System.out.println(1931);}}
    public static class Impl1932 implements DemoService {public int type() {return 1932;} public void run() {System.out.println(1932);}}
    public static class Impl1933 implements DemoService {public int type() {return 1933;} public void run() {System.out.println(1933);}}
    public static class Impl1934 implements DemoService {public int type() {return 1934;} public void run() {System.out.println(1934);}}
    public static class Impl1935 implements DemoService {public int type() {return 1935;} public void run() {System.out.println(1935);}}
    public static class Impl1936 implements DemoService {public int type() {return 1936;} public void run() {System.out.println(1936);}}
    public static class Impl1937 implements DemoService {public int type() {return 1937;} public void run() {System.out.println(1937);}}
    public static class Impl1938 implements DemoService {public int type() {return 1938;} public void run() {System.out.println(1938);}}
    public static class Impl1939 implements DemoService {public int type() {return 1939;} public void run() {System.out.println(1939);}}
    public static class Impl1940 implements DemoService {public int type() {return 1940;} public void run() {System.out.println(1940);}}
    public static class Impl1941 implements DemoService {public int type() {return 1941;} public void run() {System.out.println(1941);}}
    public static class Impl1942 implements DemoService {public int type() {return 1942;} public void run() {System.out.println(1942);}}
    public static class Impl1943 implements DemoService {public int type() {return 1943;} public void run() {System.out.println(1943);}}
    public static class Impl1944 implements DemoService {public int type() {return 1944;} public void run() {System.out.println(1944);}}
    public static class Impl1945 implements DemoService {public int type() {return 1945;} public void run() {System.out.println(1945);}}
    public static class Impl1946 implements DemoService {public int type() {return 1946;} public void run() {System.out.println(1946);}}
    public static class Impl1947 implements DemoService {public int type() {return 1947;} public void run() {System.out.println(1947);}}
    public static class Impl1948 implements DemoService {public int type() {return 1948;} public void run() {System.out.println(1948);}}
    public static class Impl1949 implements DemoService {public int type() {return 1949;} public void run() {System.out.println(1949);}}
    public static class Impl1950 implements DemoService {public int type() {return 1950;} public void run() {System.out.println(1950);}}
    public static class Impl1951 implements DemoService {public int type() {return 1951;} public void run() {System.out.println(1951);}}
    public static class Impl1952 implements DemoService {public int type() {return 1952;} public void run() {System.out.println(1952);}}
    public static class Impl1953 implements DemoService {public int type() {return 1953;} public void run() {System.out.println(1953);}}
    public static class Impl1954 implements DemoService {public int type() {return 1954;} public void run() {System.out.println(1954);}}
    public static class Impl1955 implements DemoService {public int type() {return 1955;} public void run() {System.out.println(1955);}}
    public static class Impl1956 implements DemoService {public int type() {return 1956;} public void run() {System.out.println(1956);}}
    public static class Impl1957 implements DemoService {public int type() {return 1957;} public void run() {System.out.println(1957);}}
    public static class Impl1958 implements DemoService {public int type() {return 1958;} public void run() {System.out.println(1958);}}
    public static class Impl1959 implements DemoService {public int type() {return 1959;} public void run() {System.out.println(1959);}}
    public static class Impl1960 implements DemoService {public int type() {return 1960;} public void run() {System.out.println(1960);}}
    public static class Impl1961 implements DemoService {public int type() {return 1961;} public void run() {System.out.println(1961);}}
    public static class Impl1962 implements DemoService {public int type() {return 1962;} public void run() {System.out.println(1962);}}
    public static class Impl1963 implements DemoService {public int type() {return 1963;} public void run() {System.out.println(1963);}}
    public static class Impl1964 implements DemoService {public int type() {return 1964;} public void run() {System.out.println(1964);}}
    public static class Impl1965 implements DemoService {public int type() {return 1965;} public void run() {System.out.println(1965);}}
    public static class Impl1966 implements DemoService {public int type() {return 1966;} public void run() {System.out.println(1966);}}
    public static class Impl1967 implements DemoService {public int type() {return 1967;} public void run() {System.out.println(1967);}}
    public static class Impl1968 implements DemoService {public int type() {return 1968;} public void run() {System.out.println(1968);}}
    public static class Impl1969 implements DemoService {public int type() {return 1969;} public void run() {System.out.println(1969);}}
    public static class Impl1970 implements DemoService {public int type() {return 1970;} public void run() {System.out.println(1970);}}
    public static class Impl1971 implements DemoService {public int type() {return 1971;} public void run() {System.out.println(1971);}}
    public static class Impl1972 implements DemoService {public int type() {return 1972;} public void run() {System.out.println(1972);}}
    public static class Impl1973 implements DemoService {public int type() {return 1973;} public void run() {System.out.println(1973);}}
    public static class Impl1974 implements DemoService {public int type() {return 1974;} public void run() {System.out.println(1974);}}
    public static class Impl1975 implements DemoService {public int type() {return 1975;} public void run() {System.out.println(1975);}}
    public static class Impl1976 implements DemoService {public int type() {return 1976;} public void run() {System.out.println(1976);}}
    public static class Impl1977 implements DemoService {public int type() {return 1977;} public void run() {System.out.println(1977);}}
    public static class Impl1978 implements DemoService {public int type() {return 1978;} public void run() {System.out.println(1978);}}
    public static class Impl1979 implements DemoService {public int type() {return 1979;} public void run() {System.out.println(1979);}}
    public static class Impl1980 implements DemoService {public int type() {return 1980;} public void run() {System.out.println(1980);}}
    public static class Impl1981 implements DemoService {public int type() {return 1981;} public void run() {System.out.println(1981);}}
    public static class Impl1982 implements DemoService {public int type() {return 1982;} public void run() {System.out.println(1982);}}
    public static class Impl1983 implements DemoService {public int type() {return 1983;} public void run() {System.out.println(1983);}}
    public static class Impl1984 implements DemoService {public int type() {return 1984;} public void run() {System.out.println(1984);}}
    public static class Impl1985 implements DemoService {public int type() {return 1985;} public void run() {System.out.println(1985);}}
    public static class Impl1986 implements DemoService {public int type() {return 1986;} public void run() {System.out.println(1986);}}
    public static class Impl1987 implements DemoService {public int type() {return 1987;} public void run() {System.out.println(1987);}}
    public static class Impl1988 implements DemoService {public int type() {return 1988;} public void run() {System.out.println(1988);}}
    public static class Impl1989 implements DemoService {public int type() {return 1989;} public void run() {System.out.println(1989);}}
    public static class Impl1990 implements DemoService {public int type() {return 1990;} public void run() {System.out.println(1990);}}
    public static class Impl1991 implements DemoService {public int type() {return 1991;} public void run() {System.out.println(1991);}}
    public static class Impl1992 implements DemoService {public int type() {return 1992;} public void run() {System.out.println(1992);}}
    public static class Impl1993 implements DemoService {public int type() {return 1993;} public void run() {System.out.println(1993);}}
    public static class Impl1994 implements DemoService {public int type() {return 1994;} public void run() {System.out.println(1994);}}
    public static class Impl1995 implements DemoService {public int type() {return 1995;} public void run() {System.out.println(1995);}}
    public static class Impl1996 implements DemoService {public int type() {return 1996;} public void run() {System.out.println(1996);}}
    public static class Impl1997 implements DemoService {public int type() {return 1997;} public void run() {System.out.println(1997);}}
    public static class Impl1998 implements DemoService {public int type() {return 1998;} public void run() {System.out.println(1998);}}
    public static class Impl1999 implements DemoService {public int type() {return 1999;} public void run() {System.out.println(1999);}}
    public static class Impl2000 implements DemoService {public int type() {return 2000;} public void run() {System.out.println(2000);}}
    public static class Impl2001 implements DemoService {public int type() {return 2001;} public void run() {System.out.println(2001);}}
    public static class Impl2002 implements DemoService {public int type() {return 2002;} public void run() {System.out.println(2002);}}
    public static class Impl2003 implements DemoService {public int type() {return 2003;} public void run() {System.out.println(2003);}}
    public static class Impl2004 implements DemoService {public int type() {return 2004;} public void run() {System.out.println(2004);}}
    public static class Impl2005 implements DemoService {public int type() {return 2005;} public void run() {System.out.println(2005);}}
    public static class Impl2006 implements DemoService {public int type() {return 2006;} public void run() {System.out.println(2006);}}
    public static class Impl2007 implements DemoService {public int type() {return 2007;} public void run() {System.out.println(2007);}}
    public static class Impl2008 implements DemoService {public int type() {return 2008;} public void run() {System.out.println(2008);}}
    public static class Impl2009 implements DemoService {public int type() {return 2009;} public void run() {System.out.println(2009);}}
    public static class Impl2010 implements DemoService {public int type() {return 2010;} public void run() {System.out.println(2010);}}
    public static class Impl2011 implements DemoService {public int type() {return 2011;} public void run() {System.out.println(2011);}}
    public static class Impl2012 implements DemoService {public int type() {return 2012;} public void run() {System.out.println(2012);}}
    public static class Impl2013 implements DemoService {public int type() {return 2013;} public void run() {System.out.println(2013);}}
    public static class Impl2014 implements DemoService {public int type() {return 2014;} public void run() {System.out.println(2014);}}
    public static class Impl2015 implements DemoService {public int type() {return 2015;} public void run() {System.out.println(2015);}}
    public static class Impl2016 implements DemoService {public int type() {return 2016;} public void run() {System.out.println(2016);}}
    public static class Impl2017 implements DemoService {public int type() {return 2017;} public void run() {System.out.println(2017);}}
    public static class Impl2018 implements DemoService {public int type() {return 2018;} public void run() {System.out.println(2018);}}
    public static class Impl2019 implements DemoService {public int type() {return 2019;} public void run() {System.out.println(2019);}}
    public static class Impl2020 implements DemoService {public int type() {return 2020;} public void run() {System.out.println(2020);}}
    public static class Impl2021 implements DemoService {public int type() {return 2021;} public void run() {System.out.println(2021);}}
    public static class Impl2022 implements DemoService {public int type() {return 2022;} public void run() {System.out.println(2022);}}
    public static class Impl2023 implements DemoService {public int type() {return 2023;} public void run() {System.out.println(2023);}}
    public static class Impl2024 implements DemoService {public int type() {return 2024;} public void run() {System.out.println(2024);}}
    public static class Impl2025 implements DemoService {public int type() {return 2025;} public void run() {System.out.println(2025);}}
    public static class Impl2026 implements DemoService {public int type() {return 2026;} public void run() {System.out.println(2026);}}
    public static class Impl2027 implements DemoService {public int type() {return 2027;} public void run() {System.out.println(2027);}}
    public static class Impl2028 implements DemoService {public int type() {return 2028;} public void run() {System.out.println(2028);}}
    public static class Impl2029 implements DemoService {public int type() {return 2029;} public void run() {System.out.println(2029);}}
    public static class Impl2030 implements DemoService {public int type() {return 2030;} public void run() {System.out.println(2030);}}
    public static class Impl2031 implements DemoService {public int type() {return 2031;} public void run() {System.out.println(2031);}}
    public static class Impl2032 implements DemoService {public int type() {return 2032;} public void run() {System.out.println(2032);}}
    public static class Impl2033 implements DemoService {public int type() {return 2033;} public void run() {System.out.println(2033);}}
    public static class Impl2034 implements DemoService {public int type() {return 2034;} public void run() {System.out.println(2034);}}
    public static class Impl2035 implements DemoService {public int type() {return 2035;} public void run() {System.out.println(2035);}}
    public static class Impl2036 implements DemoService {public int type() {return 2036;} public void run() {System.out.println(2036);}}
    public static class Impl2037 implements DemoService {public int type() {return 2037;} public void run() {System.out.println(2037);}}
    public static class Impl2038 implements DemoService {public int type() {return 2038;} public void run() {System.out.println(2038);}}
    public static class Impl2039 implements DemoService {public int type() {return 2039;} public void run() {System.out.println(2039);}}
    public static class Impl2040 implements DemoService {public int type() {return 2040;} public void run() {System.out.println(2040);}}
    public static class Impl2041 implements DemoService {public int type() {return 2041;} public void run() {System.out.println(2041);}}
    public static class Impl2042 implements DemoService {public int type() {return 2042;} public void run() {System.out.println(2042);}}
    public static class Impl2043 implements DemoService {public int type() {return 2043;} public void run() {System.out.println(2043);}}
    public static class Impl2044 implements DemoService {public int type() {return 2044;} public void run() {System.out.println(2044);}}
    public static class Impl2045 implements DemoService {public int type() {return 2045;} public void run() {System.out.println(2045);}}
    public static class Impl2046 implements DemoService {public int type() {return 2046;} public void run() {System.out.println(2046);}}
    public static class Impl2047 implements DemoService {public int type() {return 2047;} public void run() {System.out.println(2047);}}
    public static class Impl2048 implements DemoService {public int type() {return 2048;} public void run() {System.out.println(2048);}}
    public static class Impl2049 implements DemoService {public int type() {return 2049;} public void run() {System.out.println(2049);}}
    public static class Impl2050 implements DemoService {public int type() {return 2050;} public void run() {System.out.println(2050);}}
    public static class Impl2051 implements DemoService {public int type() {return 2051;} public void run() {System.out.println(2051);}}
    public static class Impl2052 implements DemoService {public int type() {return 2052;} public void run() {System.out.println(2052);}}
    public static class Impl2053 implements DemoService {public int type() {return 2053;} public void run() {System.out.println(2053);}}
    public static class Impl2054 implements DemoService {public int type() {return 2054;} public void run() {System.out.println(2054);}}
    public static class Impl2055 implements DemoService {public int type() {return 2055;} public void run() {System.out.println(2055);}}
    public static class Impl2056 implements DemoService {public int type() {return 2056;} public void run() {System.out.println(2056);}}
    public static class Impl2057 implements DemoService {public int type() {return 2057;} public void run() {System.out.println(2057);}}
    public static class Impl2058 implements DemoService {public int type() {return 2058;} public void run() {System.out.println(2058);}}
    public static class Impl2059 implements DemoService {public int type() {return 2059;} public void run() {System.out.println(2059);}}
    public static class Impl2060 implements DemoService {public int type() {return 2060;} public void run() {System.out.println(2060);}}
    public static class Impl2061 implements DemoService {public int type() {return 2061;} public void run() {System.out.println(2061);}}
    public static class Impl2062 implements DemoService {public int type() {return 2062;} public void run() {System.out.println(2062);}}
    public static class Impl2063 implements DemoService {public int type() {return 2063;} public void run() {System.out.println(2063);}}
    public static class Impl2064 implements DemoService {public int type() {return 2064;} public void run() {System.out.println(2064);}}
    public static class Impl2065 implements DemoService {public int type() {return 2065;} public void run() {System.out.println(2065);}}
    public static class Impl2066 implements DemoService {public int type() {return 2066;} public void run() {System.out.println(2066);}}
    public static class Impl2067 implements DemoService {public int type() {return 2067;} public void run() {System.out.println(2067);}}
    public static class Impl2068 implements DemoService {public int type() {return 2068;} public void run() {System.out.println(2068);}}
    public static class Impl2069 implements DemoService {public int type() {return 2069;} public void run() {System.out.println(2069);}}
    public static class Impl2070 implements DemoService {public int type() {return 2070;} public void run() {System.out.println(2070);}}
    public static class Impl2071 implements DemoService {public int type() {return 2071;} public void run() {System.out.println(2071);}}
    public static class Impl2072 implements DemoService {public int type() {return 2072;} public void run() {System.out.println(2072);}}
    public static class Impl2073 implements DemoService {public int type() {return 2073;} public void run() {System.out.println(2073);}}
    public static class Impl2074 implements DemoService {public int type() {return 2074;} public void run() {System.out.println(2074);}}
    public static class Impl2075 implements DemoService {public int type() {return 2075;} public void run() {System.out.println(2075);}}
    public static class Impl2076 implements DemoService {public int type() {return 2076;} public void run() {System.out.println(2076);}}
    public static class Impl2077 implements DemoService {public int type() {return 2077;} public void run() {System.out.println(2077);}}
    public static class Impl2078 implements DemoService {public int type() {return 2078;} public void run() {System.out.println(2078);}}
    public static class Impl2079 implements DemoService {public int type() {return 2079;} public void run() {System.out.println(2079);}}
    public static class Impl2080 implements DemoService {public int type() {return 2080;} public void run() {System.out.println(2080);}}
    public static class Impl2081 implements DemoService {public int type() {return 2081;} public void run() {System.out.println(2081);}}
    public static class Impl2082 implements DemoService {public int type() {return 2082;} public void run() {System.out.println(2082);}}
    public static class Impl2083 implements DemoService {public int type() {return 2083;} public void run() {System.out.println(2083);}}
    public static class Impl2084 implements DemoService {public int type() {return 2084;} public void run() {System.out.println(2084);}}
    public static class Impl2085 implements DemoService {public int type() {return 2085;} public void run() {System.out.println(2085);}}
    public static class Impl2086 implements DemoService {public int type() {return 2086;} public void run() {System.out.println(2086);}}
    public static class Impl2087 implements DemoService {public int type() {return 2087;} public void run() {System.out.println(2087);}}
    public static class Impl2088 implements DemoService {public int type() {return 2088;} public void run() {System.out.println(2088);}}
    public static class Impl2089 implements DemoService {public int type() {return 2089;} public void run() {System.out.println(2089);}}
    public static class Impl2090 implements DemoService {public int type() {return 2090;} public void run() {System.out.println(2090);}}
    public static class Impl2091 implements DemoService {public int type() {return 2091;} public void run() {System.out.println(2091);}}
    public static class Impl2092 implements DemoService {public int type() {return 2092;} public void run() {System.out.println(2092);}}
    public static class Impl2093 implements DemoService {public int type() {return 2093;} public void run() {System.out.println(2093);}}
    public static class Impl2094 implements DemoService {public int type() {return 2094;} public void run() {System.out.println(2094);}}
    public static class Impl2095 implements DemoService {public int type() {return 2095;} public void run() {System.out.println(2095);}}
    public static class Impl2096 implements DemoService {public int type() {return 2096;} public void run() {System.out.println(2096);}}
    public static class Impl2097 implements DemoService {public int type() {return 2097;} public void run() {System.out.println(2097);}}
    public static class Impl2098 implements DemoService {public int type() {return 2098;} public void run() {System.out.println(2098);}}
    public static class Impl2099 implements DemoService {public int type() {return 2099;} public void run() {System.out.println(2099);}}
    public static class Impl2100 implements DemoService {public int type() {return 2100;} public void run() {System.out.println(2100);}}
    public static class Impl2101 implements DemoService {public int type() {return 2101;} public void run() {System.out.println(2101);}}
    public static class Impl2102 implements DemoService {public int type() {return 2102;} public void run() {System.out.println(2102);}}
    public static class Impl2103 implements DemoService {public int type() {return 2103;} public void run() {System.out.println(2103);}}
    public static class Impl2104 implements DemoService {public int type() {return 2104;} public void run() {System.out.println(2104);}}
    public static class Impl2105 implements DemoService {public int type() {return 2105;} public void run() {System.out.println(2105);}}
    public static class Impl2106 implements DemoService {public int type() {return 2106;} public void run() {System.out.println(2106);}}
    public static class Impl2107 implements DemoService {public int type() {return 2107;} public void run() {System.out.println(2107);}}
    public static class Impl2108 implements DemoService {public int type() {return 2108;} public void run() {System.out.println(2108);}}
    public static class Impl2109 implements DemoService {public int type() {return 2109;} public void run() {System.out.println(2109);}}
    public static class Impl2110 implements DemoService {public int type() {return 2110;} public void run() {System.out.println(2110);}}
    public static class Impl2111 implements DemoService {public int type() {return 2111;} public void run() {System.out.println(2111);}}
    public static class Impl2112 implements DemoService {public int type() {return 2112;} public void run() {System.out.println(2112);}}
    public static class Impl2113 implements DemoService {public int type() {return 2113;} public void run() {System.out.println(2113);}}
    public static class Impl2114 implements DemoService {public int type() {return 2114;} public void run() {System.out.println(2114);}}
    public static class Impl2115 implements DemoService {public int type() {return 2115;} public void run() {System.out.println(2115);}}
    public static class Impl2116 implements DemoService {public int type() {return 2116;} public void run() {System.out.println(2116);}}
    public static class Impl2117 implements DemoService {public int type() {return 2117;} public void run() {System.out.println(2117);}}
    public static class Impl2118 implements DemoService {public int type() {return 2118;} public void run() {System.out.println(2118);}}
    public static class Impl2119 implements DemoService {public int type() {return 2119;} public void run() {System.out.println(2119);}}
    public static class Impl2120 implements DemoService {public int type() {return 2120;} public void run() {System.out.println(2120);}}
    public static class Impl2121 implements DemoService {public int type() {return 2121;} public void run() {System.out.println(2121);}}
    public static class Impl2122 implements DemoService {public int type() {return 2122;} public void run() {System.out.println(2122);}}
    public static class Impl2123 implements DemoService {public int type() {return 2123;} public void run() {System.out.println(2123);}}
    public static class Impl2124 implements DemoService {public int type() {return 2124;} public void run() {System.out.println(2124);}}
    public static class Impl2125 implements DemoService {public int type() {return 2125;} public void run() {System.out.println(2125);}}
    public static class Impl2126 implements DemoService {public int type() {return 2126;} public void run() {System.out.println(2126);}}
    public static class Impl2127 implements DemoService {public int type() {return 2127;} public void run() {System.out.println(2127);}}
    public static class Impl2128 implements DemoService {public int type() {return 2128;} public void run() {System.out.println(2128);}}
    public static class Impl2129 implements DemoService {public int type() {return 2129;} public void run() {System.out.println(2129);}}
    public static class Impl2130 implements DemoService {public int type() {return 2130;} public void run() {System.out.println(2130);}}
    public static class Impl2131 implements DemoService {public int type() {return 2131;} public void run() {System.out.println(2131);}}
    public static class Impl2132 implements DemoService {public int type() {return 2132;} public void run() {System.out.println(2132);}}
    public static class Impl2133 implements DemoService {public int type() {return 2133;} public void run() {System.out.println(2133);}}
    public static class Impl2134 implements DemoService {public int type() {return 2134;} public void run() {System.out.println(2134);}}
    public static class Impl2135 implements DemoService {public int type() {return 2135;} public void run() {System.out.println(2135);}}
    public static class Impl2136 implements DemoService {public int type() {return 2136;} public void run() {System.out.println(2136);}}
    public static class Impl2137 implements DemoService {public int type() {return 2137;} public void run() {System.out.println(2137);}}
    public static class Impl2138 implements DemoService {public int type() {return 2138;} public void run() {System.out.println(2138);}}
    public static class Impl2139 implements DemoService {public int type() {return 2139;} public void run() {System.out.println(2139);}}
    public static class Impl2140 implements DemoService {public int type() {return 2140;} public void run() {System.out.println(2140);}}
    public static class Impl2141 implements DemoService {public int type() {return 2141;} public void run() {System.out.println(2141);}}
    public static class Impl2142 implements DemoService {public int type() {return 2142;} public void run() {System.out.println(2142);}}
    public static class Impl2143 implements DemoService {public int type() {return 2143;} public void run() {System.out.println(2143);}}
    public static class Impl2144 implements DemoService {public int type() {return 2144;} public void run() {System.out.println(2144);}}
    public static class Impl2145 implements DemoService {public int type() {return 2145;} public void run() {System.out.println(2145);}}
    public static class Impl2146 implements DemoService {public int type() {return 2146;} public void run() {System.out.println(2146);}}
    public static class Impl2147 implements DemoService {public int type() {return 2147;} public void run() {System.out.println(2147);}}
    public static class Impl2148 implements DemoService {public int type() {return 2148;} public void run() {System.out.println(2148);}}
    public static class Impl2149 implements DemoService {public int type() {return 2149;} public void run() {System.out.println(2149);}}
    public static class Impl2150 implements DemoService {public int type() {return 2150;} public void run() {System.out.println(2150);}}
    public static class Impl2151 implements DemoService {public int type() {return 2151;} public void run() {System.out.println(2151);}}
    public static class Impl2152 implements DemoService {public int type() {return 2152;} public void run() {System.out.println(2152);}}
    public static class Impl2153 implements DemoService {public int type() {return 2153;} public void run() {System.out.println(2153);}}
    public static class Impl2154 implements DemoService {public int type() {return 2154;} public void run() {System.out.println(2154);}}
    public static class Impl2155 implements DemoService {public int type() {return 2155;} public void run() {System.out.println(2155);}}
    public static class Impl2156 implements DemoService {public int type() {return 2156;} public void run() {System.out.println(2156);}}
    public static class Impl2157 implements DemoService {public int type() {return 2157;} public void run() {System.out.println(2157);}}
    public static class Impl2158 implements DemoService {public int type() {return 2158;} public void run() {System.out.println(2158);}}
    public static class Impl2159 implements DemoService {public int type() {return 2159;} public void run() {System.out.println(2159);}}
    public static class Impl2160 implements DemoService {public int type() {return 2160;} public void run() {System.out.println(2160);}}
    public static class Impl2161 implements DemoService {public int type() {return 2161;} public void run() {System.out.println(2161);}}
    public static class Impl2162 implements DemoService {public int type() {return 2162;} public void run() {System.out.println(2162);}}
    public static class Impl2163 implements DemoService {public int type() {return 2163;} public void run() {System.out.println(2163);}}
    public static class Impl2164 implements DemoService {public int type() {return 2164;} public void run() {System.out.println(2164);}}
    public static class Impl2165 implements DemoService {public int type() {return 2165;} public void run() {System.out.println(2165);}}
    public static class Impl2166 implements DemoService {public int type() {return 2166;} public void run() {System.out.println(2166);}}
    public static class Impl2167 implements DemoService {public int type() {return 2167;} public void run() {System.out.println(2167);}}
    public static class Impl2168 implements DemoService {public int type() {return 2168;} public void run() {System.out.println(2168);}}
    public static class Impl2169 implements DemoService {public int type() {return 2169;} public void run() {System.out.println(2169);}}
    public static class Impl2170 implements DemoService {public int type() {return 2170;} public void run() {System.out.println(2170);}}
    public static class Impl2171 implements DemoService {public int type() {return 2171;} public void run() {System.out.println(2171);}}
    public static class Impl2172 implements DemoService {public int type() {return 2172;} public void run() {System.out.println(2172);}}
    public static class Impl2173 implements DemoService {public int type() {return 2173;} public void run() {System.out.println(2173);}}
    public static class Impl2174 implements DemoService {public int type() {return 2174;} public void run() {System.out.println(2174);}}
    public static class Impl2175 implements DemoService {public int type() {return 2175;} public void run() {System.out.println(2175);}}
    public static class Impl2176 implements DemoService {public int type() {return 2176;} public void run() {System.out.println(2176);}}
    public static class Impl2177 implements DemoService {public int type() {return 2177;} public void run() {System.out.println(2177);}}
    public static class Impl2178 implements DemoService {public int type() {return 2178;} public void run() {System.out.println(2178);}}
    public static class Impl2179 implements DemoService {public int type() {return 2179;} public void run() {System.out.println(2179);}}
    public static class Impl2180 implements DemoService {public int type() {return 2180;} public void run() {System.out.println(2180);}}
    public static class Impl2181 implements DemoService {public int type() {return 2181;} public void run() {System.out.println(2181);}}
    public static class Impl2182 implements DemoService {public int type() {return 2182;} public void run() {System.out.println(2182);}}
    public static class Impl2183 implements DemoService {public int type() {return 2183;} public void run() {System.out.println(2183);}}
    public static class Impl2184 implements DemoService {public int type() {return 2184;} public void run() {System.out.println(2184);}}
    public static class Impl2185 implements DemoService {public int type() {return 2185;} public void run() {System.out.println(2185);}}
    public static class Impl2186 implements DemoService {public int type() {return 2186;} public void run() {System.out.println(2186);}}
    public static class Impl2187 implements DemoService {public int type() {return 2187;} public void run() {System.out.println(2187);}}
    public static class Impl2188 implements DemoService {public int type() {return 2188;} public void run() {System.out.println(2188);}}
    public static class Impl2189 implements DemoService {public int type() {return 2189;} public void run() {System.out.println(2189);}}
    public static class Impl2190 implements DemoService {public int type() {return 2190;} public void run() {System.out.println(2190);}}
    public static class Impl2191 implements DemoService {public int type() {return 2191;} public void run() {System.out.println(2191);}}
    public static class Impl2192 implements DemoService {public int type() {return 2192;} public void run() {System.out.println(2192);}}
    public static class Impl2193 implements DemoService {public int type() {return 2193;} public void run() {System.out.println(2193);}}
    public static class Impl2194 implements DemoService {public int type() {return 2194;} public void run() {System.out.println(2194);}}
    public static class Impl2195 implements DemoService {public int type() {return 2195;} public void run() {System.out.println(2195);}}
    public static class Impl2196 implements DemoService {public int type() {return 2196;} public void run() {System.out.println(2196);}}
    public static class Impl2197 implements DemoService {public int type() {return 2197;} public void run() {System.out.println(2197);}}
    public static class Impl2198 implements DemoService {public int type() {return 2198;} public void run() {System.out.println(2198);}}
    public static class Impl2199 implements DemoService {public int type() {return 2199;} public void run() {System.out.println(2199);}}
    public static class Impl2200 implements DemoService {public int type() {return 2200;} public void run() {System.out.println(2200);}}
    public static class Impl2201 implements DemoService {public int type() {return 2201;} public void run() {System.out.println(2201);}}
    public static class Impl2202 implements DemoService {public int type() {return 2202;} public void run() {System.out.println(2202);}}
    public static class Impl2203 implements DemoService {public int type() {return 2203;} public void run() {System.out.println(2203);}}
    public static class Impl2204 implements DemoService {public int type() {return 2204;} public void run() {System.out.println(2204);}}
    public static class Impl2205 implements DemoService {public int type() {return 2205;} public void run() {System.out.println(2205);}}
    public static class Impl2206 implements DemoService {public int type() {return 2206;} public void run() {System.out.println(2206);}}
    public static class Impl2207 implements DemoService {public int type() {return 2207;} public void run() {System.out.println(2207);}}
    public static class Impl2208 implements DemoService {public int type() {return 2208;} public void run() {System.out.println(2208);}}
    public static class Impl2209 implements DemoService {public int type() {return 2209;} public void run() {System.out.println(2209);}}
    public static class Impl2210 implements DemoService {public int type() {return 2210;} public void run() {System.out.println(2210);}}
    public static class Impl2211 implements DemoService {public int type() {return 2211;} public void run() {System.out.println(2211);}}
    public static class Impl2212 implements DemoService {public int type() {return 2212;} public void run() {System.out.println(2212);}}
    public static class Impl2213 implements DemoService {public int type() {return 2213;} public void run() {System.out.println(2213);}}
    public static class Impl2214 implements DemoService {public int type() {return 2214;} public void run() {System.out.println(2214);}}
    public static class Impl2215 implements DemoService {public int type() {return 2215;} public void run() {System.out.println(2215);}}
    public static class Impl2216 implements DemoService {public int type() {return 2216;} public void run() {System.out.println(2216);}}
    public static class Impl2217 implements DemoService {public int type() {return 2217;} public void run() {System.out.println(2217);}}
    public static class Impl2218 implements DemoService {public int type() {return 2218;} public void run() {System.out.println(2218);}}
    public static class Impl2219 implements DemoService {public int type() {return 2219;} public void run() {System.out.println(2219);}}
    public static class Impl2220 implements DemoService {public int type() {return 2220;} public void run() {System.out.println(2220);}}
    public static class Impl2221 implements DemoService {public int type() {return 2221;} public void run() {System.out.println(2221);}}
    public static class Impl2222 implements DemoService {public int type() {return 2222;} public void run() {System.out.println(2222);}}
    public static class Impl2223 implements DemoService {public int type() {return 2223;} public void run() {System.out.println(2223);}}
    public static class Impl2224 implements DemoService {public int type() {return 2224;} public void run() {System.out.println(2224);}}
    public static class Impl2225 implements DemoService {public int type() {return 2225;} public void run() {System.out.println(2225);}}
    public static class Impl2226 implements DemoService {public int type() {return 2226;} public void run() {System.out.println(2226);}}
    public static class Impl2227 implements DemoService {public int type() {return 2227;} public void run() {System.out.println(2227);}}
    public static class Impl2228 implements DemoService {public int type() {return 2228;} public void run() {System.out.println(2228);}}
    public static class Impl2229 implements DemoService {public int type() {return 2229;} public void run() {System.out.println(2229);}}
    public static class Impl2230 implements DemoService {public int type() {return 2230;} public void run() {System.out.println(2230);}}
    public static class Impl2231 implements DemoService {public int type() {return 2231;} public void run() {System.out.println(2231);}}
    public static class Impl2232 implements DemoService {public int type() {return 2232;} public void run() {System.out.println(2232);}}
    public static class Impl2233 implements DemoService {public int type() {return 2233;} public void run() {System.out.println(2233);}}
    public static class Impl2234 implements DemoService {public int type() {return 2234;} public void run() {System.out.println(2234);}}
    public static class Impl2235 implements DemoService {public int type() {return 2235;} public void run() {System.out.println(2235);}}
    public static class Impl2236 implements DemoService {public int type() {return 2236;} public void run() {System.out.println(2236);}}
    public static class Impl2237 implements DemoService {public int type() {return 2237;} public void run() {System.out.println(2237);}}
    public static class Impl2238 implements DemoService {public int type() {return 2238;} public void run() {System.out.println(2238);}}
    public static class Impl2239 implements DemoService {public int type() {return 2239;} public void run() {System.out.println(2239);}}
    public static class Impl2240 implements DemoService {public int type() {return 2240;} public void run() {System.out.println(2240);}}
    public static class Impl2241 implements DemoService {public int type() {return 2241;} public void run() {System.out.println(2241);}}
    public static class Impl2242 implements DemoService {public int type() {return 2242;} public void run() {System.out.println(2242);}}
    public static class Impl2243 implements DemoService {public int type() {return 2243;} public void run() {System.out.println(2243);}}
    public static class Impl2244 implements DemoService {public int type() {return 2244;} public void run() {System.out.println(2244);}}
    public static class Impl2245 implements DemoService {public int type() {return 2245;} public void run() {System.out.println(2245);}}
    public static class Impl2246 implements DemoService {public int type() {return 2246;} public void run() {System.out.println(2246);}}
    public static class Impl2247 implements DemoService {public int type() {return 2247;} public void run() {System.out.println(2247);}}
    public static class Impl2248 implements DemoService {public int type() {return 2248;} public void run() {System.out.println(2248);}}
    public static class Impl2249 implements DemoService {public int type() {return 2249;} public void run() {System.out.println(2249);}}
    public static class Impl2250 implements DemoService {public int type() {return 2250;} public void run() {System.out.println(2250);}}
    public static class Impl2251 implements DemoService {public int type() {return 2251;} public void run() {System.out.println(2251);}}
    public static class Impl2252 implements DemoService {public int type() {return 2252;} public void run() {System.out.println(2252);}}
    public static class Impl2253 implements DemoService {public int type() {return 2253;} public void run() {System.out.println(2253);}}
    public static class Impl2254 implements DemoService {public int type() {return 2254;} public void run() {System.out.println(2254);}}
    public static class Impl2255 implements DemoService {public int type() {return 2255;} public void run() {System.out.println(2255);}}
    public static class Impl2256 implements DemoService {public int type() {return 2256;} public void run() {System.out.println(2256);}}
    public static class Impl2257 implements DemoService {public int type() {return 2257;} public void run() {System.out.println(2257);}}
    public static class Impl2258 implements DemoService {public int type() {return 2258;} public void run() {System.out.println(2258);}}
    public static class Impl2259 implements DemoService {public int type() {return 2259;} public void run() {System.out.println(2259);}}
    public static class Impl2260 implements DemoService {public int type() {return 2260;} public void run() {System.out.println(2260);}}
    public static class Impl2261 implements DemoService {public int type() {return 2261;} public void run() {System.out.println(2261);}}
    public static class Impl2262 implements DemoService {public int type() {return 2262;} public void run() {System.out.println(2262);}}
    public static class Impl2263 implements DemoService {public int type() {return 2263;} public void run() {System.out.println(2263);}}
    public static class Impl2264 implements DemoService {public int type() {return 2264;} public void run() {System.out.println(2264);}}
    public static class Impl2265 implements DemoService {public int type() {return 2265;} public void run() {System.out.println(2265);}}
    public static class Impl2266 implements DemoService {public int type() {return 2266;} public void run() {System.out.println(2266);}}
    public static class Impl2267 implements DemoService {public int type() {return 2267;} public void run() {System.out.println(2267);}}
    public static class Impl2268 implements DemoService {public int type() {return 2268;} public void run() {System.out.println(2268);}}
    public static class Impl2269 implements DemoService {public int type() {return 2269;} public void run() {System.out.println(2269);}}
    public static class Impl2270 implements DemoService {public int type() {return 2270;} public void run() {System.out.println(2270);}}
    public static class Impl2271 implements DemoService {public int type() {return 2271;} public void run() {System.out.println(2271);}}
    public static class Impl2272 implements DemoService {public int type() {return 2272;} public void run() {System.out.println(2272);}}
    public static class Impl2273 implements DemoService {public int type() {return 2273;} public void run() {System.out.println(2273);}}
    public static class Impl2274 implements DemoService {public int type() {return 2274;} public void run() {System.out.println(2274);}}
    public static class Impl2275 implements DemoService {public int type() {return 2275;} public void run() {System.out.println(2275);}}
    public static class Impl2276 implements DemoService {public int type() {return 2276;} public void run() {System.out.println(2276);}}
    public static class Impl2277 implements DemoService {public int type() {return 2277;} public void run() {System.out.println(2277);}}
    public static class Impl2278 implements DemoService {public int type() {return 2278;} public void run() {System.out.println(2278);}}
    public static class Impl2279 implements DemoService {public int type() {return 2279;} public void run() {System.out.println(2279);}}
    public static class Impl2280 implements DemoService {public int type() {return 2280;} public void run() {System.out.println(2280);}}
    public static class Impl2281 implements DemoService {public int type() {return 2281;} public void run() {System.out.println(2281);}}
    public static class Impl2282 implements DemoService {public int type() {return 2282;} public void run() {System.out.println(2282);}}
    public static class Impl2283 implements DemoService {public int type() {return 2283;} public void run() {System.out.println(2283);}}
    public static class Impl2284 implements DemoService {public int type() {return 2284;} public void run() {System.out.println(2284);}}
    public static class Impl2285 implements DemoService {public int type() {return 2285;} public void run() {System.out.println(2285);}}
    public static class Impl2286 implements DemoService {public int type() {return 2286;} public void run() {System.out.println(2286);}}
    public static class Impl2287 implements DemoService {public int type() {return 2287;} public void run() {System.out.println(2287);}}
    public static class Impl2288 implements DemoService {public int type() {return 2288;} public void run() {System.out.println(2288);}}
    public static class Impl2289 implements DemoService {public int type() {return 2289;} public void run() {System.out.println(2289);}}
    public static class Impl2290 implements DemoService {public int type() {return 2290;} public void run() {System.out.println(2290);}}
    public static class Impl2291 implements DemoService {public int type() {return 2291;} public void run() {System.out.println(2291);}}
    public static class Impl2292 implements DemoService {public int type() {return 2292;} public void run() {System.out.println(2292);}}
    public static class Impl2293 implements DemoService {public int type() {return 2293;} public void run() {System.out.println(2293);}}
    public static class Impl2294 implements DemoService {public int type() {return 2294;} public void run() {System.out.println(2294);}}
    public static class Impl2295 implements DemoService {public int type() {return 2295;} public void run() {System.out.println(2295);}}
    public static class Impl2296 implements DemoService {public int type() {return 2296;} public void run() {System.out.println(2296);}}
    public static class Impl2297 implements DemoService {public int type() {return 2297;} public void run() {System.out.println(2297);}}
    public static class Impl2298 implements DemoService {public int type() {return 2298;} public void run() {System.out.println(2298);}}
    public static class Impl2299 implements DemoService {public int type() {return 2299;} public void run() {System.out.println(2299);}}
    public static class Impl2300 implements DemoService {public int type() {return 2300;} public void run() {System.out.println(2300);}}
    public static class Impl2301 implements DemoService {public int type() {return 2301;} public void run() {System.out.println(2301);}}
    public static class Impl2302 implements DemoService {public int type() {return 2302;} public void run() {System.out.println(2302);}}
    public static class Impl2303 implements DemoService {public int type() {return 2303;} public void run() {System.out.println(2303);}}
    public static class Impl2304 implements DemoService {public int type() {return 2304;} public void run() {System.out.println(2304);}}
    public static class Impl2305 implements DemoService {public int type() {return 2305;} public void run() {System.out.println(2305);}}
    public static class Impl2306 implements DemoService {public int type() {return 2306;} public void run() {System.out.println(2306);}}
    public static class Impl2307 implements DemoService {public int type() {return 2307;} public void run() {System.out.println(2307);}}
    public static class Impl2308 implements DemoService {public int type() {return 2308;} public void run() {System.out.println(2308);}}
    public static class Impl2309 implements DemoService {public int type() {return 2309;} public void run() {System.out.println(2309);}}
    public static class Impl2310 implements DemoService {public int type() {return 2310;} public void run() {System.out.println(2310);}}
    public static class Impl2311 implements DemoService {public int type() {return 2311;} public void run() {System.out.println(2311);}}
    public static class Impl2312 implements DemoService {public int type() {return 2312;} public void run() {System.out.println(2312);}}
    public static class Impl2313 implements DemoService {public int type() {return 2313;} public void run() {System.out.println(2313);}}
    public static class Impl2314 implements DemoService {public int type() {return 2314;} public void run() {System.out.println(2314);}}
    public static class Impl2315 implements DemoService {public int type() {return 2315;} public void run() {System.out.println(2315);}}
    public static class Impl2316 implements DemoService {public int type() {return 2316;} public void run() {System.out.println(2316);}}
    public static class Impl2317 implements DemoService {public int type() {return 2317;} public void run() {System.out.println(2317);}}
    public static class Impl2318 implements DemoService {public int type() {return 2318;} public void run() {System.out.println(2318);}}
    public static class Impl2319 implements DemoService {public int type() {return 2319;} public void run() {System.out.println(2319);}}
    public static class Impl2320 implements DemoService {public int type() {return 2320;} public void run() {System.out.println(2320);}}
    public static class Impl2321 implements DemoService {public int type() {return 2321;} public void run() {System.out.println(2321);}}
    public static class Impl2322 implements DemoService {public int type() {return 2322;} public void run() {System.out.println(2322);}}
    public static class Impl2323 implements DemoService {public int type() {return 2323;} public void run() {System.out.println(2323);}}
    public static class Impl2324 implements DemoService {public int type() {return 2324;} public void run() {System.out.println(2324);}}
    public static class Impl2325 implements DemoService {public int type() {return 2325;} public void run() {System.out.println(2325);}}
    public static class Impl2326 implements DemoService {public int type() {return 2326;} public void run() {System.out.println(2326);}}
    public static class Impl2327 implements DemoService {public int type() {return 2327;} public void run() {System.out.println(2327);}}
    public static class Impl2328 implements DemoService {public int type() {return 2328;} public void run() {System.out.println(2328);}}
    public static class Impl2329 implements DemoService {public int type() {return 2329;} public void run() {System.out.println(2329);}}
    public static class Impl2330 implements DemoService {public int type() {return 2330;} public void run() {System.out.println(2330);}}
    public static class Impl2331 implements DemoService {public int type() {return 2331;} public void run() {System.out.println(2331);}}
    public static class Impl2332 implements DemoService {public int type() {return 2332;} public void run() {System.out.println(2332);}}
    public static class Impl2333 implements DemoService {public int type() {return 2333;} public void run() {System.out.println(2333);}}
    public static class Impl2334 implements DemoService {public int type() {return 2334;} public void run() {System.out.println(2334);}}
    public static class Impl2335 implements DemoService {public int type() {return 2335;} public void run() {System.out.println(2335);}}
    public static class Impl2336 implements DemoService {public int type() {return 2336;} public void run() {System.out.println(2336);}}
    public static class Impl2337 implements DemoService {public int type() {return 2337;} public void run() {System.out.println(2337);}}
    public static class Impl2338 implements DemoService {public int type() {return 2338;} public void run() {System.out.println(2338);}}
    public static class Impl2339 implements DemoService {public int type() {return 2339;} public void run() {System.out.println(2339);}}
    public static class Impl2340 implements DemoService {public int type() {return 2340;} public void run() {System.out.println(2340);}}
    public static class Impl2341 implements DemoService {public int type() {return 2341;} public void run() {System.out.println(2341);}}
    public static class Impl2342 implements DemoService {public int type() {return 2342;} public void run() {System.out.println(2342);}}
    public static class Impl2343 implements DemoService {public int type() {return 2343;} public void run() {System.out.println(2343);}}
    public static class Impl2344 implements DemoService {public int type() {return 2344;} public void run() {System.out.println(2344);}}
    public static class Impl2345 implements DemoService {public int type() {return 2345;} public void run() {System.out.println(2345);}}
    public static class Impl2346 implements DemoService {public int type() {return 2346;} public void run() {System.out.println(2346);}}
    public static class Impl2347 implements DemoService {public int type() {return 2347;} public void run() {System.out.println(2347);}}
    public static class Impl2348 implements DemoService {public int type() {return 2348;} public void run() {System.out.println(2348);}}
    public static class Impl2349 implements DemoService {public int type() {return 2349;} public void run() {System.out.println(2349);}}
    public static class Impl2350 implements DemoService {public int type() {return 2350;} public void run() {System.out.println(2350);}}
    public static class Impl2351 implements DemoService {public int type() {return 2351;} public void run() {System.out.println(2351);}}
    public static class Impl2352 implements DemoService {public int type() {return 2352;} public void run() {System.out.println(2352);}}
    public static class Impl2353 implements DemoService {public int type() {return 2353;} public void run() {System.out.println(2353);}}
    public static class Impl2354 implements DemoService {public int type() {return 2354;} public void run() {System.out.println(2354);}}
    public static class Impl2355 implements DemoService {public int type() {return 2355;} public void run() {System.out.println(2355);}}
    public static class Impl2356 implements DemoService {public int type() {return 2356;} public void run() {System.out.println(2356);}}
    public static class Impl2357 implements DemoService {public int type() {return 2357;} public void run() {System.out.println(2357);}}
    public static class Impl2358 implements DemoService {public int type() {return 2358;} public void run() {System.out.println(2358);}}
    public static class Impl2359 implements DemoService {public int type() {return 2359;} public void run() {System.out.println(2359);}}
    public static class Impl2360 implements DemoService {public int type() {return 2360;} public void run() {System.out.println(2360);}}
    public static class Impl2361 implements DemoService {public int type() {return 2361;} public void run() {System.out.println(2361);}}
    public static class Impl2362 implements DemoService {public int type() {return 2362;} public void run() {System.out.println(2362);}}
    public static class Impl2363 implements DemoService {public int type() {return 2363;} public void run() {System.out.println(2363);}}
    public static class Impl2364 implements DemoService {public int type() {return 2364;} public void run() {System.out.println(2364);}}
    public static class Impl2365 implements DemoService {public int type() {return 2365;} public void run() {System.out.println(2365);}}
    public static class Impl2366 implements DemoService {public int type() {return 2366;} public void run() {System.out.println(2366);}}
    public static class Impl2367 implements DemoService {public int type() {return 2367;} public void run() {System.out.println(2367);}}
    public static class Impl2368 implements DemoService {public int type() {return 2368;} public void run() {System.out.println(2368);}}
    public static class Impl2369 implements DemoService {public int type() {return 2369;} public void run() {System.out.println(2369);}}
    public static class Impl2370 implements DemoService {public int type() {return 2370;} public void run() {System.out.println(2370);}}
    public static class Impl2371 implements DemoService {public int type() {return 2371;} public void run() {System.out.println(2371);}}
    public static class Impl2372 implements DemoService {public int type() {return 2372;} public void run() {System.out.println(2372);}}
    public static class Impl2373 implements DemoService {public int type() {return 2373;} public void run() {System.out.println(2373);}}
    public static class Impl2374 implements DemoService {public int type() {return 2374;} public void run() {System.out.println(2374);}}
    public static class Impl2375 implements DemoService {public int type() {return 2375;} public void run() {System.out.println(2375);}}
    public static class Impl2376 implements DemoService {public int type() {return 2376;} public void run() {System.out.println(2376);}}
    public static class Impl2377 implements DemoService {public int type() {return 2377;} public void run() {System.out.println(2377);}}
    public static class Impl2378 implements DemoService {public int type() {return 2378;} public void run() {System.out.println(2378);}}
    public static class Impl2379 implements DemoService {public int type() {return 2379;} public void run() {System.out.println(2379);}}
    public static class Impl2380 implements DemoService {public int type() {return 2380;} public void run() {System.out.println(2380);}}
    public static class Impl2381 implements DemoService {public int type() {return 2381;} public void run() {System.out.println(2381);}}
    public static class Impl2382 implements DemoService {public int type() {return 2382;} public void run() {System.out.println(2382);}}
    public static class Impl2383 implements DemoService {public int type() {return 2383;} public void run() {System.out.println(2383);}}
    public static class Impl2384 implements DemoService {public int type() {return 2384;} public void run() {System.out.println(2384);}}
    public static class Impl2385 implements DemoService {public int type() {return 2385;} public void run() {System.out.println(2385);}}
    public static class Impl2386 implements DemoService {public int type() {return 2386;} public void run() {System.out.println(2386);}}
    public static class Impl2387 implements DemoService {public int type() {return 2387;} public void run() {System.out.println(2387);}}
    public static class Impl2388 implements DemoService {public int type() {return 2388;} public void run() {System.out.println(2388);}}
    public static class Impl2389 implements DemoService {public int type() {return 2389;} public void run() {System.out.println(2389);}}
    public static class Impl2390 implements DemoService {public int type() {return 2390;} public void run() {System.out.println(2390);}}
    public static class Impl2391 implements DemoService {public int type() {return 2391;} public void run() {System.out.println(2391);}}
    public static class Impl2392 implements DemoService {public int type() {return 2392;} public void run() {System.out.println(2392);}}
    public static class Impl2393 implements DemoService {public int type() {return 2393;} public void run() {System.out.println(2393);}}
    public static class Impl2394 implements DemoService {public int type() {return 2394;} public void run() {System.out.println(2394);}}
    public static class Impl2395 implements DemoService {public int type() {return 2395;} public void run() {System.out.println(2395);}}
    public static class Impl2396 implements DemoService {public int type() {return 2396;} public void run() {System.out.println(2396);}}
    public static class Impl2397 implements DemoService {public int type() {return 2397;} public void run() {System.out.println(2397);}}
    public static class Impl2398 implements DemoService {public int type() {return 2398;} public void run() {System.out.println(2398);}}
    public static class Impl2399 implements DemoService {public int type() {return 2399;} public void run() {System.out.println(2399);}}
    public static class Impl2400 implements DemoService {public int type() {return 2400;} public void run() {System.out.println(2400);}}
    public static class Impl2401 implements DemoService {public int type() {return 2401;} public void run() {System.out.println(2401);}}
    public static class Impl2402 implements DemoService {public int type() {return 2402;} public void run() {System.out.println(2402);}}
    public static class Impl2403 implements DemoService {public int type() {return 2403;} public void run() {System.out.println(2403);}}
    public static class Impl2404 implements DemoService {public int type() {return 2404;} public void run() {System.out.println(2404);}}
    public static class Impl2405 implements DemoService {public int type() {return 2405;} public void run() {System.out.println(2405);}}
    public static class Impl2406 implements DemoService {public int type() {return 2406;} public void run() {System.out.println(2406);}}
    public static class Impl2407 implements DemoService {public int type() {return 2407;} public void run() {System.out.println(2407);}}
    public static class Impl2408 implements DemoService {public int type() {return 2408;} public void run() {System.out.println(2408);}}
    public static class Impl2409 implements DemoService {public int type() {return 2409;} public void run() {System.out.println(2409);}}
    public static class Impl2410 implements DemoService {public int type() {return 2410;} public void run() {System.out.println(2410);}}
    public static class Impl2411 implements DemoService {public int type() {return 2411;} public void run() {System.out.println(2411);}}
    public static class Impl2412 implements DemoService {public int type() {return 2412;} public void run() {System.out.println(2412);}}
    public static class Impl2413 implements DemoService {public int type() {return 2413;} public void run() {System.out.println(2413);}}
    public static class Impl2414 implements DemoService {public int type() {return 2414;} public void run() {System.out.println(2414);}}
    public static class Impl2415 implements DemoService {public int type() {return 2415;} public void run() {System.out.println(2415);}}
    public static class Impl2416 implements DemoService {public int type() {return 2416;} public void run() {System.out.println(2416);}}
    public static class Impl2417 implements DemoService {public int type() {return 2417;} public void run() {System.out.println(2417);}}
    public static class Impl2418 implements DemoService {public int type() {return 2418;} public void run() {System.out.println(2418);}}
    public static class Impl2419 implements DemoService {public int type() {return 2419;} public void run() {System.out.println(2419);}}
    public static class Impl2420 implements DemoService {public int type() {return 2420;} public void run() {System.out.println(2420);}}
    public static class Impl2421 implements DemoService {public int type() {return 2421;} public void run() {System.out.println(2421);}}
    public static class Impl2422 implements DemoService {public int type() {return 2422;} public void run() {System.out.println(2422);}}
    public static class Impl2423 implements DemoService {public int type() {return 2423;} public void run() {System.out.println(2423);}}
    public static class Impl2424 implements DemoService {public int type() {return 2424;} public void run() {System.out.println(2424);}}
    public static class Impl2425 implements DemoService {public int type() {return 2425;} public void run() {System.out.println(2425);}}
    public static class Impl2426 implements DemoService {public int type() {return 2426;} public void run() {System.out.println(2426);}}
    public static class Impl2427 implements DemoService {public int type() {return 2427;} public void run() {System.out.println(2427);}}
    public static class Impl2428 implements DemoService {public int type() {return 2428;} public void run() {System.out.println(2428);}}
    public static class Impl2429 implements DemoService {public int type() {return 2429;} public void run() {System.out.println(2429);}}
    public static class Impl2430 implements DemoService {public int type() {return 2430;} public void run() {System.out.println(2430);}}
    public static class Impl2431 implements DemoService {public int type() {return 2431;} public void run() {System.out.println(2431);}}
    public static class Impl2432 implements DemoService {public int type() {return 2432;} public void run() {System.out.println(2432);}}
    public static class Impl2433 implements DemoService {public int type() {return 2433;} public void run() {System.out.println(2433);}}
    public static class Impl2434 implements DemoService {public int type() {return 2434;} public void run() {System.out.println(2434);}}
    public static class Impl2435 implements DemoService {public int type() {return 2435;} public void run() {System.out.println(2435);}}
    public static class Impl2436 implements DemoService {public int type() {return 2436;} public void run() {System.out.println(2436);}}
    public static class Impl2437 implements DemoService {public int type() {return 2437;} public void run() {System.out.println(2437);}}
    public static class Impl2438 implements DemoService {public int type() {return 2438;} public void run() {System.out.println(2438);}}
    public static class Impl2439 implements DemoService {public int type() {return 2439;} public void run() {System.out.println(2439);}}
    public static class Impl2440 implements DemoService {public int type() {return 2440;} public void run() {System.out.println(2440);}}
    public static class Impl2441 implements DemoService {public int type() {return 2441;} public void run() {System.out.println(2441);}}
    public static class Impl2442 implements DemoService {public int type() {return 2442;} public void run() {System.out.println(2442);}}
    public static class Impl2443 implements DemoService {public int type() {return 2443;} public void run() {System.out.println(2443);}}
    public static class Impl2444 implements DemoService {public int type() {return 2444;} public void run() {System.out.println(2444);}}
    public static class Impl2445 implements DemoService {public int type() {return 2445;} public void run() {System.out.println(2445);}}
    public static class Impl2446 implements DemoService {public int type() {return 2446;} public void run() {System.out.println(2446);}}
    public static class Impl2447 implements DemoService {public int type() {return 2447;} public void run() {System.out.println(2447);}}
    public static class Impl2448 implements DemoService {public int type() {return 2448;} public void run() {System.out.println(2448);}}
    public static class Impl2449 implements DemoService {public int type() {return 2449;} public void run() {System.out.println(2449);}}
    public static class Impl2450 implements DemoService {public int type() {return 2450;} public void run() {System.out.println(2450);}}
    public static class Impl2451 implements DemoService {public int type() {return 2451;} public void run() {System.out.println(2451);}}
    public static class Impl2452 implements DemoService {public int type() {return 2452;} public void run() {System.out.println(2452);}}
    public static class Impl2453 implements DemoService {public int type() {return 2453;} public void run() {System.out.println(2453);}}
    public static class Impl2454 implements DemoService {public int type() {return 2454;} public void run() {System.out.println(2454);}}
    public static class Impl2455 implements DemoService {public int type() {return 2455;} public void run() {System.out.println(2455);}}
    public static class Impl2456 implements DemoService {public int type() {return 2456;} public void run() {System.out.println(2456);}}
    public static class Impl2457 implements DemoService {public int type() {return 2457;} public void run() {System.out.println(2457);}}
    public static class Impl2458 implements DemoService {public int type() {return 2458;} public void run() {System.out.println(2458);}}
    public static class Impl2459 implements DemoService {public int type() {return 2459;} public void run() {System.out.println(2459);}}
    public static class Impl2460 implements DemoService {public int type() {return 2460;} public void run() {System.out.println(2460);}}
    public static class Impl2461 implements DemoService {public int type() {return 2461;} public void run() {System.out.println(2461);}}
    public static class Impl2462 implements DemoService {public int type() {return 2462;} public void run() {System.out.println(2462);}}
    public static class Impl2463 implements DemoService {public int type() {return 2463;} public void run() {System.out.println(2463);}}
    public static class Impl2464 implements DemoService {public int type() {return 2464;} public void run() {System.out.println(2464);}}
    public static class Impl2465 implements DemoService {public int type() {return 2465;} public void run() {System.out.println(2465);}}
    public static class Impl2466 implements DemoService {public int type() {return 2466;} public void run() {System.out.println(2466);}}
    public static class Impl2467 implements DemoService {public int type() {return 2467;} public void run() {System.out.println(2467);}}
    public static class Impl2468 implements DemoService {public int type() {return 2468;} public void run() {System.out.println(2468);}}
    public static class Impl2469 implements DemoService {public int type() {return 2469;} public void run() {System.out.println(2469);}}
    public static class Impl2470 implements DemoService {public int type() {return 2470;} public void run() {System.out.println(2470);}}
    public static class Impl2471 implements DemoService {public int type() {return 2471;} public void run() {System.out.println(2471);}}
    public static class Impl2472 implements DemoService {public int type() {return 2472;} public void run() {System.out.println(2472);}}
    public static class Impl2473 implements DemoService {public int type() {return 2473;} public void run() {System.out.println(2473);}}
    public static class Impl2474 implements DemoService {public int type() {return 2474;} public void run() {System.out.println(2474);}}
    public static class Impl2475 implements DemoService {public int type() {return 2475;} public void run() {System.out.println(2475);}}
    public static class Impl2476 implements DemoService {public int type() {return 2476;} public void run() {System.out.println(2476);}}
    public static class Impl2477 implements DemoService {public int type() {return 2477;} public void run() {System.out.println(2477);}}
    public static class Impl2478 implements DemoService {public int type() {return 2478;} public void run() {System.out.println(2478);}}
    public static class Impl2479 implements DemoService {public int type() {return 2479;} public void run() {System.out.println(2479);}}
    public static class Impl2480 implements DemoService {public int type() {return 2480;} public void run() {System.out.println(2480);}}
    public static class Impl2481 implements DemoService {public int type() {return 2481;} public void run() {System.out.println(2481);}}
    public static class Impl2482 implements DemoService {public int type() {return 2482;} public void run() {System.out.println(2482);}}
    public static class Impl2483 implements DemoService {public int type() {return 2483;} public void run() {System.out.println(2483);}}
    public static class Impl2484 implements DemoService {public int type() {return 2484;} public void run() {System.out.println(2484);}}
    public static class Impl2485 implements DemoService {public int type() {return 2485;} public void run() {System.out.println(2485);}}
    public static class Impl2486 implements DemoService {public int type() {return 2486;} public void run() {System.out.println(2486);}}
    public static class Impl2487 implements DemoService {public int type() {return 2487;} public void run() {System.out.println(2487);}}
    public static class Impl2488 implements DemoService {public int type() {return 2488;} public void run() {System.out.println(2488);}}
    public static class Impl2489 implements DemoService {public int type() {return 2489;} public void run() {System.out.println(2489);}}
    public static class Impl2490 implements DemoService {public int type() {return 2490;} public void run() {System.out.println(2490);}}
    public static class Impl2491 implements DemoService {public int type() {return 2491;} public void run() {System.out.println(2491);}}
    public static class Impl2492 implements DemoService {public int type() {return 2492;} public void run() {System.out.println(2492);}}
    public static class Impl2493 implements DemoService {public int type() {return 2493;} public void run() {System.out.println(2493);}}
    public static class Impl2494 implements DemoService {public int type() {return 2494;} public void run() {System.out.println(2494);}}
    public static class Impl2495 implements DemoService {public int type() {return 2495;} public void run() {System.out.println(2495);}}
    public static class Impl2496 implements DemoService {public int type() {return 2496;} public void run() {System.out.println(2496);}}
    public static class Impl2497 implements DemoService {public int type() {return 2497;} public void run() {System.out.println(2497);}}
    public static class Impl2498 implements DemoService {public int type() {return 2498;} public void run() {System.out.println(2498);}}
    public static class Impl2499 implements DemoService {public int type() {return 2499;} public void run() {System.out.println(2499);}}
    public static class Impl2500 implements DemoService {public int type() {return 2500;} public void run() {System.out.println(2500);}}
    public static class Impl2501 implements DemoService {public int type() {return 2501;} public void run() {System.out.println(2501);}}
    public static class Impl2502 implements DemoService {public int type() {return 2502;} public void run() {System.out.println(2502);}}
    public static class Impl2503 implements DemoService {public int type() {return 2503;} public void run() {System.out.println(2503);}}
    public static class Impl2504 implements DemoService {public int type() {return 2504;} public void run() {System.out.println(2504);}}
    public static class Impl2505 implements DemoService {public int type() {return 2505;} public void run() {System.out.println(2505);}}
    public static class Impl2506 implements DemoService {public int type() {return 2506;} public void run() {System.out.println(2506);}}
    public static class Impl2507 implements DemoService {public int type() {return 2507;} public void run() {System.out.println(2507);}}
    public static class Impl2508 implements DemoService {public int type() {return 2508;} public void run() {System.out.println(2508);}}
    public static class Impl2509 implements DemoService {public int type() {return 2509;} public void run() {System.out.println(2509);}}
    public static class Impl2510 implements DemoService {public int type() {return 2510;} public void run() {System.out.println(2510);}}
    public static class Impl2511 implements DemoService {public int type() {return 2511;} public void run() {System.out.println(2511);}}
    public static class Impl2512 implements DemoService {public int type() {return 2512;} public void run() {System.out.println(2512);}}
    public static class Impl2513 implements DemoService {public int type() {return 2513;} public void run() {System.out.println(2513);}}
    public static class Impl2514 implements DemoService {public int type() {return 2514;} public void run() {System.out.println(2514);}}
    public static class Impl2515 implements DemoService {public int type() {return 2515;} public void run() {System.out.println(2515);}}
    public static class Impl2516 implements DemoService {public int type() {return 2516;} public void run() {System.out.println(2516);}}
    public static class Impl2517 implements DemoService {public int type() {return 2517;} public void run() {System.out.println(2517);}}
    public static class Impl2518 implements DemoService {public int type() {return 2518;} public void run() {System.out.println(2518);}}
    public static class Impl2519 implements DemoService {public int type() {return 2519;} public void run() {System.out.println(2519);}}
    public static class Impl2520 implements DemoService {public int type() {return 2520;} public void run() {System.out.println(2520);}}
    public static class Impl2521 implements DemoService {public int type() {return 2521;} public void run() {System.out.println(2521);}}
    public static class Impl2522 implements DemoService {public int type() {return 2522;} public void run() {System.out.println(2522);}}
    public static class Impl2523 implements DemoService {public int type() {return 2523;} public void run() {System.out.println(2523);}}
    public static class Impl2524 implements DemoService {public int type() {return 2524;} public void run() {System.out.println(2524);}}
    public static class Impl2525 implements DemoService {public int type() {return 2525;} public void run() {System.out.println(2525);}}
    public static class Impl2526 implements DemoService {public int type() {return 2526;} public void run() {System.out.println(2526);}}
    public static class Impl2527 implements DemoService {public int type() {return 2527;} public void run() {System.out.println(2527);}}
    public static class Impl2528 implements DemoService {public int type() {return 2528;} public void run() {System.out.println(2528);}}
    public static class Impl2529 implements DemoService {public int type() {return 2529;} public void run() {System.out.println(2529);}}
    public static class Impl2530 implements DemoService {public int type() {return 2530;} public void run() {System.out.println(2530);}}
    public static class Impl2531 implements DemoService {public int type() {return 2531;} public void run() {System.out.println(2531);}}
    public static class Impl2532 implements DemoService {public int type() {return 2532;} public void run() {System.out.println(2532);}}
    public static class Impl2533 implements DemoService {public int type() {return 2533;} public void run() {System.out.println(2533);}}
    public static class Impl2534 implements DemoService {public int type() {return 2534;} public void run() {System.out.println(2534);}}
    public static class Impl2535 implements DemoService {public int type() {return 2535;} public void run() {System.out.println(2535);}}
    public static class Impl2536 implements DemoService {public int type() {return 2536;} public void run() {System.out.println(2536);}}
    public static class Impl2537 implements DemoService {public int type() {return 2537;} public void run() {System.out.println(2537);}}
    public static class Impl2538 implements DemoService {public int type() {return 2538;} public void run() {System.out.println(2538);}}
    public static class Impl2539 implements DemoService {public int type() {return 2539;} public void run() {System.out.println(2539);}}
    public static class Impl2540 implements DemoService {public int type() {return 2540;} public void run() {System.out.println(2540);}}
    public static class Impl2541 implements DemoService {public int type() {return 2541;} public void run() {System.out.println(2541);}}
    public static class Impl2542 implements DemoService {public int type() {return 2542;} public void run() {System.out.println(2542);}}
    public static class Impl2543 implements DemoService {public int type() {return 2543;} public void run() {System.out.println(2543);}}
    public static class Impl2544 implements DemoService {public int type() {return 2544;} public void run() {System.out.println(2544);}}
    public static class Impl2545 implements DemoService {public int type() {return 2545;} public void run() {System.out.println(2545);}}
    public static class Impl2546 implements DemoService {public int type() {return 2546;} public void run() {System.out.println(2546);}}
    public static class Impl2547 implements DemoService {public int type() {return 2547;} public void run() {System.out.println(2547);}}
    public static class Impl2548 implements DemoService {public int type() {return 2548;} public void run() {System.out.println(2548);}}
    public static class Impl2549 implements DemoService {public int type() {return 2549;} public void run() {System.out.println(2549);}}
    public static class Impl2550 implements DemoService {public int type() {return 2550;} public void run() {System.out.println(2550);}}
    public static class Impl2551 implements DemoService {public int type() {return 2551;} public void run() {System.out.println(2551);}}
    public static class Impl2552 implements DemoService {public int type() {return 2552;} public void run() {System.out.println(2552);}}
    public static class Impl2553 implements DemoService {public int type() {return 2553;} public void run() {System.out.println(2553);}}
    public static class Impl2554 implements DemoService {public int type() {return 2554;} public void run() {System.out.println(2554);}}
    public static class Impl2555 implements DemoService {public int type() {return 2555;} public void run() {System.out.println(2555);}}
    public static class Impl2556 implements DemoService {public int type() {return 2556;} public void run() {System.out.println(2556);}}
    public static class Impl2557 implements DemoService {public int type() {return 2557;} public void run() {System.out.println(2557);}}
    public static class Impl2558 implements DemoService {public int type() {return 2558;} public void run() {System.out.println(2558);}}
    public static class Impl2559 implements DemoService {public int type() {return 2559;} public void run() {System.out.println(2559);}}
    public static class Impl2560 implements DemoService {public int type() {return 2560;} public void run() {System.out.println(2560);}}
    public static class Impl2561 implements DemoService {public int type() {return 2561;} public void run() {System.out.println(2561);}}
    public static class Impl2562 implements DemoService {public int type() {return 2562;} public void run() {System.out.println(2562);}}
    public static class Impl2563 implements DemoService {public int type() {return 2563;} public void run() {System.out.println(2563);}}
    public static class Impl2564 implements DemoService {public int type() {return 2564;} public void run() {System.out.println(2564);}}
    public static class Impl2565 implements DemoService {public int type() {return 2565;} public void run() {System.out.println(2565);}}
    public static class Impl2566 implements DemoService {public int type() {return 2566;} public void run() {System.out.println(2566);}}
    public static class Impl2567 implements DemoService {public int type() {return 2567;} public void run() {System.out.println(2567);}}
    public static class Impl2568 implements DemoService {public int type() {return 2568;} public void run() {System.out.println(2568);}}
    public static class Impl2569 implements DemoService {public int type() {return 2569;} public void run() {System.out.println(2569);}}
    public static class Impl2570 implements DemoService {public int type() {return 2570;} public void run() {System.out.println(2570);}}
    public static class Impl2571 implements DemoService {public int type() {return 2571;} public void run() {System.out.println(2571);}}
    public static class Impl2572 implements DemoService {public int type() {return 2572;} public void run() {System.out.println(2572);}}
    public static class Impl2573 implements DemoService {public int type() {return 2573;} public void run() {System.out.println(2573);}}
    public static class Impl2574 implements DemoService {public int type() {return 2574;} public void run() {System.out.println(2574);}}
    public static class Impl2575 implements DemoService {public int type() {return 2575;} public void run() {System.out.println(2575);}}
    public static class Impl2576 implements DemoService {public int type() {return 2576;} public void run() {System.out.println(2576);}}
    public static class Impl2577 implements DemoService {public int type() {return 2577;} public void run() {System.out.println(2577);}}
    public static class Impl2578 implements DemoService {public int type() {return 2578;} public void run() {System.out.println(2578);}}
    public static class Impl2579 implements DemoService {public int type() {return 2579;} public void run() {System.out.println(2579);}}
    public static class Impl2580 implements DemoService {public int type() {return 2580;} public void run() {System.out.println(2580);}}
    public static class Impl2581 implements DemoService {public int type() {return 2581;} public void run() {System.out.println(2581);}}
    public static class Impl2582 implements DemoService {public int type() {return 2582;} public void run() {System.out.println(2582);}}
    public static class Impl2583 implements DemoService {public int type() {return 2583;} public void run() {System.out.println(2583);}}
    public static class Impl2584 implements DemoService {public int type() {return 2584;} public void run() {System.out.println(2584);}}
    public static class Impl2585 implements DemoService {public int type() {return 2585;} public void run() {System.out.println(2585);}}
    public static class Impl2586 implements DemoService {public int type() {return 2586;} public void run() {System.out.println(2586);}}
    public static class Impl2587 implements DemoService {public int type() {return 2587;} public void run() {System.out.println(2587);}}
    public static class Impl2588 implements DemoService {public int type() {return 2588;} public void run() {System.out.println(2588);}}
    public static class Impl2589 implements DemoService {public int type() {return 2589;} public void run() {System.out.println(2589);}}
    public static class Impl2590 implements DemoService {public int type() {return 2590;} public void run() {System.out.println(2590);}}
    public static class Impl2591 implements DemoService {public int type() {return 2591;} public void run() {System.out.println(2591);}}
    public static class Impl2592 implements DemoService {public int type() {return 2592;} public void run() {System.out.println(2592);}}
    public static class Impl2593 implements DemoService {public int type() {return 2593;} public void run() {System.out.println(2593);}}
    public static class Impl2594 implements DemoService {public int type() {return 2594;} public void run() {System.out.println(2594);}}
    public static class Impl2595 implements DemoService {public int type() {return 2595;} public void run() {System.out.println(2595);}}
    public static class Impl2596 implements DemoService {public int type() {return 2596;} public void run() {System.out.println(2596);}}
    public static class Impl2597 implements DemoService {public int type() {return 2597;} public void run() {System.out.println(2597);}}
    public static class Impl2598 implements DemoService {public int type() {return 2598;} public void run() {System.out.println(2598);}}
    public static class Impl2599 implements DemoService {public int type() {return 2599;} public void run() {System.out.println(2599);}}
    public static class Impl2600 implements DemoService {public int type() {return 2600;} public void run() {System.out.println(2600);}}
    public static class Impl2601 implements DemoService {public int type() {return 2601;} public void run() {System.out.println(2601);}}
    public static class Impl2602 implements DemoService {public int type() {return 2602;} public void run() {System.out.println(2602);}}
    public static class Impl2603 implements DemoService {public int type() {return 2603;} public void run() {System.out.println(2603);}}
    public static class Impl2604 implements DemoService {public int type() {return 2604;} public void run() {System.out.println(2604);}}
    public static class Impl2605 implements DemoService {public int type() {return 2605;} public void run() {System.out.println(2605);}}
    public static class Impl2606 implements DemoService {public int type() {return 2606;} public void run() {System.out.println(2606);}}
    public static class Impl2607 implements DemoService {public int type() {return 2607;} public void run() {System.out.println(2607);}}
    public static class Impl2608 implements DemoService {public int type() {return 2608;} public void run() {System.out.println(2608);}}
    public static class Impl2609 implements DemoService {public int type() {return 2609;} public void run() {System.out.println(2609);}}
    public static class Impl2610 implements DemoService {public int type() {return 2610;} public void run() {System.out.println(2610);}}
    public static class Impl2611 implements DemoService {public int type() {return 2611;} public void run() {System.out.println(2611);}}
    public static class Impl2612 implements DemoService {public int type() {return 2612;} public void run() {System.out.println(2612);}}
    public static class Impl2613 implements DemoService {public int type() {return 2613;} public void run() {System.out.println(2613);}}
    public static class Impl2614 implements DemoService {public int type() {return 2614;} public void run() {System.out.println(2614);}}
    public static class Impl2615 implements DemoService {public int type() {return 2615;} public void run() {System.out.println(2615);}}
    public static class Impl2616 implements DemoService {public int type() {return 2616;} public void run() {System.out.println(2616);}}
    public static class Impl2617 implements DemoService {public int type() {return 2617;} public void run() {System.out.println(2617);}}
    public static class Impl2618 implements DemoService {public int type() {return 2618;} public void run() {System.out.println(2618);}}
    public static class Impl2619 implements DemoService {public int type() {return 2619;} public void run() {System.out.println(2619);}}
    public static class Impl2620 implements DemoService {public int type() {return 2620;} public void run() {System.out.println(2620);}}
    public static class Impl2621 implements DemoService {public int type() {return 2621;} public void run() {System.out.println(2621);}}
    public static class Impl2622 implements DemoService {public int type() {return 2622;} public void run() {System.out.println(2622);}}
    public static class Impl2623 implements DemoService {public int type() {return 2623;} public void run() {System.out.println(2623);}}
    public static class Impl2624 implements DemoService {public int type() {return 2624;} public void run() {System.out.println(2624);}}
    public static class Impl2625 implements DemoService {public int type() {return 2625;} public void run() {System.out.println(2625);}}
    public static class Impl2626 implements DemoService {public int type() {return 2626;} public void run() {System.out.println(2626);}}
    public static class Impl2627 implements DemoService {public int type() {return 2627;} public void run() {System.out.println(2627);}}
    public static class Impl2628 implements DemoService {public int type() {return 2628;} public void run() {System.out.println(2628);}}
    public static class Impl2629 implements DemoService {public int type() {return 2629;} public void run() {System.out.println(2629);}}
    public static class Impl2630 implements DemoService {public int type() {return 2630;} public void run() {System.out.println(2630);}}
    public static class Impl2631 implements DemoService {public int type() {return 2631;} public void run() {System.out.println(2631);}}
    public static class Impl2632 implements DemoService {public int type() {return 2632;} public void run() {System.out.println(2632);}}
    public static class Impl2633 implements DemoService {public int type() {return 2633;} public void run() {System.out.println(2633);}}
    public static class Impl2634 implements DemoService {public int type() {return 2634;} public void run() {System.out.println(2634);}}
    public static class Impl2635 implements DemoService {public int type() {return 2635;} public void run() {System.out.println(2635);}}
    public static class Impl2636 implements DemoService {public int type() {return 2636;} public void run() {System.out.println(2636);}}
    public static class Impl2637 implements DemoService {public int type() {return 2637;} public void run() {System.out.println(2637);}}
    public static class Impl2638 implements DemoService {public int type() {return 2638;} public void run() {System.out.println(2638);}}
    public static class Impl2639 implements DemoService {public int type() {return 2639;} public void run() {System.out.println(2639);}}
    public static class Impl2640 implements DemoService {public int type() {return 2640;} public void run() {System.out.println(2640);}}
    public static class Impl2641 implements DemoService {public int type() {return 2641;} public void run() {System.out.println(2641);}}
    public static class Impl2642 implements DemoService {public int type() {return 2642;} public void run() {System.out.println(2642);}}
    public static class Impl2643 implements DemoService {public int type() {return 2643;} public void run() {System.out.println(2643);}}
    public static class Impl2644 implements DemoService {public int type() {return 2644;} public void run() {System.out.println(2644);}}
    public static class Impl2645 implements DemoService {public int type() {return 2645;} public void run() {System.out.println(2645);}}
    public static class Impl2646 implements DemoService {public int type() {return 2646;} public void run() {System.out.println(2646);}}
    public static class Impl2647 implements DemoService {public int type() {return 2647;} public void run() {System.out.println(2647);}}
    public static class Impl2648 implements DemoService {public int type() {return 2648;} public void run() {System.out.println(2648);}}
    public static class Impl2649 implements DemoService {public int type() {return 2649;} public void run() {System.out.println(2649);}}
    public static class Impl2650 implements DemoService {public int type() {return 2650;} public void run() {System.out.println(2650);}}
    public static class Impl2651 implements DemoService {public int type() {return 2651;} public void run() {System.out.println(2651);}}
    public static class Impl2652 implements DemoService {public int type() {return 2652;} public void run() {System.out.println(2652);}}
    public static class Impl2653 implements DemoService {public int type() {return 2653;} public void run() {System.out.println(2653);}}
    public static class Impl2654 implements DemoService {public int type() {return 2654;} public void run() {System.out.println(2654);}}
    public static class Impl2655 implements DemoService {public int type() {return 2655;} public void run() {System.out.println(2655);}}
    public static class Impl2656 implements DemoService {public int type() {return 2656;} public void run() {System.out.println(2656);}}
    public static class Impl2657 implements DemoService {public int type() {return 2657;} public void run() {System.out.println(2657);}}
    public static class Impl2658 implements DemoService {public int type() {return 2658;} public void run() {System.out.println(2658);}}
    public static class Impl2659 implements DemoService {public int type() {return 2659;} public void run() {System.out.println(2659);}}
    public static class Impl2660 implements DemoService {public int type() {return 2660;} public void run() {System.out.println(2660);}}
    public static class Impl2661 implements DemoService {public int type() {return 2661;} public void run() {System.out.println(2661);}}
    public static class Impl2662 implements DemoService {public int type() {return 2662;} public void run() {System.out.println(2662);}}
    public static class Impl2663 implements DemoService {public int type() {return 2663;} public void run() {System.out.println(2663);}}
    public static class Impl2664 implements DemoService {public int type() {return 2664;} public void run() {System.out.println(2664);}}
    public static class Impl2665 implements DemoService {public int type() {return 2665;} public void run() {System.out.println(2665);}}
    public static class Impl2666 implements DemoService {public int type() {return 2666;} public void run() {System.out.println(2666);}}
    public static class Impl2667 implements DemoService {public int type() {return 2667;} public void run() {System.out.println(2667);}}
    public static class Impl2668 implements DemoService {public int type() {return 2668;} public void run() {System.out.println(2668);}}
    public static class Impl2669 implements DemoService {public int type() {return 2669;} public void run() {System.out.println(2669);}}
    public static class Impl2670 implements DemoService {public int type() {return 2670;} public void run() {System.out.println(2670);}}
    public static class Impl2671 implements DemoService {public int type() {return 2671;} public void run() {System.out.println(2671);}}
    public static class Impl2672 implements DemoService {public int type() {return 2672;} public void run() {System.out.println(2672);}}
    public static class Impl2673 implements DemoService {public int type() {return 2673;} public void run() {System.out.println(2673);}}
    public static class Impl2674 implements DemoService {public int type() {return 2674;} public void run() {System.out.println(2674);}}
    public static class Impl2675 implements DemoService {public int type() {return 2675;} public void run() {System.out.println(2675);}}
    public static class Impl2676 implements DemoService {public int type() {return 2676;} public void run() {System.out.println(2676);}}
    public static class Impl2677 implements DemoService {public int type() {return 2677;} public void run() {System.out.println(2677);}}
    public static class Impl2678 implements DemoService {public int type() {return 2678;} public void run() {System.out.println(2678);}}
    public static class Impl2679 implements DemoService {public int type() {return 2679;} public void run() {System.out.println(2679);}}
    public static class Impl2680 implements DemoService {public int type() {return 2680;} public void run() {System.out.println(2680);}}
    public static class Impl2681 implements DemoService {public int type() {return 2681;} public void run() {System.out.println(2681);}}
    public static class Impl2682 implements DemoService {public int type() {return 2682;} public void run() {System.out.println(2682);}}
    public static class Impl2683 implements DemoService {public int type() {return 2683;} public void run() {System.out.println(2683);}}
    public static class Impl2684 implements DemoService {public int type() {return 2684;} public void run() {System.out.println(2684);}}
    public static class Impl2685 implements DemoService {public int type() {return 2685;} public void run() {System.out.println(2685);}}
    public static class Impl2686 implements DemoService {public int type() {return 2686;} public void run() {System.out.println(2686);}}
    public static class Impl2687 implements DemoService {public int type() {return 2687;} public void run() {System.out.println(2687);}}
    public static class Impl2688 implements DemoService {public int type() {return 2688;} public void run() {System.out.println(2688);}}
    public static class Impl2689 implements DemoService {public int type() {return 2689;} public void run() {System.out.println(2689);}}
    public static class Impl2690 implements DemoService {public int type() {return 2690;} public void run() {System.out.println(2690);}}
    public static class Impl2691 implements DemoService {public int type() {return 2691;} public void run() {System.out.println(2691);}}
    public static class Impl2692 implements DemoService {public int type() {return 2692;} public void run() {System.out.println(2692);}}
    public static class Impl2693 implements DemoService {public int type() {return 2693;} public void run() {System.out.println(2693);}}
    public static class Impl2694 implements DemoService {public int type() {return 2694;} public void run() {System.out.println(2694);}}
    public static class Impl2695 implements DemoService {public int type() {return 2695;} public void run() {System.out.println(2695);}}
    public static class Impl2696 implements DemoService {public int type() {return 2696;} public void run() {System.out.println(2696);}}
    public static class Impl2697 implements DemoService {public int type() {return 2697;} public void run() {System.out.println(2697);}}
    public static class Impl2698 implements DemoService {public int type() {return 2698;} public void run() {System.out.println(2698);}}
    public static class Impl2699 implements DemoService {public int type() {return 2699;} public void run() {System.out.println(2699);}}
    public static class Impl2700 implements DemoService {public int type() {return 2700;} public void run() {System.out.println(2700);}}
    public static class Impl2701 implements DemoService {public int type() {return 2701;} public void run() {System.out.println(2701);}}
    public static class Impl2702 implements DemoService {public int type() {return 2702;} public void run() {System.out.println(2702);}}
    public static class Impl2703 implements DemoService {public int type() {return 2703;} public void run() {System.out.println(2703);}}
    public static class Impl2704 implements DemoService {public int type() {return 2704;} public void run() {System.out.println(2704);}}
    public static class Impl2705 implements DemoService {public int type() {return 2705;} public void run() {System.out.println(2705);}}
    public static class Impl2706 implements DemoService {public int type() {return 2706;} public void run() {System.out.println(2706);}}
    public static class Impl2707 implements DemoService {public int type() {return 2707;} public void run() {System.out.println(2707);}}
    public static class Impl2708 implements DemoService {public int type() {return 2708;} public void run() {System.out.println(2708);}}
    public static class Impl2709 implements DemoService {public int type() {return 2709;} public void run() {System.out.println(2709);}}
    public static class Impl2710 implements DemoService {public int type() {return 2710;} public void run() {System.out.println(2710);}}
    public static class Impl2711 implements DemoService {public int type() {return 2711;} public void run() {System.out.println(2711);}}
    public static class Impl2712 implements DemoService {public int type() {return 2712;} public void run() {System.out.println(2712);}}
    public static class Impl2713 implements DemoService {public int type() {return 2713;} public void run() {System.out.println(2713);}}
    public static class Impl2714 implements DemoService {public int type() {return 2714;} public void run() {System.out.println(2714);}}
    public static class Impl2715 implements DemoService {public int type() {return 2715;} public void run() {System.out.println(2715);}}
    public static class Impl2716 implements DemoService {public int type() {return 2716;} public void run() {System.out.println(2716);}}
    public static class Impl2717 implements DemoService {public int type() {return 2717;} public void run() {System.out.println(2717);}}
    public static class Impl2718 implements DemoService {public int type() {return 2718;} public void run() {System.out.println(2718);}}
    public static class Impl2719 implements DemoService {public int type() {return 2719;} public void run() {System.out.println(2719);}}
    public static class Impl2720 implements DemoService {public int type() {return 2720;} public void run() {System.out.println(2720);}}
    public static class Impl2721 implements DemoService {public int type() {return 2721;} public void run() {System.out.println(2721);}}
    public static class Impl2722 implements DemoService {public int type() {return 2722;} public void run() {System.out.println(2722);}}
    public static class Impl2723 implements DemoService {public int type() {return 2723;} public void run() {System.out.println(2723);}}
    public static class Impl2724 implements DemoService {public int type() {return 2724;} public void run() {System.out.println(2724);}}
    public static class Impl2725 implements DemoService {public int type() {return 2725;} public void run() {System.out.println(2725);}}
    public static class Impl2726 implements DemoService {public int type() {return 2726;} public void run() {System.out.println(2726);}}
    public static class Impl2727 implements DemoService {public int type() {return 2727;} public void run() {System.out.println(2727);}}
    public static class Impl2728 implements DemoService {public int type() {return 2728;} public void run() {System.out.println(2728);}}
    public static class Impl2729 implements DemoService {public int type() {return 2729;} public void run() {System.out.println(2729);}}
    public static class Impl2730 implements DemoService {public int type() {return 2730;} public void run() {System.out.println(2730);}}
    public static class Impl2731 implements DemoService {public int type() {return 2731;} public void run() {System.out.println(2731);}}
    public static class Impl2732 implements DemoService {public int type() {return 2732;} public void run() {System.out.println(2732);}}
    public static class Impl2733 implements DemoService {public int type() {return 2733;} public void run() {System.out.println(2733);}}
    public static class Impl2734 implements DemoService {public int type() {return 2734;} public void run() {System.out.println(2734);}}
    public static class Impl2735 implements DemoService {public int type() {return 2735;} public void run() {System.out.println(2735);}}
    public static class Impl2736 implements DemoService {public int type() {return 2736;} public void run() {System.out.println(2736);}}
    public static class Impl2737 implements DemoService {public int type() {return 2737;} public void run() {System.out.println(2737);}}
    public static class Impl2738 implements DemoService {public int type() {return 2738;} public void run() {System.out.println(2738);}}
    public static class Impl2739 implements DemoService {public int type() {return 2739;} public void run() {System.out.println(2739);}}
    public static class Impl2740 implements DemoService {public int type() {return 2740;} public void run() {System.out.println(2740);}}
    public static class Impl2741 implements DemoService {public int type() {return 2741;} public void run() {System.out.println(2741);}}
    public static class Impl2742 implements DemoService {public int type() {return 2742;} public void run() {System.out.println(2742);}}
    public static class Impl2743 implements DemoService {public int type() {return 2743;} public void run() {System.out.println(2743);}}
    public static class Impl2744 implements DemoService {public int type() {return 2744;} public void run() {System.out.println(2744);}}
    public static class Impl2745 implements DemoService {public int type() {return 2745;} public void run() {System.out.println(2745);}}
    public static class Impl2746 implements DemoService {public int type() {return 2746;} public void run() {System.out.println(2746);}}
    public static class Impl2747 implements DemoService {public int type() {return 2747;} public void run() {System.out.println(2747);}}
    public static class Impl2748 implements DemoService {public int type() {return 2748;} public void run() {System.out.println(2748);}}
    public static class Impl2749 implements DemoService {public int type() {return 2749;} public void run() {System.out.println(2749);}}
    public static class Impl2750 implements DemoService {public int type() {return 2750;} public void run() {System.out.println(2750);}}
    public static class Impl2751 implements DemoService {public int type() {return 2751;} public void run() {System.out.println(2751);}}
    public static class Impl2752 implements DemoService {public int type() {return 2752;} public void run() {System.out.println(2752);}}
    public static class Impl2753 implements DemoService {public int type() {return 2753;} public void run() {System.out.println(2753);}}
    public static class Impl2754 implements DemoService {public int type() {return 2754;} public void run() {System.out.println(2754);}}
    public static class Impl2755 implements DemoService {public int type() {return 2755;} public void run() {System.out.println(2755);}}
    public static class Impl2756 implements DemoService {public int type() {return 2756;} public void run() {System.out.println(2756);}}
    public static class Impl2757 implements DemoService {public int type() {return 2757;} public void run() {System.out.println(2757);}}
    public static class Impl2758 implements DemoService {public int type() {return 2758;} public void run() {System.out.println(2758);}}
    public static class Impl2759 implements DemoService {public int type() {return 2759;} public void run() {System.out.println(2759);}}
    public static class Impl2760 implements DemoService {public int type() {return 2760;} public void run() {System.out.println(2760);}}
    public static class Impl2761 implements DemoService {public int type() {return 2761;} public void run() {System.out.println(2761);}}
    public static class Impl2762 implements DemoService {public int type() {return 2762;} public void run() {System.out.println(2762);}}
    public static class Impl2763 implements DemoService {public int type() {return 2763;} public void run() {System.out.println(2763);}}
    public static class Impl2764 implements DemoService {public int type() {return 2764;} public void run() {System.out.println(2764);}}
    public static class Impl2765 implements DemoService {public int type() {return 2765;} public void run() {System.out.println(2765);}}
    public static class Impl2766 implements DemoService {public int type() {return 2766;} public void run() {System.out.println(2766);}}
    public static class Impl2767 implements DemoService {public int type() {return 2767;} public void run() {System.out.println(2767);}}
    public static class Impl2768 implements DemoService {public int type() {return 2768;} public void run() {System.out.println(2768);}}
    public static class Impl2769 implements DemoService {public int type() {return 2769;} public void run() {System.out.println(2769);}}
    public static class Impl2770 implements DemoService {public int type() {return 2770;} public void run() {System.out.println(2770);}}
    public static class Impl2771 implements DemoService {public int type() {return 2771;} public void run() {System.out.println(2771);}}
    public static class Impl2772 implements DemoService {public int type() {return 2772;} public void run() {System.out.println(2772);}}
    public static class Impl2773 implements DemoService {public int type() {return 2773;} public void run() {System.out.println(2773);}}
    public static class Impl2774 implements DemoService {public int type() {return 2774;} public void run() {System.out.println(2774);}}
    public static class Impl2775 implements DemoService {public int type() {return 2775;} public void run() {System.out.println(2775);}}
    public static class Impl2776 implements DemoService {public int type() {return 2776;} public void run() {System.out.println(2776);}}
    public static class Impl2777 implements DemoService {public int type() {return 2777;} public void run() {System.out.println(2777);}}
    public static class Impl2778 implements DemoService {public int type() {return 2778;} public void run() {System.out.println(2778);}}
    public static class Impl2779 implements DemoService {public int type() {return 2779;} public void run() {System.out.println(2779);}}
    public static class Impl2780 implements DemoService {public int type() {return 2780;} public void run() {System.out.println(2780);}}
    public static class Impl2781 implements DemoService {public int type() {return 2781;} public void run() {System.out.println(2781);}}
    public static class Impl2782 implements DemoService {public int type() {return 2782;} public void run() {System.out.println(2782);}}
    public static class Impl2783 implements DemoService {public int type() {return 2783;} public void run() {System.out.println(2783);}}
    public static class Impl2784 implements DemoService {public int type() {return 2784;} public void run() {System.out.println(2784);}}
    public static class Impl2785 implements DemoService {public int type() {return 2785;} public void run() {System.out.println(2785);}}
    public static class Impl2786 implements DemoService {public int type() {return 2786;} public void run() {System.out.println(2786);}}
    public static class Impl2787 implements DemoService {public int type() {return 2787;} public void run() {System.out.println(2787);}}
    public static class Impl2788 implements DemoService {public int type() {return 2788;} public void run() {System.out.println(2788);}}
    public static class Impl2789 implements DemoService {public int type() {return 2789;} public void run() {System.out.println(2789);}}
    public static class Impl2790 implements DemoService {public int type() {return 2790;} public void run() {System.out.println(2790);}}
    public static class Impl2791 implements DemoService {public int type() {return 2791;} public void run() {System.out.println(2791);}}
    public static class Impl2792 implements DemoService {public int type() {return 2792;} public void run() {System.out.println(2792);}}
    public static class Impl2793 implements DemoService {public int type() {return 2793;} public void run() {System.out.println(2793);}}
    public static class Impl2794 implements DemoService {public int type() {return 2794;} public void run() {System.out.println(2794);}}
    public static class Impl2795 implements DemoService {public int type() {return 2795;} public void run() {System.out.println(2795);}}
    public static class Impl2796 implements DemoService {public int type() {return 2796;} public void run() {System.out.println(2796);}}
    public static class Impl2797 implements DemoService {public int type() {return 2797;} public void run() {System.out.println(2797);}}
    public static class Impl2798 implements DemoService {public int type() {return 2798;} public void run() {System.out.println(2798);}}
    public static class Impl2799 implements DemoService {public int type() {return 2799;} public void run() {System.out.println(2799);}}
    public static class Impl2800 implements DemoService {public int type() {return 2800;} public void run() {System.out.println(2800);}}
    public static class Impl2801 implements DemoService {public int type() {return 2801;} public void run() {System.out.println(2801);}}
    public static class Impl2802 implements DemoService {public int type() {return 2802;} public void run() {System.out.println(2802);}}
    public static class Impl2803 implements DemoService {public int type() {return 2803;} public void run() {System.out.println(2803);}}
    public static class Impl2804 implements DemoService {public int type() {return 2804;} public void run() {System.out.println(2804);}}
    public static class Impl2805 implements DemoService {public int type() {return 2805;} public void run() {System.out.println(2805);}}
    public static class Impl2806 implements DemoService {public int type() {return 2806;} public void run() {System.out.println(2806);}}
    public static class Impl2807 implements DemoService {public int type() {return 2807;} public void run() {System.out.println(2807);}}
    public static class Impl2808 implements DemoService {public int type() {return 2808;} public void run() {System.out.println(2808);}}
    public static class Impl2809 implements DemoService {public int type() {return 2809;} public void run() {System.out.println(2809);}}
    public static class Impl2810 implements DemoService {public int type() {return 2810;} public void run() {System.out.println(2810);}}
    public static class Impl2811 implements DemoService {public int type() {return 2811;} public void run() {System.out.println(2811);}}
    public static class Impl2812 implements DemoService {public int type() {return 2812;} public void run() {System.out.println(2812);}}
    public static class Impl2813 implements DemoService {public int type() {return 2813;} public void run() {System.out.println(2813);}}
    public static class Impl2814 implements DemoService {public int type() {return 2814;} public void run() {System.out.println(2814);}}
    public static class Impl2815 implements DemoService {public int type() {return 2815;} public void run() {System.out.println(2815);}}
    public static class Impl2816 implements DemoService {public int type() {return 2816;} public void run() {System.out.println(2816);}}
    public static class Impl2817 implements DemoService {public int type() {return 2817;} public void run() {System.out.println(2817);}}
    public static class Impl2818 implements DemoService {public int type() {return 2818;} public void run() {System.out.println(2818);}}
    public static class Impl2819 implements DemoService {public int type() {return 2819;} public void run() {System.out.println(2819);}}
    public static class Impl2820 implements DemoService {public int type() {return 2820;} public void run() {System.out.println(2820);}}
    public static class Impl2821 implements DemoService {public int type() {return 2821;} public void run() {System.out.println(2821);}}
    public static class Impl2822 implements DemoService {public int type() {return 2822;} public void run() {System.out.println(2822);}}
    public static class Impl2823 implements DemoService {public int type() {return 2823;} public void run() {System.out.println(2823);}}
    public static class Impl2824 implements DemoService {public int type() {return 2824;} public void run() {System.out.println(2824);}}
    public static class Impl2825 implements DemoService {public int type() {return 2825;} public void run() {System.out.println(2825);}}
    public static class Impl2826 implements DemoService {public int type() {return 2826;} public void run() {System.out.println(2826);}}
    public static class Impl2827 implements DemoService {public int type() {return 2827;} public void run() {System.out.println(2827);}}
    public static class Impl2828 implements DemoService {public int type() {return 2828;} public void run() {System.out.println(2828);}}
    public static class Impl2829 implements DemoService {public int type() {return 2829;} public void run() {System.out.println(2829);}}
    public static class Impl2830 implements DemoService {public int type() {return 2830;} public void run() {System.out.println(2830);}}
    public static class Impl2831 implements DemoService {public int type() {return 2831;} public void run() {System.out.println(2831);}}
    public static class Impl2832 implements DemoService {public int type() {return 2832;} public void run() {System.out.println(2832);}}
    public static class Impl2833 implements DemoService {public int type() {return 2833;} public void run() {System.out.println(2833);}}
    public static class Impl2834 implements DemoService {public int type() {return 2834;} public void run() {System.out.println(2834);}}
    public static class Impl2835 implements DemoService {public int type() {return 2835;} public void run() {System.out.println(2835);}}
    public static class Impl2836 implements DemoService {public int type() {return 2836;} public void run() {System.out.println(2836);}}
    public static class Impl2837 implements DemoService {public int type() {return 2837;} public void run() {System.out.println(2837);}}
    public static class Impl2838 implements DemoService {public int type() {return 2838;} public void run() {System.out.println(2838);}}
    public static class Impl2839 implements DemoService {public int type() {return 2839;} public void run() {System.out.println(2839);}}
    public static class Impl2840 implements DemoService {public int type() {return 2840;} public void run() {System.out.println(2840);}}
    public static class Impl2841 implements DemoService {public int type() {return 2841;} public void run() {System.out.println(2841);}}
    public static class Impl2842 implements DemoService {public int type() {return 2842;} public void run() {System.out.println(2842);}}
    public static class Impl2843 implements DemoService {public int type() {return 2843;} public void run() {System.out.println(2843);}}
    public static class Impl2844 implements DemoService {public int type() {return 2844;} public void run() {System.out.println(2844);}}
    public static class Impl2845 implements DemoService {public int type() {return 2845;} public void run() {System.out.println(2845);}}
    public static class Impl2846 implements DemoService {public int type() {return 2846;} public void run() {System.out.println(2846);}}
    public static class Impl2847 implements DemoService {public int type() {return 2847;} public void run() {System.out.println(2847);}}
    public static class Impl2848 implements DemoService {public int type() {return 2848;} public void run() {System.out.println(2848);}}
    public static class Impl2849 implements DemoService {public int type() {return 2849;} public void run() {System.out.println(2849);}}
    public static class Impl2850 implements DemoService {public int type() {return 2850;} public void run() {System.out.println(2850);}}
    public static class Impl2851 implements DemoService {public int type() {return 2851;} public void run() {System.out.println(2851);}}
    public static class Impl2852 implements DemoService {public int type() {return 2852;} public void run() {System.out.println(2852);}}
    public static class Impl2853 implements DemoService {public int type() {return 2853;} public void run() {System.out.println(2853);}}
    public static class Impl2854 implements DemoService {public int type() {return 2854;} public void run() {System.out.println(2854);}}
    public static class Impl2855 implements DemoService {public int type() {return 2855;} public void run() {System.out.println(2855);}}
    public static class Impl2856 implements DemoService {public int type() {return 2856;} public void run() {System.out.println(2856);}}
    public static class Impl2857 implements DemoService {public int type() {return 2857;} public void run() {System.out.println(2857);}}
    public static class Impl2858 implements DemoService {public int type() {return 2858;} public void run() {System.out.println(2858);}}
    public static class Impl2859 implements DemoService {public int type() {return 2859;} public void run() {System.out.println(2859);}}
    public static class Impl2860 implements DemoService {public int type() {return 2860;} public void run() {System.out.println(2860);}}
    public static class Impl2861 implements DemoService {public int type() {return 2861;} public void run() {System.out.println(2861);}}
    public static class Impl2862 implements DemoService {public int type() {return 2862;} public void run() {System.out.println(2862);}}
    public static class Impl2863 implements DemoService {public int type() {return 2863;} public void run() {System.out.println(2863);}}
    public static class Impl2864 implements DemoService {public int type() {return 2864;} public void run() {System.out.println(2864);}}
    public static class Impl2865 implements DemoService {public int type() {return 2865;} public void run() {System.out.println(2865);}}
    public static class Impl2866 implements DemoService {public int type() {return 2866;} public void run() {System.out.println(2866);}}
    public static class Impl2867 implements DemoService {public int type() {return 2867;} public void run() {System.out.println(2867);}}
    public static class Impl2868 implements DemoService {public int type() {return 2868;} public void run() {System.out.println(2868);}}
    public static class Impl2869 implements DemoService {public int type() {return 2869;} public void run() {System.out.println(2869);}}
    public static class Impl2870 implements DemoService {public int type() {return 2870;} public void run() {System.out.println(2870);}}
    public static class Impl2871 implements DemoService {public int type() {return 2871;} public void run() {System.out.println(2871);}}
    public static class Impl2872 implements DemoService {public int type() {return 2872;} public void run() {System.out.println(2872);}}
    public static class Impl2873 implements DemoService {public int type() {return 2873;} public void run() {System.out.println(2873);}}
    public static class Impl2874 implements DemoService {public int type() {return 2874;} public void run() {System.out.println(2874);}}
    public static class Impl2875 implements DemoService {public int type() {return 2875;} public void run() {System.out.println(2875);}}
    public static class Impl2876 implements DemoService {public int type() {return 2876;} public void run() {System.out.println(2876);}}
    public static class Impl2877 implements DemoService {public int type() {return 2877;} public void run() {System.out.println(2877);}}
    public static class Impl2878 implements DemoService {public int type() {return 2878;} public void run() {System.out.println(2878);}}
    public static class Impl2879 implements DemoService {public int type() {return 2879;} public void run() {System.out.println(2879);}}
    public static class Impl2880 implements DemoService {public int type() {return 2880;} public void run() {System.out.println(2880);}}
    public static class Impl2881 implements DemoService {public int type() {return 2881;} public void run() {System.out.println(2881);}}
    public static class Impl2882 implements DemoService {public int type() {return 2882;} public void run() {System.out.println(2882);}}
    public static class Impl2883 implements DemoService {public int type() {return 2883;} public void run() {System.out.println(2883);}}
    public static class Impl2884 implements DemoService {public int type() {return 2884;} public void run() {System.out.println(2884);}}
    public static class Impl2885 implements DemoService {public int type() {return 2885;} public void run() {System.out.println(2885);}}
    public static class Impl2886 implements DemoService {public int type() {return 2886;} public void run() {System.out.println(2886);}}
    public static class Impl2887 implements DemoService {public int type() {return 2887;} public void run() {System.out.println(2887);}}
    public static class Impl2888 implements DemoService {public int type() {return 2888;} public void run() {System.out.println(2888);}}
    public static class Impl2889 implements DemoService {public int type() {return 2889;} public void run() {System.out.println(2889);}}
    public static class Impl2890 implements DemoService {public int type() {return 2890;} public void run() {System.out.println(2890);}}
    public static class Impl2891 implements DemoService {public int type() {return 2891;} public void run() {System.out.println(2891);}}
    public static class Impl2892 implements DemoService {public int type() {return 2892;} public void run() {System.out.println(2892);}}
    public static class Impl2893 implements DemoService {public int type() {return 2893;} public void run() {System.out.println(2893);}}
    public static class Impl2894 implements DemoService {public int type() {return 2894;} public void run() {System.out.println(2894);}}
    public static class Impl2895 implements DemoService {public int type() {return 2895;} public void run() {System.out.println(2895);}}
    public static class Impl2896 implements DemoService {public int type() {return 2896;} public void run() {System.out.println(2896);}}
    public static class Impl2897 implements DemoService {public int type() {return 2897;} public void run() {System.out.println(2897);}}
    public static class Impl2898 implements DemoService {public int type() {return 2898;} public void run() {System.out.println(2898);}}
    public static class Impl2899 implements DemoService {public int type() {return 2899;} public void run() {System.out.println(2899);}}
    public static class Impl2900 implements DemoService {public int type() {return 2900;} public void run() {System.out.println(2900);}}
    public static class Impl2901 implements DemoService {public int type() {return 2901;} public void run() {System.out.println(2901);}}
    public static class Impl2902 implements DemoService {public int type() {return 2902;} public void run() {System.out.println(2902);}}
    public static class Impl2903 implements DemoService {public int type() {return 2903;} public void run() {System.out.println(2903);}}
    public static class Impl2904 implements DemoService {public int type() {return 2904;} public void run() {System.out.println(2904);}}
    public static class Impl2905 implements DemoService {public int type() {return 2905;} public void run() {System.out.println(2905);}}
    public static class Impl2906 implements DemoService {public int type() {return 2906;} public void run() {System.out.println(2906);}}
    public static class Impl2907 implements DemoService {public int type() {return 2907;} public void run() {System.out.println(2907);}}
    public static class Impl2908 implements DemoService {public int type() {return 2908;} public void run() {System.out.println(2908);}}
    public static class Impl2909 implements DemoService {public int type() {return 2909;} public void run() {System.out.println(2909);}}
    public static class Impl2910 implements DemoService {public int type() {return 2910;} public void run() {System.out.println(2910);}}
    public static class Impl2911 implements DemoService {public int type() {return 2911;} public void run() {System.out.println(2911);}}
    public static class Impl2912 implements DemoService {public int type() {return 2912;} public void run() {System.out.println(2912);}}
    public static class Impl2913 implements DemoService {public int type() {return 2913;} public void run() {System.out.println(2913);}}
    public static class Impl2914 implements DemoService {public int type() {return 2914;} public void run() {System.out.println(2914);}}
    public static class Impl2915 implements DemoService {public int type() {return 2915;} public void run() {System.out.println(2915);}}
    public static class Impl2916 implements DemoService {public int type() {return 2916;} public void run() {System.out.println(2916);}}
    public static class Impl2917 implements DemoService {public int type() {return 2917;} public void run() {System.out.println(2917);}}
    public static class Impl2918 implements DemoService {public int type() {return 2918;} public void run() {System.out.println(2918);}}
    public static class Impl2919 implements DemoService {public int type() {return 2919;} public void run() {System.out.println(2919);}}
    public static class Impl2920 implements DemoService {public int type() {return 2920;} public void run() {System.out.println(2920);}}
    public static class Impl2921 implements DemoService {public int type() {return 2921;} public void run() {System.out.println(2921);}}
    public static class Impl2922 implements DemoService {public int type() {return 2922;} public void run() {System.out.println(2922);}}
    public static class Impl2923 implements DemoService {public int type() {return 2923;} public void run() {System.out.println(2923);}}
    public static class Impl2924 implements DemoService {public int type() {return 2924;} public void run() {System.out.println(2924);}}
    public static class Impl2925 implements DemoService {public int type() {return 2925;} public void run() {System.out.println(2925);}}
    public static class Impl2926 implements DemoService {public int type() {return 2926;} public void run() {System.out.println(2926);}}
    public static class Impl2927 implements DemoService {public int type() {return 2927;} public void run() {System.out.println(2927);}}
    public static class Impl2928 implements DemoService {public int type() {return 2928;} public void run() {System.out.println(2928);}}
    public static class Impl2929 implements DemoService {public int type() {return 2929;} public void run() {System.out.println(2929);}}
    public static class Impl2930 implements DemoService {public int type() {return 2930;} public void run() {System.out.println(2930);}}
    public static class Impl2931 implements DemoService {public int type() {return 2931;} public void run() {System.out.println(2931);}}
    public static class Impl2932 implements DemoService {public int type() {return 2932;} public void run() {System.out.println(2932);}}
    public static class Impl2933 implements DemoService {public int type() {return 2933;} public void run() {System.out.println(2933);}}
    public static class Impl2934 implements DemoService {public int type() {return 2934;} public void run() {System.out.println(2934);}}
    public static class Impl2935 implements DemoService {public int type() {return 2935;} public void run() {System.out.println(2935);}}
    public static class Impl2936 implements DemoService {public int type() {return 2936;} public void run() {System.out.println(2936);}}
    public static class Impl2937 implements DemoService {public int type() {return 2937;} public void run() {System.out.println(2937);}}
    public static class Impl2938 implements DemoService {public int type() {return 2938;} public void run() {System.out.println(2938);}}
    public static class Impl2939 implements DemoService {public int type() {return 2939;} public void run() {System.out.println(2939);}}
    public static class Impl2940 implements DemoService {public int type() {return 2940;} public void run() {System.out.println(2940);}}
    public static class Impl2941 implements DemoService {public int type() {return 2941;} public void run() {System.out.println(2941);}}
    public static class Impl2942 implements DemoService {public int type() {return 2942;} public void run() {System.out.println(2942);}}
    public static class Impl2943 implements DemoService {public int type() {return 2943;} public void run() {System.out.println(2943);}}
    public static class Impl2944 implements DemoService {public int type() {return 2944;} public void run() {System.out.println(2944);}}
    public static class Impl2945 implements DemoService {public int type() {return 2945;} public void run() {System.out.println(2945);}}
    public static class Impl2946 implements DemoService {public int type() {return 2946;} public void run() {System.out.println(2946);}}
    public static class Impl2947 implements DemoService {public int type() {return 2947;} public void run() {System.out.println(2947);}}
    public static class Impl2948 implements DemoService {public int type() {return 2948;} public void run() {System.out.println(2948);}}
    public static class Impl2949 implements DemoService {public int type() {return 2949;} public void run() {System.out.println(2949);}}
    public static class Impl2950 implements DemoService {public int type() {return 2950;} public void run() {System.out.println(2950);}}
    public static class Impl2951 implements DemoService {public int type() {return 2951;} public void run() {System.out.println(2951);}}
    public static class Impl2952 implements DemoService {public int type() {return 2952;} public void run() {System.out.println(2952);}}
    public static class Impl2953 implements DemoService {public int type() {return 2953;} public void run() {System.out.println(2953);}}
    public static class Impl2954 implements DemoService {public int type() {return 2954;} public void run() {System.out.println(2954);}}
    public static class Impl2955 implements DemoService {public int type() {return 2955;} public void run() {System.out.println(2955);}}
    public static class Impl2956 implements DemoService {public int type() {return 2956;} public void run() {System.out.println(2956);}}
    public static class Impl2957 implements DemoService {public int type() {return 2957;} public void run() {System.out.println(2957);}}
    public static class Impl2958 implements DemoService {public int type() {return 2958;} public void run() {System.out.println(2958);}}
    public static class Impl2959 implements DemoService {public int type() {return 2959;} public void run() {System.out.println(2959);}}
    public static class Impl2960 implements DemoService {public int type() {return 2960;} public void run() {System.out.println(2960);}}
    public static class Impl2961 implements DemoService {public int type() {return 2961;} public void run() {System.out.println(2961);}}
    public static class Impl2962 implements DemoService {public int type() {return 2962;} public void run() {System.out.println(2962);}}
    public static class Impl2963 implements DemoService {public int type() {return 2963;} public void run() {System.out.println(2963);}}
    public static class Impl2964 implements DemoService {public int type() {return 2964;} public void run() {System.out.println(2964);}}
    public static class Impl2965 implements DemoService {public int type() {return 2965;} public void run() {System.out.println(2965);}}
    public static class Impl2966 implements DemoService {public int type() {return 2966;} public void run() {System.out.println(2966);}}
    public static class Impl2967 implements DemoService {public int type() {return 2967;} public void run() {System.out.println(2967);}}
    public static class Impl2968 implements DemoService {public int type() {return 2968;} public void run() {System.out.println(2968);}}
    public static class Impl2969 implements DemoService {public int type() {return 2969;} public void run() {System.out.println(2969);}}
    public static class Impl2970 implements DemoService {public int type() {return 2970;} public void run() {System.out.println(2970);}}
    public static class Impl2971 implements DemoService {public int type() {return 2971;} public void run() {System.out.println(2971);}}
    public static class Impl2972 implements DemoService {public int type() {return 2972;} public void run() {System.out.println(2972);}}
    public static class Impl2973 implements DemoService {public int type() {return 2973;} public void run() {System.out.println(2973);}}
    public static class Impl2974 implements DemoService {public int type() {return 2974;} public void run() {System.out.println(2974);}}
    public static class Impl2975 implements DemoService {public int type() {return 2975;} public void run() {System.out.println(2975);}}
    public static class Impl2976 implements DemoService {public int type() {return 2976;} public void run() {System.out.println(2976);}}
    public static class Impl2977 implements DemoService {public int type() {return 2977;} public void run() {System.out.println(2977);}}
    public static class Impl2978 implements DemoService {public int type() {return 2978;} public void run() {System.out.println(2978);}}
    public static class Impl2979 implements DemoService {public int type() {return 2979;} public void run() {System.out.println(2979);}}
    public static class Impl2980 implements DemoService {public int type() {return 2980;} public void run() {System.out.println(2980);}}
    public static class Impl2981 implements DemoService {public int type() {return 2981;} public void run() {System.out.println(2981);}}
    public static class Impl2982 implements DemoService {public int type() {return 2982;} public void run() {System.out.println(2982);}}
    public static class Impl2983 implements DemoService {public int type() {return 2983;} public void run() {System.out.println(2983);}}
    public static class Impl2984 implements DemoService {public int type() {return 2984;} public void run() {System.out.println(2984);}}
    public static class Impl2985 implements DemoService {public int type() {return 2985;} public void run() {System.out.println(2985);}}
    public static class Impl2986 implements DemoService {public int type() {return 2986;} public void run() {System.out.println(2986);}}
    public static class Impl2987 implements DemoService {public int type() {return 2987;} public void run() {System.out.println(2987);}}
    public static class Impl2988 implements DemoService {public int type() {return 2988;} public void run() {System.out.println(2988);}}
    public static class Impl2989 implements DemoService {public int type() {return 2989;} public void run() {System.out.println(2989);}}
    public static class Impl2990 implements DemoService {public int type() {return 2990;} public void run() {System.out.println(2990);}}
    public static class Impl2991 implements DemoService {public int type() {return 2991;} public void run() {System.out.println(2991);}}
    public static class Impl2992 implements DemoService {public int type() {return 2992;} public void run() {System.out.println(2992);}}
    public static class Impl2993 implements DemoService {public int type() {return 2993;} public void run() {System.out.println(2993);}}
    public static class Impl2994 implements DemoService {public int type() {return 2994;} public void run() {System.out.println(2994);}}
    public static class Impl2995 implements DemoService {public int type() {return 2995;} public void run() {System.out.println(2995);}}
    public static class Impl2996 implements DemoService {public int type() {return 2996;} public void run() {System.out.println(2996);}}
    public static class Impl2997 implements DemoService {public int type() {return 2997;} public void run() {System.out.println(2997);}}
    public static class Impl2998 implements DemoService {public int type() {return 2998;} public void run() {System.out.println(2998);}}
    public static class Impl2999 implements DemoService {public int type() {return 2999;} public void run() {System.out.println(2999);}}
    public static class Impl3000 implements DemoService {public int type() {return 3000;} public void run() {System.out.println(3000);}}
    public static class Impl3001 implements DemoService {public int type() {return 3001;} public void run() {System.out.println(3001);}}
    public static class Impl3002 implements DemoService {public int type() {return 3002;} public void run() {System.out.println(3002);}}
    public static class Impl3003 implements DemoService {public int type() {return 3003;} public void run() {System.out.println(3003);}}
    public static class Impl3004 implements DemoService {public int type() {return 3004;} public void run() {System.out.println(3004);}}
    public static class Impl3005 implements DemoService {public int type() {return 3005;} public void run() {System.out.println(3005);}}
    public static class Impl3006 implements DemoService {public int type() {return 3006;} public void run() {System.out.println(3006);}}
    public static class Impl3007 implements DemoService {public int type() {return 3007;} public void run() {System.out.println(3007);}}
    public static class Impl3008 implements DemoService {public int type() {return 3008;} public void run() {System.out.println(3008);}}
    public static class Impl3009 implements DemoService {public int type() {return 3009;} public void run() {System.out.println(3009);}}
    public static class Impl3010 implements DemoService {public int type() {return 3010;} public void run() {System.out.println(3010);}}
    public static class Impl3011 implements DemoService {public int type() {return 3011;} public void run() {System.out.println(3011);}}
    public static class Impl3012 implements DemoService {public int type() {return 3012;} public void run() {System.out.println(3012);}}
    public static class Impl3013 implements DemoService {public int type() {return 3013;} public void run() {System.out.println(3013);}}
    public static class Impl3014 implements DemoService {public int type() {return 3014;} public void run() {System.out.println(3014);}}
    public static class Impl3015 implements DemoService {public int type() {return 3015;} public void run() {System.out.println(3015);}}
    public static class Impl3016 implements DemoService {public int type() {return 3016;} public void run() {System.out.println(3016);}}
    public static class Impl3017 implements DemoService {public int type() {return 3017;} public void run() {System.out.println(3017);}}
    public static class Impl3018 implements DemoService {public int type() {return 3018;} public void run() {System.out.println(3018);}}
    public static class Impl3019 implements DemoService {public int type() {return 3019;} public void run() {System.out.println(3019);}}
    public static class Impl3020 implements DemoService {public int type() {return 3020;} public void run() {System.out.println(3020);}}
    public static class Impl3021 implements DemoService {public int type() {return 3021;} public void run() {System.out.println(3021);}}
    public static class Impl3022 implements DemoService {public int type() {return 3022;} public void run() {System.out.println(3022);}}
    public static class Impl3023 implements DemoService {public int type() {return 3023;} public void run() {System.out.println(3023);}}
    public static class Impl3024 implements DemoService {public int type() {return 3024;} public void run() {System.out.println(3024);}}
    public static class Impl3025 implements DemoService {public int type() {return 3025;} public void run() {System.out.println(3025);}}
    public static class Impl3026 implements DemoService {public int type() {return 3026;} public void run() {System.out.println(3026);}}
    public static class Impl3027 implements DemoService {public int type() {return 3027;} public void run() {System.out.println(3027);}}
    public static class Impl3028 implements DemoService {public int type() {return 3028;} public void run() {System.out.println(3028);}}
    public static class Impl3029 implements DemoService {public int type() {return 3029;} public void run() {System.out.println(3029);}}
    public static class Impl3030 implements DemoService {public int type() {return 3030;} public void run() {System.out.println(3030);}}
    public static class Impl3031 implements DemoService {public int type() {return 3031;} public void run() {System.out.println(3031);}}
    public static class Impl3032 implements DemoService {public int type() {return 3032;} public void run() {System.out.println(3032);}}
    public static class Impl3033 implements DemoService {public int type() {return 3033;} public void run() {System.out.println(3033);}}
    public static class Impl3034 implements DemoService {public int type() {return 3034;} public void run() {System.out.println(3034);}}
    public static class Impl3035 implements DemoService {public int type() {return 3035;} public void run() {System.out.println(3035);}}
    public static class Impl3036 implements DemoService {public int type() {return 3036;} public void run() {System.out.println(3036);}}
    public static class Impl3037 implements DemoService {public int type() {return 3037;} public void run() {System.out.println(3037);}}
    public static class Impl3038 implements DemoService {public int type() {return 3038;} public void run() {System.out.println(3038);}}
    public static class Impl3039 implements DemoService {public int type() {return 3039;} public void run() {System.out.println(3039);}}
    public static class Impl3040 implements DemoService {public int type() {return 3040;} public void run() {System.out.println(3040);}}
    public static class Impl3041 implements DemoService {public int type() {return 3041;} public void run() {System.out.println(3041);}}
    public static class Impl3042 implements DemoService {public int type() {return 3042;} public void run() {System.out.println(3042);}}
    public static class Impl3043 implements DemoService {public int type() {return 3043;} public void run() {System.out.println(3043);}}
    public static class Impl3044 implements DemoService {public int type() {return 3044;} public void run() {System.out.println(3044);}}
    public static class Impl3045 implements DemoService {public int type() {return 3045;} public void run() {System.out.println(3045);}}
    public static class Impl3046 implements DemoService {public int type() {return 3046;} public void run() {System.out.println(3046);}}
    public static class Impl3047 implements DemoService {public int type() {return 3047;} public void run() {System.out.println(3047);}}
    public static class Impl3048 implements DemoService {public int type() {return 3048;} public void run() {System.out.println(3048);}}
    public static class Impl3049 implements DemoService {public int type() {return 3049;} public void run() {System.out.println(3049);}}
    public static class Impl3050 implements DemoService {public int type() {return 3050;} public void run() {System.out.println(3050);}}
    public static class Impl3051 implements DemoService {public int type() {return 3051;} public void run() {System.out.println(3051);}}
    public static class Impl3052 implements DemoService {public int type() {return 3052;} public void run() {System.out.println(3052);}}
    public static class Impl3053 implements DemoService {public int type() {return 3053;} public void run() {System.out.println(3053);}}
    public static class Impl3054 implements DemoService {public int type() {return 3054;} public void run() {System.out.println(3054);}}
    public static class Impl3055 implements DemoService {public int type() {return 3055;} public void run() {System.out.println(3055);}}
    public static class Impl3056 implements DemoService {public int type() {return 3056;} public void run() {System.out.println(3056);}}
    public static class Impl3057 implements DemoService {public int type() {return 3057;} public void run() {System.out.println(3057);}}
    public static class Impl3058 implements DemoService {public int type() {return 3058;} public void run() {System.out.println(3058);}}
    public static class Impl3059 implements DemoService {public int type() {return 3059;} public void run() {System.out.println(3059);}}
    public static class Impl3060 implements DemoService {public int type() {return 3060;} public void run() {System.out.println(3060);}}
    public static class Impl3061 implements DemoService {public int type() {return 3061;} public void run() {System.out.println(3061);}}
    public static class Impl3062 implements DemoService {public int type() {return 3062;} public void run() {System.out.println(3062);}}
    public static class Impl3063 implements DemoService {public int type() {return 3063;} public void run() {System.out.println(3063);}}
    public static class Impl3064 implements DemoService {public int type() {return 3064;} public void run() {System.out.println(3064);}}
    public static class Impl3065 implements DemoService {public int type() {return 3065;} public void run() {System.out.println(3065);}}
    public static class Impl3066 implements DemoService {public int type() {return 3066;} public void run() {System.out.println(3066);}}
    public static class Impl3067 implements DemoService {public int type() {return 3067;} public void run() {System.out.println(3067);}}
    public static class Impl3068 implements DemoService {public int type() {return 3068;} public void run() {System.out.println(3068);}}
    public static class Impl3069 implements DemoService {public int type() {return 3069;} public void run() {System.out.println(3069);}}
    public static class Impl3070 implements DemoService {public int type() {return 3070;} public void run() {System.out.println(3070);}}
    public static class Impl3071 implements DemoService {public int type() {return 3071;} public void run() {System.out.println(3071);}}
    public static class Impl3072 implements DemoService {public int type() {return 3072;} public void run() {System.out.println(3072);}}
    public static class Impl3073 implements DemoService {public int type() {return 3073;} public void run() {System.out.println(3073);}}
    public static class Impl3074 implements DemoService {public int type() {return 3074;} public void run() {System.out.println(3074);}}
    public static class Impl3075 implements DemoService {public int type() {return 3075;} public void run() {System.out.println(3075);}}
    public static class Impl3076 implements DemoService {public int type() {return 3076;} public void run() {System.out.println(3076);}}
    public static class Impl3077 implements DemoService {public int type() {return 3077;} public void run() {System.out.println(3077);}}
    public static class Impl3078 implements DemoService {public int type() {return 3078;} public void run() {System.out.println(3078);}}
    public static class Impl3079 implements DemoService {public int type() {return 3079;} public void run() {System.out.println(3079);}}
    public static class Impl3080 implements DemoService {public int type() {return 3080;} public void run() {System.out.println(3080);}}
    public static class Impl3081 implements DemoService {public int type() {return 3081;} public void run() {System.out.println(3081);}}
    public static class Impl3082 implements DemoService {public int type() {return 3082;} public void run() {System.out.println(3082);}}
    public static class Impl3083 implements DemoService {public int type() {return 3083;} public void run() {System.out.println(3083);}}
    public static class Impl3084 implements DemoService {public int type() {return 3084;} public void run() {System.out.println(3084);}}
    public static class Impl3085 implements DemoService {public int type() {return 3085;} public void run() {System.out.println(3085);}}
    public static class Impl3086 implements DemoService {public int type() {return 3086;} public void run() {System.out.println(3086);}}
    public static class Impl3087 implements DemoService {public int type() {return 3087;} public void run() {System.out.println(3087);}}
    public static class Impl3088 implements DemoService {public int type() {return 3088;} public void run() {System.out.println(3088);}}
    public static class Impl3089 implements DemoService {public int type() {return 3089;} public void run() {System.out.println(3089);}}
    public static class Impl3090 implements DemoService {public int type() {return 3090;} public void run() {System.out.println(3090);}}
    public static class Impl3091 implements DemoService {public int type() {return 3091;} public void run() {System.out.println(3091);}}
    public static class Impl3092 implements DemoService {public int type() {return 3092;} public void run() {System.out.println(3092);}}
    public static class Impl3093 implements DemoService {public int type() {return 3093;} public void run() {System.out.println(3093);}}
    public static class Impl3094 implements DemoService {public int type() {return 3094;} public void run() {System.out.println(3094);}}
    public static class Impl3095 implements DemoService {public int type() {return 3095;} public void run() {System.out.println(3095);}}
    public static class Impl3096 implements DemoService {public int type() {return 3096;} public void run() {System.out.println(3096);}}
    public static class Impl3097 implements DemoService {public int type() {return 3097;} public void run() {System.out.println(3097);}}
    public static class Impl3098 implements DemoService {public int type() {return 3098;} public void run() {System.out.println(3098);}}
    public static class Impl3099 implements DemoService {public int type() {return 3099;} public void run() {System.out.println(3099);}}
    public static class Impl3100 implements DemoService {public int type() {return 3100;} public void run() {System.out.println(3100);}}
    public static class Impl3101 implements DemoService {public int type() {return 3101;} public void run() {System.out.println(3101);}}
    public static class Impl3102 implements DemoService {public int type() {return 3102;} public void run() {System.out.println(3102);}}
    public static class Impl3103 implements DemoService {public int type() {return 3103;} public void run() {System.out.println(3103);}}
    public static class Impl3104 implements DemoService {public int type() {return 3104;} public void run() {System.out.println(3104);}}
    public static class Impl3105 implements DemoService {public int type() {return 3105;} public void run() {System.out.println(3105);}}
    public static class Impl3106 implements DemoService {public int type() {return 3106;} public void run() {System.out.println(3106);}}
    public static class Impl3107 implements DemoService {public int type() {return 3107;} public void run() {System.out.println(3107);}}
    public static class Impl3108 implements DemoService {public int type() {return 3108;} public void run() {System.out.println(3108);}}
    public static class Impl3109 implements DemoService {public int type() {return 3109;} public void run() {System.out.println(3109);}}
    public static class Impl3110 implements DemoService {public int type() {return 3110;} public void run() {System.out.println(3110);}}
    public static class Impl3111 implements DemoService {public int type() {return 3111;} public void run() {System.out.println(3111);}}
    public static class Impl3112 implements DemoService {public int type() {return 3112;} public void run() {System.out.println(3112);}}
    public static class Impl3113 implements DemoService {public int type() {return 3113;} public void run() {System.out.println(3113);}}
    public static class Impl3114 implements DemoService {public int type() {return 3114;} public void run() {System.out.println(3114);}}
    public static class Impl3115 implements DemoService {public int type() {return 3115;} public void run() {System.out.println(3115);}}
    public static class Impl3116 implements DemoService {public int type() {return 3116;} public void run() {System.out.println(3116);}}
    public static class Impl3117 implements DemoService {public int type() {return 3117;} public void run() {System.out.println(3117);}}
    public static class Impl3118 implements DemoService {public int type() {return 3118;} public void run() {System.out.println(3118);}}
    public static class Impl3119 implements DemoService {public int type() {return 3119;} public void run() {System.out.println(3119);}}
    public static class Impl3120 implements DemoService {public int type() {return 3120;} public void run() {System.out.println(3120);}}
    public static class Impl3121 implements DemoService {public int type() {return 3121;} public void run() {System.out.println(3121);}}
    public static class Impl3122 implements DemoService {public int type() {return 3122;} public void run() {System.out.println(3122);}}
    public static class Impl3123 implements DemoService {public int type() {return 3123;} public void run() {System.out.println(3123);}}
    public static class Impl3124 implements DemoService {public int type() {return 3124;} public void run() {System.out.println(3124);}}
    public static class Impl3125 implements DemoService {public int type() {return 3125;} public void run() {System.out.println(3125);}}
    public static class Impl3126 implements DemoService {public int type() {return 3126;} public void run() {System.out.println(3126);}}
    public static class Impl3127 implements DemoService {public int type() {return 3127;} public void run() {System.out.println(3127);}}
    public static class Impl3128 implements DemoService {public int type() {return 3128;} public void run() {System.out.println(3128);}}
    public static class Impl3129 implements DemoService {public int type() {return 3129;} public void run() {System.out.println(3129);}}
    public static class Impl3130 implements DemoService {public int type() {return 3130;} public void run() {System.out.println(3130);}}
    public static class Impl3131 implements DemoService {public int type() {return 3131;} public void run() {System.out.println(3131);}}
    public static class Impl3132 implements DemoService {public int type() {return 3132;} public void run() {System.out.println(3132);}}
    public static class Impl3133 implements DemoService {public int type() {return 3133;} public void run() {System.out.println(3133);}}
    public static class Impl3134 implements DemoService {public int type() {return 3134;} public void run() {System.out.println(3134);}}
    public static class Impl3135 implements DemoService {public int type() {return 3135;} public void run() {System.out.println(3135);}}
    public static class Impl3136 implements DemoService {public int type() {return 3136;} public void run() {System.out.println(3136);}}
    public static class Impl3137 implements DemoService {public int type() {return 3137;} public void run() {System.out.println(3137);}}
    public static class Impl3138 implements DemoService {public int type() {return 3138;} public void run() {System.out.println(3138);}}
    public static class Impl3139 implements DemoService {public int type() {return 3139;} public void run() {System.out.println(3139);}}
    public static class Impl3140 implements DemoService {public int type() {return 3140;} public void run() {System.out.println(3140);}}
    public static class Impl3141 implements DemoService {public int type() {return 3141;} public void run() {System.out.println(3141);}}
    public static class Impl3142 implements DemoService {public int type() {return 3142;} public void run() {System.out.println(3142);}}
    public static class Impl3143 implements DemoService {public int type() {return 3143;} public void run() {System.out.println(3143);}}
    public static class Impl3144 implements DemoService {public int type() {return 3144;} public void run() {System.out.println(3144);}}
    public static class Impl3145 implements DemoService {public int type() {return 3145;} public void run() {System.out.println(3145);}}
    public static class Impl3146 implements DemoService {public int type() {return 3146;} public void run() {System.out.println(3146);}}
    public static class Impl3147 implements DemoService {public int type() {return 3147;} public void run() {System.out.println(3147);}}
    public static class Impl3148 implements DemoService {public int type() {return 3148;} public void run() {System.out.println(3148);}}
    public static class Impl3149 implements DemoService {public int type() {return 3149;} public void run() {System.out.println(3149);}}
    public static class Impl3150 implements DemoService {public int type() {return 3150;} public void run() {System.out.println(3150);}}
    public static class Impl3151 implements DemoService {public int type() {return 3151;} public void run() {System.out.println(3151);}}
    public static class Impl3152 implements DemoService {public int type() {return 3152;} public void run() {System.out.println(3152);}}
    public static class Impl3153 implements DemoService {public int type() {return 3153;} public void run() {System.out.println(3153);}}
    public static class Impl3154 implements DemoService {public int type() {return 3154;} public void run() {System.out.println(3154);}}
    public static class Impl3155 implements DemoService {public int type() {return 3155;} public void run() {System.out.println(3155);}}
    public static class Impl3156 implements DemoService {public int type() {return 3156;} public void run() {System.out.println(3156);}}
    public static class Impl3157 implements DemoService {public int type() {return 3157;} public void run() {System.out.println(3157);}}
    public static class Impl3158 implements DemoService {public int type() {return 3158;} public void run() {System.out.println(3158);}}
    public static class Impl3159 implements DemoService {public int type() {return 3159;} public void run() {System.out.println(3159);}}
    public static class Impl3160 implements DemoService {public int type() {return 3160;} public void run() {System.out.println(3160);}}
    public static class Impl3161 implements DemoService {public int type() {return 3161;} public void run() {System.out.println(3161);}}
    public static class Impl3162 implements DemoService {public int type() {return 3162;} public void run() {System.out.println(3162);}}
    public static class Impl3163 implements DemoService {public int type() {return 3163;} public void run() {System.out.println(3163);}}
    public static class Impl3164 implements DemoService {public int type() {return 3164;} public void run() {System.out.println(3164);}}
    public static class Impl3165 implements DemoService {public int type() {return 3165;} public void run() {System.out.println(3165);}}
    public static class Impl3166 implements DemoService {public int type() {return 3166;} public void run() {System.out.println(3166);}}
    public static class Impl3167 implements DemoService {public int type() {return 3167;} public void run() {System.out.println(3167);}}
    public static class Impl3168 implements DemoService {public int type() {return 3168;} public void run() {System.out.println(3168);}}
    public static class Impl3169 implements DemoService {public int type() {return 3169;} public void run() {System.out.println(3169);}}
    public static class Impl3170 implements DemoService {public int type() {return 3170;} public void run() {System.out.println(3170);}}
    public static class Impl3171 implements DemoService {public int type() {return 3171;} public void run() {System.out.println(3171);}}
    public static class Impl3172 implements DemoService {public int type() {return 3172;} public void run() {System.out.println(3172);}}
    public static class Impl3173 implements DemoService {public int type() {return 3173;} public void run() {System.out.println(3173);}}
    public static class Impl3174 implements DemoService {public int type() {return 3174;} public void run() {System.out.println(3174);}}
    public static class Impl3175 implements DemoService {public int type() {return 3175;} public void run() {System.out.println(3175);}}
    public static class Impl3176 implements DemoService {public int type() {return 3176;} public void run() {System.out.println(3176);}}
    public static class Impl3177 implements DemoService {public int type() {return 3177;} public void run() {System.out.println(3177);}}
    public static class Impl3178 implements DemoService {public int type() {return 3178;} public void run() {System.out.println(3178);}}
    public static class Impl3179 implements DemoService {public int type() {return 3179;} public void run() {System.out.println(3179);}}
    public static class Impl3180 implements DemoService {public int type() {return 3180;} public void run() {System.out.println(3180);}}
    public static class Impl3181 implements DemoService {public int type() {return 3181;} public void run() {System.out.println(3181);}}
    public static class Impl3182 implements DemoService {public int type() {return 3182;} public void run() {System.out.println(3182);}}
    public static class Impl3183 implements DemoService {public int type() {return 3183;} public void run() {System.out.println(3183);}}
    public static class Impl3184 implements DemoService {public int type() {return 3184;} public void run() {System.out.println(3184);}}
    public static class Impl3185 implements DemoService {public int type() {return 3185;} public void run() {System.out.println(3185);}}
    public static class Impl3186 implements DemoService {public int type() {return 3186;} public void run() {System.out.println(3186);}}
    public static class Impl3187 implements DemoService {public int type() {return 3187;} public void run() {System.out.println(3187);}}
    public static class Impl3188 implements DemoService {public int type() {return 3188;} public void run() {System.out.println(3188);}}
    public static class Impl3189 implements DemoService {public int type() {return 3189;} public void run() {System.out.println(3189);}}
    public static class Impl3190 implements DemoService {public int type() {return 3190;} public void run() {System.out.println(3190);}}
    public static class Impl3191 implements DemoService {public int type() {return 3191;} public void run() {System.out.println(3191);}}
    public static class Impl3192 implements DemoService {public int type() {return 3192;} public void run() {System.out.println(3192);}}
    public static class Impl3193 implements DemoService {public int type() {return 3193;} public void run() {System.out.println(3193);}}
    public static class Impl3194 implements DemoService {public int type() {return 3194;} public void run() {System.out.println(3194);}}
    public static class Impl3195 implements DemoService {public int type() {return 3195;} public void run() {System.out.println(3195);}}
    public static class Impl3196 implements DemoService {public int type() {return 3196;} public void run() {System.out.println(3196);}}
    public static class Impl3197 implements DemoService {public int type() {return 3197;} public void run() {System.out.println(3197);}}
    public static class Impl3198 implements DemoService {public int type() {return 3198;} public void run() {System.out.println(3198);}}
    public static class Impl3199 implements DemoService {public int type() {return 3199;} public void run() {System.out.println(3199);}}
    public static class Impl3200 implements DemoService {public int type() {return 3200;} public void run() {System.out.println(3200);}}
    public static class Impl3201 implements DemoService {public int type() {return 3201;} public void run() {System.out.println(3201);}}
    public static class Impl3202 implements DemoService {public int type() {return 3202;} public void run() {System.out.println(3202);}}
    public static class Impl3203 implements DemoService {public int type() {return 3203;} public void run() {System.out.println(3203);}}
    public static class Impl3204 implements DemoService {public int type() {return 3204;} public void run() {System.out.println(3204);}}
    public static class Impl3205 implements DemoService {public int type() {return 3205;} public void run() {System.out.println(3205);}}
    public static class Impl3206 implements DemoService {public int type() {return 3206;} public void run() {System.out.println(3206);}}
    public static class Impl3207 implements DemoService {public int type() {return 3207;} public void run() {System.out.println(3207);}}
    public static class Impl3208 implements DemoService {public int type() {return 3208;} public void run() {System.out.println(3208);}}
    public static class Impl3209 implements DemoService {public int type() {return 3209;} public void run() {System.out.println(3209);}}
    public static class Impl3210 implements DemoService {public int type() {return 3210;} public void run() {System.out.println(3210);}}
    public static class Impl3211 implements DemoService {public int type() {return 3211;} public void run() {System.out.println(3211);}}
    public static class Impl3212 implements DemoService {public int type() {return 3212;} public void run() {System.out.println(3212);}}
    public static class Impl3213 implements DemoService {public int type() {return 3213;} public void run() {System.out.println(3213);}}
    public static class Impl3214 implements DemoService {public int type() {return 3214;} public void run() {System.out.println(3214);}}
    public static class Impl3215 implements DemoService {public int type() {return 3215;} public void run() {System.out.println(3215);}}
    public static class Impl3216 implements DemoService {public int type() {return 3216;} public void run() {System.out.println(3216);}}
    public static class Impl3217 implements DemoService {public int type() {return 3217;} public void run() {System.out.println(3217);}}
    public static class Impl3218 implements DemoService {public int type() {return 3218;} public void run() {System.out.println(3218);}}
    public static class Impl3219 implements DemoService {public int type() {return 3219;} public void run() {System.out.println(3219);}}
    public static class Impl3220 implements DemoService {public int type() {return 3220;} public void run() {System.out.println(3220);}}
    public static class Impl3221 implements DemoService {public int type() {return 3221;} public void run() {System.out.println(3221);}}
    public static class Impl3222 implements DemoService {public int type() {return 3222;} public void run() {System.out.println(3222);}}
    public static class Impl3223 implements DemoService {public int type() {return 3223;} public void run() {System.out.println(3223);}}
    public static class Impl3224 implements DemoService {public int type() {return 3224;} public void run() {System.out.println(3224);}}
    public static class Impl3225 implements DemoService {public int type() {return 3225;} public void run() {System.out.println(3225);}}
    public static class Impl3226 implements DemoService {public int type() {return 3226;} public void run() {System.out.println(3226);}}
    public static class Impl3227 implements DemoService {public int type() {return 3227;} public void run() {System.out.println(3227);}}
    public static class Impl3228 implements DemoService {public int type() {return 3228;} public void run() {System.out.println(3228);}}
    public static class Impl3229 implements DemoService {public int type() {return 3229;} public void run() {System.out.println(3229);}}
    public static class Impl3230 implements DemoService {public int type() {return 3230;} public void run() {System.out.println(3230);}}
    public static class Impl3231 implements DemoService {public int type() {return 3231;} public void run() {System.out.println(3231);}}
    public static class Impl3232 implements DemoService {public int type() {return 3232;} public void run() {System.out.println(3232);}}
    public static class Impl3233 implements DemoService {public int type() {return 3233;} public void run() {System.out.println(3233);}}
    public static class Impl3234 implements DemoService {public int type() {return 3234;} public void run() {System.out.println(3234);}}
    public static class Impl3235 implements DemoService {public int type() {return 3235;} public void run() {System.out.println(3235);}}
    public static class Impl3236 implements DemoService {public int type() {return 3236;} public void run() {System.out.println(3236);}}
    public static class Impl3237 implements DemoService {public int type() {return 3237;} public void run() {System.out.println(3237);}}
    public static class Impl3238 implements DemoService {public int type() {return 3238;} public void run() {System.out.println(3238);}}
    public static class Impl3239 implements DemoService {public int type() {return 3239;} public void run() {System.out.println(3239);}}
    public static class Impl3240 implements DemoService {public int type() {return 3240;} public void run() {System.out.println(3240);}}
    public static class Impl3241 implements DemoService {public int type() {return 3241;} public void run() {System.out.println(3241);}}
    public static class Impl3242 implements DemoService {public int type() {return 3242;} public void run() {System.out.println(3242);}}
    public static class Impl3243 implements DemoService {public int type() {return 3243;} public void run() {System.out.println(3243);}}
    public static class Impl3244 implements DemoService {public int type() {return 3244;} public void run() {System.out.println(3244);}}
    public static class Impl3245 implements DemoService {public int type() {return 3245;} public void run() {System.out.println(3245);}}
    public static class Impl3246 implements DemoService {public int type() {return 3246;} public void run() {System.out.println(3246);}}
    public static class Impl3247 implements DemoService {public int type() {return 3247;} public void run() {System.out.println(3247);}}
    public static class Impl3248 implements DemoService {public int type() {return 3248;} public void run() {System.out.println(3248);}}
    public static class Impl3249 implements DemoService {public int type() {return 3249;} public void run() {System.out.println(3249);}}
    public static class Impl3250 implements DemoService {public int type() {return 3250;} public void run() {System.out.println(3250);}}
    public static class Impl3251 implements DemoService {public int type() {return 3251;} public void run() {System.out.println(3251);}}
    public static class Impl3252 implements DemoService {public int type() {return 3252;} public void run() {System.out.println(3252);}}
    public static class Impl3253 implements DemoService {public int type() {return 3253;} public void run() {System.out.println(3253);}}
    public static class Impl3254 implements DemoService {public int type() {return 3254;} public void run() {System.out.println(3254);}}
    public static class Impl3255 implements DemoService {public int type() {return 3255;} public void run() {System.out.println(3255);}}
    public static class Impl3256 implements DemoService {public int type() {return 3256;} public void run() {System.out.println(3256);}}
    public static class Impl3257 implements DemoService {public int type() {return 3257;} public void run() {System.out.println(3257);}}
    public static class Impl3258 implements DemoService {public int type() {return 3258;} public void run() {System.out.println(3258);}}
    public static class Impl3259 implements DemoService {public int type() {return 3259;} public void run() {System.out.println(3259);}}
    public static class Impl3260 implements DemoService {public int type() {return 3260;} public void run() {System.out.println(3260);}}
    public static class Impl3261 implements DemoService {public int type() {return 3261;} public void run() {System.out.println(3261);}}
    public static class Impl3262 implements DemoService {public int type() {return 3262;} public void run() {System.out.println(3262);}}
    public static class Impl3263 implements DemoService {public int type() {return 3263;} public void run() {System.out.println(3263);}}
    public static class Impl3264 implements DemoService {public int type() {return 3264;} public void run() {System.out.println(3264);}}
    public static class Impl3265 implements DemoService {public int type() {return 3265;} public void run() {System.out.println(3265);}}
    public static class Impl3266 implements DemoService {public int type() {return 3266;} public void run() {System.out.println(3266);}}
    public static class Impl3267 implements DemoService {public int type() {return 3267;} public void run() {System.out.println(3267);}}
    public static class Impl3268 implements DemoService {public int type() {return 3268;} public void run() {System.out.println(3268);}}
    public static class Impl3269 implements DemoService {public int type() {return 3269;} public void run() {System.out.println(3269);}}
    public static class Impl3270 implements DemoService {public int type() {return 3270;} public void run() {System.out.println(3270);}}
    public static class Impl3271 implements DemoService {public int type() {return 3271;} public void run() {System.out.println(3271);}}
    public static class Impl3272 implements DemoService {public int type() {return 3272;} public void run() {System.out.println(3272);}}
    public static class Impl3273 implements DemoService {public int type() {return 3273;} public void run() {System.out.println(3273);}}
    public static class Impl3274 implements DemoService {public int type() {return 3274;} public void run() {System.out.println(3274);}}
    public static class Impl3275 implements DemoService {public int type() {return 3275;} public void run() {System.out.println(3275);}}
    public static class Impl3276 implements DemoService {public int type() {return 3276;} public void run() {System.out.println(3276);}}
    public static class Impl3277 implements DemoService {public int type() {return 3277;} public void run() {System.out.println(3277);}}
    public static class Impl3278 implements DemoService {public int type() {return 3278;} public void run() {System.out.println(3278);}}
    public static class Impl3279 implements DemoService {public int type() {return 3279;} public void run() {System.out.println(3279);}}
    public static class Impl3280 implements DemoService {public int type() {return 3280;} public void run() {System.out.println(3280);}}
    public static class Impl3281 implements DemoService {public int type() {return 3281;} public void run() {System.out.println(3281);}}
    public static class Impl3282 implements DemoService {public int type() {return 3282;} public void run() {System.out.println(3282);}}
    public static class Impl3283 implements DemoService {public int type() {return 3283;} public void run() {System.out.println(3283);}}
    public static class Impl3284 implements DemoService {public int type() {return 3284;} public void run() {System.out.println(3284);}}
    public static class Impl3285 implements DemoService {public int type() {return 3285;} public void run() {System.out.println(3285);}}
    public static class Impl3286 implements DemoService {public int type() {return 3286;} public void run() {System.out.println(3286);}}
    public static class Impl3287 implements DemoService {public int type() {return 3287;} public void run() {System.out.println(3287);}}
    public static class Impl3288 implements DemoService {public int type() {return 3288;} public void run() {System.out.println(3288);}}
    public static class Impl3289 implements DemoService {public int type() {return 3289;} public void run() {System.out.println(3289);}}
    public static class Impl3290 implements DemoService {public int type() {return 3290;} public void run() {System.out.println(3290);}}
    public static class Impl3291 implements DemoService {public int type() {return 3291;} public void run() {System.out.println(3291);}}
    public static class Impl3292 implements DemoService {public int type() {return 3292;} public void run() {System.out.println(3292);}}
    public static class Impl3293 implements DemoService {public int type() {return 3293;} public void run() {System.out.println(3293);}}
    public static class Impl3294 implements DemoService {public int type() {return 3294;} public void run() {System.out.println(3294);}}
    public static class Impl3295 implements DemoService {public int type() {return 3295;} public void run() {System.out.println(3295);}}
    public static class Impl3296 implements DemoService {public int type() {return 3296;} public void run() {System.out.println(3296);}}
    public static class Impl3297 implements DemoService {public int type() {return 3297;} public void run() {System.out.println(3297);}}
    public static class Impl3298 implements DemoService {public int type() {return 3298;} public void run() {System.out.println(3298);}}
    public static class Impl3299 implements DemoService {public int type() {return 3299;} public void run() {System.out.println(3299);}}
    public static class Impl3300 implements DemoService {public int type() {return 3300;} public void run() {System.out.println(3300);}}
    public static class Impl3301 implements DemoService {public int type() {return 3301;} public void run() {System.out.println(3301);}}
    public static class Impl3302 implements DemoService {public int type() {return 3302;} public void run() {System.out.println(3302);}}
    public static class Impl3303 implements DemoService {public int type() {return 3303;} public void run() {System.out.println(3303);}}
    public static class Impl3304 implements DemoService {public int type() {return 3304;} public void run() {System.out.println(3304);}}
    public static class Impl3305 implements DemoService {public int type() {return 3305;} public void run() {System.out.println(3305);}}
    public static class Impl3306 implements DemoService {public int type() {return 3306;} public void run() {System.out.println(3306);}}
    public static class Impl3307 implements DemoService {public int type() {return 3307;} public void run() {System.out.println(3307);}}
    public static class Impl3308 implements DemoService {public int type() {return 3308;} public void run() {System.out.println(3308);}}
    public static class Impl3309 implements DemoService {public int type() {return 3309;} public void run() {System.out.println(3309);}}
    public static class Impl3310 implements DemoService {public int type() {return 3310;} public void run() {System.out.println(3310);}}
    public static class Impl3311 implements DemoService {public int type() {return 3311;} public void run() {System.out.println(3311);}}
    public static class Impl3312 implements DemoService {public int type() {return 3312;} public void run() {System.out.println(3312);}}
    public static class Impl3313 implements DemoService {public int type() {return 3313;} public void run() {System.out.println(3313);}}
    public static class Impl3314 implements DemoService {public int type() {return 3314;} public void run() {System.out.println(3314);}}
    public static class Impl3315 implements DemoService {public int type() {return 3315;} public void run() {System.out.println(3315);}}
    public static class Impl3316 implements DemoService {public int type() {return 3316;} public void run() {System.out.println(3316);}}
    public static class Impl3317 implements DemoService {public int type() {return 3317;} public void run() {System.out.println(3317);}}
    public static class Impl3318 implements DemoService {public int type() {return 3318;} public void run() {System.out.println(3318);}}
    public static class Impl3319 implements DemoService {public int type() {return 3319;} public void run() {System.out.println(3319);}}
    public static class Impl3320 implements DemoService {public int type() {return 3320;} public void run() {System.out.println(3320);}}
    public static class Impl3321 implements DemoService {public int type() {return 3321;} public void run() {System.out.println(3321);}}
    public static class Impl3322 implements DemoService {public int type() {return 3322;} public void run() {System.out.println(3322);}}
    public static class Impl3323 implements DemoService {public int type() {return 3323;} public void run() {System.out.println(3323);}}
    public static class Impl3324 implements DemoService {public int type() {return 3324;} public void run() {System.out.println(3324);}}
    public static class Impl3325 implements DemoService {public int type() {return 3325;} public void run() {System.out.println(3325);}}
    public static class Impl3326 implements DemoService {public int type() {return 3326;} public void run() {System.out.println(3326);}}
    public static class Impl3327 implements DemoService {public int type() {return 3327;} public void run() {System.out.println(3327);}}
    public static class Impl3328 implements DemoService {public int type() {return 3328;} public void run() {System.out.println(3328);}}
    public static class Impl3329 implements DemoService {public int type() {return 3329;} public void run() {System.out.println(3329);}}
    public static class Impl3330 implements DemoService {public int type() {return 3330;} public void run() {System.out.println(3330);}}
    public static class Impl3331 implements DemoService {public int type() {return 3331;} public void run() {System.out.println(3331);}}
    public static class Impl3332 implements DemoService {public int type() {return 3332;} public void run() {System.out.println(3332);}}
    public static class Impl3333 implements DemoService {public int type() {return 3333;} public void run() {System.out.println(3333);}}
    public static class Impl3334 implements DemoService {public int type() {return 3334;} public void run() {System.out.println(3334);}}
    public static class Impl3335 implements DemoService {public int type() {return 3335;} public void run() {System.out.println(3335);}}
    public static class Impl3336 implements DemoService {public int type() {return 3336;} public void run() {System.out.println(3336);}}
    public static class Impl3337 implements DemoService {public int type() {return 3337;} public void run() {System.out.println(3337);}}
    public static class Impl3338 implements DemoService {public int type() {return 3338;} public void run() {System.out.println(3338);}}
    public static class Impl3339 implements DemoService {public int type() {return 3339;} public void run() {System.out.println(3339);}}
    public static class Impl3340 implements DemoService {public int type() {return 3340;} public void run() {System.out.println(3340);}}
    public static class Impl3341 implements DemoService {public int type() {return 3341;} public void run() {System.out.println(3341);}}
    public static class Impl3342 implements DemoService {public int type() {return 3342;} public void run() {System.out.println(3342);}}
    public static class Impl3343 implements DemoService {public int type() {return 3343;} public void run() {System.out.println(3343);}}
    public static class Impl3344 implements DemoService {public int type() {return 3344;} public void run() {System.out.println(3344);}}
    public static class Impl3345 implements DemoService {public int type() {return 3345;} public void run() {System.out.println(3345);}}
    public static class Impl3346 implements DemoService {public int type() {return 3346;} public void run() {System.out.println(3346);}}
    public static class Impl3347 implements DemoService {public int type() {return 3347;} public void run() {System.out.println(3347);}}
    public static class Impl3348 implements DemoService {public int type() {return 3348;} public void run() {System.out.println(3348);}}
    public static class Impl3349 implements DemoService {public int type() {return 3349;} public void run() {System.out.println(3349);}}
    public static class Impl3350 implements DemoService {public int type() {return 3350;} public void run() {System.out.println(3350);}}
    public static class Impl3351 implements DemoService {public int type() {return 3351;} public void run() {System.out.println(3351);}}
    public static class Impl3352 implements DemoService {public int type() {return 3352;} public void run() {System.out.println(3352);}}
    public static class Impl3353 implements DemoService {public int type() {return 3353;} public void run() {System.out.println(3353);}}
    public static class Impl3354 implements DemoService {public int type() {return 3354;} public void run() {System.out.println(3354);}}
    public static class Impl3355 implements DemoService {public int type() {return 3355;} public void run() {System.out.println(3355);}}
    public static class Impl3356 implements DemoService {public int type() {return 3356;} public void run() {System.out.println(3356);}}
    public static class Impl3357 implements DemoService {public int type() {return 3357;} public void run() {System.out.println(3357);}}
    public static class Impl3358 implements DemoService {public int type() {return 3358;} public void run() {System.out.println(3358);}}
    public static class Impl3359 implements DemoService {public int type() {return 3359;} public void run() {System.out.println(3359);}}
    public static class Impl3360 implements DemoService {public int type() {return 3360;} public void run() {System.out.println(3360);}}
    public static class Impl3361 implements DemoService {public int type() {return 3361;} public void run() {System.out.println(3361);}}
    public static class Impl3362 implements DemoService {public int type() {return 3362;} public void run() {System.out.println(3362);}}
    public static class Impl3363 implements DemoService {public int type() {return 3363;} public void run() {System.out.println(3363);}}
    public static class Impl3364 implements DemoService {public int type() {return 3364;} public void run() {System.out.println(3364);}}
    public static class Impl3365 implements DemoService {public int type() {return 3365;} public void run() {System.out.println(3365);}}
    public static class Impl3366 implements DemoService {public int type() {return 3366;} public void run() {System.out.println(3366);}}
    public static class Impl3367 implements DemoService {public int type() {return 3367;} public void run() {System.out.println(3367);}}
    public static class Impl3368 implements DemoService {public int type() {return 3368;} public void run() {System.out.println(3368);}}
    public static class Impl3369 implements DemoService {public int type() {return 3369;} public void run() {System.out.println(3369);}}
    public static class Impl3370 implements DemoService {public int type() {return 3370;} public void run() {System.out.println(3370);}}
    public static class Impl3371 implements DemoService {public int type() {return 3371;} public void run() {System.out.println(3371);}}
    public static class Impl3372 implements DemoService {public int type() {return 3372;} public void run() {System.out.println(3372);}}
    public static class Impl3373 implements DemoService {public int type() {return 3373;} public void run() {System.out.println(3373);}}
    public static class Impl3374 implements DemoService {public int type() {return 3374;} public void run() {System.out.println(3374);}}
    public static class Impl3375 implements DemoService {public int type() {return 3375;} public void run() {System.out.println(3375);}}
    public static class Impl3376 implements DemoService {public int type() {return 3376;} public void run() {System.out.println(3376);}}
    public static class Impl3377 implements DemoService {public int type() {return 3377;} public void run() {System.out.println(3377);}}
    public static class Impl3378 implements DemoService {public int type() {return 3378;} public void run() {System.out.println(3378);}}
    public static class Impl3379 implements DemoService {public int type() {return 3379;} public void run() {System.out.println(3379);}}
    public static class Impl3380 implements DemoService {public int type() {return 3380;} public void run() {System.out.println(3380);}}
    public static class Impl3381 implements DemoService {public int type() {return 3381;} public void run() {System.out.println(3381);}}
    public static class Impl3382 implements DemoService {public int type() {return 3382;} public void run() {System.out.println(3382);}}
    public static class Impl3383 implements DemoService {public int type() {return 3383;} public void run() {System.out.println(3383);}}
    public static class Impl3384 implements DemoService {public int type() {return 3384;} public void run() {System.out.println(3384);}}
    public static class Impl3385 implements DemoService {public int type() {return 3385;} public void run() {System.out.println(3385);}}
    public static class Impl3386 implements DemoService {public int type() {return 3386;} public void run() {System.out.println(3386);}}
    public static class Impl3387 implements DemoService {public int type() {return 3387;} public void run() {System.out.println(3387);}}
    public static class Impl3388 implements DemoService {public int type() {return 3388;} public void run() {System.out.println(3388);}}
    public static class Impl3389 implements DemoService {public int type() {return 3389;} public void run() {System.out.println(3389);}}
    public static class Impl3390 implements DemoService {public int type() {return 3390;} public void run() {System.out.println(3390);}}
    public static class Impl3391 implements DemoService {public int type() {return 3391;} public void run() {System.out.println(3391);}}
    public static class Impl3392 implements DemoService {public int type() {return 3392;} public void run() {System.out.println(3392);}}
    public static class Impl3393 implements DemoService {public int type() {return 3393;} public void run() {System.out.println(3393);}}
    public static class Impl3394 implements DemoService {public int type() {return 3394;} public void run() {System.out.println(3394);}}
    public static class Impl3395 implements DemoService {public int type() {return 3395;} public void run() {System.out.println(3395);}}
    public static class Impl3396 implements DemoService {public int type() {return 3396;} public void run() {System.out.println(3396);}}
    public static class Impl3397 implements DemoService {public int type() {return 3397;} public void run() {System.out.println(3397);}}
    public static class Impl3398 implements DemoService {public int type() {return 3398;} public void run() {System.out.println(3398);}}
    public static class Impl3399 implements DemoService {public int type() {return 3399;} public void run() {System.out.println(3399);}}
    public static class Impl3400 implements DemoService {public int type() {return 3400;} public void run() {System.out.println(3400);}}
    public static class Impl3401 implements DemoService {public int type() {return 3401;} public void run() {System.out.println(3401);}}
    public static class Impl3402 implements DemoService {public int type() {return 3402;} public void run() {System.out.println(3402);}}
    public static class Impl3403 implements DemoService {public int type() {return 3403;} public void run() {System.out.println(3403);}}
    public static class Impl3404 implements DemoService {public int type() {return 3404;} public void run() {System.out.println(3404);}}
    public static class Impl3405 implements DemoService {public int type() {return 3405;} public void run() {System.out.println(3405);}}
    public static class Impl3406 implements DemoService {public int type() {return 3406;} public void run() {System.out.println(3406);}}
    public static class Impl3407 implements DemoService {public int type() {return 3407;} public void run() {System.out.println(3407);}}
    public static class Impl3408 implements DemoService {public int type() {return 3408;} public void run() {System.out.println(3408);}}
    public static class Impl3409 implements DemoService {public int type() {return 3409;} public void run() {System.out.println(3409);}}
    public static class Impl3410 implements DemoService {public int type() {return 3410;} public void run() {System.out.println(3410);}}
    public static class Impl3411 implements DemoService {public int type() {return 3411;} public void run() {System.out.println(3411);}}
    public static class Impl3412 implements DemoService {public int type() {return 3412;} public void run() {System.out.println(3412);}}
    public static class Impl3413 implements DemoService {public int type() {return 3413;} public void run() {System.out.println(3413);}}
    public static class Impl3414 implements DemoService {public int type() {return 3414;} public void run() {System.out.println(3414);}}
    public static class Impl3415 implements DemoService {public int type() {return 3415;} public void run() {System.out.println(3415);}}
    public static class Impl3416 implements DemoService {public int type() {return 3416;} public void run() {System.out.println(3416);}}
    public static class Impl3417 implements DemoService {public int type() {return 3417;} public void run() {System.out.println(3417);}}
    public static class Impl3418 implements DemoService {public int type() {return 3418;} public void run() {System.out.println(3418);}}
    public static class Impl3419 implements DemoService {public int type() {return 3419;} public void run() {System.out.println(3419);}}
    public static class Impl3420 implements DemoService {public int type() {return 3420;} public void run() {System.out.println(3420);}}
    public static class Impl3421 implements DemoService {public int type() {return 3421;} public void run() {System.out.println(3421);}}
    public static class Impl3422 implements DemoService {public int type() {return 3422;} public void run() {System.out.println(3422);}}
    public static class Impl3423 implements DemoService {public int type() {return 3423;} public void run() {System.out.println(3423);}}
    public static class Impl3424 implements DemoService {public int type() {return 3424;} public void run() {System.out.println(3424);}}
    public static class Impl3425 implements DemoService {public int type() {return 3425;} public void run() {System.out.println(3425);}}
    public static class Impl3426 implements DemoService {public int type() {return 3426;} public void run() {System.out.println(3426);}}
    public static class Impl3427 implements DemoService {public int type() {return 3427;} public void run() {System.out.println(3427);}}
    public static class Impl3428 implements DemoService {public int type() {return 3428;} public void run() {System.out.println(3428);}}
    public static class Impl3429 implements DemoService {public int type() {return 3429;} public void run() {System.out.println(3429);}}
    public static class Impl3430 implements DemoService {public int type() {return 3430;} public void run() {System.out.println(3430);}}
    public static class Impl3431 implements DemoService {public int type() {return 3431;} public void run() {System.out.println(3431);}}
    public static class Impl3432 implements DemoService {public int type() {return 3432;} public void run() {System.out.println(3432);}}
    public static class Impl3433 implements DemoService {public int type() {return 3433;} public void run() {System.out.println(3433);}}
    public static class Impl3434 implements DemoService {public int type() {return 3434;} public void run() {System.out.println(3434);}}
    public static class Impl3435 implements DemoService {public int type() {return 3435;} public void run() {System.out.println(3435);}}
    public static class Impl3436 implements DemoService {public int type() {return 3436;} public void run() {System.out.println(3436);}}
    public static class Impl3437 implements DemoService {public int type() {return 3437;} public void run() {System.out.println(3437);}}
    public static class Impl3438 implements DemoService {public int type() {return 3438;} public void run() {System.out.println(3438);}}
    public static class Impl3439 implements DemoService {public int type() {return 3439;} public void run() {System.out.println(3439);}}
    public static class Impl3440 implements DemoService {public int type() {return 3440;} public void run() {System.out.println(3440);}}
    public static class Impl3441 implements DemoService {public int type() {return 3441;} public void run() {System.out.println(3441);}}
    public static class Impl3442 implements DemoService {public int type() {return 3442;} public void run() {System.out.println(3442);}}
    public static class Impl3443 implements DemoService {public int type() {return 3443;} public void run() {System.out.println(3443);}}
    public static class Impl3444 implements DemoService {public int type() {return 3444;} public void run() {System.out.println(3444);}}
    public static class Impl3445 implements DemoService {public int type() {return 3445;} public void run() {System.out.println(3445);}}
    public static class Impl3446 implements DemoService {public int type() {return 3446;} public void run() {System.out.println(3446);}}
    public static class Impl3447 implements DemoService {public int type() {return 3447;} public void run() {System.out.println(3447);}}
    public static class Impl3448 implements DemoService {public int type() {return 3448;} public void run() {System.out.println(3448);}}
    public static class Impl3449 implements DemoService {public int type() {return 3449;} public void run() {System.out.println(3449);}}
    public static class Impl3450 implements DemoService {public int type() {return 3450;} public void run() {System.out.println(3450);}}
    public static class Impl3451 implements DemoService {public int type() {return 3451;} public void run() {System.out.println(3451);}}
    public static class Impl3452 implements DemoService {public int type() {return 3452;} public void run() {System.out.println(3452);}}
    public static class Impl3453 implements DemoService {public int type() {return 3453;} public void run() {System.out.println(3453);}}
    public static class Impl3454 implements DemoService {public int type() {return 3454;} public void run() {System.out.println(3454);}}
    public static class Impl3455 implements DemoService {public int type() {return 3455;} public void run() {System.out.println(3455);}}
    public static class Impl3456 implements DemoService {public int type() {return 3456;} public void run() {System.out.println(3456);}}
    public static class Impl3457 implements DemoService {public int type() {return 3457;} public void run() {System.out.println(3457);}}
    public static class Impl3458 implements DemoService {public int type() {return 3458;} public void run() {System.out.println(3458);}}
    public static class Impl3459 implements DemoService {public int type() {return 3459;} public void run() {System.out.println(3459);}}
    public static class Impl3460 implements DemoService {public int type() {return 3460;} public void run() {System.out.println(3460);}}
    public static class Impl3461 implements DemoService {public int type() {return 3461;} public void run() {System.out.println(3461);}}
    public static class Impl3462 implements DemoService {public int type() {return 3462;} public void run() {System.out.println(3462);}}
    public static class Impl3463 implements DemoService {public int type() {return 3463;} public void run() {System.out.println(3463);}}
    public static class Impl3464 implements DemoService {public int type() {return 3464;} public void run() {System.out.println(3464);}}
    public static class Impl3465 implements DemoService {public int type() {return 3465;} public void run() {System.out.println(3465);}}
    public static class Impl3466 implements DemoService {public int type() {return 3466;} public void run() {System.out.println(3466);}}
    public static class Impl3467 implements DemoService {public int type() {return 3467;} public void run() {System.out.println(3467);}}
    public static class Impl3468 implements DemoService {public int type() {return 3468;} public void run() {System.out.println(3468);}}
    public static class Impl3469 implements DemoService {public int type() {return 3469;} public void run() {System.out.println(3469);}}
    public static class Impl3470 implements DemoService {public int type() {return 3470;} public void run() {System.out.println(3470);}}
    public static class Impl3471 implements DemoService {public int type() {return 3471;} public void run() {System.out.println(3471);}}
    public static class Impl3472 implements DemoService {public int type() {return 3472;} public void run() {System.out.println(3472);}}
    public static class Impl3473 implements DemoService {public int type() {return 3473;} public void run() {System.out.println(3473);}}
    public static class Impl3474 implements DemoService {public int type() {return 3474;} public void run() {System.out.println(3474);}}
    public static class Impl3475 implements DemoService {public int type() {return 3475;} public void run() {System.out.println(3475);}}
    public static class Impl3476 implements DemoService {public int type() {return 3476;} public void run() {System.out.println(3476);}}
    public static class Impl3477 implements DemoService {public int type() {return 3477;} public void run() {System.out.println(3477);}}
    public static class Impl3478 implements DemoService {public int type() {return 3478;} public void run() {System.out.println(3478);}}
    public static class Impl3479 implements DemoService {public int type() {return 3479;} public void run() {System.out.println(3479);}}
    public static class Impl3480 implements DemoService {public int type() {return 3480;} public void run() {System.out.println(3480);}}
    public static class Impl3481 implements DemoService {public int type() {return 3481;} public void run() {System.out.println(3481);}}
    public static class Impl3482 implements DemoService {public int type() {return 3482;} public void run() {System.out.println(3482);}}
    public static class Impl3483 implements DemoService {public int type() {return 3483;} public void run() {System.out.println(3483);}}
    public static class Impl3484 implements DemoService {public int type() {return 3484;} public void run() {System.out.println(3484);}}
    public static class Impl3485 implements DemoService {public int type() {return 3485;} public void run() {System.out.println(3485);}}
    public static class Impl3486 implements DemoService {public int type() {return 3486;} public void run() {System.out.println(3486);}}
    public static class Impl3487 implements DemoService {public int type() {return 3487;} public void run() {System.out.println(3487);}}
    public static class Impl3488 implements DemoService {public int type() {return 3488;} public void run() {System.out.println(3488);}}
    public static class Impl3489 implements DemoService {public int type() {return 3489;} public void run() {System.out.println(3489);}}
    public static class Impl3490 implements DemoService {public int type() {return 3490;} public void run() {System.out.println(3490);}}
    public static class Impl3491 implements DemoService {public int type() {return 3491;} public void run() {System.out.println(3491);}}
    public static class Impl3492 implements DemoService {public int type() {return 3492;} public void run() {System.out.println(3492);}}
    public static class Impl3493 implements DemoService {public int type() {return 3493;} public void run() {System.out.println(3493);}}
    public static class Impl3494 implements DemoService {public int type() {return 3494;} public void run() {System.out.println(3494);}}
    public static class Impl3495 implements DemoService {public int type() {return 3495;} public void run() {System.out.println(3495);}}
    public static class Impl3496 implements DemoService {public int type() {return 3496;} public void run() {System.out.println(3496);}}
    public static class Impl3497 implements DemoService {public int type() {return 3497;} public void run() {System.out.println(3497);}}
    public static class Impl3498 implements DemoService {public int type() {return 3498;} public void run() {System.out.println(3498);}}
    public static class Impl3499 implements DemoService {public int type() {return 3499;} public void run() {System.out.println(3499);}}
    public static class Impl3500 implements DemoService {public int type() {return 3500;} public void run() {System.out.println(3500);}}
    public static class Impl3501 implements DemoService {public int type() {return 3501;} public void run() {System.out.println(3501);}}
    public static class Impl3502 implements DemoService {public int type() {return 3502;} public void run() {System.out.println(3502);}}
    public static class Impl3503 implements DemoService {public int type() {return 3503;} public void run() {System.out.println(3503);}}
    public static class Impl3504 implements DemoService {public int type() {return 3504;} public void run() {System.out.println(3504);}}
    public static class Impl3505 implements DemoService {public int type() {return 3505;} public void run() {System.out.println(3505);}}
    public static class Impl3506 implements DemoService {public int type() {return 3506;} public void run() {System.out.println(3506);}}
    public static class Impl3507 implements DemoService {public int type() {return 3507;} public void run() {System.out.println(3507);}}
    public static class Impl3508 implements DemoService {public int type() {return 3508;} public void run() {System.out.println(3508);}}
    public static class Impl3509 implements DemoService {public int type() {return 3509;} public void run() {System.out.println(3509);}}
    public static class Impl3510 implements DemoService {public int type() {return 3510;} public void run() {System.out.println(3510);}}
    public static class Impl3511 implements DemoService {public int type() {return 3511;} public void run() {System.out.println(3511);}}
    public static class Impl3512 implements DemoService {public int type() {return 3512;} public void run() {System.out.println(3512);}}
    public static class Impl3513 implements DemoService {public int type() {return 3513;} public void run() {System.out.println(3513);}}
    public static class Impl3514 implements DemoService {public int type() {return 3514;} public void run() {System.out.println(3514);}}
    public static class Impl3515 implements DemoService {public int type() {return 3515;} public void run() {System.out.println(3515);}}
    public static class Impl3516 implements DemoService {public int type() {return 3516;} public void run() {System.out.println(3516);}}
    public static class Impl3517 implements DemoService {public int type() {return 3517;} public void run() {System.out.println(3517);}}
    public static class Impl3518 implements DemoService {public int type() {return 3518;} public void run() {System.out.println(3518);}}
    public static class Impl3519 implements DemoService {public int type() {return 3519;} public void run() {System.out.println(3519);}}
    public static class Impl3520 implements DemoService {public int type() {return 3520;} public void run() {System.out.println(3520);}}
    public static class Impl3521 implements DemoService {public int type() {return 3521;} public void run() {System.out.println(3521);}}
    public static class Impl3522 implements DemoService {public int type() {return 3522;} public void run() {System.out.println(3522);}}
    public static class Impl3523 implements DemoService {public int type() {return 3523;} public void run() {System.out.println(3523);}}
    public static class Impl3524 implements DemoService {public int type() {return 3524;} public void run() {System.out.println(3524);}}
    public static class Impl3525 implements DemoService {public int type() {return 3525;} public void run() {System.out.println(3525);}}
    public static class Impl3526 implements DemoService {public int type() {return 3526;} public void run() {System.out.println(3526);}}
    public static class Impl3527 implements DemoService {public int type() {return 3527;} public void run() {System.out.println(3527);}}
    public static class Impl3528 implements DemoService {public int type() {return 3528;} public void run() {System.out.println(3528);}}
    public static class Impl3529 implements DemoService {public int type() {return 3529;} public void run() {System.out.println(3529);}}
    public static class Impl3530 implements DemoService {public int type() {return 3530;} public void run() {System.out.println(3530);}}
    public static class Impl3531 implements DemoService {public int type() {return 3531;} public void run() {System.out.println(3531);}}
    public static class Impl3532 implements DemoService {public int type() {return 3532;} public void run() {System.out.println(3532);}}
    public static class Impl3533 implements DemoService {public int type() {return 3533;} public void run() {System.out.println(3533);}}
    public static class Impl3534 implements DemoService {public int type() {return 3534;} public void run() {System.out.println(3534);}}
    public static class Impl3535 implements DemoService {public int type() {return 3535;} public void run() {System.out.println(3535);}}
    public static class Impl3536 implements DemoService {public int type() {return 3536;} public void run() {System.out.println(3536);}}
    public static class Impl3537 implements DemoService {public int type() {return 3537;} public void run() {System.out.println(3537);}}
    public static class Impl3538 implements DemoService {public int type() {return 3538;} public void run() {System.out.println(3538);}}
    public static class Impl3539 implements DemoService {public int type() {return 3539;} public void run() {System.out.println(3539);}}
    public static class Impl3540 implements DemoService {public int type() {return 3540;} public void run() {System.out.println(3540);}}
    public static class Impl3541 implements DemoService {public int type() {return 3541;} public void run() {System.out.println(3541);}}
    public static class Impl3542 implements DemoService {public int type() {return 3542;} public void run() {System.out.println(3542);}}
    public static class Impl3543 implements DemoService {public int type() {return 3543;} public void run() {System.out.println(3543);}}
    public static class Impl3544 implements DemoService {public int type() {return 3544;} public void run() {System.out.println(3544);}}
    public static class Impl3545 implements DemoService {public int type() {return 3545;} public void run() {System.out.println(3545);}}
    public static class Impl3546 implements DemoService {public int type() {return 3546;} public void run() {System.out.println(3546);}}
    public static class Impl3547 implements DemoService {public int type() {return 3547;} public void run() {System.out.println(3547);}}
    public static class Impl3548 implements DemoService {public int type() {return 3548;} public void run() {System.out.println(3548);}}
    public static class Impl3549 implements DemoService {public int type() {return 3549;} public void run() {System.out.println(3549);}}
    public static class Impl3550 implements DemoService {public int type() {return 3550;} public void run() {System.out.println(3550);}}
    public static class Impl3551 implements DemoService {public int type() {return 3551;} public void run() {System.out.println(3551);}}
    public static class Impl3552 implements DemoService {public int type() {return 3552;} public void run() {System.out.println(3552);}}
    public static class Impl3553 implements DemoService {public int type() {return 3553;} public void run() {System.out.println(3553);}}
    public static class Impl3554 implements DemoService {public int type() {return 3554;} public void run() {System.out.println(3554);}}
    public static class Impl3555 implements DemoService {public int type() {return 3555;} public void run() {System.out.println(3555);}}
    public static class Impl3556 implements DemoService {public int type() {return 3556;} public void run() {System.out.println(3556);}}
    public static class Impl3557 implements DemoService {public int type() {return 3557;} public void run() {System.out.println(3557);}}
    public static class Impl3558 implements DemoService {public int type() {return 3558;} public void run() {System.out.println(3558);}}
    public static class Impl3559 implements DemoService {public int type() {return 3559;} public void run() {System.out.println(3559);}}
    public static class Impl3560 implements DemoService {public int type() {return 3560;} public void run() {System.out.println(3560);}}
    public static class Impl3561 implements DemoService {public int type() {return 3561;} public void run() {System.out.println(3561);}}
    public static class Impl3562 implements DemoService {public int type() {return 3562;} public void run() {System.out.println(3562);}}
    public static class Impl3563 implements DemoService {public int type() {return 3563;} public void run() {System.out.println(3563);}}
    public static class Impl3564 implements DemoService {public int type() {return 3564;} public void run() {System.out.println(3564);}}
    public static class Impl3565 implements DemoService {public int type() {return 3565;} public void run() {System.out.println(3565);}}
    public static class Impl3566 implements DemoService {public int type() {return 3566;} public void run() {System.out.println(3566);}}
    public static class Impl3567 implements DemoService {public int type() {return 3567;} public void run() {System.out.println(3567);}}
    public static class Impl3568 implements DemoService {public int type() {return 3568;} public void run() {System.out.println(3568);}}
    public static class Impl3569 implements DemoService {public int type() {return 3569;} public void run() {System.out.println(3569);}}
    public static class Impl3570 implements DemoService {public int type() {return 3570;} public void run() {System.out.println(3570);}}
    public static class Impl3571 implements DemoService {public int type() {return 3571;} public void run() {System.out.println(3571);}}
    public static class Impl3572 implements DemoService {public int type() {return 3572;} public void run() {System.out.println(3572);}}
    public static class Impl3573 implements DemoService {public int type() {return 3573;} public void run() {System.out.println(3573);}}
    public static class Impl3574 implements DemoService {public int type() {return 3574;} public void run() {System.out.println(3574);}}
    public static class Impl3575 implements DemoService {public int type() {return 3575;} public void run() {System.out.println(3575);}}
    public static class Impl3576 implements DemoService {public int type() {return 3576;} public void run() {System.out.println(3576);}}
    public static class Impl3577 implements DemoService {public int type() {return 3577;} public void run() {System.out.println(3577);}}
    public static class Impl3578 implements DemoService {public int type() {return 3578;} public void run() {System.out.println(3578);}}
    public static class Impl3579 implements DemoService {public int type() {return 3579;} public void run() {System.out.println(3579);}}
    public static class Impl3580 implements DemoService {public int type() {return 3580;} public void run() {System.out.println(3580);}}
    public static class Impl3581 implements DemoService {public int type() {return 3581;} public void run() {System.out.println(3581);}}
    public static class Impl3582 implements DemoService {public int type() {return 3582;} public void run() {System.out.println(3582);}}
    public static class Impl3583 implements DemoService {public int type() {return 3583;} public void run() {System.out.println(3583);}}
    public static class Impl3584 implements DemoService {public int type() {return 3584;} public void run() {System.out.println(3584);}}
    public static class Impl3585 implements DemoService {public int type() {return 3585;} public void run() {System.out.println(3585);}}
    public static class Impl3586 implements DemoService {public int type() {return 3586;} public void run() {System.out.println(3586);}}
    public static class Impl3587 implements DemoService {public int type() {return 3587;} public void run() {System.out.println(3587);}}
    public static class Impl3588 implements DemoService {public int type() {return 3588;} public void run() {System.out.println(3588);}}
    public static class Impl3589 implements DemoService {public int type() {return 3589;} public void run() {System.out.println(3589);}}
    public static class Impl3590 implements DemoService {public int type() {return 3590;} public void run() {System.out.println(3590);}}
    public static class Impl3591 implements DemoService {public int type() {return 3591;} public void run() {System.out.println(3591);}}
    public static class Impl3592 implements DemoService {public int type() {return 3592;} public void run() {System.out.println(3592);}}
    public static class Impl3593 implements DemoService {public int type() {return 3593;} public void run() {System.out.println(3593);}}
    public static class Impl3594 implements DemoService {public int type() {return 3594;} public void run() {System.out.println(3594);}}
    public static class Impl3595 implements DemoService {public int type() {return 3595;} public void run() {System.out.println(3595);}}
    public static class Impl3596 implements DemoService {public int type() {return 3596;} public void run() {System.out.println(3596);}}
    public static class Impl3597 implements DemoService {public int type() {return 3597;} public void run() {System.out.println(3597);}}
    public static class Impl3598 implements DemoService {public int type() {return 3598;} public void run() {System.out.println(3598);}}
    public static class Impl3599 implements DemoService {public int type() {return 3599;} public void run() {System.out.println(3599);}}
    public static class Impl3600 implements DemoService {public int type() {return 3600;} public void run() {System.out.println(3600);}}
    public static class Impl3601 implements DemoService {public int type() {return 3601;} public void run() {System.out.println(3601);}}
    public static class Impl3602 implements DemoService {public int type() {return 3602;} public void run() {System.out.println(3602);}}
    public static class Impl3603 implements DemoService {public int type() {return 3603;} public void run() {System.out.println(3603);}}
    public static class Impl3604 implements DemoService {public int type() {return 3604;} public void run() {System.out.println(3604);}}
    public static class Impl3605 implements DemoService {public int type() {return 3605;} public void run() {System.out.println(3605);}}
    public static class Impl3606 implements DemoService {public int type() {return 3606;} public void run() {System.out.println(3606);}}
    public static class Impl3607 implements DemoService {public int type() {return 3607;} public void run() {System.out.println(3607);}}
    public static class Impl3608 implements DemoService {public int type() {return 3608;} public void run() {System.out.println(3608);}}
    public static class Impl3609 implements DemoService {public int type() {return 3609;} public void run() {System.out.println(3609);}}
    public static class Impl3610 implements DemoService {public int type() {return 3610;} public void run() {System.out.println(3610);}}
    public static class Impl3611 implements DemoService {public int type() {return 3611;} public void run() {System.out.println(3611);}}
    public static class Impl3612 implements DemoService {public int type() {return 3612;} public void run() {System.out.println(3612);}}
    public static class Impl3613 implements DemoService {public int type() {return 3613;} public void run() {System.out.println(3613);}}
    public static class Impl3614 implements DemoService {public int type() {return 3614;} public void run() {System.out.println(3614);}}
    public static class Impl3615 implements DemoService {public int type() {return 3615;} public void run() {System.out.println(3615);}}
    public static class Impl3616 implements DemoService {public int type() {return 3616;} public void run() {System.out.println(3616);}}
    public static class Impl3617 implements DemoService {public int type() {return 3617;} public void run() {System.out.println(3617);}}
    public static class Impl3618 implements DemoService {public int type() {return 3618;} public void run() {System.out.println(3618);}}
    public static class Impl3619 implements DemoService {public int type() {return 3619;} public void run() {System.out.println(3619);}}
    public static class Impl3620 implements DemoService {public int type() {return 3620;} public void run() {System.out.println(3620);}}
    public static class Impl3621 implements DemoService {public int type() {return 3621;} public void run() {System.out.println(3621);}}
    public static class Impl3622 implements DemoService {public int type() {return 3622;} public void run() {System.out.println(3622);}}
    public static class Impl3623 implements DemoService {public int type() {return 3623;} public void run() {System.out.println(3623);}}
    public static class Impl3624 implements DemoService {public int type() {return 3624;} public void run() {System.out.println(3624);}}
    public static class Impl3625 implements DemoService {public int type() {return 3625;} public void run() {System.out.println(3625);}}
    public static class Impl3626 implements DemoService {public int type() {return 3626;} public void run() {System.out.println(3626);}}
    public static class Impl3627 implements DemoService {public int type() {return 3627;} public void run() {System.out.println(3627);}}
    public static class Impl3628 implements DemoService {public int type() {return 3628;} public void run() {System.out.println(3628);}}
    public static class Impl3629 implements DemoService {public int type() {return 3629;} public void run() {System.out.println(3629);}}
    public static class Impl3630 implements DemoService {public int type() {return 3630;} public void run() {System.out.println(3630);}}
    public static class Impl3631 implements DemoService {public int type() {return 3631;} public void run() {System.out.println(3631);}}
    public static class Impl3632 implements DemoService {public int type() {return 3632;} public void run() {System.out.println(3632);}}
    public static class Impl3633 implements DemoService {public int type() {return 3633;} public void run() {System.out.println(3633);}}
    public static class Impl3634 implements DemoService {public int type() {return 3634;} public void run() {System.out.println(3634);}}
    public static class Impl3635 implements DemoService {public int type() {return 3635;} public void run() {System.out.println(3635);}}
    public static class Impl3636 implements DemoService {public int type() {return 3636;} public void run() {System.out.println(3636);}}
    public static class Impl3637 implements DemoService {public int type() {return 3637;} public void run() {System.out.println(3637);}}
    public static class Impl3638 implements DemoService {public int type() {return 3638;} public void run() {System.out.println(3638);}}
    public static class Impl3639 implements DemoService {public int type() {return 3639;} public void run() {System.out.println(3639);}}
    public static class Impl3640 implements DemoService {public int type() {return 3640;} public void run() {System.out.println(3640);}}
    public static class Impl3641 implements DemoService {public int type() {return 3641;} public void run() {System.out.println(3641);}}
    public static class Impl3642 implements DemoService {public int type() {return 3642;} public void run() {System.out.println(3642);}}
    public static class Impl3643 implements DemoService {public int type() {return 3643;} public void run() {System.out.println(3643);}}
    public static class Impl3644 implements DemoService {public int type() {return 3644;} public void run() {System.out.println(3644);}}
    public static class Impl3645 implements DemoService {public int type() {return 3645;} public void run() {System.out.println(3645);}}
    public static class Impl3646 implements DemoService {public int type() {return 3646;} public void run() {System.out.println(3646);}}
    public static class Impl3647 implements DemoService {public int type() {return 3647;} public void run() {System.out.println(3647);}}
    public static class Impl3648 implements DemoService {public int type() {return 3648;} public void run() {System.out.println(3648);}}
    public static class Impl3649 implements DemoService {public int type() {return 3649;} public void run() {System.out.println(3649);}}
    public static class Impl3650 implements DemoService {public int type() {return 3650;} public void run() {System.out.println(3650);}}
    public static class Impl3651 implements DemoService {public int type() {return 3651;} public void run() {System.out.println(3651);}}
    public static class Impl3652 implements DemoService {public int type() {return 3652;} public void run() {System.out.println(3652);}}
    public static class Impl3653 implements DemoService {public int type() {return 3653;} public void run() {System.out.println(3653);}}
    public static class Impl3654 implements DemoService {public int type() {return 3654;} public void run() {System.out.println(3654);}}
    public static class Impl3655 implements DemoService {public int type() {return 3655;} public void run() {System.out.println(3655);}}
    public static class Impl3656 implements DemoService {public int type() {return 3656;} public void run() {System.out.println(3656);}}
    public static class Impl3657 implements DemoService {public int type() {return 3657;} public void run() {System.out.println(3657);}}
    public static class Impl3658 implements DemoService {public int type() {return 3658;} public void run() {System.out.println(3658);}}
    public static class Impl3659 implements DemoService {public int type() {return 3659;} public void run() {System.out.println(3659);}}
    public static class Impl3660 implements DemoService {public int type() {return 3660;} public void run() {System.out.println(3660);}}
    public static class Impl3661 implements DemoService {public int type() {return 3661;} public void run() {System.out.println(3661);}}
    public static class Impl3662 implements DemoService {public int type() {return 3662;} public void run() {System.out.println(3662);}}
    public static class Impl3663 implements DemoService {public int type() {return 3663;} public void run() {System.out.println(3663);}}
    public static class Impl3664 implements DemoService {public int type() {return 3664;} public void run() {System.out.println(3664);}}
    public static class Impl3665 implements DemoService {public int type() {return 3665;} public void run() {System.out.println(3665);}}
    public static class Impl3666 implements DemoService {public int type() {return 3666;} public void run() {System.out.println(3666);}}
    public static class Impl3667 implements DemoService {public int type() {return 3667;} public void run() {System.out.println(3667);}}
    public static class Impl3668 implements DemoService {public int type() {return 3668;} public void run() {System.out.println(3668);}}
    public static class Impl3669 implements DemoService {public int type() {return 3669;} public void run() {System.out.println(3669);}}
    public static class Impl3670 implements DemoService {public int type() {return 3670;} public void run() {System.out.println(3670);}}
    public static class Impl3671 implements DemoService {public int type() {return 3671;} public void run() {System.out.println(3671);}}
    public static class Impl3672 implements DemoService {public int type() {return 3672;} public void run() {System.out.println(3672);}}
    public static class Impl3673 implements DemoService {public int type() {return 3673;} public void run() {System.out.println(3673);}}
    public static class Impl3674 implements DemoService {public int type() {return 3674;} public void run() {System.out.println(3674);}}
    public static class Impl3675 implements DemoService {public int type() {return 3675;} public void run() {System.out.println(3675);}}
    public static class Impl3676 implements DemoService {public int type() {return 3676;} public void run() {System.out.println(3676);}}
    public static class Impl3677 implements DemoService {public int type() {return 3677;} public void run() {System.out.println(3677);}}
    public static class Impl3678 implements DemoService {public int type() {return 3678;} public void run() {System.out.println(3678);}}
    public static class Impl3679 implements DemoService {public int type() {return 3679;} public void run() {System.out.println(3679);}}
    public static class Impl3680 implements DemoService {public int type() {return 3680;} public void run() {System.out.println(3680);}}
    public static class Impl3681 implements DemoService {public int type() {return 3681;} public void run() {System.out.println(3681);}}
    public static class Impl3682 implements DemoService {public int type() {return 3682;} public void run() {System.out.println(3682);}}
    public static class Impl3683 implements DemoService {public int type() {return 3683;} public void run() {System.out.println(3683);}}
    public static class Impl3684 implements DemoService {public int type() {return 3684;} public void run() {System.out.println(3684);}}
    public static class Impl3685 implements DemoService {public int type() {return 3685;} public void run() {System.out.println(3685);}}
    public static class Impl3686 implements DemoService {public int type() {return 3686;} public void run() {System.out.println(3686);}}
    public static class Impl3687 implements DemoService {public int type() {return 3687;} public void run() {System.out.println(3687);}}
    public static class Impl3688 implements DemoService {public int type() {return 3688;} public void run() {System.out.println(3688);}}
    public static class Impl3689 implements DemoService {public int type() {return 3689;} public void run() {System.out.println(3689);}}
    public static class Impl3690 implements DemoService {public int type() {return 3690;} public void run() {System.out.println(3690);}}
    public static class Impl3691 implements DemoService {public int type() {return 3691;} public void run() {System.out.println(3691);}}
    public static class Impl3692 implements DemoService {public int type() {return 3692;} public void run() {System.out.println(3692);}}
    public static class Impl3693 implements DemoService {public int type() {return 3693;} public void run() {System.out.println(3693);}}
    public static class Impl3694 implements DemoService {public int type() {return 3694;} public void run() {System.out.println(3694);}}
    public static class Impl3695 implements DemoService {public int type() {return 3695;} public void run() {System.out.println(3695);}}
    public static class Impl3696 implements DemoService {public int type() {return 3696;} public void run() {System.out.println(3696);}}
    public static class Impl3697 implements DemoService {public int type() {return 3697;} public void run() {System.out.println(3697);}}
    public static class Impl3698 implements DemoService {public int type() {return 3698;} public void run() {System.out.println(3698);}}
    public static class Impl3699 implements DemoService {public int type() {return 3699;} public void run() {System.out.println(3699);}}
    public static class Impl3700 implements DemoService {public int type() {return 3700;} public void run() {System.out.println(3700);}}
    public static class Impl3701 implements DemoService {public int type() {return 3701;} public void run() {System.out.println(3701);}}
    public static class Impl3702 implements DemoService {public int type() {return 3702;} public void run() {System.out.println(3702);}}
    public static class Impl3703 implements DemoService {public int type() {return 3703;} public void run() {System.out.println(3703);}}
    public static class Impl3704 implements DemoService {public int type() {return 3704;} public void run() {System.out.println(3704);}}
    public static class Impl3705 implements DemoService {public int type() {return 3705;} public void run() {System.out.println(3705);}}
    public static class Impl3706 implements DemoService {public int type() {return 3706;} public void run() {System.out.println(3706);}}
    public static class Impl3707 implements DemoService {public int type() {return 3707;} public void run() {System.out.println(3707);}}
    public static class Impl3708 implements DemoService {public int type() {return 3708;} public void run() {System.out.println(3708);}}
    public static class Impl3709 implements DemoService {public int type() {return 3709;} public void run() {System.out.println(3709);}}
    public static class Impl3710 implements DemoService {public int type() {return 3710;} public void run() {System.out.println(3710);}}
    public static class Impl3711 implements DemoService {public int type() {return 3711;} public void run() {System.out.println(3711);}}
    public static class Impl3712 implements DemoService {public int type() {return 3712;} public void run() {System.out.println(3712);}}
    public static class Impl3713 implements DemoService {public int type() {return 3713;} public void run() {System.out.println(3713);}}
    public static class Impl3714 implements DemoService {public int type() {return 3714;} public void run() {System.out.println(3714);}}
    public static class Impl3715 implements DemoService {public int type() {return 3715;} public void run() {System.out.println(3715);}}
    public static class Impl3716 implements DemoService {public int type() {return 3716;} public void run() {System.out.println(3716);}}
    public static class Impl3717 implements DemoService {public int type() {return 3717;} public void run() {System.out.println(3717);}}
    public static class Impl3718 implements DemoService {public int type() {return 3718;} public void run() {System.out.println(3718);}}
    public static class Impl3719 implements DemoService {public int type() {return 3719;} public void run() {System.out.println(3719);}}
    public static class Impl3720 implements DemoService {public int type() {return 3720;} public void run() {System.out.println(3720);}}
    public static class Impl3721 implements DemoService {public int type() {return 3721;} public void run() {System.out.println(3721);}}
    public static class Impl3722 implements DemoService {public int type() {return 3722;} public void run() {System.out.println(3722);}}
    public static class Impl3723 implements DemoService {public int type() {return 3723;} public void run() {System.out.println(3723);}}
    public static class Impl3724 implements DemoService {public int type() {return 3724;} public void run() {System.out.println(3724);}}
    public static class Impl3725 implements DemoService {public int type() {return 3725;} public void run() {System.out.println(3725);}}
    public static class Impl3726 implements DemoService {public int type() {return 3726;} public void run() {System.out.println(3726);}}
    public static class Impl3727 implements DemoService {public int type() {return 3727;} public void run() {System.out.println(3727);}}
    public static class Impl3728 implements DemoService {public int type() {return 3728;} public void run() {System.out.println(3728);}}
    public static class Impl3729 implements DemoService {public int type() {return 3729;} public void run() {System.out.println(3729);}}
    public static class Impl3730 implements DemoService {public int type() {return 3730;} public void run() {System.out.println(3730);}}
    public static class Impl3731 implements DemoService {public int type() {return 3731;} public void run() {System.out.println(3731);}}
    public static class Impl3732 implements DemoService {public int type() {return 3732;} public void run() {System.out.println(3732);}}
    public static class Impl3733 implements DemoService {public int type() {return 3733;} public void run() {System.out.println(3733);}}
    public static class Impl3734 implements DemoService {public int type() {return 3734;} public void run() {System.out.println(3734);}}
    public static class Impl3735 implements DemoService {public int type() {return 3735;} public void run() {System.out.println(3735);}}
    public static class Impl3736 implements DemoService {public int type() {return 3736;} public void run() {System.out.println(3736);}}
    public static class Impl3737 implements DemoService {public int type() {return 3737;} public void run() {System.out.println(3737);}}
    public static class Impl3738 implements DemoService {public int type() {return 3738;} public void run() {System.out.println(3738);}}
    public static class Impl3739 implements DemoService {public int type() {return 3739;} public void run() {System.out.println(3739);}}
    public static class Impl3740 implements DemoService {public int type() {return 3740;} public void run() {System.out.println(3740);}}
    public static class Impl3741 implements DemoService {public int type() {return 3741;} public void run() {System.out.println(3741);}}
    public static class Impl3742 implements DemoService {public int type() {return 3742;} public void run() {System.out.println(3742);}}
    public static class Impl3743 implements DemoService {public int type() {return 3743;} public void run() {System.out.println(3743);}}
    public static class Impl3744 implements DemoService {public int type() {return 3744;} public void run() {System.out.println(3744);}}
    public static class Impl3745 implements DemoService {public int type() {return 3745;} public void run() {System.out.println(3745);}}
    public static class Impl3746 implements DemoService {public int type() {return 3746;} public void run() {System.out.println(3746);}}
    public static class Impl3747 implements DemoService {public int type() {return 3747;} public void run() {System.out.println(3747);}}
    public static class Impl3748 implements DemoService {public int type() {return 3748;} public void run() {System.out.println(3748);}}
    public static class Impl3749 implements DemoService {public int type() {return 3749;} public void run() {System.out.println(3749);}}
    public static class Impl3750 implements DemoService {public int type() {return 3750;} public void run() {System.out.println(3750);}}
    public static class Impl3751 implements DemoService {public int type() {return 3751;} public void run() {System.out.println(3751);}}
    public static class Impl3752 implements DemoService {public int type() {return 3752;} public void run() {System.out.println(3752);}}
    public static class Impl3753 implements DemoService {public int type() {return 3753;} public void run() {System.out.println(3753);}}
    public static class Impl3754 implements DemoService {public int type() {return 3754;} public void run() {System.out.println(3754);}}
    public static class Impl3755 implements DemoService {public int type() {return 3755;} public void run() {System.out.println(3755);}}
    public static class Impl3756 implements DemoService {public int type() {return 3756;} public void run() {System.out.println(3756);}}
    public static class Impl3757 implements DemoService {public int type() {return 3757;} public void run() {System.out.println(3757);}}
    public static class Impl3758 implements DemoService {public int type() {return 3758;} public void run() {System.out.println(3758);}}
    public static class Impl3759 implements DemoService {public int type() {return 3759;} public void run() {System.out.println(3759);}}
    public static class Impl3760 implements DemoService {public int type() {return 3760;} public void run() {System.out.println(3760);}}
    public static class Impl3761 implements DemoService {public int type() {return 3761;} public void run() {System.out.println(3761);}}
    public static class Impl3762 implements DemoService {public int type() {return 3762;} public void run() {System.out.println(3762);}}
    public static class Impl3763 implements DemoService {public int type() {return 3763;} public void run() {System.out.println(3763);}}
    public static class Impl3764 implements DemoService {public int type() {return 3764;} public void run() {System.out.println(3764);}}
    public static class Impl3765 implements DemoService {public int type() {return 3765;} public void run() {System.out.println(3765);}}
    public static class Impl3766 implements DemoService {public int type() {return 3766;} public void run() {System.out.println(3766);}}
    public static class Impl3767 implements DemoService {public int type() {return 3767;} public void run() {System.out.println(3767);}}
    public static class Impl3768 implements DemoService {public int type() {return 3768;} public void run() {System.out.println(3768);}}
    public static class Impl3769 implements DemoService {public int type() {return 3769;} public void run() {System.out.println(3769);}}
    public static class Impl3770 implements DemoService {public int type() {return 3770;} public void run() {System.out.println(3770);}}
    public static class Impl3771 implements DemoService {public int type() {return 3771;} public void run() {System.out.println(3771);}}
    public static class Impl3772 implements DemoService {public int type() {return 3772;} public void run() {System.out.println(3772);}}
    public static class Impl3773 implements DemoService {public int type() {return 3773;} public void run() {System.out.println(3773);}}
    public static class Impl3774 implements DemoService {public int type() {return 3774;} public void run() {System.out.println(3774);}}
    public static class Impl3775 implements DemoService {public int type() {return 3775;} public void run() {System.out.println(3775);}}
    public static class Impl3776 implements DemoService {public int type() {return 3776;} public void run() {System.out.println(3776);}}
    public static class Impl3777 implements DemoService {public int type() {return 3777;} public void run() {System.out.println(3777);}}
    public static class Impl3778 implements DemoService {public int type() {return 3778;} public void run() {System.out.println(3778);}}
    public static class Impl3779 implements DemoService {public int type() {return 3779;} public void run() {System.out.println(3779);}}
    public static class Impl3780 implements DemoService {public int type() {return 3780;} public void run() {System.out.println(3780);}}
    public static class Impl3781 implements DemoService {public int type() {return 3781;} public void run() {System.out.println(3781);}}
    public static class Impl3782 implements DemoService {public int type() {return 3782;} public void run() {System.out.println(3782);}}
    public static class Impl3783 implements DemoService {public int type() {return 3783;} public void run() {System.out.println(3783);}}
    public static class Impl3784 implements DemoService {public int type() {return 3784;} public void run() {System.out.println(3784);}}
    public static class Impl3785 implements DemoService {public int type() {return 3785;} public void run() {System.out.println(3785);}}
    public static class Impl3786 implements DemoService {public int type() {return 3786;} public void run() {System.out.println(3786);}}
    public static class Impl3787 implements DemoService {public int type() {return 3787;} public void run() {System.out.println(3787);}}
    public static class Impl3788 implements DemoService {public int type() {return 3788;} public void run() {System.out.println(3788);}}
    public static class Impl3789 implements DemoService {public int type() {return 3789;} public void run() {System.out.println(3789);}}
    public static class Impl3790 implements DemoService {public int type() {return 3790;} public void run() {System.out.println(3790);}}
    public static class Impl3791 implements DemoService {public int type() {return 3791;} public void run() {System.out.println(3791);}}
    public static class Impl3792 implements DemoService {public int type() {return 3792;} public void run() {System.out.println(3792);}}
    public static class Impl3793 implements DemoService {public int type() {return 3793;} public void run() {System.out.println(3793);}}
    public static class Impl3794 implements DemoService {public int type() {return 3794;} public void run() {System.out.println(3794);}}
    public static class Impl3795 implements DemoService {public int type() {return 3795;} public void run() {System.out.println(3795);}}
    public static class Impl3796 implements DemoService {public int type() {return 3796;} public void run() {System.out.println(3796);}}
    public static class Impl3797 implements DemoService {public int type() {return 3797;} public void run() {System.out.println(3797);}}
    public static class Impl3798 implements DemoService {public int type() {return 3798;} public void run() {System.out.println(3798);}}
    public static class Impl3799 implements DemoService {public int type() {return 3799;} public void run() {System.out.println(3799);}}
    public static class Impl3800 implements DemoService {public int type() {return 3800;} public void run() {System.out.println(3800);}}
    public static class Impl3801 implements DemoService {public int type() {return 3801;} public void run() {System.out.println(3801);}}
    public static class Impl3802 implements DemoService {public int type() {return 3802;} public void run() {System.out.println(3802);}}
    public static class Impl3803 implements DemoService {public int type() {return 3803;} public void run() {System.out.println(3803);}}
    public static class Impl3804 implements DemoService {public int type() {return 3804;} public void run() {System.out.println(3804);}}
    public static class Impl3805 implements DemoService {public int type() {return 3805;} public void run() {System.out.println(3805);}}
    public static class Impl3806 implements DemoService {public int type() {return 3806;} public void run() {System.out.println(3806);}}
    public static class Impl3807 implements DemoService {public int type() {return 3807;} public void run() {System.out.println(3807);}}
    public static class Impl3808 implements DemoService {public int type() {return 3808;} public void run() {System.out.println(3808);}}
    public static class Impl3809 implements DemoService {public int type() {return 3809;} public void run() {System.out.println(3809);}}
    public static class Impl3810 implements DemoService {public int type() {return 3810;} public void run() {System.out.println(3810);}}
    public static class Impl3811 implements DemoService {public int type() {return 3811;} public void run() {System.out.println(3811);}}
    public static class Impl3812 implements DemoService {public int type() {return 3812;} public void run() {System.out.println(3812);}}
    public static class Impl3813 implements DemoService {public int type() {return 3813;} public void run() {System.out.println(3813);}}
    public static class Impl3814 implements DemoService {public int type() {return 3814;} public void run() {System.out.println(3814);}}
    public static class Impl3815 implements DemoService {public int type() {return 3815;} public void run() {System.out.println(3815);}}
    public static class Impl3816 implements DemoService {public int type() {return 3816;} public void run() {System.out.println(3816);}}
    public static class Impl3817 implements DemoService {public int type() {return 3817;} public void run() {System.out.println(3817);}}
    public static class Impl3818 implements DemoService {public int type() {return 3818;} public void run() {System.out.println(3818);}}
    public static class Impl3819 implements DemoService {public int type() {return 3819;} public void run() {System.out.println(3819);}}
    public static class Impl3820 implements DemoService {public int type() {return 3820;} public void run() {System.out.println(3820);}}
    public static class Impl3821 implements DemoService {public int type() {return 3821;} public void run() {System.out.println(3821);}}
    public static class Impl3822 implements DemoService {public int type() {return 3822;} public void run() {System.out.println(3822);}}
    public static class Impl3823 implements DemoService {public int type() {return 3823;} public void run() {System.out.println(3823);}}
    public static class Impl3824 implements DemoService {public int type() {return 3824;} public void run() {System.out.println(3824);}}
    public static class Impl3825 implements DemoService {public int type() {return 3825;} public void run() {System.out.println(3825);}}
    public static class Impl3826 implements DemoService {public int type() {return 3826;} public void run() {System.out.println(3826);}}
    public static class Impl3827 implements DemoService {public int type() {return 3827;} public void run() {System.out.println(3827);}}
    public static class Impl3828 implements DemoService {public int type() {return 3828;} public void run() {System.out.println(3828);}}
    public static class Impl3829 implements DemoService {public int type() {return 3829;} public void run() {System.out.println(3829);}}
    public static class Impl3830 implements DemoService {public int type() {return 3830;} public void run() {System.out.println(3830);}}
    public static class Impl3831 implements DemoService {public int type() {return 3831;} public void run() {System.out.println(3831);}}
    public static class Impl3832 implements DemoService {public int type() {return 3832;} public void run() {System.out.println(3832);}}
    public static class Impl3833 implements DemoService {public int type() {return 3833;} public void run() {System.out.println(3833);}}
    public static class Impl3834 implements DemoService {public int type() {return 3834;} public void run() {System.out.println(3834);}}
    public static class Impl3835 implements DemoService {public int type() {return 3835;} public void run() {System.out.println(3835);}}
    public static class Impl3836 implements DemoService {public int type() {return 3836;} public void run() {System.out.println(3836);}}
    public static class Impl3837 implements DemoService {public int type() {return 3837;} public void run() {System.out.println(3837);}}
    public static class Impl3838 implements DemoService {public int type() {return 3838;} public void run() {System.out.println(3838);}}
    public static class Impl3839 implements DemoService {public int type() {return 3839;} public void run() {System.out.println(3839);}}
    public static class Impl3840 implements DemoService {public int type() {return 3840;} public void run() {System.out.println(3840);}}
    public static class Impl3841 implements DemoService {public int type() {return 3841;} public void run() {System.out.println(3841);}}
    public static class Impl3842 implements DemoService {public int type() {return 3842;} public void run() {System.out.println(3842);}}
    public static class Impl3843 implements DemoService {public int type() {return 3843;} public void run() {System.out.println(3843);}}
    public static class Impl3844 implements DemoService {public int type() {return 3844;} public void run() {System.out.println(3844);}}
    public static class Impl3845 implements DemoService {public int type() {return 3845;} public void run() {System.out.println(3845);}}
    public static class Impl3846 implements DemoService {public int type() {return 3846;} public void run() {System.out.println(3846);}}
    public static class Impl3847 implements DemoService {public int type() {return 3847;} public void run() {System.out.println(3847);}}
    public static class Impl3848 implements DemoService {public int type() {return 3848;} public void run() {System.out.println(3848);}}
    public static class Impl3849 implements DemoService {public int type() {return 3849;} public void run() {System.out.println(3849);}}
    public static class Impl3850 implements DemoService {public int type() {return 3850;} public void run() {System.out.println(3850);}}
    public static class Impl3851 implements DemoService {public int type() {return 3851;} public void run() {System.out.println(3851);}}
    public static class Impl3852 implements DemoService {public int type() {return 3852;} public void run() {System.out.println(3852);}}
    public static class Impl3853 implements DemoService {public int type() {return 3853;} public void run() {System.out.println(3853);}}
    public static class Impl3854 implements DemoService {public int type() {return 3854;} public void run() {System.out.println(3854);}}
    public static class Impl3855 implements DemoService {public int type() {return 3855;} public void run() {System.out.println(3855);}}
    public static class Impl3856 implements DemoService {public int type() {return 3856;} public void run() {System.out.println(3856);}}
    public static class Impl3857 implements DemoService {public int type() {return 3857;} public void run() {System.out.println(3857);}}
    public static class Impl3858 implements DemoService {public int type() {return 3858;} public void run() {System.out.println(3858);}}
    public static class Impl3859 implements DemoService {public int type() {return 3859;} public void run() {System.out.println(3859);}}
    public static class Impl3860 implements DemoService {public int type() {return 3860;} public void run() {System.out.println(3860);}}
    public static class Impl3861 implements DemoService {public int type() {return 3861;} public void run() {System.out.println(3861);}}
    public static class Impl3862 implements DemoService {public int type() {return 3862;} public void run() {System.out.println(3862);}}
    public static class Impl3863 implements DemoService {public int type() {return 3863;} public void run() {System.out.println(3863);}}
    public static class Impl3864 implements DemoService {public int type() {return 3864;} public void run() {System.out.println(3864);}}
    public static class Impl3865 implements DemoService {public int type() {return 3865;} public void run() {System.out.println(3865);}}
    public static class Impl3866 implements DemoService {public int type() {return 3866;} public void run() {System.out.println(3866);}}
    public static class Impl3867 implements DemoService {public int type() {return 3867;} public void run() {System.out.println(3867);}}
    public static class Impl3868 implements DemoService {public int type() {return 3868;} public void run() {System.out.println(3868);}}
    public static class Impl3869 implements DemoService {public int type() {return 3869;} public void run() {System.out.println(3869);}}
    public static class Impl3870 implements DemoService {public int type() {return 3870;} public void run() {System.out.println(3870);}}
    public static class Impl3871 implements DemoService {public int type() {return 3871;} public void run() {System.out.println(3871);}}
    public static class Impl3872 implements DemoService {public int type() {return 3872;} public void run() {System.out.println(3872);}}
    public static class Impl3873 implements DemoService {public int type() {return 3873;} public void run() {System.out.println(3873);}}
    public static class Impl3874 implements DemoService {public int type() {return 3874;} public void run() {System.out.println(3874);}}
    public static class Impl3875 implements DemoService {public int type() {return 3875;} public void run() {System.out.println(3875);}}
    public static class Impl3876 implements DemoService {public int type() {return 3876;} public void run() {System.out.println(3876);}}
    public static class Impl3877 implements DemoService {public int type() {return 3877;} public void run() {System.out.println(3877);}}
    public static class Impl3878 implements DemoService {public int type() {return 3878;} public void run() {System.out.println(3878);}}
    public static class Impl3879 implements DemoService {public int type() {return 3879;} public void run() {System.out.println(3879);}}
    public static class Impl3880 implements DemoService {public int type() {return 3880;} public void run() {System.out.println(3880);}}
    public static class Impl3881 implements DemoService {public int type() {return 3881;} public void run() {System.out.println(3881);}}
    public static class Impl3882 implements DemoService {public int type() {return 3882;} public void run() {System.out.println(3882);}}
    public static class Impl3883 implements DemoService {public int type() {return 3883;} public void run() {System.out.println(3883);}}
    public static class Impl3884 implements DemoService {public int type() {return 3884;} public void run() {System.out.println(3884);}}
    public static class Impl3885 implements DemoService {public int type() {return 3885;} public void run() {System.out.println(3885);}}
    public static class Impl3886 implements DemoService {public int type() {return 3886;} public void run() {System.out.println(3886);}}
    public static class Impl3887 implements DemoService {public int type() {return 3887;} public void run() {System.out.println(3887);}}
    public static class Impl3888 implements DemoService {public int type() {return 3888;} public void run() {System.out.println(3888);}}
    public static class Impl3889 implements DemoService {public int type() {return 3889;} public void run() {System.out.println(3889);}}
    public static class Impl3890 implements DemoService {public int type() {return 3890;} public void run() {System.out.println(3890);}}
    public static class Impl3891 implements DemoService {public int type() {return 3891;} public void run() {System.out.println(3891);}}
    public static class Impl3892 implements DemoService {public int type() {return 3892;} public void run() {System.out.println(3892);}}
    public static class Impl3893 implements DemoService {public int type() {return 3893;} public void run() {System.out.println(3893);}}
    public static class Impl3894 implements DemoService {public int type() {return 3894;} public void run() {System.out.println(3894);}}
    public static class Impl3895 implements DemoService {public int type() {return 3895;} public void run() {System.out.println(3895);}}
    public static class Impl3896 implements DemoService {public int type() {return 3896;} public void run() {System.out.println(3896);}}
    public static class Impl3897 implements DemoService {public int type() {return 3897;} public void run() {System.out.println(3897);}}
    public static class Impl3898 implements DemoService {public int type() {return 3898;} public void run() {System.out.println(3898);}}
    public static class Impl3899 implements DemoService {public int type() {return 3899;} public void run() {System.out.println(3899);}}
    public static class Impl3900 implements DemoService {public int type() {return 3900;} public void run() {System.out.println(3900);}}
    public static class Impl3901 implements DemoService {public int type() {return 3901;} public void run() {System.out.println(3901);}}
    public static class Impl3902 implements DemoService {public int type() {return 3902;} public void run() {System.out.println(3902);}}
    public static class Impl3903 implements DemoService {public int type() {return 3903;} public void run() {System.out.println(3903);}}
    public static class Impl3904 implements DemoService {public int type() {return 3904;} public void run() {System.out.println(3904);}}
    public static class Impl3905 implements DemoService {public int type() {return 3905;} public void run() {System.out.println(3905);}}
    public static class Impl3906 implements DemoService {public int type() {return 3906;} public void run() {System.out.println(3906);}}
    public static class Impl3907 implements DemoService {public int type() {return 3907;} public void run() {System.out.println(3907);}}
    public static class Impl3908 implements DemoService {public int type() {return 3908;} public void run() {System.out.println(3908);}}
    public static class Impl3909 implements DemoService {public int type() {return 3909;} public void run() {System.out.println(3909);}}
    public static class Impl3910 implements DemoService {public int type() {return 3910;} public void run() {System.out.println(3910);}}
    public static class Impl3911 implements DemoService {public int type() {return 3911;} public void run() {System.out.println(3911);}}
    public static class Impl3912 implements DemoService {public int type() {return 3912;} public void run() {System.out.println(3912);}}
    public static class Impl3913 implements DemoService {public int type() {return 3913;} public void run() {System.out.println(3913);}}
    public static class Impl3914 implements DemoService {public int type() {return 3914;} public void run() {System.out.println(3914);}}
    public static class Impl3915 implements DemoService {public int type() {return 3915;} public void run() {System.out.println(3915);}}
    public static class Impl3916 implements DemoService {public int type() {return 3916;} public void run() {System.out.println(3916);}}
    public static class Impl3917 implements DemoService {public int type() {return 3917;} public void run() {System.out.println(3917);}}
    public static class Impl3918 implements DemoService {public int type() {return 3918;} public void run() {System.out.println(3918);}}
    public static class Impl3919 implements DemoService {public int type() {return 3919;} public void run() {System.out.println(3919);}}
    public static class Impl3920 implements DemoService {public int type() {return 3920;} public void run() {System.out.println(3920);}}
    public static class Impl3921 implements DemoService {public int type() {return 3921;} public void run() {System.out.println(3921);}}
    public static class Impl3922 implements DemoService {public int type() {return 3922;} public void run() {System.out.println(3922);}}
    public static class Impl3923 implements DemoService {public int type() {return 3923;} public void run() {System.out.println(3923);}}
    public static class Impl3924 implements DemoService {public int type() {return 3924;} public void run() {System.out.println(3924);}}
    public static class Impl3925 implements DemoService {public int type() {return 3925;} public void run() {System.out.println(3925);}}
    public static class Impl3926 implements DemoService {public int type() {return 3926;} public void run() {System.out.println(3926);}}
    public static class Impl3927 implements DemoService {public int type() {return 3927;} public void run() {System.out.println(3927);}}
    public static class Impl3928 implements DemoService {public int type() {return 3928;} public void run() {System.out.println(3928);}}
    public static class Impl3929 implements DemoService {public int type() {return 3929;} public void run() {System.out.println(3929);}}
    public static class Impl3930 implements DemoService {public int type() {return 3930;} public void run() {System.out.println(3930);}}
    public static class Impl3931 implements DemoService {public int type() {return 3931;} public void run() {System.out.println(3931);}}
    public static class Impl3932 implements DemoService {public int type() {return 3932;} public void run() {System.out.println(3932);}}
    public static class Impl3933 implements DemoService {public int type() {return 3933;} public void run() {System.out.println(3933);}}
    public static class Impl3934 implements DemoService {public int type() {return 3934;} public void run() {System.out.println(3934);}}
    public static class Impl3935 implements DemoService {public int type() {return 3935;} public void run() {System.out.println(3935);}}
    public static class Impl3936 implements DemoService {public int type() {return 3936;} public void run() {System.out.println(3936);}}
    public static class Impl3937 implements DemoService {public int type() {return 3937;} public void run() {System.out.println(3937);}}
    public static class Impl3938 implements DemoService {public int type() {return 3938;} public void run() {System.out.println(3938);}}
    public static class Impl3939 implements DemoService {public int type() {return 3939;} public void run() {System.out.println(3939);}}
    public static class Impl3940 implements DemoService {public int type() {return 3940;} public void run() {System.out.println(3940);}}
    public static class Impl3941 implements DemoService {public int type() {return 3941;} public void run() {System.out.println(3941);}}
    public static class Impl3942 implements DemoService {public int type() {return 3942;} public void run() {System.out.println(3942);}}
    public static class Impl3943 implements DemoService {public int type() {return 3943;} public void run() {System.out.println(3943);}}
    public static class Impl3944 implements DemoService {public int type() {return 3944;} public void run() {System.out.println(3944);}}
    public static class Impl3945 implements DemoService {public int type() {return 3945;} public void run() {System.out.println(3945);}}
    public static class Impl3946 implements DemoService {public int type() {return 3946;} public void run() {System.out.println(3946);}}
    public static class Impl3947 implements DemoService {public int type() {return 3947;} public void run() {System.out.println(3947);}}
    public static class Impl3948 implements DemoService {public int type() {return 3948;} public void run() {System.out.println(3948);}}
    public static class Impl3949 implements DemoService {public int type() {return 3949;} public void run() {System.out.println(3949);}}
    public static class Impl3950 implements DemoService {public int type() {return 3950;} public void run() {System.out.println(3950);}}
    public static class Impl3951 implements DemoService {public int type() {return 3951;} public void run() {System.out.println(3951);}}
    public static class Impl3952 implements DemoService {public int type() {return 3952;} public void run() {System.out.println(3952);}}
    public static class Impl3953 implements DemoService {public int type() {return 3953;} public void run() {System.out.println(3953);}}
    public static class Impl3954 implements DemoService {public int type() {return 3954;} public void run() {System.out.println(3954);}}
    public static class Impl3955 implements DemoService {public int type() {return 3955;} public void run() {System.out.println(3955);}}
    public static class Impl3956 implements DemoService {public int type() {return 3956;} public void run() {System.out.println(3956);}}
    public static class Impl3957 implements DemoService {public int type() {return 3957;} public void run() {System.out.println(3957);}}
    public static class Impl3958 implements DemoService {public int type() {return 3958;} public void run() {System.out.println(3958);}}
    public static class Impl3959 implements DemoService {public int type() {return 3959;} public void run() {System.out.println(3959);}}
    public static class Impl3960 implements DemoService {public int type() {return 3960;} public void run() {System.out.println(3960);}}
    public static class Impl3961 implements DemoService {public int type() {return 3961;} public void run() {System.out.println(3961);}}
    public static class Impl3962 implements DemoService {public int type() {return 3962;} public void run() {System.out.println(3962);}}
    public static class Impl3963 implements DemoService {public int type() {return 3963;} public void run() {System.out.println(3963);}}
    public static class Impl3964 implements DemoService {public int type() {return 3964;} public void run() {System.out.println(3964);}}
    public static class Impl3965 implements DemoService {public int type() {return 3965;} public void run() {System.out.println(3965);}}
    public static class Impl3966 implements DemoService {public int type() {return 3966;} public void run() {System.out.println(3966);}}
    public static class Impl3967 implements DemoService {public int type() {return 3967;} public void run() {System.out.println(3967);}}
    public static class Impl3968 implements DemoService {public int type() {return 3968;} public void run() {System.out.println(3968);}}
    public static class Impl3969 implements DemoService {public int type() {return 3969;} public void run() {System.out.println(3969);}}
    public static class Impl3970 implements DemoService {public int type() {return 3970;} public void run() {System.out.println(3970);}}
    public static class Impl3971 implements DemoService {public int type() {return 3971;} public void run() {System.out.println(3971);}}
    public static class Impl3972 implements DemoService {public int type() {return 3972;} public void run() {System.out.println(3972);}}
    public static class Impl3973 implements DemoService {public int type() {return 3973;} public void run() {System.out.println(3973);}}
    public static class Impl3974 implements DemoService {public int type() {return 3974;} public void run() {System.out.println(3974);}}
    public static class Impl3975 implements DemoService {public int type() {return 3975;} public void run() {System.out.println(3975);}}
    public static class Impl3976 implements DemoService {public int type() {return 3976;} public void run() {System.out.println(3976);}}
    public static class Impl3977 implements DemoService {public int type() {return 3977;} public void run() {System.out.println(3977);}}
    public static class Impl3978 implements DemoService {public int type() {return 3978;} public void run() {System.out.println(3978);}}
    public static class Impl3979 implements DemoService {public int type() {return 3979;} public void run() {System.out.println(3979);}}
    public static class Impl3980 implements DemoService {public int type() {return 3980;} public void run() {System.out.println(3980);}}
    public static class Impl3981 implements DemoService {public int type() {return 3981;} public void run() {System.out.println(3981);}}
    public static class Impl3982 implements DemoService {public int type() {return 3982;} public void run() {System.out.println(3982);}}
    public static class Impl3983 implements DemoService {public int type() {return 3983;} public void run() {System.out.println(3983);}}
    public static class Impl3984 implements DemoService {public int type() {return 3984;} public void run() {System.out.println(3984);}}
    public static class Impl3985 implements DemoService {public int type() {return 3985;} public void run() {System.out.println(3985);}}
    public static class Impl3986 implements DemoService {public int type() {return 3986;} public void run() {System.out.println(3986);}}
    public static class Impl3987 implements DemoService {public int type() {return 3987;} public void run() {System.out.println(3987);}}
    public static class Impl3988 implements DemoService {public int type() {return 3988;} public void run() {System.out.println(3988);}}
    public static class Impl3989 implements DemoService {public int type() {return 3989;} public void run() {System.out.println(3989);}}
    public static class Impl3990 implements DemoService {public int type() {return 3990;} public void run() {System.out.println(3990);}}
    public static class Impl3991 implements DemoService {public int type() {return 3991;} public void run() {System.out.println(3991);}}
    public static class Impl3992 implements DemoService {public int type() {return 3992;} public void run() {System.out.println(3992);}}
    public static class Impl3993 implements DemoService {public int type() {return 3993;} public void run() {System.out.println(3993);}}
    public static class Impl3994 implements DemoService {public int type() {return 3994;} public void run() {System.out.println(3994);}}
    public static class Impl3995 implements DemoService {public int type() {return 3995;} public void run() {System.out.println(3995);}}
    public static class Impl3996 implements DemoService {public int type() {return 3996;} public void run() {System.out.println(3996);}}
    public static class Impl3997 implements DemoService {public int type() {return 3997;} public void run() {System.out.println(3997);}}
    public static class Impl3998 implements DemoService {public int type() {return 3998;} public void run() {System.out.println(3998);}}
    public static class Impl3999 implements DemoService {public int type() {return 3999;} public void run() {System.out.println(3999);}}
    public static class Impl4000 implements DemoService {public int type() {return 4000;} public void run() {System.out.println(4000);}}
    public static class Impl4001 implements DemoService {public int type() {return 4001;} public void run() {System.out.println(4001);}}
    public static class Impl4002 implements DemoService {public int type() {return 4002;} public void run() {System.out.println(4002);}}
    public static class Impl4003 implements DemoService {public int type() {return 4003;} public void run() {System.out.println(4003);}}
    public static class Impl4004 implements DemoService {public int type() {return 4004;} public void run() {System.out.println(4004);}}
    public static class Impl4005 implements DemoService {public int type() {return 4005;} public void run() {System.out.println(4005);}}
    public static class Impl4006 implements DemoService {public int type() {return 4006;} public void run() {System.out.println(4006);}}
    public static class Impl4007 implements DemoService {public int type() {return 4007;} public void run() {System.out.println(4007);}}
    public static class Impl4008 implements DemoService {public int type() {return 4008;} public void run() {System.out.println(4008);}}
    public static class Impl4009 implements DemoService {public int type() {return 4009;} public void run() {System.out.println(4009);}}
    public static class Impl4010 implements DemoService {public int type() {return 4010;} public void run() {System.out.println(4010);}}
    public static class Impl4011 implements DemoService {public int type() {return 4011;} public void run() {System.out.println(4011);}}
    public static class Impl4012 implements DemoService {public int type() {return 4012;} public void run() {System.out.println(4012);}}
    public static class Impl4013 implements DemoService {public int type() {return 4013;} public void run() {System.out.println(4013);}}
    public static class Impl4014 implements DemoService {public int type() {return 4014;} public void run() {System.out.println(4014);}}
    public static class Impl4015 implements DemoService {public int type() {return 4015;} public void run() {System.out.println(4015);}}
    public static class Impl4016 implements DemoService {public int type() {return 4016;} public void run() {System.out.println(4016);}}
    public static class Impl4017 implements DemoService {public int type() {return 4017;} public void run() {System.out.println(4017);}}
    public static class Impl4018 implements DemoService {public int type() {return 4018;} public void run() {System.out.println(4018);}}
    public static class Impl4019 implements DemoService {public int type() {return 4019;} public void run() {System.out.println(4019);}}
    public static class Impl4020 implements DemoService {public int type() {return 4020;} public void run() {System.out.println(4020);}}
    public static class Impl4021 implements DemoService {public int type() {return 4021;} public void run() {System.out.println(4021);}}
    public static class Impl4022 implements DemoService {public int type() {return 4022;} public void run() {System.out.println(4022);}}
    public static class Impl4023 implements DemoService {public int type() {return 4023;} public void run() {System.out.println(4023);}}
    public static class Impl4024 implements DemoService {public int type() {return 4024;} public void run() {System.out.println(4024);}}
    public static class Impl4025 implements DemoService {public int type() {return 4025;} public void run() {System.out.println(4025);}}
    public static class Impl4026 implements DemoService {public int type() {return 4026;} public void run() {System.out.println(4026);}}
    public static class Impl4027 implements DemoService {public int type() {return 4027;} public void run() {System.out.println(4027);}}
    public static class Impl4028 implements DemoService {public int type() {return 4028;} public void run() {System.out.println(4028);}}
    public static class Impl4029 implements DemoService {public int type() {return 4029;} public void run() {System.out.println(4029);}}
    public static class Impl4030 implements DemoService {public int type() {return 4030;} public void run() {System.out.println(4030);}}
    public static class Impl4031 implements DemoService {public int type() {return 4031;} public void run() {System.out.println(4031);}}
    public static class Impl4032 implements DemoService {public int type() {return 4032;} public void run() {System.out.println(4032);}}
    public static class Impl4033 implements DemoService {public int type() {return 4033;} public void run() {System.out.println(4033);}}
    public static class Impl4034 implements DemoService {public int type() {return 4034;} public void run() {System.out.println(4034);}}
    public static class Impl4035 implements DemoService {public int type() {return 4035;} public void run() {System.out.println(4035);}}
    public static class Impl4036 implements DemoService {public int type() {return 4036;} public void run() {System.out.println(4036);}}
    public static class Impl4037 implements DemoService {public int type() {return 4037;} public void run() {System.out.println(4037);}}
    public static class Impl4038 implements DemoService {public int type() {return 4038;} public void run() {System.out.println(4038);}}
    public static class Impl4039 implements DemoService {public int type() {return 4039;} public void run() {System.out.println(4039);}}
    public static class Impl4040 implements DemoService {public int type() {return 4040;} public void run() {System.out.println(4040);}}
    public static class Impl4041 implements DemoService {public int type() {return 4041;} public void run() {System.out.println(4041);}}
    public static class Impl4042 implements DemoService {public int type() {return 4042;} public void run() {System.out.println(4042);}}
    public static class Impl4043 implements DemoService {public int type() {return 4043;} public void run() {System.out.println(4043);}}
    public static class Impl4044 implements DemoService {public int type() {return 4044;} public void run() {System.out.println(4044);}}
    public static class Impl4045 implements DemoService {public int type() {return 4045;} public void run() {System.out.println(4045);}}
    public static class Impl4046 implements DemoService {public int type() {return 4046;} public void run() {System.out.println(4046);}}
    public static class Impl4047 implements DemoService {public int type() {return 4047;} public void run() {System.out.println(4047);}}
    public static class Impl4048 implements DemoService {public int type() {return 4048;} public void run() {System.out.println(4048);}}
    public static class Impl4049 implements DemoService {public int type() {return 4049;} public void run() {System.out.println(4049);}}
    public static class Impl4050 implements DemoService {public int type() {return 4050;} public void run() {System.out.println(4050);}}
    public static class Impl4051 implements DemoService {public int type() {return 4051;} public void run() {System.out.println(4051);}}
    public static class Impl4052 implements DemoService {public int type() {return 4052;} public void run() {System.out.println(4052);}}
    public static class Impl4053 implements DemoService {public int type() {return 4053;} public void run() {System.out.println(4053);}}
    public static class Impl4054 implements DemoService {public int type() {return 4054;} public void run() {System.out.println(4054);}}
    public static class Impl4055 implements DemoService {public int type() {return 4055;} public void run() {System.out.println(4055);}}
    public static class Impl4056 implements DemoService {public int type() {return 4056;} public void run() {System.out.println(4056);}}
    public static class Impl4057 implements DemoService {public int type() {return 4057;} public void run() {System.out.println(4057);}}
    public static class Impl4058 implements DemoService {public int type() {return 4058;} public void run() {System.out.println(4058);}}
    public static class Impl4059 implements DemoService {public int type() {return 4059;} public void run() {System.out.println(4059);}}
    public static class Impl4060 implements DemoService {public int type() {return 4060;} public void run() {System.out.println(4060);}}
    public static class Impl4061 implements DemoService {public int type() {return 4061;} public void run() {System.out.println(4061);}}
    public static class Impl4062 implements DemoService {public int type() {return 4062;} public void run() {System.out.println(4062);}}
    public static class Impl4063 implements DemoService {public int type() {return 4063;} public void run() {System.out.println(4063);}}
    public static class Impl4064 implements DemoService {public int type() {return 4064;} public void run() {System.out.println(4064);}}
    public static class Impl4065 implements DemoService {public int type() {return 4065;} public void run() {System.out.println(4065);}}
    public static class Impl4066 implements DemoService {public int type() {return 4066;} public void run() {System.out.println(4066);}}
    public static class Impl4067 implements DemoService {public int type() {return 4067;} public void run() {System.out.println(4067);}}
    public static class Impl4068 implements DemoService {public int type() {return 4068;} public void run() {System.out.println(4068);}}
    public static class Impl4069 implements DemoService {public int type() {return 4069;} public void run() {System.out.println(4069);}}
    public static class Impl4070 implements DemoService {public int type() {return 4070;} public void run() {System.out.println(4070);}}
    public static class Impl4071 implements DemoService {public int type() {return 4071;} public void run() {System.out.println(4071);}}
    public static class Impl4072 implements DemoService {public int type() {return 4072;} public void run() {System.out.println(4072);}}
    public static class Impl4073 implements DemoService {public int type() {return 4073;} public void run() {System.out.println(4073);}}
    public static class Impl4074 implements DemoService {public int type() {return 4074;} public void run() {System.out.println(4074);}}
    public static class Impl4075 implements DemoService {public int type() {return 4075;} public void run() {System.out.println(4075);}}
    public static class Impl4076 implements DemoService {public int type() {return 4076;} public void run() {System.out.println(4076);}}
    public static class Impl4077 implements DemoService {public int type() {return 4077;} public void run() {System.out.println(4077);}}
    public static class Impl4078 implements DemoService {public int type() {return 4078;} public void run() {System.out.println(4078);}}
    public static class Impl4079 implements DemoService {public int type() {return 4079;} public void run() {System.out.println(4079);}}
    public static class Impl4080 implements DemoService {public int type() {return 4080;} public void run() {System.out.println(4080);}}
    public static class Impl4081 implements DemoService {public int type() {return 4081;} public void run() {System.out.println(4081);}}
    public static class Impl4082 implements DemoService {public int type() {return 4082;} public void run() {System.out.println(4082);}}
    public static class Impl4083 implements DemoService {public int type() {return 4083;} public void run() {System.out.println(4083);}}
    public static class Impl4084 implements DemoService {public int type() {return 4084;} public void run() {System.out.println(4084);}}
    public static class Impl4085 implements DemoService {public int type() {return 4085;} public void run() {System.out.println(4085);}}
    public static class Impl4086 implements DemoService {public int type() {return 4086;} public void run() {System.out.println(4086);}}
    public static class Impl4087 implements DemoService {public int type() {return 4087;} public void run() {System.out.println(4087);}}
    public static class Impl4088 implements DemoService {public int type() {return 4088;} public void run() {System.out.println(4088);}}
    public static class Impl4089 implements DemoService {public int type() {return 4089;} public void run() {System.out.println(4089);}}
    public static class Impl4090 implements DemoService {public int type() {return 4090;} public void run() {System.out.println(4090);}}
    public static class Impl4091 implements DemoService {public int type() {return 4091;} public void run() {System.out.println(4091);}}
    public static class Impl4092 implements DemoService {public int type() {return 4092;} public void run() {System.out.println(4092);}}
    public static class Impl4093 implements DemoService {public int type() {return 4093;} public void run() {System.out.println(4093);}}
    public static class Impl4094 implements DemoService {public int type() {return 4094;} public void run() {System.out.println(4094);}}
    public static class Impl4095 implements DemoService {public int type() {return 4095;} public void run() {System.out.println(4095);}}
    public static class Impl4096 implements DemoService {public int type() {return 4096;} public void run() {System.out.println(4096);}}
    public static class Impl4097 implements DemoService {public int type() {return 4097;} public void run() {System.out.println(4097);}}
    public static class Impl4098 implements DemoService {public int type() {return 4098;} public void run() {System.out.println(4098);}}
    public static class Impl4099 implements DemoService {public int type() {return 4099;} public void run() {System.out.println(4099);}}
    public static class Impl4100 implements DemoService {public int type() {return 4100;} public void run() {System.out.println(4100);}}
    public static class Impl4101 implements DemoService {public int type() {return 4101;} public void run() {System.out.println(4101);}}
    public static class Impl4102 implements DemoService {public int type() {return 4102;} public void run() {System.out.println(4102);}}
    public static class Impl4103 implements DemoService {public int type() {return 4103;} public void run() {System.out.println(4103);}}
    public static class Impl4104 implements DemoService {public int type() {return 4104;} public void run() {System.out.println(4104);}}
    public static class Impl4105 implements DemoService {public int type() {return 4105;} public void run() {System.out.println(4105);}}
    public static class Impl4106 implements DemoService {public int type() {return 4106;} public void run() {System.out.println(4106);}}
    public static class Impl4107 implements DemoService {public int type() {return 4107;} public void run() {System.out.println(4107);}}
    public static class Impl4108 implements DemoService {public int type() {return 4108;} public void run() {System.out.println(4108);}}
    public static class Impl4109 implements DemoService {public int type() {return 4109;} public void run() {System.out.println(4109);}}
    public static class Impl4110 implements DemoService {public int type() {return 4110;} public void run() {System.out.println(4110);}}
    public static class Impl4111 implements DemoService {public int type() {return 4111;} public void run() {System.out.println(4111);}}
    public static class Impl4112 implements DemoService {public int type() {return 4112;} public void run() {System.out.println(4112);}}
    public static class Impl4113 implements DemoService {public int type() {return 4113;} public void run() {System.out.println(4113);}}
    public static class Impl4114 implements DemoService {public int type() {return 4114;} public void run() {System.out.println(4114);}}
    public static class Impl4115 implements DemoService {public int type() {return 4115;} public void run() {System.out.println(4115);}}
    public static class Impl4116 implements DemoService {public int type() {return 4116;} public void run() {System.out.println(4116);}}
    public static class Impl4117 implements DemoService {public int type() {return 4117;} public void run() {System.out.println(4117);}}
    public static class Impl4118 implements DemoService {public int type() {return 4118;} public void run() {System.out.println(4118);}}
    public static class Impl4119 implements DemoService {public int type() {return 4119;} public void run() {System.out.println(4119);}}
    public static class Impl4120 implements DemoService {public int type() {return 4120;} public void run() {System.out.println(4120);}}
    public static class Impl4121 implements DemoService {public int type() {return 4121;} public void run() {System.out.println(4121);}}
    public static class Impl4122 implements DemoService {public int type() {return 4122;} public void run() {System.out.println(4122);}}
    public static class Impl4123 implements DemoService {public int type() {return 4123;} public void run() {System.out.println(4123);}}
    public static class Impl4124 implements DemoService {public int type() {return 4124;} public void run() {System.out.println(4124);}}
    public static class Impl4125 implements DemoService {public int type() {return 4125;} public void run() {System.out.println(4125);}}
    public static class Impl4126 implements DemoService {public int type() {return 4126;} public void run() {System.out.println(4126);}}
    public static class Impl4127 implements DemoService {public int type() {return 4127;} public void run() {System.out.println(4127);}}
    public static class Impl4128 implements DemoService {public int type() {return 4128;} public void run() {System.out.println(4128);}}
    public static class Impl4129 implements DemoService {public int type() {return 4129;} public void run() {System.out.println(4129);}}
    public static class Impl4130 implements DemoService {public int type() {return 4130;} public void run() {System.out.println(4130);}}
    public static class Impl4131 implements DemoService {public int type() {return 4131;} public void run() {System.out.println(4131);}}
    public static class Impl4132 implements DemoService {public int type() {return 4132;} public void run() {System.out.println(4132);}}
    public static class Impl4133 implements DemoService {public int type() {return 4133;} public void run() {System.out.println(4133);}}
    public static class Impl4134 implements DemoService {public int type() {return 4134;} public void run() {System.out.println(4134);}}
    public static class Impl4135 implements DemoService {public int type() {return 4135;} public void run() {System.out.println(4135);}}
    public static class Impl4136 implements DemoService {public int type() {return 4136;} public void run() {System.out.println(4136);}}
    public static class Impl4137 implements DemoService {public int type() {return 4137;} public void run() {System.out.println(4137);}}
    public static class Impl4138 implements DemoService {public int type() {return 4138;} public void run() {System.out.println(4138);}}
    public static class Impl4139 implements DemoService {public int type() {return 4139;} public void run() {System.out.println(4139);}}
    public static class Impl4140 implements DemoService {public int type() {return 4140;} public void run() {System.out.println(4140);}}
    public static class Impl4141 implements DemoService {public int type() {return 4141;} public void run() {System.out.println(4141);}}
    public static class Impl4142 implements DemoService {public int type() {return 4142;} public void run() {System.out.println(4142);}}
    public static class Impl4143 implements DemoService {public int type() {return 4143;} public void run() {System.out.println(4143);}}
    public static class Impl4144 implements DemoService {public int type() {return 4144;} public void run() {System.out.println(4144);}}
    public static class Impl4145 implements DemoService {public int type() {return 4145;} public void run() {System.out.println(4145);}}
    public static class Impl4146 implements DemoService {public int type() {return 4146;} public void run() {System.out.println(4146);}}
    public static class Impl4147 implements DemoService {public int type() {return 4147;} public void run() {System.out.println(4147);}}
    public static class Impl4148 implements DemoService {public int type() {return 4148;} public void run() {System.out.println(4148);}}
    public static class Impl4149 implements DemoService {public int type() {return 4149;} public void run() {System.out.println(4149);}}
    public static class Impl4150 implements DemoService {public int type() {return 4150;} public void run() {System.out.println(4150);}}
    public static class Impl4151 implements DemoService {public int type() {return 4151;} public void run() {System.out.println(4151);}}
    public static class Impl4152 implements DemoService {public int type() {return 4152;} public void run() {System.out.println(4152);}}
    public static class Impl4153 implements DemoService {public int type() {return 4153;} public void run() {System.out.println(4153);}}
    public static class Impl4154 implements DemoService {public int type() {return 4154;} public void run() {System.out.println(4154);}}
    public static class Impl4155 implements DemoService {public int type() {return 4155;} public void run() {System.out.println(4155);}}
    public static class Impl4156 implements DemoService {public int type() {return 4156;} public void run() {System.out.println(4156);}}
    public static class Impl4157 implements DemoService {public int type() {return 4157;} public void run() {System.out.println(4157);}}
    public static class Impl4158 implements DemoService {public int type() {return 4158;} public void run() {System.out.println(4158);}}
    public static class Impl4159 implements DemoService {public int type() {return 4159;} public void run() {System.out.println(4159);}}
    public static class Impl4160 implements DemoService {public int type() {return 4160;} public void run() {System.out.println(4160);}}
    public static class Impl4161 implements DemoService {public int type() {return 4161;} public void run() {System.out.println(4161);}}
    public static class Impl4162 implements DemoService {public int type() {return 4162;} public void run() {System.out.println(4162);}}
    public static class Impl4163 implements DemoService {public int type() {return 4163;} public void run() {System.out.println(4163);}}
    public static class Impl4164 implements DemoService {public int type() {return 4164;} public void run() {System.out.println(4164);}}
    public static class Impl4165 implements DemoService {public int type() {return 4165;} public void run() {System.out.println(4165);}}
    public static class Impl4166 implements DemoService {public int type() {return 4166;} public void run() {System.out.println(4166);}}
    public static class Impl4167 implements DemoService {public int type() {return 4167;} public void run() {System.out.println(4167);}}
    public static class Impl4168 implements DemoService {public int type() {return 4168;} public void run() {System.out.println(4168);}}
    public static class Impl4169 implements DemoService {public int type() {return 4169;} public void run() {System.out.println(4169);}}
    public static class Impl4170 implements DemoService {public int type() {return 4170;} public void run() {System.out.println(4170);}}
    public static class Impl4171 implements DemoService {public int type() {return 4171;} public void run() {System.out.println(4171);}}
    public static class Impl4172 implements DemoService {public int type() {return 4172;} public void run() {System.out.println(4172);}}
    public static class Impl4173 implements DemoService {public int type() {return 4173;} public void run() {System.out.println(4173);}}
    public static class Impl4174 implements DemoService {public int type() {return 4174;} public void run() {System.out.println(4174);}}
    public static class Impl4175 implements DemoService {public int type() {return 4175;} public void run() {System.out.println(4175);}}
    public static class Impl4176 implements DemoService {public int type() {return 4176;} public void run() {System.out.println(4176);}}
    public static class Impl4177 implements DemoService {public int type() {return 4177;} public void run() {System.out.println(4177);}}
    public static class Impl4178 implements DemoService {public int type() {return 4178;} public void run() {System.out.println(4178);}}
    public static class Impl4179 implements DemoService {public int type() {return 4179;} public void run() {System.out.println(4179);}}
    public static class Impl4180 implements DemoService {public int type() {return 4180;} public void run() {System.out.println(4180);}}
    public static class Impl4181 implements DemoService {public int type() {return 4181;} public void run() {System.out.println(4181);}}
    public static class Impl4182 implements DemoService {public int type() {return 4182;} public void run() {System.out.println(4182);}}
    public static class Impl4183 implements DemoService {public int type() {return 4183;} public void run() {System.out.println(4183);}}
    public static class Impl4184 implements DemoService {public int type() {return 4184;} public void run() {System.out.println(4184);}}
    public static class Impl4185 implements DemoService {public int type() {return 4185;} public void run() {System.out.println(4185);}}
    public static class Impl4186 implements DemoService {public int type() {return 4186;} public void run() {System.out.println(4186);}}
    public static class Impl4187 implements DemoService {public int type() {return 4187;} public void run() {System.out.println(4187);}}
    public static class Impl4188 implements DemoService {public int type() {return 4188;} public void run() {System.out.println(4188);}}
    public static class Impl4189 implements DemoService {public int type() {return 4189;} public void run() {System.out.println(4189);}}
    public static class Impl4190 implements DemoService {public int type() {return 4190;} public void run() {System.out.println(4190);}}
    public static class Impl4191 implements DemoService {public int type() {return 4191;} public void run() {System.out.println(4191);}}
    public static class Impl4192 implements DemoService {public int type() {return 4192;} public void run() {System.out.println(4192);}}
    public static class Impl4193 implements DemoService {public int type() {return 4193;} public void run() {System.out.println(4193);}}
    public static class Impl4194 implements DemoService {public int type() {return 4194;} public void run() {System.out.println(4194);}}
    public static class Impl4195 implements DemoService {public int type() {return 4195;} public void run() {System.out.println(4195);}}
    public static class Impl4196 implements DemoService {public int type() {return 4196;} public void run() {System.out.println(4196);}}
    public static class Impl4197 implements DemoService {public int type() {return 4197;} public void run() {System.out.println(4197);}}
    public static class Impl4198 implements DemoService {public int type() {return 4198;} public void run() {System.out.println(4198);}}
    public static class Impl4199 implements DemoService {public int type() {return 4199;} public void run() {System.out.println(4199);}}
    public static class Impl4200 implements DemoService {public int type() {return 4200;} public void run() {System.out.println(4200);}}
    public static class Impl4201 implements DemoService {public int type() {return 4201;} public void run() {System.out.println(4201);}}
    public static class Impl4202 implements DemoService {public int type() {return 4202;} public void run() {System.out.println(4202);}}
    public static class Impl4203 implements DemoService {public int type() {return 4203;} public void run() {System.out.println(4203);}}
    public static class Impl4204 implements DemoService {public int type() {return 4204;} public void run() {System.out.println(4204);}}
    public static class Impl4205 implements DemoService {public int type() {return 4205;} public void run() {System.out.println(4205);}}
    public static class Impl4206 implements DemoService {public int type() {return 4206;} public void run() {System.out.println(4206);}}
    public static class Impl4207 implements DemoService {public int type() {return 4207;} public void run() {System.out.println(4207);}}
    public static class Impl4208 implements DemoService {public int type() {return 4208;} public void run() {System.out.println(4208);}}
    public static class Impl4209 implements DemoService {public int type() {return 4209;} public void run() {System.out.println(4209);}}
    public static class Impl4210 implements DemoService {public int type() {return 4210;} public void run() {System.out.println(4210);}}
    public static class Impl4211 implements DemoService {public int type() {return 4211;} public void run() {System.out.println(4211);}}
    public static class Impl4212 implements DemoService {public int type() {return 4212;} public void run() {System.out.println(4212);}}
    public static class Impl4213 implements DemoService {public int type() {return 4213;} public void run() {System.out.println(4213);}}
    public static class Impl4214 implements DemoService {public int type() {return 4214;} public void run() {System.out.println(4214);}}
    public static class Impl4215 implements DemoService {public int type() {return 4215;} public void run() {System.out.println(4215);}}
    public static class Impl4216 implements DemoService {public int type() {return 4216;} public void run() {System.out.println(4216);}}
    public static class Impl4217 implements DemoService {public int type() {return 4217;} public void run() {System.out.println(4217);}}
    public static class Impl4218 implements DemoService {public int type() {return 4218;} public void run() {System.out.println(4218);}}
    public static class Impl4219 implements DemoService {public int type() {return 4219;} public void run() {System.out.println(4219);}}
    public static class Impl4220 implements DemoService {public int type() {return 4220;} public void run() {System.out.println(4220);}}
    public static class Impl4221 implements DemoService {public int type() {return 4221;} public void run() {System.out.println(4221);}}
    public static class Impl4222 implements DemoService {public int type() {return 4222;} public void run() {System.out.println(4222);}}
    public static class Impl4223 implements DemoService {public int type() {return 4223;} public void run() {System.out.println(4223);}}
    public static class Impl4224 implements DemoService {public int type() {return 4224;} public void run() {System.out.println(4224);}}
    public static class Impl4225 implements DemoService {public int type() {return 4225;} public void run() {System.out.println(4225);}}
    public static class Impl4226 implements DemoService {public int type() {return 4226;} public void run() {System.out.println(4226);}}
    public static class Impl4227 implements DemoService {public int type() {return 4227;} public void run() {System.out.println(4227);}}
    public static class Impl4228 implements DemoService {public int type() {return 4228;} public void run() {System.out.println(4228);}}
    public static class Impl4229 implements DemoService {public int type() {return 4229;} public void run() {System.out.println(4229);}}
    public static class Impl4230 implements DemoService {public int type() {return 4230;} public void run() {System.out.println(4230);}}
    public static class Impl4231 implements DemoService {public int type() {return 4231;} public void run() {System.out.println(4231);}}
    public static class Impl4232 implements DemoService {public int type() {return 4232;} public void run() {System.out.println(4232);}}
    public static class Impl4233 implements DemoService {public int type() {return 4233;} public void run() {System.out.println(4233);}}
    public static class Impl4234 implements DemoService {public int type() {return 4234;} public void run() {System.out.println(4234);}}
    public static class Impl4235 implements DemoService {public int type() {return 4235;} public void run() {System.out.println(4235);}}
    public static class Impl4236 implements DemoService {public int type() {return 4236;} public void run() {System.out.println(4236);}}
    public static class Impl4237 implements DemoService {public int type() {return 4237;} public void run() {System.out.println(4237);}}
    public static class Impl4238 implements DemoService {public int type() {return 4238;} public void run() {System.out.println(4238);}}
    public static class Impl4239 implements DemoService {public int type() {return 4239;} public void run() {System.out.println(4239);}}
    public static class Impl4240 implements DemoService {public int type() {return 4240;} public void run() {System.out.println(4240);}}
    public static class Impl4241 implements DemoService {public int type() {return 4241;} public void run() {System.out.println(4241);}}
    public static class Impl4242 implements DemoService {public int type() {return 4242;} public void run() {System.out.println(4242);}}
    public static class Impl4243 implements DemoService {public int type() {return 4243;} public void run() {System.out.println(4243);}}
    public static class Impl4244 implements DemoService {public int type() {return 4244;} public void run() {System.out.println(4244);}}
    public static class Impl4245 implements DemoService {public int type() {return 4245;} public void run() {System.out.println(4245);}}
    public static class Impl4246 implements DemoService {public int type() {return 4246;} public void run() {System.out.println(4246);}}
    public static class Impl4247 implements DemoService {public int type() {return 4247;} public void run() {System.out.println(4247);}}
    public static class Impl4248 implements DemoService {public int type() {return 4248;} public void run() {System.out.println(4248);}}
    public static class Impl4249 implements DemoService {public int type() {return 4249;} public void run() {System.out.println(4249);}}
    public static class Impl4250 implements DemoService {public int type() {return 4250;} public void run() {System.out.println(4250);}}
    public static class Impl4251 implements DemoService {public int type() {return 4251;} public void run() {System.out.println(4251);}}
    public static class Impl4252 implements DemoService {public int type() {return 4252;} public void run() {System.out.println(4252);}}
    public static class Impl4253 implements DemoService {public int type() {return 4253;} public void run() {System.out.println(4253);}}
    public static class Impl4254 implements DemoService {public int type() {return 4254;} public void run() {System.out.println(4254);}}
    public static class Impl4255 implements DemoService {public int type() {return 4255;} public void run() {System.out.println(4255);}}
    public static class Impl4256 implements DemoService {public int type() {return 4256;} public void run() {System.out.println(4256);}}
    public static class Impl4257 implements DemoService {public int type() {return 4257;} public void run() {System.out.println(4257);}}
    public static class Impl4258 implements DemoService {public int type() {return 4258;} public void run() {System.out.println(4258);}}
    public static class Impl4259 implements DemoService {public int type() {return 4259;} public void run() {System.out.println(4259);}}
    public static class Impl4260 implements DemoService {public int type() {return 4260;} public void run() {System.out.println(4260);}}
    public static class Impl4261 implements DemoService {public int type() {return 4261;} public void run() {System.out.println(4261);}}
    public static class Impl4262 implements DemoService {public int type() {return 4262;} public void run() {System.out.println(4262);}}
    public static class Impl4263 implements DemoService {public int type() {return 4263;} public void run() {System.out.println(4263);}}
    public static class Impl4264 implements DemoService {public int type() {return 4264;} public void run() {System.out.println(4264);}}
    public static class Impl4265 implements DemoService {public int type() {return 4265;} public void run() {System.out.println(4265);}}
    public static class Impl4266 implements DemoService {public int type() {return 4266;} public void run() {System.out.println(4266);}}
    public static class Impl4267 implements DemoService {public int type() {return 4267;} public void run() {System.out.println(4267);}}
    public static class Impl4268 implements DemoService {public int type() {return 4268;} public void run() {System.out.println(4268);}}
    public static class Impl4269 implements DemoService {public int type() {return 4269;} public void run() {System.out.println(4269);}}
    public static class Impl4270 implements DemoService {public int type() {return 4270;} public void run() {System.out.println(4270);}}
    public static class Impl4271 implements DemoService {public int type() {return 4271;} public void run() {System.out.println(4271);}}
    public static class Impl4272 implements DemoService {public int type() {return 4272;} public void run() {System.out.println(4272);}}
    public static class Impl4273 implements DemoService {public int type() {return 4273;} public void run() {System.out.println(4273);}}
    public static class Impl4274 implements DemoService {public int type() {return 4274;} public void run() {System.out.println(4274);}}
    public static class Impl4275 implements DemoService {public int type() {return 4275;} public void run() {System.out.println(4275);}}
    public static class Impl4276 implements DemoService {public int type() {return 4276;} public void run() {System.out.println(4276);}}
    public static class Impl4277 implements DemoService {public int type() {return 4277;} public void run() {System.out.println(4277);}}
    public static class Impl4278 implements DemoService {public int type() {return 4278;} public void run() {System.out.println(4278);}}
    public static class Impl4279 implements DemoService {public int type() {return 4279;} public void run() {System.out.println(4279);}}
    public static class Impl4280 implements DemoService {public int type() {return 4280;} public void run() {System.out.println(4280);}}
    public static class Impl4281 implements DemoService {public int type() {return 4281;} public void run() {System.out.println(4281);}}
    public static class Impl4282 implements DemoService {public int type() {return 4282;} public void run() {System.out.println(4282);}}
    public static class Impl4283 implements DemoService {public int type() {return 4283;} public void run() {System.out.println(4283);}}
    public static class Impl4284 implements DemoService {public int type() {return 4284;} public void run() {System.out.println(4284);}}
    public static class Impl4285 implements DemoService {public int type() {return 4285;} public void run() {System.out.println(4285);}}
    public static class Impl4286 implements DemoService {public int type() {return 4286;} public void run() {System.out.println(4286);}}
    public static class Impl4287 implements DemoService {public int type() {return 4287;} public void run() {System.out.println(4287);}}
    public static class Impl4288 implements DemoService {public int type() {return 4288;} public void run() {System.out.println(4288);}}
    public static class Impl4289 implements DemoService {public int type() {return 4289;} public void run() {System.out.println(4289);}}
    public static class Impl4290 implements DemoService {public int type() {return 4290;} public void run() {System.out.println(4290);}}
    public static class Impl4291 implements DemoService {public int type() {return 4291;} public void run() {System.out.println(4291);}}
    public static class Impl4292 implements DemoService {public int type() {return 4292;} public void run() {System.out.println(4292);}}
    public static class Impl4293 implements DemoService {public int type() {return 4293;} public void run() {System.out.println(4293);}}
    public static class Impl4294 implements DemoService {public int type() {return 4294;} public void run() {System.out.println(4294);}}
    public static class Impl4295 implements DemoService {public int type() {return 4295;} public void run() {System.out.println(4295);}}
    public static class Impl4296 implements DemoService {public int type() {return 4296;} public void run() {System.out.println(4296);}}
    public static class Impl4297 implements DemoService {public int type() {return 4297;} public void run() {System.out.println(4297);}}
    public static class Impl4298 implements DemoService {public int type() {return 4298;} public void run() {System.out.println(4298);}}
    public static class Impl4299 implements DemoService {public int type() {return 4299;} public void run() {System.out.println(4299);}}
    public static class Impl4300 implements DemoService {public int type() {return 4300;} public void run() {System.out.println(4300);}}
    public static class Impl4301 implements DemoService {public int type() {return 4301;} public void run() {System.out.println(4301);}}
    public static class Impl4302 implements DemoService {public int type() {return 4302;} public void run() {System.out.println(4302);}}
    public static class Impl4303 implements DemoService {public int type() {return 4303;} public void run() {System.out.println(4303);}}
    public static class Impl4304 implements DemoService {public int type() {return 4304;} public void run() {System.out.println(4304);}}
    public static class Impl4305 implements DemoService {public int type() {return 4305;} public void run() {System.out.println(4305);}}
    public static class Impl4306 implements DemoService {public int type() {return 4306;} public void run() {System.out.println(4306);}}
    public static class Impl4307 implements DemoService {public int type() {return 4307;} public void run() {System.out.println(4307);}}
    public static class Impl4308 implements DemoService {public int type() {return 4308;} public void run() {System.out.println(4308);}}
    public static class Impl4309 implements DemoService {public int type() {return 4309;} public void run() {System.out.println(4309);}}
    public static class Impl4310 implements DemoService {public int type() {return 4310;} public void run() {System.out.println(4310);}}
    public static class Impl4311 implements DemoService {public int type() {return 4311;} public void run() {System.out.println(4311);}}
    public static class Impl4312 implements DemoService {public int type() {return 4312;} public void run() {System.out.println(4312);}}
    public static class Impl4313 implements DemoService {public int type() {return 4313;} public void run() {System.out.println(4313);}}
    public static class Impl4314 implements DemoService {public int type() {return 4314;} public void run() {System.out.println(4314);}}
    public static class Impl4315 implements DemoService {public int type() {return 4315;} public void run() {System.out.println(4315);}}
    public static class Impl4316 implements DemoService {public int type() {return 4316;} public void run() {System.out.println(4316);}}
    public static class Impl4317 implements DemoService {public int type() {return 4317;} public void run() {System.out.println(4317);}}
    public static class Impl4318 implements DemoService {public int type() {return 4318;} public void run() {System.out.println(4318);}}
    public static class Impl4319 implements DemoService {public int type() {return 4319;} public void run() {System.out.println(4319);}}
    public static class Impl4320 implements DemoService {public int type() {return 4320;} public void run() {System.out.println(4320);}}
    public static class Impl4321 implements DemoService {public int type() {return 4321;} public void run() {System.out.println(4321);}}
    public static class Impl4322 implements DemoService {public int type() {return 4322;} public void run() {System.out.println(4322);}}
    public static class Impl4323 implements DemoService {public int type() {return 4323;} public void run() {System.out.println(4323);}}
    public static class Impl4324 implements DemoService {public int type() {return 4324;} public void run() {System.out.println(4324);}}
    public static class Impl4325 implements DemoService {public int type() {return 4325;} public void run() {System.out.println(4325);}}
    public static class Impl4326 implements DemoService {public int type() {return 4326;} public void run() {System.out.println(4326);}}
    public static class Impl4327 implements DemoService {public int type() {return 4327;} public void run() {System.out.println(4327);}}
    public static class Impl4328 implements DemoService {public int type() {return 4328;} public void run() {System.out.println(4328);}}
    public static class Impl4329 implements DemoService {public int type() {return 4329;} public void run() {System.out.println(4329);}}
    public static class Impl4330 implements DemoService {public int type() {return 4330;} public void run() {System.out.println(4330);}}
    public static class Impl4331 implements DemoService {public int type() {return 4331;} public void run() {System.out.println(4331);}}
    public static class Impl4332 implements DemoService {public int type() {return 4332;} public void run() {System.out.println(4332);}}
    public static class Impl4333 implements DemoService {public int type() {return 4333;} public void run() {System.out.println(4333);}}
    public static class Impl4334 implements DemoService {public int type() {return 4334;} public void run() {System.out.println(4334);}}
    public static class Impl4335 implements DemoService {public int type() {return 4335;} public void run() {System.out.println(4335);}}
    public static class Impl4336 implements DemoService {public int type() {return 4336;} public void run() {System.out.println(4336);}}
    public static class Impl4337 implements DemoService {public int type() {return 4337;} public void run() {System.out.println(4337);}}
    public static class Impl4338 implements DemoService {public int type() {return 4338;} public void run() {System.out.println(4338);}}
    public static class Impl4339 implements DemoService {public int type() {return 4339;} public void run() {System.out.println(4339);}}
    public static class Impl4340 implements DemoService {public int type() {return 4340;} public void run() {System.out.println(4340);}}
    public static class Impl4341 implements DemoService {public int type() {return 4341;} public void run() {System.out.println(4341);}}
    public static class Impl4342 implements DemoService {public int type() {return 4342;} public void run() {System.out.println(4342);}}
    public static class Impl4343 implements DemoService {public int type() {return 4343;} public void run() {System.out.println(4343);}}
    public static class Impl4344 implements DemoService {public int type() {return 4344;} public void run() {System.out.println(4344);}}
    public static class Impl4345 implements DemoService {public int type() {return 4345;} public void run() {System.out.println(4345);}}
    public static class Impl4346 implements DemoService {public int type() {return 4346;} public void run() {System.out.println(4346);}}
    public static class Impl4347 implements DemoService {public int type() {return 4347;} public void run() {System.out.println(4347);}}
    public static class Impl4348 implements DemoService {public int type() {return 4348;} public void run() {System.out.println(4348);}}
    public static class Impl4349 implements DemoService {public int type() {return 4349;} public void run() {System.out.println(4349);}}
    public static class Impl4350 implements DemoService {public int type() {return 4350;} public void run() {System.out.println(4350);}}
    public static class Impl4351 implements DemoService {public int type() {return 4351;} public void run() {System.out.println(4351);}}
    public static class Impl4352 implements DemoService {public int type() {return 4352;} public void run() {System.out.println(4352);}}
    public static class Impl4353 implements DemoService {public int type() {return 4353;} public void run() {System.out.println(4353);}}
    public static class Impl4354 implements DemoService {public int type() {return 4354;} public void run() {System.out.println(4354);}}
    public static class Impl4355 implements DemoService {public int type() {return 4355;} public void run() {System.out.println(4355);}}
    public static class Impl4356 implements DemoService {public int type() {return 4356;} public void run() {System.out.println(4356);}}
    public static class Impl4357 implements DemoService {public int type() {return 4357;} public void run() {System.out.println(4357);}}
    public static class Impl4358 implements DemoService {public int type() {return 4358;} public void run() {System.out.println(4358);}}
    public static class Impl4359 implements DemoService {public int type() {return 4359;} public void run() {System.out.println(4359);}}
    public static class Impl4360 implements DemoService {public int type() {return 4360;} public void run() {System.out.println(4360);}}
    public static class Impl4361 implements DemoService {public int type() {return 4361;} public void run() {System.out.println(4361);}}
    public static class Impl4362 implements DemoService {public int type() {return 4362;} public void run() {System.out.println(4362);}}
    public static class Impl4363 implements DemoService {public int type() {return 4363;} public void run() {System.out.println(4363);}}
    public static class Impl4364 implements DemoService {public int type() {return 4364;} public void run() {System.out.println(4364);}}
    public static class Impl4365 implements DemoService {public int type() {return 4365;} public void run() {System.out.println(4365);}}
    public static class Impl4366 implements DemoService {public int type() {return 4366;} public void run() {System.out.println(4366);}}
    public static class Impl4367 implements DemoService {public int type() {return 4367;} public void run() {System.out.println(4367);}}
    public static class Impl4368 implements DemoService {public int type() {return 4368;} public void run() {System.out.println(4368);}}
    public static class Impl4369 implements DemoService {public int type() {return 4369;} public void run() {System.out.println(4369);}}
    public static class Impl4370 implements DemoService {public int type() {return 4370;} public void run() {System.out.println(4370);}}
    public static class Impl4371 implements DemoService {public int type() {return 4371;} public void run() {System.out.println(4371);}}
    public static class Impl4372 implements DemoService {public int type() {return 4372;} public void run() {System.out.println(4372);}}
    public static class Impl4373 implements DemoService {public int type() {return 4373;} public void run() {System.out.println(4373);}}
    public static class Impl4374 implements DemoService {public int type() {return 4374;} public void run() {System.out.println(4374);}}
    public static class Impl4375 implements DemoService {public int type() {return 4375;} public void run() {System.out.println(4375);}}
    public static class Impl4376 implements DemoService {public int type() {return 4376;} public void run() {System.out.println(4376);}}
    public static class Impl4377 implements DemoService {public int type() {return 4377;} public void run() {System.out.println(4377);}}
    public static class Impl4378 implements DemoService {public int type() {return 4378;} public void run() {System.out.println(4378);}}
    public static class Impl4379 implements DemoService {public int type() {return 4379;} public void run() {System.out.println(4379);}}
    public static class Impl4380 implements DemoService {public int type() {return 4380;} public void run() {System.out.println(4380);}}
    public static class Impl4381 implements DemoService {public int type() {return 4381;} public void run() {System.out.println(4381);}}
    public static class Impl4382 implements DemoService {public int type() {return 4382;} public void run() {System.out.println(4382);}}
    public static class Impl4383 implements DemoService {public int type() {return 4383;} public void run() {System.out.println(4383);}}
    public static class Impl4384 implements DemoService {public int type() {return 4384;} public void run() {System.out.println(4384);}}
    public static class Impl4385 implements DemoService {public int type() {return 4385;} public void run() {System.out.println(4385);}}
    public static class Impl4386 implements DemoService {public int type() {return 4386;} public void run() {System.out.println(4386);}}
    public static class Impl4387 implements DemoService {public int type() {return 4387;} public void run() {System.out.println(4387);}}
    public static class Impl4388 implements DemoService {public int type() {return 4388;} public void run() {System.out.println(4388);}}
    public static class Impl4389 implements DemoService {public int type() {return 4389;} public void run() {System.out.println(4389);}}
    public static class Impl4390 implements DemoService {public int type() {return 4390;} public void run() {System.out.println(4390);}}
    public static class Impl4391 implements DemoService {public int type() {return 4391;} public void run() {System.out.println(4391);}}
    public static class Impl4392 implements DemoService {public int type() {return 4392;} public void run() {System.out.println(4392);}}
    public static class Impl4393 implements DemoService {public int type() {return 4393;} public void run() {System.out.println(4393);}}
    public static class Impl4394 implements DemoService {public int type() {return 4394;} public void run() {System.out.println(4394);}}
    public static class Impl4395 implements DemoService {public int type() {return 4395;} public void run() {System.out.println(4395);}}
    public static class Impl4396 implements DemoService {public int type() {return 4396;} public void run() {System.out.println(4396);}}
    public static class Impl4397 implements DemoService {public int type() {return 4397;} public void run() {System.out.println(4397);}}
    public static class Impl4398 implements DemoService {public int type() {return 4398;} public void run() {System.out.println(4398);}}
    public static class Impl4399 implements DemoService {public int type() {return 4399;} public void run() {System.out.println(4399);}}
    public static class Impl4400 implements DemoService {public int type() {return 4400;} public void run() {System.out.println(4400);}}
    public static class Impl4401 implements DemoService {public int type() {return 4401;} public void run() {System.out.println(4401);}}
    public static class Impl4402 implements DemoService {public int type() {return 4402;} public void run() {System.out.println(4402);}}
    public static class Impl4403 implements DemoService {public int type() {return 4403;} public void run() {System.out.println(4403);}}
    public static class Impl4404 implements DemoService {public int type() {return 4404;} public void run() {System.out.println(4404);}}
    public static class Impl4405 implements DemoService {public int type() {return 4405;} public void run() {System.out.println(4405);}}
    public static class Impl4406 implements DemoService {public int type() {return 4406;} public void run() {System.out.println(4406);}}
    public static class Impl4407 implements DemoService {public int type() {return 4407;} public void run() {System.out.println(4407);}}
    public static class Impl4408 implements DemoService {public int type() {return 4408;} public void run() {System.out.println(4408);}}
    public static class Impl4409 implements DemoService {public int type() {return 4409;} public void run() {System.out.println(4409);}}
    public static class Impl4410 implements DemoService {public int type() {return 4410;} public void run() {System.out.println(4410);}}
    public static class Impl4411 implements DemoService {public int type() {return 4411;} public void run() {System.out.println(4411);}}
    public static class Impl4412 implements DemoService {public int type() {return 4412;} public void run() {System.out.println(4412);}}
    public static class Impl4413 implements DemoService {public int type() {return 4413;} public void run() {System.out.println(4413);}}
    public static class Impl4414 implements DemoService {public int type() {return 4414;} public void run() {System.out.println(4414);}}
    public static class Impl4415 implements DemoService {public int type() {return 4415;} public void run() {System.out.println(4415);}}
    public static class Impl4416 implements DemoService {public int type() {return 4416;} public void run() {System.out.println(4416);}}
    public static class Impl4417 implements DemoService {public int type() {return 4417;} public void run() {System.out.println(4417);}}
    public static class Impl4418 implements DemoService {public int type() {return 4418;} public void run() {System.out.println(4418);}}
    public static class Impl4419 implements DemoService {public int type() {return 4419;} public void run() {System.out.println(4419);}}
    public static class Impl4420 implements DemoService {public int type() {return 4420;} public void run() {System.out.println(4420);}}
    public static class Impl4421 implements DemoService {public int type() {return 4421;} public void run() {System.out.println(4421);}}
    public static class Impl4422 implements DemoService {public int type() {return 4422;} public void run() {System.out.println(4422);}}
    public static class Impl4423 implements DemoService {public int type() {return 4423;} public void run() {System.out.println(4423);}}
    public static class Impl4424 implements DemoService {public int type() {return 4424;} public void run() {System.out.println(4424);}}
    public static class Impl4425 implements DemoService {public int type() {return 4425;} public void run() {System.out.println(4425);}}
    public static class Impl4426 implements DemoService {public int type() {return 4426;} public void run() {System.out.println(4426);}}
    public static class Impl4427 implements DemoService {public int type() {return 4427;} public void run() {System.out.println(4427);}}
    public static class Impl4428 implements DemoService {public int type() {return 4428;} public void run() {System.out.println(4428);}}
    public static class Impl4429 implements DemoService {public int type() {return 4429;} public void run() {System.out.println(4429);}}
    public static class Impl4430 implements DemoService {public int type() {return 4430;} public void run() {System.out.println(4430);}}
    public static class Impl4431 implements DemoService {public int type() {return 4431;} public void run() {System.out.println(4431);}}
    public static class Impl4432 implements DemoService {public int type() {return 4432;} public void run() {System.out.println(4432);}}
    public static class Impl4433 implements DemoService {public int type() {return 4433;} public void run() {System.out.println(4433);}}
    public static class Impl4434 implements DemoService {public int type() {return 4434;} public void run() {System.out.println(4434);}}
    public static class Impl4435 implements DemoService {public int type() {return 4435;} public void run() {System.out.println(4435);}}
    public static class Impl4436 implements DemoService {public int type() {return 4436;} public void run() {System.out.println(4436);}}
    public static class Impl4437 implements DemoService {public int type() {return 4437;} public void run() {System.out.println(4437);}}
    public static class Impl4438 implements DemoService {public int type() {return 4438;} public void run() {System.out.println(4438);}}
    public static class Impl4439 implements DemoService {public int type() {return 4439;} public void run() {System.out.println(4439);}}
    public static class Impl4440 implements DemoService {public int type() {return 4440;} public void run() {System.out.println(4440);}}
    public static class Impl4441 implements DemoService {public int type() {return 4441;} public void run() {System.out.println(4441);}}
    public static class Impl4442 implements DemoService {public int type() {return 4442;} public void run() {System.out.println(4442);}}
    public static class Impl4443 implements DemoService {public int type() {return 4443;} public void run() {System.out.println(4443);}}
    public static class Impl4444 implements DemoService {public int type() {return 4444;} public void run() {System.out.println(4444);}}
    public static class Impl4445 implements DemoService {public int type() {return 4445;} public void run() {System.out.println(4445);}}
    public static class Impl4446 implements DemoService {public int type() {return 4446;} public void run() {System.out.println(4446);}}
    public static class Impl4447 implements DemoService {public int type() {return 4447;} public void run() {System.out.println(4447);}}
    public static class Impl4448 implements DemoService {public int type() {return 4448;} public void run() {System.out.println(4448);}}
    public static class Impl4449 implements DemoService {public int type() {return 4449;} public void run() {System.out.println(4449);}}
    public static class Impl4450 implements DemoService {public int type() {return 4450;} public void run() {System.out.println(4450);}}
    public static class Impl4451 implements DemoService {public int type() {return 4451;} public void run() {System.out.println(4451);}}
    public static class Impl4452 implements DemoService {public int type() {return 4452;} public void run() {System.out.println(4452);}}
    public static class Impl4453 implements DemoService {public int type() {return 4453;} public void run() {System.out.println(4453);}}
    public static class Impl4454 implements DemoService {public int type() {return 4454;} public void run() {System.out.println(4454);}}
    public static class Impl4455 implements DemoService {public int type() {return 4455;} public void run() {System.out.println(4455);}}
    public static class Impl4456 implements DemoService {public int type() {return 4456;} public void run() {System.out.println(4456);}}
    public static class Impl4457 implements DemoService {public int type() {return 4457;} public void run() {System.out.println(4457);}}
    public static class Impl4458 implements DemoService {public int type() {return 4458;} public void run() {System.out.println(4458);}}
    public static class Impl4459 implements DemoService {public int type() {return 4459;} public void run() {System.out.println(4459);}}
    public static class Impl4460 implements DemoService {public int type() {return 4460;} public void run() {System.out.println(4460);}}
    public static class Impl4461 implements DemoService {public int type() {return 4461;} public void run() {System.out.println(4461);}}
    public static class Impl4462 implements DemoService {public int type() {return 4462;} public void run() {System.out.println(4462);}}
    public static class Impl4463 implements DemoService {public int type() {return 4463;} public void run() {System.out.println(4463);}}
    public static class Impl4464 implements DemoService {public int type() {return 4464;} public void run() {System.out.println(4464);}}
    public static class Impl4465 implements DemoService {public int type() {return 4465;} public void run() {System.out.println(4465);}}
    public static class Impl4466 implements DemoService {public int type() {return 4466;} public void run() {System.out.println(4466);}}
    public static class Impl4467 implements DemoService {public int type() {return 4467;} public void run() {System.out.println(4467);}}
    public static class Impl4468 implements DemoService {public int type() {return 4468;} public void run() {System.out.println(4468);}}
    public static class Impl4469 implements DemoService {public int type() {return 4469;} public void run() {System.out.println(4469);}}
    public static class Impl4470 implements DemoService {public int type() {return 4470;} public void run() {System.out.println(4470);}}
    public static class Impl4471 implements DemoService {public int type() {return 4471;} public void run() {System.out.println(4471);}}
    public static class Impl4472 implements DemoService {public int type() {return 4472;} public void run() {System.out.println(4472);}}
    public static class Impl4473 implements DemoService {public int type() {return 4473;} public void run() {System.out.println(4473);}}
    public static class Impl4474 implements DemoService {public int type() {return 4474;} public void run() {System.out.println(4474);}}
    public static class Impl4475 implements DemoService {public int type() {return 4475;} public void run() {System.out.println(4475);}}
    public static class Impl4476 implements DemoService {public int type() {return 4476;} public void run() {System.out.println(4476);}}
    public static class Impl4477 implements DemoService {public int type() {return 4477;} public void run() {System.out.println(4477);}}
    public static class Impl4478 implements DemoService {public int type() {return 4478;} public void run() {System.out.println(4478);}}
    public static class Impl4479 implements DemoService {public int type() {return 4479;} public void run() {System.out.println(4479);}}
    public static class Impl4480 implements DemoService {public int type() {return 4480;} public void run() {System.out.println(4480);}}
    public static class Impl4481 implements DemoService {public int type() {return 4481;} public void run() {System.out.println(4481);}}
    public static class Impl4482 implements DemoService {public int type() {return 4482;} public void run() {System.out.println(4482);}}
    public static class Impl4483 implements DemoService {public int type() {return 4483;} public void run() {System.out.println(4483);}}
    public static class Impl4484 implements DemoService {public int type() {return 4484;} public void run() {System.out.println(4484);}}
    public static class Impl4485 implements DemoService {public int type() {return 4485;} public void run() {System.out.println(4485);}}
    public static class Impl4486 implements DemoService {public int type() {return 4486;} public void run() {System.out.println(4486);}}
    public static class Impl4487 implements DemoService {public int type() {return 4487;} public void run() {System.out.println(4487);}}
    public static class Impl4488 implements DemoService {public int type() {return 4488;} public void run() {System.out.println(4488);}}
    public static class Impl4489 implements DemoService {public int type() {return 4489;} public void run() {System.out.println(4489);}}
    public static class Impl4490 implements DemoService {public int type() {return 4490;} public void run() {System.out.println(4490);}}
    public static class Impl4491 implements DemoService {public int type() {return 4491;} public void run() {System.out.println(4491);}}
    public static class Impl4492 implements DemoService {public int type() {return 4492;} public void run() {System.out.println(4492);}}
    public static class Impl4493 implements DemoService {public int type() {return 4493;} public void run() {System.out.println(4493);}}
    public static class Impl4494 implements DemoService {public int type() {return 4494;} public void run() {System.out.println(4494);}}
    public static class Impl4495 implements DemoService {public int type() {return 4495;} public void run() {System.out.println(4495);}}
    public static class Impl4496 implements DemoService {public int type() {return 4496;} public void run() {System.out.println(4496);}}
    public static class Impl4497 implements DemoService {public int type() {return 4497;} public void run() {System.out.println(4497);}}
    public static class Impl4498 implements DemoService {public int type() {return 4498;} public void run() {System.out.println(4498);}}
    public static class Impl4499 implements DemoService {public int type() {return 4499;} public void run() {System.out.println(4499);}}
    public static class Impl4500 implements DemoService {public int type() {return 4500;} public void run() {System.out.println(4500);}}
    public static class Impl4501 implements DemoService {public int type() {return 4501;} public void run() {System.out.println(4501);}}
    public static class Impl4502 implements DemoService {public int type() {return 4502;} public void run() {System.out.println(4502);}}
    public static class Impl4503 implements DemoService {public int type() {return 4503;} public void run() {System.out.println(4503);}}
    public static class Impl4504 implements DemoService {public int type() {return 4504;} public void run() {System.out.println(4504);}}
    public static class Impl4505 implements DemoService {public int type() {return 4505;} public void run() {System.out.println(4505);}}
    public static class Impl4506 implements DemoService {public int type() {return 4506;} public void run() {System.out.println(4506);}}
    public static class Impl4507 implements DemoService {public int type() {return 4507;} public void run() {System.out.println(4507);}}
    public static class Impl4508 implements DemoService {public int type() {return 4508;} public void run() {System.out.println(4508);}}
    public static class Impl4509 implements DemoService {public int type() {return 4509;} public void run() {System.out.println(4509);}}
    public static class Impl4510 implements DemoService {public int type() {return 4510;} public void run() {System.out.println(4510);}}
    public static class Impl4511 implements DemoService {public int type() {return 4511;} public void run() {System.out.println(4511);}}
    public static class Impl4512 implements DemoService {public int type() {return 4512;} public void run() {System.out.println(4512);}}
    public static class Impl4513 implements DemoService {public int type() {return 4513;} public void run() {System.out.println(4513);}}
    public static class Impl4514 implements DemoService {public int type() {return 4514;} public void run() {System.out.println(4514);}}
    public static class Impl4515 implements DemoService {public int type() {return 4515;} public void run() {System.out.println(4515);}}
    public static class Impl4516 implements DemoService {public int type() {return 4516;} public void run() {System.out.println(4516);}}
    public static class Impl4517 implements DemoService {public int type() {return 4517;} public void run() {System.out.println(4517);}}
    public static class Impl4518 implements DemoService {public int type() {return 4518;} public void run() {System.out.println(4518);}}
    public static class Impl4519 implements DemoService {public int type() {return 4519;} public void run() {System.out.println(4519);}}
    public static class Impl4520 implements DemoService {public int type() {return 4520;} public void run() {System.out.println(4520);}}
    public static class Impl4521 implements DemoService {public int type() {return 4521;} public void run() {System.out.println(4521);}}
    public static class Impl4522 implements DemoService {public int type() {return 4522;} public void run() {System.out.println(4522);}}
    public static class Impl4523 implements DemoService {public int type() {return 4523;} public void run() {System.out.println(4523);}}
    public static class Impl4524 implements DemoService {public int type() {return 4524;} public void run() {System.out.println(4524);}}
    public static class Impl4525 implements DemoService {public int type() {return 4525;} public void run() {System.out.println(4525);}}
    public static class Impl4526 implements DemoService {public int type() {return 4526;} public void run() {System.out.println(4526);}}
    public static class Impl4527 implements DemoService {public int type() {return 4527;} public void run() {System.out.println(4527);}}
    public static class Impl4528 implements DemoService {public int type() {return 4528;} public void run() {System.out.println(4528);}}
    public static class Impl4529 implements DemoService {public int type() {return 4529;} public void run() {System.out.println(4529);}}
    public static class Impl4530 implements DemoService {public int type() {return 4530;} public void run() {System.out.println(4530);}}
    public static class Impl4531 implements DemoService {public int type() {return 4531;} public void run() {System.out.println(4531);}}
    public static class Impl4532 implements DemoService {public int type() {return 4532;} public void run() {System.out.println(4532);}}
    public static class Impl4533 implements DemoService {public int type() {return 4533;} public void run() {System.out.println(4533);}}
    public static class Impl4534 implements DemoService {public int type() {return 4534;} public void run() {System.out.println(4534);}}
    public static class Impl4535 implements DemoService {public int type() {return 4535;} public void run() {System.out.println(4535);}}
    public static class Impl4536 implements DemoService {public int type() {return 4536;} public void run() {System.out.println(4536);}}
    public static class Impl4537 implements DemoService {public int type() {return 4537;} public void run() {System.out.println(4537);}}
    public static class Impl4538 implements DemoService {public int type() {return 4538;} public void run() {System.out.println(4538);}}
    public static class Impl4539 implements DemoService {public int type() {return 4539;} public void run() {System.out.println(4539);}}
    public static class Impl4540 implements DemoService {public int type() {return 4540;} public void run() {System.out.println(4540);}}
    public static class Impl4541 implements DemoService {public int type() {return 4541;} public void run() {System.out.println(4541);}}
    public static class Impl4542 implements DemoService {public int type() {return 4542;} public void run() {System.out.println(4542);}}
    public static class Impl4543 implements DemoService {public int type() {return 4543;} public void run() {System.out.println(4543);}}
    public static class Impl4544 implements DemoService {public int type() {return 4544;} public void run() {System.out.println(4544);}}
    public static class Impl4545 implements DemoService {public int type() {return 4545;} public void run() {System.out.println(4545);}}
    public static class Impl4546 implements DemoService {public int type() {return 4546;} public void run() {System.out.println(4546);}}
    public static class Impl4547 implements DemoService {public int type() {return 4547;} public void run() {System.out.println(4547);}}
    public static class Impl4548 implements DemoService {public int type() {return 4548;} public void run() {System.out.println(4548);}}
    public static class Impl4549 implements DemoService {public int type() {return 4549;} public void run() {System.out.println(4549);}}
    public static class Impl4550 implements DemoService {public int type() {return 4550;} public void run() {System.out.println(4550);}}
    public static class Impl4551 implements DemoService {public int type() {return 4551;} public void run() {System.out.println(4551);}}
    public static class Impl4552 implements DemoService {public int type() {return 4552;} public void run() {System.out.println(4552);}}
    public static class Impl4553 implements DemoService {public int type() {return 4553;} public void run() {System.out.println(4553);}}
    public static class Impl4554 implements DemoService {public int type() {return 4554;} public void run() {System.out.println(4554);}}
    public static class Impl4555 implements DemoService {public int type() {return 4555;} public void run() {System.out.println(4555);}}
    public static class Impl4556 implements DemoService {public int type() {return 4556;} public void run() {System.out.println(4556);}}
    public static class Impl4557 implements DemoService {public int type() {return 4557;} public void run() {System.out.println(4557);}}
    public static class Impl4558 implements DemoService {public int type() {return 4558;} public void run() {System.out.println(4558);}}
    public static class Impl4559 implements DemoService {public int type() {return 4559;} public void run() {System.out.println(4559);}}
    public static class Impl4560 implements DemoService {public int type() {return 4560;} public void run() {System.out.println(4560);}}
    public static class Impl4561 implements DemoService {public int type() {return 4561;} public void run() {System.out.println(4561);}}
    public static class Impl4562 implements DemoService {public int type() {return 4562;} public void run() {System.out.println(4562);}}
    public static class Impl4563 implements DemoService {public int type() {return 4563;} public void run() {System.out.println(4563);}}
    public static class Impl4564 implements DemoService {public int type() {return 4564;} public void run() {System.out.println(4564);}}
    public static class Impl4565 implements DemoService {public int type() {return 4565;} public void run() {System.out.println(4565);}}
    public static class Impl4566 implements DemoService {public int type() {return 4566;} public void run() {System.out.println(4566);}}
    public static class Impl4567 implements DemoService {public int type() {return 4567;} public void run() {System.out.println(4567);}}
    public static class Impl4568 implements DemoService {public int type() {return 4568;} public void run() {System.out.println(4568);}}
    public static class Impl4569 implements DemoService {public int type() {return 4569;} public void run() {System.out.println(4569);}}
    public static class Impl4570 implements DemoService {public int type() {return 4570;} public void run() {System.out.println(4570);}}
    public static class Impl4571 implements DemoService {public int type() {return 4571;} public void run() {System.out.println(4571);}}
    public static class Impl4572 implements DemoService {public int type() {return 4572;} public void run() {System.out.println(4572);}}
    public static class Impl4573 implements DemoService {public int type() {return 4573;} public void run() {System.out.println(4573);}}
    public static class Impl4574 implements DemoService {public int type() {return 4574;} public void run() {System.out.println(4574);}}
    public static class Impl4575 implements DemoService {public int type() {return 4575;} public void run() {System.out.println(4575);}}
    public static class Impl4576 implements DemoService {public int type() {return 4576;} public void run() {System.out.println(4576);}}
    public static class Impl4577 implements DemoService {public int type() {return 4577;} public void run() {System.out.println(4577);}}
    public static class Impl4578 implements DemoService {public int type() {return 4578;} public void run() {System.out.println(4578);}}
    public static class Impl4579 implements DemoService {public int type() {return 4579;} public void run() {System.out.println(4579);}}
    public static class Impl4580 implements DemoService {public int type() {return 4580;} public void run() {System.out.println(4580);}}
    public static class Impl4581 implements DemoService {public int type() {return 4581;} public void run() {System.out.println(4581);}}
    public static class Impl4582 implements DemoService {public int type() {return 4582;} public void run() {System.out.println(4582);}}
    public static class Impl4583 implements DemoService {public int type() {return 4583;} public void run() {System.out.println(4583);}}
    public static class Impl4584 implements DemoService {public int type() {return 4584;} public void run() {System.out.println(4584);}}
    public static class Impl4585 implements DemoService {public int type() {return 4585;} public void run() {System.out.println(4585);}}
    public static class Impl4586 implements DemoService {public int type() {return 4586;} public void run() {System.out.println(4586);}}
    public static class Impl4587 implements DemoService {public int type() {return 4587;} public void run() {System.out.println(4587);}}
    public static class Impl4588 implements DemoService {public int type() {return 4588;} public void run() {System.out.println(4588);}}
    public static class Impl4589 implements DemoService {public int type() {return 4589;} public void run() {System.out.println(4589);}}
    public static class Impl4590 implements DemoService {public int type() {return 4590;} public void run() {System.out.println(4590);}}
    public static class Impl4591 implements DemoService {public int type() {return 4591;} public void run() {System.out.println(4591);}}
    public static class Impl4592 implements DemoService {public int type() {return 4592;} public void run() {System.out.println(4592);}}
    public static class Impl4593 implements DemoService {public int type() {return 4593;} public void run() {System.out.println(4593);}}
    public static class Impl4594 implements DemoService {public int type() {return 4594;} public void run() {System.out.println(4594);}}
    public static class Impl4595 implements DemoService {public int type() {return 4595;} public void run() {System.out.println(4595);}}
    public static class Impl4596 implements DemoService {public int type() {return 4596;} public void run() {System.out.println(4596);}}
    public static class Impl4597 implements DemoService {public int type() {return 4597;} public void run() {System.out.println(4597);}}
    public static class Impl4598 implements DemoService {public int type() {return 4598;} public void run() {System.out.println(4598);}}
    public static class Impl4599 implements DemoService {public int type() {return 4599;} public void run() {System.out.println(4599);}}
    public static class Impl4600 implements DemoService {public int type() {return 4600;} public void run() {System.out.println(4600);}}
    public static class Impl4601 implements DemoService {public int type() {return 4601;} public void run() {System.out.println(4601);}}
    public static class Impl4602 implements DemoService {public int type() {return 4602;} public void run() {System.out.println(4602);}}
    public static class Impl4603 implements DemoService {public int type() {return 4603;} public void run() {System.out.println(4603);}}
    public static class Impl4604 implements DemoService {public int type() {return 4604;} public void run() {System.out.println(4604);}}
    public static class Impl4605 implements DemoService {public int type() {return 4605;} public void run() {System.out.println(4605);}}
    public static class Impl4606 implements DemoService {public int type() {return 4606;} public void run() {System.out.println(4606);}}
    public static class Impl4607 implements DemoService {public int type() {return 4607;} public void run() {System.out.println(4607);}}
    public static class Impl4608 implements DemoService {public int type() {return 4608;} public void run() {System.out.println(4608);}}
    public static class Impl4609 implements DemoService {public int type() {return 4609;} public void run() {System.out.println(4609);}}
    public static class Impl4610 implements DemoService {public int type() {return 4610;} public void run() {System.out.println(4610);}}
    public static class Impl4611 implements DemoService {public int type() {return 4611;} public void run() {System.out.println(4611);}}
    public static class Impl4612 implements DemoService {public int type() {return 4612;} public void run() {System.out.println(4612);}}
    public static class Impl4613 implements DemoService {public int type() {return 4613;} public void run() {System.out.println(4613);}}
    public static class Impl4614 implements DemoService {public int type() {return 4614;} public void run() {System.out.println(4614);}}
    public static class Impl4615 implements DemoService {public int type() {return 4615;} public void run() {System.out.println(4615);}}
    public static class Impl4616 implements DemoService {public int type() {return 4616;} public void run() {System.out.println(4616);}}
    public static class Impl4617 implements DemoService {public int type() {return 4617;} public void run() {System.out.println(4617);}}
    public static class Impl4618 implements DemoService {public int type() {return 4618;} public void run() {System.out.println(4618);}}
    public static class Impl4619 implements DemoService {public int type() {return 4619;} public void run() {System.out.println(4619);}}
    public static class Impl4620 implements DemoService {public int type() {return 4620;} public void run() {System.out.println(4620);}}
    public static class Impl4621 implements DemoService {public int type() {return 4621;} public void run() {System.out.println(4621);}}
    public static class Impl4622 implements DemoService {public int type() {return 4622;} public void run() {System.out.println(4622);}}
    public static class Impl4623 implements DemoService {public int type() {return 4623;} public void run() {System.out.println(4623);}}
    public static class Impl4624 implements DemoService {public int type() {return 4624;} public void run() {System.out.println(4624);}}
    public static class Impl4625 implements DemoService {public int type() {return 4625;} public void run() {System.out.println(4625);}}
    public static class Impl4626 implements DemoService {public int type() {return 4626;} public void run() {System.out.println(4626);}}
    public static class Impl4627 implements DemoService {public int type() {return 4627;} public void run() {System.out.println(4627);}}
    public static class Impl4628 implements DemoService {public int type() {return 4628;} public void run() {System.out.println(4628);}}
    public static class Impl4629 implements DemoService {public int type() {return 4629;} public void run() {System.out.println(4629);}}
    public static class Impl4630 implements DemoService {public int type() {return 4630;} public void run() {System.out.println(4630);}}
    public static class Impl4631 implements DemoService {public int type() {return 4631;} public void run() {System.out.println(4631);}}
    public static class Impl4632 implements DemoService {public int type() {return 4632;} public void run() {System.out.println(4632);}}
    public static class Impl4633 implements DemoService {public int type() {return 4633;} public void run() {System.out.println(4633);}}
    public static class Impl4634 implements DemoService {public int type() {return 4634;} public void run() {System.out.println(4634);}}
    public static class Impl4635 implements DemoService {public int type() {return 4635;} public void run() {System.out.println(4635);}}
    public static class Impl4636 implements DemoService {public int type() {return 4636;} public void run() {System.out.println(4636);}}
    public static class Impl4637 implements DemoService {public int type() {return 4637;} public void run() {System.out.println(4637);}}
    public static class Impl4638 implements DemoService {public int type() {return 4638;} public void run() {System.out.println(4638);}}
    public static class Impl4639 implements DemoService {public int type() {return 4639;} public void run() {System.out.println(4639);}}
    public static class Impl4640 implements DemoService {public int type() {return 4640;} public void run() {System.out.println(4640);}}
    public static class Impl4641 implements DemoService {public int type() {return 4641;} public void run() {System.out.println(4641);}}
    public static class Impl4642 implements DemoService {public int type() {return 4642;} public void run() {System.out.println(4642);}}
    public static class Impl4643 implements DemoService {public int type() {return 4643;} public void run() {System.out.println(4643);}}
    public static class Impl4644 implements DemoService {public int type() {return 4644;} public void run() {System.out.println(4644);}}
    public static class Impl4645 implements DemoService {public int type() {return 4645;} public void run() {System.out.println(4645);}}
    public static class Impl4646 implements DemoService {public int type() {return 4646;} public void run() {System.out.println(4646);}}
    public static class Impl4647 implements DemoService {public int type() {return 4647;} public void run() {System.out.println(4647);}}
    public static class Impl4648 implements DemoService {public int type() {return 4648;} public void run() {System.out.println(4648);}}
    public static class Impl4649 implements DemoService {public int type() {return 4649;} public void run() {System.out.println(4649);}}
    public static class Impl4650 implements DemoService {public int type() {return 4650;} public void run() {System.out.println(4650);}}
    public static class Impl4651 implements DemoService {public int type() {return 4651;} public void run() {System.out.println(4651);}}
    public static class Impl4652 implements DemoService {public int type() {return 4652;} public void run() {System.out.println(4652);}}
    public static class Impl4653 implements DemoService {public int type() {return 4653;} public void run() {System.out.println(4653);}}
    public static class Impl4654 implements DemoService {public int type() {return 4654;} public void run() {System.out.println(4654);}}
    public static class Impl4655 implements DemoService {public int type() {return 4655;} public void run() {System.out.println(4655);}}
    public static class Impl4656 implements DemoService {public int type() {return 4656;} public void run() {System.out.println(4656);}}
    public static class Impl4657 implements DemoService {public int type() {return 4657;} public void run() {System.out.println(4657);}}
    public static class Impl4658 implements DemoService {public int type() {return 4658;} public void run() {System.out.println(4658);}}
    public static class Impl4659 implements DemoService {public int type() {return 4659;} public void run() {System.out.println(4659);}}
    public static class Impl4660 implements DemoService {public int type() {return 4660;} public void run() {System.out.println(4660);}}
    public static class Impl4661 implements DemoService {public int type() {return 4661;} public void run() {System.out.println(4661);}}
    public static class Impl4662 implements DemoService {public int type() {return 4662;} public void run() {System.out.println(4662);}}
    public static class Impl4663 implements DemoService {public int type() {return 4663;} public void run() {System.out.println(4663);}}
    public static class Impl4664 implements DemoService {public int type() {return 4664;} public void run() {System.out.println(4664);}}
    public static class Impl4665 implements DemoService {public int type() {return 4665;} public void run() {System.out.println(4665);}}
    public static class Impl4666 implements DemoService {public int type() {return 4666;} public void run() {System.out.println(4666);}}
    public static class Impl4667 implements DemoService {public int type() {return 4667;} public void run() {System.out.println(4667);}}
    public static class Impl4668 implements DemoService {public int type() {return 4668;} public void run() {System.out.println(4668);}}
    public static class Impl4669 implements DemoService {public int type() {return 4669;} public void run() {System.out.println(4669);}}
    public static class Impl4670 implements DemoService {public int type() {return 4670;} public void run() {System.out.println(4670);}}
    public static class Impl4671 implements DemoService {public int type() {return 4671;} public void run() {System.out.println(4671);}}
    public static class Impl4672 implements DemoService {public int type() {return 4672;} public void run() {System.out.println(4672);}}
    public static class Impl4673 implements DemoService {public int type() {return 4673;} public void run() {System.out.println(4673);}}
    public static class Impl4674 implements DemoService {public int type() {return 4674;} public void run() {System.out.println(4674);}}
    public static class Impl4675 implements DemoService {public int type() {return 4675;} public void run() {System.out.println(4675);}}
    public static class Impl4676 implements DemoService {public int type() {return 4676;} public void run() {System.out.println(4676);}}
    public static class Impl4677 implements DemoService {public int type() {return 4677;} public void run() {System.out.println(4677);}}
    public static class Impl4678 implements DemoService {public int type() {return 4678;} public void run() {System.out.println(4678);}}
    public static class Impl4679 implements DemoService {public int type() {return 4679;} public void run() {System.out.println(4679);}}
    public static class Impl4680 implements DemoService {public int type() {return 4680;} public void run() {System.out.println(4680);}}
    public static class Impl4681 implements DemoService {public int type() {return 4681;} public void run() {System.out.println(4681);}}
    public static class Impl4682 implements DemoService {public int type() {return 4682;} public void run() {System.out.println(4682);}}
    public static class Impl4683 implements DemoService {public int type() {return 4683;} public void run() {System.out.println(4683);}}
    public static class Impl4684 implements DemoService {public int type() {return 4684;} public void run() {System.out.println(4684);}}
    public static class Impl4685 implements DemoService {public int type() {return 4685;} public void run() {System.out.println(4685);}}
    public static class Impl4686 implements DemoService {public int type() {return 4686;} public void run() {System.out.println(4686);}}
    public static class Impl4687 implements DemoService {public int type() {return 4687;} public void run() {System.out.println(4687);}}
    public static class Impl4688 implements DemoService {public int type() {return 4688;} public void run() {System.out.println(4688);}}
    public static class Impl4689 implements DemoService {public int type() {return 4689;} public void run() {System.out.println(4689);}}
    public static class Impl4690 implements DemoService {public int type() {return 4690;} public void run() {System.out.println(4690);}}
    public static class Impl4691 implements DemoService {public int type() {return 4691;} public void run() {System.out.println(4691);}}
    public static class Impl4692 implements DemoService {public int type() {return 4692;} public void run() {System.out.println(4692);}}
    public static class Impl4693 implements DemoService {public int type() {return 4693;} public void run() {System.out.println(4693);}}
    public static class Impl4694 implements DemoService {public int type() {return 4694;} public void run() {System.out.println(4694);}}
    public static class Impl4695 implements DemoService {public int type() {return 4695;} public void run() {System.out.println(4695);}}
    public static class Impl4696 implements DemoService {public int type() {return 4696;} public void run() {System.out.println(4696);}}
    public static class Impl4697 implements DemoService {public int type() {return 4697;} public void run() {System.out.println(4697);}}
    public static class Impl4698 implements DemoService {public int type() {return 4698;} public void run() {System.out.println(4698);}}
    public static class Impl4699 implements DemoService {public int type() {return 4699;} public void run() {System.out.println(4699);}}
    public static class Impl4700 implements DemoService {public int type() {return 4700;} public void run() {System.out.println(4700);}}
    public static class Impl4701 implements DemoService {public int type() {return 4701;} public void run() {System.out.println(4701);}}
    public static class Impl4702 implements DemoService {public int type() {return 4702;} public void run() {System.out.println(4702);}}
    public static class Impl4703 implements DemoService {public int type() {return 4703;} public void run() {System.out.println(4703);}}
    public static class Impl4704 implements DemoService {public int type() {return 4704;} public void run() {System.out.println(4704);}}
    public static class Impl4705 implements DemoService {public int type() {return 4705;} public void run() {System.out.println(4705);}}
    public static class Impl4706 implements DemoService {public int type() {return 4706;} public void run() {System.out.println(4706);}}
    public static class Impl4707 implements DemoService {public int type() {return 4707;} public void run() {System.out.println(4707);}}
    public static class Impl4708 implements DemoService {public int type() {return 4708;} public void run() {System.out.println(4708);}}
    public static class Impl4709 implements DemoService {public int type() {return 4709;} public void run() {System.out.println(4709);}}
    public static class Impl4710 implements DemoService {public int type() {return 4710;} public void run() {System.out.println(4710);}}
    public static class Impl4711 implements DemoService {public int type() {return 4711;} public void run() {System.out.println(4711);}}
    public static class Impl4712 implements DemoService {public int type() {return 4712;} public void run() {System.out.println(4712);}}
    public static class Impl4713 implements DemoService {public int type() {return 4713;} public void run() {System.out.println(4713);}}
    public static class Impl4714 implements DemoService {public int type() {return 4714;} public void run() {System.out.println(4714);}}
    public static class Impl4715 implements DemoService {public int type() {return 4715;} public void run() {System.out.println(4715);}}
    public static class Impl4716 implements DemoService {public int type() {return 4716;} public void run() {System.out.println(4716);}}
    public static class Impl4717 implements DemoService {public int type() {return 4717;} public void run() {System.out.println(4717);}}
    public static class Impl4718 implements DemoService {public int type() {return 4718;} public void run() {System.out.println(4718);}}
    public static class Impl4719 implements DemoService {public int type() {return 4719;} public void run() {System.out.println(4719);}}
    public static class Impl4720 implements DemoService {public int type() {return 4720;} public void run() {System.out.println(4720);}}
    public static class Impl4721 implements DemoService {public int type() {return 4721;} public void run() {System.out.println(4721);}}
    public static class Impl4722 implements DemoService {public int type() {return 4722;} public void run() {System.out.println(4722);}}
    public static class Impl4723 implements DemoService {public int type() {return 4723;} public void run() {System.out.println(4723);}}
    public static class Impl4724 implements DemoService {public int type() {return 4724;} public void run() {System.out.println(4724);}}
    public static class Impl4725 implements DemoService {public int type() {return 4725;} public void run() {System.out.println(4725);}}
    public static class Impl4726 implements DemoService {public int type() {return 4726;} public void run() {System.out.println(4726);}}
    public static class Impl4727 implements DemoService {public int type() {return 4727;} public void run() {System.out.println(4727);}}
    public static class Impl4728 implements DemoService {public int type() {return 4728;} public void run() {System.out.println(4728);}}
    public static class Impl4729 implements DemoService {public int type() {return 4729;} public void run() {System.out.println(4729);}}
    public static class Impl4730 implements DemoService {public int type() {return 4730;} public void run() {System.out.println(4730);}}
    public static class Impl4731 implements DemoService {public int type() {return 4731;} public void run() {System.out.println(4731);}}
    public static class Impl4732 implements DemoService {public int type() {return 4732;} public void run() {System.out.println(4732);}}
    public static class Impl4733 implements DemoService {public int type() {return 4733;} public void run() {System.out.println(4733);}}
    public static class Impl4734 implements DemoService {public int type() {return 4734;} public void run() {System.out.println(4734);}}
    public static class Impl4735 implements DemoService {public int type() {return 4735;} public void run() {System.out.println(4735);}}
    public static class Impl4736 implements DemoService {public int type() {return 4736;} public void run() {System.out.println(4736);}}
    public static class Impl4737 implements DemoService {public int type() {return 4737;} public void run() {System.out.println(4737);}}
    public static class Impl4738 implements DemoService {public int type() {return 4738;} public void run() {System.out.println(4738);}}
    public static class Impl4739 implements DemoService {public int type() {return 4739;} public void run() {System.out.println(4739);}}
    public static class Impl4740 implements DemoService {public int type() {return 4740;} public void run() {System.out.println(4740);}}
    public static class Impl4741 implements DemoService {public int type() {return 4741;} public void run() {System.out.println(4741);}}
    public static class Impl4742 implements DemoService {public int type() {return 4742;} public void run() {System.out.println(4742);}}
    public static class Impl4743 implements DemoService {public int type() {return 4743;} public void run() {System.out.println(4743);}}
    public static class Impl4744 implements DemoService {public int type() {return 4744;} public void run() {System.out.println(4744);}}
    public static class Impl4745 implements DemoService {public int type() {return 4745;} public void run() {System.out.println(4745);}}
    public static class Impl4746 implements DemoService {public int type() {return 4746;} public void run() {System.out.println(4746);}}
    public static class Impl4747 implements DemoService {public int type() {return 4747;} public void run() {System.out.println(4747);}}
    public static class Impl4748 implements DemoService {public int type() {return 4748;} public void run() {System.out.println(4748);}}
    public static class Impl4749 implements DemoService {public int type() {return 4749;} public void run() {System.out.println(4749);}}
    public static class Impl4750 implements DemoService {public int type() {return 4750;} public void run() {System.out.println(4750);}}
    public static class Impl4751 implements DemoService {public int type() {return 4751;} public void run() {System.out.println(4751);}}
    public static class Impl4752 implements DemoService {public int type() {return 4752;} public void run() {System.out.println(4752);}}
    public static class Impl4753 implements DemoService {public int type() {return 4753;} public void run() {System.out.println(4753);}}
    public static class Impl4754 implements DemoService {public int type() {return 4754;} public void run() {System.out.println(4754);}}
    public static class Impl4755 implements DemoService {public int type() {return 4755;} public void run() {System.out.println(4755);}}
    public static class Impl4756 implements DemoService {public int type() {return 4756;} public void run() {System.out.println(4756);}}
    public static class Impl4757 implements DemoService {public int type() {return 4757;} public void run() {System.out.println(4757);}}
    public static class Impl4758 implements DemoService {public int type() {return 4758;} public void run() {System.out.println(4758);}}
    public static class Impl4759 implements DemoService {public int type() {return 4759;} public void run() {System.out.println(4759);}}
    public static class Impl4760 implements DemoService {public int type() {return 4760;} public void run() {System.out.println(4760);}}
    public static class Impl4761 implements DemoService {public int type() {return 4761;} public void run() {System.out.println(4761);}}
    public static class Impl4762 implements DemoService {public int type() {return 4762;} public void run() {System.out.println(4762);}}
    public static class Impl4763 implements DemoService {public int type() {return 4763;} public void run() {System.out.println(4763);}}
    public static class Impl4764 implements DemoService {public int type() {return 4764;} public void run() {System.out.println(4764);}}
    public static class Impl4765 implements DemoService {public int type() {return 4765;} public void run() {System.out.println(4765);}}
    public static class Impl4766 implements DemoService {public int type() {return 4766;} public void run() {System.out.println(4766);}}
    public static class Impl4767 implements DemoService {public int type() {return 4767;} public void run() {System.out.println(4767);}}
    public static class Impl4768 implements DemoService {public int type() {return 4768;} public void run() {System.out.println(4768);}}
    public static class Impl4769 implements DemoService {public int type() {return 4769;} public void run() {System.out.println(4769);}}
    public static class Impl4770 implements DemoService {public int type() {return 4770;} public void run() {System.out.println(4770);}}
    public static class Impl4771 implements DemoService {public int type() {return 4771;} public void run() {System.out.println(4771);}}
    public static class Impl4772 implements DemoService {public int type() {return 4772;} public void run() {System.out.println(4772);}}
    public static class Impl4773 implements DemoService {public int type() {return 4773;} public void run() {System.out.println(4773);}}
    public static class Impl4774 implements DemoService {public int type() {return 4774;} public void run() {System.out.println(4774);}}
    public static class Impl4775 implements DemoService {public int type() {return 4775;} public void run() {System.out.println(4775);}}
    public static class Impl4776 implements DemoService {public int type() {return 4776;} public void run() {System.out.println(4776);}}
    public static class Impl4777 implements DemoService {public int type() {return 4777;} public void run() {System.out.println(4777);}}
    public static class Impl4778 implements DemoService {public int type() {return 4778;} public void run() {System.out.println(4778);}}
    public static class Impl4779 implements DemoService {public int type() {return 4779;} public void run() {System.out.println(4779);}}
    public static class Impl4780 implements DemoService {public int type() {return 4780;} public void run() {System.out.println(4780);}}
    public static class Impl4781 implements DemoService {public int type() {return 4781;} public void run() {System.out.println(4781);}}
    public static class Impl4782 implements DemoService {public int type() {return 4782;} public void run() {System.out.println(4782);}}
    public static class Impl4783 implements DemoService {public int type() {return 4783;} public void run() {System.out.println(4783);}}
    public static class Impl4784 implements DemoService {public int type() {return 4784;} public void run() {System.out.println(4784);}}
    public static class Impl4785 implements DemoService {public int type() {return 4785;} public void run() {System.out.println(4785);}}
    public static class Impl4786 implements DemoService {public int type() {return 4786;} public void run() {System.out.println(4786);}}
    public static class Impl4787 implements DemoService {public int type() {return 4787;} public void run() {System.out.println(4787);}}
    public static class Impl4788 implements DemoService {public int type() {return 4788;} public void run() {System.out.println(4788);}}
    public static class Impl4789 implements DemoService {public int type() {return 4789;} public void run() {System.out.println(4789);}}
    public static class Impl4790 implements DemoService {public int type() {return 4790;} public void run() {System.out.println(4790);}}
    public static class Impl4791 implements DemoService {public int type() {return 4791;} public void run() {System.out.println(4791);}}
    public static class Impl4792 implements DemoService {public int type() {return 4792;} public void run() {System.out.println(4792);}}
    public static class Impl4793 implements DemoService {public int type() {return 4793;} public void run() {System.out.println(4793);}}
    public static class Impl4794 implements DemoService {public int type() {return 4794;} public void run() {System.out.println(4794);}}
    public static class Impl4795 implements DemoService {public int type() {return 4795;} public void run() {System.out.println(4795);}}
    public static class Impl4796 implements DemoService {public int type() {return 4796;} public void run() {System.out.println(4796);}}
    public static class Impl4797 implements DemoService {public int type() {return 4797;} public void run() {System.out.println(4797);}}
    public static class Impl4798 implements DemoService {public int type() {return 4798;} public void run() {System.out.println(4798);}}
    public static class Impl4799 implements DemoService {public int type() {return 4799;} public void run() {System.out.println(4799);}}
    public static class Impl4800 implements DemoService {public int type() {return 4800;} public void run() {System.out.println(4800);}}
    public static class Impl4801 implements DemoService {public int type() {return 4801;} public void run() {System.out.println(4801);}}
    public static class Impl4802 implements DemoService {public int type() {return 4802;} public void run() {System.out.println(4802);}}
    public static class Impl4803 implements DemoService {public int type() {return 4803;} public void run() {System.out.println(4803);}}
    public static class Impl4804 implements DemoService {public int type() {return 4804;} public void run() {System.out.println(4804);}}
    public static class Impl4805 implements DemoService {public int type() {return 4805;} public void run() {System.out.println(4805);}}
    public static class Impl4806 implements DemoService {public int type() {return 4806;} public void run() {System.out.println(4806);}}
    public static class Impl4807 implements DemoService {public int type() {return 4807;} public void run() {System.out.println(4807);}}
    public static class Impl4808 implements DemoService {public int type() {return 4808;} public void run() {System.out.println(4808);}}
    public static class Impl4809 implements DemoService {public int type() {return 4809;} public void run() {System.out.println(4809);}}
    public static class Impl4810 implements DemoService {public int type() {return 4810;} public void run() {System.out.println(4810);}}
    public static class Impl4811 implements DemoService {public int type() {return 4811;} public void run() {System.out.println(4811);}}
    public static class Impl4812 implements DemoService {public int type() {return 4812;} public void run() {System.out.println(4812);}}
    public static class Impl4813 implements DemoService {public int type() {return 4813;} public void run() {System.out.println(4813);}}
    public static class Impl4814 implements DemoService {public int type() {return 4814;} public void run() {System.out.println(4814);}}
    public static class Impl4815 implements DemoService {public int type() {return 4815;} public void run() {System.out.println(4815);}}
    public static class Impl4816 implements DemoService {public int type() {return 4816;} public void run() {System.out.println(4816);}}
    public static class Impl4817 implements DemoService {public int type() {return 4817;} public void run() {System.out.println(4817);}}
    public static class Impl4818 implements DemoService {public int type() {return 4818;} public void run() {System.out.println(4818);}}
    public static class Impl4819 implements DemoService {public int type() {return 4819;} public void run() {System.out.println(4819);}}
    public static class Impl4820 implements DemoService {public int type() {return 4820;} public void run() {System.out.println(4820);}}
    public static class Impl4821 implements DemoService {public int type() {return 4821;} public void run() {System.out.println(4821);}}
    public static class Impl4822 implements DemoService {public int type() {return 4822;} public void run() {System.out.println(4822);}}
    public static class Impl4823 implements DemoService {public int type() {return 4823;} public void run() {System.out.println(4823);}}
    public static class Impl4824 implements DemoService {public int type() {return 4824;} public void run() {System.out.println(4824);}}
    public static class Impl4825 implements DemoService {public int type() {return 4825;} public void run() {System.out.println(4825);}}
    public static class Impl4826 implements DemoService {public int type() {return 4826;} public void run() {System.out.println(4826);}}
    public static class Impl4827 implements DemoService {public int type() {return 4827;} public void run() {System.out.println(4827);}}
    public static class Impl4828 implements DemoService {public int type() {return 4828;} public void run() {System.out.println(4828);}}
    public static class Impl4829 implements DemoService {public int type() {return 4829;} public void run() {System.out.println(4829);}}
    public static class Impl4830 implements DemoService {public int type() {return 4830;} public void run() {System.out.println(4830);}}
    public static class Impl4831 implements DemoService {public int type() {return 4831;} public void run() {System.out.println(4831);}}
    public static class Impl4832 implements DemoService {public int type() {return 4832;} public void run() {System.out.println(4832);}}
    public static class Impl4833 implements DemoService {public int type() {return 4833;} public void run() {System.out.println(4833);}}
    public static class Impl4834 implements DemoService {public int type() {return 4834;} public void run() {System.out.println(4834);}}
    public static class Impl4835 implements DemoService {public int type() {return 4835;} public void run() {System.out.println(4835);}}
    public static class Impl4836 implements DemoService {public int type() {return 4836;} public void run() {System.out.println(4836);}}
    public static class Impl4837 implements DemoService {public int type() {return 4837;} public void run() {System.out.println(4837);}}
    public static class Impl4838 implements DemoService {public int type() {return 4838;} public void run() {System.out.println(4838);}}
    public static class Impl4839 implements DemoService {public int type() {return 4839;} public void run() {System.out.println(4839);}}
    public static class Impl4840 implements DemoService {public int type() {return 4840;} public void run() {System.out.println(4840);}}
    public static class Impl4841 implements DemoService {public int type() {return 4841;} public void run() {System.out.println(4841);}}
    public static class Impl4842 implements DemoService {public int type() {return 4842;} public void run() {System.out.println(4842);}}
    public static class Impl4843 implements DemoService {public int type() {return 4843;} public void run() {System.out.println(4843);}}
    public static class Impl4844 implements DemoService {public int type() {return 4844;} public void run() {System.out.println(4844);}}
    public static class Impl4845 implements DemoService {public int type() {return 4845;} public void run() {System.out.println(4845);}}
    public static class Impl4846 implements DemoService {public int type() {return 4846;} public void run() {System.out.println(4846);}}
    public static class Impl4847 implements DemoService {public int type() {return 4847;} public void run() {System.out.println(4847);}}
    public static class Impl4848 implements DemoService {public int type() {return 4848;} public void run() {System.out.println(4848);}}
    public static class Impl4849 implements DemoService {public int type() {return 4849;} public void run() {System.out.println(4849);}}
    public static class Impl4850 implements DemoService {public int type() {return 4850;} public void run() {System.out.println(4850);}}
    public static class Impl4851 implements DemoService {public int type() {return 4851;} public void run() {System.out.println(4851);}}
    public static class Impl4852 implements DemoService {public int type() {return 4852;} public void run() {System.out.println(4852);}}
    public static class Impl4853 implements DemoService {public int type() {return 4853;} public void run() {System.out.println(4853);}}
    public static class Impl4854 implements DemoService {public int type() {return 4854;} public void run() {System.out.println(4854);}}
    public static class Impl4855 implements DemoService {public int type() {return 4855;} public void run() {System.out.println(4855);}}
    public static class Impl4856 implements DemoService {public int type() {return 4856;} public void run() {System.out.println(4856);}}
    public static class Impl4857 implements DemoService {public int type() {return 4857;} public void run() {System.out.println(4857);}}
    public static class Impl4858 implements DemoService {public int type() {return 4858;} public void run() {System.out.println(4858);}}
    public static class Impl4859 implements DemoService {public int type() {return 4859;} public void run() {System.out.println(4859);}}
    public static class Impl4860 implements DemoService {public int type() {return 4860;} public void run() {System.out.println(4860);}}
    public static class Impl4861 implements DemoService {public int type() {return 4861;} public void run() {System.out.println(4861);}}
    public static class Impl4862 implements DemoService {public int type() {return 4862;} public void run() {System.out.println(4862);}}
    public static class Impl4863 implements DemoService {public int type() {return 4863;} public void run() {System.out.println(4863);}}
    public static class Impl4864 implements DemoService {public int type() {return 4864;} public void run() {System.out.println(4864);}}
    public static class Impl4865 implements DemoService {public int type() {return 4865;} public void run() {System.out.println(4865);}}
    public static class Impl4866 implements DemoService {public int type() {return 4866;} public void run() {System.out.println(4866);}}
    public static class Impl4867 implements DemoService {public int type() {return 4867;} public void run() {System.out.println(4867);}}
    public static class Impl4868 implements DemoService {public int type() {return 4868;} public void run() {System.out.println(4868);}}
    public static class Impl4869 implements DemoService {public int type() {return 4869;} public void run() {System.out.println(4869);}}
    public static class Impl4870 implements DemoService {public int type() {return 4870;} public void run() {System.out.println(4870);}}
    public static class Impl4871 implements DemoService {public int type() {return 4871;} public void run() {System.out.println(4871);}}
    public static class Impl4872 implements DemoService {public int type() {return 4872;} public void run() {System.out.println(4872);}}
    public static class Impl4873 implements DemoService {public int type() {return 4873;} public void run() {System.out.println(4873);}}
    public static class Impl4874 implements DemoService {public int type() {return 4874;} public void run() {System.out.println(4874);}}
    public static class Impl4875 implements DemoService {public int type() {return 4875;} public void run() {System.out.println(4875);}}
    public static class Impl4876 implements DemoService {public int type() {return 4876;} public void run() {System.out.println(4876);}}
    public static class Impl4877 implements DemoService {public int type() {return 4877;} public void run() {System.out.println(4877);}}
    public static class Impl4878 implements DemoService {public int type() {return 4878;} public void run() {System.out.println(4878);}}
    public static class Impl4879 implements DemoService {public int type() {return 4879;} public void run() {System.out.println(4879);}}
    public static class Impl4880 implements DemoService {public int type() {return 4880;} public void run() {System.out.println(4880);}}
    public static class Impl4881 implements DemoService {public int type() {return 4881;} public void run() {System.out.println(4881);}}
    public static class Impl4882 implements DemoService {public int type() {return 4882;} public void run() {System.out.println(4882);}}
    public static class Impl4883 implements DemoService {public int type() {return 4883;} public void run() {System.out.println(4883);}}
    public static class Impl4884 implements DemoService {public int type() {return 4884;} public void run() {System.out.println(4884);}}
    public static class Impl4885 implements DemoService {public int type() {return 4885;} public void run() {System.out.println(4885);}}
    public static class Impl4886 implements DemoService {public int type() {return 4886;} public void run() {System.out.println(4886);}}
    public static class Impl4887 implements DemoService {public int type() {return 4887;} public void run() {System.out.println(4887);}}
    public static class Impl4888 implements DemoService {public int type() {return 4888;} public void run() {System.out.println(4888);}}
    public static class Impl4889 implements DemoService {public int type() {return 4889;} public void run() {System.out.println(4889);}}
    public static class Impl4890 implements DemoService {public int type() {return 4890;} public void run() {System.out.println(4890);}}
    public static class Impl4891 implements DemoService {public int type() {return 4891;} public void run() {System.out.println(4891);}}
    public static class Impl4892 implements DemoService {public int type() {return 4892;} public void run() {System.out.println(4892);}}
    public static class Impl4893 implements DemoService {public int type() {return 4893;} public void run() {System.out.println(4893);}}
    public static class Impl4894 implements DemoService {public int type() {return 4894;} public void run() {System.out.println(4894);}}
    public static class Impl4895 implements DemoService {public int type() {return 4895;} public void run() {System.out.println(4895);}}
    public static class Impl4896 implements DemoService {public int type() {return 4896;} public void run() {System.out.println(4896);}}
    public static class Impl4897 implements DemoService {public int type() {return 4897;} public void run() {System.out.println(4897);}}
    public static class Impl4898 implements DemoService {public int type() {return 4898;} public void run() {System.out.println(4898);}}
    public static class Impl4899 implements DemoService {public int type() {return 4899;} public void run() {System.out.println(4899);}}
    public static class Impl4900 implements DemoService {public int type() {return 4900;} public void run() {System.out.println(4900);}}
    public static class Impl4901 implements DemoService {public int type() {return 4901;} public void run() {System.out.println(4901);}}
    public static class Impl4902 implements DemoService {public int type() {return 4902;} public void run() {System.out.println(4902);}}
    public static class Impl4903 implements DemoService {public int type() {return 4903;} public void run() {System.out.println(4903);}}
    public static class Impl4904 implements DemoService {public int type() {return 4904;} public void run() {System.out.println(4904);}}
    public static class Impl4905 implements DemoService {public int type() {return 4905;} public void run() {System.out.println(4905);}}
    public static class Impl4906 implements DemoService {public int type() {return 4906;} public void run() {System.out.println(4906);}}
    public static class Impl4907 implements DemoService {public int type() {return 4907;} public void run() {System.out.println(4907);}}
    public static class Impl4908 implements DemoService {public int type() {return 4908;} public void run() {System.out.println(4908);}}
    public static class Impl4909 implements DemoService {public int type() {return 4909;} public void run() {System.out.println(4909);}}
    public static class Impl4910 implements DemoService {public int type() {return 4910;} public void run() {System.out.println(4910);}}
    public static class Impl4911 implements DemoService {public int type() {return 4911;} public void run() {System.out.println(4911);}}
    public static class Impl4912 implements DemoService {public int type() {return 4912;} public void run() {System.out.println(4912);}}
    public static class Impl4913 implements DemoService {public int type() {return 4913;} public void run() {System.out.println(4913);}}
    public static class Impl4914 implements DemoService {public int type() {return 4914;} public void run() {System.out.println(4914);}}
    public static class Impl4915 implements DemoService {public int type() {return 4915;} public void run() {System.out.println(4915);}}
    public static class Impl4916 implements DemoService {public int type() {return 4916;} public void run() {System.out.println(4916);}}
    public static class Impl4917 implements DemoService {public int type() {return 4917;} public void run() {System.out.println(4917);}}
    public static class Impl4918 implements DemoService {public int type() {return 4918;} public void run() {System.out.println(4918);}}
    public static class Impl4919 implements DemoService {public int type() {return 4919;} public void run() {System.out.println(4919);}}
    public static class Impl4920 implements DemoService {public int type() {return 4920;} public void run() {System.out.println(4920);}}
    public static class Impl4921 implements DemoService {public int type() {return 4921;} public void run() {System.out.println(4921);}}
    public static class Impl4922 implements DemoService {public int type() {return 4922;} public void run() {System.out.println(4922);}}
    public static class Impl4923 implements DemoService {public int type() {return 4923;} public void run() {System.out.println(4923);}}
    public static class Impl4924 implements DemoService {public int type() {return 4924;} public void run() {System.out.println(4924);}}
    public static class Impl4925 implements DemoService {public int type() {return 4925;} public void run() {System.out.println(4925);}}
    public static class Impl4926 implements DemoService {public int type() {return 4926;} public void run() {System.out.println(4926);}}
    public static class Impl4927 implements DemoService {public int type() {return 4927;} public void run() {System.out.println(4927);}}
    public static class Impl4928 implements DemoService {public int type() {return 4928;} public void run() {System.out.println(4928);}}
    public static class Impl4929 implements DemoService {public int type() {return 4929;} public void run() {System.out.println(4929);}}
    public static class Impl4930 implements DemoService {public int type() {return 4930;} public void run() {System.out.println(4930);}}
    public static class Impl4931 implements DemoService {public int type() {return 4931;} public void run() {System.out.println(4931);}}
    public static class Impl4932 implements DemoService {public int type() {return 4932;} public void run() {System.out.println(4932);}}
    public static class Impl4933 implements DemoService {public int type() {return 4933;} public void run() {System.out.println(4933);}}
    public static class Impl4934 implements DemoService {public int type() {return 4934;} public void run() {System.out.println(4934);}}
    public static class Impl4935 implements DemoService {public int type() {return 4935;} public void run() {System.out.println(4935);}}
    public static class Impl4936 implements DemoService {public int type() {return 4936;} public void run() {System.out.println(4936);}}
    public static class Impl4937 implements DemoService {public int type() {return 4937;} public void run() {System.out.println(4937);}}
    public static class Impl4938 implements DemoService {public int type() {return 4938;} public void run() {System.out.println(4938);}}
    public static class Impl4939 implements DemoService {public int type() {return 4939;} public void run() {System.out.println(4939);}}
    public static class Impl4940 implements DemoService {public int type() {return 4940;} public void run() {System.out.println(4940);}}
    public static class Impl4941 implements DemoService {public int type() {return 4941;} public void run() {System.out.println(4941);}}
    public static class Impl4942 implements DemoService {public int type() {return 4942;} public void run() {System.out.println(4942);}}
    public static class Impl4943 implements DemoService {public int type() {return 4943;} public void run() {System.out.println(4943);}}
    public static class Impl4944 implements DemoService {public int type() {return 4944;} public void run() {System.out.println(4944);}}
    public static class Impl4945 implements DemoService {public int type() {return 4945;} public void run() {System.out.println(4945);}}
    public static class Impl4946 implements DemoService {public int type() {return 4946;} public void run() {System.out.println(4946);}}
    public static class Impl4947 implements DemoService {public int type() {return 4947;} public void run() {System.out.println(4947);}}
    public static class Impl4948 implements DemoService {public int type() {return 4948;} public void run() {System.out.println(4948);}}
    public static class Impl4949 implements DemoService {public int type() {return 4949;} public void run() {System.out.println(4949);}}
    public static class Impl4950 implements DemoService {public int type() {return 4950;} public void run() {System.out.println(4950);}}
    public static class Impl4951 implements DemoService {public int type() {return 4951;} public void run() {System.out.println(4951);}}
    public static class Impl4952 implements DemoService {public int type() {return 4952;} public void run() {System.out.println(4952);}}
    public static class Impl4953 implements DemoService {public int type() {return 4953;} public void run() {System.out.println(4953);}}
    public static class Impl4954 implements DemoService {public int type() {return 4954;} public void run() {System.out.println(4954);}}
    public static class Impl4955 implements DemoService {public int type() {return 4955;} public void run() {System.out.println(4955);}}
    public static class Impl4956 implements DemoService {public int type() {return 4956;} public void run() {System.out.println(4956);}}
    public static class Impl4957 implements DemoService {public int type() {return 4957;} public void run() {System.out.println(4957);}}
    public static class Impl4958 implements DemoService {public int type() {return 4958;} public void run() {System.out.println(4958);}}
    public static class Impl4959 implements DemoService {public int type() {return 4959;} public void run() {System.out.println(4959);}}
    public static class Impl4960 implements DemoService {public int type() {return 4960;} public void run() {System.out.println(4960);}}
    public static class Impl4961 implements DemoService {public int type() {return 4961;} public void run() {System.out.println(4961);}}
    public static class Impl4962 implements DemoService {public int type() {return 4962;} public void run() {System.out.println(4962);}}
    public static class Impl4963 implements DemoService {public int type() {return 4963;} public void run() {System.out.println(4963);}}
    public static class Impl4964 implements DemoService {public int type() {return 4964;} public void run() {System.out.println(4964);}}
    public static class Impl4965 implements DemoService {public int type() {return 4965;} public void run() {System.out.println(4965);}}
    public static class Impl4966 implements DemoService {public int type() {return 4966;} public void run() {System.out.println(4966);}}
    public static class Impl4967 implements DemoService {public int type() {return 4967;} public void run() {System.out.println(4967);}}
    public static class Impl4968 implements DemoService {public int type() {return 4968;} public void run() {System.out.println(4968);}}
    public static class Impl4969 implements DemoService {public int type() {return 4969;} public void run() {System.out.println(4969);}}
    public static class Impl4970 implements DemoService {public int type() {return 4970;} public void run() {System.out.println(4970);}}
    public static class Impl4971 implements DemoService {public int type() {return 4971;} public void run() {System.out.println(4971);}}
    public static class Impl4972 implements DemoService {public int type() {return 4972;} public void run() {System.out.println(4972);}}
    public static class Impl4973 implements DemoService {public int type() {return 4973;} public void run() {System.out.println(4973);}}
    public static class Impl4974 implements DemoService {public int type() {return 4974;} public void run() {System.out.println(4974);}}
    public static class Impl4975 implements DemoService {public int type() {return 4975;} public void run() {System.out.println(4975);}}
    public static class Impl4976 implements DemoService {public int type() {return 4976;} public void run() {System.out.println(4976);}}
    public static class Impl4977 implements DemoService {public int type() {return 4977;} public void run() {System.out.println(4977);}}
    public static class Impl4978 implements DemoService {public int type() {return 4978;} public void run() {System.out.println(4978);}}
    public static class Impl4979 implements DemoService {public int type() {return 4979;} public void run() {System.out.println(4979);}}
    public static class Impl4980 implements DemoService {public int type() {return 4980;} public void run() {System.out.println(4980);}}
    public static class Impl4981 implements DemoService {public int type() {return 4981;} public void run() {System.out.println(4981);}}
    public static class Impl4982 implements DemoService {public int type() {return 4982;} public void run() {System.out.println(4982);}}
    public static class Impl4983 implements DemoService {public int type() {return 4983;} public void run() {System.out.println(4983);}}
    public static class Impl4984 implements DemoService {public int type() {return 4984;} public void run() {System.out.println(4984);}}
    public static class Impl4985 implements DemoService {public int type() {return 4985;} public void run() {System.out.println(4985);}}
    public static class Impl4986 implements DemoService {public int type() {return 4986;} public void run() {System.out.println(4986);}}
    public static class Impl4987 implements DemoService {public int type() {return 4987;} public void run() {System.out.println(4987);}}
    public static class Impl4988 implements DemoService {public int type() {return 4988;} public void run() {System.out.println(4988);}}
    public static class Impl4989 implements DemoService {public int type() {return 4989;} public void run() {System.out.println(4989);}}
    public static class Impl4990 implements DemoService {public int type() {return 4990;} public void run() {System.out.println(4990);}}
    public static class Impl4991 implements DemoService {public int type() {return 4991;} public void run() {System.out.println(4991);}}
    public static class Impl4992 implements DemoService {public int type() {return 4992;} public void run() {System.out.println(4992);}}
    public static class Impl4993 implements DemoService {public int type() {return 4993;} public void run() {System.out.println(4993);}}
    public static class Impl4994 implements DemoService {public int type() {return 4994;} public void run() {System.out.println(4994);}}
    public static class Impl4995 implements DemoService {public int type() {return 4995;} public void run() {System.out.println(4995);}}
    public static class Impl4996 implements DemoService {public int type() {return 4996;} public void run() {System.out.println(4996);}}
    public static class Impl4997 implements DemoService {public int type() {return 4997;} public void run() {System.out.println(4997);}}
    public static class Impl4998 implements DemoService {public int type() {return 4998;} public void run() {System.out.println(4998);}}
    public static class Impl4999 implements DemoService {public int type() {return 4999;} public void run() {System.out.println(4999);}}
    public static class Impl5000 implements DemoService {public int type() {return 5000;} public void run() {System.out.println(5000);}}
    public static class Impl5001 implements DemoService {public int type() {return 5001;} public void run() {System.out.println(5001);}}
    public static class Impl5002 implements DemoService {public int type() {return 5002;} public void run() {System.out.println(5002);}}
    public static class Impl5003 implements DemoService {public int type() {return 5003;} public void run() {System.out.println(5003);}}
    public static class Impl5004 implements DemoService {public int type() {return 5004;} public void run() {System.out.println(5004);}}
    public static class Impl5005 implements DemoService {public int type() {return 5005;} public void run() {System.out.println(5005);}}
    public static class Impl5006 implements DemoService {public int type() {return 5006;} public void run() {System.out.println(5006);}}
    public static class Impl5007 implements DemoService {public int type() {return 5007;} public void run() {System.out.println(5007);}}
    public static class Impl5008 implements DemoService {public int type() {return 5008;} public void run() {System.out.println(5008);}}
    public static class Impl5009 implements DemoService {public int type() {return 5009;} public void run() {System.out.println(5009);}}
    public static class Impl5010 implements DemoService {public int type() {return 5010;} public void run() {System.out.println(5010);}}
    public static class Impl5011 implements DemoService {public int type() {return 5011;} public void run() {System.out.println(5011);}}
    public static class Impl5012 implements DemoService {public int type() {return 5012;} public void run() {System.out.println(5012);}}
    public static class Impl5013 implements DemoService {public int type() {return 5013;} public void run() {System.out.println(5013);}}
    public static class Impl5014 implements DemoService {public int type() {return 5014;} public void run() {System.out.println(5014);}}
    public static class Impl5015 implements DemoService {public int type() {return 5015;} public void run() {System.out.println(5015);}}
    public static class Impl5016 implements DemoService {public int type() {return 5016;} public void run() {System.out.println(5016);}}
    public static class Impl5017 implements DemoService {public int type() {return 5017;} public void run() {System.out.println(5017);}}
    public static class Impl5018 implements DemoService {public int type() {return 5018;} public void run() {System.out.println(5018);}}
    public static class Impl5019 implements DemoService {public int type() {return 5019;} public void run() {System.out.println(5019);}}
    public static class Impl5020 implements DemoService {public int type() {return 5020;} public void run() {System.out.println(5020);}}
    public static class Impl5021 implements DemoService {public int type() {return 5021;} public void run() {System.out.println(5021);}}
    public static class Impl5022 implements DemoService {public int type() {return 5022;} public void run() {System.out.println(5022);}}
    public static class Impl5023 implements DemoService {public int type() {return 5023;} public void run() {System.out.println(5023);}}
    public static class Impl5024 implements DemoService {public int type() {return 5024;} public void run() {System.out.println(5024);}}
    public static class Impl5025 implements DemoService {public int type() {return 5025;} public void run() {System.out.println(5025);}}
    public static class Impl5026 implements DemoService {public int type() {return 5026;} public void run() {System.out.println(5026);}}
    public static class Impl5027 implements DemoService {public int type() {return 5027;} public void run() {System.out.println(5027);}}
    public static class Impl5028 implements DemoService {public int type() {return 5028;} public void run() {System.out.println(5028);}}
    public static class Impl5029 implements DemoService {public int type() {return 5029;} public void run() {System.out.println(5029);}}
    public static class Impl5030 implements DemoService {public int type() {return 5030;} public void run() {System.out.println(5030);}}
    public static class Impl5031 implements DemoService {public int type() {return 5031;} public void run() {System.out.println(5031);}}
    public static class Impl5032 implements DemoService {public int type() {return 5032;} public void run() {System.out.println(5032);}}
    public static class Impl5033 implements DemoService {public int type() {return 5033;} public void run() {System.out.println(5033);}}
    public static class Impl5034 implements DemoService {public int type() {return 5034;} public void run() {System.out.println(5034);}}
    public static class Impl5035 implements DemoService {public int type() {return 5035;} public void run() {System.out.println(5035);}}
    public static class Impl5036 implements DemoService {public int type() {return 5036;} public void run() {System.out.println(5036);}}
    public static class Impl5037 implements DemoService {public int type() {return 5037;} public void run() {System.out.println(5037);}}
    public static class Impl5038 implements DemoService {public int type() {return 5038;} public void run() {System.out.println(5038);}}
    public static class Impl5039 implements DemoService {public int type() {return 5039;} public void run() {System.out.println(5039);}}
    public static class Impl5040 implements DemoService {public int type() {return 5040;} public void run() {System.out.println(5040);}}
    public static class Impl5041 implements DemoService {public int type() {return 5041;} public void run() {System.out.println(5041);}}
    public static class Impl5042 implements DemoService {public int type() {return 5042;} public void run() {System.out.println(5042);}}
    public static class Impl5043 implements DemoService {public int type() {return 5043;} public void run() {System.out.println(5043);}}
    public static class Impl5044 implements DemoService {public int type() {return 5044;} public void run() {System.out.println(5044);}}
    public static class Impl5045 implements DemoService {public int type() {return 5045;} public void run() {System.out.println(5045);}}
    public static class Impl5046 implements DemoService {public int type() {return 5046;} public void run() {System.out.println(5046);}}
    public static class Impl5047 implements DemoService {public int type() {return 5047;} public void run() {System.out.println(5047);}}
    public static class Impl5048 implements DemoService {public int type() {return 5048;} public void run() {System.out.println(5048);}}
    public static class Impl5049 implements DemoService {public int type() {return 5049;} public void run() {System.out.println(5049);}}
    public static class Impl5050 implements DemoService {public int type() {return 5050;} public void run() {System.out.println(5050);}}
    public static class Impl5051 implements DemoService {public int type() {return 5051;} public void run() {System.out.println(5051);}}
    public static class Impl5052 implements DemoService {public int type() {return 5052;} public void run() {System.out.println(5052);}}
    public static class Impl5053 implements DemoService {public int type() {return 5053;} public void run() {System.out.println(5053);}}
    public static class Impl5054 implements DemoService {public int type() {return 5054;} public void run() {System.out.println(5054);}}
    public static class Impl5055 implements DemoService {public int type() {return 5055;} public void run() {System.out.println(5055);}}
    public static class Impl5056 implements DemoService {public int type() {return 5056;} public void run() {System.out.println(5056);}}
    public static class Impl5057 implements DemoService {public int type() {return 5057;} public void run() {System.out.println(5057);}}
    public static class Impl5058 implements DemoService {public int type() {return 5058;} public void run() {System.out.println(5058);}}
    public static class Impl5059 implements DemoService {public int type() {return 5059;} public void run() {System.out.println(5059);}}
    public static class Impl5060 implements DemoService {public int type() {return 5060;} public void run() {System.out.println(5060);}}
    public static class Impl5061 implements DemoService {public int type() {return 5061;} public void run() {System.out.println(5061);}}
    public static class Impl5062 implements DemoService {public int type() {return 5062;} public void run() {System.out.println(5062);}}
    public static class Impl5063 implements DemoService {public int type() {return 5063;} public void run() {System.out.println(5063);}}
    public static class Impl5064 implements DemoService {public int type() {return 5064;} public void run() {System.out.println(5064);}}
    public static class Impl5065 implements DemoService {public int type() {return 5065;} public void run() {System.out.println(5065);}}
    public static class Impl5066 implements DemoService {public int type() {return 5066;} public void run() {System.out.println(5066);}}
    public static class Impl5067 implements DemoService {public int type() {return 5067;} public void run() {System.out.println(5067);}}
    public static class Impl5068 implements DemoService {public int type() {return 5068;} public void run() {System.out.println(5068);}}
    public static class Impl5069 implements DemoService {public int type() {return 5069;} public void run() {System.out.println(5069);}}
    public static class Impl5070 implements DemoService {public int type() {return 5070;} public void run() {System.out.println(5070);}}
    public static class Impl5071 implements DemoService {public int type() {return 5071;} public void run() {System.out.println(5071);}}
    public static class Impl5072 implements DemoService {public int type() {return 5072;} public void run() {System.out.println(5072);}}
    public static class Impl5073 implements DemoService {public int type() {return 5073;} public void run() {System.out.println(5073);}}
    public static class Impl5074 implements DemoService {public int type() {return 5074;} public void run() {System.out.println(5074);}}
    public static class Impl5075 implements DemoService {public int type() {return 5075;} public void run() {System.out.println(5075);}}
    public static class Impl5076 implements DemoService {public int type() {return 5076;} public void run() {System.out.println(5076);}}
    public static class Impl5077 implements DemoService {public int type() {return 5077;} public void run() {System.out.println(5077);}}
    public static class Impl5078 implements DemoService {public int type() {return 5078;} public void run() {System.out.println(5078);}}
    public static class Impl5079 implements DemoService {public int type() {return 5079;} public void run() {System.out.println(5079);}}
    public static class Impl5080 implements DemoService {public int type() {return 5080;} public void run() {System.out.println(5080);}}
    public static class Impl5081 implements DemoService {public int type() {return 5081;} public void run() {System.out.println(5081);}}
    public static class Impl5082 implements DemoService {public int type() {return 5082;} public void run() {System.out.println(5082);}}
    public static class Impl5083 implements DemoService {public int type() {return 5083;} public void run() {System.out.println(5083);}}
    public static class Impl5084 implements DemoService {public int type() {return 5084;} public void run() {System.out.println(5084);}}
    public static class Impl5085 implements DemoService {public int type() {return 5085;} public void run() {System.out.println(5085);}}
    public static class Impl5086 implements DemoService {public int type() {return 5086;} public void run() {System.out.println(5086);}}
    public static class Impl5087 implements DemoService {public int type() {return 5087;} public void run() {System.out.println(5087);}}
    public static class Impl5088 implements DemoService {public int type() {return 5088;} public void run() {System.out.println(5088);}}
    public static class Impl5089 implements DemoService {public int type() {return 5089;} public void run() {System.out.println(5089);}}
    public static class Impl5090 implements DemoService {public int type() {return 5090;} public void run() {System.out.println(5090);}}
    public static class Impl5091 implements DemoService {public int type() {return 5091;} public void run() {System.out.println(5091);}}
    public static class Impl5092 implements DemoService {public int type() {return 5092;} public void run() {System.out.println(5092);}}
    public static class Impl5093 implements DemoService {public int type() {return 5093;} public void run() {System.out.println(5093);}}
    public static class Impl5094 implements DemoService {public int type() {return 5094;} public void run() {System.out.println(5094);}}
    public static class Impl5095 implements DemoService {public int type() {return 5095;} public void run() {System.out.println(5095);}}
    public static class Impl5096 implements DemoService {public int type() {return 5096;} public void run() {System.out.println(5096);}}
    public static class Impl5097 implements DemoService {public int type() {return 5097;} public void run() {System.out.println(5097);}}
    public static class Impl5098 implements DemoService {public int type() {return 5098;} public void run() {System.out.println(5098);}}
    public static class Impl5099 implements DemoService {public int type() {return 5099;} public void run() {System.out.println(5099);}}
    public static class Impl5100 implements DemoService {public int type() {return 5100;} public void run() {System.out.println(5100);}}
    public static class Impl5101 implements DemoService {public int type() {return 5101;} public void run() {System.out.println(5101);}}
    public static class Impl5102 implements DemoService {public int type() {return 5102;} public void run() {System.out.println(5102);}}
    public static class Impl5103 implements DemoService {public int type() {return 5103;} public void run() {System.out.println(5103);}}
    public static class Impl5104 implements DemoService {public int type() {return 5104;} public void run() {System.out.println(5104);}}
    public static class Impl5105 implements DemoService {public int type() {return 5105;} public void run() {System.out.println(5105);}}
    public static class Impl5106 implements DemoService {public int type() {return 5106;} public void run() {System.out.println(5106);}}
    public static class Impl5107 implements DemoService {public int type() {return 5107;} public void run() {System.out.println(5107);}}
    public static class Impl5108 implements DemoService {public int type() {return 5108;} public void run() {System.out.println(5108);}}
    public static class Impl5109 implements DemoService {public int type() {return 5109;} public void run() {System.out.println(5109);}}
    public static class Impl5110 implements DemoService {public int type() {return 5110;} public void run() {System.out.println(5110);}}
    public static class Impl5111 implements DemoService {public int type() {return 5111;} public void run() {System.out.println(5111);}}
    public static class Impl5112 implements DemoService {public int type() {return 5112;} public void run() {System.out.println(5112);}}
    public static class Impl5113 implements DemoService {public int type() {return 5113;} public void run() {System.out.println(5113);}}
    public static class Impl5114 implements DemoService {public int type() {return 5114;} public void run() {System.out.println(5114);}}
    public static class Impl5115 implements DemoService {public int type() {return 5115;} public void run() {System.out.println(5115);}}
    public static class Impl5116 implements DemoService {public int type() {return 5116;} public void run() {System.out.println(5116);}}
    public static class Impl5117 implements DemoService {public int type() {return 5117;} public void run() {System.out.println(5117);}}
    public static class Impl5118 implements DemoService {public int type() {return 5118;} public void run() {System.out.println(5118);}}
    public static class Impl5119 implements DemoService {public int type() {return 5119;} public void run() {System.out.println(5119);}}
    public static class Impl5120 implements DemoService {public int type() {return 5120;} public void run() {System.out.println(5120);}}
    public static class Impl5121 implements DemoService {public int type() {return 5121;} public void run() {System.out.println(5121);}}
    public static class Impl5122 implements DemoService {public int type() {return 5122;} public void run() {System.out.println(5122);}}
    public static class Impl5123 implements DemoService {public int type() {return 5123;} public void run() {System.out.println(5123);}}
    public static class Impl5124 implements DemoService {public int type() {return 5124;} public void run() {System.out.println(5124);}}
    public static class Impl5125 implements DemoService {public int type() {return 5125;} public void run() {System.out.println(5125);}}
    public static class Impl5126 implements DemoService {public int type() {return 5126;} public void run() {System.out.println(5126);}}
    public static class Impl5127 implements DemoService {public int type() {return 5127;} public void run() {System.out.println(5127);}}
    public static class Impl5128 implements DemoService {public int type() {return 5128;} public void run() {System.out.println(5128);}}
    public static class Impl5129 implements DemoService {public int type() {return 5129;} public void run() {System.out.println(5129);}}
    public static class Impl5130 implements DemoService {public int type() {return 5130;} public void run() {System.out.println(5130);}}
    public static class Impl5131 implements DemoService {public int type() {return 5131;} public void run() {System.out.println(5131);}}
    public static class Impl5132 implements DemoService {public int type() {return 5132;} public void run() {System.out.println(5132);}}
    public static class Impl5133 implements DemoService {public int type() {return 5133;} public void run() {System.out.println(5133);}}
    public static class Impl5134 implements DemoService {public int type() {return 5134;} public void run() {System.out.println(5134);}}
    public static class Impl5135 implements DemoService {public int type() {return 5135;} public void run() {System.out.println(5135);}}
    public static class Impl5136 implements DemoService {public int type() {return 5136;} public void run() {System.out.println(5136);}}
    public static class Impl5137 implements DemoService {public int type() {return 5137;} public void run() {System.out.println(5137);}}
    public static class Impl5138 implements DemoService {public int type() {return 5138;} public void run() {System.out.println(5138);}}
    public static class Impl5139 implements DemoService {public int type() {return 5139;} public void run() {System.out.println(5139);}}
    public static class Impl5140 implements DemoService {public int type() {return 5140;} public void run() {System.out.println(5140);}}
    public static class Impl5141 implements DemoService {public int type() {return 5141;} public void run() {System.out.println(5141);}}
    public static class Impl5142 implements DemoService {public int type() {return 5142;} public void run() {System.out.println(5142);}}
    public static class Impl5143 implements DemoService {public int type() {return 5143;} public void run() {System.out.println(5143);}}
    public static class Impl5144 implements DemoService {public int type() {return 5144;} public void run() {System.out.println(5144);}}
    public static class Impl5145 implements DemoService {public int type() {return 5145;} public void run() {System.out.println(5145);}}
    public static class Impl5146 implements DemoService {public int type() {return 5146;} public void run() {System.out.println(5146);}}
    public static class Impl5147 implements DemoService {public int type() {return 5147;} public void run() {System.out.println(5147);}}
    public static class Impl5148 implements DemoService {public int type() {return 5148;} public void run() {System.out.println(5148);}}
    public static class Impl5149 implements DemoService {public int type() {return 5149;} public void run() {System.out.println(5149);}}
    public static class Impl5150 implements DemoService {public int type() {return 5150;} public void run() {System.out.println(5150);}}
    public static class Impl5151 implements DemoService {public int type() {return 5151;} public void run() {System.out.println(5151);}}
    public static class Impl5152 implements DemoService {public int type() {return 5152;} public void run() {System.out.println(5152);}}
    public static class Impl5153 implements DemoService {public int type() {return 5153;} public void run() {System.out.println(5153);}}
    public static class Impl5154 implements DemoService {public int type() {return 5154;} public void run() {System.out.println(5154);}}
    public static class Impl5155 implements DemoService {public int type() {return 5155;} public void run() {System.out.println(5155);}}
    public static class Impl5156 implements DemoService {public int type() {return 5156;} public void run() {System.out.println(5156);}}
    public static class Impl5157 implements DemoService {public int type() {return 5157;} public void run() {System.out.println(5157);}}
    public static class Impl5158 implements DemoService {public int type() {return 5158;} public void run() {System.out.println(5158);}}
    public static class Impl5159 implements DemoService {public int type() {return 5159;} public void run() {System.out.println(5159);}}
    public static class Impl5160 implements DemoService {public int type() {return 5160;} public void run() {System.out.println(5160);}}
    public static class Impl5161 implements DemoService {public int type() {return 5161;} public void run() {System.out.println(5161);}}
    public static class Impl5162 implements DemoService {public int type() {return 5162;} public void run() {System.out.println(5162);}}
    public static class Impl5163 implements DemoService {public int type() {return 5163;} public void run() {System.out.println(5163);}}
    public static class Impl5164 implements DemoService {public int type() {return 5164;} public void run() {System.out.println(5164);}}
    public static class Impl5165 implements DemoService {public int type() {return 5165;} public void run() {System.out.println(5165);}}
    public static class Impl5166 implements DemoService {public int type() {return 5166;} public void run() {System.out.println(5166);}}
    public static class Impl5167 implements DemoService {public int type() {return 5167;} public void run() {System.out.println(5167);}}
    public static class Impl5168 implements DemoService {public int type() {return 5168;} public void run() {System.out.println(5168);}}
    public static class Impl5169 implements DemoService {public int type() {return 5169;} public void run() {System.out.println(5169);}}
    public static class Impl5170 implements DemoService {public int type() {return 5170;} public void run() {System.out.println(5170);}}
    public static class Impl5171 implements DemoService {public int type() {return 5171;} public void run() {System.out.println(5171);}}
    public static class Impl5172 implements DemoService {public int type() {return 5172;} public void run() {System.out.println(5172);}}
    public static class Impl5173 implements DemoService {public int type() {return 5173;} public void run() {System.out.println(5173);}}
    public static class Impl5174 implements DemoService {public int type() {return 5174;} public void run() {System.out.println(5174);}}
    public static class Impl5175 implements DemoService {public int type() {return 5175;} public void run() {System.out.println(5175);}}
    public static class Impl5176 implements DemoService {public int type() {return 5176;} public void run() {System.out.println(5176);}}
    public static class Impl5177 implements DemoService {public int type() {return 5177;} public void run() {System.out.println(5177);}}
    public static class Impl5178 implements DemoService {public int type() {return 5178;} public void run() {System.out.println(5178);}}
    public static class Impl5179 implements DemoService {public int type() {return 5179;} public void run() {System.out.println(5179);}}
    public static class Impl5180 implements DemoService {public int type() {return 5180;} public void run() {System.out.println(5180);}}
    public static class Impl5181 implements DemoService {public int type() {return 5181;} public void run() {System.out.println(5181);}}
    public static class Impl5182 implements DemoService {public int type() {return 5182;} public void run() {System.out.println(5182);}}
    public static class Impl5183 implements DemoService {public int type() {return 5183;} public void run() {System.out.println(5183);}}
    public static class Impl5184 implements DemoService {public int type() {return 5184;} public void run() {System.out.println(5184);}}
    public static class Impl5185 implements DemoService {public int type() {return 5185;} public void run() {System.out.println(5185);}}
    public static class Impl5186 implements DemoService {public int type() {return 5186;} public void run() {System.out.println(5186);}}
    public static class Impl5187 implements DemoService {public int type() {return 5187;} public void run() {System.out.println(5187);}}
    public static class Impl5188 implements DemoService {public int type() {return 5188;} public void run() {System.out.println(5188);}}
    public static class Impl5189 implements DemoService {public int type() {return 5189;} public void run() {System.out.println(5189);}}
    public static class Impl5190 implements DemoService {public int type() {return 5190;} public void run() {System.out.println(5190);}}
    public static class Impl5191 implements DemoService {public int type() {return 5191;} public void run() {System.out.println(5191);}}
    public static class Impl5192 implements DemoService {public int type() {return 5192;} public void run() {System.out.println(5192);}}
    public static class Impl5193 implements DemoService {public int type() {return 5193;} public void run() {System.out.println(5193);}}
    public static class Impl5194 implements DemoService {public int type() {return 5194;} public void run() {System.out.println(5194);}}
    public static class Impl5195 implements DemoService {public int type() {return 5195;} public void run() {System.out.println(5195);}}
    public static class Impl5196 implements DemoService {public int type() {return 5196;} public void run() {System.out.println(5196);}}
    public static class Impl5197 implements DemoService {public int type() {return 5197;} public void run() {System.out.println(5197);}}
    public static class Impl5198 implements DemoService {public int type() {return 5198;} public void run() {System.out.println(5198);}}
    public static class Impl5199 implements DemoService {public int type() {return 5199;} public void run() {System.out.println(5199);}}
    public static class Impl5200 implements DemoService {public int type() {return 5200;} public void run() {System.out.println(5200);}}
    public static class Impl5201 implements DemoService {public int type() {return 5201;} public void run() {System.out.println(5201);}}
    public static class Impl5202 implements DemoService {public int type() {return 5202;} public void run() {System.out.println(5202);}}
    public static class Impl5203 implements DemoService {public int type() {return 5203;} public void run() {System.out.println(5203);}}
    public static class Impl5204 implements DemoService {public int type() {return 5204;} public void run() {System.out.println(5204);}}
    public static class Impl5205 implements DemoService {public int type() {return 5205;} public void run() {System.out.println(5205);}}
    public static class Impl5206 implements DemoService {public int type() {return 5206;} public void run() {System.out.println(5206);}}
    public static class Impl5207 implements DemoService {public int type() {return 5207;} public void run() {System.out.println(5207);}}
    public static class Impl5208 implements DemoService {public int type() {return 5208;} public void run() {System.out.println(5208);}}
    public static class Impl5209 implements DemoService {public int type() {return 5209;} public void run() {System.out.println(5209);}}
    public static class Impl5210 implements DemoService {public int type() {return 5210;} public void run() {System.out.println(5210);}}
    public static class Impl5211 implements DemoService {public int type() {return 5211;} public void run() {System.out.println(5211);}}
    public static class Impl5212 implements DemoService {public int type() {return 5212;} public void run() {System.out.println(5212);}}
    public static class Impl5213 implements DemoService {public int type() {return 5213;} public void run() {System.out.println(5213);}}
    public static class Impl5214 implements DemoService {public int type() {return 5214;} public void run() {System.out.println(5214);}}
    public static class Impl5215 implements DemoService {public int type() {return 5215;} public void run() {System.out.println(5215);}}
    public static class Impl5216 implements DemoService {public int type() {return 5216;} public void run() {System.out.println(5216);}}
    public static class Impl5217 implements DemoService {public int type() {return 5217;} public void run() {System.out.println(5217);}}
    public static class Impl5218 implements DemoService {public int type() {return 5218;} public void run() {System.out.println(5218);}}
    public static class Impl5219 implements DemoService {public int type() {return 5219;} public void run() {System.out.println(5219);}}
    public static class Impl5220 implements DemoService {public int type() {return 5220;} public void run() {System.out.println(5220);}}
    public static class Impl5221 implements DemoService {public int type() {return 5221;} public void run() {System.out.println(5221);}}
    public static class Impl5222 implements DemoService {public int type() {return 5222;} public void run() {System.out.println(5222);}}
    public static class Impl5223 implements DemoService {public int type() {return 5223;} public void run() {System.out.println(5223);}}
    public static class Impl5224 implements DemoService {public int type() {return 5224;} public void run() {System.out.println(5224);}}
    public static class Impl5225 implements DemoService {public int type() {return 5225;} public void run() {System.out.println(5225);}}
    public static class Impl5226 implements DemoService {public int type() {return 5226;} public void run() {System.out.println(5226);}}
    public static class Impl5227 implements DemoService {public int type() {return 5227;} public void run() {System.out.println(5227);}}
    public static class Impl5228 implements DemoService {public int type() {return 5228;} public void run() {System.out.println(5228);}}
    public static class Impl5229 implements DemoService {public int type() {return 5229;} public void run() {System.out.println(5229);}}
    public static class Impl5230 implements DemoService {public int type() {return 5230;} public void run() {System.out.println(5230);}}
    public static class Impl5231 implements DemoService {public int type() {return 5231;} public void run() {System.out.println(5231);}}
    public static class Impl5232 implements DemoService {public int type() {return 5232;} public void run() {System.out.println(5232);}}
    public static class Impl5233 implements DemoService {public int type() {return 5233;} public void run() {System.out.println(5233);}}
    public static class Impl5234 implements DemoService {public int type() {return 5234;} public void run() {System.out.println(5234);}}
    public static class Impl5235 implements DemoService {public int type() {return 5235;} public void run() {System.out.println(5235);}}
    public static class Impl5236 implements DemoService {public int type() {return 5236;} public void run() {System.out.println(5236);}}
    public static class Impl5237 implements DemoService {public int type() {return 5237;} public void run() {System.out.println(5237);}}
    public static class Impl5238 implements DemoService {public int type() {return 5238;} public void run() {System.out.println(5238);}}
    public static class Impl5239 implements DemoService {public int type() {return 5239;} public void run() {System.out.println(5239);}}
    public static class Impl5240 implements DemoService {public int type() {return 5240;} public void run() {System.out.println(5240);}}
    public static class Impl5241 implements DemoService {public int type() {return 5241;} public void run() {System.out.println(5241);}}
    public static class Impl5242 implements DemoService {public int type() {return 5242;} public void run() {System.out.println(5242);}}
    public static class Impl5243 implements DemoService {public int type() {return 5243;} public void run() {System.out.println(5243);}}
    public static class Impl5244 implements DemoService {public int type() {return 5244;} public void run() {System.out.println(5244);}}
    public static class Impl5245 implements DemoService {public int type() {return 5245;} public void run() {System.out.println(5245);}}
    public static class Impl5246 implements DemoService {public int type() {return 5246;} public void run() {System.out.println(5246);}}
    public static class Impl5247 implements DemoService {public int type() {return 5247;} public void run() {System.out.println(5247);}}
    public static class Impl5248 implements DemoService {public int type() {return 5248;} public void run() {System.out.println(5248);}}
    public static class Impl5249 implements DemoService {public int type() {return 5249;} public void run() {System.out.println(5249);}}
    public static class Impl5250 implements DemoService {public int type() {return 5250;} public void run() {System.out.println(5250);}}
    public static class Impl5251 implements DemoService {public int type() {return 5251;} public void run() {System.out.println(5251);}}
    public static class Impl5252 implements DemoService {public int type() {return 5252;} public void run() {System.out.println(5252);}}
    public static class Impl5253 implements DemoService {public int type() {return 5253;} public void run() {System.out.println(5253);}}
    public static class Impl5254 implements DemoService {public int type() {return 5254;} public void run() {System.out.println(5254);}}
    public static class Impl5255 implements DemoService {public int type() {return 5255;} public void run() {System.out.println(5255);}}
    public static class Impl5256 implements DemoService {public int type() {return 5256;} public void run() {System.out.println(5256);}}
    public static class Impl5257 implements DemoService {public int type() {return 5257;} public void run() {System.out.println(5257);}}
    public static class Impl5258 implements DemoService {public int type() {return 5258;} public void run() {System.out.println(5258);}}
    public static class Impl5259 implements DemoService {public int type() {return 5259;} public void run() {System.out.println(5259);}}
    public static class Impl5260 implements DemoService {public int type() {return 5260;} public void run() {System.out.println(5260);}}
    public static class Impl5261 implements DemoService {public int type() {return 5261;} public void run() {System.out.println(5261);}}
    public static class Impl5262 implements DemoService {public int type() {return 5262;} public void run() {System.out.println(5262);}}
    public static class Impl5263 implements DemoService {public int type() {return 5263;} public void run() {System.out.println(5263);}}
    public static class Impl5264 implements DemoService {public int type() {return 5264;} public void run() {System.out.println(5264);}}
    public static class Impl5265 implements DemoService {public int type() {return 5265;} public void run() {System.out.println(5265);}}
    public static class Impl5266 implements DemoService {public int type() {return 5266;} public void run() {System.out.println(5266);}}
    public static class Impl5267 implements DemoService {public int type() {return 5267;} public void run() {System.out.println(5267);}}
    public static class Impl5268 implements DemoService {public int type() {return 5268;} public void run() {System.out.println(5268);}}
    public static class Impl5269 implements DemoService {public int type() {return 5269;} public void run() {System.out.println(5269);}}
    public static class Impl5270 implements DemoService {public int type() {return 5270;} public void run() {System.out.println(5270);}}
    public static class Impl5271 implements DemoService {public int type() {return 5271;} public void run() {System.out.println(5271);}}
    public static class Impl5272 implements DemoService {public int type() {return 5272;} public void run() {System.out.println(5272);}}
    public static class Impl5273 implements DemoService {public int type() {return 5273;} public void run() {System.out.println(5273);}}
    public static class Impl5274 implements DemoService {public int type() {return 5274;} public void run() {System.out.println(5274);}}
    public static class Impl5275 implements DemoService {public int type() {return 5275;} public void run() {System.out.println(5275);}}
    public static class Impl5276 implements DemoService {public int type() {return 5276;} public void run() {System.out.println(5276);}}
    public static class Impl5277 implements DemoService {public int type() {return 5277;} public void run() {System.out.println(5277);}}
    public static class Impl5278 implements DemoService {public int type() {return 5278;} public void run() {System.out.println(5278);}}
    public static class Impl5279 implements DemoService {public int type() {return 5279;} public void run() {System.out.println(5279);}}
    public static class Impl5280 implements DemoService {public int type() {return 5280;} public void run() {System.out.println(5280);}}
    public static class Impl5281 implements DemoService {public int type() {return 5281;} public void run() {System.out.println(5281);}}
    public static class Impl5282 implements DemoService {public int type() {return 5282;} public void run() {System.out.println(5282);}}
    public static class Impl5283 implements DemoService {public int type() {return 5283;} public void run() {System.out.println(5283);}}
    public static class Impl5284 implements DemoService {public int type() {return 5284;} public void run() {System.out.println(5284);}}
    public static class Impl5285 implements DemoService {public int type() {return 5285;} public void run() {System.out.println(5285);}}
    public static class Impl5286 implements DemoService {public int type() {return 5286;} public void run() {System.out.println(5286);}}
    public static class Impl5287 implements DemoService {public int type() {return 5287;} public void run() {System.out.println(5287);}}
    public static class Impl5288 implements DemoService {public int type() {return 5288;} public void run() {System.out.println(5288);}}
    public static class Impl5289 implements DemoService {public int type() {return 5289;} public void run() {System.out.println(5289);}}
    public static class Impl5290 implements DemoService {public int type() {return 5290;} public void run() {System.out.println(5290);}}
    public static class Impl5291 implements DemoService {public int type() {return 5291;} public void run() {System.out.println(5291);}}
    public static class Impl5292 implements DemoService {public int type() {return 5292;} public void run() {System.out.println(5292);}}
    public static class Impl5293 implements DemoService {public int type() {return 5293;} public void run() {System.out.println(5293);}}
    public static class Impl5294 implements DemoService {public int type() {return 5294;} public void run() {System.out.println(5294);}}
    public static class Impl5295 implements DemoService {public int type() {return 5295;} public void run() {System.out.println(5295);}}
    public static class Impl5296 implements DemoService {public int type() {return 5296;} public void run() {System.out.println(5296);}}
    public static class Impl5297 implements DemoService {public int type() {return 5297;} public void run() {System.out.println(5297);}}
    public static class Impl5298 implements DemoService {public int type() {return 5298;} public void run() {System.out.println(5298);}}
    public static class Impl5299 implements DemoService {public int type() {return 5299;} public void run() {System.out.println(5299);}}
    public static class Impl5300 implements DemoService {public int type() {return 5300;} public void run() {System.out.println(5300);}}
    public static class Impl5301 implements DemoService {public int type() {return 5301;} public void run() {System.out.println(5301);}}
    public static class Impl5302 implements DemoService {public int type() {return 5302;} public void run() {System.out.println(5302);}}
    public static class Impl5303 implements DemoService {public int type() {return 5303;} public void run() {System.out.println(5303);}}
    public static class Impl5304 implements DemoService {public int type() {return 5304;} public void run() {System.out.println(5304);}}
    public static class Impl5305 implements DemoService {public int type() {return 5305;} public void run() {System.out.println(5305);}}
    public static class Impl5306 implements DemoService {public int type() {return 5306;} public void run() {System.out.println(5306);}}
    public static class Impl5307 implements DemoService {public int type() {return 5307;} public void run() {System.out.println(5307);}}
    public static class Impl5308 implements DemoService {public int type() {return 5308;} public void run() {System.out.println(5308);}}
    public static class Impl5309 implements DemoService {public int type() {return 5309;} public void run() {System.out.println(5309);}}
    public static class Impl5310 implements DemoService {public int type() {return 5310;} public void run() {System.out.println(5310);}}
    public static class Impl5311 implements DemoService {public int type() {return 5311;} public void run() {System.out.println(5311);}}
    public static class Impl5312 implements DemoService {public int type() {return 5312;} public void run() {System.out.println(5312);}}
    public static class Impl5313 implements DemoService {public int type() {return 5313;} public void run() {System.out.println(5313);}}
    public static class Impl5314 implements DemoService {public int type() {return 5314;} public void run() {System.out.println(5314);}}
    public static class Impl5315 implements DemoService {public int type() {return 5315;} public void run() {System.out.println(5315);}}
    public static class Impl5316 implements DemoService {public int type() {return 5316;} public void run() {System.out.println(5316);}}
    public static class Impl5317 implements DemoService {public int type() {return 5317;} public void run() {System.out.println(5317);}}
    public static class Impl5318 implements DemoService {public int type() {return 5318;} public void run() {System.out.println(5318);}}
    public static class Impl5319 implements DemoService {public int type() {return 5319;} public void run() {System.out.println(5319);}}
    public static class Impl5320 implements DemoService {public int type() {return 5320;} public void run() {System.out.println(5320);}}
    public static class Impl5321 implements DemoService {public int type() {return 5321;} public void run() {System.out.println(5321);}}
    public static class Impl5322 implements DemoService {public int type() {return 5322;} public void run() {System.out.println(5322);}}
    public static class Impl5323 implements DemoService {public int type() {return 5323;} public void run() {System.out.println(5323);}}
    public static class Impl5324 implements DemoService {public int type() {return 5324;} public void run() {System.out.println(5324);}}
    public static class Impl5325 implements DemoService {public int type() {return 5325;} public void run() {System.out.println(5325);}}
    public static class Impl5326 implements DemoService {public int type() {return 5326;} public void run() {System.out.println(5326);}}
    public static class Impl5327 implements DemoService {public int type() {return 5327;} public void run() {System.out.println(5327);}}
    public static class Impl5328 implements DemoService {public int type() {return 5328;} public void run() {System.out.println(5328);}}
    public static class Impl5329 implements DemoService {public int type() {return 5329;} public void run() {System.out.println(5329);}}
    public static class Impl5330 implements DemoService {public int type() {return 5330;} public void run() {System.out.println(5330);}}
    public static class Impl5331 implements DemoService {public int type() {return 5331;} public void run() {System.out.println(5331);}}
    public static class Impl5332 implements DemoService {public int type() {return 5332;} public void run() {System.out.println(5332);}}
    public static class Impl5333 implements DemoService {public int type() {return 5333;} public void run() {System.out.println(5333);}}
    public static class Impl5334 implements DemoService {public int type() {return 5334;} public void run() {System.out.println(5334);}}
    public static class Impl5335 implements DemoService {public int type() {return 5335;} public void run() {System.out.println(5335);}}
    public static class Impl5336 implements DemoService {public int type() {return 5336;} public void run() {System.out.println(5336);}}
    public static class Impl5337 implements DemoService {public int type() {return 5337;} public void run() {System.out.println(5337);}}
    public static class Impl5338 implements DemoService {public int type() {return 5338;} public void run() {System.out.println(5338);}}
    public static class Impl5339 implements DemoService {public int type() {return 5339;} public void run() {System.out.println(5339);}}
    public static class Impl5340 implements DemoService {public int type() {return 5340;} public void run() {System.out.println(5340);}}
    public static class Impl5341 implements DemoService {public int type() {return 5341;} public void run() {System.out.println(5341);}}
    public static class Impl5342 implements DemoService {public int type() {return 5342;} public void run() {System.out.println(5342);}}
    public static class Impl5343 implements DemoService {public int type() {return 5343;} public void run() {System.out.println(5343);}}
    public static class Impl5344 implements DemoService {public int type() {return 5344;} public void run() {System.out.println(5344);}}
    public static class Impl5345 implements DemoService {public int type() {return 5345;} public void run() {System.out.println(5345);}}
    public static class Impl5346 implements DemoService {public int type() {return 5346;} public void run() {System.out.println(5346);}}
    public static class Impl5347 implements DemoService {public int type() {return 5347;} public void run() {System.out.println(5347);}}
    public static class Impl5348 implements DemoService {public int type() {return 5348;} public void run() {System.out.println(5348);}}
    public static class Impl5349 implements DemoService {public int type() {return 5349;} public void run() {System.out.println(5349);}}
    public static class Impl5350 implements DemoService {public int type() {return 5350;} public void run() {System.out.println(5350);}}
    public static class Impl5351 implements DemoService {public int type() {return 5351;} public void run() {System.out.println(5351);}}
    public static class Impl5352 implements DemoService {public int type() {return 5352;} public void run() {System.out.println(5352);}}
    public static class Impl5353 implements DemoService {public int type() {return 5353;} public void run() {System.out.println(5353);}}
    public static class Impl5354 implements DemoService {public int type() {return 5354;} public void run() {System.out.println(5354);}}
    public static class Impl5355 implements DemoService {public int type() {return 5355;} public void run() {System.out.println(5355);}}
    public static class Impl5356 implements DemoService {public int type() {return 5356;} public void run() {System.out.println(5356);}}
    public static class Impl5357 implements DemoService {public int type() {return 5357;} public void run() {System.out.println(5357);}}
    public static class Impl5358 implements DemoService {public int type() {return 5358;} public void run() {System.out.println(5358);}}
    public static class Impl5359 implements DemoService {public int type() {return 5359;} public void run() {System.out.println(5359);}}
    public static class Impl5360 implements DemoService {public int type() {return 5360;} public void run() {System.out.println(5360);}}
    public static class Impl5361 implements DemoService {public int type() {return 5361;} public void run() {System.out.println(5361);}}
    public static class Impl5362 implements DemoService {public int type() {return 5362;} public void run() {System.out.println(5362);}}
    public static class Impl5363 implements DemoService {public int type() {return 5363;} public void run() {System.out.println(5363);}}
    public static class Impl5364 implements DemoService {public int type() {return 5364;} public void run() {System.out.println(5364);}}
    public static class Impl5365 implements DemoService {public int type() {return 5365;} public void run() {System.out.println(5365);}}
    public static class Impl5366 implements DemoService {public int type() {return 5366;} public void run() {System.out.println(5366);}}
    public static class Impl5367 implements DemoService {public int type() {return 5367;} public void run() {System.out.println(5367);}}
    public static class Impl5368 implements DemoService {public int type() {return 5368;} public void run() {System.out.println(5368);}}
    public static class Impl5369 implements DemoService {public int type() {return 5369;} public void run() {System.out.println(5369);}}
    public static class Impl5370 implements DemoService {public int type() {return 5370;} public void run() {System.out.println(5370);}}
    public static class Impl5371 implements DemoService {public int type() {return 5371;} public void run() {System.out.println(5371);}}
    public static class Impl5372 implements DemoService {public int type() {return 5372;} public void run() {System.out.println(5372);}}
    public static class Impl5373 implements DemoService {public int type() {return 5373;} public void run() {System.out.println(5373);}}
    public static class Impl5374 implements DemoService {public int type() {return 5374;} public void run() {System.out.println(5374);}}
    public static class Impl5375 implements DemoService {public int type() {return 5375;} public void run() {System.out.println(5375);}}
    public static class Impl5376 implements DemoService {public int type() {return 5376;} public void run() {System.out.println(5376);}}
    public static class Impl5377 implements DemoService {public int type() {return 5377;} public void run() {System.out.println(5377);}}
    public static class Impl5378 implements DemoService {public int type() {return 5378;} public void run() {System.out.println(5378);}}
    public static class Impl5379 implements DemoService {public int type() {return 5379;} public void run() {System.out.println(5379);}}
    public static class Impl5380 implements DemoService {public int type() {return 5380;} public void run() {System.out.println(5380);}}
    public static class Impl5381 implements DemoService {public int type() {return 5381;} public void run() {System.out.println(5381);}}
    public static class Impl5382 implements DemoService {public int type() {return 5382;} public void run() {System.out.println(5382);}}
    public static class Impl5383 implements DemoService {public int type() {return 5383;} public void run() {System.out.println(5383);}}
    public static class Impl5384 implements DemoService {public int type() {return 5384;} public void run() {System.out.println(5384);}}
    public static class Impl5385 implements DemoService {public int type() {return 5385;} public void run() {System.out.println(5385);}}
    public static class Impl5386 implements DemoService {public int type() {return 5386;} public void run() {System.out.println(5386);}}
    public static class Impl5387 implements DemoService {public int type() {return 5387;} public void run() {System.out.println(5387);}}
    public static class Impl5388 implements DemoService {public int type() {return 5388;} public void run() {System.out.println(5388);}}
    public static class Impl5389 implements DemoService {public int type() {return 5389;} public void run() {System.out.println(5389);}}
    public static class Impl5390 implements DemoService {public int type() {return 5390;} public void run() {System.out.println(5390);}}
    public static class Impl5391 implements DemoService {public int type() {return 5391;} public void run() {System.out.println(5391);}}
    public static class Impl5392 implements DemoService {public int type() {return 5392;} public void run() {System.out.println(5392);}}
    public static class Impl5393 implements DemoService {public int type() {return 5393;} public void run() {System.out.println(5393);}}
    public static class Impl5394 implements DemoService {public int type() {return 5394;} public void run() {System.out.println(5394);}}
    public static class Impl5395 implements DemoService {public int type() {return 5395;} public void run() {System.out.println(5395);}}
    public static class Impl5396 implements DemoService {public int type() {return 5396;} public void run() {System.out.println(5396);}}
    public static class Impl5397 implements DemoService {public int type() {return 5397;} public void run() {System.out.println(5397);}}
    public static class Impl5398 implements DemoService {public int type() {return 5398;} public void run() {System.out.println(5398);}}
    public static class Impl5399 implements DemoService {public int type() {return 5399;} public void run() {System.out.println(5399);}}
    public static class Impl5400 implements DemoService {public int type() {return 5400;} public void run() {System.out.println(5400);}}
    public static class Impl5401 implements DemoService {public int type() {return 5401;} public void run() {System.out.println(5401);}}
    public static class Impl5402 implements DemoService {public int type() {return 5402;} public void run() {System.out.println(5402);}}
    public static class Impl5403 implements DemoService {public int type() {return 5403;} public void run() {System.out.println(5403);}}
    public static class Impl5404 implements DemoService {public int type() {return 5404;} public void run() {System.out.println(5404);}}
    public static class Impl5405 implements DemoService {public int type() {return 5405;} public void run() {System.out.println(5405);}}
    public static class Impl5406 implements DemoService {public int type() {return 5406;} public void run() {System.out.println(5406);}}
    public static class Impl5407 implements DemoService {public int type() {return 5407;} public void run() {System.out.println(5407);}}
    public static class Impl5408 implements DemoService {public int type() {return 5408;} public void run() {System.out.println(5408);}}
    public static class Impl5409 implements DemoService {public int type() {return 5409;} public void run() {System.out.println(5409);}}
    public static class Impl5410 implements DemoService {public int type() {return 5410;} public void run() {System.out.println(5410);}}
    public static class Impl5411 implements DemoService {public int type() {return 5411;} public void run() {System.out.println(5411);}}
    public static class Impl5412 implements DemoService {public int type() {return 5412;} public void run() {System.out.println(5412);}}
    public static class Impl5413 implements DemoService {public int type() {return 5413;} public void run() {System.out.println(5413);}}
    public static class Impl5414 implements DemoService {public int type() {return 5414;} public void run() {System.out.println(5414);}}
    public static class Impl5415 implements DemoService {public int type() {return 5415;} public void run() {System.out.println(5415);}}
    public static class Impl5416 implements DemoService {public int type() {return 5416;} public void run() {System.out.println(5416);}}
    public static class Impl5417 implements DemoService {public int type() {return 5417;} public void run() {System.out.println(5417);}}
    public static class Impl5418 implements DemoService {public int type() {return 5418;} public void run() {System.out.println(5418);}}
    public static class Impl5419 implements DemoService {public int type() {return 5419;} public void run() {System.out.println(5419);}}
    public static class Impl5420 implements DemoService {public int type() {return 5420;} public void run() {System.out.println(5420);}}
    public static class Impl5421 implements DemoService {public int type() {return 5421;} public void run() {System.out.println(5421);}}
    public static class Impl5422 implements DemoService {public int type() {return 5422;} public void run() {System.out.println(5422);}}
    public static class Impl5423 implements DemoService {public int type() {return 5423;} public void run() {System.out.println(5423);}}
    public static class Impl5424 implements DemoService {public int type() {return 5424;} public void run() {System.out.println(5424);}}
    public static class Impl5425 implements DemoService {public int type() {return 5425;} public void run() {System.out.println(5425);}}
    public static class Impl5426 implements DemoService {public int type() {return 5426;} public void run() {System.out.println(5426);}}
    public static class Impl5427 implements DemoService {public int type() {return 5427;} public void run() {System.out.println(5427);}}
    public static class Impl5428 implements DemoService {public int type() {return 5428;} public void run() {System.out.println(5428);}}
    public static class Impl5429 implements DemoService {public int type() {return 5429;} public void run() {System.out.println(5429);}}
    public static class Impl5430 implements DemoService {public int type() {return 5430;} public void run() {System.out.println(5430);}}
    public static class Impl5431 implements DemoService {public int type() {return 5431;} public void run() {System.out.println(5431);}}
    public static class Impl5432 implements DemoService {public int type() {return 5432;} public void run() {System.out.println(5432);}}
    public static class Impl5433 implements DemoService {public int type() {return 5433;} public void run() {System.out.println(5433);}}
    public static class Impl5434 implements DemoService {public int type() {return 5434;} public void run() {System.out.println(5434);}}
    public static class Impl5435 implements DemoService {public int type() {return 5435;} public void run() {System.out.println(5435);}}
    public static class Impl5436 implements DemoService {public int type() {return 5436;} public void run() {System.out.println(5436);}}
    public static class Impl5437 implements DemoService {public int type() {return 5437;} public void run() {System.out.println(5437);}}
    public static class Impl5438 implements DemoService {public int type() {return 5438;} public void run() {System.out.println(5438);}}
    public static class Impl5439 implements DemoService {public int type() {return 5439;} public void run() {System.out.println(5439);}}
    public static class Impl5440 implements DemoService {public int type() {return 5440;} public void run() {System.out.println(5440);}}
    public static class Impl5441 implements DemoService {public int type() {return 5441;} public void run() {System.out.println(5441);}}
    public static class Impl5442 implements DemoService {public int type() {return 5442;} public void run() {System.out.println(5442);}}
    public static class Impl5443 implements DemoService {public int type() {return 5443;} public void run() {System.out.println(5443);}}
    public static class Impl5444 implements DemoService {public int type() {return 5444;} public void run() {System.out.println(5444);}}
    public static class Impl5445 implements DemoService {public int type() {return 5445;} public void run() {System.out.println(5445);}}
    public static class Impl5446 implements DemoService {public int type() {return 5446;} public void run() {System.out.println(5446);}}
    public static class Impl5447 implements DemoService {public int type() {return 5447;} public void run() {System.out.println(5447);}}
    public static class Impl5448 implements DemoService {public int type() {return 5448;} public void run() {System.out.println(5448);}}
    public static class Impl5449 implements DemoService {public int type() {return 5449;} public void run() {System.out.println(5449);}}
    public static class Impl5450 implements DemoService {public int type() {return 5450;} public void run() {System.out.println(5450);}}
    public static class Impl5451 implements DemoService {public int type() {return 5451;} public void run() {System.out.println(5451);}}
    public static class Impl5452 implements DemoService {public int type() {return 5452;} public void run() {System.out.println(5452);}}
    public static class Impl5453 implements DemoService {public int type() {return 5453;} public void run() {System.out.println(5453);}}
    public static class Impl5454 implements DemoService {public int type() {return 5454;} public void run() {System.out.println(5454);}}
    public static class Impl5455 implements DemoService {public int type() {return 5455;} public void run() {System.out.println(5455);}}
    public static class Impl5456 implements DemoService {public int type() {return 5456;} public void run() {System.out.println(5456);}}
    public static class Impl5457 implements DemoService {public int type() {return 5457;} public void run() {System.out.println(5457);}}
    public static class Impl5458 implements DemoService {public int type() {return 5458;} public void run() {System.out.println(5458);}}
    public static class Impl5459 implements DemoService {public int type() {return 5459;} public void run() {System.out.println(5459);}}
    public static class Impl5460 implements DemoService {public int type() {return 5460;} public void run() {System.out.println(5460);}}
    public static class Impl5461 implements DemoService {public int type() {return 5461;} public void run() {System.out.println(5461);}}
    public static class Impl5462 implements DemoService {public int type() {return 5462;} public void run() {System.out.println(5462);}}
    public static class Impl5463 implements DemoService {public int type() {return 5463;} public void run() {System.out.println(5463);}}
    public static class Impl5464 implements DemoService {public int type() {return 5464;} public void run() {System.out.println(5464);}}
    public static class Impl5465 implements DemoService {public int type() {return 5465;} public void run() {System.out.println(5465);}}
    public static class Impl5466 implements DemoService {public int type() {return 5466;} public void run() {System.out.println(5466);}}
    public static class Impl5467 implements DemoService {public int type() {return 5467;} public void run() {System.out.println(5467);}}
    public static class Impl5468 implements DemoService {public int type() {return 5468;} public void run() {System.out.println(5468);}}
    public static class Impl5469 implements DemoService {public int type() {return 5469;} public void run() {System.out.println(5469);}}
    public static class Impl5470 implements DemoService {public int type() {return 5470;} public void run() {System.out.println(5470);}}
    public static class Impl5471 implements DemoService {public int type() {return 5471;} public void run() {System.out.println(5471);}}
    public static class Impl5472 implements DemoService {public int type() {return 5472;} public void run() {System.out.println(5472);}}
    public static class Impl5473 implements DemoService {public int type() {return 5473;} public void run() {System.out.println(5473);}}
    public static class Impl5474 implements DemoService {public int type() {return 5474;} public void run() {System.out.println(5474);}}
    public static class Impl5475 implements DemoService {public int type() {return 5475;} public void run() {System.out.println(5475);}}
    public static class Impl5476 implements DemoService {public int type() {return 5476;} public void run() {System.out.println(5476);}}
    public static class Impl5477 implements DemoService {public int type() {return 5477;} public void run() {System.out.println(5477);}}
    public static class Impl5478 implements DemoService {public int type() {return 5478;} public void run() {System.out.println(5478);}}
    public static class Impl5479 implements DemoService {public int type() {return 5479;} public void run() {System.out.println(5479);}}
    public static class Impl5480 implements DemoService {public int type() {return 5480;} public void run() {System.out.println(5480);}}
    public static class Impl5481 implements DemoService {public int type() {return 5481;} public void run() {System.out.println(5481);}}
    public static class Impl5482 implements DemoService {public int type() {return 5482;} public void run() {System.out.println(5482);}}
    public static class Impl5483 implements DemoService {public int type() {return 5483;} public void run() {System.out.println(5483);}}
    public static class Impl5484 implements DemoService {public int type() {return 5484;} public void run() {System.out.println(5484);}}
    public static class Impl5485 implements DemoService {public int type() {return 5485;} public void run() {System.out.println(5485);}}
    public static class Impl5486 implements DemoService {public int type() {return 5486;} public void run() {System.out.println(5486);}}
    public static class Impl5487 implements DemoService {public int type() {return 5487;} public void run() {System.out.println(5487);}}
    public static class Impl5488 implements DemoService {public int type() {return 5488;} public void run() {System.out.println(5488);}}
    public static class Impl5489 implements DemoService {public int type() {return 5489;} public void run() {System.out.println(5489);}}
    public static class Impl5490 implements DemoService {public int type() {return 5490;} public void run() {System.out.println(5490);}}
    public static class Impl5491 implements DemoService {public int type() {return 5491;} public void run() {System.out.println(5491);}}
    public static class Impl5492 implements DemoService {public int type() {return 5492;} public void run() {System.out.println(5492);}}
    public static class Impl5493 implements DemoService {public int type() {return 5493;} public void run() {System.out.println(5493);}}
    public static class Impl5494 implements DemoService {public int type() {return 5494;} public void run() {System.out.println(5494);}}
    public static class Impl5495 implements DemoService {public int type() {return 5495;} public void run() {System.out.println(5495);}}
    public static class Impl5496 implements DemoService {public int type() {return 5496;} public void run() {System.out.println(5496);}}
    public static class Impl5497 implements DemoService {public int type() {return 5497;} public void run() {System.out.println(5497);}}
    public static class Impl5498 implements DemoService {public int type() {return 5498;} public void run() {System.out.println(5498);}}
    public static class Impl5499 implements DemoService {public int type() {return 5499;} public void run() {System.out.println(5499);}}
    public static class Impl5500 implements DemoService {public int type() {return 5500;} public void run() {System.out.println(5500);}}
    public static class Impl5501 implements DemoService {public int type() {return 5501;} public void run() {System.out.println(5501);}}
    public static class Impl5502 implements DemoService {public int type() {return 5502;} public void run() {System.out.println(5502);}}
    public static class Impl5503 implements DemoService {public int type() {return 5503;} public void run() {System.out.println(5503);}}
    public static class Impl5504 implements DemoService {public int type() {return 5504;} public void run() {System.out.println(5504);}}
    public static class Impl5505 implements DemoService {public int type() {return 5505;} public void run() {System.out.println(5505);}}
    public static class Impl5506 implements DemoService {public int type() {return 5506;} public void run() {System.out.println(5506);}}
    public static class Impl5507 implements DemoService {public int type() {return 5507;} public void run() {System.out.println(5507);}}
    public static class Impl5508 implements DemoService {public int type() {return 5508;} public void run() {System.out.println(5508);}}
    public static class Impl5509 implements DemoService {public int type() {return 5509;} public void run() {System.out.println(5509);}}
    public static class Impl5510 implements DemoService {public int type() {return 5510;} public void run() {System.out.println(5510);}}
    public static class Impl5511 implements DemoService {public int type() {return 5511;} public void run() {System.out.println(5511);}}
    public static class Impl5512 implements DemoService {public int type() {return 5512;} public void run() {System.out.println(5512);}}
    public static class Impl5513 implements DemoService {public int type() {return 5513;} public void run() {System.out.println(5513);}}
    public static class Impl5514 implements DemoService {public int type() {return 5514;} public void run() {System.out.println(5514);}}
    public static class Impl5515 implements DemoService {public int type() {return 5515;} public void run() {System.out.println(5515);}}
    public static class Impl5516 implements DemoService {public int type() {return 5516;} public void run() {System.out.println(5516);}}
    public static class Impl5517 implements DemoService {public int type() {return 5517;} public void run() {System.out.println(5517);}}
    public static class Impl5518 implements DemoService {public int type() {return 5518;} public void run() {System.out.println(5518);}}
    public static class Impl5519 implements DemoService {public int type() {return 5519;} public void run() {System.out.println(5519);}}
    public static class Impl5520 implements DemoService {public int type() {return 5520;} public void run() {System.out.println(5520);}}
    public static class Impl5521 implements DemoService {public int type() {return 5521;} public void run() {System.out.println(5521);}}
    public static class Impl5522 implements DemoService {public int type() {return 5522;} public void run() {System.out.println(5522);}}
    public static class Impl5523 implements DemoService {public int type() {return 5523;} public void run() {System.out.println(5523);}}
    public static class Impl5524 implements DemoService {public int type() {return 5524;} public void run() {System.out.println(5524);}}
    public static class Impl5525 implements DemoService {public int type() {return 5525;} public void run() {System.out.println(5525);}}
    public static class Impl5526 implements DemoService {public int type() {return 5526;} public void run() {System.out.println(5526);}}
    public static class Impl5527 implements DemoService {public int type() {return 5527;} public void run() {System.out.println(5527);}}
    public static class Impl5528 implements DemoService {public int type() {return 5528;} public void run() {System.out.println(5528);}}
    public static class Impl5529 implements DemoService {public int type() {return 5529;} public void run() {System.out.println(5529);}}
    public static class Impl5530 implements DemoService {public int type() {return 5530;} public void run() {System.out.println(5530);}}
    public static class Impl5531 implements DemoService {public int type() {return 5531;} public void run() {System.out.println(5531);}}
    public static class Impl5532 implements DemoService {public int type() {return 5532;} public void run() {System.out.println(5532);}}
    public static class Impl5533 implements DemoService {public int type() {return 5533;} public void run() {System.out.println(5533);}}
    public static class Impl5534 implements DemoService {public int type() {return 5534;} public void run() {System.out.println(5534);}}
    public static class Impl5535 implements DemoService {public int type() {return 5535;} public void run() {System.out.println(5535);}}
    public static class Impl5536 implements DemoService {public int type() {return 5536;} public void run() {System.out.println(5536);}}
    public static class Impl5537 implements DemoService {public int type() {return 5537;} public void run() {System.out.println(5537);}}
    public static class Impl5538 implements DemoService {public int type() {return 5538;} public void run() {System.out.println(5538);}}
    public static class Impl5539 implements DemoService {public int type() {return 5539;} public void run() {System.out.println(5539);}}
    public static class Impl5540 implements DemoService {public int type() {return 5540;} public void run() {System.out.println(5540);}}
    public static class Impl5541 implements DemoService {public int type() {return 5541;} public void run() {System.out.println(5541);}}
    public static class Impl5542 implements DemoService {public int type() {return 5542;} public void run() {System.out.println(5542);}}
    public static class Impl5543 implements DemoService {public int type() {return 5543;} public void run() {System.out.println(5543);}}
    public static class Impl5544 implements DemoService {public int type() {return 5544;} public void run() {System.out.println(5544);}}
    public static class Impl5545 implements DemoService {public int type() {return 5545;} public void run() {System.out.println(5545);}}
    public static class Impl5546 implements DemoService {public int type() {return 5546;} public void run() {System.out.println(5546);}}
    public static class Impl5547 implements DemoService {public int type() {return 5547;} public void run() {System.out.println(5547);}}
    public static class Impl5548 implements DemoService {public int type() {return 5548;} public void run() {System.out.println(5548);}}
    public static class Impl5549 implements DemoService {public int type() {return 5549;} public void run() {System.out.println(5549);}}
    public static class Impl5550 implements DemoService {public int type() {return 5550;} public void run() {System.out.println(5550);}}
    public static class Impl5551 implements DemoService {public int type() {return 5551;} public void run() {System.out.println(5551);}}
    public static class Impl5552 implements DemoService {public int type() {return 5552;} public void run() {System.out.println(5552);}}
    public static class Impl5553 implements DemoService {public int type() {return 5553;} public void run() {System.out.println(5553);}}
    public static class Impl5554 implements DemoService {public int type() {return 5554;} public void run() {System.out.println(5554);}}
    public static class Impl5555 implements DemoService {public int type() {return 5555;} public void run() {System.out.println(5555);}}
    public static class Impl5556 implements DemoService {public int type() {return 5556;} public void run() {System.out.println(5556);}}
    public static class Impl5557 implements DemoService {public int type() {return 5557;} public void run() {System.out.println(5557);}}
    public static class Impl5558 implements DemoService {public int type() {return 5558;} public void run() {System.out.println(5558);}}
    public static class Impl5559 implements DemoService {public int type() {return 5559;} public void run() {System.out.println(5559);}}
    public static class Impl5560 implements DemoService {public int type() {return 5560;} public void run() {System.out.println(5560);}}
    public static class Impl5561 implements DemoService {public int type() {return 5561;} public void run() {System.out.println(5561);}}
    public static class Impl5562 implements DemoService {public int type() {return 5562;} public void run() {System.out.println(5562);}}
    public static class Impl5563 implements DemoService {public int type() {return 5563;} public void run() {System.out.println(5563);}}
    public static class Impl5564 implements DemoService {public int type() {return 5564;} public void run() {System.out.println(5564);}}
    public static class Impl5565 implements DemoService {public int type() {return 5565;} public void run() {System.out.println(5565);}}
    public static class Impl5566 implements DemoService {public int type() {return 5566;} public void run() {System.out.println(5566);}}
    public static class Impl5567 implements DemoService {public int type() {return 5567;} public void run() {System.out.println(5567);}}
    public static class Impl5568 implements DemoService {public int type() {return 5568;} public void run() {System.out.println(5568);}}
    public static class Impl5569 implements DemoService {public int type() {return 5569;} public void run() {System.out.println(5569);}}
    public static class Impl5570 implements DemoService {public int type() {return 5570;} public void run() {System.out.println(5570);}}
    public static class Impl5571 implements DemoService {public int type() {return 5571;} public void run() {System.out.println(5571);}}
    public static class Impl5572 implements DemoService {public int type() {return 5572;} public void run() {System.out.println(5572);}}
    public static class Impl5573 implements DemoService {public int type() {return 5573;} public void run() {System.out.println(5573);}}
    public static class Impl5574 implements DemoService {public int type() {return 5574;} public void run() {System.out.println(5574);}}
    public static class Impl5575 implements DemoService {public int type() {return 5575;} public void run() {System.out.println(5575);}}
    public static class Impl5576 implements DemoService {public int type() {return 5576;} public void run() {System.out.println(5576);}}
    public static class Impl5577 implements DemoService {public int type() {return 5577;} public void run() {System.out.println(5577);}}
    public static class Impl5578 implements DemoService {public int type() {return 5578;} public void run() {System.out.println(5578);}}
    public static class Impl5579 implements DemoService {public int type() {return 5579;} public void run() {System.out.println(5579);}}
    public static class Impl5580 implements DemoService {public int type() {return 5580;} public void run() {System.out.println(5580);}}
    public static class Impl5581 implements DemoService {public int type() {return 5581;} public void run() {System.out.println(5581);}}
    public static class Impl5582 implements DemoService {public int type() {return 5582;} public void run() {System.out.println(5582);}}
    public static class Impl5583 implements DemoService {public int type() {return 5583;} public void run() {System.out.println(5583);}}
    public static class Impl5584 implements DemoService {public int type() {return 5584;} public void run() {System.out.println(5584);}}
    public static class Impl5585 implements DemoService {public int type() {return 5585;} public void run() {System.out.println(5585);}}
    public static class Impl5586 implements DemoService {public int type() {return 5586;} public void run() {System.out.println(5586);}}
    public static class Impl5587 implements DemoService {public int type() {return 5587;} public void run() {System.out.println(5587);}}
    public static class Impl5588 implements DemoService {public int type() {return 5588;} public void run() {System.out.println(5588);}}
    public static class Impl5589 implements DemoService {public int type() {return 5589;} public void run() {System.out.println(5589);}}
    public static class Impl5590 implements DemoService {public int type() {return 5590;} public void run() {System.out.println(5590);}}
    public static class Impl5591 implements DemoService {public int type() {return 5591;} public void run() {System.out.println(5591);}}
    public static class Impl5592 implements DemoService {public int type() {return 5592;} public void run() {System.out.println(5592);}}
    public static class Impl5593 implements DemoService {public int type() {return 5593;} public void run() {System.out.println(5593);}}
    public static class Impl5594 implements DemoService {public int type() {return 5594;} public void run() {System.out.println(5594);}}
    public static class Impl5595 implements DemoService {public int type() {return 5595;} public void run() {System.out.println(5595);}}
    public static class Impl5596 implements DemoService {public int type() {return 5596;} public void run() {System.out.println(5596);}}
    public static class Impl5597 implements DemoService {public int type() {return 5597;} public void run() {System.out.println(5597);}}
    public static class Impl5598 implements DemoService {public int type() {return 5598;} public void run() {System.out.println(5598);}}
    public static class Impl5599 implements DemoService {public int type() {return 5599;} public void run() {System.out.println(5599);}}
    public static class Impl5600 implements DemoService {public int type() {return 5600;} public void run() {System.out.println(5600);}}
    public static class Impl5601 implements DemoService {public int type() {return 5601;} public void run() {System.out.println(5601);}}
    public static class Impl5602 implements DemoService {public int type() {return 5602;} public void run() {System.out.println(5602);}}
    public static class Impl5603 implements DemoService {public int type() {return 5603;} public void run() {System.out.println(5603);}}
    public static class Impl5604 implements DemoService {public int type() {return 5604;} public void run() {System.out.println(5604);}}
    public static class Impl5605 implements DemoService {public int type() {return 5605;} public void run() {System.out.println(5605);}}
    public static class Impl5606 implements DemoService {public int type() {return 5606;} public void run() {System.out.println(5606);}}
    public static class Impl5607 implements DemoService {public int type() {return 5607;} public void run() {System.out.println(5607);}}
    public static class Impl5608 implements DemoService {public int type() {return 5608;} public void run() {System.out.println(5608);}}
    public static class Impl5609 implements DemoService {public int type() {return 5609;} public void run() {System.out.println(5609);}}
    public static class Impl5610 implements DemoService {public int type() {return 5610;} public void run() {System.out.println(5610);}}
    public static class Impl5611 implements DemoService {public int type() {return 5611;} public void run() {System.out.println(5611);}}
    public static class Impl5612 implements DemoService {public int type() {return 5612;} public void run() {System.out.println(5612);}}
    public static class Impl5613 implements DemoService {public int type() {return 5613;} public void run() {System.out.println(5613);}}
    public static class Impl5614 implements DemoService {public int type() {return 5614;} public void run() {System.out.println(5614);}}
    public static class Impl5615 implements DemoService {public int type() {return 5615;} public void run() {System.out.println(5615);}}
    public static class Impl5616 implements DemoService {public int type() {return 5616;} public void run() {System.out.println(5616);}}
    public static class Impl5617 implements DemoService {public int type() {return 5617;} public void run() {System.out.println(5617);}}
    public static class Impl5618 implements DemoService {public int type() {return 5618;} public void run() {System.out.println(5618);}}
    public static class Impl5619 implements DemoService {public int type() {return 5619;} public void run() {System.out.println(5619);}}
    public static class Impl5620 implements DemoService {public int type() {return 5620;} public void run() {System.out.println(5620);}}
    public static class Impl5621 implements DemoService {public int type() {return 5621;} public void run() {System.out.println(5621);}}
    public static class Impl5622 implements DemoService {public int type() {return 5622;} public void run() {System.out.println(5622);}}
    public static class Impl5623 implements DemoService {public int type() {return 5623;} public void run() {System.out.println(5623);}}
    public static class Impl5624 implements DemoService {public int type() {return 5624;} public void run() {System.out.println(5624);}}
    public static class Impl5625 implements DemoService {public int type() {return 5625;} public void run() {System.out.println(5625);}}
    public static class Impl5626 implements DemoService {public int type() {return 5626;} public void run() {System.out.println(5626);}}
    public static class Impl5627 implements DemoService {public int type() {return 5627;} public void run() {System.out.println(5627);}}
    public static class Impl5628 implements DemoService {public int type() {return 5628;} public void run() {System.out.println(5628);}}
    public static class Impl5629 implements DemoService {public int type() {return 5629;} public void run() {System.out.println(5629);}}
    public static class Impl5630 implements DemoService {public int type() {return 5630;} public void run() {System.out.println(5630);}}
    public static class Impl5631 implements DemoService {public int type() {return 5631;} public void run() {System.out.println(5631);}}
    public static class Impl5632 implements DemoService {public int type() {return 5632;} public void run() {System.out.println(5632);}}
    public static class Impl5633 implements DemoService {public int type() {return 5633;} public void run() {System.out.println(5633);}}
    public static class Impl5634 implements DemoService {public int type() {return 5634;} public void run() {System.out.println(5634);}}
    public static class Impl5635 implements DemoService {public int type() {return 5635;} public void run() {System.out.println(5635);}}
    public static class Impl5636 implements DemoService {public int type() {return 5636;} public void run() {System.out.println(5636);}}
    public static class Impl5637 implements DemoService {public int type() {return 5637;} public void run() {System.out.println(5637);}}
    public static class Impl5638 implements DemoService {public int type() {return 5638;} public void run() {System.out.println(5638);}}
    public static class Impl5639 implements DemoService {public int type() {return 5639;} public void run() {System.out.println(5639);}}
    public static class Impl5640 implements DemoService {public int type() {return 5640;} public void run() {System.out.println(5640);}}
    public static class Impl5641 implements DemoService {public int type() {return 5641;} public void run() {System.out.println(5641);}}
    public static class Impl5642 implements DemoService {public int type() {return 5642;} public void run() {System.out.println(5642);}}
    public static class Impl5643 implements DemoService {public int type() {return 5643;} public void run() {System.out.println(5643);}}
    public static class Impl5644 implements DemoService {public int type() {return 5644;} public void run() {System.out.println(5644);}}
    public static class Impl5645 implements DemoService {public int type() {return 5645;} public void run() {System.out.println(5645);}}
    public static class Impl5646 implements DemoService {public int type() {return 5646;} public void run() {System.out.println(5646);}}
    public static class Impl5647 implements DemoService {public int type() {return 5647;} public void run() {System.out.println(5647);}}
    public static class Impl5648 implements DemoService {public int type() {return 5648;} public void run() {System.out.println(5648);}}
    public static class Impl5649 implements DemoService {public int type() {return 5649;} public void run() {System.out.println(5649);}}
    public static class Impl5650 implements DemoService {public int type() {return 5650;} public void run() {System.out.println(5650);}}
    public static class Impl5651 implements DemoService {public int type() {return 5651;} public void run() {System.out.println(5651);}}
    public static class Impl5652 implements DemoService {public int type() {return 5652;} public void run() {System.out.println(5652);}}
    public static class Impl5653 implements DemoService {public int type() {return 5653;} public void run() {System.out.println(5653);}}
    public static class Impl5654 implements DemoService {public int type() {return 5654;} public void run() {System.out.println(5654);}}
    public static class Impl5655 implements DemoService {public int type() {return 5655;} public void run() {System.out.println(5655);}}
    public static class Impl5656 implements DemoService {public int type() {return 5656;} public void run() {System.out.println(5656);}}
    public static class Impl5657 implements DemoService {public int type() {return 5657;} public void run() {System.out.println(5657);}}
    public static class Impl5658 implements DemoService {public int type() {return 5658;} public void run() {System.out.println(5658);}}
    public static class Impl5659 implements DemoService {public int type() {return 5659;} public void run() {System.out.println(5659);}}
    public static class Impl5660 implements DemoService {public int type() {return 5660;} public void run() {System.out.println(5660);}}
    public static class Impl5661 implements DemoService {public int type() {return 5661;} public void run() {System.out.println(5661);}}
    public static class Impl5662 implements DemoService {public int type() {return 5662;} public void run() {System.out.println(5662);}}
    public static class Impl5663 implements DemoService {public int type() {return 5663;} public void run() {System.out.println(5663);}}
    public static class Impl5664 implements DemoService {public int type() {return 5664;} public void run() {System.out.println(5664);}}
    public static class Impl5665 implements DemoService {public int type() {return 5665;} public void run() {System.out.println(5665);}}
    public static class Impl5666 implements DemoService {public int type() {return 5666;} public void run() {System.out.println(5666);}}
    public static class Impl5667 implements DemoService {public int type() {return 5667;} public void run() {System.out.println(5667);}}
    public static class Impl5668 implements DemoService {public int type() {return 5668;} public void run() {System.out.println(5668);}}
    public static class Impl5669 implements DemoService {public int type() {return 5669;} public void run() {System.out.println(5669);}}
    public static class Impl5670 implements DemoService {public int type() {return 5670;} public void run() {System.out.println(5670);}}
    public static class Impl5671 implements DemoService {public int type() {return 5671;} public void run() {System.out.println(5671);}}
    public static class Impl5672 implements DemoService {public int type() {return 5672;} public void run() {System.out.println(5672);}}
    public static class Impl5673 implements DemoService {public int type() {return 5673;} public void run() {System.out.println(5673);}}
    public static class Impl5674 implements DemoService {public int type() {return 5674;} public void run() {System.out.println(5674);}}
    public static class Impl5675 implements DemoService {public int type() {return 5675;} public void run() {System.out.println(5675);}}
    public static class Impl5676 implements DemoService {public int type() {return 5676;} public void run() {System.out.println(5676);}}
    public static class Impl5677 implements DemoService {public int type() {return 5677;} public void run() {System.out.println(5677);}}
    public static class Impl5678 implements DemoService {public int type() {return 5678;} public void run() {System.out.println(5678);}}
    public static class Impl5679 implements DemoService {public int type() {return 5679;} public void run() {System.out.println(5679);}}
    public static class Impl5680 implements DemoService {public int type() {return 5680;} public void run() {System.out.println(5680);}}
    public static class Impl5681 implements DemoService {public int type() {return 5681;} public void run() {System.out.println(5681);}}
    public static class Impl5682 implements DemoService {public int type() {return 5682;} public void run() {System.out.println(5682);}}
    public static class Impl5683 implements DemoService {public int type() {return 5683;} public void run() {System.out.println(5683);}}
    public static class Impl5684 implements DemoService {public int type() {return 5684;} public void run() {System.out.println(5684);}}
    public static class Impl5685 implements DemoService {public int type() {return 5685;} public void run() {System.out.println(5685);}}
    public static class Impl5686 implements DemoService {public int type() {return 5686;} public void run() {System.out.println(5686);}}
    public static class Impl5687 implements DemoService {public int type() {return 5687;} public void run() {System.out.println(5687);}}
    public static class Impl5688 implements DemoService {public int type() {return 5688;} public void run() {System.out.println(5688);}}
    public static class Impl5689 implements DemoService {public int type() {return 5689;} public void run() {System.out.println(5689);}}
    public static class Impl5690 implements DemoService {public int type() {return 5690;} public void run() {System.out.println(5690);}}
    public static class Impl5691 implements DemoService {public int type() {return 5691;} public void run() {System.out.println(5691);}}
    public static class Impl5692 implements DemoService {public int type() {return 5692;} public void run() {System.out.println(5692);}}
    public static class Impl5693 implements DemoService {public int type() {return 5693;} public void run() {System.out.println(5693);}}
    public static class Impl5694 implements DemoService {public int type() {return 5694;} public void run() {System.out.println(5694);}}
    public static class Impl5695 implements DemoService {public int type() {return 5695;} public void run() {System.out.println(5695);}}
    public static class Impl5696 implements DemoService {public int type() {return 5696;} public void run() {System.out.println(5696);}}
    public static class Impl5697 implements DemoService {public int type() {return 5697;} public void run() {System.out.println(5697);}}
    public static class Impl5698 implements DemoService {public int type() {return 5698;} public void run() {System.out.println(5698);}}
    public static class Impl5699 implements DemoService {public int type() {return 5699;} public void run() {System.out.println(5699);}}
    public static class Impl5700 implements DemoService {public int type() {return 5700;} public void run() {System.out.println(5700);}}
    public static class Impl5701 implements DemoService {public int type() {return 5701;} public void run() {System.out.println(5701);}}
    public static class Impl5702 implements DemoService {public int type() {return 5702;} public void run() {System.out.println(5702);}}
    public static class Impl5703 implements DemoService {public int type() {return 5703;} public void run() {System.out.println(5703);}}
    public static class Impl5704 implements DemoService {public int type() {return 5704;} public void run() {System.out.println(5704);}}
    public static class Impl5705 implements DemoService {public int type() {return 5705;} public void run() {System.out.println(5705);}}
    public static class Impl5706 implements DemoService {public int type() {return 5706;} public void run() {System.out.println(5706);}}
    public static class Impl5707 implements DemoService {public int type() {return 5707;} public void run() {System.out.println(5707);}}
    public static class Impl5708 implements DemoService {public int type() {return 5708;} public void run() {System.out.println(5708);}}
    public static class Impl5709 implements DemoService {public int type() {return 5709;} public void run() {System.out.println(5709);}}
    public static class Impl5710 implements DemoService {public int type() {return 5710;} public void run() {System.out.println(5710);}}
    public static class Impl5711 implements DemoService {public int type() {return 5711;} public void run() {System.out.println(5711);}}
    public static class Impl5712 implements DemoService {public int type() {return 5712;} public void run() {System.out.println(5712);}}
    public static class Impl5713 implements DemoService {public int type() {return 5713;} public void run() {System.out.println(5713);}}
    public static class Impl5714 implements DemoService {public int type() {return 5714;} public void run() {System.out.println(5714);}}
    public static class Impl5715 implements DemoService {public int type() {return 5715;} public void run() {System.out.println(5715);}}
    public static class Impl5716 implements DemoService {public int type() {return 5716;} public void run() {System.out.println(5716);}}
    public static class Impl5717 implements DemoService {public int type() {return 5717;} public void run() {System.out.println(5717);}}
    public static class Impl5718 implements DemoService {public int type() {return 5718;} public void run() {System.out.println(5718);}}
    public static class Impl5719 implements DemoService {public int type() {return 5719;} public void run() {System.out.println(5719);}}
    public static class Impl5720 implements DemoService {public int type() {return 5720;} public void run() {System.out.println(5720);}}
    public static class Impl5721 implements DemoService {public int type() {return 5721;} public void run() {System.out.println(5721);}}
    public static class Impl5722 implements DemoService {public int type() {return 5722;} public void run() {System.out.println(5722);}}
    public static class Impl5723 implements DemoService {public int type() {return 5723;} public void run() {System.out.println(5723);}}
    public static class Impl5724 implements DemoService {public int type() {return 5724;} public void run() {System.out.println(5724);}}
    public static class Impl5725 implements DemoService {public int type() {return 5725;} public void run() {System.out.println(5725);}}
    public static class Impl5726 implements DemoService {public int type() {return 5726;} public void run() {System.out.println(5726);}}
    public static class Impl5727 implements DemoService {public int type() {return 5727;} public void run() {System.out.println(5727);}}
    public static class Impl5728 implements DemoService {public int type() {return 5728;} public void run() {System.out.println(5728);}}
    public static class Impl5729 implements DemoService {public int type() {return 5729;} public void run() {System.out.println(5729);}}
    public static class Impl5730 implements DemoService {public int type() {return 5730;} public void run() {System.out.println(5730);}}
    public static class Impl5731 implements DemoService {public int type() {return 5731;} public void run() {System.out.println(5731);}}
    public static class Impl5732 implements DemoService {public int type() {return 5732;} public void run() {System.out.println(5732);}}
    public static class Impl5733 implements DemoService {public int type() {return 5733;} public void run() {System.out.println(5733);}}
    public static class Impl5734 implements DemoService {public int type() {return 5734;} public void run() {System.out.println(5734);}}
    public static class Impl5735 implements DemoService {public int type() {return 5735;} public void run() {System.out.println(5735);}}
    public static class Impl5736 implements DemoService {public int type() {return 5736;} public void run() {System.out.println(5736);}}
    public static class Impl5737 implements DemoService {public int type() {return 5737;} public void run() {System.out.println(5737);}}
    public static class Impl5738 implements DemoService {public int type() {return 5738;} public void run() {System.out.println(5738);}}
    public static class Impl5739 implements DemoService {public int type() {return 5739;} public void run() {System.out.println(5739);}}
    public static class Impl5740 implements DemoService {public int type() {return 5740;} public void run() {System.out.println(5740);}}
    public static class Impl5741 implements DemoService {public int type() {return 5741;} public void run() {System.out.println(5741);}}
    public static class Impl5742 implements DemoService {public int type() {return 5742;} public void run() {System.out.println(5742);}}
    public static class Impl5743 implements DemoService {public int type() {return 5743;} public void run() {System.out.println(5743);}}
    public static class Impl5744 implements DemoService {public int type() {return 5744;} public void run() {System.out.println(5744);}}
    public static class Impl5745 implements DemoService {public int type() {return 5745;} public void run() {System.out.println(5745);}}
    public static class Impl5746 implements DemoService {public int type() {return 5746;} public void run() {System.out.println(5746);}}
    public static class Impl5747 implements DemoService {public int type() {return 5747;} public void run() {System.out.println(5747);}}
    public static class Impl5748 implements DemoService {public int type() {return 5748;} public void run() {System.out.println(5748);}}
    public static class Impl5749 implements DemoService {public int type() {return 5749;} public void run() {System.out.println(5749);}}
    public static class Impl5750 implements DemoService {public int type() {return 5750;} public void run() {System.out.println(5750);}}
    public static class Impl5751 implements DemoService {public int type() {return 5751;} public void run() {System.out.println(5751);}}
    public static class Impl5752 implements DemoService {public int type() {return 5752;} public void run() {System.out.println(5752);}}
    public static class Impl5753 implements DemoService {public int type() {return 5753;} public void run() {System.out.println(5753);}}
    public static class Impl5754 implements DemoService {public int type() {return 5754;} public void run() {System.out.println(5754);}}
    public static class Impl5755 implements DemoService {public int type() {return 5755;} public void run() {System.out.println(5755);}}
    public static class Impl5756 implements DemoService {public int type() {return 5756;} public void run() {System.out.println(5756);}}
    public static class Impl5757 implements DemoService {public int type() {return 5757;} public void run() {System.out.println(5757);}}
    public static class Impl5758 implements DemoService {public int type() {return 5758;} public void run() {System.out.println(5758);}}
    public static class Impl5759 implements DemoService {public int type() {return 5759;} public void run() {System.out.println(5759);}}
    public static class Impl5760 implements DemoService {public int type() {return 5760;} public void run() {System.out.println(5760);}}
    public static class Impl5761 implements DemoService {public int type() {return 5761;} public void run() {System.out.println(5761);}}
    public static class Impl5762 implements DemoService {public int type() {return 5762;} public void run() {System.out.println(5762);}}
    public static class Impl5763 implements DemoService {public int type() {return 5763;} public void run() {System.out.println(5763);}}
    public static class Impl5764 implements DemoService {public int type() {return 5764;} public void run() {System.out.println(5764);}}
    public static class Impl5765 implements DemoService {public int type() {return 5765;} public void run() {System.out.println(5765);}}
    public static class Impl5766 implements DemoService {public int type() {return 5766;} public void run() {System.out.println(5766);}}
    public static class Impl5767 implements DemoService {public int type() {return 5767;} public void run() {System.out.println(5767);}}
    public static class Impl5768 implements DemoService {public int type() {return 5768;} public void run() {System.out.println(5768);}}
    public static class Impl5769 implements DemoService {public int type() {return 5769;} public void run() {System.out.println(5769);}}
    public static class Impl5770 implements DemoService {public int type() {return 5770;} public void run() {System.out.println(5770);}}
    public static class Impl5771 implements DemoService {public int type() {return 5771;} public void run() {System.out.println(5771);}}
    public static class Impl5772 implements DemoService {public int type() {return 5772;} public void run() {System.out.println(5772);}}
    public static class Impl5773 implements DemoService {public int type() {return 5773;} public void run() {System.out.println(5773);}}
    public static class Impl5774 implements DemoService {public int type() {return 5774;} public void run() {System.out.println(5774);}}
    public static class Impl5775 implements DemoService {public int type() {return 5775;} public void run() {System.out.println(5775);}}
    public static class Impl5776 implements DemoService {public int type() {return 5776;} public void run() {System.out.println(5776);}}
    public static class Impl5777 implements DemoService {public int type() {return 5777;} public void run() {System.out.println(5777);}}
    public static class Impl5778 implements DemoService {public int type() {return 5778;} public void run() {System.out.println(5778);}}
    public static class Impl5779 implements DemoService {public int type() {return 5779;} public void run() {System.out.println(5779);}}
    public static class Impl5780 implements DemoService {public int type() {return 5780;} public void run() {System.out.println(5780);}}
    public static class Impl5781 implements DemoService {public int type() {return 5781;} public void run() {System.out.println(5781);}}
    public static class Impl5782 implements DemoService {public int type() {return 5782;} public void run() {System.out.println(5782);}}
    public static class Impl5783 implements DemoService {public int type() {return 5783;} public void run() {System.out.println(5783);}}
    public static class Impl5784 implements DemoService {public int type() {return 5784;} public void run() {System.out.println(5784);}}
    public static class Impl5785 implements DemoService {public int type() {return 5785;} public void run() {System.out.println(5785);}}
    public static class Impl5786 implements DemoService {public int type() {return 5786;} public void run() {System.out.println(5786);}}
    public static class Impl5787 implements DemoService {public int type() {return 5787;} public void run() {System.out.println(5787);}}
    public static class Impl5788 implements DemoService {public int type() {return 5788;} public void run() {System.out.println(5788);}}
    public static class Impl5789 implements DemoService {public int type() {return 5789;} public void run() {System.out.println(5789);}}
    public static class Impl5790 implements DemoService {public int type() {return 5790;} public void run() {System.out.println(5790);}}
    public static class Impl5791 implements DemoService {public int type() {return 5791;} public void run() {System.out.println(5791);}}
    public static class Impl5792 implements DemoService {public int type() {return 5792;} public void run() {System.out.println(5792);}}
    public static class Impl5793 implements DemoService {public int type() {return 5793;} public void run() {System.out.println(5793);}}
    public static class Impl5794 implements DemoService {public int type() {return 5794;} public void run() {System.out.println(5794);}}
    public static class Impl5795 implements DemoService {public int type() {return 5795;} public void run() {System.out.println(5795);}}
    public static class Impl5796 implements DemoService {public int type() {return 5796;} public void run() {System.out.println(5796);}}
    public static class Impl5797 implements DemoService {public int type() {return 5797;} public void run() {System.out.println(5797);}}
    public static class Impl5798 implements DemoService {public int type() {return 5798;} public void run() {System.out.println(5798);}}
    public static class Impl5799 implements DemoService {public int type() {return 5799;} public void run() {System.out.println(5799);}}
    public static class Impl5800 implements DemoService {public int type() {return 5800;} public void run() {System.out.println(5800);}}
    public static class Impl5801 implements DemoService {public int type() {return 5801;} public void run() {System.out.println(5801);}}
    public static class Impl5802 implements DemoService {public int type() {return 5802;} public void run() {System.out.println(5802);}}
    public static class Impl5803 implements DemoService {public int type() {return 5803;} public void run() {System.out.println(5803);}}
    public static class Impl5804 implements DemoService {public int type() {return 5804;} public void run() {System.out.println(5804);}}
    public static class Impl5805 implements DemoService {public int type() {return 5805;} public void run() {System.out.println(5805);}}
    public static class Impl5806 implements DemoService {public int type() {return 5806;} public void run() {System.out.println(5806);}}
    public static class Impl5807 implements DemoService {public int type() {return 5807;} public void run() {System.out.println(5807);}}
    public static class Impl5808 implements DemoService {public int type() {return 5808;} public void run() {System.out.println(5808);}}
    public static class Impl5809 implements DemoService {public int type() {return 5809;} public void run() {System.out.println(5809);}}
    public static class Impl5810 implements DemoService {public int type() {return 5810;} public void run() {System.out.println(5810);}}
    public static class Impl5811 implements DemoService {public int type() {return 5811;} public void run() {System.out.println(5811);}}
    public static class Impl5812 implements DemoService {public int type() {return 5812;} public void run() {System.out.println(5812);}}
    public static class Impl5813 implements DemoService {public int type() {return 5813;} public void run() {System.out.println(5813);}}
    public static class Impl5814 implements DemoService {public int type() {return 5814;} public void run() {System.out.println(5814);}}
    public static class Impl5815 implements DemoService {public int type() {return 5815;} public void run() {System.out.println(5815);}}
    public static class Impl5816 implements DemoService {public int type() {return 5816;} public void run() {System.out.println(5816);}}
    public static class Impl5817 implements DemoService {public int type() {return 5817;} public void run() {System.out.println(5817);}}
    public static class Impl5818 implements DemoService {public int type() {return 5818;} public void run() {System.out.println(5818);}}
    public static class Impl5819 implements DemoService {public int type() {return 5819;} public void run() {System.out.println(5819);}}
    public static class Impl5820 implements DemoService {public int type() {return 5820;} public void run() {System.out.println(5820);}}
    public static class Impl5821 implements DemoService {public int type() {return 5821;} public void run() {System.out.println(5821);}}
    public static class Impl5822 implements DemoService {public int type() {return 5822;} public void run() {System.out.println(5822);}}
    public static class Impl5823 implements DemoService {public int type() {return 5823;} public void run() {System.out.println(5823);}}
    public static class Impl5824 implements DemoService {public int type() {return 5824;} public void run() {System.out.println(5824);}}
    public static class Impl5825 implements DemoService {public int type() {return 5825;} public void run() {System.out.println(5825);}}
    public static class Impl5826 implements DemoService {public int type() {return 5826;} public void run() {System.out.println(5826);}}
    public static class Impl5827 implements DemoService {public int type() {return 5827;} public void run() {System.out.println(5827);}}
    public static class Impl5828 implements DemoService {public int type() {return 5828;} public void run() {System.out.println(5828);}}
    public static class Impl5829 implements DemoService {public int type() {return 5829;} public void run() {System.out.println(5829);}}
    public static class Impl5830 implements DemoService {public int type() {return 5830;} public void run() {System.out.println(5830);}}
    public static class Impl5831 implements DemoService {public int type() {return 5831;} public void run() {System.out.println(5831);}}
    public static class Impl5832 implements DemoService {public int type() {return 5832;} public void run() {System.out.println(5832);}}
    public static class Impl5833 implements DemoService {public int type() {return 5833;} public void run() {System.out.println(5833);}}
    public static class Impl5834 implements DemoService {public int type() {return 5834;} public void run() {System.out.println(5834);}}
    public static class Impl5835 implements DemoService {public int type() {return 5835;} public void run() {System.out.println(5835);}}
    public static class Impl5836 implements DemoService {public int type() {return 5836;} public void run() {System.out.println(5836);}}
    public static class Impl5837 implements DemoService {public int type() {return 5837;} public void run() {System.out.println(5837);}}
    public static class Impl5838 implements DemoService {public int type() {return 5838;} public void run() {System.out.println(5838);}}
    public static class Impl5839 implements DemoService {public int type() {return 5839;} public void run() {System.out.println(5839);}}
    public static class Impl5840 implements DemoService {public int type() {return 5840;} public void run() {System.out.println(5840);}}
    public static class Impl5841 implements DemoService {public int type() {return 5841;} public void run() {System.out.println(5841);}}
    public static class Impl5842 implements DemoService {public int type() {return 5842;} public void run() {System.out.println(5842);}}
    public static class Impl5843 implements DemoService {public int type() {return 5843;} public void run() {System.out.println(5843);}}
    public static class Impl5844 implements DemoService {public int type() {return 5844;} public void run() {System.out.println(5844);}}
    public static class Impl5845 implements DemoService {public int type() {return 5845;} public void run() {System.out.println(5845);}}
    public static class Impl5846 implements DemoService {public int type() {return 5846;} public void run() {System.out.println(5846);}}
    public static class Impl5847 implements DemoService {public int type() {return 5847;} public void run() {System.out.println(5847);}}
    public static class Impl5848 implements DemoService {public int type() {return 5848;} public void run() {System.out.println(5848);}}
    public static class Impl5849 implements DemoService {public int type() {return 5849;} public void run() {System.out.println(5849);}}
    public static class Impl5850 implements DemoService {public int type() {return 5850;} public void run() {System.out.println(5850);}}
    public static class Impl5851 implements DemoService {public int type() {return 5851;} public void run() {System.out.println(5851);}}
    public static class Impl5852 implements DemoService {public int type() {return 5852;} public void run() {System.out.println(5852);}}
    public static class Impl5853 implements DemoService {public int type() {return 5853;} public void run() {System.out.println(5853);}}
    public static class Impl5854 implements DemoService {public int type() {return 5854;} public void run() {System.out.println(5854);}}
    public static class Impl5855 implements DemoService {public int type() {return 5855;} public void run() {System.out.println(5855);}}
    public static class Impl5856 implements DemoService {public int type() {return 5856;} public void run() {System.out.println(5856);}}
    public static class Impl5857 implements DemoService {public int type() {return 5857;} public void run() {System.out.println(5857);}}
    public static class Impl5858 implements DemoService {public int type() {return 5858;} public void run() {System.out.println(5858);}}
    public static class Impl5859 implements DemoService {public int type() {return 5859;} public void run() {System.out.println(5859);}}
    public static class Impl5860 implements DemoService {public int type() {return 5860;} public void run() {System.out.println(5860);}}
    public static class Impl5861 implements DemoService {public int type() {return 5861;} public void run() {System.out.println(5861);}}
    public static class Impl5862 implements DemoService {public int type() {return 5862;} public void run() {System.out.println(5862);}}
    public static class Impl5863 implements DemoService {public int type() {return 5863;} public void run() {System.out.println(5863);}}
    public static class Impl5864 implements DemoService {public int type() {return 5864;} public void run() {System.out.println(5864);}}
    public static class Impl5865 implements DemoService {public int type() {return 5865;} public void run() {System.out.println(5865);}}
    public static class Impl5866 implements DemoService {public int type() {return 5866;} public void run() {System.out.println(5866);}}
    public static class Impl5867 implements DemoService {public int type() {return 5867;} public void run() {System.out.println(5867);}}
    public static class Impl5868 implements DemoService {public int type() {return 5868;} public void run() {System.out.println(5868);}}
    public static class Impl5869 implements DemoService {public int type() {return 5869;} public void run() {System.out.println(5869);}}
    public static class Impl5870 implements DemoService {public int type() {return 5870;} public void run() {System.out.println(5870);}}
    public static class Impl5871 implements DemoService {public int type() {return 5871;} public void run() {System.out.println(5871);}}
    public static class Impl5872 implements DemoService {public int type() {return 5872;} public void run() {System.out.println(5872);}}
    public static class Impl5873 implements DemoService {public int type() {return 5873;} public void run() {System.out.println(5873);}}
    public static class Impl5874 implements DemoService {public int type() {return 5874;} public void run() {System.out.println(5874);}}
    public static class Impl5875 implements DemoService {public int type() {return 5875;} public void run() {System.out.println(5875);}}
    public static class Impl5876 implements DemoService {public int type() {return 5876;} public void run() {System.out.println(5876);}}
    public static class Impl5877 implements DemoService {public int type() {return 5877;} public void run() {System.out.println(5877);}}
    public static class Impl5878 implements DemoService {public int type() {return 5878;} public void run() {System.out.println(5878);}}
    public static class Impl5879 implements DemoService {public int type() {return 5879;} public void run() {System.out.println(5879);}}
    public static class Impl5880 implements DemoService {public int type() {return 5880;} public void run() {System.out.println(5880);}}
    public static class Impl5881 implements DemoService {public int type() {return 5881;} public void run() {System.out.println(5881);}}
    public static class Impl5882 implements DemoService {public int type() {return 5882;} public void run() {System.out.println(5882);}}
    public static class Impl5883 implements DemoService {public int type() {return 5883;} public void run() {System.out.println(5883);}}
    public static class Impl5884 implements DemoService {public int type() {return 5884;} public void run() {System.out.println(5884);}}
    public static class Impl5885 implements DemoService {public int type() {return 5885;} public void run() {System.out.println(5885);}}
    public static class Impl5886 implements DemoService {public int type() {return 5886;} public void run() {System.out.println(5886);}}
    public static class Impl5887 implements DemoService {public int type() {return 5887;} public void run() {System.out.println(5887);}}
    public static class Impl5888 implements DemoService {public int type() {return 5888;} public void run() {System.out.println(5888);}}
    public static class Impl5889 implements DemoService {public int type() {return 5889;} public void run() {System.out.println(5889);}}
    public static class Impl5890 implements DemoService {public int type() {return 5890;} public void run() {System.out.println(5890);}}
    public static class Impl5891 implements DemoService {public int type() {return 5891;} public void run() {System.out.println(5891);}}
    public static class Impl5892 implements DemoService {public int type() {return 5892;} public void run() {System.out.println(5892);}}
    public static class Impl5893 implements DemoService {public int type() {return 5893;} public void run() {System.out.println(5893);}}
    public static class Impl5894 implements DemoService {public int type() {return 5894;} public void run() {System.out.println(5894);}}
    public static class Impl5895 implements DemoService {public int type() {return 5895;} public void run() {System.out.println(5895);}}
    public static class Impl5896 implements DemoService {public int type() {return 5896;} public void run() {System.out.println(5896);}}
    public static class Impl5897 implements DemoService {public int type() {return 5897;} public void run() {System.out.println(5897);}}
    public static class Impl5898 implements DemoService {public int type() {return 5898;} public void run() {System.out.println(5898);}}
    public static class Impl5899 implements DemoService {public int type() {return 5899;} public void run() {System.out.println(5899);}}
    public static class Impl5900 implements DemoService {public int type() {return 5900;} public void run() {System.out.println(5900);}}
    public static class Impl5901 implements DemoService {public int type() {return 5901;} public void run() {System.out.println(5901);}}
    public static class Impl5902 implements DemoService {public int type() {return 5902;} public void run() {System.out.println(5902);}}
    public static class Impl5903 implements DemoService {public int type() {return 5903;} public void run() {System.out.println(5903);}}
    public static class Impl5904 implements DemoService {public int type() {return 5904;} public void run() {System.out.println(5904);}}
    public static class Impl5905 implements DemoService {public int type() {return 5905;} public void run() {System.out.println(5905);}}
    public static class Impl5906 implements DemoService {public int type() {return 5906;} public void run() {System.out.println(5906);}}
    public static class Impl5907 implements DemoService {public int type() {return 5907;} public void run() {System.out.println(5907);}}
    public static class Impl5908 implements DemoService {public int type() {return 5908;} public void run() {System.out.println(5908);}}
    public static class Impl5909 implements DemoService {public int type() {return 5909;} public void run() {System.out.println(5909);}}
    public static class Impl5910 implements DemoService {public int type() {return 5910;} public void run() {System.out.println(5910);}}
    public static class Impl5911 implements DemoService {public int type() {return 5911;} public void run() {System.out.println(5911);}}
    public static class Impl5912 implements DemoService {public int type() {return 5912;} public void run() {System.out.println(5912);}}
    public static class Impl5913 implements DemoService {public int type() {return 5913;} public void run() {System.out.println(5913);}}
    public static class Impl5914 implements DemoService {public int type() {return 5914;} public void run() {System.out.println(5914);}}
    public static class Impl5915 implements DemoService {public int type() {return 5915;} public void run() {System.out.println(5915);}}
    public static class Impl5916 implements DemoService {public int type() {return 5916;} public void run() {System.out.println(5916);}}
    public static class Impl5917 implements DemoService {public int type() {return 5917;} public void run() {System.out.println(5917);}}
    public static class Impl5918 implements DemoService {public int type() {return 5918;} public void run() {System.out.println(5918);}}
    public static class Impl5919 implements DemoService {public int type() {return 5919;} public void run() {System.out.println(5919);}}
    public static class Impl5920 implements DemoService {public int type() {return 5920;} public void run() {System.out.println(5920);}}
    public static class Impl5921 implements DemoService {public int type() {return 5921;} public void run() {System.out.println(5921);}}
    public static class Impl5922 implements DemoService {public int type() {return 5922;} public void run() {System.out.println(5922);}}
    public static class Impl5923 implements DemoService {public int type() {return 5923;} public void run() {System.out.println(5923);}}
    public static class Impl5924 implements DemoService {public int type() {return 5924;} public void run() {System.out.println(5924);}}
    public static class Impl5925 implements DemoService {public int type() {return 5925;} public void run() {System.out.println(5925);}}
    public static class Impl5926 implements DemoService {public int type() {return 5926;} public void run() {System.out.println(5926);}}
    public static class Impl5927 implements DemoService {public int type() {return 5927;} public void run() {System.out.println(5927);}}
    public static class Impl5928 implements DemoService {public int type() {return 5928;} public void run() {System.out.println(5928);}}
    public static class Impl5929 implements DemoService {public int type() {return 5929;} public void run() {System.out.println(5929);}}
    public static class Impl5930 implements DemoService {public int type() {return 5930;} public void run() {System.out.println(5930);}}
    public static class Impl5931 implements DemoService {public int type() {return 5931;} public void run() {System.out.println(5931);}}
    public static class Impl5932 implements DemoService {public int type() {return 5932;} public void run() {System.out.println(5932);}}
    public static class Impl5933 implements DemoService {public int type() {return 5933;} public void run() {System.out.println(5933);}}
    public static class Impl5934 implements DemoService {public int type() {return 5934;} public void run() {System.out.println(5934);}}
    public static class Impl5935 implements DemoService {public int type() {return 5935;} public void run() {System.out.println(5935);}}
    public static class Impl5936 implements DemoService {public int type() {return 5936;} public void run() {System.out.println(5936);}}
    public static class Impl5937 implements DemoService {public int type() {return 5937;} public void run() {System.out.println(5937);}}
    public static class Impl5938 implements DemoService {public int type() {return 5938;} public void run() {System.out.println(5938);}}
    public static class Impl5939 implements DemoService {public int type() {return 5939;} public void run() {System.out.println(5939);}}
    public static class Impl5940 implements DemoService {public int type() {return 5940;} public void run() {System.out.println(5940);}}
    public static class Impl5941 implements DemoService {public int type() {return 5941;} public void run() {System.out.println(5941);}}
    public static class Impl5942 implements DemoService {public int type() {return 5942;} public void run() {System.out.println(5942);}}
    public static class Impl5943 implements DemoService {public int type() {return 5943;} public void run() {System.out.println(5943);}}
    public static class Impl5944 implements DemoService {public int type() {return 5944;} public void run() {System.out.println(5944);}}
    public static class Impl5945 implements DemoService {public int type() {return 5945;} public void run() {System.out.println(5945);}}
    public static class Impl5946 implements DemoService {public int type() {return 5946;} public void run() {System.out.println(5946);}}
    public static class Impl5947 implements DemoService {public int type() {return 5947;} public void run() {System.out.println(5947);}}
    public static class Impl5948 implements DemoService {public int type() {return 5948;} public void run() {System.out.println(5948);}}
    public static class Impl5949 implements DemoService {public int type() {return 5949;} public void run() {System.out.println(5949);}}
    public static class Impl5950 implements DemoService {public int type() {return 5950;} public void run() {System.out.println(5950);}}
    public static class Impl5951 implements DemoService {public int type() {return 5951;} public void run() {System.out.println(5951);}}
    public static class Impl5952 implements DemoService {public int type() {return 5952;} public void run() {System.out.println(5952);}}
    public static class Impl5953 implements DemoService {public int type() {return 5953;} public void run() {System.out.println(5953);}}
    public static class Impl5954 implements DemoService {public int type() {return 5954;} public void run() {System.out.println(5954);}}
    public static class Impl5955 implements DemoService {public int type() {return 5955;} public void run() {System.out.println(5955);}}
    public static class Impl5956 implements DemoService {public int type() {return 5956;} public void run() {System.out.println(5956);}}
    public static class Impl5957 implements DemoService {public int type() {return 5957;} public void run() {System.out.println(5957);}}
    public static class Impl5958 implements DemoService {public int type() {return 5958;} public void run() {System.out.println(5958);}}
    public static class Impl5959 implements DemoService {public int type() {return 5959;} public void run() {System.out.println(5959);}}
    public static class Impl5960 implements DemoService {public int type() {return 5960;} public void run() {System.out.println(5960);}}
    public static class Impl5961 implements DemoService {public int type() {return 5961;} public void run() {System.out.println(5961);}}
    public static class Impl5962 implements DemoService {public int type() {return 5962;} public void run() {System.out.println(5962);}}
    public static class Impl5963 implements DemoService {public int type() {return 5963;} public void run() {System.out.println(5963);}}
    public static class Impl5964 implements DemoService {public int type() {return 5964;} public void run() {System.out.println(5964);}}
    public static class Impl5965 implements DemoService {public int type() {return 5965;} public void run() {System.out.println(5965);}}
    public static class Impl5966 implements DemoService {public int type() {return 5966;} public void run() {System.out.println(5966);}}
    public static class Impl5967 implements DemoService {public int type() {return 5967;} public void run() {System.out.println(5967);}}
    public static class Impl5968 implements DemoService {public int type() {return 5968;} public void run() {System.out.println(5968);}}
    public static class Impl5969 implements DemoService {public int type() {return 5969;} public void run() {System.out.println(5969);}}
    public static class Impl5970 implements DemoService {public int type() {return 5970;} public void run() {System.out.println(5970);}}
    public static class Impl5971 implements DemoService {public int type() {return 5971;} public void run() {System.out.println(5971);}}
    public static class Impl5972 implements DemoService {public int type() {return 5972;} public void run() {System.out.println(5972);}}
    public static class Impl5973 implements DemoService {public int type() {return 5973;} public void run() {System.out.println(5973);}}
    public static class Impl5974 implements DemoService {public int type() {return 5974;} public void run() {System.out.println(5974);}}
    public static class Impl5975 implements DemoService {public int type() {return 5975;} public void run() {System.out.println(5975);}}
    public static class Impl5976 implements DemoService {public int type() {return 5976;} public void run() {System.out.println(5976);}}
    public static class Impl5977 implements DemoService {public int type() {return 5977;} public void run() {System.out.println(5977);}}
    public static class Impl5978 implements DemoService {public int type() {return 5978;} public void run() {System.out.println(5978);}}
    public static class Impl5979 implements DemoService {public int type() {return 5979;} public void run() {System.out.println(5979);}}
    public static class Impl5980 implements DemoService {public int type() {return 5980;} public void run() {System.out.println(5980);}}
    public static class Impl5981 implements DemoService {public int type() {return 5981;} public void run() {System.out.println(5981);}}
    public static class Impl5982 implements DemoService {public int type() {return 5982;} public void run() {System.out.println(5982);}}
    public static class Impl5983 implements DemoService {public int type() {return 5983;} public void run() {System.out.println(5983);}}
    public static class Impl5984 implements DemoService {public int type() {return 5984;} public void run() {System.out.println(5984);}}
    public static class Impl5985 implements DemoService {public int type() {return 5985;} public void run() {System.out.println(5985);}}
    public static class Impl5986 implements DemoService {public int type() {return 5986;} public void run() {System.out.println(5986);}}
    public static class Impl5987 implements DemoService {public int type() {return 5987;} public void run() {System.out.println(5987);}}
    public static class Impl5988 implements DemoService {public int type() {return 5988;} public void run() {System.out.println(5988);}}
    public static class Impl5989 implements DemoService {public int type() {return 5989;} public void run() {System.out.println(5989);}}
    public static class Impl5990 implements DemoService {public int type() {return 5990;} public void run() {System.out.println(5990);}}
    public static class Impl5991 implements DemoService {public int type() {return 5991;} public void run() {System.out.println(5991);}}
    public static class Impl5992 implements DemoService {public int type() {return 5992;} public void run() {System.out.println(5992);}}
    public static class Impl5993 implements DemoService {public int type() {return 5993;} public void run() {System.out.println(5993);}}
    public static class Impl5994 implements DemoService {public int type() {return 5994;} public void run() {System.out.println(5994);}}
    public static class Impl5995 implements DemoService {public int type() {return 5995;} public void run() {System.out.println(5995);}}
    public static class Impl5996 implements DemoService {public int type() {return 5996;} public void run() {System.out.println(5996);}}
    public static class Impl5997 implements DemoService {public int type() {return 5997;} public void run() {System.out.println(5997);}}
    public static class Impl5998 implements DemoService {public int type() {return 5998;} public void run() {System.out.println(5998);}}
    public static class Impl5999 implements DemoService {public int type() {return 5999;} public void run() {System.out.println(5999);}}
    public static class Impl6000 implements DemoService {public int type() {return 6000;} public void run() {System.out.println(6000);}}
    public static class Impl6001 implements DemoService {public int type() {return 6001;} public void run() {System.out.println(6001);}}
    public static class Impl6002 implements DemoService {public int type() {return 6002;} public void run() {System.out.println(6002);}}
    public static class Impl6003 implements DemoService {public int type() {return 6003;} public void run() {System.out.println(6003);}}
    public static class Impl6004 implements DemoService {public int type() {return 6004;} public void run() {System.out.println(6004);}}
    public static class Impl6005 implements DemoService {public int type() {return 6005;} public void run() {System.out.println(6005);}}
    public static class Impl6006 implements DemoService {public int type() {return 6006;} public void run() {System.out.println(6006);}}
    public static class Impl6007 implements DemoService {public int type() {return 6007;} public void run() {System.out.println(6007);}}
    public static class Impl6008 implements DemoService {public int type() {return 6008;} public void run() {System.out.println(6008);}}
    public static class Impl6009 implements DemoService {public int type() {return 6009;} public void run() {System.out.println(6009);}}
    public static class Impl6010 implements DemoService {public int type() {return 6010;} public void run() {System.out.println(6010);}}
    public static class Impl6011 implements DemoService {public int type() {return 6011;} public void run() {System.out.println(6011);}}
    public static class Impl6012 implements DemoService {public int type() {return 6012;} public void run() {System.out.println(6012);}}
    public static class Impl6013 implements DemoService {public int type() {return 6013;} public void run() {System.out.println(6013);}}
    public static class Impl6014 implements DemoService {public int type() {return 6014;} public void run() {System.out.println(6014);}}
    public static class Impl6015 implements DemoService {public int type() {return 6015;} public void run() {System.out.println(6015);}}
    public static class Impl6016 implements DemoService {public int type() {return 6016;} public void run() {System.out.println(6016);}}
    public static class Impl6017 implements DemoService {public int type() {return 6017;} public void run() {System.out.println(6017);}}
    public static class Impl6018 implements DemoService {public int type() {return 6018;} public void run() {System.out.println(6018);}}
    public static class Impl6019 implements DemoService {public int type() {return 6019;} public void run() {System.out.println(6019);}}
    public static class Impl6020 implements DemoService {public int type() {return 6020;} public void run() {System.out.println(6020);}}
    public static class Impl6021 implements DemoService {public int type() {return 6021;} public void run() {System.out.println(6021);}}
    public static class Impl6022 implements DemoService {public int type() {return 6022;} public void run() {System.out.println(6022);}}
    public static class Impl6023 implements DemoService {public int type() {return 6023;} public void run() {System.out.println(6023);}}
    public static class Impl6024 implements DemoService {public int type() {return 6024;} public void run() {System.out.println(6024);}}
    public static class Impl6025 implements DemoService {public int type() {return 6025;} public void run() {System.out.println(6025);}}
    public static class Impl6026 implements DemoService {public int type() {return 6026;} public void run() {System.out.println(6026);}}
    public static class Impl6027 implements DemoService {public int type() {return 6027;} public void run() {System.out.println(6027);}}
    public static class Impl6028 implements DemoService {public int type() {return 6028;} public void run() {System.out.println(6028);}}
    public static class Impl6029 implements DemoService {public int type() {return 6029;} public void run() {System.out.println(6029);}}
    public static class Impl6030 implements DemoService {public int type() {return 6030;} public void run() {System.out.println(6030);}}
    public static class Impl6031 implements DemoService {public int type() {return 6031;} public void run() {System.out.println(6031);}}
    public static class Impl6032 implements DemoService {public int type() {return 6032;} public void run() {System.out.println(6032);}}
    public static class Impl6033 implements DemoService {public int type() {return 6033;} public void run() {System.out.println(6033);}}
    public static class Impl6034 implements DemoService {public int type() {return 6034;} public void run() {System.out.println(6034);}}
    public static class Impl6035 implements DemoService {public int type() {return 6035;} public void run() {System.out.println(6035);}}
    public static class Impl6036 implements DemoService {public int type() {return 6036;} public void run() {System.out.println(6036);}}
    public static class Impl6037 implements DemoService {public int type() {return 6037;} public void run() {System.out.println(6037);}}
    public static class Impl6038 implements DemoService {public int type() {return 6038;} public void run() {System.out.println(6038);}}
    public static class Impl6039 implements DemoService {public int type() {return 6039;} public void run() {System.out.println(6039);}}
    public static class Impl6040 implements DemoService {public int type() {return 6040;} public void run() {System.out.println(6040);}}
    public static class Impl6041 implements DemoService {public int type() {return 6041;} public void run() {System.out.println(6041);}}
    public static class Impl6042 implements DemoService {public int type() {return 6042;} public void run() {System.out.println(6042);}}
    public static class Impl6043 implements DemoService {public int type() {return 6043;} public void run() {System.out.println(6043);}}
    public static class Impl6044 implements DemoService {public int type() {return 6044;} public void run() {System.out.println(6044);}}
    public static class Impl6045 implements DemoService {public int type() {return 6045;} public void run() {System.out.println(6045);}}
    public static class Impl6046 implements DemoService {public int type() {return 6046;} public void run() {System.out.println(6046);}}
    public static class Impl6047 implements DemoService {public int type() {return 6047;} public void run() {System.out.println(6047);}}
    public static class Impl6048 implements DemoService {public int type() {return 6048;} public void run() {System.out.println(6048);}}
    public static class Impl6049 implements DemoService {public int type() {return 6049;} public void run() {System.out.println(6049);}}
    public static class Impl6050 implements DemoService {public int type() {return 6050;} public void run() {System.out.println(6050);}}
    public static class Impl6051 implements DemoService {public int type() {return 6051;} public void run() {System.out.println(6051);}}
    public static class Impl6052 implements DemoService {public int type() {return 6052;} public void run() {System.out.println(6052);}}
    public static class Impl6053 implements DemoService {public int type() {return 6053;} public void run() {System.out.println(6053);}}
    public static class Impl6054 implements DemoService {public int type() {return 6054;} public void run() {System.out.println(6054);}}
    public static class Impl6055 implements DemoService {public int type() {return 6055;} public void run() {System.out.println(6055);}}
    public static class Impl6056 implements DemoService {public int type() {return 6056;} public void run() {System.out.println(6056);}}
    public static class Impl6057 implements DemoService {public int type() {return 6057;} public void run() {System.out.println(6057);}}
    public static class Impl6058 implements DemoService {public int type() {return 6058;} public void run() {System.out.println(6058);}}
    public static class Impl6059 implements DemoService {public int type() {return 6059;} public void run() {System.out.println(6059);}}
    public static class Impl6060 implements DemoService {public int type() {return 6060;} public void run() {System.out.println(6060);}}
    public static class Impl6061 implements DemoService {public int type() {return 6061;} public void run() {System.out.println(6061);}}
    public static class Impl6062 implements DemoService {public int type() {return 6062;} public void run() {System.out.println(6062);}}
    public static class Impl6063 implements DemoService {public int type() {return 6063;} public void run() {System.out.println(6063);}}
    public static class Impl6064 implements DemoService {public int type() {return 6064;} public void run() {System.out.println(6064);}}
    public static class Impl6065 implements DemoService {public int type() {return 6065;} public void run() {System.out.println(6065);}}
    public static class Impl6066 implements DemoService {public int type() {return 6066;} public void run() {System.out.println(6066);}}
    public static class Impl6067 implements DemoService {public int type() {return 6067;} public void run() {System.out.println(6067);}}
    public static class Impl6068 implements DemoService {public int type() {return 6068;} public void run() {System.out.println(6068);}}
    public static class Impl6069 implements DemoService {public int type() {return 6069;} public void run() {System.out.println(6069);}}
    public static class Impl6070 implements DemoService {public int type() {return 6070;} public void run() {System.out.println(6070);}}
    public static class Impl6071 implements DemoService {public int type() {return 6071;} public void run() {System.out.println(6071);}}
    public static class Impl6072 implements DemoService {public int type() {return 6072;} public void run() {System.out.println(6072);}}
    public static class Impl6073 implements DemoService {public int type() {return 6073;} public void run() {System.out.println(6073);}}
    public static class Impl6074 implements DemoService {public int type() {return 6074;} public void run() {System.out.println(6074);}}
    public static class Impl6075 implements DemoService {public int type() {return 6075;} public void run() {System.out.println(6075);}}
    public static class Impl6076 implements DemoService {public int type() {return 6076;} public void run() {System.out.println(6076);}}
    public static class Impl6077 implements DemoService {public int type() {return 6077;} public void run() {System.out.println(6077);}}
    public static class Impl6078 implements DemoService {public int type() {return 6078;} public void run() {System.out.println(6078);}}
    public static class Impl6079 implements DemoService {public int type() {return 6079;} public void run() {System.out.println(6079);}}
    public static class Impl6080 implements DemoService {public int type() {return 6080;} public void run() {System.out.println(6080);}}
    public static class Impl6081 implements DemoService {public int type() {return 6081;} public void run() {System.out.println(6081);}}
    public static class Impl6082 implements DemoService {public int type() {return 6082;} public void run() {System.out.println(6082);}}
    public static class Impl6083 implements DemoService {public int type() {return 6083;} public void run() {System.out.println(6083);}}
    public static class Impl6084 implements DemoService {public int type() {return 6084;} public void run() {System.out.println(6084);}}
    public static class Impl6085 implements DemoService {public int type() {return 6085;} public void run() {System.out.println(6085);}}
    public static class Impl6086 implements DemoService {public int type() {return 6086;} public void run() {System.out.println(6086);}}
    public static class Impl6087 implements DemoService {public int type() {return 6087;} public void run() {System.out.println(6087);}}
    public static class Impl6088 implements DemoService {public int type() {return 6088;} public void run() {System.out.println(6088);}}
    public static class Impl6089 implements DemoService {public int type() {return 6089;} public void run() {System.out.println(6089);}}
    public static class Impl6090 implements DemoService {public int type() {return 6090;} public void run() {System.out.println(6090);}}
    public static class Impl6091 implements DemoService {public int type() {return 6091;} public void run() {System.out.println(6091);}}
    public static class Impl6092 implements DemoService {public int type() {return 6092;} public void run() {System.out.println(6092);}}
    public static class Impl6093 implements DemoService {public int type() {return 6093;} public void run() {System.out.println(6093);}}
    public static class Impl6094 implements DemoService {public int type() {return 6094;} public void run() {System.out.println(6094);}}
    public static class Impl6095 implements DemoService {public int type() {return 6095;} public void run() {System.out.println(6095);}}
    public static class Impl6096 implements DemoService {public int type() {return 6096;} public void run() {System.out.println(6096);}}
    public static class Impl6097 implements DemoService {public int type() {return 6097;} public void run() {System.out.println(6097);}}
    public static class Impl6098 implements DemoService {public int type() {return 6098;} public void run() {System.out.println(6098);}}
    public static class Impl6099 implements DemoService {public int type() {return 6099;} public void run() {System.out.println(6099);}}
    public static class Impl6100 implements DemoService {public int type() {return 6100;} public void run() {System.out.println(6100);}}
    public static class Impl6101 implements DemoService {public int type() {return 6101;} public void run() {System.out.println(6101);}}
    public static class Impl6102 implements DemoService {public int type() {return 6102;} public void run() {System.out.println(6102);}}
    public static class Impl6103 implements DemoService {public int type() {return 6103;} public void run() {System.out.println(6103);}}
    public static class Impl6104 implements DemoService {public int type() {return 6104;} public void run() {System.out.println(6104);}}
    public static class Impl6105 implements DemoService {public int type() {return 6105;} public void run() {System.out.println(6105);}}
    public static class Impl6106 implements DemoService {public int type() {return 6106;} public void run() {System.out.println(6106);}}
    public static class Impl6107 implements DemoService {public int type() {return 6107;} public void run() {System.out.println(6107);}}
    public static class Impl6108 implements DemoService {public int type() {return 6108;} public void run() {System.out.println(6108);}}
    public static class Impl6109 implements DemoService {public int type() {return 6109;} public void run() {System.out.println(6109);}}
    public static class Impl6110 implements DemoService {public int type() {return 6110;} public void run() {System.out.println(6110);}}
    public static class Impl6111 implements DemoService {public int type() {return 6111;} public void run() {System.out.println(6111);}}
    public static class Impl6112 implements DemoService {public int type() {return 6112;} public void run() {System.out.println(6112);}}
    public static class Impl6113 implements DemoService {public int type() {return 6113;} public void run() {System.out.println(6113);}}
    public static class Impl6114 implements DemoService {public int type() {return 6114;} public void run() {System.out.println(6114);}}
    public static class Impl6115 implements DemoService {public int type() {return 6115;} public void run() {System.out.println(6115);}}
    public static class Impl6116 implements DemoService {public int type() {return 6116;} public void run() {System.out.println(6116);}}
    public static class Impl6117 implements DemoService {public int type() {return 6117;} public void run() {System.out.println(6117);}}
    public static class Impl6118 implements DemoService {public int type() {return 6118;} public void run() {System.out.println(6118);}}
    public static class Impl6119 implements DemoService {public int type() {return 6119;} public void run() {System.out.println(6119);}}
    public static class Impl6120 implements DemoService {public int type() {return 6120;} public void run() {System.out.println(6120);}}
    public static class Impl6121 implements DemoService {public int type() {return 6121;} public void run() {System.out.println(6121);}}
    public static class Impl6122 implements DemoService {public int type() {return 6122;} public void run() {System.out.println(6122);}}
    public static class Impl6123 implements DemoService {public int type() {return 6123;} public void run() {System.out.println(6123);}}
    public static class Impl6124 implements DemoService {public int type() {return 6124;} public void run() {System.out.println(6124);}}
    public static class Impl6125 implements DemoService {public int type() {return 6125;} public void run() {System.out.println(6125);}}
    public static class Impl6126 implements DemoService {public int type() {return 6126;} public void run() {System.out.println(6126);}}
    public static class Impl6127 implements DemoService {public int type() {return 6127;} public void run() {System.out.println(6127);}}
    public static class Impl6128 implements DemoService {public int type() {return 6128;} public void run() {System.out.println(6128);}}
    public static class Impl6129 implements DemoService {public int type() {return 6129;} public void run() {System.out.println(6129);}}
    public static class Impl6130 implements DemoService {public int type() {return 6130;} public void run() {System.out.println(6130);}}
    public static class Impl6131 implements DemoService {public int type() {return 6131;} public void run() {System.out.println(6131);}}
    public static class Impl6132 implements DemoService {public int type() {return 6132;} public void run() {System.out.println(6132);}}
    public static class Impl6133 implements DemoService {public int type() {return 6133;} public void run() {System.out.println(6133);}}
    public static class Impl6134 implements DemoService {public int type() {return 6134;} public void run() {System.out.println(6134);}}
    public static class Impl6135 implements DemoService {public int type() {return 6135;} public void run() {System.out.println(6135);}}
    public static class Impl6136 implements DemoService {public int type() {return 6136;} public void run() {System.out.println(6136);}}
    public static class Impl6137 implements DemoService {public int type() {return 6137;} public void run() {System.out.println(6137);}}
    public static class Impl6138 implements DemoService {public int type() {return 6138;} public void run() {System.out.println(6138);}}
    public static class Impl6139 implements DemoService {public int type() {return 6139;} public void run() {System.out.println(6139);}}
    public static class Impl6140 implements DemoService {public int type() {return 6140;} public void run() {System.out.println(6140);}}
    public static class Impl6141 implements DemoService {public int type() {return 6141;} public void run() {System.out.println(6141);}}
    public static class Impl6142 implements DemoService {public int type() {return 6142;} public void run() {System.out.println(6142);}}
    public static class Impl6143 implements DemoService {public int type() {return 6143;} public void run() {System.out.println(6143);}}
    public static class Impl6144 implements DemoService {public int type() {return 6144;} public void run() {System.out.println(6144);}}
    public static class Impl6145 implements DemoService {public int type() {return 6145;} public void run() {System.out.println(6145);}}
    public static class Impl6146 implements DemoService {public int type() {return 6146;} public void run() {System.out.println(6146);}}
    public static class Impl6147 implements DemoService {public int type() {return 6147;} public void run() {System.out.println(6147);}}
    public static class Impl6148 implements DemoService {public int type() {return 6148;} public void run() {System.out.println(6148);}}
    public static class Impl6149 implements DemoService {public int type() {return 6149;} public void run() {System.out.println(6149);}}
    public static class Impl6150 implements DemoService {public int type() {return 6150;} public void run() {System.out.println(6150);}}
    public static class Impl6151 implements DemoService {public int type() {return 6151;} public void run() {System.out.println(6151);}}
    public static class Impl6152 implements DemoService {public int type() {return 6152;} public void run() {System.out.println(6152);}}
    public static class Impl6153 implements DemoService {public int type() {return 6153;} public void run() {System.out.println(6153);}}
    public static class Impl6154 implements DemoService {public int type() {return 6154;} public void run() {System.out.println(6154);}}
    public static class Impl6155 implements DemoService {public int type() {return 6155;} public void run() {System.out.println(6155);}}
    public static class Impl6156 implements DemoService {public int type() {return 6156;} public void run() {System.out.println(6156);}}
    public static class Impl6157 implements DemoService {public int type() {return 6157;} public void run() {System.out.println(6157);}}
    public static class Impl6158 implements DemoService {public int type() {return 6158;} public void run() {System.out.println(6158);}}
    public static class Impl6159 implements DemoService {public int type() {return 6159;} public void run() {System.out.println(6159);}}
    public static class Impl6160 implements DemoService {public int type() {return 6160;} public void run() {System.out.println(6160);}}
    public static class Impl6161 implements DemoService {public int type() {return 6161;} public void run() {System.out.println(6161);}}
    public static class Impl6162 implements DemoService {public int type() {return 6162;} public void run() {System.out.println(6162);}}
    public static class Impl6163 implements DemoService {public int type() {return 6163;} public void run() {System.out.println(6163);}}
    public static class Impl6164 implements DemoService {public int type() {return 6164;} public void run() {System.out.println(6164);}}
    public static class Impl6165 implements DemoService {public int type() {return 6165;} public void run() {System.out.println(6165);}}
    public static class Impl6166 implements DemoService {public int type() {return 6166;} public void run() {System.out.println(6166);}}
    public static class Impl6167 implements DemoService {public int type() {return 6167;} public void run() {System.out.println(6167);}}
    public static class Impl6168 implements DemoService {public int type() {return 6168;} public void run() {System.out.println(6168);}}
    public static class Impl6169 implements DemoService {public int type() {return 6169;} public void run() {System.out.println(6169);}}
    public static class Impl6170 implements DemoService {public int type() {return 6170;} public void run() {System.out.println(6170);}}
    public static class Impl6171 implements DemoService {public int type() {return 6171;} public void run() {System.out.println(6171);}}
    public static class Impl6172 implements DemoService {public int type() {return 6172;} public void run() {System.out.println(6172);}}
    public static class Impl6173 implements DemoService {public int type() {return 6173;} public void run() {System.out.println(6173);}}
    public static class Impl6174 implements DemoService {public int type() {return 6174;} public void run() {System.out.println(6174);}}
    public static class Impl6175 implements DemoService {public int type() {return 6175;} public void run() {System.out.println(6175);}}
    public static class Impl6176 implements DemoService {public int type() {return 6176;} public void run() {System.out.println(6176);}}
    public static class Impl6177 implements DemoService {public int type() {return 6177;} public void run() {System.out.println(6177);}}
    public static class Impl6178 implements DemoService {public int type() {return 6178;} public void run() {System.out.println(6178);}}
    public static class Impl6179 implements DemoService {public int type() {return 6179;} public void run() {System.out.println(6179);}}
    public static class Impl6180 implements DemoService {public int type() {return 6180;} public void run() {System.out.println(6180);}}
    public static class Impl6181 implements DemoService {public int type() {return 6181;} public void run() {System.out.println(6181);}}
    public static class Impl6182 implements DemoService {public int type() {return 6182;} public void run() {System.out.println(6182);}}
    public static class Impl6183 implements DemoService {public int type() {return 6183;} public void run() {System.out.println(6183);}}
    public static class Impl6184 implements DemoService {public int type() {return 6184;} public void run() {System.out.println(6184);}}
    public static class Impl6185 implements DemoService {public int type() {return 6185;} public void run() {System.out.println(6185);}}
    public static class Impl6186 implements DemoService {public int type() {return 6186;} public void run() {System.out.println(6186);}}
    public static class Impl6187 implements DemoService {public int type() {return 6187;} public void run() {System.out.println(6187);}}
    public static class Impl6188 implements DemoService {public int type() {return 6188;} public void run() {System.out.println(6188);}}
    public static class Impl6189 implements DemoService {public int type() {return 6189;} public void run() {System.out.println(6189);}}
    public static class Impl6190 implements DemoService {public int type() {return 6190;} public void run() {System.out.println(6190);}}
    public static class Impl6191 implements DemoService {public int type() {return 6191;} public void run() {System.out.println(6191);}}
    public static class Impl6192 implements DemoService {public int type() {return 6192;} public void run() {System.out.println(6192);}}
    public static class Impl6193 implements DemoService {public int type() {return 6193;} public void run() {System.out.println(6193);}}
    public static class Impl6194 implements DemoService {public int type() {return 6194;} public void run() {System.out.println(6194);}}
    public static class Impl6195 implements DemoService {public int type() {return 6195;} public void run() {System.out.println(6195);}}
    public static class Impl6196 implements DemoService {public int type() {return 6196;} public void run() {System.out.println(6196);}}
    public static class Impl6197 implements DemoService {public int type() {return 6197;} public void run() {System.out.println(6197);}}
    public static class Impl6198 implements DemoService {public int type() {return 6198;} public void run() {System.out.println(6198);}}
    public static class Impl6199 implements DemoService {public int type() {return 6199;} public void run() {System.out.println(6199);}}
    public static class Impl6200 implements DemoService {public int type() {return 6200;} public void run() {System.out.println(6200);}}
    public static class Impl6201 implements DemoService {public int type() {return 6201;} public void run() {System.out.println(6201);}}
    public static class Impl6202 implements DemoService {public int type() {return 6202;} public void run() {System.out.println(6202);}}
    public static class Impl6203 implements DemoService {public int type() {return 6203;} public void run() {System.out.println(6203);}}
    public static class Impl6204 implements DemoService {public int type() {return 6204;} public void run() {System.out.println(6204);}}
    public static class Impl6205 implements DemoService {public int type() {return 6205;} public void run() {System.out.println(6205);}}
    public static class Impl6206 implements DemoService {public int type() {return 6206;} public void run() {System.out.println(6206);}}
    public static class Impl6207 implements DemoService {public int type() {return 6207;} public void run() {System.out.println(6207);}}
    public static class Impl6208 implements DemoService {public int type() {return 6208;} public void run() {System.out.println(6208);}}
    public static class Impl6209 implements DemoService {public int type() {return 6209;} public void run() {System.out.println(6209);}}
    public static class Impl6210 implements DemoService {public int type() {return 6210;} public void run() {System.out.println(6210);}}
    public static class Impl6211 implements DemoService {public int type() {return 6211;} public void run() {System.out.println(6211);}}
    public static class Impl6212 implements DemoService {public int type() {return 6212;} public void run() {System.out.println(6212);}}
    public static class Impl6213 implements DemoService {public int type() {return 6213;} public void run() {System.out.println(6213);}}
    public static class Impl6214 implements DemoService {public int type() {return 6214;} public void run() {System.out.println(6214);}}
    public static class Impl6215 implements DemoService {public int type() {return 6215;} public void run() {System.out.println(6215);}}
    public static class Impl6216 implements DemoService {public int type() {return 6216;} public void run() {System.out.println(6216);}}
    public static class Impl6217 implements DemoService {public int type() {return 6217;} public void run() {System.out.println(6217);}}
    public static class Impl6218 implements DemoService {public int type() {return 6218;} public void run() {System.out.println(6218);}}
    public static class Impl6219 implements DemoService {public int type() {return 6219;} public void run() {System.out.println(6219);}}
    public static class Impl6220 implements DemoService {public int type() {return 6220;} public void run() {System.out.println(6220);}}
    public static class Impl6221 implements DemoService {public int type() {return 6221;} public void run() {System.out.println(6221);}}
    public static class Impl6222 implements DemoService {public int type() {return 6222;} public void run() {System.out.println(6222);}}
    public static class Impl6223 implements DemoService {public int type() {return 6223;} public void run() {System.out.println(6223);}}
    public static class Impl6224 implements DemoService {public int type() {return 6224;} public void run() {System.out.println(6224);}}
    public static class Impl6225 implements DemoService {public int type() {return 6225;} public void run() {System.out.println(6225);}}
    public static class Impl6226 implements DemoService {public int type() {return 6226;} public void run() {System.out.println(6226);}}
    public static class Impl6227 implements DemoService {public int type() {return 6227;} public void run() {System.out.println(6227);}}
    public static class Impl6228 implements DemoService {public int type() {return 6228;} public void run() {System.out.println(6228);}}
    public static class Impl6229 implements DemoService {public int type() {return 6229;} public void run() {System.out.println(6229);}}
    public static class Impl6230 implements DemoService {public int type() {return 6230;} public void run() {System.out.println(6230);}}
    public static class Impl6231 implements DemoService {public int type() {return 6231;} public void run() {System.out.println(6231);}}
    public static class Impl6232 implements DemoService {public int type() {return 6232;} public void run() {System.out.println(6232);}}
    public static class Impl6233 implements DemoService {public int type() {return 6233;} public void run() {System.out.println(6233);}}
    public static class Impl6234 implements DemoService {public int type() {return 6234;} public void run() {System.out.println(6234);}}
    public static class Impl6235 implements DemoService {public int type() {return 6235;} public void run() {System.out.println(6235);}}
    public static class Impl6236 implements DemoService {public int type() {return 6236;} public void run() {System.out.println(6236);}}
    public static class Impl6237 implements DemoService {public int type() {return 6237;} public void run() {System.out.println(6237);}}
    public static class Impl6238 implements DemoService {public int type() {return 6238;} public void run() {System.out.println(6238);}}
    public static class Impl6239 implements DemoService {public int type() {return 6239;} public void run() {System.out.println(6239);}}
    public static class Impl6240 implements DemoService {public int type() {return 6240;} public void run() {System.out.println(6240);}}
    public static class Impl6241 implements DemoService {public int type() {return 6241;} public void run() {System.out.println(6241);}}
    public static class Impl6242 implements DemoService {public int type() {return 6242;} public void run() {System.out.println(6242);}}
    public static class Impl6243 implements DemoService {public int type() {return 6243;} public void run() {System.out.println(6243);}}
    public static class Impl6244 implements DemoService {public int type() {return 6244;} public void run() {System.out.println(6244);}}
    public static class Impl6245 implements DemoService {public int type() {return 6245;} public void run() {System.out.println(6245);}}
    public static class Impl6246 implements DemoService {public int type() {return 6246;} public void run() {System.out.println(6246);}}
    public static class Impl6247 implements DemoService {public int type() {return 6247;} public void run() {System.out.println(6247);}}
    public static class Impl6248 implements DemoService {public int type() {return 6248;} public void run() {System.out.println(6248);}}
    public static class Impl6249 implements DemoService {public int type() {return 6249;} public void run() {System.out.println(6249);}}
    public static class Impl6250 implements DemoService {public int type() {return 6250;} public void run() {System.out.println(6250);}}
    public static class Impl6251 implements DemoService {public int type() {return 6251;} public void run() {System.out.println(6251);}}
    public static class Impl6252 implements DemoService {public int type() {return 6252;} public void run() {System.out.println(6252);}}
    public static class Impl6253 implements DemoService {public int type() {return 6253;} public void run() {System.out.println(6253);}}
    public static class Impl6254 implements DemoService {public int type() {return 6254;} public void run() {System.out.println(6254);}}
    public static class Impl6255 implements DemoService {public int type() {return 6255;} public void run() {System.out.println(6255);}}
    public static class Impl6256 implements DemoService {public int type() {return 6256;} public void run() {System.out.println(6256);}}
    public static class Impl6257 implements DemoService {public int type() {return 6257;} public void run() {System.out.println(6257);}}
    public static class Impl6258 implements DemoService {public int type() {return 6258;} public void run() {System.out.println(6258);}}
    public static class Impl6259 implements DemoService {public int type() {return 6259;} public void run() {System.out.println(6259);}}
    public static class Impl6260 implements DemoService {public int type() {return 6260;} public void run() {System.out.println(6260);}}
    public static class Impl6261 implements DemoService {public int type() {return 6261;} public void run() {System.out.println(6261);}}
    public static class Impl6262 implements DemoService {public int type() {return 6262;} public void run() {System.out.println(6262);}}
    public static class Impl6263 implements DemoService {public int type() {return 6263;} public void run() {System.out.println(6263);}}
    public static class Impl6264 implements DemoService {public int type() {return 6264;} public void run() {System.out.println(6264);}}
    public static class Impl6265 implements DemoService {public int type() {return 6265;} public void run() {System.out.println(6265);}}
    public static class Impl6266 implements DemoService {public int type() {return 6266;} public void run() {System.out.println(6266);}}
    public static class Impl6267 implements DemoService {public int type() {return 6267;} public void run() {System.out.println(6267);}}
    public static class Impl6268 implements DemoService {public int type() {return 6268;} public void run() {System.out.println(6268);}}
    public static class Impl6269 implements DemoService {public int type() {return 6269;} public void run() {System.out.println(6269);}}
    public static class Impl6270 implements DemoService {public int type() {return 6270;} public void run() {System.out.println(6270);}}
    public static class Impl6271 implements DemoService {public int type() {return 6271;} public void run() {System.out.println(6271);}}
    public static class Impl6272 implements DemoService {public int type() {return 6272;} public void run() {System.out.println(6272);}}
    public static class Impl6273 implements DemoService {public int type() {return 6273;} public void run() {System.out.println(6273);}}
    public static class Impl6274 implements DemoService {public int type() {return 6274;} public void run() {System.out.println(6274);}}
    public static class Impl6275 implements DemoService {public int type() {return 6275;} public void run() {System.out.println(6275);}}
    public static class Impl6276 implements DemoService {public int type() {return 6276;} public void run() {System.out.println(6276);}}
    public static class Impl6277 implements DemoService {public int type() {return 6277;} public void run() {System.out.println(6277);}}
    public static class Impl6278 implements DemoService {public int type() {return 6278;} public void run() {System.out.println(6278);}}
    public static class Impl6279 implements DemoService {public int type() {return 6279;} public void run() {System.out.println(6279);}}
    public static class Impl6280 implements DemoService {public int type() {return 6280;} public void run() {System.out.println(6280);}}
    public static class Impl6281 implements DemoService {public int type() {return 6281;} public void run() {System.out.println(6281);}}
    public static class Impl6282 implements DemoService {public int type() {return 6282;} public void run() {System.out.println(6282);}}
    public static class Impl6283 implements DemoService {public int type() {return 6283;} public void run() {System.out.println(6283);}}
    public static class Impl6284 implements DemoService {public int type() {return 6284;} public void run() {System.out.println(6284);}}
    public static class Impl6285 implements DemoService {public int type() {return 6285;} public void run() {System.out.println(6285);}}
    public static class Impl6286 implements DemoService {public int type() {return 6286;} public void run() {System.out.println(6286);}}
    public static class Impl6287 implements DemoService {public int type() {return 6287;} public void run() {System.out.println(6287);}}
    public static class Impl6288 implements DemoService {public int type() {return 6288;} public void run() {System.out.println(6288);}}
    public static class Impl6289 implements DemoService {public int type() {return 6289;} public void run() {System.out.println(6289);}}
    public static class Impl6290 implements DemoService {public int type() {return 6290;} public void run() {System.out.println(6290);}}
    public static class Impl6291 implements DemoService {public int type() {return 6291;} public void run() {System.out.println(6291);}}
    public static class Impl6292 implements DemoService {public int type() {return 6292;} public void run() {System.out.println(6292);}}
    public static class Impl6293 implements DemoService {public int type() {return 6293;} public void run() {System.out.println(6293);}}
    public static class Impl6294 implements DemoService {public int type() {return 6294;} public void run() {System.out.println(6294);}}
    public static class Impl6295 implements DemoService {public int type() {return 6295;} public void run() {System.out.println(6295);}}
    public static class Impl6296 implements DemoService {public int type() {return 6296;} public void run() {System.out.println(6296);}}
    public static class Impl6297 implements DemoService {public int type() {return 6297;} public void run() {System.out.println(6297);}}
    public static class Impl6298 implements DemoService {public int type() {return 6298;} public void run() {System.out.println(6298);}}
    public static class Impl6299 implements DemoService {public int type() {return 6299;} public void run() {System.out.println(6299);}}
    public static class Impl6300 implements DemoService {public int type() {return 6300;} public void run() {System.out.println(6300);}}
    public static class Impl6301 implements DemoService {public int type() {return 6301;} public void run() {System.out.println(6301);}}
    public static class Impl6302 implements DemoService {public int type() {return 6302;} public void run() {System.out.println(6302);}}
    public static class Impl6303 implements DemoService {public int type() {return 6303;} public void run() {System.out.println(6303);}}
    public static class Impl6304 implements DemoService {public int type() {return 6304;} public void run() {System.out.println(6304);}}
    public static class Impl6305 implements DemoService {public int type() {return 6305;} public void run() {System.out.println(6305);}}
    public static class Impl6306 implements DemoService {public int type() {return 6306;} public void run() {System.out.println(6306);}}
    public static class Impl6307 implements DemoService {public int type() {return 6307;} public void run() {System.out.println(6307);}}
    public static class Impl6308 implements DemoService {public int type() {return 6308;} public void run() {System.out.println(6308);}}
    public static class Impl6309 implements DemoService {public int type() {return 6309;} public void run() {System.out.println(6309);}}
    public static class Impl6310 implements DemoService {public int type() {return 6310;} public void run() {System.out.println(6310);}}
    public static class Impl6311 implements DemoService {public int type() {return 6311;} public void run() {System.out.println(6311);}}
    public static class Impl6312 implements DemoService {public int type() {return 6312;} public void run() {System.out.println(6312);}}
    public static class Impl6313 implements DemoService {public int type() {return 6313;} public void run() {System.out.println(6313);}}
    public static class Impl6314 implements DemoService {public int type() {return 6314;} public void run() {System.out.println(6314);}}
    public static class Impl6315 implements DemoService {public int type() {return 6315;} public void run() {System.out.println(6315);}}
    public static class Impl6316 implements DemoService {public int type() {return 6316;} public void run() {System.out.println(6316);}}
    public static class Impl6317 implements DemoService {public int type() {return 6317;} public void run() {System.out.println(6317);}}
    public static class Impl6318 implements DemoService {public int type() {return 6318;} public void run() {System.out.println(6318);}}
    public static class Impl6319 implements DemoService {public int type() {return 6319;} public void run() {System.out.println(6319);}}
    public static class Impl6320 implements DemoService {public int type() {return 6320;} public void run() {System.out.println(6320);}}
    public static class Impl6321 implements DemoService {public int type() {return 6321;} public void run() {System.out.println(6321);}}
    public static class Impl6322 implements DemoService {public int type() {return 6322;} public void run() {System.out.println(6322);}}
    public static class Impl6323 implements DemoService {public int type() {return 6323;} public void run() {System.out.println(6323);}}
    public static class Impl6324 implements DemoService {public int type() {return 6324;} public void run() {System.out.println(6324);}}
    public static class Impl6325 implements DemoService {public int type() {return 6325;} public void run() {System.out.println(6325);}}
    public static class Impl6326 implements DemoService {public int type() {return 6326;} public void run() {System.out.println(6326);}}
    public static class Impl6327 implements DemoService {public int type() {return 6327;} public void run() {System.out.println(6327);}}
    public static class Impl6328 implements DemoService {public int type() {return 6328;} public void run() {System.out.println(6328);}}
    public static class Impl6329 implements DemoService {public int type() {return 6329;} public void run() {System.out.println(6329);}}
    public static class Impl6330 implements DemoService {public int type() {return 6330;} public void run() {System.out.println(6330);}}
    public static class Impl6331 implements DemoService {public int type() {return 6331;} public void run() {System.out.println(6331);}}
    public static class Impl6332 implements DemoService {public int type() {return 6332;} public void run() {System.out.println(6332);}}
    public static class Impl6333 implements DemoService {public int type() {return 6333;} public void run() {System.out.println(6333);}}
    public static class Impl6334 implements DemoService {public int type() {return 6334;} public void run() {System.out.println(6334);}}
    public static class Impl6335 implements DemoService {public int type() {return 6335;} public void run() {System.out.println(6335);}}
    public static class Impl6336 implements DemoService {public int type() {return 6336;} public void run() {System.out.println(6336);}}
    public static class Impl6337 implements DemoService {public int type() {return 6337;} public void run() {System.out.println(6337);}}
    public static class Impl6338 implements DemoService {public int type() {return 6338;} public void run() {System.out.println(6338);}}
    public static class Impl6339 implements DemoService {public int type() {return 6339;} public void run() {System.out.println(6339);}}
    public static class Impl6340 implements DemoService {public int type() {return 6340;} public void run() {System.out.println(6340);}}
    public static class Impl6341 implements DemoService {public int type() {return 6341;} public void run() {System.out.println(6341);}}
    public static class Impl6342 implements DemoService {public int type() {return 6342;} public void run() {System.out.println(6342);}}
    public static class Impl6343 implements DemoService {public int type() {return 6343;} public void run() {System.out.println(6343);}}
    public static class Impl6344 implements DemoService {public int type() {return 6344;} public void run() {System.out.println(6344);}}
    public static class Impl6345 implements DemoService {public int type() {return 6345;} public void run() {System.out.println(6345);}}
    public static class Impl6346 implements DemoService {public int type() {return 6346;} public void run() {System.out.println(6346);}}
    public static class Impl6347 implements DemoService {public int type() {return 6347;} public void run() {System.out.println(6347);}}
    public static class Impl6348 implements DemoService {public int type() {return 6348;} public void run() {System.out.println(6348);}}
    public static class Impl6349 implements DemoService {public int type() {return 6349;} public void run() {System.out.println(6349);}}
    public static class Impl6350 implements DemoService {public int type() {return 6350;} public void run() {System.out.println(6350);}}
    public static class Impl6351 implements DemoService {public int type() {return 6351;} public void run() {System.out.println(6351);}}
    public static class Impl6352 implements DemoService {public int type() {return 6352;} public void run() {System.out.println(6352);}}
    public static class Impl6353 implements DemoService {public int type() {return 6353;} public void run() {System.out.println(6353);}}
    public static class Impl6354 implements DemoService {public int type() {return 6354;} public void run() {System.out.println(6354);}}
    public static class Impl6355 implements DemoService {public int type() {return 6355;} public void run() {System.out.println(6355);}}
    public static class Impl6356 implements DemoService {public int type() {return 6356;} public void run() {System.out.println(6356);}}
    public static class Impl6357 implements DemoService {public int type() {return 6357;} public void run() {System.out.println(6357);}}
    public static class Impl6358 implements DemoService {public int type() {return 6358;} public void run() {System.out.println(6358);}}
    public static class Impl6359 implements DemoService {public int type() {return 6359;} public void run() {System.out.println(6359);}}
    public static class Impl6360 implements DemoService {public int type() {return 6360;} public void run() {System.out.println(6360);}}
    public static class Impl6361 implements DemoService {public int type() {return 6361;} public void run() {System.out.println(6361);}}
    public static class Impl6362 implements DemoService {public int type() {return 6362;} public void run() {System.out.println(6362);}}
    public static class Impl6363 implements DemoService {public int type() {return 6363;} public void run() {System.out.println(6363);}}
    public static class Impl6364 implements DemoService {public int type() {return 6364;} public void run() {System.out.println(6364);}}
    public static class Impl6365 implements DemoService {public int type() {return 6365;} public void run() {System.out.println(6365);}}
    public static class Impl6366 implements DemoService {public int type() {return 6366;} public void run() {System.out.println(6366);}}
    public static class Impl6367 implements DemoService {public int type() {return 6367;} public void run() {System.out.println(6367);}}
    public static class Impl6368 implements DemoService {public int type() {return 6368;} public void run() {System.out.println(6368);}}
    public static class Impl6369 implements DemoService {public int type() {return 6369;} public void run() {System.out.println(6369);}}
    public static class Impl6370 implements DemoService {public int type() {return 6370;} public void run() {System.out.println(6370);}}
    public static class Impl6371 implements DemoService {public int type() {return 6371;} public void run() {System.out.println(6371);}}
    public static class Impl6372 implements DemoService {public int type() {return 6372;} public void run() {System.out.println(6372);}}
    public static class Impl6373 implements DemoService {public int type() {return 6373;} public void run() {System.out.println(6373);}}
    public static class Impl6374 implements DemoService {public int type() {return 6374;} public void run() {System.out.println(6374);}}
    public static class Impl6375 implements DemoService {public int type() {return 6375;} public void run() {System.out.println(6375);}}
    public static class Impl6376 implements DemoService {public int type() {return 6376;} public void run() {System.out.println(6376);}}
    public static class Impl6377 implements DemoService {public int type() {return 6377;} public void run() {System.out.println(6377);}}
    public static class Impl6378 implements DemoService {public int type() {return 6378;} public void run() {System.out.println(6378);}}
    public static class Impl6379 implements DemoService {public int type() {return 6379;} public void run() {System.out.println(6379);}}
    public static class Impl6380 implements DemoService {public int type() {return 6380;} public void run() {System.out.println(6380);}}
    public static class Impl6381 implements DemoService {public int type() {return 6381;} public void run() {System.out.println(6381);}}
    public static class Impl6382 implements DemoService {public int type() {return 6382;} public void run() {System.out.println(6382);}}
    public static class Impl6383 implements DemoService {public int type() {return 6383;} public void run() {System.out.println(6383);}}
    public static class Impl6384 implements DemoService {public int type() {return 6384;} public void run() {System.out.println(6384);}}
    public static class Impl6385 implements DemoService {public int type() {return 6385;} public void run() {System.out.println(6385);}}
    public static class Impl6386 implements DemoService {public int type() {return 6386;} public void run() {System.out.println(6386);}}
    public static class Impl6387 implements DemoService {public int type() {return 6387;} public void run() {System.out.println(6387);}}
    public static class Impl6388 implements DemoService {public int type() {return 6388;} public void run() {System.out.println(6388);}}
    public static class Impl6389 implements DemoService {public int type() {return 6389;} public void run() {System.out.println(6389);}}
    public static class Impl6390 implements DemoService {public int type() {return 6390;} public void run() {System.out.println(6390);}}
    public static class Impl6391 implements DemoService {public int type() {return 6391;} public void run() {System.out.println(6391);}}
    public static class Impl6392 implements DemoService {public int type() {return 6392;} public void run() {System.out.println(6392);}}
    public static class Impl6393 implements DemoService {public int type() {return 6393;} public void run() {System.out.println(6393);}}
    public static class Impl6394 implements DemoService {public int type() {return 6394;} public void run() {System.out.println(6394);}}
    public static class Impl6395 implements DemoService {public int type() {return 6395;} public void run() {System.out.println(6395);}}
    public static class Impl6396 implements DemoService {public int type() {return 6396;} public void run() {System.out.println(6396);}}
    public static class Impl6397 implements DemoService {public int type() {return 6397;} public void run() {System.out.println(6397);}}
    public static class Impl6398 implements DemoService {public int type() {return 6398;} public void run() {System.out.println(6398);}}
    public static class Impl6399 implements DemoService {public int type() {return 6399;} public void run() {System.out.println(6399);}}
    public static class Impl6400 implements DemoService {public int type() {return 6400;} public void run() {System.out.println(6400);}}
    public static class Impl6401 implements DemoService {public int type() {return 6401;} public void run() {System.out.println(6401);}}
    public static class Impl6402 implements DemoService {public int type() {return 6402;} public void run() {System.out.println(6402);}}
    public static class Impl6403 implements DemoService {public int type() {return 6403;} public void run() {System.out.println(6403);}}
    public static class Impl6404 implements DemoService {public int type() {return 6404;} public void run() {System.out.println(6404);}}
    public static class Impl6405 implements DemoService {public int type() {return 6405;} public void run() {System.out.println(6405);}}
    public static class Impl6406 implements DemoService {public int type() {return 6406;} public void run() {System.out.println(6406);}}
    public static class Impl6407 implements DemoService {public int type() {return 6407;} public void run() {System.out.println(6407);}}
    public static class Impl6408 implements DemoService {public int type() {return 6408;} public void run() {System.out.println(6408);}}
    public static class Impl6409 implements DemoService {public int type() {return 6409;} public void run() {System.out.println(6409);}}
    public static class Impl6410 implements DemoService {public int type() {return 6410;} public void run() {System.out.println(6410);}}
    public static class Impl6411 implements DemoService {public int type() {return 6411;} public void run() {System.out.println(6411);}}
    public static class Impl6412 implements DemoService {public int type() {return 6412;} public void run() {System.out.println(6412);}}
    public static class Impl6413 implements DemoService {public int type() {return 6413;} public void run() {System.out.println(6413);}}
    public static class Impl6414 implements DemoService {public int type() {return 6414;} public void run() {System.out.println(6414);}}
    public static class Impl6415 implements DemoService {public int type() {return 6415;} public void run() {System.out.println(6415);}}
    public static class Impl6416 implements DemoService {public int type() {return 6416;} public void run() {System.out.println(6416);}}
    public static class Impl6417 implements DemoService {public int type() {return 6417;} public void run() {System.out.println(6417);}}
    public static class Impl6418 implements DemoService {public int type() {return 6418;} public void run() {System.out.println(6418);}}
    public static class Impl6419 implements DemoService {public int type() {return 6419;} public void run() {System.out.println(6419);}}
    public static class Impl6420 implements DemoService {public int type() {return 6420;} public void run() {System.out.println(6420);}}
    public static class Impl6421 implements DemoService {public int type() {return 6421;} public void run() {System.out.println(6421);}}
    public static class Impl6422 implements DemoService {public int type() {return 6422;} public void run() {System.out.println(6422);}}
    public static class Impl6423 implements DemoService {public int type() {return 6423;} public void run() {System.out.println(6423);}}
    public static class Impl6424 implements DemoService {public int type() {return 6424;} public void run() {System.out.println(6424);}}
    public static class Impl6425 implements DemoService {public int type() {return 6425;} public void run() {System.out.println(6425);}}
    public static class Impl6426 implements DemoService {public int type() {return 6426;} public void run() {System.out.println(6426);}}
    public static class Impl6427 implements DemoService {public int type() {return 6427;} public void run() {System.out.println(6427);}}
    public static class Impl6428 implements DemoService {public int type() {return 6428;} public void run() {System.out.println(6428);}}
    public static class Impl6429 implements DemoService {public int type() {return 6429;} public void run() {System.out.println(6429);}}
    public static class Impl6430 implements DemoService {public int type() {return 6430;} public void run() {System.out.println(6430);}}
    public static class Impl6431 implements DemoService {public int type() {return 6431;} public void run() {System.out.println(6431);}}
    public static class Impl6432 implements DemoService {public int type() {return 6432;} public void run() {System.out.println(6432);}}
    public static class Impl6433 implements DemoService {public int type() {return 6433;} public void run() {System.out.println(6433);}}
    public static class Impl6434 implements DemoService {public int type() {return 6434;} public void run() {System.out.println(6434);}}
    public static class Impl6435 implements DemoService {public int type() {return 6435;} public void run() {System.out.println(6435);}}
    public static class Impl6436 implements DemoService {public int type() {return 6436;} public void run() {System.out.println(6436);}}
    public static class Impl6437 implements DemoService {public int type() {return 6437;} public void run() {System.out.println(6437);}}
    public static class Impl6438 implements DemoService {public int type() {return 6438;} public void run() {System.out.println(6438);}}
    public static class Impl6439 implements DemoService {public int type() {return 6439;} public void run() {System.out.println(6439);}}
    public static class Impl6440 implements DemoService {public int type() {return 6440;} public void run() {System.out.println(6440);}}
    public static class Impl6441 implements DemoService {public int type() {return 6441;} public void run() {System.out.println(6441);}}
    public static class Impl6442 implements DemoService {public int type() {return 6442;} public void run() {System.out.println(6442);}}
    public static class Impl6443 implements DemoService {public int type() {return 6443;} public void run() {System.out.println(6443);}}
    public static class Impl6444 implements DemoService {public int type() {return 6444;} public void run() {System.out.println(6444);}}
    public static class Impl6445 implements DemoService {public int type() {return 6445;} public void run() {System.out.println(6445);}}
    public static class Impl6446 implements DemoService {public int type() {return 6446;} public void run() {System.out.println(6446);}}
    public static class Impl6447 implements DemoService {public int type() {return 6447;} public void run() {System.out.println(6447);}}
    public static class Impl6448 implements DemoService {public int type() {return 6448;} public void run() {System.out.println(6448);}}
    public static class Impl6449 implements DemoService {public int type() {return 6449;} public void run() {System.out.println(6449);}}
    public static class Impl6450 implements DemoService {public int type() {return 6450;} public void run() {System.out.println(6450);}}
    public static class Impl6451 implements DemoService {public int type() {return 6451;} public void run() {System.out.println(6451);}}
    public static class Impl6452 implements DemoService {public int type() {return 6452;} public void run() {System.out.println(6452);}}
    public static class Impl6453 implements DemoService {public int type() {return 6453;} public void run() {System.out.println(6453);}}
    public static class Impl6454 implements DemoService {public int type() {return 6454;} public void run() {System.out.println(6454);}}
    public static class Impl6455 implements DemoService {public int type() {return 6455;} public void run() {System.out.println(6455);}}
    public static class Impl6456 implements DemoService {public int type() {return 6456;} public void run() {System.out.println(6456);}}
    public static class Impl6457 implements DemoService {public int type() {return 6457;} public void run() {System.out.println(6457);}}
    public static class Impl6458 implements DemoService {public int type() {return 6458;} public void run() {System.out.println(6458);}}
    public static class Impl6459 implements DemoService {public int type() {return 6459;} public void run() {System.out.println(6459);}}
    public static class Impl6460 implements DemoService {public int type() {return 6460;} public void run() {System.out.println(6460);}}
    public static class Impl6461 implements DemoService {public int type() {return 6461;} public void run() {System.out.println(6461);}}
    public static class Impl6462 implements DemoService {public int type() {return 6462;} public void run() {System.out.println(6462);}}
    public static class Impl6463 implements DemoService {public int type() {return 6463;} public void run() {System.out.println(6463);}}
    public static class Impl6464 implements DemoService {public int type() {return 6464;} public void run() {System.out.println(6464);}}
    public static class Impl6465 implements DemoService {public int type() {return 6465;} public void run() {System.out.println(6465);}}
    public static class Impl6466 implements DemoService {public int type() {return 6466;} public void run() {System.out.println(6466);}}
    public static class Impl6467 implements DemoService {public int type() {return 6467;} public void run() {System.out.println(6467);}}
    public static class Impl6468 implements DemoService {public int type() {return 6468;} public void run() {System.out.println(6468);}}
    public static class Impl6469 implements DemoService {public int type() {return 6469;} public void run() {System.out.println(6469);}}
    public static class Impl6470 implements DemoService {public int type() {return 6470;} public void run() {System.out.println(6470);}}
    public static class Impl6471 implements DemoService {public int type() {return 6471;} public void run() {System.out.println(6471);}}
    public static class Impl6472 implements DemoService {public int type() {return 6472;} public void run() {System.out.println(6472);}}
    public static class Impl6473 implements DemoService {public int type() {return 6473;} public void run() {System.out.println(6473);}}
    public static class Impl6474 implements DemoService {public int type() {return 6474;} public void run() {System.out.println(6474);}}
    public static class Impl6475 implements DemoService {public int type() {return 6475;} public void run() {System.out.println(6475);}}
    public static class Impl6476 implements DemoService {public int type() {return 6476;} public void run() {System.out.println(6476);}}
    public static class Impl6477 implements DemoService {public int type() {return 6477;} public void run() {System.out.println(6477);}}
    public static class Impl6478 implements DemoService {public int type() {return 6478;} public void run() {System.out.println(6478);}}
    public static class Impl6479 implements DemoService {public int type() {return 6479;} public void run() {System.out.println(6479);}}
    public static class Impl6480 implements DemoService {public int type() {return 6480;} public void run() {System.out.println(6480);}}
    public static class Impl6481 implements DemoService {public int type() {return 6481;} public void run() {System.out.println(6481);}}
    public static class Impl6482 implements DemoService {public int type() {return 6482;} public void run() {System.out.println(6482);}}
    public static class Impl6483 implements DemoService {public int type() {return 6483;} public void run() {System.out.println(6483);}}
    public static class Impl6484 implements DemoService {public int type() {return 6484;} public void run() {System.out.println(6484);}}
    public static class Impl6485 implements DemoService {public int type() {return 6485;} public void run() {System.out.println(6485);}}
    public static class Impl6486 implements DemoService {public int type() {return 6486;} public void run() {System.out.println(6486);}}
    public static class Impl6487 implements DemoService {public int type() {return 6487;} public void run() {System.out.println(6487);}}
    public static class Impl6488 implements DemoService {public int type() {return 6488;} public void run() {System.out.println(6488);}}
    public static class Impl6489 implements DemoService {public int type() {return 6489;} public void run() {System.out.println(6489);}}
    public static class Impl6490 implements DemoService {public int type() {return 6490;} public void run() {System.out.println(6490);}}
    public static class Impl6491 implements DemoService {public int type() {return 6491;} public void run() {System.out.println(6491);}}
    public static class Impl6492 implements DemoService {public int type() {return 6492;} public void run() {System.out.println(6492);}}
    public static class Impl6493 implements DemoService {public int type() {return 6493;} public void run() {System.out.println(6493);}}
    public static class Impl6494 implements DemoService {public int type() {return 6494;} public void run() {System.out.println(6494);}}
    public static class Impl6495 implements DemoService {public int type() {return 6495;} public void run() {System.out.println(6495);}}
    public static class Impl6496 implements DemoService {public int type() {return 6496;} public void run() {System.out.println(6496);}}
    public static class Impl6497 implements DemoService {public int type() {return 6497;} public void run() {System.out.println(6497);}}
    public static class Impl6498 implements DemoService {public int type() {return 6498;} public void run() {System.out.println(6498);}}
    public static class Impl6499 implements DemoService {public int type() {return 6499;} public void run() {System.out.println(6499);}}
    public static class Impl6500 implements DemoService {public int type() {return 6500;} public void run() {System.out.println(6500);}}
    public static class Impl6501 implements DemoService {public int type() {return 6501;} public void run() {System.out.println(6501);}}
    public static class Impl6502 implements DemoService {public int type() {return 6502;} public void run() {System.out.println(6502);}}
    public static class Impl6503 implements DemoService {public int type() {return 6503;} public void run() {System.out.println(6503);}}
    public static class Impl6504 implements DemoService {public int type() {return 6504;} public void run() {System.out.println(6504);}}
    public static class Impl6505 implements DemoService {public int type() {return 6505;} public void run() {System.out.println(6505);}}
    public static class Impl6506 implements DemoService {public int type() {return 6506;} public void run() {System.out.println(6506);}}
    public static class Impl6507 implements DemoService {public int type() {return 6507;} public void run() {System.out.println(6507);}}
    public static class Impl6508 implements DemoService {public int type() {return 6508;} public void run() {System.out.println(6508);}}
    public static class Impl6509 implements DemoService {public int type() {return 6509;} public void run() {System.out.println(6509);}}
    public static class Impl6510 implements DemoService {public int type() {return 6510;} public void run() {System.out.println(6510);}}
    public static class Impl6511 implements DemoService {public int type() {return 6511;} public void run() {System.out.println(6511);}}
    public static class Impl6512 implements DemoService {public int type() {return 6512;} public void run() {System.out.println(6512);}}
    public static class Impl6513 implements DemoService {public int type() {return 6513;} public void run() {System.out.println(6513);}}
    public static class Impl6514 implements DemoService {public int type() {return 6514;} public void run() {System.out.println(6514);}}
    public static class Impl6515 implements DemoService {public int type() {return 6515;} public void run() {System.out.println(6515);}}
    public static class Impl6516 implements DemoService {public int type() {return 6516;} public void run() {System.out.println(6516);}}
    public static class Impl6517 implements DemoService {public int type() {return 6517;} public void run() {System.out.println(6517);}}
    public static class Impl6518 implements DemoService {public int type() {return 6518;} public void run() {System.out.println(6518);}}
    public static class Impl6519 implements DemoService {public int type() {return 6519;} public void run() {System.out.println(6519);}}
    public static class Impl6520 implements DemoService {public int type() {return 6520;} public void run() {System.out.println(6520);}}
    public static class Impl6521 implements DemoService {public int type() {return 6521;} public void run() {System.out.println(6521);}}
    public static class Impl6522 implements DemoService {public int type() {return 6522;} public void run() {System.out.println(6522);}}
    public static class Impl6523 implements DemoService {public int type() {return 6523;} public void run() {System.out.println(6523);}}
    public static class Impl6524 implements DemoService {public int type() {return 6524;} public void run() {System.out.println(6524);}}
    public static class Impl6525 implements DemoService {public int type() {return 6525;} public void run() {System.out.println(6525);}}
    public static class Impl6526 implements DemoService {public int type() {return 6526;} public void run() {System.out.println(6526);}}
    public static class Impl6527 implements DemoService {public int type() {return 6527;} public void run() {System.out.println(6527);}}
    public static class Impl6528 implements DemoService {public int type() {return 6528;} public void run() {System.out.println(6528);}}
    public static class Impl6529 implements DemoService {public int type() {return 6529;} public void run() {System.out.println(6529);}}
    public static class Impl6530 implements DemoService {public int type() {return 6530;} public void run() {System.out.println(6530);}}
    public static class Impl6531 implements DemoService {public int type() {return 6531;} public void run() {System.out.println(6531);}}
    public static class Impl6532 implements DemoService {public int type() {return 6532;} public void run() {System.out.println(6532);}}
    public static class Impl6533 implements DemoService {public int type() {return 6533;} public void run() {System.out.println(6533);}}
    public static class Impl6534 implements DemoService {public int type() {return 6534;} public void run() {System.out.println(6534);}}
    public static class Impl6535 implements DemoService {public int type() {return 6535;} public void run() {System.out.println(6535);}}
    public static class Impl6536 implements DemoService {public int type() {return 6536;} public void run() {System.out.println(6536);}}
    public static class Impl6537 implements DemoService {public int type() {return 6537;} public void run() {System.out.println(6537);}}
    public static class Impl6538 implements DemoService {public int type() {return 6538;} public void run() {System.out.println(6538);}}
    public static class Impl6539 implements DemoService {public int type() {return 6539;} public void run() {System.out.println(6539);}}
    public static class Impl6540 implements DemoService {public int type() {return 6540;} public void run() {System.out.println(6540);}}
    public static class Impl6541 implements DemoService {public int type() {return 6541;} public void run() {System.out.println(6541);}}
    public static class Impl6542 implements DemoService {public int type() {return 6542;} public void run() {System.out.println(6542);}}
    public static class Impl6543 implements DemoService {public int type() {return 6543;} public void run() {System.out.println(6543);}}
    public static class Impl6544 implements DemoService {public int type() {return 6544;} public void run() {System.out.println(6544);}}
    public static class Impl6545 implements DemoService {public int type() {return 6545;} public void run() {System.out.println(6545);}}
    public static class Impl6546 implements DemoService {public int type() {return 6546;} public void run() {System.out.println(6546);}}
    public static class Impl6547 implements DemoService {public int type() {return 6547;} public void run() {System.out.println(6547);}}
    public static class Impl6548 implements DemoService {public int type() {return 6548;} public void run() {System.out.println(6548);}}
    public static class Impl6549 implements DemoService {public int type() {return 6549;} public void run() {System.out.println(6549);}}
    public static class Impl6550 implements DemoService {public int type() {return 6550;} public void run() {System.out.println(6550);}}
    public static class Impl6551 implements DemoService {public int type() {return 6551;} public void run() {System.out.println(6551);}}
    public static class Impl6552 implements DemoService {public int type() {return 6552;} public void run() {System.out.println(6552);}}
    public static class Impl6553 implements DemoService {public int type() {return 6553;} public void run() {System.out.println(6553);}}
    public static class Impl6554 implements DemoService {public int type() {return 6554;} public void run() {System.out.println(6554);}}
    public static class Impl6555 implements DemoService {public int type() {return 6555;} public void run() {System.out.println(6555);}}
    public static class Impl6556 implements DemoService {public int type() {return 6556;} public void run() {System.out.println(6556);}}
    public static class Impl6557 implements DemoService {public int type() {return 6557;} public void run() {System.out.println(6557);}}
    public static class Impl6558 implements DemoService {public int type() {return 6558;} public void run() {System.out.println(6558);}}
    public static class Impl6559 implements DemoService {public int type() {return 6559;} public void run() {System.out.println(6559);}}
    public static class Impl6560 implements DemoService {public int type() {return 6560;} public void run() {System.out.println(6560);}}
    public static class Impl6561 implements DemoService {public int type() {return 6561;} public void run() {System.out.println(6561);}}
    public static class Impl6562 implements DemoService {public int type() {return 6562;} public void run() {System.out.println(6562);}}
    public static class Impl6563 implements DemoService {public int type() {return 6563;} public void run() {System.out.println(6563);}}
    public static class Impl6564 implements DemoService {public int type() {return 6564;} public void run() {System.out.println(6564);}}
    public static class Impl6565 implements DemoService {public int type() {return 6565;} public void run() {System.out.println(6565);}}
    public static class Impl6566 implements DemoService {public int type() {return 6566;} public void run() {System.out.println(6566);}}
    public static class Impl6567 implements DemoService {public int type() {return 6567;} public void run() {System.out.println(6567);}}
    public static class Impl6568 implements DemoService {public int type() {return 6568;} public void run() {System.out.println(6568);}}
    public static class Impl6569 implements DemoService {public int type() {return 6569;} public void run() {System.out.println(6569);}}
    public static class Impl6570 implements DemoService {public int type() {return 6570;} public void run() {System.out.println(6570);}}
    public static class Impl6571 implements DemoService {public int type() {return 6571;} public void run() {System.out.println(6571);}}
    public static class Impl6572 implements DemoService {public int type() {return 6572;} public void run() {System.out.println(6572);}}
    public static class Impl6573 implements DemoService {public int type() {return 6573;} public void run() {System.out.println(6573);}}
    public static class Impl6574 implements DemoService {public int type() {return 6574;} public void run() {System.out.println(6574);}}
    public static class Impl6575 implements DemoService {public int type() {return 6575;} public void run() {System.out.println(6575);}}
    public static class Impl6576 implements DemoService {public int type() {return 6576;} public void run() {System.out.println(6576);}}
    public static class Impl6577 implements DemoService {public int type() {return 6577;} public void run() {System.out.println(6577);}}
    public static class Impl6578 implements DemoService {public int type() {return 6578;} public void run() {System.out.println(6578);}}
    public static class Impl6579 implements DemoService {public int type() {return 6579;} public void run() {System.out.println(6579);}}
    public static class Impl6580 implements DemoService {public int type() {return 6580;} public void run() {System.out.println(6580);}}
    public static class Impl6581 implements DemoService {public int type() {return 6581;} public void run() {System.out.println(6581);}}
    public static class Impl6582 implements DemoService {public int type() {return 6582;} public void run() {System.out.println(6582);}}
    public static class Impl6583 implements DemoService {public int type() {return 6583;} public void run() {System.out.println(6583);}}
    public static class Impl6584 implements DemoService {public int type() {return 6584;} public void run() {System.out.println(6584);}}
    public static class Impl6585 implements DemoService {public int type() {return 6585;} public void run() {System.out.println(6585);}}
    public static class Impl6586 implements DemoService {public int type() {return 6586;} public void run() {System.out.println(6586);}}
    public static class Impl6587 implements DemoService {public int type() {return 6587;} public void run() {System.out.println(6587);}}
    public static class Impl6588 implements DemoService {public int type() {return 6588;} public void run() {System.out.println(6588);}}
    public static class Impl6589 implements DemoService {public int type() {return 6589;} public void run() {System.out.println(6589);}}
    public static class Impl6590 implements DemoService {public int type() {return 6590;} public void run() {System.out.println(6590);}}
    public static class Impl6591 implements DemoService {public int type() {return 6591;} public void run() {System.out.println(6591);}}
    public static class Impl6592 implements DemoService {public int type() {return 6592;} public void run() {System.out.println(6592);}}
    public static class Impl6593 implements DemoService {public int type() {return 6593;} public void run() {System.out.println(6593);}}
    public static class Impl6594 implements DemoService {public int type() {return 6594;} public void run() {System.out.println(6594);}}
    public static class Impl6595 implements DemoService {public int type() {return 6595;} public void run() {System.out.println(6595);}}
    public static class Impl6596 implements DemoService {public int type() {return 6596;} public void run() {System.out.println(6596);}}
    public static class Impl6597 implements DemoService {public int type() {return 6597;} public void run() {System.out.println(6597);}}
    public static class Impl6598 implements DemoService {public int type() {return 6598;} public void run() {System.out.println(6598);}}
    public static class Impl6599 implements DemoService {public int type() {return 6599;} public void run() {System.out.println(6599);}}
    public static class Impl6600 implements DemoService {public int type() {return 6600;} public void run() {System.out.println(6600);}}
    public static class Impl6601 implements DemoService {public int type() {return 6601;} public void run() {System.out.println(6601);}}
    public static class Impl6602 implements DemoService {public int type() {return 6602;} public void run() {System.out.println(6602);}}
    public static class Impl6603 implements DemoService {public int type() {return 6603;} public void run() {System.out.println(6603);}}
    public static class Impl6604 implements DemoService {public int type() {return 6604;} public void run() {System.out.println(6604);}}
    public static class Impl6605 implements DemoService {public int type() {return 6605;} public void run() {System.out.println(6605);}}
    public static class Impl6606 implements DemoService {public int type() {return 6606;} public void run() {System.out.println(6606);}}
    public static class Impl6607 implements DemoService {public int type() {return 6607;} public void run() {System.out.println(6607);}}
    public static class Impl6608 implements DemoService {public int type() {return 6608;} public void run() {System.out.println(6608);}}
    public static class Impl6609 implements DemoService {public int type() {return 6609;} public void run() {System.out.println(6609);}}
    public static class Impl6610 implements DemoService {public int type() {return 6610;} public void run() {System.out.println(6610);}}
    public static class Impl6611 implements DemoService {public int type() {return 6611;} public void run() {System.out.println(6611);}}
    public static class Impl6612 implements DemoService {public int type() {return 6612;} public void run() {System.out.println(6612);}}
    public static class Impl6613 implements DemoService {public int type() {return 6613;} public void run() {System.out.println(6613);}}
    public static class Impl6614 implements DemoService {public int type() {return 6614;} public void run() {System.out.println(6614);}}
    public static class Impl6615 implements DemoService {public int type() {return 6615;} public void run() {System.out.println(6615);}}
    public static class Impl6616 implements DemoService {public int type() {return 6616;} public void run() {System.out.println(6616);}}
    public static class Impl6617 implements DemoService {public int type() {return 6617;} public void run() {System.out.println(6617);}}
    public static class Impl6618 implements DemoService {public int type() {return 6618;} public void run() {System.out.println(6618);}}
    public static class Impl6619 implements DemoService {public int type() {return 6619;} public void run() {System.out.println(6619);}}
    public static class Impl6620 implements DemoService {public int type() {return 6620;} public void run() {System.out.println(6620);}}
    public static class Impl6621 implements DemoService {public int type() {return 6621;} public void run() {System.out.println(6621);}}
    public static class Impl6622 implements DemoService {public int type() {return 6622;} public void run() {System.out.println(6622);}}
    public static class Impl6623 implements DemoService {public int type() {return 6623;} public void run() {System.out.println(6623);}}
    public static class Impl6624 implements DemoService {public int type() {return 6624;} public void run() {System.out.println(6624);}}
    public static class Impl6625 implements DemoService {public int type() {return 6625;} public void run() {System.out.println(6625);}}
    public static class Impl6626 implements DemoService {public int type() {return 6626;} public void run() {System.out.println(6626);}}
    public static class Impl6627 implements DemoService {public int type() {return 6627;} public void run() {System.out.println(6627);}}
    public static class Impl6628 implements DemoService {public int type() {return 6628;} public void run() {System.out.println(6628);}}
    public static class Impl6629 implements DemoService {public int type() {return 6629;} public void run() {System.out.println(6629);}}
    public static class Impl6630 implements DemoService {public int type() {return 6630;} public void run() {System.out.println(6630);}}
    public static class Impl6631 implements DemoService {public int type() {return 6631;} public void run() {System.out.println(6631);}}
    public static class Impl6632 implements DemoService {public int type() {return 6632;} public void run() {System.out.println(6632);}}
    public static class Impl6633 implements DemoService {public int type() {return 6633;} public void run() {System.out.println(6633);}}
    public static class Impl6634 implements DemoService {public int type() {return 6634;} public void run() {System.out.println(6634);}}
    public static class Impl6635 implements DemoService {public int type() {return 6635;} public void run() {System.out.println(6635);}}
    public static class Impl6636 implements DemoService {public int type() {return 6636;} public void run() {System.out.println(6636);}}
    public static class Impl6637 implements DemoService {public int type() {return 6637;} public void run() {System.out.println(6637);}}
    public static class Impl6638 implements DemoService {public int type() {return 6638;} public void run() {System.out.println(6638);}}
    public static class Impl6639 implements DemoService {public int type() {return 6639;} public void run() {System.out.println(6639);}}
    public static class Impl6640 implements DemoService {public int type() {return 6640;} public void run() {System.out.println(6640);}}
    public static class Impl6641 implements DemoService {public int type() {return 6641;} public void run() {System.out.println(6641);}}
    public static class Impl6642 implements DemoService {public int type() {return 6642;} public void run() {System.out.println(6642);}}
    public static class Impl6643 implements DemoService {public int type() {return 6643;} public void run() {System.out.println(6643);}}
    public static class Impl6644 implements DemoService {public int type() {return 6644;} public void run() {System.out.println(6644);}}
    public static class Impl6645 implements DemoService {public int type() {return 6645;} public void run() {System.out.println(6645);}}
    public static class Impl6646 implements DemoService {public int type() {return 6646;} public void run() {System.out.println(6646);}}
    public static class Impl6647 implements DemoService {public int type() {return 6647;} public void run() {System.out.println(6647);}}
    public static class Impl6648 implements DemoService {public int type() {return 6648;} public void run() {System.out.println(6648);}}
    public static class Impl6649 implements DemoService {public int type() {return 6649;} public void run() {System.out.println(6649);}}
    public static class Impl6650 implements DemoService {public int type() {return 6650;} public void run() {System.out.println(6650);}}
    public static class Impl6651 implements DemoService {public int type() {return 6651;} public void run() {System.out.println(6651);}}
    public static class Impl6652 implements DemoService {public int type() {return 6652;} public void run() {System.out.println(6652);}}
    public static class Impl6653 implements DemoService {public int type() {return 6653;} public void run() {System.out.println(6653);}}
    public static class Impl6654 implements DemoService {public int type() {return 6654;} public void run() {System.out.println(6654);}}
    public static class Impl6655 implements DemoService {public int type() {return 6655;} public void run() {System.out.println(6655);}}
    public static class Impl6656 implements DemoService {public int type() {return 6656;} public void run() {System.out.println(6656);}}
    public static class Impl6657 implements DemoService {public int type() {return 6657;} public void run() {System.out.println(6657);}}
    public static class Impl6658 implements DemoService {public int type() {return 6658;} public void run() {System.out.println(6658);}}
    public static class Impl6659 implements DemoService {public int type() {return 6659;} public void run() {System.out.println(6659);}}
    public static class Impl6660 implements DemoService {public int type() {return 6660;} public void run() {System.out.println(6660);}}
    public static class Impl6661 implements DemoService {public int type() {return 6661;} public void run() {System.out.println(6661);}}
    public static class Impl6662 implements DemoService {public int type() {return 6662;} public void run() {System.out.println(6662);}}
    public static class Impl6663 implements DemoService {public int type() {return 6663;} public void run() {System.out.println(6663);}}
    public static class Impl6664 implements DemoService {public int type() {return 6664;} public void run() {System.out.println(6664);}}
    public static class Impl6665 implements DemoService {public int type() {return 6665;} public void run() {System.out.println(6665);}}
    public static class Impl6666 implements DemoService {public int type() {return 6666;} public void run() {System.out.println(6666);}}
    public static class Impl6667 implements DemoService {public int type() {return 6667;} public void run() {System.out.println(6667);}}
    public static class Impl6668 implements DemoService {public int type() {return 6668;} public void run() {System.out.println(6668);}}
    public static class Impl6669 implements DemoService {public int type() {return 6669;} public void run() {System.out.println(6669);}}
    public static class Impl6670 implements DemoService {public int type() {return 6670;} public void run() {System.out.println(6670);}}
    public static class Impl6671 implements DemoService {public int type() {return 6671;} public void run() {System.out.println(6671);}}
    public static class Impl6672 implements DemoService {public int type() {return 6672;} public void run() {System.out.println(6672);}}
    public static class Impl6673 implements DemoService {public int type() {return 6673;} public void run() {System.out.println(6673);}}
    public static class Impl6674 implements DemoService {public int type() {return 6674;} public void run() {System.out.println(6674);}}
    public static class Impl6675 implements DemoService {public int type() {return 6675;} public void run() {System.out.println(6675);}}
    public static class Impl6676 implements DemoService {public int type() {return 6676;} public void run() {System.out.println(6676);}}
    public static class Impl6677 implements DemoService {public int type() {return 6677;} public void run() {System.out.println(6677);}}
    public static class Impl6678 implements DemoService {public int type() {return 6678;} public void run() {System.out.println(6678);}}
    public static class Impl6679 implements DemoService {public int type() {return 6679;} public void run() {System.out.println(6679);}}
    public static class Impl6680 implements DemoService {public int type() {return 6680;} public void run() {System.out.println(6680);}}
    public static class Impl6681 implements DemoService {public int type() {return 6681;} public void run() {System.out.println(6681);}}
    public static class Impl6682 implements DemoService {public int type() {return 6682;} public void run() {System.out.println(6682);}}
    public static class Impl6683 implements DemoService {public int type() {return 6683;} public void run() {System.out.println(6683);}}
    public static class Impl6684 implements DemoService {public int type() {return 6684;} public void run() {System.out.println(6684);}}
    public static class Impl6685 implements DemoService {public int type() {return 6685;} public void run() {System.out.println(6685);}}
    public static class Impl6686 implements DemoService {public int type() {return 6686;} public void run() {System.out.println(6686);}}
    public static class Impl6687 implements DemoService {public int type() {return 6687;} public void run() {System.out.println(6687);}}
    public static class Impl6688 implements DemoService {public int type() {return 6688;} public void run() {System.out.println(6688);}}
    public static class Impl6689 implements DemoService {public int type() {return 6689;} public void run() {System.out.println(6689);}}
    public static class Impl6690 implements DemoService {public int type() {return 6690;} public void run() {System.out.println(6690);}}
    public static class Impl6691 implements DemoService {public int type() {return 6691;} public void run() {System.out.println(6691);}}
    public static class Impl6692 implements DemoService {public int type() {return 6692;} public void run() {System.out.println(6692);}}
    public static class Impl6693 implements DemoService {public int type() {return 6693;} public void run() {System.out.println(6693);}}
    public static class Impl6694 implements DemoService {public int type() {return 6694;} public void run() {System.out.println(6694);}}
    public static class Impl6695 implements DemoService {public int type() {return 6695;} public void run() {System.out.println(6695);}}
    public static class Impl6696 implements DemoService {public int type() {return 6696;} public void run() {System.out.println(6696);}}
    public static class Impl6697 implements DemoService {public int type() {return 6697;} public void run() {System.out.println(6697);}}
    public static class Impl6698 implements DemoService {public int type() {return 6698;} public void run() {System.out.println(6698);}}
    public static class Impl6699 implements DemoService {public int type() {return 6699;} public void run() {System.out.println(6699);}}
    public static class Impl6700 implements DemoService {public int type() {return 6700;} public void run() {System.out.println(6700);}}
    public static class Impl6701 implements DemoService {public int type() {return 6701;} public void run() {System.out.println(6701);}}
    public static class Impl6702 implements DemoService {public int type() {return 6702;} public void run() {System.out.println(6702);}}
    public static class Impl6703 implements DemoService {public int type() {return 6703;} public void run() {System.out.println(6703);}}
    public static class Impl6704 implements DemoService {public int type() {return 6704;} public void run() {System.out.println(6704);}}
    public static class Impl6705 implements DemoService {public int type() {return 6705;} public void run() {System.out.println(6705);}}
    public static class Impl6706 implements DemoService {public int type() {return 6706;} public void run() {System.out.println(6706);}}
    public static class Impl6707 implements DemoService {public int type() {return 6707;} public void run() {System.out.println(6707);}}
    public static class Impl6708 implements DemoService {public int type() {return 6708;} public void run() {System.out.println(6708);}}
    public static class Impl6709 implements DemoService {public int type() {return 6709;} public void run() {System.out.println(6709);}}
    public static class Impl6710 implements DemoService {public int type() {return 6710;} public void run() {System.out.println(6710);}}
    public static class Impl6711 implements DemoService {public int type() {return 6711;} public void run() {System.out.println(6711);}}
    public static class Impl6712 implements DemoService {public int type() {return 6712;} public void run() {System.out.println(6712);}}
    public static class Impl6713 implements DemoService {public int type() {return 6713;} public void run() {System.out.println(6713);}}
    public static class Impl6714 implements DemoService {public int type() {return 6714;} public void run() {System.out.println(6714);}}
    public static class Impl6715 implements DemoService {public int type() {return 6715;} public void run() {System.out.println(6715);}}
    public static class Impl6716 implements DemoService {public int type() {return 6716;} public void run() {System.out.println(6716);}}
    public static class Impl6717 implements DemoService {public int type() {return 6717;} public void run() {System.out.println(6717);}}
    public static class Impl6718 implements DemoService {public int type() {return 6718;} public void run() {System.out.println(6718);}}
    public static class Impl6719 implements DemoService {public int type() {return 6719;} public void run() {System.out.println(6719);}}
    public static class Impl6720 implements DemoService {public int type() {return 6720;} public void run() {System.out.println(6720);}}
    public static class Impl6721 implements DemoService {public int type() {return 6721;} public void run() {System.out.println(6721);}}
    public static class Impl6722 implements DemoService {public int type() {return 6722;} public void run() {System.out.println(6722);}}
    public static class Impl6723 implements DemoService {public int type() {return 6723;} public void run() {System.out.println(6723);}}
    public static class Impl6724 implements DemoService {public int type() {return 6724;} public void run() {System.out.println(6724);}}
    public static class Impl6725 implements DemoService {public int type() {return 6725;} public void run() {System.out.println(6725);}}
    public static class Impl6726 implements DemoService {public int type() {return 6726;} public void run() {System.out.println(6726);}}
    public static class Impl6727 implements DemoService {public int type() {return 6727;} public void run() {System.out.println(6727);}}
    public static class Impl6728 implements DemoService {public int type() {return 6728;} public void run() {System.out.println(6728);}}
    public static class Impl6729 implements DemoService {public int type() {return 6729;} public void run() {System.out.println(6729);}}
    public static class Impl6730 implements DemoService {public int type() {return 6730;} public void run() {System.out.println(6730);}}
    public static class Impl6731 implements DemoService {public int type() {return 6731;} public void run() {System.out.println(6731);}}
    public static class Impl6732 implements DemoService {public int type() {return 6732;} public void run() {System.out.println(6732);}}
    public static class Impl6733 implements DemoService {public int type() {return 6733;} public void run() {System.out.println(6733);}}
    public static class Impl6734 implements DemoService {public int type() {return 6734;} public void run() {System.out.println(6734);}}
    public static class Impl6735 implements DemoService {public int type() {return 6735;} public void run() {System.out.println(6735);}}
    public static class Impl6736 implements DemoService {public int type() {return 6736;} public void run() {System.out.println(6736);}}
    public static class Impl6737 implements DemoService {public int type() {return 6737;} public void run() {System.out.println(6737);}}
    public static class Impl6738 implements DemoService {public int type() {return 6738;} public void run() {System.out.println(6738);}}
    public static class Impl6739 implements DemoService {public int type() {return 6739;} public void run() {System.out.println(6739);}}
    public static class Impl6740 implements DemoService {public int type() {return 6740;} public void run() {System.out.println(6740);}}
    public static class Impl6741 implements DemoService {public int type() {return 6741;} public void run() {System.out.println(6741);}}
    public static class Impl6742 implements DemoService {public int type() {return 6742;} public void run() {System.out.println(6742);}}
    public static class Impl6743 implements DemoService {public int type() {return 6743;} public void run() {System.out.println(6743);}}
    public static class Impl6744 implements DemoService {public int type() {return 6744;} public void run() {System.out.println(6744);}}
    public static class Impl6745 implements DemoService {public int type() {return 6745;} public void run() {System.out.println(6745);}}
    public static class Impl6746 implements DemoService {public int type() {return 6746;} public void run() {System.out.println(6746);}}
    public static class Impl6747 implements DemoService {public int type() {return 6747;} public void run() {System.out.println(6747);}}
    public static class Impl6748 implements DemoService {public int type() {return 6748;} public void run() {System.out.println(6748);}}
    public static class Impl6749 implements DemoService {public int type() {return 6749;} public void run() {System.out.println(6749);}}
    public static class Impl6750 implements DemoService {public int type() {return 6750;} public void run() {System.out.println(6750);}}
    public static class Impl6751 implements DemoService {public int type() {return 6751;} public void run() {System.out.println(6751);}}
    public static class Impl6752 implements DemoService {public int type() {return 6752;} public void run() {System.out.println(6752);}}
    public static class Impl6753 implements DemoService {public int type() {return 6753;} public void run() {System.out.println(6753);}}
    public static class Impl6754 implements DemoService {public int type() {return 6754;} public void run() {System.out.println(6754);}}
    public static class Impl6755 implements DemoService {public int type() {return 6755;} public void run() {System.out.println(6755);}}
    public static class Impl6756 implements DemoService {public int type() {return 6756;} public void run() {System.out.println(6756);}}
    public static class Impl6757 implements DemoService {public int type() {return 6757;} public void run() {System.out.println(6757);}}
    public static class Impl6758 implements DemoService {public int type() {return 6758;} public void run() {System.out.println(6758);}}
    public static class Impl6759 implements DemoService {public int type() {return 6759;} public void run() {System.out.println(6759);}}
    public static class Impl6760 implements DemoService {public int type() {return 6760;} public void run() {System.out.println(6760);}}
    public static class Impl6761 implements DemoService {public int type() {return 6761;} public void run() {System.out.println(6761);}}
    public static class Impl6762 implements DemoService {public int type() {return 6762;} public void run() {System.out.println(6762);}}
    public static class Impl6763 implements DemoService {public int type() {return 6763;} public void run() {System.out.println(6763);}}
    public static class Impl6764 implements DemoService {public int type() {return 6764;} public void run() {System.out.println(6764);}}
    public static class Impl6765 implements DemoService {public int type() {return 6765;} public void run() {System.out.println(6765);}}
    public static class Impl6766 implements DemoService {public int type() {return 6766;} public void run() {System.out.println(6766);}}
    public static class Impl6767 implements DemoService {public int type() {return 6767;} public void run() {System.out.println(6767);}}
    public static class Impl6768 implements DemoService {public int type() {return 6768;} public void run() {System.out.println(6768);}}
    public static class Impl6769 implements DemoService {public int type() {return 6769;} public void run() {System.out.println(6769);}}
    public static class Impl6770 implements DemoService {public int type() {return 6770;} public void run() {System.out.println(6770);}}
    public static class Impl6771 implements DemoService {public int type() {return 6771;} public void run() {System.out.println(6771);}}
    public static class Impl6772 implements DemoService {public int type() {return 6772;} public void run() {System.out.println(6772);}}
    public static class Impl6773 implements DemoService {public int type() {return 6773;} public void run() {System.out.println(6773);}}
    public static class Impl6774 implements DemoService {public int type() {return 6774;} public void run() {System.out.println(6774);}}
    public static class Impl6775 implements DemoService {public int type() {return 6775;} public void run() {System.out.println(6775);}}
    public static class Impl6776 implements DemoService {public int type() {return 6776;} public void run() {System.out.println(6776);}}
    public static class Impl6777 implements DemoService {public int type() {return 6777;} public void run() {System.out.println(6777);}}
    public static class Impl6778 implements DemoService {public int type() {return 6778;} public void run() {System.out.println(6778);}}
    public static class Impl6779 implements DemoService {public int type() {return 6779;} public void run() {System.out.println(6779);}}
    public static class Impl6780 implements DemoService {public int type() {return 6780;} public void run() {System.out.println(6780);}}
    public static class Impl6781 implements DemoService {public int type() {return 6781;} public void run() {System.out.println(6781);}}
    public static class Impl6782 implements DemoService {public int type() {return 6782;} public void run() {System.out.println(6782);}}
    public static class Impl6783 implements DemoService {public int type() {return 6783;} public void run() {System.out.println(6783);}}
    public static class Impl6784 implements DemoService {public int type() {return 6784;} public void run() {System.out.println(6784);}}
    public static class Impl6785 implements DemoService {public int type() {return 6785;} public void run() {System.out.println(6785);}}
    public static class Impl6786 implements DemoService {public int type() {return 6786;} public void run() {System.out.println(6786);}}
    public static class Impl6787 implements DemoService {public int type() {return 6787;} public void run() {System.out.println(6787);}}
    public static class Impl6788 implements DemoService {public int type() {return 6788;} public void run() {System.out.println(6788);}}
    public static class Impl6789 implements DemoService {public int type() {return 6789;} public void run() {System.out.println(6789);}}
    public static class Impl6790 implements DemoService {public int type() {return 6790;} public void run() {System.out.println(6790);}}
    public static class Impl6791 implements DemoService {public int type() {return 6791;} public void run() {System.out.println(6791);}}
    public static class Impl6792 implements DemoService {public int type() {return 6792;} public void run() {System.out.println(6792);}}
    public static class Impl6793 implements DemoService {public int type() {return 6793;} public void run() {System.out.println(6793);}}
    public static class Impl6794 implements DemoService {public int type() {return 6794;} public void run() {System.out.println(6794);}}
    public static class Impl6795 implements DemoService {public int type() {return 6795;} public void run() {System.out.println(6795);}}
    public static class Impl6796 implements DemoService {public int type() {return 6796;} public void run() {System.out.println(6796);}}
    public static class Impl6797 implements DemoService {public int type() {return 6797;} public void run() {System.out.println(6797);}}
    public static class Impl6798 implements DemoService {public int type() {return 6798;} public void run() {System.out.println(6798);}}
    public static class Impl6799 implements DemoService {public int type() {return 6799;} public void run() {System.out.println(6799);}}
    public static class Impl6800 implements DemoService {public int type() {return 6800;} public void run() {System.out.println(6800);}}
    public static class Impl6801 implements DemoService {public int type() {return 6801;} public void run() {System.out.println(6801);}}
    public static class Impl6802 implements DemoService {public int type() {return 6802;} public void run() {System.out.println(6802);}}
    public static class Impl6803 implements DemoService {public int type() {return 6803;} public void run() {System.out.println(6803);}}
    public static class Impl6804 implements DemoService {public int type() {return 6804;} public void run() {System.out.println(6804);}}
    public static class Impl6805 implements DemoService {public int type() {return 6805;} public void run() {System.out.println(6805);}}
    public static class Impl6806 implements DemoService {public int type() {return 6806;} public void run() {System.out.println(6806);}}
    public static class Impl6807 implements DemoService {public int type() {return 6807;} public void run() {System.out.println(6807);}}
    public static class Impl6808 implements DemoService {public int type() {return 6808;} public void run() {System.out.println(6808);}}
    public static class Impl6809 implements DemoService {public int type() {return 6809;} public void run() {System.out.println(6809);}}
    public static class Impl6810 implements DemoService {public int type() {return 6810;} public void run() {System.out.println(6810);}}
    public static class Impl6811 implements DemoService {public int type() {return 6811;} public void run() {System.out.println(6811);}}
    public static class Impl6812 implements DemoService {public int type() {return 6812;} public void run() {System.out.println(6812);}}
    public static class Impl6813 implements DemoService {public int type() {return 6813;} public void run() {System.out.println(6813);}}
    public static class Impl6814 implements DemoService {public int type() {return 6814;} public void run() {System.out.println(6814);}}
    public static class Impl6815 implements DemoService {public int type() {return 6815;} public void run() {System.out.println(6815);}}
    public static class Impl6816 implements DemoService {public int type() {return 6816;} public void run() {System.out.println(6816);}}
    public static class Impl6817 implements DemoService {public int type() {return 6817;} public void run() {System.out.println(6817);}}
    public static class Impl6818 implements DemoService {public int type() {return 6818;} public void run() {System.out.println(6818);}}
    public static class Impl6819 implements DemoService {public int type() {return 6819;} public void run() {System.out.println(6819);}}
    public static class Impl6820 implements DemoService {public int type() {return 6820;} public void run() {System.out.println(6820);}}
    public static class Impl6821 implements DemoService {public int type() {return 6821;} public void run() {System.out.println(6821);}}
    public static class Impl6822 implements DemoService {public int type() {return 6822;} public void run() {System.out.println(6822);}}
    public static class Impl6823 implements DemoService {public int type() {return 6823;} public void run() {System.out.println(6823);}}
    public static class Impl6824 implements DemoService {public int type() {return 6824;} public void run() {System.out.println(6824);}}
    public static class Impl6825 implements DemoService {public int type() {return 6825;} public void run() {System.out.println(6825);}}
    public static class Impl6826 implements DemoService {public int type() {return 6826;} public void run() {System.out.println(6826);}}
    public static class Impl6827 implements DemoService {public int type() {return 6827;} public void run() {System.out.println(6827);}}
    public static class Impl6828 implements DemoService {public int type() {return 6828;} public void run() {System.out.println(6828);}}
    public static class Impl6829 implements DemoService {public int type() {return 6829;} public void run() {System.out.println(6829);}}
    public static class Impl6830 implements DemoService {public int type() {return 6830;} public void run() {System.out.println(6830);}}
    public static class Impl6831 implements DemoService {public int type() {return 6831;} public void run() {System.out.println(6831);}}
    public static class Impl6832 implements DemoService {public int type() {return 6832;} public void run() {System.out.println(6832);}}
    public static class Impl6833 implements DemoService {public int type() {return 6833;} public void run() {System.out.println(6833);}}
    public static class Impl6834 implements DemoService {public int type() {return 6834;} public void run() {System.out.println(6834);}}
    public static class Impl6835 implements DemoService {public int type() {return 6835;} public void run() {System.out.println(6835);}}
    public static class Impl6836 implements DemoService {public int type() {return 6836;} public void run() {System.out.println(6836);}}
    public static class Impl6837 implements DemoService {public int type() {return 6837;} public void run() {System.out.println(6837);}}
    public static class Impl6838 implements DemoService {public int type() {return 6838;} public void run() {System.out.println(6838);}}
    public static class Impl6839 implements DemoService {public int type() {return 6839;} public void run() {System.out.println(6839);}}
    public static class Impl6840 implements DemoService {public int type() {return 6840;} public void run() {System.out.println(6840);}}
    public static class Impl6841 implements DemoService {public int type() {return 6841;} public void run() {System.out.println(6841);}}
    public static class Impl6842 implements DemoService {public int type() {return 6842;} public void run() {System.out.println(6842);}}
    public static class Impl6843 implements DemoService {public int type() {return 6843;} public void run() {System.out.println(6843);}}
    public static class Impl6844 implements DemoService {public int type() {return 6844;} public void run() {System.out.println(6844);}}
    public static class Impl6845 implements DemoService {public int type() {return 6845;} public void run() {System.out.println(6845);}}
    public static class Impl6846 implements DemoService {public int type() {return 6846;} public void run() {System.out.println(6846);}}
    public static class Impl6847 implements DemoService {public int type() {return 6847;} public void run() {System.out.println(6847);}}
    public static class Impl6848 implements DemoService {public int type() {return 6848;} public void run() {System.out.println(6848);}}
    public static class Impl6849 implements DemoService {public int type() {return 6849;} public void run() {System.out.println(6849);}}
    public static class Impl6850 implements DemoService {public int type() {return 6850;} public void run() {System.out.println(6850);}}
    public static class Impl6851 implements DemoService {public int type() {return 6851;} public void run() {System.out.println(6851);}}
    public static class Impl6852 implements DemoService {public int type() {return 6852;} public void run() {System.out.println(6852);}}
    public static class Impl6853 implements DemoService {public int type() {return 6853;} public void run() {System.out.println(6853);}}
    public static class Impl6854 implements DemoService {public int type() {return 6854;} public void run() {System.out.println(6854);}}
    public static class Impl6855 implements DemoService {public int type() {return 6855;} public void run() {System.out.println(6855);}}
    public static class Impl6856 implements DemoService {public int type() {return 6856;} public void run() {System.out.println(6856);}}
    public static class Impl6857 implements DemoService {public int type() {return 6857;} public void run() {System.out.println(6857);}}
    public static class Impl6858 implements DemoService {public int type() {return 6858;} public void run() {System.out.println(6858);}}
    public static class Impl6859 implements DemoService {public int type() {return 6859;} public void run() {System.out.println(6859);}}
    public static class Impl6860 implements DemoService {public int type() {return 6860;} public void run() {System.out.println(6860);}}
    public static class Impl6861 implements DemoService {public int type() {return 6861;} public void run() {System.out.println(6861);}}
    public static class Impl6862 implements DemoService {public int type() {return 6862;} public void run() {System.out.println(6862);}}
    public static class Impl6863 implements DemoService {public int type() {return 6863;} public void run() {System.out.println(6863);}}
    public static class Impl6864 implements DemoService {public int type() {return 6864;} public void run() {System.out.println(6864);}}
    public static class Impl6865 implements DemoService {public int type() {return 6865;} public void run() {System.out.println(6865);}}
    public static class Impl6866 implements DemoService {public int type() {return 6866;} public void run() {System.out.println(6866);}}
    public static class Impl6867 implements DemoService {public int type() {return 6867;} public void run() {System.out.println(6867);}}
    public static class Impl6868 implements DemoService {public int type() {return 6868;} public void run() {System.out.println(6868);}}
    public static class Impl6869 implements DemoService {public int type() {return 6869;} public void run() {System.out.println(6869);}}
    public static class Impl6870 implements DemoService {public int type() {return 6870;} public void run() {System.out.println(6870);}}
    public static class Impl6871 implements DemoService {public int type() {return 6871;} public void run() {System.out.println(6871);}}
    public static class Impl6872 implements DemoService {public int type() {return 6872;} public void run() {System.out.println(6872);}}
    public static class Impl6873 implements DemoService {public int type() {return 6873;} public void run() {System.out.println(6873);}}
    public static class Impl6874 implements DemoService {public int type() {return 6874;} public void run() {System.out.println(6874);}}
    public static class Impl6875 implements DemoService {public int type() {return 6875;} public void run() {System.out.println(6875);}}
    public static class Impl6876 implements DemoService {public int type() {return 6876;} public void run() {System.out.println(6876);}}
    public static class Impl6877 implements DemoService {public int type() {return 6877;} public void run() {System.out.println(6877);}}
    public static class Impl6878 implements DemoService {public int type() {return 6878;} public void run() {System.out.println(6878);}}
    public static class Impl6879 implements DemoService {public int type() {return 6879;} public void run() {System.out.println(6879);}}
    public static class Impl6880 implements DemoService {public int type() {return 6880;} public void run() {System.out.println(6880);}}
    public static class Impl6881 implements DemoService {public int type() {return 6881;} public void run() {System.out.println(6881);}}
    public static class Impl6882 implements DemoService {public int type() {return 6882;} public void run() {System.out.println(6882);}}
    public static class Impl6883 implements DemoService {public int type() {return 6883;} public void run() {System.out.println(6883);}}
    public static class Impl6884 implements DemoService {public int type() {return 6884;} public void run() {System.out.println(6884);}}
    public static class Impl6885 implements DemoService {public int type() {return 6885;} public void run() {System.out.println(6885);}}
    public static class Impl6886 implements DemoService {public int type() {return 6886;} public void run() {System.out.println(6886);}}
    public static class Impl6887 implements DemoService {public int type() {return 6887;} public void run() {System.out.println(6887);}}
    public static class Impl6888 implements DemoService {public int type() {return 6888;} public void run() {System.out.println(6888);}}
    public static class Impl6889 implements DemoService {public int type() {return 6889;} public void run() {System.out.println(6889);}}
    public static class Impl6890 implements DemoService {public int type() {return 6890;} public void run() {System.out.println(6890);}}
    public static class Impl6891 implements DemoService {public int type() {return 6891;} public void run() {System.out.println(6891);}}
    public static class Impl6892 implements DemoService {public int type() {return 6892;} public void run() {System.out.println(6892);}}
    public static class Impl6893 implements DemoService {public int type() {return 6893;} public void run() {System.out.println(6893);}}
    public static class Impl6894 implements DemoService {public int type() {return 6894;} public void run() {System.out.println(6894);}}
    public static class Impl6895 implements DemoService {public int type() {return 6895;} public void run() {System.out.println(6895);}}
    public static class Impl6896 implements DemoService {public int type() {return 6896;} public void run() {System.out.println(6896);}}
    public static class Impl6897 implements DemoService {public int type() {return 6897;} public void run() {System.out.println(6897);}}
    public static class Impl6898 implements DemoService {public int type() {return 6898;} public void run() {System.out.println(6898);}}
    public static class Impl6899 implements DemoService {public int type() {return 6899;} public void run() {System.out.println(6899);}}
    public static class Impl6900 implements DemoService {public int type() {return 6900;} public void run() {System.out.println(6900);}}
    public static class Impl6901 implements DemoService {public int type() {return 6901;} public void run() {System.out.println(6901);}}
    public static class Impl6902 implements DemoService {public int type() {return 6902;} public void run() {System.out.println(6902);}}
    public static class Impl6903 implements DemoService {public int type() {return 6903;} public void run() {System.out.println(6903);}}
    public static class Impl6904 implements DemoService {public int type() {return 6904;} public void run() {System.out.println(6904);}}
    public static class Impl6905 implements DemoService {public int type() {return 6905;} public void run() {System.out.println(6905);}}
    public static class Impl6906 implements DemoService {public int type() {return 6906;} public void run() {System.out.println(6906);}}
    public static class Impl6907 implements DemoService {public int type() {return 6907;} public void run() {System.out.println(6907);}}
    public static class Impl6908 implements DemoService {public int type() {return 6908;} public void run() {System.out.println(6908);}}
    public static class Impl6909 implements DemoService {public int type() {return 6909;} public void run() {System.out.println(6909);}}
    public static class Impl6910 implements DemoService {public int type() {return 6910;} public void run() {System.out.println(6910);}}
    public static class Impl6911 implements DemoService {public int type() {return 6911;} public void run() {System.out.println(6911);}}
    public static class Impl6912 implements DemoService {public int type() {return 6912;} public void run() {System.out.println(6912);}}
    public static class Impl6913 implements DemoService {public int type() {return 6913;} public void run() {System.out.println(6913);}}
    public static class Impl6914 implements DemoService {public int type() {return 6914;} public void run() {System.out.println(6914);}}
    public static class Impl6915 implements DemoService {public int type() {return 6915;} public void run() {System.out.println(6915);}}
    public static class Impl6916 implements DemoService {public int type() {return 6916;} public void run() {System.out.println(6916);}}
    public static class Impl6917 implements DemoService {public int type() {return 6917;} public void run() {System.out.println(6917);}}
    public static class Impl6918 implements DemoService {public int type() {return 6918;} public void run() {System.out.println(6918);}}
    public static class Impl6919 implements DemoService {public int type() {return 6919;} public void run() {System.out.println(6919);}}
    public static class Impl6920 implements DemoService {public int type() {return 6920;} public void run() {System.out.println(6920);}}
    public static class Impl6921 implements DemoService {public int type() {return 6921;} public void run() {System.out.println(6921);}}
    public static class Impl6922 implements DemoService {public int type() {return 6922;} public void run() {System.out.println(6922);}}
    public static class Impl6923 implements DemoService {public int type() {return 6923;} public void run() {System.out.println(6923);}}
    public static class Impl6924 implements DemoService {public int type() {return 6924;} public void run() {System.out.println(6924);}}
    public static class Impl6925 implements DemoService {public int type() {return 6925;} public void run() {System.out.println(6925);}}
    public static class Impl6926 implements DemoService {public int type() {return 6926;} public void run() {System.out.println(6926);}}
    public static class Impl6927 implements DemoService {public int type() {return 6927;} public void run() {System.out.println(6927);}}
    public static class Impl6928 implements DemoService {public int type() {return 6928;} public void run() {System.out.println(6928);}}
    public static class Impl6929 implements DemoService {public int type() {return 6929;} public void run() {System.out.println(6929);}}
    public static class Impl6930 implements DemoService {public int type() {return 6930;} public void run() {System.out.println(6930);}}
    public static class Impl6931 implements DemoService {public int type() {return 6931;} public void run() {System.out.println(6931);}}
    public static class Impl6932 implements DemoService {public int type() {return 6932;} public void run() {System.out.println(6932);}}
    public static class Impl6933 implements DemoService {public int type() {return 6933;} public void run() {System.out.println(6933);}}
    public static class Impl6934 implements DemoService {public int type() {return 6934;} public void run() {System.out.println(6934);}}
    public static class Impl6935 implements DemoService {public int type() {return 6935;} public void run() {System.out.println(6935);}}
    public static class Impl6936 implements DemoService {public int type() {return 6936;} public void run() {System.out.println(6936);}}
    public static class Impl6937 implements DemoService {public int type() {return 6937;} public void run() {System.out.println(6937);}}
    public static class Impl6938 implements DemoService {public int type() {return 6938;} public void run() {System.out.println(6938);}}
    public static class Impl6939 implements DemoService {public int type() {return 6939;} public void run() {System.out.println(6939);}}
    public static class Impl6940 implements DemoService {public int type() {return 6940;} public void run() {System.out.println(6940);}}
    public static class Impl6941 implements DemoService {public int type() {return 6941;} public void run() {System.out.println(6941);}}
    public static class Impl6942 implements DemoService {public int type() {return 6942;} public void run() {System.out.println(6942);}}
    public static class Impl6943 implements DemoService {public int type() {return 6943;} public void run() {System.out.println(6943);}}
    public static class Impl6944 implements DemoService {public int type() {return 6944;} public void run() {System.out.println(6944);}}
    public static class Impl6945 implements DemoService {public int type() {return 6945;} public void run() {System.out.println(6945);}}
    public static class Impl6946 implements DemoService {public int type() {return 6946;} public void run() {System.out.println(6946);}}
    public static class Impl6947 implements DemoService {public int type() {return 6947;} public void run() {System.out.println(6947);}}
    public static class Impl6948 implements DemoService {public int type() {return 6948;} public void run() {System.out.println(6948);}}
    public static class Impl6949 implements DemoService {public int type() {return 6949;} public void run() {System.out.println(6949);}}
    public static class Impl6950 implements DemoService {public int type() {return 6950;} public void run() {System.out.println(6950);}}
    public static class Impl6951 implements DemoService {public int type() {return 6951;} public void run() {System.out.println(6951);}}
    public static class Impl6952 implements DemoService {public int type() {return 6952;} public void run() {System.out.println(6952);}}
    public static class Impl6953 implements DemoService {public int type() {return 6953;} public void run() {System.out.println(6953);}}
    public static class Impl6954 implements DemoService {public int type() {return 6954;} public void run() {System.out.println(6954);}}
    public static class Impl6955 implements DemoService {public int type() {return 6955;} public void run() {System.out.println(6955);}}
    public static class Impl6956 implements DemoService {public int type() {return 6956;} public void run() {System.out.println(6956);}}
    public static class Impl6957 implements DemoService {public int type() {return 6957;} public void run() {System.out.println(6957);}}
    public static class Impl6958 implements DemoService {public int type() {return 6958;} public void run() {System.out.println(6958);}}
    public static class Impl6959 implements DemoService {public int type() {return 6959;} public void run() {System.out.println(6959);}}
    public static class Impl6960 implements DemoService {public int type() {return 6960;} public void run() {System.out.println(6960);}}
    public static class Impl6961 implements DemoService {public int type() {return 6961;} public void run() {System.out.println(6961);}}
    public static class Impl6962 implements DemoService {public int type() {return 6962;} public void run() {System.out.println(6962);}}
    public static class Impl6963 implements DemoService {public int type() {return 6963;} public void run() {System.out.println(6963);}}
    public static class Impl6964 implements DemoService {public int type() {return 6964;} public void run() {System.out.println(6964);}}
    public static class Impl6965 implements DemoService {public int type() {return 6965;} public void run() {System.out.println(6965);}}
    public static class Impl6966 implements DemoService {public int type() {return 6966;} public void run() {System.out.println(6966);}}
    public static class Impl6967 implements DemoService {public int type() {return 6967;} public void run() {System.out.println(6967);}}
    public static class Impl6968 implements DemoService {public int type() {return 6968;} public void run() {System.out.println(6968);}}
    public static class Impl6969 implements DemoService {public int type() {return 6969;} public void run() {System.out.println(6969);}}
    public static class Impl6970 implements DemoService {public int type() {return 6970;} public void run() {System.out.println(6970);}}
    public static class Impl6971 implements DemoService {public int type() {return 6971;} public void run() {System.out.println(6971);}}
    public static class Impl6972 implements DemoService {public int type() {return 6972;} public void run() {System.out.println(6972);}}
    public static class Impl6973 implements DemoService {public int type() {return 6973;} public void run() {System.out.println(6973);}}
    public static class Impl6974 implements DemoService {public int type() {return 6974;} public void run() {System.out.println(6974);}}
    public static class Impl6975 implements DemoService {public int type() {return 6975;} public void run() {System.out.println(6975);}}
    public static class Impl6976 implements DemoService {public int type() {return 6976;} public void run() {System.out.println(6976);}}
    public static class Impl6977 implements DemoService {public int type() {return 6977;} public void run() {System.out.println(6977);}}
    public static class Impl6978 implements DemoService {public int type() {return 6978;} public void run() {System.out.println(6978);}}
    public static class Impl6979 implements DemoService {public int type() {return 6979;} public void run() {System.out.println(6979);}}
    public static class Impl6980 implements DemoService {public int type() {return 6980;} public void run() {System.out.println(6980);}}
    public static class Impl6981 implements DemoService {public int type() {return 6981;} public void run() {System.out.println(6981);}}
    public static class Impl6982 implements DemoService {public int type() {return 6982;} public void run() {System.out.println(6982);}}
    public static class Impl6983 implements DemoService {public int type() {return 6983;} public void run() {System.out.println(6983);}}
    public static class Impl6984 implements DemoService {public int type() {return 6984;} public void run() {System.out.println(6984);}}
    public static class Impl6985 implements DemoService {public int type() {return 6985;} public void run() {System.out.println(6985);}}
    public static class Impl6986 implements DemoService {public int type() {return 6986;} public void run() {System.out.println(6986);}}
    public static class Impl6987 implements DemoService {public int type() {return 6987;} public void run() {System.out.println(6987);}}
    public static class Impl6988 implements DemoService {public int type() {return 6988;} public void run() {System.out.println(6988);}}
    public static class Impl6989 implements DemoService {public int type() {return 6989;} public void run() {System.out.println(6989);}}
    public static class Impl6990 implements DemoService {public int type() {return 6990;} public void run() {System.out.println(6990);}}
    public static class Impl6991 implements DemoService {public int type() {return 6991;} public void run() {System.out.println(6991);}}
    public static class Impl6992 implements DemoService {public int type() {return 6992;} public void run() {System.out.println(6992);}}
    public static class Impl6993 implements DemoService {public int type() {return 6993;} public void run() {System.out.println(6993);}}
    public static class Impl6994 implements DemoService {public int type() {return 6994;} public void run() {System.out.println(6994);}}
    public static class Impl6995 implements DemoService {public int type() {return 6995;} public void run() {System.out.println(6995);}}
    public static class Impl6996 implements DemoService {public int type() {return 6996;} public void run() {System.out.println(6996);}}
    public static class Impl6997 implements DemoService {public int type() {return 6997;} public void run() {System.out.println(6997);}}
    public static class Impl6998 implements DemoService {public int type() {return 6998;} public void run() {System.out.println(6998);}}
    public static class Impl6999 implements DemoService {public int type() {return 6999;} public void run() {System.out.println(6999);}}
    public static class Impl7000 implements DemoService {public int type() {return 7000;} public void run() {System.out.println(7000);}}
    public static class Impl7001 implements DemoService {public int type() {return 7001;} public void run() {System.out.println(7001);}}
    public static class Impl7002 implements DemoService {public int type() {return 7002;} public void run() {System.out.println(7002);}}
    public static class Impl7003 implements DemoService {public int type() {return 7003;} public void run() {System.out.println(7003);}}
    public static class Impl7004 implements DemoService {public int type() {return 7004;} public void run() {System.out.println(7004);}}
    public static class Impl7005 implements DemoService {public int type() {return 7005;} public void run() {System.out.println(7005);}}
    public static class Impl7006 implements DemoService {public int type() {return 7006;} public void run() {System.out.println(7006);}}
    public static class Impl7007 implements DemoService {public int type() {return 7007;} public void run() {System.out.println(7007);}}
    public static class Impl7008 implements DemoService {public int type() {return 7008;} public void run() {System.out.println(7008);}}
    public static class Impl7009 implements DemoService {public int type() {return 7009;} public void run() {System.out.println(7009);}}
    public static class Impl7010 implements DemoService {public int type() {return 7010;} public void run() {System.out.println(7010);}}
    public static class Impl7011 implements DemoService {public int type() {return 7011;} public void run() {System.out.println(7011);}}
    public static class Impl7012 implements DemoService {public int type() {return 7012;} public void run() {System.out.println(7012);}}
    public static class Impl7013 implements DemoService {public int type() {return 7013;} public void run() {System.out.println(7013);}}
    public static class Impl7014 implements DemoService {public int type() {return 7014;} public void run() {System.out.println(7014);}}
    public static class Impl7015 implements DemoService {public int type() {return 7015;} public void run() {System.out.println(7015);}}
    public static class Impl7016 implements DemoService {public int type() {return 7016;} public void run() {System.out.println(7016);}}
    public static class Impl7017 implements DemoService {public int type() {return 7017;} public void run() {System.out.println(7017);}}
    public static class Impl7018 implements DemoService {public int type() {return 7018;} public void run() {System.out.println(7018);}}
    public static class Impl7019 implements DemoService {public int type() {return 7019;} public void run() {System.out.println(7019);}}
    public static class Impl7020 implements DemoService {public int type() {return 7020;} public void run() {System.out.println(7020);}}
    public static class Impl7021 implements DemoService {public int type() {return 7021;} public void run() {System.out.println(7021);}}
    public static class Impl7022 implements DemoService {public int type() {return 7022;} public void run() {System.out.println(7022);}}
    public static class Impl7023 implements DemoService {public int type() {return 7023;} public void run() {System.out.println(7023);}}
    public static class Impl7024 implements DemoService {public int type() {return 7024;} public void run() {System.out.println(7024);}}
    public static class Impl7025 implements DemoService {public int type() {return 7025;} public void run() {System.out.println(7025);}}
    public static class Impl7026 implements DemoService {public int type() {return 7026;} public void run() {System.out.println(7026);}}
    public static class Impl7027 implements DemoService {public int type() {return 7027;} public void run() {System.out.println(7027);}}
    public static class Impl7028 implements DemoService {public int type() {return 7028;} public void run() {System.out.println(7028);}}
    public static class Impl7029 implements DemoService {public int type() {return 7029;} public void run() {System.out.println(7029);}}
    public static class Impl7030 implements DemoService {public int type() {return 7030;} public void run() {System.out.println(7030);}}
    public static class Impl7031 implements DemoService {public int type() {return 7031;} public void run() {System.out.println(7031);}}
    public static class Impl7032 implements DemoService {public int type() {return 7032;} public void run() {System.out.println(7032);}}
    public static class Impl7033 implements DemoService {public int type() {return 7033;} public void run() {System.out.println(7033);}}
    public static class Impl7034 implements DemoService {public int type() {return 7034;} public void run() {System.out.println(7034);}}
    public static class Impl7035 implements DemoService {public int type() {return 7035;} public void run() {System.out.println(7035);}}
    public static class Impl7036 implements DemoService {public int type() {return 7036;} public void run() {System.out.println(7036);}}
    public static class Impl7037 implements DemoService {public int type() {return 7037;} public void run() {System.out.println(7037);}}
    public static class Impl7038 implements DemoService {public int type() {return 7038;} public void run() {System.out.println(7038);}}
    public static class Impl7039 implements DemoService {public int type() {return 7039;} public void run() {System.out.println(7039);}}
    public static class Impl7040 implements DemoService {public int type() {return 7040;} public void run() {System.out.println(7040);}}
    public static class Impl7041 implements DemoService {public int type() {return 7041;} public void run() {System.out.println(7041);}}
    public static class Impl7042 implements DemoService {public int type() {return 7042;} public void run() {System.out.println(7042);}}
    public static class Impl7043 implements DemoService {public int type() {return 7043;} public void run() {System.out.println(7043);}}
    public static class Impl7044 implements DemoService {public int type() {return 7044;} public void run() {System.out.println(7044);}}
    public static class Impl7045 implements DemoService {public int type() {return 7045;} public void run() {System.out.println(7045);}}
    public static class Impl7046 implements DemoService {public int type() {return 7046;} public void run() {System.out.println(7046);}}
    public static class Impl7047 implements DemoService {public int type() {return 7047;} public void run() {System.out.println(7047);}}
    public static class Impl7048 implements DemoService {public int type() {return 7048;} public void run() {System.out.println(7048);}}
    public static class Impl7049 implements DemoService {public int type() {return 7049;} public void run() {System.out.println(7049);}}
    public static class Impl7050 implements DemoService {public int type() {return 7050;} public void run() {System.out.println(7050);}}
    public static class Impl7051 implements DemoService {public int type() {return 7051;} public void run() {System.out.println(7051);}}
    public static class Impl7052 implements DemoService {public int type() {return 7052;} public void run() {System.out.println(7052);}}
    public static class Impl7053 implements DemoService {public int type() {return 7053;} public void run() {System.out.println(7053);}}
    public static class Impl7054 implements DemoService {public int type() {return 7054;} public void run() {System.out.println(7054);}}
    public static class Impl7055 implements DemoService {public int type() {return 7055;} public void run() {System.out.println(7055);}}
    public static class Impl7056 implements DemoService {public int type() {return 7056;} public void run() {System.out.println(7056);}}
    public static class Impl7057 implements DemoService {public int type() {return 7057;} public void run() {System.out.println(7057);}}
    public static class Impl7058 implements DemoService {public int type() {return 7058;} public void run() {System.out.println(7058);}}
    public static class Impl7059 implements DemoService {public int type() {return 7059;} public void run() {System.out.println(7059);}}
    public static class Impl7060 implements DemoService {public int type() {return 7060;} public void run() {System.out.println(7060);}}
    public static class Impl7061 implements DemoService {public int type() {return 7061;} public void run() {System.out.println(7061);}}
    public static class Impl7062 implements DemoService {public int type() {return 7062;} public void run() {System.out.println(7062);}}
    public static class Impl7063 implements DemoService {public int type() {return 7063;} public void run() {System.out.println(7063);}}
    public static class Impl7064 implements DemoService {public int type() {return 7064;} public void run() {System.out.println(7064);}}
    public static class Impl7065 implements DemoService {public int type() {return 7065;} public void run() {System.out.println(7065);}}
    public static class Impl7066 implements DemoService {public int type() {return 7066;} public void run() {System.out.println(7066);}}
    public static class Impl7067 implements DemoService {public int type() {return 7067;} public void run() {System.out.println(7067);}}
    public static class Impl7068 implements DemoService {public int type() {return 7068;} public void run() {System.out.println(7068);}}
    public static class Impl7069 implements DemoService {public int type() {return 7069;} public void run() {System.out.println(7069);}}
    public static class Impl7070 implements DemoService {public int type() {return 7070;} public void run() {System.out.println(7070);}}
    public static class Impl7071 implements DemoService {public int type() {return 7071;} public void run() {System.out.println(7071);}}
    public static class Impl7072 implements DemoService {public int type() {return 7072;} public void run() {System.out.println(7072);}}
    public static class Impl7073 implements DemoService {public int type() {return 7073;} public void run() {System.out.println(7073);}}
    public static class Impl7074 implements DemoService {public int type() {return 7074;} public void run() {System.out.println(7074);}}
    public static class Impl7075 implements DemoService {public int type() {return 7075;} public void run() {System.out.println(7075);}}
    public static class Impl7076 implements DemoService {public int type() {return 7076;} public void run() {System.out.println(7076);}}
    public static class Impl7077 implements DemoService {public int type() {return 7077;} public void run() {System.out.println(7077);}}
    public static class Impl7078 implements DemoService {public int type() {return 7078;} public void run() {System.out.println(7078);}}
    public static class Impl7079 implements DemoService {public int type() {return 7079;} public void run() {System.out.println(7079);}}
    public static class Impl7080 implements DemoService {public int type() {return 7080;} public void run() {System.out.println(7080);}}
    public static class Impl7081 implements DemoService {public int type() {return 7081;} public void run() {System.out.println(7081);}}
    public static class Impl7082 implements DemoService {public int type() {return 7082;} public void run() {System.out.println(7082);}}
    public static class Impl7083 implements DemoService {public int type() {return 7083;} public void run() {System.out.println(7083);}}
    public static class Impl7084 implements DemoService {public int type() {return 7084;} public void run() {System.out.println(7084);}}
    public static class Impl7085 implements DemoService {public int type() {return 7085;} public void run() {System.out.println(7085);}}
    public static class Impl7086 implements DemoService {public int type() {return 7086;} public void run() {System.out.println(7086);}}
    public static class Impl7087 implements DemoService {public int type() {return 7087;} public void run() {System.out.println(7087);}}
    public static class Impl7088 implements DemoService {public int type() {return 7088;} public void run() {System.out.println(7088);}}
    public static class Impl7089 implements DemoService {public int type() {return 7089;} public void run() {System.out.println(7089);}}
    public static class Impl7090 implements DemoService {public int type() {return 7090;} public void run() {System.out.println(7090);}}
    public static class Impl7091 implements DemoService {public int type() {return 7091;} public void run() {System.out.println(7091);}}
    public static class Impl7092 implements DemoService {public int type() {return 7092;} public void run() {System.out.println(7092);}}
    public static class Impl7093 implements DemoService {public int type() {return 7093;} public void run() {System.out.println(7093);}}
    public static class Impl7094 implements DemoService {public int type() {return 7094;} public void run() {System.out.println(7094);}}
    public static class Impl7095 implements DemoService {public int type() {return 7095;} public void run() {System.out.println(7095);}}
    public static class Impl7096 implements DemoService {public int type() {return 7096;} public void run() {System.out.println(7096);}}
    public static class Impl7097 implements DemoService {public int type() {return 7097;} public void run() {System.out.println(7097);}}
    public static class Impl7098 implements DemoService {public int type() {return 7098;} public void run() {System.out.println(7098);}}
    public static class Impl7099 implements DemoService {public int type() {return 7099;} public void run() {System.out.println(7099);}}
    public static class Impl7100 implements DemoService {public int type() {return 7100;} public void run() {System.out.println(7100);}}
    public static class Impl7101 implements DemoService {public int type() {return 7101;} public void run() {System.out.println(7101);}}
    public static class Impl7102 implements DemoService {public int type() {return 7102;} public void run() {System.out.println(7102);}}
    public static class Impl7103 implements DemoService {public int type() {return 7103;} public void run() {System.out.println(7103);}}
    public static class Impl7104 implements DemoService {public int type() {return 7104;} public void run() {System.out.println(7104);}}
    public static class Impl7105 implements DemoService {public int type() {return 7105;} public void run() {System.out.println(7105);}}
    public static class Impl7106 implements DemoService {public int type() {return 7106;} public void run() {System.out.println(7106);}}
    public static class Impl7107 implements DemoService {public int type() {return 7107;} public void run() {System.out.println(7107);}}
    public static class Impl7108 implements DemoService {public int type() {return 7108;} public void run() {System.out.println(7108);}}
    public static class Impl7109 implements DemoService {public int type() {return 7109;} public void run() {System.out.println(7109);}}
    public static class Impl7110 implements DemoService {public int type() {return 7110;} public void run() {System.out.println(7110);}}
    public static class Impl7111 implements DemoService {public int type() {return 7111;} public void run() {System.out.println(7111);}}
    public static class Impl7112 implements DemoService {public int type() {return 7112;} public void run() {System.out.println(7112);}}
    public static class Impl7113 implements DemoService {public int type() {return 7113;} public void run() {System.out.println(7113);}}
    public static class Impl7114 implements DemoService {public int type() {return 7114;} public void run() {System.out.println(7114);}}
    public static class Impl7115 implements DemoService {public int type() {return 7115;} public void run() {System.out.println(7115);}}
    public static class Impl7116 implements DemoService {public int type() {return 7116;} public void run() {System.out.println(7116);}}
    public static class Impl7117 implements DemoService {public int type() {return 7117;} public void run() {System.out.println(7117);}}
    public static class Impl7118 implements DemoService {public int type() {return 7118;} public void run() {System.out.println(7118);}}
    public static class Impl7119 implements DemoService {public int type() {return 7119;} public void run() {System.out.println(7119);}}
    public static class Impl7120 implements DemoService {public int type() {return 7120;} public void run() {System.out.println(7120);}}
    public static class Impl7121 implements DemoService {public int type() {return 7121;} public void run() {System.out.println(7121);}}
    public static class Impl7122 implements DemoService {public int type() {return 7122;} public void run() {System.out.println(7122);}}
    public static class Impl7123 implements DemoService {public int type() {return 7123;} public void run() {System.out.println(7123);}}
    public static class Impl7124 implements DemoService {public int type() {return 7124;} public void run() {System.out.println(7124);}}
    public static class Impl7125 implements DemoService {public int type() {return 7125;} public void run() {System.out.println(7125);}}
    public static class Impl7126 implements DemoService {public int type() {return 7126;} public void run() {System.out.println(7126);}}
    public static class Impl7127 implements DemoService {public int type() {return 7127;} public void run() {System.out.println(7127);}}
    public static class Impl7128 implements DemoService {public int type() {return 7128;} public void run() {System.out.println(7128);}}
    public static class Impl7129 implements DemoService {public int type() {return 7129;} public void run() {System.out.println(7129);}}
    public static class Impl7130 implements DemoService {public int type() {return 7130;} public void run() {System.out.println(7130);}}
    public static class Impl7131 implements DemoService {public int type() {return 7131;} public void run() {System.out.println(7131);}}
    public static class Impl7132 implements DemoService {public int type() {return 7132;} public void run() {System.out.println(7132);}}
    public static class Impl7133 implements DemoService {public int type() {return 7133;} public void run() {System.out.println(7133);}}
    public static class Impl7134 implements DemoService {public int type() {return 7134;} public void run() {System.out.println(7134);}}
    public static class Impl7135 implements DemoService {public int type() {return 7135;} public void run() {System.out.println(7135);}}
    public static class Impl7136 implements DemoService {public int type() {return 7136;} public void run() {System.out.println(7136);}}
    public static class Impl7137 implements DemoService {public int type() {return 7137;} public void run() {System.out.println(7137);}}
    public static class Impl7138 implements DemoService {public int type() {return 7138;} public void run() {System.out.println(7138);}}
    public static class Impl7139 implements DemoService {public int type() {return 7139;} public void run() {System.out.println(7139);}}
    public static class Impl7140 implements DemoService {public int type() {return 7140;} public void run() {System.out.println(7140);}}
    public static class Impl7141 implements DemoService {public int type() {return 7141;} public void run() {System.out.println(7141);}}
    public static class Impl7142 implements DemoService {public int type() {return 7142;} public void run() {System.out.println(7142);}}
    public static class Impl7143 implements DemoService {public int type() {return 7143;} public void run() {System.out.println(7143);}}
    public static class Impl7144 implements DemoService {public int type() {return 7144;} public void run() {System.out.println(7144);}}
    public static class Impl7145 implements DemoService {public int type() {return 7145;} public void run() {System.out.println(7145);}}
    public static class Impl7146 implements DemoService {public int type() {return 7146;} public void run() {System.out.println(7146);}}
    public static class Impl7147 implements DemoService {public int type() {return 7147;} public void run() {System.out.println(7147);}}
    public static class Impl7148 implements DemoService {public int type() {return 7148;} public void run() {System.out.println(7148);}}
    public static class Impl7149 implements DemoService {public int type() {return 7149;} public void run() {System.out.println(7149);}}
    public static class Impl7150 implements DemoService {public int type() {return 7150;} public void run() {System.out.println(7150);}}
    public static class Impl7151 implements DemoService {public int type() {return 7151;} public void run() {System.out.println(7151);}}
    public static class Impl7152 implements DemoService {public int type() {return 7152;} public void run() {System.out.println(7152);}}
    public static class Impl7153 implements DemoService {public int type() {return 7153;} public void run() {System.out.println(7153);}}
    public static class Impl7154 implements DemoService {public int type() {return 7154;} public void run() {System.out.println(7154);}}
    public static class Impl7155 implements DemoService {public int type() {return 7155;} public void run() {System.out.println(7155);}}
    public static class Impl7156 implements DemoService {public int type() {return 7156;} public void run() {System.out.println(7156);}}
    public static class Impl7157 implements DemoService {public int type() {return 7157;} public void run() {System.out.println(7157);}}
    public static class Impl7158 implements DemoService {public int type() {return 7158;} public void run() {System.out.println(7158);}}
    public static class Impl7159 implements DemoService {public int type() {return 7159;} public void run() {System.out.println(7159);}}
    public static class Impl7160 implements DemoService {public int type() {return 7160;} public void run() {System.out.println(7160);}}
    public static class Impl7161 implements DemoService {public int type() {return 7161;} public void run() {System.out.println(7161);}}
    public static class Impl7162 implements DemoService {public int type() {return 7162;} public void run() {System.out.println(7162);}}
    public static class Impl7163 implements DemoService {public int type() {return 7163;} public void run() {System.out.println(7163);}}
    public static class Impl7164 implements DemoService {public int type() {return 7164;} public void run() {System.out.println(7164);}}
    public static class Impl7165 implements DemoService {public int type() {return 7165;} public void run() {System.out.println(7165);}}
    public static class Impl7166 implements DemoService {public int type() {return 7166;} public void run() {System.out.println(7166);}}
    public static class Impl7167 implements DemoService {public int type() {return 7167;} public void run() {System.out.println(7167);}}
    public static class Impl7168 implements DemoService {public int type() {return 7168;} public void run() {System.out.println(7168);}}
    public static class Impl7169 implements DemoService {public int type() {return 7169;} public void run() {System.out.println(7169);}}
    public static class Impl7170 implements DemoService {public int type() {return 7170;} public void run() {System.out.println(7170);}}
    public static class Impl7171 implements DemoService {public int type() {return 7171;} public void run() {System.out.println(7171);}}
    public static class Impl7172 implements DemoService {public int type() {return 7172;} public void run() {System.out.println(7172);}}
    public static class Impl7173 implements DemoService {public int type() {return 7173;} public void run() {System.out.println(7173);}}
    public static class Impl7174 implements DemoService {public int type() {return 7174;} public void run() {System.out.println(7174);}}
    public static class Impl7175 implements DemoService {public int type() {return 7175;} public void run() {System.out.println(7175);}}
    public static class Impl7176 implements DemoService {public int type() {return 7176;} public void run() {System.out.println(7176);}}
    public static class Impl7177 implements DemoService {public int type() {return 7177;} public void run() {System.out.println(7177);}}
    public static class Impl7178 implements DemoService {public int type() {return 7178;} public void run() {System.out.println(7178);}}
    public static class Impl7179 implements DemoService {public int type() {return 7179;} public void run() {System.out.println(7179);}}
    public static class Impl7180 implements DemoService {public int type() {return 7180;} public void run() {System.out.println(7180);}}
    public static class Impl7181 implements DemoService {public int type() {return 7181;} public void run() {System.out.println(7181);}}
    public static class Impl7182 implements DemoService {public int type() {return 7182;} public void run() {System.out.println(7182);}}
    public static class Impl7183 implements DemoService {public int type() {return 7183;} public void run() {System.out.println(7183);}}
    public static class Impl7184 implements DemoService {public int type() {return 7184;} public void run() {System.out.println(7184);}}
    public static class Impl7185 implements DemoService {public int type() {return 7185;} public void run() {System.out.println(7185);}}
    public static class Impl7186 implements DemoService {public int type() {return 7186;} public void run() {System.out.println(7186);}}
    public static class Impl7187 implements DemoService {public int type() {return 7187;} public void run() {System.out.println(7187);}}
    public static class Impl7188 implements DemoService {public int type() {return 7188;} public void run() {System.out.println(7188);}}
    public static class Impl7189 implements DemoService {public int type() {return 7189;} public void run() {System.out.println(7189);}}
    public static class Impl7190 implements DemoService {public int type() {return 7190;} public void run() {System.out.println(7190);}}
    public static class Impl7191 implements DemoService {public int type() {return 7191;} public void run() {System.out.println(7191);}}
    public static class Impl7192 implements DemoService {public int type() {return 7192;} public void run() {System.out.println(7192);}}
    public static class Impl7193 implements DemoService {public int type() {return 7193;} public void run() {System.out.println(7193);}}
    public static class Impl7194 implements DemoService {public int type() {return 7194;} public void run() {System.out.println(7194);}}
    public static class Impl7195 implements DemoService {public int type() {return 7195;} public void run() {System.out.println(7195);}}
    public static class Impl7196 implements DemoService {public int type() {return 7196;} public void run() {System.out.println(7196);}}
    public static class Impl7197 implements DemoService {public int type() {return 7197;} public void run() {System.out.println(7197);}}
    public static class Impl7198 implements DemoService {public int type() {return 7198;} public void run() {System.out.println(7198);}}
    public static class Impl7199 implements DemoService {public int type() {return 7199;} public void run() {System.out.println(7199);}}
    public static class Impl7200 implements DemoService {public int type() {return 7200;} public void run() {System.out.println(7200);}}
    public static class Impl7201 implements DemoService {public int type() {return 7201;} public void run() {System.out.println(7201);}}
    public static class Impl7202 implements DemoService {public int type() {return 7202;} public void run() {System.out.println(7202);}}
    public static class Impl7203 implements DemoService {public int type() {return 7203;} public void run() {System.out.println(7203);}}
    public static class Impl7204 implements DemoService {public int type() {return 7204;} public void run() {System.out.println(7204);}}
    public static class Impl7205 implements DemoService {public int type() {return 7205;} public void run() {System.out.println(7205);}}
    public static class Impl7206 implements DemoService {public int type() {return 7206;} public void run() {System.out.println(7206);}}
    public static class Impl7207 implements DemoService {public int type() {return 7207;} public void run() {System.out.println(7207);}}
    public static class Impl7208 implements DemoService {public int type() {return 7208;} public void run() {System.out.println(7208);}}
    public static class Impl7209 implements DemoService {public int type() {return 7209;} public void run() {System.out.println(7209);}}
    public static class Impl7210 implements DemoService {public int type() {return 7210;} public void run() {System.out.println(7210);}}
    public static class Impl7211 implements DemoService {public int type() {return 7211;} public void run() {System.out.println(7211);}}
    public static class Impl7212 implements DemoService {public int type() {return 7212;} public void run() {System.out.println(7212);}}
    public static class Impl7213 implements DemoService {public int type() {return 7213;} public void run() {System.out.println(7213);}}
    public static class Impl7214 implements DemoService {public int type() {return 7214;} public void run() {System.out.println(7214);}}
    public static class Impl7215 implements DemoService {public int type() {return 7215;} public void run() {System.out.println(7215);}}
    public static class Impl7216 implements DemoService {public int type() {return 7216;} public void run() {System.out.println(7216);}}
    public static class Impl7217 implements DemoService {public int type() {return 7217;} public void run() {System.out.println(7217);}}
    public static class Impl7218 implements DemoService {public int type() {return 7218;} public void run() {System.out.println(7218);}}
    public static class Impl7219 implements DemoService {public int type() {return 7219;} public void run() {System.out.println(7219);}}
    public static class Impl7220 implements DemoService {public int type() {return 7220;} public void run() {System.out.println(7220);}}
    public static class Impl7221 implements DemoService {public int type() {return 7221;} public void run() {System.out.println(7221);}}
    public static class Impl7222 implements DemoService {public int type() {return 7222;} public void run() {System.out.println(7222);}}
    public static class Impl7223 implements DemoService {public int type() {return 7223;} public void run() {System.out.println(7223);}}
    public static class Impl7224 implements DemoService {public int type() {return 7224;} public void run() {System.out.println(7224);}}
    public static class Impl7225 implements DemoService {public int type() {return 7225;} public void run() {System.out.println(7225);}}
    public static class Impl7226 implements DemoService {public int type() {return 7226;} public void run() {System.out.println(7226);}}
    public static class Impl7227 implements DemoService {public int type() {return 7227;} public void run() {System.out.println(7227);}}
    public static class Impl7228 implements DemoService {public int type() {return 7228;} public void run() {System.out.println(7228);}}
    public static class Impl7229 implements DemoService {public int type() {return 7229;} public void run() {System.out.println(7229);}}
    public static class Impl7230 implements DemoService {public int type() {return 7230;} public void run() {System.out.println(7230);}}
    public static class Impl7231 implements DemoService {public int type() {return 7231;} public void run() {System.out.println(7231);}}
    public static class Impl7232 implements DemoService {public int type() {return 7232;} public void run() {System.out.println(7232);}}
    public static class Impl7233 implements DemoService {public int type() {return 7233;} public void run() {System.out.println(7233);}}
    public static class Impl7234 implements DemoService {public int type() {return 7234;} public void run() {System.out.println(7234);}}
    public static class Impl7235 implements DemoService {public int type() {return 7235;} public void run() {System.out.println(7235);}}
    public static class Impl7236 implements DemoService {public int type() {return 7236;} public void run() {System.out.println(7236);}}
    public static class Impl7237 implements DemoService {public int type() {return 7237;} public void run() {System.out.println(7237);}}
    public static class Impl7238 implements DemoService {public int type() {return 7238;} public void run() {System.out.println(7238);}}
    public static class Impl7239 implements DemoService {public int type() {return 7239;} public void run() {System.out.println(7239);}}
    public static class Impl7240 implements DemoService {public int type() {return 7240;} public void run() {System.out.println(7240);}}
    public static class Impl7241 implements DemoService {public int type() {return 7241;} public void run() {System.out.println(7241);}}
    public static class Impl7242 implements DemoService {public int type() {return 7242;} public void run() {System.out.println(7242);}}
    public static class Impl7243 implements DemoService {public int type() {return 7243;} public void run() {System.out.println(7243);}}
    public static class Impl7244 implements DemoService {public int type() {return 7244;} public void run() {System.out.println(7244);}}
    public static class Impl7245 implements DemoService {public int type() {return 7245;} public void run() {System.out.println(7245);}}
    public static class Impl7246 implements DemoService {public int type() {return 7246;} public void run() {System.out.println(7246);}}
    public static class Impl7247 implements DemoService {public int type() {return 7247;} public void run() {System.out.println(7247);}}
    public static class Impl7248 implements DemoService {public int type() {return 7248;} public void run() {System.out.println(7248);}}
    public static class Impl7249 implements DemoService {public int type() {return 7249;} public void run() {System.out.println(7249);}}
    public static class Impl7250 implements DemoService {public int type() {return 7250;} public void run() {System.out.println(7250);}}
    public static class Impl7251 implements DemoService {public int type() {return 7251;} public void run() {System.out.println(7251);}}
    public static class Impl7252 implements DemoService {public int type() {return 7252;} public void run() {System.out.println(7252);}}
    public static class Impl7253 implements DemoService {public int type() {return 7253;} public void run() {System.out.println(7253);}}
    public static class Impl7254 implements DemoService {public int type() {return 7254;} public void run() {System.out.println(7254);}}
    public static class Impl7255 implements DemoService {public int type() {return 7255;} public void run() {System.out.println(7255);}}
    public static class Impl7256 implements DemoService {public int type() {return 7256;} public void run() {System.out.println(7256);}}
    public static class Impl7257 implements DemoService {public int type() {return 7257;} public void run() {System.out.println(7257);}}
    public static class Impl7258 implements DemoService {public int type() {return 7258;} public void run() {System.out.println(7258);}}
    public static class Impl7259 implements DemoService {public int type() {return 7259;} public void run() {System.out.println(7259);}}
    public static class Impl7260 implements DemoService {public int type() {return 7260;} public void run() {System.out.println(7260);}}
    public static class Impl7261 implements DemoService {public int type() {return 7261;} public void run() {System.out.println(7261);}}
    public static class Impl7262 implements DemoService {public int type() {return 7262;} public void run() {System.out.println(7262);}}
    public static class Impl7263 implements DemoService {public int type() {return 7263;} public void run() {System.out.println(7263);}}
    public static class Impl7264 implements DemoService {public int type() {return 7264;} public void run() {System.out.println(7264);}}
    public static class Impl7265 implements DemoService {public int type() {return 7265;} public void run() {System.out.println(7265);}}
    public static class Impl7266 implements DemoService {public int type() {return 7266;} public void run() {System.out.println(7266);}}
    public static class Impl7267 implements DemoService {public int type() {return 7267;} public void run() {System.out.println(7267);}}
    public static class Impl7268 implements DemoService {public int type() {return 7268;} public void run() {System.out.println(7268);}}
    public static class Impl7269 implements DemoService {public int type() {return 7269;} public void run() {System.out.println(7269);}}
    public static class Impl7270 implements DemoService {public int type() {return 7270;} public void run() {System.out.println(7270);}}
    public static class Impl7271 implements DemoService {public int type() {return 7271;} public void run() {System.out.println(7271);}}
    public static class Impl7272 implements DemoService {public int type() {return 7272;} public void run() {System.out.println(7272);}}
    public static class Impl7273 implements DemoService {public int type() {return 7273;} public void run() {System.out.println(7273);}}
    public static class Impl7274 implements DemoService {public int type() {return 7274;} public void run() {System.out.println(7274);}}
    public static class Impl7275 implements DemoService {public int type() {return 7275;} public void run() {System.out.println(7275);}}
    public static class Impl7276 implements DemoService {public int type() {return 7276;} public void run() {System.out.println(7276);}}
    public static class Impl7277 implements DemoService {public int type() {return 7277;} public void run() {System.out.println(7277);}}
    public static class Impl7278 implements DemoService {public int type() {return 7278;} public void run() {System.out.println(7278);}}
    public static class Impl7279 implements DemoService {public int type() {return 7279;} public void run() {System.out.println(7279);}}
    public static class Impl7280 implements DemoService {public int type() {return 7280;} public void run() {System.out.println(7280);}}
    public static class Impl7281 implements DemoService {public int type() {return 7281;} public void run() {System.out.println(7281);}}
    public static class Impl7282 implements DemoService {public int type() {return 7282;} public void run() {System.out.println(7282);}}
    public static class Impl7283 implements DemoService {public int type() {return 7283;} public void run() {System.out.println(7283);}}
    public static class Impl7284 implements DemoService {public int type() {return 7284;} public void run() {System.out.println(7284);}}
    public static class Impl7285 implements DemoService {public int type() {return 7285;} public void run() {System.out.println(7285);}}
    public static class Impl7286 implements DemoService {public int type() {return 7286;} public void run() {System.out.println(7286);}}
    public static class Impl7287 implements DemoService {public int type() {return 7287;} public void run() {System.out.println(7287);}}
    public static class Impl7288 implements DemoService {public int type() {return 7288;} public void run() {System.out.println(7288);}}
    public static class Impl7289 implements DemoService {public int type() {return 7289;} public void run() {System.out.println(7289);}}
    public static class Impl7290 implements DemoService {public int type() {return 7290;} public void run() {System.out.println(7290);}}
    public static class Impl7291 implements DemoService {public int type() {return 7291;} public void run() {System.out.println(7291);}}
    public static class Impl7292 implements DemoService {public int type() {return 7292;} public void run() {System.out.println(7292);}}
    public static class Impl7293 implements DemoService {public int type() {return 7293;} public void run() {System.out.println(7293);}}
    public static class Impl7294 implements DemoService {public int type() {return 7294;} public void run() {System.out.println(7294);}}
    public static class Impl7295 implements DemoService {public int type() {return 7295;} public void run() {System.out.println(7295);}}
    public static class Impl7296 implements DemoService {public int type() {return 7296;} public void run() {System.out.println(7296);}}
    public static class Impl7297 implements DemoService {public int type() {return 7297;} public void run() {System.out.println(7297);}}
    public static class Impl7298 implements DemoService {public int type() {return 7298;} public void run() {System.out.println(7298);}}
    public static class Impl7299 implements DemoService {public int type() {return 7299;} public void run() {System.out.println(7299);}}
    public static class Impl7300 implements DemoService {public int type() {return 7300;} public void run() {System.out.println(7300);}}
    public static class Impl7301 implements DemoService {public int type() {return 7301;} public void run() {System.out.println(7301);}}
    public static class Impl7302 implements DemoService {public int type() {return 7302;} public void run() {System.out.println(7302);}}
    public static class Impl7303 implements DemoService {public int type() {return 7303;} public void run() {System.out.println(7303);}}
    public static class Impl7304 implements DemoService {public int type() {return 7304;} public void run() {System.out.println(7304);}}
    public static class Impl7305 implements DemoService {public int type() {return 7305;} public void run() {System.out.println(7305);}}
    public static class Impl7306 implements DemoService {public int type() {return 7306;} public void run() {System.out.println(7306);}}
    public static class Impl7307 implements DemoService {public int type() {return 7307;} public void run() {System.out.println(7307);}}
    public static class Impl7308 implements DemoService {public int type() {return 7308;} public void run() {System.out.println(7308);}}
    public static class Impl7309 implements DemoService {public int type() {return 7309;} public void run() {System.out.println(7309);}}
    public static class Impl7310 implements DemoService {public int type() {return 7310;} public void run() {System.out.println(7310);}}
    public static class Impl7311 implements DemoService {public int type() {return 7311;} public void run() {System.out.println(7311);}}
    public static class Impl7312 implements DemoService {public int type() {return 7312;} public void run() {System.out.println(7312);}}
    public static class Impl7313 implements DemoService {public int type() {return 7313;} public void run() {System.out.println(7313);}}
    public static class Impl7314 implements DemoService {public int type() {return 7314;} public void run() {System.out.println(7314);}}
    public static class Impl7315 implements DemoService {public int type() {return 7315;} public void run() {System.out.println(7315);}}
    public static class Impl7316 implements DemoService {public int type() {return 7316;} public void run() {System.out.println(7316);}}
    public static class Impl7317 implements DemoService {public int type() {return 7317;} public void run() {System.out.println(7317);}}
    public static class Impl7318 implements DemoService {public int type() {return 7318;} public void run() {System.out.println(7318);}}
    public static class Impl7319 implements DemoService {public int type() {return 7319;} public void run() {System.out.println(7319);}}
    public static class Impl7320 implements DemoService {public int type() {return 7320;} public void run() {System.out.println(7320);}}
    public static class Impl7321 implements DemoService {public int type() {return 7321;} public void run() {System.out.println(7321);}}
    public static class Impl7322 implements DemoService {public int type() {return 7322;} public void run() {System.out.println(7322);}}
    public static class Impl7323 implements DemoService {public int type() {return 7323;} public void run() {System.out.println(7323);}}
    public static class Impl7324 implements DemoService {public int type() {return 7324;} public void run() {System.out.println(7324);}}
    public static class Impl7325 implements DemoService {public int type() {return 7325;} public void run() {System.out.println(7325);}}
    public static class Impl7326 implements DemoService {public int type() {return 7326;} public void run() {System.out.println(7326);}}
    public static class Impl7327 implements DemoService {public int type() {return 7327;} public void run() {System.out.println(7327);}}
    public static class Impl7328 implements DemoService {public int type() {return 7328;} public void run() {System.out.println(7328);}}
    public static class Impl7329 implements DemoService {public int type() {return 7329;} public void run() {System.out.println(7329);}}
    public static class Impl7330 implements DemoService {public int type() {return 7330;} public void run() {System.out.println(7330);}}
    public static class Impl7331 implements DemoService {public int type() {return 7331;} public void run() {System.out.println(7331);}}
    public static class Impl7332 implements DemoService {public int type() {return 7332;} public void run() {System.out.println(7332);}}
    public static class Impl7333 implements DemoService {public int type() {return 7333;} public void run() {System.out.println(7333);}}
    public static class Impl7334 implements DemoService {public int type() {return 7334;} public void run() {System.out.println(7334);}}
    public static class Impl7335 implements DemoService {public int type() {return 7335;} public void run() {System.out.println(7335);}}
    public static class Impl7336 implements DemoService {public int type() {return 7336;} public void run() {System.out.println(7336);}}
    public static class Impl7337 implements DemoService {public int type() {return 7337;} public void run() {System.out.println(7337);}}
    public static class Impl7338 implements DemoService {public int type() {return 7338;} public void run() {System.out.println(7338);}}
    public static class Impl7339 implements DemoService {public int type() {return 7339;} public void run() {System.out.println(7339);}}
    public static class Impl7340 implements DemoService {public int type() {return 7340;} public void run() {System.out.println(7340);}}
    public static class Impl7341 implements DemoService {public int type() {return 7341;} public void run() {System.out.println(7341);}}
    public static class Impl7342 implements DemoService {public int type() {return 7342;} public void run() {System.out.println(7342);}}
    public static class Impl7343 implements DemoService {public int type() {return 7343;} public void run() {System.out.println(7343);}}
    public static class Impl7344 implements DemoService {public int type() {return 7344;} public void run() {System.out.println(7344);}}
    public static class Impl7345 implements DemoService {public int type() {return 7345;} public void run() {System.out.println(7345);}}
    public static class Impl7346 implements DemoService {public int type() {return 7346;} public void run() {System.out.println(7346);}}
    public static class Impl7347 implements DemoService {public int type() {return 7347;} public void run() {System.out.println(7347);}}
    public static class Impl7348 implements DemoService {public int type() {return 7348;} public void run() {System.out.println(7348);}}
    public static class Impl7349 implements DemoService {public int type() {return 7349;} public void run() {System.out.println(7349);}}
    public static class Impl7350 implements DemoService {public int type() {return 7350;} public void run() {System.out.println(7350);}}
    public static class Impl7351 implements DemoService {public int type() {return 7351;} public void run() {System.out.println(7351);}}
    public static class Impl7352 implements DemoService {public int type() {return 7352;} public void run() {System.out.println(7352);}}
    public static class Impl7353 implements DemoService {public int type() {return 7353;} public void run() {System.out.println(7353);}}
    public static class Impl7354 implements DemoService {public int type() {return 7354;} public void run() {System.out.println(7354);}}
    public static class Impl7355 implements DemoService {public int type() {return 7355;} public void run() {System.out.println(7355);}}
    public static class Impl7356 implements DemoService {public int type() {return 7356;} public void run() {System.out.println(7356);}}
    public static class Impl7357 implements DemoService {public int type() {return 7357;} public void run() {System.out.println(7357);}}
    public static class Impl7358 implements DemoService {public int type() {return 7358;} public void run() {System.out.println(7358);}}
    public static class Impl7359 implements DemoService {public int type() {return 7359;} public void run() {System.out.println(7359);}}
    public static class Impl7360 implements DemoService {public int type() {return 7360;} public void run() {System.out.println(7360);}}
    public static class Impl7361 implements DemoService {public int type() {return 7361;} public void run() {System.out.println(7361);}}
    public static class Impl7362 implements DemoService {public int type() {return 7362;} public void run() {System.out.println(7362);}}
    public static class Impl7363 implements DemoService {public int type() {return 7363;} public void run() {System.out.println(7363);}}
    public static class Impl7364 implements DemoService {public int type() {return 7364;} public void run() {System.out.println(7364);}}
    public static class Impl7365 implements DemoService {public int type() {return 7365;} public void run() {System.out.println(7365);}}
    public static class Impl7366 implements DemoService {public int type() {return 7366;} public void run() {System.out.println(7366);}}
    public static class Impl7367 implements DemoService {public int type() {return 7367;} public void run() {System.out.println(7367);}}
    public static class Impl7368 implements DemoService {public int type() {return 7368;} public void run() {System.out.println(7368);}}
    public static class Impl7369 implements DemoService {public int type() {return 7369;} public void run() {System.out.println(7369);}}
    public static class Impl7370 implements DemoService {public int type() {return 7370;} public void run() {System.out.println(7370);}}
    public static class Impl7371 implements DemoService {public int type() {return 7371;} public void run() {System.out.println(7371);}}
    public static class Impl7372 implements DemoService {public int type() {return 7372;} public void run() {System.out.println(7372);}}
    public static class Impl7373 implements DemoService {public int type() {return 7373;} public void run() {System.out.println(7373);}}
    public static class Impl7374 implements DemoService {public int type() {return 7374;} public void run() {System.out.println(7374);}}
    public static class Impl7375 implements DemoService {public int type() {return 7375;} public void run() {System.out.println(7375);}}
    public static class Impl7376 implements DemoService {public int type() {return 7376;} public void run() {System.out.println(7376);}}
    public static class Impl7377 implements DemoService {public int type() {return 7377;} public void run() {System.out.println(7377);}}
    public static class Impl7378 implements DemoService {public int type() {return 7378;} public void run() {System.out.println(7378);}}
    public static class Impl7379 implements DemoService {public int type() {return 7379;} public void run() {System.out.println(7379);}}
    public static class Impl7380 implements DemoService {public int type() {return 7380;} public void run() {System.out.println(7380);}}
    public static class Impl7381 implements DemoService {public int type() {return 7381;} public void run() {System.out.println(7381);}}
    public static class Impl7382 implements DemoService {public int type() {return 7382;} public void run() {System.out.println(7382);}}
    public static class Impl7383 implements DemoService {public int type() {return 7383;} public void run() {System.out.println(7383);}}
    public static class Impl7384 implements DemoService {public int type() {return 7384;} public void run() {System.out.println(7384);}}
    public static class Impl7385 implements DemoService {public int type() {return 7385;} public void run() {System.out.println(7385);}}
    public static class Impl7386 implements DemoService {public int type() {return 7386;} public void run() {System.out.println(7386);}}
    public static class Impl7387 implements DemoService {public int type() {return 7387;} public void run() {System.out.println(7387);}}
    public static class Impl7388 implements DemoService {public int type() {return 7388;} public void run() {System.out.println(7388);}}
    public static class Impl7389 implements DemoService {public int type() {return 7389;} public void run() {System.out.println(7389);}}
    public static class Impl7390 implements DemoService {public int type() {return 7390;} public void run() {System.out.println(7390);}}
    public static class Impl7391 implements DemoService {public int type() {return 7391;} public void run() {System.out.println(7391);}}
    public static class Impl7392 implements DemoService {public int type() {return 7392;} public void run() {System.out.println(7392);}}
    public static class Impl7393 implements DemoService {public int type() {return 7393;} public void run() {System.out.println(7393);}}
    public static class Impl7394 implements DemoService {public int type() {return 7394;} public void run() {System.out.println(7394);}}
    public static class Impl7395 implements DemoService {public int type() {return 7395;} public void run() {System.out.println(7395);}}
    public static class Impl7396 implements DemoService {public int type() {return 7396;} public void run() {System.out.println(7396);}}
    public static class Impl7397 implements DemoService {public int type() {return 7397;} public void run() {System.out.println(7397);}}
    public static class Impl7398 implements DemoService {public int type() {return 7398;} public void run() {System.out.println(7398);}}
    public static class Impl7399 implements DemoService {public int type() {return 7399;} public void run() {System.out.println(7399);}}
    public static class Impl7400 implements DemoService {public int type() {return 7400;} public void run() {System.out.println(7400);}}
    public static class Impl7401 implements DemoService {public int type() {return 7401;} public void run() {System.out.println(7401);}}
    public static class Impl7402 implements DemoService {public int type() {return 7402;} public void run() {System.out.println(7402);}}
    public static class Impl7403 implements DemoService {public int type() {return 7403;} public void run() {System.out.println(7403);}}
    public static class Impl7404 implements DemoService {public int type() {return 7404;} public void run() {System.out.println(7404);}}
    public static class Impl7405 implements DemoService {public int type() {return 7405;} public void run() {System.out.println(7405);}}
    public static class Impl7406 implements DemoService {public int type() {return 7406;} public void run() {System.out.println(7406);}}
    public static class Impl7407 implements DemoService {public int type() {return 7407;} public void run() {System.out.println(7407);}}
    public static class Impl7408 implements DemoService {public int type() {return 7408;} public void run() {System.out.println(7408);}}
    public static class Impl7409 implements DemoService {public int type() {return 7409;} public void run() {System.out.println(7409);}}
    public static class Impl7410 implements DemoService {public int type() {return 7410;} public void run() {System.out.println(7410);}}
    public static class Impl7411 implements DemoService {public int type() {return 7411;} public void run() {System.out.println(7411);}}
    public static class Impl7412 implements DemoService {public int type() {return 7412;} public void run() {System.out.println(7412);}}
    public static class Impl7413 implements DemoService {public int type() {return 7413;} public void run() {System.out.println(7413);}}
    public static class Impl7414 implements DemoService {public int type() {return 7414;} public void run() {System.out.println(7414);}}
    public static class Impl7415 implements DemoService {public int type() {return 7415;} public void run() {System.out.println(7415);}}
    public static class Impl7416 implements DemoService {public int type() {return 7416;} public void run() {System.out.println(7416);}}
    public static class Impl7417 implements DemoService {public int type() {return 7417;} public void run() {System.out.println(7417);}}
    public static class Impl7418 implements DemoService {public int type() {return 7418;} public void run() {System.out.println(7418);}}
    public static class Impl7419 implements DemoService {public int type() {return 7419;} public void run() {System.out.println(7419);}}
    public static class Impl7420 implements DemoService {public int type() {return 7420;} public void run() {System.out.println(7420);}}
    public static class Impl7421 implements DemoService {public int type() {return 7421;} public void run() {System.out.println(7421);}}
    public static class Impl7422 implements DemoService {public int type() {return 7422;} public void run() {System.out.println(7422);}}
    public static class Impl7423 implements DemoService {public int type() {return 7423;} public void run() {System.out.println(7423);}}
    public static class Impl7424 implements DemoService {public int type() {return 7424;} public void run() {System.out.println(7424);}}
    public static class Impl7425 implements DemoService {public int type() {return 7425;} public void run() {System.out.println(7425);}}
    public static class Impl7426 implements DemoService {public int type() {return 7426;} public void run() {System.out.println(7426);}}
    public static class Impl7427 implements DemoService {public int type() {return 7427;} public void run() {System.out.println(7427);}}
    public static class Impl7428 implements DemoService {public int type() {return 7428;} public void run() {System.out.println(7428);}}
    public static class Impl7429 implements DemoService {public int type() {return 7429;} public void run() {System.out.println(7429);}}
    public static class Impl7430 implements DemoService {public int type() {return 7430;} public void run() {System.out.println(7430);}}
    public static class Impl7431 implements DemoService {public int type() {return 7431;} public void run() {System.out.println(7431);}}
    public static class Impl7432 implements DemoService {public int type() {return 7432;} public void run() {System.out.println(7432);}}
    public static class Impl7433 implements DemoService {public int type() {return 7433;} public void run() {System.out.println(7433);}}
    public static class Impl7434 implements DemoService {public int type() {return 7434;} public void run() {System.out.println(7434);}}
    public static class Impl7435 implements DemoService {public int type() {return 7435;} public void run() {System.out.println(7435);}}
    public static class Impl7436 implements DemoService {public int type() {return 7436;} public void run() {System.out.println(7436);}}
    public static class Impl7437 implements DemoService {public int type() {return 7437;} public void run() {System.out.println(7437);}}
    public static class Impl7438 implements DemoService {public int type() {return 7438;} public void run() {System.out.println(7438);}}
    public static class Impl7439 implements DemoService {public int type() {return 7439;} public void run() {System.out.println(7439);}}
    public static class Impl7440 implements DemoService {public int type() {return 7440;} public void run() {System.out.println(7440);}}
    public static class Impl7441 implements DemoService {public int type() {return 7441;} public void run() {System.out.println(7441);}}
    public static class Impl7442 implements DemoService {public int type() {return 7442;} public void run() {System.out.println(7442);}}
    public static class Impl7443 implements DemoService {public int type() {return 7443;} public void run() {System.out.println(7443);}}
    public static class Impl7444 implements DemoService {public int type() {return 7444;} public void run() {System.out.println(7444);}}
    public static class Impl7445 implements DemoService {public int type() {return 7445;} public void run() {System.out.println(7445);}}
    public static class Impl7446 implements DemoService {public int type() {return 7446;} public void run() {System.out.println(7446);}}
    public static class Impl7447 implements DemoService {public int type() {return 7447;} public void run() {System.out.println(7447);}}
    public static class Impl7448 implements DemoService {public int type() {return 7448;} public void run() {System.out.println(7448);}}
    public static class Impl7449 implements DemoService {public int type() {return 7449;} public void run() {System.out.println(7449);}}
    public static class Impl7450 implements DemoService {public int type() {return 7450;} public void run() {System.out.println(7450);}}
    public static class Impl7451 implements DemoService {public int type() {return 7451;} public void run() {System.out.println(7451);}}
    public static class Impl7452 implements DemoService {public int type() {return 7452;} public void run() {System.out.println(7452);}}
    public static class Impl7453 implements DemoService {public int type() {return 7453;} public void run() {System.out.println(7453);}}
    public static class Impl7454 implements DemoService {public int type() {return 7454;} public void run() {System.out.println(7454);}}
    public static class Impl7455 implements DemoService {public int type() {return 7455;} public void run() {System.out.println(7455);}}
    public static class Impl7456 implements DemoService {public int type() {return 7456;} public void run() {System.out.println(7456);}}
    public static class Impl7457 implements DemoService {public int type() {return 7457;} public void run() {System.out.println(7457);}}
    public static class Impl7458 implements DemoService {public int type() {return 7458;} public void run() {System.out.println(7458);}}
    public static class Impl7459 implements DemoService {public int type() {return 7459;} public void run() {System.out.println(7459);}}
    public static class Impl7460 implements DemoService {public int type() {return 7460;} public void run() {System.out.println(7460);}}
    public static class Impl7461 implements DemoService {public int type() {return 7461;} public void run() {System.out.println(7461);}}
    public static class Impl7462 implements DemoService {public int type() {return 7462;} public void run() {System.out.println(7462);}}
    public static class Impl7463 implements DemoService {public int type() {return 7463;} public void run() {System.out.println(7463);}}
    public static class Impl7464 implements DemoService {public int type() {return 7464;} public void run() {System.out.println(7464);}}
    public static class Impl7465 implements DemoService {public int type() {return 7465;} public void run() {System.out.println(7465);}}
    public static class Impl7466 implements DemoService {public int type() {return 7466;} public void run() {System.out.println(7466);}}
    public static class Impl7467 implements DemoService {public int type() {return 7467;} public void run() {System.out.println(7467);}}
    public static class Impl7468 implements DemoService {public int type() {return 7468;} public void run() {System.out.println(7468);}}
    public static class Impl7469 implements DemoService {public int type() {return 7469;} public void run() {System.out.println(7469);}}
    public static class Impl7470 implements DemoService {public int type() {return 7470;} public void run() {System.out.println(7470);}}
    public static class Impl7471 implements DemoService {public int type() {return 7471;} public void run() {System.out.println(7471);}}
    public static class Impl7472 implements DemoService {public int type() {return 7472;} public void run() {System.out.println(7472);}}
    public static class Impl7473 implements DemoService {public int type() {return 7473;} public void run() {System.out.println(7473);}}
    public static class Impl7474 implements DemoService {public int type() {return 7474;} public void run() {System.out.println(7474);}}
    public static class Impl7475 implements DemoService {public int type() {return 7475;} public void run() {System.out.println(7475);}}
    public static class Impl7476 implements DemoService {public int type() {return 7476;} public void run() {System.out.println(7476);}}
    public static class Impl7477 implements DemoService {public int type() {return 7477;} public void run() {System.out.println(7477);}}
    public static class Impl7478 implements DemoService {public int type() {return 7478;} public void run() {System.out.println(7478);}}
    public static class Impl7479 implements DemoService {public int type() {return 7479;} public void run() {System.out.println(7479);}}
    public static class Impl7480 implements DemoService {public int type() {return 7480;} public void run() {System.out.println(7480);}}
    public static class Impl7481 implements DemoService {public int type() {return 7481;} public void run() {System.out.println(7481);}}
    public static class Impl7482 implements DemoService {public int type() {return 7482;} public void run() {System.out.println(7482);}}
    public static class Impl7483 implements DemoService {public int type() {return 7483;} public void run() {System.out.println(7483);}}
    public static class Impl7484 implements DemoService {public int type() {return 7484;} public void run() {System.out.println(7484);}}
    public static class Impl7485 implements DemoService {public int type() {return 7485;} public void run() {System.out.println(7485);}}
    public static class Impl7486 implements DemoService {public int type() {return 7486;} public void run() {System.out.println(7486);}}
    public static class Impl7487 implements DemoService {public int type() {return 7487;} public void run() {System.out.println(7487);}}
    public static class Impl7488 implements DemoService {public int type() {return 7488;} public void run() {System.out.println(7488);}}
    public static class Impl7489 implements DemoService {public int type() {return 7489;} public void run() {System.out.println(7489);}}
    public static class Impl7490 implements DemoService {public int type() {return 7490;} public void run() {System.out.println(7490);}}
    public static class Impl7491 implements DemoService {public int type() {return 7491;} public void run() {System.out.println(7491);}}
    public static class Impl7492 implements DemoService {public int type() {return 7492;} public void run() {System.out.println(7492);}}
    public static class Impl7493 implements DemoService {public int type() {return 7493;} public void run() {System.out.println(7493);}}
    public static class Impl7494 implements DemoService {public int type() {return 7494;} public void run() {System.out.println(7494);}}
    public static class Impl7495 implements DemoService {public int type() {return 7495;} public void run() {System.out.println(7495);}}
    public static class Impl7496 implements DemoService {public int type() {return 7496;} public void run() {System.out.println(7496);}}
    public static class Impl7497 implements DemoService {public int type() {return 7497;} public void run() {System.out.println(7497);}}
    public static class Impl7498 implements DemoService {public int type() {return 7498;} public void run() {System.out.println(7498);}}
    public static class Impl7499 implements DemoService {public int type() {return 7499;} public void run() {System.out.println(7499);}}
    public static class Impl7500 implements DemoService {public int type() {return 7500;} public void run() {System.out.println(7500);}}
    public static class Impl7501 implements DemoService {public int type() {return 7501;} public void run() {System.out.println(7501);}}
    public static class Impl7502 implements DemoService {public int type() {return 7502;} public void run() {System.out.println(7502);}}
    public static class Impl7503 implements DemoService {public int type() {return 7503;} public void run() {System.out.println(7503);}}
    public static class Impl7504 implements DemoService {public int type() {return 7504;} public void run() {System.out.println(7504);}}
    public static class Impl7505 implements DemoService {public int type() {return 7505;} public void run() {System.out.println(7505);}}
    public static class Impl7506 implements DemoService {public int type() {return 7506;} public void run() {System.out.println(7506);}}
    public static class Impl7507 implements DemoService {public int type() {return 7507;} public void run() {System.out.println(7507);}}
    public static class Impl7508 implements DemoService {public int type() {return 7508;} public void run() {System.out.println(7508);}}
    public static class Impl7509 implements DemoService {public int type() {return 7509;} public void run() {System.out.println(7509);}}
    public static class Impl7510 implements DemoService {public int type() {return 7510;} public void run() {System.out.println(7510);}}
    public static class Impl7511 implements DemoService {public int type() {return 7511;} public void run() {System.out.println(7511);}}
    public static class Impl7512 implements DemoService {public int type() {return 7512;} public void run() {System.out.println(7512);}}
    public static class Impl7513 implements DemoService {public int type() {return 7513;} public void run() {System.out.println(7513);}}
    public static class Impl7514 implements DemoService {public int type() {return 7514;} public void run() {System.out.println(7514);}}
    public static class Impl7515 implements DemoService {public int type() {return 7515;} public void run() {System.out.println(7515);}}
    public static class Impl7516 implements DemoService {public int type() {return 7516;} public void run() {System.out.println(7516);}}
    public static class Impl7517 implements DemoService {public int type() {return 7517;} public void run() {System.out.println(7517);}}
    public static class Impl7518 implements DemoService {public int type() {return 7518;} public void run() {System.out.println(7518);}}
    public static class Impl7519 implements DemoService {public int type() {return 7519;} public void run() {System.out.println(7519);}}
    public static class Impl7520 implements DemoService {public int type() {return 7520;} public void run() {System.out.println(7520);}}
    public static class Impl7521 implements DemoService {public int type() {return 7521;} public void run() {System.out.println(7521);}}
    public static class Impl7522 implements DemoService {public int type() {return 7522;} public void run() {System.out.println(7522);}}
    public static class Impl7523 implements DemoService {public int type() {return 7523;} public void run() {System.out.println(7523);}}
    public static class Impl7524 implements DemoService {public int type() {return 7524;} public void run() {System.out.println(7524);}}
    public static class Impl7525 implements DemoService {public int type() {return 7525;} public void run() {System.out.println(7525);}}
    public static class Impl7526 implements DemoService {public int type() {return 7526;} public void run() {System.out.println(7526);}}
    public static class Impl7527 implements DemoService {public int type() {return 7527;} public void run() {System.out.println(7527);}}
    public static class Impl7528 implements DemoService {public int type() {return 7528;} public void run() {System.out.println(7528);}}
    public static class Impl7529 implements DemoService {public int type() {return 7529;} public void run() {System.out.println(7529);}}
    public static class Impl7530 implements DemoService {public int type() {return 7530;} public void run() {System.out.println(7530);}}
    public static class Impl7531 implements DemoService {public int type() {return 7531;} public void run() {System.out.println(7531);}}
    public static class Impl7532 implements DemoService {public int type() {return 7532;} public void run() {System.out.println(7532);}}
    public static class Impl7533 implements DemoService {public int type() {return 7533;} public void run() {System.out.println(7533);}}
    public static class Impl7534 implements DemoService {public int type() {return 7534;} public void run() {System.out.println(7534);}}
    public static class Impl7535 implements DemoService {public int type() {return 7535;} public void run() {System.out.println(7535);}}
    public static class Impl7536 implements DemoService {public int type() {return 7536;} public void run() {System.out.println(7536);}}
    public static class Impl7537 implements DemoService {public int type() {return 7537;} public void run() {System.out.println(7537);}}
    public static class Impl7538 implements DemoService {public int type() {return 7538;} public void run() {System.out.println(7538);}}
    public static class Impl7539 implements DemoService {public int type() {return 7539;} public void run() {System.out.println(7539);}}
    public static class Impl7540 implements DemoService {public int type() {return 7540;} public void run() {System.out.println(7540);}}
    public static class Impl7541 implements DemoService {public int type() {return 7541;} public void run() {System.out.println(7541);}}
    public static class Impl7542 implements DemoService {public int type() {return 7542;} public void run() {System.out.println(7542);}}
    public static class Impl7543 implements DemoService {public int type() {return 7543;} public void run() {System.out.println(7543);}}
    public static class Impl7544 implements DemoService {public int type() {return 7544;} public void run() {System.out.println(7544);}}
    public static class Impl7545 implements DemoService {public int type() {return 7545;} public void run() {System.out.println(7545);}}
    public static class Impl7546 implements DemoService {public int type() {return 7546;} public void run() {System.out.println(7546);}}
    public static class Impl7547 implements DemoService {public int type() {return 7547;} public void run() {System.out.println(7547);}}
    public static class Impl7548 implements DemoService {public int type() {return 7548;} public void run() {System.out.println(7548);}}
    public static class Impl7549 implements DemoService {public int type() {return 7549;} public void run() {System.out.println(7549);}}
    public static class Impl7550 implements DemoService {public int type() {return 7550;} public void run() {System.out.println(7550);}}
    public static class Impl7551 implements DemoService {public int type() {return 7551;} public void run() {System.out.println(7551);}}
    public static class Impl7552 implements DemoService {public int type() {return 7552;} public void run() {System.out.println(7552);}}
    public static class Impl7553 implements DemoService {public int type() {return 7553;} public void run() {System.out.println(7553);}}
    public static class Impl7554 implements DemoService {public int type() {return 7554;} public void run() {System.out.println(7554);}}
    public static class Impl7555 implements DemoService {public int type() {return 7555;} public void run() {System.out.println(7555);}}
    public static class Impl7556 implements DemoService {public int type() {return 7556;} public void run() {System.out.println(7556);}}
    public static class Impl7557 implements DemoService {public int type() {return 7557;} public void run() {System.out.println(7557);}}
    public static class Impl7558 implements DemoService {public int type() {return 7558;} public void run() {System.out.println(7558);}}
    public static class Impl7559 implements DemoService {public int type() {return 7559;} public void run() {System.out.println(7559);}}
    public static class Impl7560 implements DemoService {public int type() {return 7560;} public void run() {System.out.println(7560);}}
    public static class Impl7561 implements DemoService {public int type() {return 7561;} public void run() {System.out.println(7561);}}
    public static class Impl7562 implements DemoService {public int type() {return 7562;} public void run() {System.out.println(7562);}}
    public static class Impl7563 implements DemoService {public int type() {return 7563;} public void run() {System.out.println(7563);}}
    public static class Impl7564 implements DemoService {public int type() {return 7564;} public void run() {System.out.println(7564);}}
    public static class Impl7565 implements DemoService {public int type() {return 7565;} public void run() {System.out.println(7565);}}
    public static class Impl7566 implements DemoService {public int type() {return 7566;} public void run() {System.out.println(7566);}}
    public static class Impl7567 implements DemoService {public int type() {return 7567;} public void run() {System.out.println(7567);}}
    public static class Impl7568 implements DemoService {public int type() {return 7568;} public void run() {System.out.println(7568);}}
    public static class Impl7569 implements DemoService {public int type() {return 7569;} public void run() {System.out.println(7569);}}
    public static class Impl7570 implements DemoService {public int type() {return 7570;} public void run() {System.out.println(7570);}}
    public static class Impl7571 implements DemoService {public int type() {return 7571;} public void run() {System.out.println(7571);}}
    public static class Impl7572 implements DemoService {public int type() {return 7572;} public void run() {System.out.println(7572);}}
    public static class Impl7573 implements DemoService {public int type() {return 7573;} public void run() {System.out.println(7573);}}
    public static class Impl7574 implements DemoService {public int type() {return 7574;} public void run() {System.out.println(7574);}}
    public static class Impl7575 implements DemoService {public int type() {return 7575;} public void run() {System.out.println(7575);}}
    public static class Impl7576 implements DemoService {public int type() {return 7576;} public void run() {System.out.println(7576);}}
    public static class Impl7577 implements DemoService {public int type() {return 7577;} public void run() {System.out.println(7577);}}
    public static class Impl7578 implements DemoService {public int type() {return 7578;} public void run() {System.out.println(7578);}}
    public static class Impl7579 implements DemoService {public int type() {return 7579;} public void run() {System.out.println(7579);}}
    public static class Impl7580 implements DemoService {public int type() {return 7580;} public void run() {System.out.println(7580);}}
    public static class Impl7581 implements DemoService {public int type() {return 7581;} public void run() {System.out.println(7581);}}
    public static class Impl7582 implements DemoService {public int type() {return 7582;} public void run() {System.out.println(7582);}}
    public static class Impl7583 implements DemoService {public int type() {return 7583;} public void run() {System.out.println(7583);}}
    public static class Impl7584 implements DemoService {public int type() {return 7584;} public void run() {System.out.println(7584);}}
    public static class Impl7585 implements DemoService {public int type() {return 7585;} public void run() {System.out.println(7585);}}
    public static class Impl7586 implements DemoService {public int type() {return 7586;} public void run() {System.out.println(7586);}}
    public static class Impl7587 implements DemoService {public int type() {return 7587;} public void run() {System.out.println(7587);}}
    public static class Impl7588 implements DemoService {public int type() {return 7588;} public void run() {System.out.println(7588);}}
    public static class Impl7589 implements DemoService {public int type() {return 7589;} public void run() {System.out.println(7589);}}
    public static class Impl7590 implements DemoService {public int type() {return 7590;} public void run() {System.out.println(7590);}}
    public static class Impl7591 implements DemoService {public int type() {return 7591;} public void run() {System.out.println(7591);}}
    public static class Impl7592 implements DemoService {public int type() {return 7592;} public void run() {System.out.println(7592);}}
    public static class Impl7593 implements DemoService {public int type() {return 7593;} public void run() {System.out.println(7593);}}
    public static class Impl7594 implements DemoService {public int type() {return 7594;} public void run() {System.out.println(7594);}}
    public static class Impl7595 implements DemoService {public int type() {return 7595;} public void run() {System.out.println(7595);}}
    public static class Impl7596 implements DemoService {public int type() {return 7596;} public void run() {System.out.println(7596);}}
    public static class Impl7597 implements DemoService {public int type() {return 7597;} public void run() {System.out.println(7597);}}
    public static class Impl7598 implements DemoService {public int type() {return 7598;} public void run() {System.out.println(7598);}}
    public static class Impl7599 implements DemoService {public int type() {return 7599;} public void run() {System.out.println(7599);}}
    public static class Impl7600 implements DemoService {public int type() {return 7600;} public void run() {System.out.println(7600);}}
    public static class Impl7601 implements DemoService {public int type() {return 7601;} public void run() {System.out.println(7601);}}
    public static class Impl7602 implements DemoService {public int type() {return 7602;} public void run() {System.out.println(7602);}}
    public static class Impl7603 implements DemoService {public int type() {return 7603;} public void run() {System.out.println(7603);}}
    public static class Impl7604 implements DemoService {public int type() {return 7604;} public void run() {System.out.println(7604);}}
    public static class Impl7605 implements DemoService {public int type() {return 7605;} public void run() {System.out.println(7605);}}
    public static class Impl7606 implements DemoService {public int type() {return 7606;} public void run() {System.out.println(7606);}}
    public static class Impl7607 implements DemoService {public int type() {return 7607;} public void run() {System.out.println(7607);}}
    public static class Impl7608 implements DemoService {public int type() {return 7608;} public void run() {System.out.println(7608);}}
    public static class Impl7609 implements DemoService {public int type() {return 7609;} public void run() {System.out.println(7609);}}
    public static class Impl7610 implements DemoService {public int type() {return 7610;} public void run() {System.out.println(7610);}}
    public static class Impl7611 implements DemoService {public int type() {return 7611;} public void run() {System.out.println(7611);}}
    public static class Impl7612 implements DemoService {public int type() {return 7612;} public void run() {System.out.println(7612);}}
    public static class Impl7613 implements DemoService {public int type() {return 7613;} public void run() {System.out.println(7613);}}
    public static class Impl7614 implements DemoService {public int type() {return 7614;} public void run() {System.out.println(7614);}}
    public static class Impl7615 implements DemoService {public int type() {return 7615;} public void run() {System.out.println(7615);}}
    public static class Impl7616 implements DemoService {public int type() {return 7616;} public void run() {System.out.println(7616);}}
    public static class Impl7617 implements DemoService {public int type() {return 7617;} public void run() {System.out.println(7617);}}
    public static class Impl7618 implements DemoService {public int type() {return 7618;} public void run() {System.out.println(7618);}}
    public static class Impl7619 implements DemoService {public int type() {return 7619;} public void run() {System.out.println(7619);}}
    public static class Impl7620 implements DemoService {public int type() {return 7620;} public void run() {System.out.println(7620);}}
    public static class Impl7621 implements DemoService {public int type() {return 7621;} public void run() {System.out.println(7621);}}
    public static class Impl7622 implements DemoService {public int type() {return 7622;} public void run() {System.out.println(7622);}}
    public static class Impl7623 implements DemoService {public int type() {return 7623;} public void run() {System.out.println(7623);}}
    public static class Impl7624 implements DemoService {public int type() {return 7624;} public void run() {System.out.println(7624);}}
    public static class Impl7625 implements DemoService {public int type() {return 7625;} public void run() {System.out.println(7625);}}
    public static class Impl7626 implements DemoService {public int type() {return 7626;} public void run() {System.out.println(7626);}}
    public static class Impl7627 implements DemoService {public int type() {return 7627;} public void run() {System.out.println(7627);}}
    public static class Impl7628 implements DemoService {public int type() {return 7628;} public void run() {System.out.println(7628);}}
    public static class Impl7629 implements DemoService {public int type() {return 7629;} public void run() {System.out.println(7629);}}
    public static class Impl7630 implements DemoService {public int type() {return 7630;} public void run() {System.out.println(7630);}}
    public static class Impl7631 implements DemoService {public int type() {return 7631;} public void run() {System.out.println(7631);}}
    public static class Impl7632 implements DemoService {public int type() {return 7632;} public void run() {System.out.println(7632);}}
    public static class Impl7633 implements DemoService {public int type() {return 7633;} public void run() {System.out.println(7633);}}
    public static class Impl7634 implements DemoService {public int type() {return 7634;} public void run() {System.out.println(7634);}}
    public static class Impl7635 implements DemoService {public int type() {return 7635;} public void run() {System.out.println(7635);}}
    public static class Impl7636 implements DemoService {public int type() {return 7636;} public void run() {System.out.println(7636);}}
    public static class Impl7637 implements DemoService {public int type() {return 7637;} public void run() {System.out.println(7637);}}
    public static class Impl7638 implements DemoService {public int type() {return 7638;} public void run() {System.out.println(7638);}}
    public static class Impl7639 implements DemoService {public int type() {return 7639;} public void run() {System.out.println(7639);}}
    public static class Impl7640 implements DemoService {public int type() {return 7640;} public void run() {System.out.println(7640);}}
    public static class Impl7641 implements DemoService {public int type() {return 7641;} public void run() {System.out.println(7641);}}
    public static class Impl7642 implements DemoService {public int type() {return 7642;} public void run() {System.out.println(7642);}}
    public static class Impl7643 implements DemoService {public int type() {return 7643;} public void run() {System.out.println(7643);}}
    public static class Impl7644 implements DemoService {public int type() {return 7644;} public void run() {System.out.println(7644);}}
    public static class Impl7645 implements DemoService {public int type() {return 7645;} public void run() {System.out.println(7645);}}
    public static class Impl7646 implements DemoService {public int type() {return 7646;} public void run() {System.out.println(7646);}}
    public static class Impl7647 implements DemoService {public int type() {return 7647;} public void run() {System.out.println(7647);}}
    public static class Impl7648 implements DemoService {public int type() {return 7648;} public void run() {System.out.println(7648);}}
    public static class Impl7649 implements DemoService {public int type() {return 7649;} public void run() {System.out.println(7649);}}
    public static class Impl7650 implements DemoService {public int type() {return 7650;} public void run() {System.out.println(7650);}}
    public static class Impl7651 implements DemoService {public int type() {return 7651;} public void run() {System.out.println(7651);}}
    public static class Impl7652 implements DemoService {public int type() {return 7652;} public void run() {System.out.println(7652);}}
    public static class Impl7653 implements DemoService {public int type() {return 7653;} public void run() {System.out.println(7653);}}
    public static class Impl7654 implements DemoService {public int type() {return 7654;} public void run() {System.out.println(7654);}}
    public static class Impl7655 implements DemoService {public int type() {return 7655;} public void run() {System.out.println(7655);}}
    public static class Impl7656 implements DemoService {public int type() {return 7656;} public void run() {System.out.println(7656);}}
    public static class Impl7657 implements DemoService {public int type() {return 7657;} public void run() {System.out.println(7657);}}
    public static class Impl7658 implements DemoService {public int type() {return 7658;} public void run() {System.out.println(7658);}}
    public static class Impl7659 implements DemoService {public int type() {return 7659;} public void run() {System.out.println(7659);}}
    public static class Impl7660 implements DemoService {public int type() {return 7660;} public void run() {System.out.println(7660);}}
    public static class Impl7661 implements DemoService {public int type() {return 7661;} public void run() {System.out.println(7661);}}
    public static class Impl7662 implements DemoService {public int type() {return 7662;} public void run() {System.out.println(7662);}}
    public static class Impl7663 implements DemoService {public int type() {return 7663;} public void run() {System.out.println(7663);}}
    public static class Impl7664 implements DemoService {public int type() {return 7664;} public void run() {System.out.println(7664);}}
    public static class Impl7665 implements DemoService {public int type() {return 7665;} public void run() {System.out.println(7665);}}
    public static class Impl7666 implements DemoService {public int type() {return 7666;} public void run() {System.out.println(7666);}}
    public static class Impl7667 implements DemoService {public int type() {return 7667;} public void run() {System.out.println(7667);}}
    public static class Impl7668 implements DemoService {public int type() {return 7668;} public void run() {System.out.println(7668);}}
    public static class Impl7669 implements DemoService {public int type() {return 7669;} public void run() {System.out.println(7669);}}
    public static class Impl7670 implements DemoService {public int type() {return 7670;} public void run() {System.out.println(7670);}}
    public static class Impl7671 implements DemoService {public int type() {return 7671;} public void run() {System.out.println(7671);}}
    public static class Impl7672 implements DemoService {public int type() {return 7672;} public void run() {System.out.println(7672);}}
    public static class Impl7673 implements DemoService {public int type() {return 7673;} public void run() {System.out.println(7673);}}
    public static class Impl7674 implements DemoService {public int type() {return 7674;} public void run() {System.out.println(7674);}}
    public static class Impl7675 implements DemoService {public int type() {return 7675;} public void run() {System.out.println(7675);}}
    public static class Impl7676 implements DemoService {public int type() {return 7676;} public void run() {System.out.println(7676);}}
    public static class Impl7677 implements DemoService {public int type() {return 7677;} public void run() {System.out.println(7677);}}
    public static class Impl7678 implements DemoService {public int type() {return 7678;} public void run() {System.out.println(7678);}}
    public static class Impl7679 implements DemoService {public int type() {return 7679;} public void run() {System.out.println(7679);}}
    public static class Impl7680 implements DemoService {public int type() {return 7680;} public void run() {System.out.println(7680);}}
    public static class Impl7681 implements DemoService {public int type() {return 7681;} public void run() {System.out.println(7681);}}
    public static class Impl7682 implements DemoService {public int type() {return 7682;} public void run() {System.out.println(7682);}}
    public static class Impl7683 implements DemoService {public int type() {return 7683;} public void run() {System.out.println(7683);}}
    public static class Impl7684 implements DemoService {public int type() {return 7684;} public void run() {System.out.println(7684);}}
    public static class Impl7685 implements DemoService {public int type() {return 7685;} public void run() {System.out.println(7685);}}
    public static class Impl7686 implements DemoService {public int type() {return 7686;} public void run() {System.out.println(7686);}}
    public static class Impl7687 implements DemoService {public int type() {return 7687;} public void run() {System.out.println(7687);}}
    public static class Impl7688 implements DemoService {public int type() {return 7688;} public void run() {System.out.println(7688);}}
    public static class Impl7689 implements DemoService {public int type() {return 7689;} public void run() {System.out.println(7689);}}
    public static class Impl7690 implements DemoService {public int type() {return 7690;} public void run() {System.out.println(7690);}}
    public static class Impl7691 implements DemoService {public int type() {return 7691;} public void run() {System.out.println(7691);}}
    public static class Impl7692 implements DemoService {public int type() {return 7692;} public void run() {System.out.println(7692);}}
    public static class Impl7693 implements DemoService {public int type() {return 7693;} public void run() {System.out.println(7693);}}
    public static class Impl7694 implements DemoService {public int type() {return 7694;} public void run() {System.out.println(7694);}}
    public static class Impl7695 implements DemoService {public int type() {return 7695;} public void run() {System.out.println(7695);}}
    public static class Impl7696 implements DemoService {public int type() {return 7696;} public void run() {System.out.println(7696);}}
    public static class Impl7697 implements DemoService {public int type() {return 7697;} public void run() {System.out.println(7697);}}
    public static class Impl7698 implements DemoService {public int type() {return 7698;} public void run() {System.out.println(7698);}}
    public static class Impl7699 implements DemoService {public int type() {return 7699;} public void run() {System.out.println(7699);}}
    public static class Impl7700 implements DemoService {public int type() {return 7700;} public void run() {System.out.println(7700);}}
    public static class Impl7701 implements DemoService {public int type() {return 7701;} public void run() {System.out.println(7701);}}
    public static class Impl7702 implements DemoService {public int type() {return 7702;} public void run() {System.out.println(7702);}}
    public static class Impl7703 implements DemoService {public int type() {return 7703;} public void run() {System.out.println(7703);}}
    public static class Impl7704 implements DemoService {public int type() {return 7704;} public void run() {System.out.println(7704);}}
    public static class Impl7705 implements DemoService {public int type() {return 7705;} public void run() {System.out.println(7705);}}
    public static class Impl7706 implements DemoService {public int type() {return 7706;} public void run() {System.out.println(7706);}}
    public static class Impl7707 implements DemoService {public int type() {return 7707;} public void run() {System.out.println(7707);}}
    public static class Impl7708 implements DemoService {public int type() {return 7708;} public void run() {System.out.println(7708);}}
    public static class Impl7709 implements DemoService {public int type() {return 7709;} public void run() {System.out.println(7709);}}
    public static class Impl7710 implements DemoService {public int type() {return 7710;} public void run() {System.out.println(7710);}}
    public static class Impl7711 implements DemoService {public int type() {return 7711;} public void run() {System.out.println(7711);}}
    public static class Impl7712 implements DemoService {public int type() {return 7712;} public void run() {System.out.println(7712);}}
    public static class Impl7713 implements DemoService {public int type() {return 7713;} public void run() {System.out.println(7713);}}
    public static class Impl7714 implements DemoService {public int type() {return 7714;} public void run() {System.out.println(7714);}}
    public static class Impl7715 implements DemoService {public int type() {return 7715;} public void run() {System.out.println(7715);}}
    public static class Impl7716 implements DemoService {public int type() {return 7716;} public void run() {System.out.println(7716);}}
    public static class Impl7717 implements DemoService {public int type() {return 7717;} public void run() {System.out.println(7717);}}
    public static class Impl7718 implements DemoService {public int type() {return 7718;} public void run() {System.out.println(7718);}}
    public static class Impl7719 implements DemoService {public int type() {return 7719;} public void run() {System.out.println(7719);}}
    public static class Impl7720 implements DemoService {public int type() {return 7720;} public void run() {System.out.println(7720);}}
    public static class Impl7721 implements DemoService {public int type() {return 7721;} public void run() {System.out.println(7721);}}
    public static class Impl7722 implements DemoService {public int type() {return 7722;} public void run() {System.out.println(7722);}}
    public static class Impl7723 implements DemoService {public int type() {return 7723;} public void run() {System.out.println(7723);}}
    public static class Impl7724 implements DemoService {public int type() {return 7724;} public void run() {System.out.println(7724);}}
    public static class Impl7725 implements DemoService {public int type() {return 7725;} public void run() {System.out.println(7725);}}
    public static class Impl7726 implements DemoService {public int type() {return 7726;} public void run() {System.out.println(7726);}}
    public static class Impl7727 implements DemoService {public int type() {return 7727;} public void run() {System.out.println(7727);}}
    public static class Impl7728 implements DemoService {public int type() {return 7728;} public void run() {System.out.println(7728);}}
    public static class Impl7729 implements DemoService {public int type() {return 7729;} public void run() {System.out.println(7729);}}
    public static class Impl7730 implements DemoService {public int type() {return 7730;} public void run() {System.out.println(7730);}}
    public static class Impl7731 implements DemoService {public int type() {return 7731;} public void run() {System.out.println(7731);}}
    public static class Impl7732 implements DemoService {public int type() {return 7732;} public void run() {System.out.println(7732);}}
    public static class Impl7733 implements DemoService {public int type() {return 7733;} public void run() {System.out.println(7733);}}
    public static class Impl7734 implements DemoService {public int type() {return 7734;} public void run() {System.out.println(7734);}}
    public static class Impl7735 implements DemoService {public int type() {return 7735;} public void run() {System.out.println(7735);}}
    public static class Impl7736 implements DemoService {public int type() {return 7736;} public void run() {System.out.println(7736);}}
    public static class Impl7737 implements DemoService {public int type() {return 7737;} public void run() {System.out.println(7737);}}
    public static class Impl7738 implements DemoService {public int type() {return 7738;} public void run() {System.out.println(7738);}}
    public static class Impl7739 implements DemoService {public int type() {return 7739;} public void run() {System.out.println(7739);}}
    public static class Impl7740 implements DemoService {public int type() {return 7740;} public void run() {System.out.println(7740);}}
    public static class Impl7741 implements DemoService {public int type() {return 7741;} public void run() {System.out.println(7741);}}
    public static class Impl7742 implements DemoService {public int type() {return 7742;} public void run() {System.out.println(7742);}}
    public static class Impl7743 implements DemoService {public int type() {return 7743;} public void run() {System.out.println(7743);}}
    public static class Impl7744 implements DemoService {public int type() {return 7744;} public void run() {System.out.println(7744);}}
    public static class Impl7745 implements DemoService {public int type() {return 7745;} public void run() {System.out.println(7745);}}
    public static class Impl7746 implements DemoService {public int type() {return 7746;} public void run() {System.out.println(7746);}}
    public static class Impl7747 implements DemoService {public int type() {return 7747;} public void run() {System.out.println(7747);}}
    public static class Impl7748 implements DemoService {public int type() {return 7748;} public void run() {System.out.println(7748);}}
    public static class Impl7749 implements DemoService {public int type() {return 7749;} public void run() {System.out.println(7749);}}
    public static class Impl7750 implements DemoService {public int type() {return 7750;} public void run() {System.out.println(7750);}}
    public static class Impl7751 implements DemoService {public int type() {return 7751;} public void run() {System.out.println(7751);}}
    public static class Impl7752 implements DemoService {public int type() {return 7752;} public void run() {System.out.println(7752);}}
    public static class Impl7753 implements DemoService {public int type() {return 7753;} public void run() {System.out.println(7753);}}
    public static class Impl7754 implements DemoService {public int type() {return 7754;} public void run() {System.out.println(7754);}}
    public static class Impl7755 implements DemoService {public int type() {return 7755;} public void run() {System.out.println(7755);}}
    public static class Impl7756 implements DemoService {public int type() {return 7756;} public void run() {System.out.println(7756);}}
    public static class Impl7757 implements DemoService {public int type() {return 7757;} public void run() {System.out.println(7757);}}
    public static class Impl7758 implements DemoService {public int type() {return 7758;} public void run() {System.out.println(7758);}}
    public static class Impl7759 implements DemoService {public int type() {return 7759;} public void run() {System.out.println(7759);}}
    public static class Impl7760 implements DemoService {public int type() {return 7760;} public void run() {System.out.println(7760);}}
    public static class Impl7761 implements DemoService {public int type() {return 7761;} public void run() {System.out.println(7761);}}
    public static class Impl7762 implements DemoService {public int type() {return 7762;} public void run() {System.out.println(7762);}}
    public static class Impl7763 implements DemoService {public int type() {return 7763;} public void run() {System.out.println(7763);}}
    public static class Impl7764 implements DemoService {public int type() {return 7764;} public void run() {System.out.println(7764);}}
    public static class Impl7765 implements DemoService {public int type() {return 7765;} public void run() {System.out.println(7765);}}
    public static class Impl7766 implements DemoService {public int type() {return 7766;} public void run() {System.out.println(7766);}}
    public static class Impl7767 implements DemoService {public int type() {return 7767;} public void run() {System.out.println(7767);}}
    public static class Impl7768 implements DemoService {public int type() {return 7768;} public void run() {System.out.println(7768);}}
    public static class Impl7769 implements DemoService {public int type() {return 7769;} public void run() {System.out.println(7769);}}
    public static class Impl7770 implements DemoService {public int type() {return 7770;} public void run() {System.out.println(7770);}}
    public static class Impl7771 implements DemoService {public int type() {return 7771;} public void run() {System.out.println(7771);}}
    public static class Impl7772 implements DemoService {public int type() {return 7772;} public void run() {System.out.println(7772);}}
    public static class Impl7773 implements DemoService {public int type() {return 7773;} public void run() {System.out.println(7773);}}
    public static class Impl7774 implements DemoService {public int type() {return 7774;} public void run() {System.out.println(7774);}}
    public static class Impl7775 implements DemoService {public int type() {return 7775;} public void run() {System.out.println(7775);}}
    public static class Impl7776 implements DemoService {public int type() {return 7776;} public void run() {System.out.println(7776);}}
    public static class Impl7777 implements DemoService {public int type() {return 7777;} public void run() {System.out.println(7777);}}
    public static class Impl7778 implements DemoService {public int type() {return 7778;} public void run() {System.out.println(7778);}}
    public static class Impl7779 implements DemoService {public int type() {return 7779;} public void run() {System.out.println(7779);}}
    public static class Impl7780 implements DemoService {public int type() {return 7780;} public void run() {System.out.println(7780);}}
    public static class Impl7781 implements DemoService {public int type() {return 7781;} public void run() {System.out.println(7781);}}
    public static class Impl7782 implements DemoService {public int type() {return 7782;} public void run() {System.out.println(7782);}}
    public static class Impl7783 implements DemoService {public int type() {return 7783;} public void run() {System.out.println(7783);}}
    public static class Impl7784 implements DemoService {public int type() {return 7784;} public void run() {System.out.println(7784);}}
    public static class Impl7785 implements DemoService {public int type() {return 7785;} public void run() {System.out.println(7785);}}
    public static class Impl7786 implements DemoService {public int type() {return 7786;} public void run() {System.out.println(7786);}}
    public static class Impl7787 implements DemoService {public int type() {return 7787;} public void run() {System.out.println(7787);}}
    public static class Impl7788 implements DemoService {public int type() {return 7788;} public void run() {System.out.println(7788);}}
    public static class Impl7789 implements DemoService {public int type() {return 7789;} public void run() {System.out.println(7789);}}
    public static class Impl7790 implements DemoService {public int type() {return 7790;} public void run() {System.out.println(7790);}}
    public static class Impl7791 implements DemoService {public int type() {return 7791;} public void run() {System.out.println(7791);}}
    public static class Impl7792 implements DemoService {public int type() {return 7792;} public void run() {System.out.println(7792);}}
    public static class Impl7793 implements DemoService {public int type() {return 7793;} public void run() {System.out.println(7793);}}
    public static class Impl7794 implements DemoService {public int type() {return 7794;} public void run() {System.out.println(7794);}}
    public static class Impl7795 implements DemoService {public int type() {return 7795;} public void run() {System.out.println(7795);}}
    public static class Impl7796 implements DemoService {public int type() {return 7796;} public void run() {System.out.println(7796);}}
    public static class Impl7797 implements DemoService {public int type() {return 7797;} public void run() {System.out.println(7797);}}
    public static class Impl7798 implements DemoService {public int type() {return 7798;} public void run() {System.out.println(7798);}}
    public static class Impl7799 implements DemoService {public int type() {return 7799;} public void run() {System.out.println(7799);}}
    public static class Impl7800 implements DemoService {public int type() {return 7800;} public void run() {System.out.println(7800);}}
    public static class Impl7801 implements DemoService {public int type() {return 7801;} public void run() {System.out.println(7801);}}
    public static class Impl7802 implements DemoService {public int type() {return 7802;} public void run() {System.out.println(7802);}}
    public static class Impl7803 implements DemoService {public int type() {return 7803;} public void run() {System.out.println(7803);}}
    public static class Impl7804 implements DemoService {public int type() {return 7804;} public void run() {System.out.println(7804);}}
    public static class Impl7805 implements DemoService {public int type() {return 7805;} public void run() {System.out.println(7805);}}
    public static class Impl7806 implements DemoService {public int type() {return 7806;} public void run() {System.out.println(7806);}}
    public static class Impl7807 implements DemoService {public int type() {return 7807;} public void run() {System.out.println(7807);}}
    public static class Impl7808 implements DemoService {public int type() {return 7808;} public void run() {System.out.println(7808);}}
    public static class Impl7809 implements DemoService {public int type() {return 7809;} public void run() {System.out.println(7809);}}
    public static class Impl7810 implements DemoService {public int type() {return 7810;} public void run() {System.out.println(7810);}}
    public static class Impl7811 implements DemoService {public int type() {return 7811;} public void run() {System.out.println(7811);}}
    public static class Impl7812 implements DemoService {public int type() {return 7812;} public void run() {System.out.println(7812);}}
    public static class Impl7813 implements DemoService {public int type() {return 7813;} public void run() {System.out.println(7813);}}
    public static class Impl7814 implements DemoService {public int type() {return 7814;} public void run() {System.out.println(7814);}}
    public static class Impl7815 implements DemoService {public int type() {return 7815;} public void run() {System.out.println(7815);}}
    public static class Impl7816 implements DemoService {public int type() {return 7816;} public void run() {System.out.println(7816);}}
    public static class Impl7817 implements DemoService {public int type() {return 7817;} public void run() {System.out.println(7817);}}
    public static class Impl7818 implements DemoService {public int type() {return 7818;} public void run() {System.out.println(7818);}}
    public static class Impl7819 implements DemoService {public int type() {return 7819;} public void run() {System.out.println(7819);}}
    public static class Impl7820 implements DemoService {public int type() {return 7820;} public void run() {System.out.println(7820);}}
    public static class Impl7821 implements DemoService {public int type() {return 7821;} public void run() {System.out.println(7821);}}
    public static class Impl7822 implements DemoService {public int type() {return 7822;} public void run() {System.out.println(7822);}}
    public static class Impl7823 implements DemoService {public int type() {return 7823;} public void run() {System.out.println(7823);}}
    public static class Impl7824 implements DemoService {public int type() {return 7824;} public void run() {System.out.println(7824);}}
    public static class Impl7825 implements DemoService {public int type() {return 7825;} public void run() {System.out.println(7825);}}
    public static class Impl7826 implements DemoService {public int type() {return 7826;} public void run() {System.out.println(7826);}}
    public static class Impl7827 implements DemoService {public int type() {return 7827;} public void run() {System.out.println(7827);}}
    public static class Impl7828 implements DemoService {public int type() {return 7828;} public void run() {System.out.println(7828);}}
    public static class Impl7829 implements DemoService {public int type() {return 7829;} public void run() {System.out.println(7829);}}
    public static class Impl7830 implements DemoService {public int type() {return 7830;} public void run() {System.out.println(7830);}}
    public static class Impl7831 implements DemoService {public int type() {return 7831;} public void run() {System.out.println(7831);}}
    public static class Impl7832 implements DemoService {public int type() {return 7832;} public void run() {System.out.println(7832);}}
    public static class Impl7833 implements DemoService {public int type() {return 7833;} public void run() {System.out.println(7833);}}
    public static class Impl7834 implements DemoService {public int type() {return 7834;} public void run() {System.out.println(7834);}}
    public static class Impl7835 implements DemoService {public int type() {return 7835;} public void run() {System.out.println(7835);}}
    public static class Impl7836 implements DemoService {public int type() {return 7836;} public void run() {System.out.println(7836);}}
    public static class Impl7837 implements DemoService {public int type() {return 7837;} public void run() {System.out.println(7837);}}
    public static class Impl7838 implements DemoService {public int type() {return 7838;} public void run() {System.out.println(7838);}}
    public static class Impl7839 implements DemoService {public int type() {return 7839;} public void run() {System.out.println(7839);}}
    public static class Impl7840 implements DemoService {public int type() {return 7840;} public void run() {System.out.println(7840);}}
    public static class Impl7841 implements DemoService {public int type() {return 7841;} public void run() {System.out.println(7841);}}
    public static class Impl7842 implements DemoService {public int type() {return 7842;} public void run() {System.out.println(7842);}}
    public static class Impl7843 implements DemoService {public int type() {return 7843;} public void run() {System.out.println(7843);}}
    public static class Impl7844 implements DemoService {public int type() {return 7844;} public void run() {System.out.println(7844);}}
    public static class Impl7845 implements DemoService {public int type() {return 7845;} public void run() {System.out.println(7845);}}
    public static class Impl7846 implements DemoService {public int type() {return 7846;} public void run() {System.out.println(7846);}}
    public static class Impl7847 implements DemoService {public int type() {return 7847;} public void run() {System.out.println(7847);}}
    public static class Impl7848 implements DemoService {public int type() {return 7848;} public void run() {System.out.println(7848);}}
    public static class Impl7849 implements DemoService {public int type() {return 7849;} public void run() {System.out.println(7849);}}
    public static class Impl7850 implements DemoService {public int type() {return 7850;} public void run() {System.out.println(7850);}}
    public static class Impl7851 implements DemoService {public int type() {return 7851;} public void run() {System.out.println(7851);}}
    public static class Impl7852 implements DemoService {public int type() {return 7852;} public void run() {System.out.println(7852);}}
    public static class Impl7853 implements DemoService {public int type() {return 7853;} public void run() {System.out.println(7853);}}
    public static class Impl7854 implements DemoService {public int type() {return 7854;} public void run() {System.out.println(7854);}}
    public static class Impl7855 implements DemoService {public int type() {return 7855;} public void run() {System.out.println(7855);}}
    public static class Impl7856 implements DemoService {public int type() {return 7856;} public void run() {System.out.println(7856);}}
    public static class Impl7857 implements DemoService {public int type() {return 7857;} public void run() {System.out.println(7857);}}
    public static class Impl7858 implements DemoService {public int type() {return 7858;} public void run() {System.out.println(7858);}}
    public static class Impl7859 implements DemoService {public int type() {return 7859;} public void run() {System.out.println(7859);}}
    public static class Impl7860 implements DemoService {public int type() {return 7860;} public void run() {System.out.println(7860);}}
    public static class Impl7861 implements DemoService {public int type() {return 7861;} public void run() {System.out.println(7861);}}
    public static class Impl7862 implements DemoService {public int type() {return 7862;} public void run() {System.out.println(7862);}}
    public static class Impl7863 implements DemoService {public int type() {return 7863;} public void run() {System.out.println(7863);}}
    public static class Impl7864 implements DemoService {public int type() {return 7864;} public void run() {System.out.println(7864);}}
    public static class Impl7865 implements DemoService {public int type() {return 7865;} public void run() {System.out.println(7865);}}
    public static class Impl7866 implements DemoService {public int type() {return 7866;} public void run() {System.out.println(7866);}}
    public static class Impl7867 implements DemoService {public int type() {return 7867;} public void run() {System.out.println(7867);}}
    public static class Impl7868 implements DemoService {public int type() {return 7868;} public void run() {System.out.println(7868);}}
    public static class Impl7869 implements DemoService {public int type() {return 7869;} public void run() {System.out.println(7869);}}
    public static class Impl7870 implements DemoService {public int type() {return 7870;} public void run() {System.out.println(7870);}}
    public static class Impl7871 implements DemoService {public int type() {return 7871;} public void run() {System.out.println(7871);}}
    public static class Impl7872 implements DemoService {public int type() {return 7872;} public void run() {System.out.println(7872);}}
    public static class Impl7873 implements DemoService {public int type() {return 7873;} public void run() {System.out.println(7873);}}
    public static class Impl7874 implements DemoService {public int type() {return 7874;} public void run() {System.out.println(7874);}}
    public static class Impl7875 implements DemoService {public int type() {return 7875;} public void run() {System.out.println(7875);}}
    public static class Impl7876 implements DemoService {public int type() {return 7876;} public void run() {System.out.println(7876);}}
    public static class Impl7877 implements DemoService {public int type() {return 7877;} public void run() {System.out.println(7877);}}
    public static class Impl7878 implements DemoService {public int type() {return 7878;} public void run() {System.out.println(7878);}}
    public static class Impl7879 implements DemoService {public int type() {return 7879;} public void run() {System.out.println(7879);}}
    public static class Impl7880 implements DemoService {public int type() {return 7880;} public void run() {System.out.println(7880);}}
    public static class Impl7881 implements DemoService {public int type() {return 7881;} public void run() {System.out.println(7881);}}
    public static class Impl7882 implements DemoService {public int type() {return 7882;} public void run() {System.out.println(7882);}}
    public static class Impl7883 implements DemoService {public int type() {return 7883;} public void run() {System.out.println(7883);}}
    public static class Impl7884 implements DemoService {public int type() {return 7884;} public void run() {System.out.println(7884);}}
    public static class Impl7885 implements DemoService {public int type() {return 7885;} public void run() {System.out.println(7885);}}
    public static class Impl7886 implements DemoService {public int type() {return 7886;} public void run() {System.out.println(7886);}}
    public static class Impl7887 implements DemoService {public int type() {return 7887;} public void run() {System.out.println(7887);}}
    public static class Impl7888 implements DemoService {public int type() {return 7888;} public void run() {System.out.println(7888);}}
    public static class Impl7889 implements DemoService {public int type() {return 7889;} public void run() {System.out.println(7889);}}
    public static class Impl7890 implements DemoService {public int type() {return 7890;} public void run() {System.out.println(7890);}}
    public static class Impl7891 implements DemoService {public int type() {return 7891;} public void run() {System.out.println(7891);}}
    public static class Impl7892 implements DemoService {public int type() {return 7892;} public void run() {System.out.println(7892);}}
    public static class Impl7893 implements DemoService {public int type() {return 7893;} public void run() {System.out.println(7893);}}
    public static class Impl7894 implements DemoService {public int type() {return 7894;} public void run() {System.out.println(7894);}}
    public static class Impl7895 implements DemoService {public int type() {return 7895;} public void run() {System.out.println(7895);}}
    public static class Impl7896 implements DemoService {public int type() {return 7896;} public void run() {System.out.println(7896);}}
    public static class Impl7897 implements DemoService {public int type() {return 7897;} public void run() {System.out.println(7897);}}
    public static class Impl7898 implements DemoService {public int type() {return 7898;} public void run() {System.out.println(7898);}}
    public static class Impl7899 implements DemoService {public int type() {return 7899;} public void run() {System.out.println(7899);}}
    public static class Impl7900 implements DemoService {public int type() {return 7900;} public void run() {System.out.println(7900);}}
    public static class Impl7901 implements DemoService {public int type() {return 7901;} public void run() {System.out.println(7901);}}
    public static class Impl7902 implements DemoService {public int type() {return 7902;} public void run() {System.out.println(7902);}}
    public static class Impl7903 implements DemoService {public int type() {return 7903;} public void run() {System.out.println(7903);}}
    public static class Impl7904 implements DemoService {public int type() {return 7904;} public void run() {System.out.println(7904);}}
    public static class Impl7905 implements DemoService {public int type() {return 7905;} public void run() {System.out.println(7905);}}
    public static class Impl7906 implements DemoService {public int type() {return 7906;} public void run() {System.out.println(7906);}}
    public static class Impl7907 implements DemoService {public int type() {return 7907;} public void run() {System.out.println(7907);}}
    public static class Impl7908 implements DemoService {public int type() {return 7908;} public void run() {System.out.println(7908);}}
    public static class Impl7909 implements DemoService {public int type() {return 7909;} public void run() {System.out.println(7909);}}
    public static class Impl7910 implements DemoService {public int type() {return 7910;} public void run() {System.out.println(7910);}}
    public static class Impl7911 implements DemoService {public int type() {return 7911;} public void run() {System.out.println(7911);}}
    public static class Impl7912 implements DemoService {public int type() {return 7912;} public void run() {System.out.println(7912);}}
    public static class Impl7913 implements DemoService {public int type() {return 7913;} public void run() {System.out.println(7913);}}
    public static class Impl7914 implements DemoService {public int type() {return 7914;} public void run() {System.out.println(7914);}}
    public static class Impl7915 implements DemoService {public int type() {return 7915;} public void run() {System.out.println(7915);}}
    public static class Impl7916 implements DemoService {public int type() {return 7916;} public void run() {System.out.println(7916);}}
    public static class Impl7917 implements DemoService {public int type() {return 7917;} public void run() {System.out.println(7917);}}
    public static class Impl7918 implements DemoService {public int type() {return 7918;} public void run() {System.out.println(7918);}}
    public static class Impl7919 implements DemoService {public int type() {return 7919;} public void run() {System.out.println(7919);}}
    public static class Impl7920 implements DemoService {public int type() {return 7920;} public void run() {System.out.println(7920);}}
    public static class Impl7921 implements DemoService {public int type() {return 7921;} public void run() {System.out.println(7921);}}
    public static class Impl7922 implements DemoService {public int type() {return 7922;} public void run() {System.out.println(7922);}}
    public static class Impl7923 implements DemoService {public int type() {return 7923;} public void run() {System.out.println(7923);}}
    public static class Impl7924 implements DemoService {public int type() {return 7924;} public void run() {System.out.println(7924);}}
    public static class Impl7925 implements DemoService {public int type() {return 7925;} public void run() {System.out.println(7925);}}
    public static class Impl7926 implements DemoService {public int type() {return 7926;} public void run() {System.out.println(7926);}}
    public static class Impl7927 implements DemoService {public int type() {return 7927;} public void run() {System.out.println(7927);}}
    public static class Impl7928 implements DemoService {public int type() {return 7928;} public void run() {System.out.println(7928);}}
    public static class Impl7929 implements DemoService {public int type() {return 7929;} public void run() {System.out.println(7929);}}
    public static class Impl7930 implements DemoService {public int type() {return 7930;} public void run() {System.out.println(7930);}}
    public static class Impl7931 implements DemoService {public int type() {return 7931;} public void run() {System.out.println(7931);}}
    public static class Impl7932 implements DemoService {public int type() {return 7932;} public void run() {System.out.println(7932);}}
    public static class Impl7933 implements DemoService {public int type() {return 7933;} public void run() {System.out.println(7933);}}
    public static class Impl7934 implements DemoService {public int type() {return 7934;} public void run() {System.out.println(7934);}}
    public static class Impl7935 implements DemoService {public int type() {return 7935;} public void run() {System.out.println(7935);}}
    public static class Impl7936 implements DemoService {public int type() {return 7936;} public void run() {System.out.println(7936);}}
    public static class Impl7937 implements DemoService {public int type() {return 7937;} public void run() {System.out.println(7937);}}
    public static class Impl7938 implements DemoService {public int type() {return 7938;} public void run() {System.out.println(7938);}}
    public static class Impl7939 implements DemoService {public int type() {return 7939;} public void run() {System.out.println(7939);}}
    public static class Impl7940 implements DemoService {public int type() {return 7940;} public void run() {System.out.println(7940);}}
    public static class Impl7941 implements DemoService {public int type() {return 7941;} public void run() {System.out.println(7941);}}
    public static class Impl7942 implements DemoService {public int type() {return 7942;} public void run() {System.out.println(7942);}}
    public static class Impl7943 implements DemoService {public int type() {return 7943;} public void run() {System.out.println(7943);}}
    public static class Impl7944 implements DemoService {public int type() {return 7944;} public void run() {System.out.println(7944);}}
    public static class Impl7945 implements DemoService {public int type() {return 7945;} public void run() {System.out.println(7945);}}
    public static class Impl7946 implements DemoService {public int type() {return 7946;} public void run() {System.out.println(7946);}}
    public static class Impl7947 implements DemoService {public int type() {return 7947;} public void run() {System.out.println(7947);}}
    public static class Impl7948 implements DemoService {public int type() {return 7948;} public void run() {System.out.println(7948);}}
    public static class Impl7949 implements DemoService {public int type() {return 7949;} public void run() {System.out.println(7949);}}
    public static class Impl7950 implements DemoService {public int type() {return 7950;} public void run() {System.out.println(7950);}}
    public static class Impl7951 implements DemoService {public int type() {return 7951;} public void run() {System.out.println(7951);}}
    public static class Impl7952 implements DemoService {public int type() {return 7952;} public void run() {System.out.println(7952);}}
    public static class Impl7953 implements DemoService {public int type() {return 7953;} public void run() {System.out.println(7953);}}
    public static class Impl7954 implements DemoService {public int type() {return 7954;} public void run() {System.out.println(7954);}}
    public static class Impl7955 implements DemoService {public int type() {return 7955;} public void run() {System.out.println(7955);}}
    public static class Impl7956 implements DemoService {public int type() {return 7956;} public void run() {System.out.println(7956);}}
    public static class Impl7957 implements DemoService {public int type() {return 7957;} public void run() {System.out.println(7957);}}
    public static class Impl7958 implements DemoService {public int type() {return 7958;} public void run() {System.out.println(7958);}}
    public static class Impl7959 implements DemoService {public int type() {return 7959;} public void run() {System.out.println(7959);}}
    public static class Impl7960 implements DemoService {public int type() {return 7960;} public void run() {System.out.println(7960);}}
    public static class Impl7961 implements DemoService {public int type() {return 7961;} public void run() {System.out.println(7961);}}
    public static class Impl7962 implements DemoService {public int type() {return 7962;} public void run() {System.out.println(7962);}}
    public static class Impl7963 implements DemoService {public int type() {return 7963;} public void run() {System.out.println(7963);}}
    public static class Impl7964 implements DemoService {public int type() {return 7964;} public void run() {System.out.println(7964);}}
    public static class Impl7965 implements DemoService {public int type() {return 7965;} public void run() {System.out.println(7965);}}
    public static class Impl7966 implements DemoService {public int type() {return 7966;} public void run() {System.out.println(7966);}}
    public static class Impl7967 implements DemoService {public int type() {return 7967;} public void run() {System.out.println(7967);}}
    public static class Impl7968 implements DemoService {public int type() {return 7968;} public void run() {System.out.println(7968);}}
    public static class Impl7969 implements DemoService {public int type() {return 7969;} public void run() {System.out.println(7969);}}
    public static class Impl7970 implements DemoService {public int type() {return 7970;} public void run() {System.out.println(7970);}}
    public static class Impl7971 implements DemoService {public int type() {return 7971;} public void run() {System.out.println(7971);}}
    public static class Impl7972 implements DemoService {public int type() {return 7972;} public void run() {System.out.println(7972);}}
    public static class Impl7973 implements DemoService {public int type() {return 7973;} public void run() {System.out.println(7973);}}
    public static class Impl7974 implements DemoService {public int type() {return 7974;} public void run() {System.out.println(7974);}}
    public static class Impl7975 implements DemoService {public int type() {return 7975;} public void run() {System.out.println(7975);}}
    public static class Impl7976 implements DemoService {public int type() {return 7976;} public void run() {System.out.println(7976);}}
    public static class Impl7977 implements DemoService {public int type() {return 7977;} public void run() {System.out.println(7977);}}
    public static class Impl7978 implements DemoService {public int type() {return 7978;} public void run() {System.out.println(7978);}}
    public static class Impl7979 implements DemoService {public int type() {return 7979;} public void run() {System.out.println(7979);}}
    public static class Impl7980 implements DemoService {public int type() {return 7980;} public void run() {System.out.println(7980);}}
    public static class Impl7981 implements DemoService {public int type() {return 7981;} public void run() {System.out.println(7981);}}
    public static class Impl7982 implements DemoService {public int type() {return 7982;} public void run() {System.out.println(7982);}}
    public static class Impl7983 implements DemoService {public int type() {return 7983;} public void run() {System.out.println(7983);}}
    public static class Impl7984 implements DemoService {public int type() {return 7984;} public void run() {System.out.println(7984);}}
    public static class Impl7985 implements DemoService {public int type() {return 7985;} public void run() {System.out.println(7985);}}
    public static class Impl7986 implements DemoService {public int type() {return 7986;} public void run() {System.out.println(7986);}}
    public static class Impl7987 implements DemoService {public int type() {return 7987;} public void run() {System.out.println(7987);}}
    public static class Impl7988 implements DemoService {public int type() {return 7988;} public void run() {System.out.println(7988);}}
    public static class Impl7989 implements DemoService {public int type() {return 7989;} public void run() {System.out.println(7989);}}
    public static class Impl7990 implements DemoService {public int type() {return 7990;} public void run() {System.out.println(7990);}}
    public static class Impl7991 implements DemoService {public int type() {return 7991;} public void run() {System.out.println(7991);}}
    public static class Impl7992 implements DemoService {public int type() {return 7992;} public void run() {System.out.println(7992);}}
    public static class Impl7993 implements DemoService {public int type() {return 7993;} public void run() {System.out.println(7993);}}
    public static class Impl7994 implements DemoService {public int type() {return 7994;} public void run() {System.out.println(7994);}}
    public static class Impl7995 implements DemoService {public int type() {return 7995;} public void run() {System.out.println(7995);}}
    public static class Impl7996 implements DemoService {public int type() {return 7996;} public void run() {System.out.println(7996);}}
    public static class Impl7997 implements DemoService {public int type() {return 7997;} public void run() {System.out.println(7997);}}
    public static class Impl7998 implements DemoService {public int type() {return 7998;} public void run() {System.out.println(7998);}}
    public static class Impl7999 implements DemoService {public int type() {return 7999;} public void run() {System.out.println(7999);}}
    public static class Impl8000 implements DemoService {public int type() {return 8000;} public void run() {System.out.println(8000);}}
    public static class Impl8001 implements DemoService {public int type() {return 8001;} public void run() {System.out.println(8001);}}
    public static class Impl8002 implements DemoService {public int type() {return 8002;} public void run() {System.out.println(8002);}}
    public static class Impl8003 implements DemoService {public int type() {return 8003;} public void run() {System.out.println(8003);}}
    public static class Impl8004 implements DemoService {public int type() {return 8004;} public void run() {System.out.println(8004);}}
    public static class Impl8005 implements DemoService {public int type() {return 8005;} public void run() {System.out.println(8005);}}
    public static class Impl8006 implements DemoService {public int type() {return 8006;} public void run() {System.out.println(8006);}}
    public static class Impl8007 implements DemoService {public int type() {return 8007;} public void run() {System.out.println(8007);}}
    public static class Impl8008 implements DemoService {public int type() {return 8008;} public void run() {System.out.println(8008);}}
    public static class Impl8009 implements DemoService {public int type() {return 8009;} public void run() {System.out.println(8009);}}
    public static class Impl8010 implements DemoService {public int type() {return 8010;} public void run() {System.out.println(8010);}}
    public static class Impl8011 implements DemoService {public int type() {return 8011;} public void run() {System.out.println(8011);}}
    public static class Impl8012 implements DemoService {public int type() {return 8012;} public void run() {System.out.println(8012);}}
    public static class Impl8013 implements DemoService {public int type() {return 8013;} public void run() {System.out.println(8013);}}
    public static class Impl8014 implements DemoService {public int type() {return 8014;} public void run() {System.out.println(8014);}}
    public static class Impl8015 implements DemoService {public int type() {return 8015;} public void run() {System.out.println(8015);}}
    public static class Impl8016 implements DemoService {public int type() {return 8016;} public void run() {System.out.println(8016);}}
    public static class Impl8017 implements DemoService {public int type() {return 8017;} public void run() {System.out.println(8017);}}
    public static class Impl8018 implements DemoService {public int type() {return 8018;} public void run() {System.out.println(8018);}}
    public static class Impl8019 implements DemoService {public int type() {return 8019;} public void run() {System.out.println(8019);}}
    public static class Impl8020 implements DemoService {public int type() {return 8020;} public void run() {System.out.println(8020);}}
    public static class Impl8021 implements DemoService {public int type() {return 8021;} public void run() {System.out.println(8021);}}
    public static class Impl8022 implements DemoService {public int type() {return 8022;} public void run() {System.out.println(8022);}}
    public static class Impl8023 implements DemoService {public int type() {return 8023;} public void run() {System.out.println(8023);}}
    public static class Impl8024 implements DemoService {public int type() {return 8024;} public void run() {System.out.println(8024);}}
    public static class Impl8025 implements DemoService {public int type() {return 8025;} public void run() {System.out.println(8025);}}
    public static class Impl8026 implements DemoService {public int type() {return 8026;} public void run() {System.out.println(8026);}}
    public static class Impl8027 implements DemoService {public int type() {return 8027;} public void run() {System.out.println(8027);}}
    public static class Impl8028 implements DemoService {public int type() {return 8028;} public void run() {System.out.println(8028);}}
    public static class Impl8029 implements DemoService {public int type() {return 8029;} public void run() {System.out.println(8029);}}
    public static class Impl8030 implements DemoService {public int type() {return 8030;} public void run() {System.out.println(8030);}}
    public static class Impl8031 implements DemoService {public int type() {return 8031;} public void run() {System.out.println(8031);}}
    public static class Impl8032 implements DemoService {public int type() {return 8032;} public void run() {System.out.println(8032);}}
    public static class Impl8033 implements DemoService {public int type() {return 8033;} public void run() {System.out.println(8033);}}
    public static class Impl8034 implements DemoService {public int type() {return 8034;} public void run() {System.out.println(8034);}}
    public static class Impl8035 implements DemoService {public int type() {return 8035;} public void run() {System.out.println(8035);}}
    public static class Impl8036 implements DemoService {public int type() {return 8036;} public void run() {System.out.println(8036);}}
    public static class Impl8037 implements DemoService {public int type() {return 8037;} public void run() {System.out.println(8037);}}
    public static class Impl8038 implements DemoService {public int type() {return 8038;} public void run() {System.out.println(8038);}}
    public static class Impl8039 implements DemoService {public int type() {return 8039;} public void run() {System.out.println(8039);}}
    public static class Impl8040 implements DemoService {public int type() {return 8040;} public void run() {System.out.println(8040);}}
    public static class Impl8041 implements DemoService {public int type() {return 8041;} public void run() {System.out.println(8041);}}
    public static class Impl8042 implements DemoService {public int type() {return 8042;} public void run() {System.out.println(8042);}}
    public static class Impl8043 implements DemoService {public int type() {return 8043;} public void run() {System.out.println(8043);}}
    public static class Impl8044 implements DemoService {public int type() {return 8044;} public void run() {System.out.println(8044);}}
    public static class Impl8045 implements DemoService {public int type() {return 8045;} public void run() {System.out.println(8045);}}
    public static class Impl8046 implements DemoService {public int type() {return 8046;} public void run() {System.out.println(8046);}}
    public static class Impl8047 implements DemoService {public int type() {return 8047;} public void run() {System.out.println(8047);}}
    public static class Impl8048 implements DemoService {public int type() {return 8048;} public void run() {System.out.println(8048);}}
    public static class Impl8049 implements DemoService {public int type() {return 8049;} public void run() {System.out.println(8049);}}
    public static class Impl8050 implements DemoService {public int type() {return 8050;} public void run() {System.out.println(8050);}}
    public static class Impl8051 implements DemoService {public int type() {return 8051;} public void run() {System.out.println(8051);}}
    public static class Impl8052 implements DemoService {public int type() {return 8052;} public void run() {System.out.println(8052);}}
    public static class Impl8053 implements DemoService {public int type() {return 8053;} public void run() {System.out.println(8053);}}
    public static class Impl8054 implements DemoService {public int type() {return 8054;} public void run() {System.out.println(8054);}}
    public static class Impl8055 implements DemoService {public int type() {return 8055;} public void run() {System.out.println(8055);}}
    public static class Impl8056 implements DemoService {public int type() {return 8056;} public void run() {System.out.println(8056);}}
    public static class Impl8057 implements DemoService {public int type() {return 8057;} public void run() {System.out.println(8057);}}
    public static class Impl8058 implements DemoService {public int type() {return 8058;} public void run() {System.out.println(8058);}}
    public static class Impl8059 implements DemoService {public int type() {return 8059;} public void run() {System.out.println(8059);}}
    public static class Impl8060 implements DemoService {public int type() {return 8060;} public void run() {System.out.println(8060);}}
    public static class Impl8061 implements DemoService {public int type() {return 8061;} public void run() {System.out.println(8061);}}
    public static class Impl8062 implements DemoService {public int type() {return 8062;} public void run() {System.out.println(8062);}}
    public static class Impl8063 implements DemoService {public int type() {return 8063;} public void run() {System.out.println(8063);}}
    public static class Impl8064 implements DemoService {public int type() {return 8064;} public void run() {System.out.println(8064);}}
    public static class Impl8065 implements DemoService {public int type() {return 8065;} public void run() {System.out.println(8065);}}
    public static class Impl8066 implements DemoService {public int type() {return 8066;} public void run() {System.out.println(8066);}}
    public static class Impl8067 implements DemoService {public int type() {return 8067;} public void run() {System.out.println(8067);}}
    public static class Impl8068 implements DemoService {public int type() {return 8068;} public void run() {System.out.println(8068);}}
    public static class Impl8069 implements DemoService {public int type() {return 8069;} public void run() {System.out.println(8069);}}
    public static class Impl8070 implements DemoService {public int type() {return 8070;} public void run() {System.out.println(8070);}}
    public static class Impl8071 implements DemoService {public int type() {return 8071;} public void run() {System.out.println(8071);}}
    public static class Impl8072 implements DemoService {public int type() {return 8072;} public void run() {System.out.println(8072);}}
    public static class Impl8073 implements DemoService {public int type() {return 8073;} public void run() {System.out.println(8073);}}
    public static class Impl8074 implements DemoService {public int type() {return 8074;} public void run() {System.out.println(8074);}}
    public static class Impl8075 implements DemoService {public int type() {return 8075;} public void run() {System.out.println(8075);}}
    public static class Impl8076 implements DemoService {public int type() {return 8076;} public void run() {System.out.println(8076);}}
    public static class Impl8077 implements DemoService {public int type() {return 8077;} public void run() {System.out.println(8077);}}
    public static class Impl8078 implements DemoService {public int type() {return 8078;} public void run() {System.out.println(8078);}}
    public static class Impl8079 implements DemoService {public int type() {return 8079;} public void run() {System.out.println(8079);}}
    public static class Impl8080 implements DemoService {public int type() {return 8080;} public void run() {System.out.println(8080);}}
    public static class Impl8081 implements DemoService {public int type() {return 8081;} public void run() {System.out.println(8081);}}
    public static class Impl8082 implements DemoService {public int type() {return 8082;} public void run() {System.out.println(8082);}}
    public static class Impl8083 implements DemoService {public int type() {return 8083;} public void run() {System.out.println(8083);}}
    public static class Impl8084 implements DemoService {public int type() {return 8084;} public void run() {System.out.println(8084);}}
    public static class Impl8085 implements DemoService {public int type() {return 8085;} public void run() {System.out.println(8085);}}
    public static class Impl8086 implements DemoService {public int type() {return 8086;} public void run() {System.out.println(8086);}}
    public static class Impl8087 implements DemoService {public int type() {return 8087;} public void run() {System.out.println(8087);}}
    public static class Impl8088 implements DemoService {public int type() {return 8088;} public void run() {System.out.println(8088);}}
    public static class Impl8089 implements DemoService {public int type() {return 8089;} public void run() {System.out.println(8089);}}
    public static class Impl8090 implements DemoService {public int type() {return 8090;} public void run() {System.out.println(8090);}}
    public static class Impl8091 implements DemoService {public int type() {return 8091;} public void run() {System.out.println(8091);}}
    public static class Impl8092 implements DemoService {public int type() {return 8092;} public void run() {System.out.println(8092);}}
    public static class Impl8093 implements DemoService {public int type() {return 8093;} public void run() {System.out.println(8093);}}
    public static class Impl8094 implements DemoService {public int type() {return 8094;} public void run() {System.out.println(8094);}}
    public static class Impl8095 implements DemoService {public int type() {return 8095;} public void run() {System.out.println(8095);}}
    public static class Impl8096 implements DemoService {public int type() {return 8096;} public void run() {System.out.println(8096);}}
    public static class Impl8097 implements DemoService {public int type() {return 8097;} public void run() {System.out.println(8097);}}
    public static class Impl8098 implements DemoService {public int type() {return 8098;} public void run() {System.out.println(8098);}}
    public static class Impl8099 implements DemoService {public int type() {return 8099;} public void run() {System.out.println(8099);}}
    public static class Impl8100 implements DemoService {public int type() {return 8100;} public void run() {System.out.println(8100);}}
    public static class Impl8101 implements DemoService {public int type() {return 8101;} public void run() {System.out.println(8101);}}
    public static class Impl8102 implements DemoService {public int type() {return 8102;} public void run() {System.out.println(8102);}}
    public static class Impl8103 implements DemoService {public int type() {return 8103;} public void run() {System.out.println(8103);}}
    public static class Impl8104 implements DemoService {public int type() {return 8104;} public void run() {System.out.println(8104);}}
    public static class Impl8105 implements DemoService {public int type() {return 8105;} public void run() {System.out.println(8105);}}
    public static class Impl8106 implements DemoService {public int type() {return 8106;} public void run() {System.out.println(8106);}}
    public static class Impl8107 implements DemoService {public int type() {return 8107;} public void run() {System.out.println(8107);}}
    public static class Impl8108 implements DemoService {public int type() {return 8108;} public void run() {System.out.println(8108);}}
    public static class Impl8109 implements DemoService {public int type() {return 8109;} public void run() {System.out.println(8109);}}
    public static class Impl8110 implements DemoService {public int type() {return 8110;} public void run() {System.out.println(8110);}}
    public static class Impl8111 implements DemoService {public int type() {return 8111;} public void run() {System.out.println(8111);}}
    public static class Impl8112 implements DemoService {public int type() {return 8112;} public void run() {System.out.println(8112);}}
    public static class Impl8113 implements DemoService {public int type() {return 8113;} public void run() {System.out.println(8113);}}
    public static class Impl8114 implements DemoService {public int type() {return 8114;} public void run() {System.out.println(8114);}}
    public static class Impl8115 implements DemoService {public int type() {return 8115;} public void run() {System.out.println(8115);}}
    public static class Impl8116 implements DemoService {public int type() {return 8116;} public void run() {System.out.println(8116);}}
    public static class Impl8117 implements DemoService {public int type() {return 8117;} public void run() {System.out.println(8117);}}
    public static class Impl8118 implements DemoService {public int type() {return 8118;} public void run() {System.out.println(8118);}}
    public static class Impl8119 implements DemoService {public int type() {return 8119;} public void run() {System.out.println(8119);}}
    public static class Impl8120 implements DemoService {public int type() {return 8120;} public void run() {System.out.println(8120);}}
    public static class Impl8121 implements DemoService {public int type() {return 8121;} public void run() {System.out.println(8121);}}
    public static class Impl8122 implements DemoService {public int type() {return 8122;} public void run() {System.out.println(8122);}}
    public static class Impl8123 implements DemoService {public int type() {return 8123;} public void run() {System.out.println(8123);}}
    public static class Impl8124 implements DemoService {public int type() {return 8124;} public void run() {System.out.println(8124);}}
    public static class Impl8125 implements DemoService {public int type() {return 8125;} public void run() {System.out.println(8125);}}
    public static class Impl8126 implements DemoService {public int type() {return 8126;} public void run() {System.out.println(8126);}}
    public static class Impl8127 implements DemoService {public int type() {return 8127;} public void run() {System.out.println(8127);}}
    public static class Impl8128 implements DemoService {public int type() {return 8128;} public void run() {System.out.println(8128);}}
    public static class Impl8129 implements DemoService {public int type() {return 8129;} public void run() {System.out.println(8129);}}
    public static class Impl8130 implements DemoService {public int type() {return 8130;} public void run() {System.out.println(8130);}}
    public static class Impl8131 implements DemoService {public int type() {return 8131;} public void run() {System.out.println(8131);}}
    public static class Impl8132 implements DemoService {public int type() {return 8132;} public void run() {System.out.println(8132);}}
    public static class Impl8133 implements DemoService {public int type() {return 8133;} public void run() {System.out.println(8133);}}
    public static class Impl8134 implements DemoService {public int type() {return 8134;} public void run() {System.out.println(8134);}}
    public static class Impl8135 implements DemoService {public int type() {return 8135;} public void run() {System.out.println(8135);}}
    public static class Impl8136 implements DemoService {public int type() {return 8136;} public void run() {System.out.println(8136);}}
    public static class Impl8137 implements DemoService {public int type() {return 8137;} public void run() {System.out.println(8137);}}
    public static class Impl8138 implements DemoService {public int type() {return 8138;} public void run() {System.out.println(8138);}}
    public static class Impl8139 implements DemoService {public int type() {return 8139;} public void run() {System.out.println(8139);}}
    public static class Impl8140 implements DemoService {public int type() {return 8140;} public void run() {System.out.println(8140);}}
    public static class Impl8141 implements DemoService {public int type() {return 8141;} public void run() {System.out.println(8141);}}
    public static class Impl8142 implements DemoService {public int type() {return 8142;} public void run() {System.out.println(8142);}}
    public static class Impl8143 implements DemoService {public int type() {return 8143;} public void run() {System.out.println(8143);}}
    public static class Impl8144 implements DemoService {public int type() {return 8144;} public void run() {System.out.println(8144);}}
    public static class Impl8145 implements DemoService {public int type() {return 8145;} public void run() {System.out.println(8145);}}
    public static class Impl8146 implements DemoService {public int type() {return 8146;} public void run() {System.out.println(8146);}}
    public static class Impl8147 implements DemoService {public int type() {return 8147;} public void run() {System.out.println(8147);}}
    public static class Impl8148 implements DemoService {public int type() {return 8148;} public void run() {System.out.println(8148);}}
    public static class Impl8149 implements DemoService {public int type() {return 8149;} public void run() {System.out.println(8149);}}
    public static class Impl8150 implements DemoService {public int type() {return 8150;} public void run() {System.out.println(8150);}}
    public static class Impl8151 implements DemoService {public int type() {return 8151;} public void run() {System.out.println(8151);}}
    public static class Impl8152 implements DemoService {public int type() {return 8152;} public void run() {System.out.println(8152);}}
    public static class Impl8153 implements DemoService {public int type() {return 8153;} public void run() {System.out.println(8153);}}
    public static class Impl8154 implements DemoService {public int type() {return 8154;} public void run() {System.out.println(8154);}}
    public static class Impl8155 implements DemoService {public int type() {return 8155;} public void run() {System.out.println(8155);}}
    public static class Impl8156 implements DemoService {public int type() {return 8156;} public void run() {System.out.println(8156);}}
    public static class Impl8157 implements DemoService {public int type() {return 8157;} public void run() {System.out.println(8157);}}
    public static class Impl8158 implements DemoService {public int type() {return 8158;} public void run() {System.out.println(8158);}}
    public static class Impl8159 implements DemoService {public int type() {return 8159;} public void run() {System.out.println(8159);}}
    public static class Impl8160 implements DemoService {public int type() {return 8160;} public void run() {System.out.println(8160);}}
    public static class Impl8161 implements DemoService {public int type() {return 8161;} public void run() {System.out.println(8161);}}
    public static class Impl8162 implements DemoService {public int type() {return 8162;} public void run() {System.out.println(8162);}}
    public static class Impl8163 implements DemoService {public int type() {return 8163;} public void run() {System.out.println(8163);}}
    public static class Impl8164 implements DemoService {public int type() {return 8164;} public void run() {System.out.println(8164);}}
    public static class Impl8165 implements DemoService {public int type() {return 8165;} public void run() {System.out.println(8165);}}
    public static class Impl8166 implements DemoService {public int type() {return 8166;} public void run() {System.out.println(8166);}}
    public static class Impl8167 implements DemoService {public int type() {return 8167;} public void run() {System.out.println(8167);}}
    public static class Impl8168 implements DemoService {public int type() {return 8168;} public void run() {System.out.println(8168);}}
    public static class Impl8169 implements DemoService {public int type() {return 8169;} public void run() {System.out.println(8169);}}
    public static class Impl8170 implements DemoService {public int type() {return 8170;} public void run() {System.out.println(8170);}}
    public static class Impl8171 implements DemoService {public int type() {return 8171;} public void run() {System.out.println(8171);}}
    public static class Impl8172 implements DemoService {public int type() {return 8172;} public void run() {System.out.println(8172);}}
    public static class Impl8173 implements DemoService {public int type() {return 8173;} public void run() {System.out.println(8173);}}
    public static class Impl8174 implements DemoService {public int type() {return 8174;} public void run() {System.out.println(8174);}}
    public static class Impl8175 implements DemoService {public int type() {return 8175;} public void run() {System.out.println(8175);}}
    public static class Impl8176 implements DemoService {public int type() {return 8176;} public void run() {System.out.println(8176);}}
    public static class Impl8177 implements DemoService {public int type() {return 8177;} public void run() {System.out.println(8177);}}
    public static class Impl8178 implements DemoService {public int type() {return 8178;} public void run() {System.out.println(8178);}}
    public static class Impl8179 implements DemoService {public int type() {return 8179;} public void run() {System.out.println(8179);}}
    public static class Impl8180 implements DemoService {public int type() {return 8180;} public void run() {System.out.println(8180);}}
    public static class Impl8181 implements DemoService {public int type() {return 8181;} public void run() {System.out.println(8181);}}
    public static class Impl8182 implements DemoService {public int type() {return 8182;} public void run() {System.out.println(8182);}}
    public static class Impl8183 implements DemoService {public int type() {return 8183;} public void run() {System.out.println(8183);}}
    public static class Impl8184 implements DemoService {public int type() {return 8184;} public void run() {System.out.println(8184);}}
    public static class Impl8185 implements DemoService {public int type() {return 8185;} public void run() {System.out.println(8185);}}
    public static class Impl8186 implements DemoService {public int type() {return 8186;} public void run() {System.out.println(8186);}}
    public static class Impl8187 implements DemoService {public int type() {return 8187;} public void run() {System.out.println(8187);}}
    public static class Impl8188 implements DemoService {public int type() {return 8188;} public void run() {System.out.println(8188);}}
    public static class Impl8189 implements DemoService {public int type() {return 8189;} public void run() {System.out.println(8189);}}
    public static class Impl8190 implements DemoService {public int type() {return 8190;} public void run() {System.out.println(8190);}}
    public static class Impl8191 implements DemoService {public int type() {return 8191;} public void run() {System.out.println(8191);}}
    public static class Impl8192 implements DemoService {public int type() {return 8192;} public void run() {System.out.println(8192);}}
    public static class Impl8193 implements DemoService {public int type() {return 8193;} public void run() {System.out.println(8193);}}
    public static class Impl8194 implements DemoService {public int type() {return 8194;} public void run() {System.out.println(8194);}}
    public static class Impl8195 implements DemoService {public int type() {return 8195;} public void run() {System.out.println(8195);}}
    public static class Impl8196 implements DemoService {public int type() {return 8196;} public void run() {System.out.println(8196);}}
    public static class Impl8197 implements DemoService {public int type() {return 8197;} public void run() {System.out.println(8197);}}
    public static class Impl8198 implements DemoService {public int type() {return 8198;} public void run() {System.out.println(8198);}}
    public static class Impl8199 implements DemoService {public int type() {return 8199;} public void run() {System.out.println(8199);}}
    public static class Impl8200 implements DemoService {public int type() {return 8200;} public void run() {System.out.println(8200);}}
    public static class Impl8201 implements DemoService {public int type() {return 8201;} public void run() {System.out.println(8201);}}
    public static class Impl8202 implements DemoService {public int type() {return 8202;} public void run() {System.out.println(8202);}}
    public static class Impl8203 implements DemoService {public int type() {return 8203;} public void run() {System.out.println(8203);}}
    public static class Impl8204 implements DemoService {public int type() {return 8204;} public void run() {System.out.println(8204);}}
    public static class Impl8205 implements DemoService {public int type() {return 8205;} public void run() {System.out.println(8205);}}
    public static class Impl8206 implements DemoService {public int type() {return 8206;} public void run() {System.out.println(8206);}}
    public static class Impl8207 implements DemoService {public int type() {return 8207;} public void run() {System.out.println(8207);}}
    public static class Impl8208 implements DemoService {public int type() {return 8208;} public void run() {System.out.println(8208);}}
    public static class Impl8209 implements DemoService {public int type() {return 8209;} public void run() {System.out.println(8209);}}
    public static class Impl8210 implements DemoService {public int type() {return 8210;} public void run() {System.out.println(8210);}}
    public static class Impl8211 implements DemoService {public int type() {return 8211;} public void run() {System.out.println(8211);}}
    public static class Impl8212 implements DemoService {public int type() {return 8212;} public void run() {System.out.println(8212);}}
    public static class Impl8213 implements DemoService {public int type() {return 8213;} public void run() {System.out.println(8213);}}
    public static class Impl8214 implements DemoService {public int type() {return 8214;} public void run() {System.out.println(8214);}}
    public static class Impl8215 implements DemoService {public int type() {return 8215;} public void run() {System.out.println(8215);}}
    public static class Impl8216 implements DemoService {public int type() {return 8216;} public void run() {System.out.println(8216);}}
    public static class Impl8217 implements DemoService {public int type() {return 8217;} public void run() {System.out.println(8217);}}
    public static class Impl8218 implements DemoService {public int type() {return 8218;} public void run() {System.out.println(8218);}}
    public static class Impl8219 implements DemoService {public int type() {return 8219;} public void run() {System.out.println(8219);}}
    public static class Impl8220 implements DemoService {public int type() {return 8220;} public void run() {System.out.println(8220);}}
    public static class Impl8221 implements DemoService {public int type() {return 8221;} public void run() {System.out.println(8221);}}
    public static class Impl8222 implements DemoService {public int type() {return 8222;} public void run() {System.out.println(8222);}}
    public static class Impl8223 implements DemoService {public int type() {return 8223;} public void run() {System.out.println(8223);}}
    public static class Impl8224 implements DemoService {public int type() {return 8224;} public void run() {System.out.println(8224);}}
    public static class Impl8225 implements DemoService {public int type() {return 8225;} public void run() {System.out.println(8225);}}
    public static class Impl8226 implements DemoService {public int type() {return 8226;} public void run() {System.out.println(8226);}}
    public static class Impl8227 implements DemoService {public int type() {return 8227;} public void run() {System.out.println(8227);}}
    public static class Impl8228 implements DemoService {public int type() {return 8228;} public void run() {System.out.println(8228);}}
    public static class Impl8229 implements DemoService {public int type() {return 8229;} public void run() {System.out.println(8229);}}
    public static class Impl8230 implements DemoService {public int type() {return 8230;} public void run() {System.out.println(8230);}}
    public static class Impl8231 implements DemoService {public int type() {return 8231;} public void run() {System.out.println(8231);}}
    public static class Impl8232 implements DemoService {public int type() {return 8232;} public void run() {System.out.println(8232);}}
    public static class Impl8233 implements DemoService {public int type() {return 8233;} public void run() {System.out.println(8233);}}
    public static class Impl8234 implements DemoService {public int type() {return 8234;} public void run() {System.out.println(8234);}}
    public static class Impl8235 implements DemoService {public int type() {return 8235;} public void run() {System.out.println(8235);}}
    public static class Impl8236 implements DemoService {public int type() {return 8236;} public void run() {System.out.println(8236);}}
    public static class Impl8237 implements DemoService {public int type() {return 8237;} public void run() {System.out.println(8237);}}
    public static class Impl8238 implements DemoService {public int type() {return 8238;} public void run() {System.out.println(8238);}}
    public static class Impl8239 implements DemoService {public int type() {return 8239;} public void run() {System.out.println(8239);}}
    public static class Impl8240 implements DemoService {public int type() {return 8240;} public void run() {System.out.println(8240);}}
    public static class Impl8241 implements DemoService {public int type() {return 8241;} public void run() {System.out.println(8241);}}
    public static class Impl8242 implements DemoService {public int type() {return 8242;} public void run() {System.out.println(8242);}}
    public static class Impl8243 implements DemoService {public int type() {return 8243;} public void run() {System.out.println(8243);}}
    public static class Impl8244 implements DemoService {public int type() {return 8244;} public void run() {System.out.println(8244);}}
    public static class Impl8245 implements DemoService {public int type() {return 8245;} public void run() {System.out.println(8245);}}
    public static class Impl8246 implements DemoService {public int type() {return 8246;} public void run() {System.out.println(8246);}}
    public static class Impl8247 implements DemoService {public int type() {return 8247;} public void run() {System.out.println(8247);}}
    public static class Impl8248 implements DemoService {public int type() {return 8248;} public void run() {System.out.println(8248);}}
    public static class Impl8249 implements DemoService {public int type() {return 8249;} public void run() {System.out.println(8249);}}
    public static class Impl8250 implements DemoService {public int type() {return 8250;} public void run() {System.out.println(8250);}}
    public static class Impl8251 implements DemoService {public int type() {return 8251;} public void run() {System.out.println(8251);}}
    public static class Impl8252 implements DemoService {public int type() {return 8252;} public void run() {System.out.println(8252);}}
    public static class Impl8253 implements DemoService {public int type() {return 8253;} public void run() {System.out.println(8253);}}
    public static class Impl8254 implements DemoService {public int type() {return 8254;} public void run() {System.out.println(8254);}}
    public static class Impl8255 implements DemoService {public int type() {return 8255;} public void run() {System.out.println(8255);}}
    public static class Impl8256 implements DemoService {public int type() {return 8256;} public void run() {System.out.println(8256);}}
    public static class Impl8257 implements DemoService {public int type() {return 8257;} public void run() {System.out.println(8257);}}
    public static class Impl8258 implements DemoService {public int type() {return 8258;} public void run() {System.out.println(8258);}}
    public static class Impl8259 implements DemoService {public int type() {return 8259;} public void run() {System.out.println(8259);}}
    public static class Impl8260 implements DemoService {public int type() {return 8260;} public void run() {System.out.println(8260);}}
    public static class Impl8261 implements DemoService {public int type() {return 8261;} public void run() {System.out.println(8261);}}
    public static class Impl8262 implements DemoService {public int type() {return 8262;} public void run() {System.out.println(8262);}}
    public static class Impl8263 implements DemoService {public int type() {return 8263;} public void run() {System.out.println(8263);}}
    public static class Impl8264 implements DemoService {public int type() {return 8264;} public void run() {System.out.println(8264);}}
    public static class Impl8265 implements DemoService {public int type() {return 8265;} public void run() {System.out.println(8265);}}
    public static class Impl8266 implements DemoService {public int type() {return 8266;} public void run() {System.out.println(8266);}}
    public static class Impl8267 implements DemoService {public int type() {return 8267;} public void run() {System.out.println(8267);}}
    public static class Impl8268 implements DemoService {public int type() {return 8268;} public void run() {System.out.println(8268);}}
    public static class Impl8269 implements DemoService {public int type() {return 8269;} public void run() {System.out.println(8269);}}
    public static class Impl8270 implements DemoService {public int type() {return 8270;} public void run() {System.out.println(8270);}}
    public static class Impl8271 implements DemoService {public int type() {return 8271;} public void run() {System.out.println(8271);}}
    public static class Impl8272 implements DemoService {public int type() {return 8272;} public void run() {System.out.println(8272);}}
    public static class Impl8273 implements DemoService {public int type() {return 8273;} public void run() {System.out.println(8273);}}
    public static class Impl8274 implements DemoService {public int type() {return 8274;} public void run() {System.out.println(8274);}}
    public static class Impl8275 implements DemoService {public int type() {return 8275;} public void run() {System.out.println(8275);}}
    public static class Impl8276 implements DemoService {public int type() {return 8276;} public void run() {System.out.println(8276);}}
    public static class Impl8277 implements DemoService {public int type() {return 8277;} public void run() {System.out.println(8277);}}
    public static class Impl8278 implements DemoService {public int type() {return 8278;} public void run() {System.out.println(8278);}}
    public static class Impl8279 implements DemoService {public int type() {return 8279;} public void run() {System.out.println(8279);}}
    public static class Impl8280 implements DemoService {public int type() {return 8280;} public void run() {System.out.println(8280);}}
    public static class Impl8281 implements DemoService {public int type() {return 8281;} public void run() {System.out.println(8281);}}
    public static class Impl8282 implements DemoService {public int type() {return 8282;} public void run() {System.out.println(8282);}}
    public static class Impl8283 implements DemoService {public int type() {return 8283;} public void run() {System.out.println(8283);}}
    public static class Impl8284 implements DemoService {public int type() {return 8284;} public void run() {System.out.println(8284);}}
    public static class Impl8285 implements DemoService {public int type() {return 8285;} public void run() {System.out.println(8285);}}
    public static class Impl8286 implements DemoService {public int type() {return 8286;} public void run() {System.out.println(8286);}}
    public static class Impl8287 implements DemoService {public int type() {return 8287;} public void run() {System.out.println(8287);}}
    public static class Impl8288 implements DemoService {public int type() {return 8288;} public void run() {System.out.println(8288);}}
    public static class Impl8289 implements DemoService {public int type() {return 8289;} public void run() {System.out.println(8289);}}
    public static class Impl8290 implements DemoService {public int type() {return 8290;} public void run() {System.out.println(8290);}}
    public static class Impl8291 implements DemoService {public int type() {return 8291;} public void run() {System.out.println(8291);}}
    public static class Impl8292 implements DemoService {public int type() {return 8292;} public void run() {System.out.println(8292);}}
    public static class Impl8293 implements DemoService {public int type() {return 8293;} public void run() {System.out.println(8293);}}
    public static class Impl8294 implements DemoService {public int type() {return 8294;} public void run() {System.out.println(8294);}}
    public static class Impl8295 implements DemoService {public int type() {return 8295;} public void run() {System.out.println(8295);}}
    public static class Impl8296 implements DemoService {public int type() {return 8296;} public void run() {System.out.println(8296);}}
    public static class Impl8297 implements DemoService {public int type() {return 8297;} public void run() {System.out.println(8297);}}
    public static class Impl8298 implements DemoService {public int type() {return 8298;} public void run() {System.out.println(8298);}}
    public static class Impl8299 implements DemoService {public int type() {return 8299;} public void run() {System.out.println(8299);}}
    public static class Impl8300 implements DemoService {public int type() {return 8300;} public void run() {System.out.println(8300);}}
    public static class Impl8301 implements DemoService {public int type() {return 8301;} public void run() {System.out.println(8301);}}
    public static class Impl8302 implements DemoService {public int type() {return 8302;} public void run() {System.out.println(8302);}}
    public static class Impl8303 implements DemoService {public int type() {return 8303;} public void run() {System.out.println(8303);}}
    public static class Impl8304 implements DemoService {public int type() {return 8304;} public void run() {System.out.println(8304);}}
    public static class Impl8305 implements DemoService {public int type() {return 8305;} public void run() {System.out.println(8305);}}
    public static class Impl8306 implements DemoService {public int type() {return 8306;} public void run() {System.out.println(8306);}}
    public static class Impl8307 implements DemoService {public int type() {return 8307;} public void run() {System.out.println(8307);}}
    public static class Impl8308 implements DemoService {public int type() {return 8308;} public void run() {System.out.println(8308);}}
    public static class Impl8309 implements DemoService {public int type() {return 8309;} public void run() {System.out.println(8309);}}
    public static class Impl8310 implements DemoService {public int type() {return 8310;} public void run() {System.out.println(8310);}}
    public static class Impl8311 implements DemoService {public int type() {return 8311;} public void run() {System.out.println(8311);}}
    public static class Impl8312 implements DemoService {public int type() {return 8312;} public void run() {System.out.println(8312);}}
    public static class Impl8313 implements DemoService {public int type() {return 8313;} public void run() {System.out.println(8313);}}
    public static class Impl8314 implements DemoService {public int type() {return 8314;} public void run() {System.out.println(8314);}}
    public static class Impl8315 implements DemoService {public int type() {return 8315;} public void run() {System.out.println(8315);}}
    public static class Impl8316 implements DemoService {public int type() {return 8316;} public void run() {System.out.println(8316);}}
    public static class Impl8317 implements DemoService {public int type() {return 8317;} public void run() {System.out.println(8317);}}
    public static class Impl8318 implements DemoService {public int type() {return 8318;} public void run() {System.out.println(8318);}}
    public static class Impl8319 implements DemoService {public int type() {return 8319;} public void run() {System.out.println(8319);}}
    public static class Impl8320 implements DemoService {public int type() {return 8320;} public void run() {System.out.println(8320);}}
    public static class Impl8321 implements DemoService {public int type() {return 8321;} public void run() {System.out.println(8321);}}
    public static class Impl8322 implements DemoService {public int type() {return 8322;} public void run() {System.out.println(8322);}}
    public static class Impl8323 implements DemoService {public int type() {return 8323;} public void run() {System.out.println(8323);}}
    public static class Impl8324 implements DemoService {public int type() {return 8324;} public void run() {System.out.println(8324);}}
    public static class Impl8325 implements DemoService {public int type() {return 8325;} public void run() {System.out.println(8325);}}
    public static class Impl8326 implements DemoService {public int type() {return 8326;} public void run() {System.out.println(8326);}}
    public static class Impl8327 implements DemoService {public int type() {return 8327;} public void run() {System.out.println(8327);}}
    public static class Impl8328 implements DemoService {public int type() {return 8328;} public void run() {System.out.println(8328);}}
    public static class Impl8329 implements DemoService {public int type() {return 8329;} public void run() {System.out.println(8329);}}
    public static class Impl8330 implements DemoService {public int type() {return 8330;} public void run() {System.out.println(8330);}}
    public static class Impl8331 implements DemoService {public int type() {return 8331;} public void run() {System.out.println(8331);}}
    public static class Impl8332 implements DemoService {public int type() {return 8332;} public void run() {System.out.println(8332);}}
    public static class Impl8333 implements DemoService {public int type() {return 8333;} public void run() {System.out.println(8333);}}
    public static class Impl8334 implements DemoService {public int type() {return 8334;} public void run() {System.out.println(8334);}}
    public static class Impl8335 implements DemoService {public int type() {return 8335;} public void run() {System.out.println(8335);}}
    public static class Impl8336 implements DemoService {public int type() {return 8336;} public void run() {System.out.println(8336);}}
    public static class Impl8337 implements DemoService {public int type() {return 8337;} public void run() {System.out.println(8337);}}
    public static class Impl8338 implements DemoService {public int type() {return 8338;} public void run() {System.out.println(8338);}}
    public static class Impl8339 implements DemoService {public int type() {return 8339;} public void run() {System.out.println(8339);}}
    public static class Impl8340 implements DemoService {public int type() {return 8340;} public void run() {System.out.println(8340);}}
    public static class Impl8341 implements DemoService {public int type() {return 8341;} public void run() {System.out.println(8341);}}
    public static class Impl8342 implements DemoService {public int type() {return 8342;} public void run() {System.out.println(8342);}}
    public static class Impl8343 implements DemoService {public int type() {return 8343;} public void run() {System.out.println(8343);}}
    public static class Impl8344 implements DemoService {public int type() {return 8344;} public void run() {System.out.println(8344);}}
    public static class Impl8345 implements DemoService {public int type() {return 8345;} public void run() {System.out.println(8345);}}
    public static class Impl8346 implements DemoService {public int type() {return 8346;} public void run() {System.out.println(8346);}}
    public static class Impl8347 implements DemoService {public int type() {return 8347;} public void run() {System.out.println(8347);}}
    public static class Impl8348 implements DemoService {public int type() {return 8348;} public void run() {System.out.println(8348);}}
    public static class Impl8349 implements DemoService {public int type() {return 8349;} public void run() {System.out.println(8349);}}
    public static class Impl8350 implements DemoService {public int type() {return 8350;} public void run() {System.out.println(8350);}}
    public static class Impl8351 implements DemoService {public int type() {return 8351;} public void run() {System.out.println(8351);}}
    public static class Impl8352 implements DemoService {public int type() {return 8352;} public void run() {System.out.println(8352);}}
    public static class Impl8353 implements DemoService {public int type() {return 8353;} public void run() {System.out.println(8353);}}
    public static class Impl8354 implements DemoService {public int type() {return 8354;} public void run() {System.out.println(8354);}}
    public static class Impl8355 implements DemoService {public int type() {return 8355;} public void run() {System.out.println(8355);}}
    public static class Impl8356 implements DemoService {public int type() {return 8356;} public void run() {System.out.println(8356);}}
    public static class Impl8357 implements DemoService {public int type() {return 8357;} public void run() {System.out.println(8357);}}
    public static class Impl8358 implements DemoService {public int type() {return 8358;} public void run() {System.out.println(8358);}}
    public static class Impl8359 implements DemoService {public int type() {return 8359;} public void run() {System.out.println(8359);}}
    public static class Impl8360 implements DemoService {public int type() {return 8360;} public void run() {System.out.println(8360);}}
    public static class Impl8361 implements DemoService {public int type() {return 8361;} public void run() {System.out.println(8361);}}
    public static class Impl8362 implements DemoService {public int type() {return 8362;} public void run() {System.out.println(8362);}}
    public static class Impl8363 implements DemoService {public int type() {return 8363;} public void run() {System.out.println(8363);}}
    public static class Impl8364 implements DemoService {public int type() {return 8364;} public void run() {System.out.println(8364);}}
    public static class Impl8365 implements DemoService {public int type() {return 8365;} public void run() {System.out.println(8365);}}
    public static class Impl8366 implements DemoService {public int type() {return 8366;} public void run() {System.out.println(8366);}}
    public static class Impl8367 implements DemoService {public int type() {return 8367;} public void run() {System.out.println(8367);}}
    public static class Impl8368 implements DemoService {public int type() {return 8368;} public void run() {System.out.println(8368);}}
    public static class Impl8369 implements DemoService {public int type() {return 8369;} public void run() {System.out.println(8369);}}
    public static class Impl8370 implements DemoService {public int type() {return 8370;} public void run() {System.out.println(8370);}}
    public static class Impl8371 implements DemoService {public int type() {return 8371;} public void run() {System.out.println(8371);}}
    public static class Impl8372 implements DemoService {public int type() {return 8372;} public void run() {System.out.println(8372);}}
    public static class Impl8373 implements DemoService {public int type() {return 8373;} public void run() {System.out.println(8373);}}
    public static class Impl8374 implements DemoService {public int type() {return 8374;} public void run() {System.out.println(8374);}}
    public static class Impl8375 implements DemoService {public int type() {return 8375;} public void run() {System.out.println(8375);}}
    public static class Impl8376 implements DemoService {public int type() {return 8376;} public void run() {System.out.println(8376);}}
    public static class Impl8377 implements DemoService {public int type() {return 8377;} public void run() {System.out.println(8377);}}
    public static class Impl8378 implements DemoService {public int type() {return 8378;} public void run() {System.out.println(8378);}}
    public static class Impl8379 implements DemoService {public int type() {return 8379;} public void run() {System.out.println(8379);}}
    public static class Impl8380 implements DemoService {public int type() {return 8380;} public void run() {System.out.println(8380);}}
    public static class Impl8381 implements DemoService {public int type() {return 8381;} public void run() {System.out.println(8381);}}
    public static class Impl8382 implements DemoService {public int type() {return 8382;} public void run() {System.out.println(8382);}}
    public static class Impl8383 implements DemoService {public int type() {return 8383;} public void run() {System.out.println(8383);}}
    public static class Impl8384 implements DemoService {public int type() {return 8384;} public void run() {System.out.println(8384);}}
    public static class Impl8385 implements DemoService {public int type() {return 8385;} public void run() {System.out.println(8385);}}
    public static class Impl8386 implements DemoService {public int type() {return 8386;} public void run() {System.out.println(8386);}}
    public static class Impl8387 implements DemoService {public int type() {return 8387;} public void run() {System.out.println(8387);}}
    public static class Impl8388 implements DemoService {public int type() {return 8388;} public void run() {System.out.println(8388);}}
    public static class Impl8389 implements DemoService {public int type() {return 8389;} public void run() {System.out.println(8389);}}
    public static class Impl8390 implements DemoService {public int type() {return 8390;} public void run() {System.out.println(8390);}}
    public static class Impl8391 implements DemoService {public int type() {return 8391;} public void run() {System.out.println(8391);}}
    public static class Impl8392 implements DemoService {public int type() {return 8392;} public void run() {System.out.println(8392);}}
    public static class Impl8393 implements DemoService {public int type() {return 8393;} public void run() {System.out.println(8393);}}
    public static class Impl8394 implements DemoService {public int type() {return 8394;} public void run() {System.out.println(8394);}}
    public static class Impl8395 implements DemoService {public int type() {return 8395;} public void run() {System.out.println(8395);}}
    public static class Impl8396 implements DemoService {public int type() {return 8396;} public void run() {System.out.println(8396);}}
    public static class Impl8397 implements DemoService {public int type() {return 8397;} public void run() {System.out.println(8397);}}
    public static class Impl8398 implements DemoService {public int type() {return 8398;} public void run() {System.out.println(8398);}}
    public static class Impl8399 implements DemoService {public int type() {return 8399;} public void run() {System.out.println(8399);}}
    public static class Impl8400 implements DemoService {public int type() {return 8400;} public void run() {System.out.println(8400);}}
    public static class Impl8401 implements DemoService {public int type() {return 8401;} public void run() {System.out.println(8401);}}
    public static class Impl8402 implements DemoService {public int type() {return 8402;} public void run() {System.out.println(8402);}}
    public static class Impl8403 implements DemoService {public int type() {return 8403;} public void run() {System.out.println(8403);}}
    public static class Impl8404 implements DemoService {public int type() {return 8404;} public void run() {System.out.println(8404);}}
    public static class Impl8405 implements DemoService {public int type() {return 8405;} public void run() {System.out.println(8405);}}
    public static class Impl8406 implements DemoService {public int type() {return 8406;} public void run() {System.out.println(8406);}}
    public static class Impl8407 implements DemoService {public int type() {return 8407;} public void run() {System.out.println(8407);}}
    public static class Impl8408 implements DemoService {public int type() {return 8408;} public void run() {System.out.println(8408);}}
    public static class Impl8409 implements DemoService {public int type() {return 8409;} public void run() {System.out.println(8409);}}
    public static class Impl8410 implements DemoService {public int type() {return 8410;} public void run() {System.out.println(8410);}}
    public static class Impl8411 implements DemoService {public int type() {return 8411;} public void run() {System.out.println(8411);}}
    public static class Impl8412 implements DemoService {public int type() {return 8412;} public void run() {System.out.println(8412);}}
    public static class Impl8413 implements DemoService {public int type() {return 8413;} public void run() {System.out.println(8413);}}
    public static class Impl8414 implements DemoService {public int type() {return 8414;} public void run() {System.out.println(8414);}}
    public static class Impl8415 implements DemoService {public int type() {return 8415;} public void run() {System.out.println(8415);}}
    public static class Impl8416 implements DemoService {public int type() {return 8416;} public void run() {System.out.println(8416);}}
    public static class Impl8417 implements DemoService {public int type() {return 8417;} public void run() {System.out.println(8417);}}
    public static class Impl8418 implements DemoService {public int type() {return 8418;} public void run() {System.out.println(8418);}}
    public static class Impl8419 implements DemoService {public int type() {return 8419;} public void run() {System.out.println(8419);}}
    public static class Impl8420 implements DemoService {public int type() {return 8420;} public void run() {System.out.println(8420);}}
    public static class Impl8421 implements DemoService {public int type() {return 8421;} public void run() {System.out.println(8421);}}
    public static class Impl8422 implements DemoService {public int type() {return 8422;} public void run() {System.out.println(8422);}}
    public static class Impl8423 implements DemoService {public int type() {return 8423;} public void run() {System.out.println(8423);}}
    public static class Impl8424 implements DemoService {public int type() {return 8424;} public void run() {System.out.println(8424);}}
    public static class Impl8425 implements DemoService {public int type() {return 8425;} public void run() {System.out.println(8425);}}
    public static class Impl8426 implements DemoService {public int type() {return 8426;} public void run() {System.out.println(8426);}}
    public static class Impl8427 implements DemoService {public int type() {return 8427;} public void run() {System.out.println(8427);}}
    public static class Impl8428 implements DemoService {public int type() {return 8428;} public void run() {System.out.println(8428);}}
    public static class Impl8429 implements DemoService {public int type() {return 8429;} public void run() {System.out.println(8429);}}
    public static class Impl8430 implements DemoService {public int type() {return 8430;} public void run() {System.out.println(8430);}}
    public static class Impl8431 implements DemoService {public int type() {return 8431;} public void run() {System.out.println(8431);}}
    public static class Impl8432 implements DemoService {public int type() {return 8432;} public void run() {System.out.println(8432);}}
    public static class Impl8433 implements DemoService {public int type() {return 8433;} public void run() {System.out.println(8433);}}
    public static class Impl8434 implements DemoService {public int type() {return 8434;} public void run() {System.out.println(8434);}}
    public static class Impl8435 implements DemoService {public int type() {return 8435;} public void run() {System.out.println(8435);}}
    public static class Impl8436 implements DemoService {public int type() {return 8436;} public void run() {System.out.println(8436);}}
    public static class Impl8437 implements DemoService {public int type() {return 8437;} public void run() {System.out.println(8437);}}
    public static class Impl8438 implements DemoService {public int type() {return 8438;} public void run() {System.out.println(8438);}}
    public static class Impl8439 implements DemoService {public int type() {return 8439;} public void run() {System.out.println(8439);}}
    public static class Impl8440 implements DemoService {public int type() {return 8440;} public void run() {System.out.println(8440);}}
    public static class Impl8441 implements DemoService {public int type() {return 8441;} public void run() {System.out.println(8441);}}
    public static class Impl8442 implements DemoService {public int type() {return 8442;} public void run() {System.out.println(8442);}}
    public static class Impl8443 implements DemoService {public int type() {return 8443;} public void run() {System.out.println(8443);}}
    public static class Impl8444 implements DemoService {public int type() {return 8444;} public void run() {System.out.println(8444);}}
    public static class Impl8445 implements DemoService {public int type() {return 8445;} public void run() {System.out.println(8445);}}
    public static class Impl8446 implements DemoService {public int type() {return 8446;} public void run() {System.out.println(8446);}}
    public static class Impl8447 implements DemoService {public int type() {return 8447;} public void run() {System.out.println(8447);}}
    public static class Impl8448 implements DemoService {public int type() {return 8448;} public void run() {System.out.println(8448);}}
    public static class Impl8449 implements DemoService {public int type() {return 8449;} public void run() {System.out.println(8449);}}
    public static class Impl8450 implements DemoService {public int type() {return 8450;} public void run() {System.out.println(8450);}}
    public static class Impl8451 implements DemoService {public int type() {return 8451;} public void run() {System.out.println(8451);}}
    public static class Impl8452 implements DemoService {public int type() {return 8452;} public void run() {System.out.println(8452);}}
    public static class Impl8453 implements DemoService {public int type() {return 8453;} public void run() {System.out.println(8453);}}
    public static class Impl8454 implements DemoService {public int type() {return 8454;} public void run() {System.out.println(8454);}}
    public static class Impl8455 implements DemoService {public int type() {return 8455;} public void run() {System.out.println(8455);}}
    public static class Impl8456 implements DemoService {public int type() {return 8456;} public void run() {System.out.println(8456);}}
    public static class Impl8457 implements DemoService {public int type() {return 8457;} public void run() {System.out.println(8457);}}
    public static class Impl8458 implements DemoService {public int type() {return 8458;} public void run() {System.out.println(8458);}}
    public static class Impl8459 implements DemoService {public int type() {return 8459;} public void run() {System.out.println(8459);}}
    public static class Impl8460 implements DemoService {public int type() {return 8460;} public void run() {System.out.println(8460);}}
    public static class Impl8461 implements DemoService {public int type() {return 8461;} public void run() {System.out.println(8461);}}
    public static class Impl8462 implements DemoService {public int type() {return 8462;} public void run() {System.out.println(8462);}}
    public static class Impl8463 implements DemoService {public int type() {return 8463;} public void run() {System.out.println(8463);}}
    public static class Impl8464 implements DemoService {public int type() {return 8464;} public void run() {System.out.println(8464);}}
    public static class Impl8465 implements DemoService {public int type() {return 8465;} public void run() {System.out.println(8465);}}
    public static class Impl8466 implements DemoService {public int type() {return 8466;} public void run() {System.out.println(8466);}}
    public static class Impl8467 implements DemoService {public int type() {return 8467;} public void run() {System.out.println(8467);}}
    public static class Impl8468 implements DemoService {public int type() {return 8468;} public void run() {System.out.println(8468);}}
    public static class Impl8469 implements DemoService {public int type() {return 8469;} public void run() {System.out.println(8469);}}
    public static class Impl8470 implements DemoService {public int type() {return 8470;} public void run() {System.out.println(8470);}}
    public static class Impl8471 implements DemoService {public int type() {return 8471;} public void run() {System.out.println(8471);}}
    public static class Impl8472 implements DemoService {public int type() {return 8472;} public void run() {System.out.println(8472);}}
    public static class Impl8473 implements DemoService {public int type() {return 8473;} public void run() {System.out.println(8473);}}
    public static class Impl8474 implements DemoService {public int type() {return 8474;} public void run() {System.out.println(8474);}}
    public static class Impl8475 implements DemoService {public int type() {return 8475;} public void run() {System.out.println(8475);}}
    public static class Impl8476 implements DemoService {public int type() {return 8476;} public void run() {System.out.println(8476);}}
    public static class Impl8477 implements DemoService {public int type() {return 8477;} public void run() {System.out.println(8477);}}
    public static class Impl8478 implements DemoService {public int type() {return 8478;} public void run() {System.out.println(8478);}}
    public static class Impl8479 implements DemoService {public int type() {return 8479;} public void run() {System.out.println(8479);}}
    public static class Impl8480 implements DemoService {public int type() {return 8480;} public void run() {System.out.println(8480);}}
    public static class Impl8481 implements DemoService {public int type() {return 8481;} public void run() {System.out.println(8481);}}
    public static class Impl8482 implements DemoService {public int type() {return 8482;} public void run() {System.out.println(8482);}}
    public static class Impl8483 implements DemoService {public int type() {return 8483;} public void run() {System.out.println(8483);}}
    public static class Impl8484 implements DemoService {public int type() {return 8484;} public void run() {System.out.println(8484);}}
    public static class Impl8485 implements DemoService {public int type() {return 8485;} public void run() {System.out.println(8485);}}
    public static class Impl8486 implements DemoService {public int type() {return 8486;} public void run() {System.out.println(8486);}}
    public static class Impl8487 implements DemoService {public int type() {return 8487;} public void run() {System.out.println(8487);}}
    public static class Impl8488 implements DemoService {public int type() {return 8488;} public void run() {System.out.println(8488);}}
    public static class Impl8489 implements DemoService {public int type() {return 8489;} public void run() {System.out.println(8489);}}
    public static class Impl8490 implements DemoService {public int type() {return 8490;} public void run() {System.out.println(8490);}}
    public static class Impl8491 implements DemoService {public int type() {return 8491;} public void run() {System.out.println(8491);}}
    public static class Impl8492 implements DemoService {public int type() {return 8492;} public void run() {System.out.println(8492);}}
    public static class Impl8493 implements DemoService {public int type() {return 8493;} public void run() {System.out.println(8493);}}
    public static class Impl8494 implements DemoService {public int type() {return 8494;} public void run() {System.out.println(8494);}}
    public static class Impl8495 implements DemoService {public int type() {return 8495;} public void run() {System.out.println(8495);}}
    public static class Impl8496 implements DemoService {public int type() {return 8496;} public void run() {System.out.println(8496);}}
    public static class Impl8497 implements DemoService {public int type() {return 8497;} public void run() {System.out.println(8497);}}
    public static class Impl8498 implements DemoService {public int type() {return 8498;} public void run() {System.out.println(8498);}}
    public static class Impl8499 implements DemoService {public int type() {return 8499;} public void run() {System.out.println(8499);}}
    public static class Impl8500 implements DemoService {public int type() {return 8500;} public void run() {System.out.println(8500);}}
    public static class Impl8501 implements DemoService {public int type() {return 8501;} public void run() {System.out.println(8501);}}
    public static class Impl8502 implements DemoService {public int type() {return 8502;} public void run() {System.out.println(8502);}}
    public static class Impl8503 implements DemoService {public int type() {return 8503;} public void run() {System.out.println(8503);}}
    public static class Impl8504 implements DemoService {public int type() {return 8504;} public void run() {System.out.println(8504);}}
    public static class Impl8505 implements DemoService {public int type() {return 8505;} public void run() {System.out.println(8505);}}
    public static class Impl8506 implements DemoService {public int type() {return 8506;} public void run() {System.out.println(8506);}}
    public static class Impl8507 implements DemoService {public int type() {return 8507;} public void run() {System.out.println(8507);}}
    public static class Impl8508 implements DemoService {public int type() {return 8508;} public void run() {System.out.println(8508);}}
    public static class Impl8509 implements DemoService {public int type() {return 8509;} public void run() {System.out.println(8509);}}
    public static class Impl8510 implements DemoService {public int type() {return 8510;} public void run() {System.out.println(8510);}}
    public static class Impl8511 implements DemoService {public int type() {return 8511;} public void run() {System.out.println(8511);}}
    public static class Impl8512 implements DemoService {public int type() {return 8512;} public void run() {System.out.println(8512);}}
    public static class Impl8513 implements DemoService {public int type() {return 8513;} public void run() {System.out.println(8513);}}
    public static class Impl8514 implements DemoService {public int type() {return 8514;} public void run() {System.out.println(8514);}}
    public static class Impl8515 implements DemoService {public int type() {return 8515;} public void run() {System.out.println(8515);}}
    public static class Impl8516 implements DemoService {public int type() {return 8516;} public void run() {System.out.println(8516);}}
    public static class Impl8517 implements DemoService {public int type() {return 8517;} public void run() {System.out.println(8517);}}
    public static class Impl8518 implements DemoService {public int type() {return 8518;} public void run() {System.out.println(8518);}}
    public static class Impl8519 implements DemoService {public int type() {return 8519;} public void run() {System.out.println(8519);}}
    public static class Impl8520 implements DemoService {public int type() {return 8520;} public void run() {System.out.println(8520);}}
    public static class Impl8521 implements DemoService {public int type() {return 8521;} public void run() {System.out.println(8521);}}
    public static class Impl8522 implements DemoService {public int type() {return 8522;} public void run() {System.out.println(8522);}}
    public static class Impl8523 implements DemoService {public int type() {return 8523;} public void run() {System.out.println(8523);}}
    public static class Impl8524 implements DemoService {public int type() {return 8524;} public void run() {System.out.println(8524);}}
    public static class Impl8525 implements DemoService {public int type() {return 8525;} public void run() {System.out.println(8525);}}
    public static class Impl8526 implements DemoService {public int type() {return 8526;} public void run() {System.out.println(8526);}}
    public static class Impl8527 implements DemoService {public int type() {return 8527;} public void run() {System.out.println(8527);}}
    public static class Impl8528 implements DemoService {public int type() {return 8528;} public void run() {System.out.println(8528);}}
    public static class Impl8529 implements DemoService {public int type() {return 8529;} public void run() {System.out.println(8529);}}
    public static class Impl8530 implements DemoService {public int type() {return 8530;} public void run() {System.out.println(8530);}}
    public static class Impl8531 implements DemoService {public int type() {return 8531;} public void run() {System.out.println(8531);}}
    public static class Impl8532 implements DemoService {public int type() {return 8532;} public void run() {System.out.println(8532);}}
    public static class Impl8533 implements DemoService {public int type() {return 8533;} public void run() {System.out.println(8533);}}
    public static class Impl8534 implements DemoService {public int type() {return 8534;} public void run() {System.out.println(8534);}}
    public static class Impl8535 implements DemoService {public int type() {return 8535;} public void run() {System.out.println(8535);}}
    public static class Impl8536 implements DemoService {public int type() {return 8536;} public void run() {System.out.println(8536);}}
    public static class Impl8537 implements DemoService {public int type() {return 8537;} public void run() {System.out.println(8537);}}
    public static class Impl8538 implements DemoService {public int type() {return 8538;} public void run() {System.out.println(8538);}}
    public static class Impl8539 implements DemoService {public int type() {return 8539;} public void run() {System.out.println(8539);}}
    public static class Impl8540 implements DemoService {public int type() {return 8540;} public void run() {System.out.println(8540);}}
    public static class Impl8541 implements DemoService {public int type() {return 8541;} public void run() {System.out.println(8541);}}
    public static class Impl8542 implements DemoService {public int type() {return 8542;} public void run() {System.out.println(8542);}}
    public static class Impl8543 implements DemoService {public int type() {return 8543;} public void run() {System.out.println(8543);}}
    public static class Impl8544 implements DemoService {public int type() {return 8544;} public void run() {System.out.println(8544);}}
    public static class Impl8545 implements DemoService {public int type() {return 8545;} public void run() {System.out.println(8545);}}
    public static class Impl8546 implements DemoService {public int type() {return 8546;} public void run() {System.out.println(8546);}}
    public static class Impl8547 implements DemoService {public int type() {return 8547;} public void run() {System.out.println(8547);}}
    public static class Impl8548 implements DemoService {public int type() {return 8548;} public void run() {System.out.println(8548);}}
    public static class Impl8549 implements DemoService {public int type() {return 8549;} public void run() {System.out.println(8549);}}
    public static class Impl8550 implements DemoService {public int type() {return 8550;} public void run() {System.out.println(8550);}}
    public static class Impl8551 implements DemoService {public int type() {return 8551;} public void run() {System.out.println(8551);}}
    public static class Impl8552 implements DemoService {public int type() {return 8552;} public void run() {System.out.println(8552);}}
    public static class Impl8553 implements DemoService {public int type() {return 8553;} public void run() {System.out.println(8553);}}
    public static class Impl8554 implements DemoService {public int type() {return 8554;} public void run() {System.out.println(8554);}}
    public static class Impl8555 implements DemoService {public int type() {return 8555;} public void run() {System.out.println(8555);}}
    public static class Impl8556 implements DemoService {public int type() {return 8556;} public void run() {System.out.println(8556);}}
    public static class Impl8557 implements DemoService {public int type() {return 8557;} public void run() {System.out.println(8557);}}
    public static class Impl8558 implements DemoService {public int type() {return 8558;} public void run() {System.out.println(8558);}}
    public static class Impl8559 implements DemoService {public int type() {return 8559;} public void run() {System.out.println(8559);}}
    public static class Impl8560 implements DemoService {public int type() {return 8560;} public void run() {System.out.println(8560);}}
    public static class Impl8561 implements DemoService {public int type() {return 8561;} public void run() {System.out.println(8561);}}
    public static class Impl8562 implements DemoService {public int type() {return 8562;} public void run() {System.out.println(8562);}}
    public static class Impl8563 implements DemoService {public int type() {return 8563;} public void run() {System.out.println(8563);}}
    public static class Impl8564 implements DemoService {public int type() {return 8564;} public void run() {System.out.println(8564);}}
    public static class Impl8565 implements DemoService {public int type() {return 8565;} public void run() {System.out.println(8565);}}
    public static class Impl8566 implements DemoService {public int type() {return 8566;} public void run() {System.out.println(8566);}}
    public static class Impl8567 implements DemoService {public int type() {return 8567;} public void run() {System.out.println(8567);}}
    public static class Impl8568 implements DemoService {public int type() {return 8568;} public void run() {System.out.println(8568);}}
    public static class Impl8569 implements DemoService {public int type() {return 8569;} public void run() {System.out.println(8569);}}
    public static class Impl8570 implements DemoService {public int type() {return 8570;} public void run() {System.out.println(8570);}}
    public static class Impl8571 implements DemoService {public int type() {return 8571;} public void run() {System.out.println(8571);}}
    public static class Impl8572 implements DemoService {public int type() {return 8572;} public void run() {System.out.println(8572);}}
    public static class Impl8573 implements DemoService {public int type() {return 8573;} public void run() {System.out.println(8573);}}
    public static class Impl8574 implements DemoService {public int type() {return 8574;} public void run() {System.out.println(8574);}}
    public static class Impl8575 implements DemoService {public int type() {return 8575;} public void run() {System.out.println(8575);}}
    public static class Impl8576 implements DemoService {public int type() {return 8576;} public void run() {System.out.println(8576);}}
    public static class Impl8577 implements DemoService {public int type() {return 8577;} public void run() {System.out.println(8577);}}
    public static class Impl8578 implements DemoService {public int type() {return 8578;} public void run() {System.out.println(8578);}}
    public static class Impl8579 implements DemoService {public int type() {return 8579;} public void run() {System.out.println(8579);}}
    public static class Impl8580 implements DemoService {public int type() {return 8580;} public void run() {System.out.println(8580);}}
    public static class Impl8581 implements DemoService {public int type() {return 8581;} public void run() {System.out.println(8581);}}
    public static class Impl8582 implements DemoService {public int type() {return 8582;} public void run() {System.out.println(8582);}}
    public static class Impl8583 implements DemoService {public int type() {return 8583;} public void run() {System.out.println(8583);}}
    public static class Impl8584 implements DemoService {public int type() {return 8584;} public void run() {System.out.println(8584);}}
    public static class Impl8585 implements DemoService {public int type() {return 8585;} public void run() {System.out.println(8585);}}
    public static class Impl8586 implements DemoService {public int type() {return 8586;} public void run() {System.out.println(8586);}}
    public static class Impl8587 implements DemoService {public int type() {return 8587;} public void run() {System.out.println(8587);}}
    public static class Impl8588 implements DemoService {public int type() {return 8588;} public void run() {System.out.println(8588);}}
    public static class Impl8589 implements DemoService {public int type() {return 8589;} public void run() {System.out.println(8589);}}
    public static class Impl8590 implements DemoService {public int type() {return 8590;} public void run() {System.out.println(8590);}}
    public static class Impl8591 implements DemoService {public int type() {return 8591;} public void run() {System.out.println(8591);}}
    public static class Impl8592 implements DemoService {public int type() {return 8592;} public void run() {System.out.println(8592);}}
    public static class Impl8593 implements DemoService {public int type() {return 8593;} public void run() {System.out.println(8593);}}
    public static class Impl8594 implements DemoService {public int type() {return 8594;} public void run() {System.out.println(8594);}}
    public static class Impl8595 implements DemoService {public int type() {return 8595;} public void run() {System.out.println(8595);}}
    public static class Impl8596 implements DemoService {public int type() {return 8596;} public void run() {System.out.println(8596);}}
    public static class Impl8597 implements DemoService {public int type() {return 8597;} public void run() {System.out.println(8597);}}
    public static class Impl8598 implements DemoService {public int type() {return 8598;} public void run() {System.out.println(8598);}}
    public static class Impl8599 implements DemoService {public int type() {return 8599;} public void run() {System.out.println(8599);}}
    public static class Impl8600 implements DemoService {public int type() {return 8600;} public void run() {System.out.println(8600);}}
    public static class Impl8601 implements DemoService {public int type() {return 8601;} public void run() {System.out.println(8601);}}
    public static class Impl8602 implements DemoService {public int type() {return 8602;} public void run() {System.out.println(8602);}}
    public static class Impl8603 implements DemoService {public int type() {return 8603;} public void run() {System.out.println(8603);}}
    public static class Impl8604 implements DemoService {public int type() {return 8604;} public void run() {System.out.println(8604);}}
    public static class Impl8605 implements DemoService {public int type() {return 8605;} public void run() {System.out.println(8605);}}
    public static class Impl8606 implements DemoService {public int type() {return 8606;} public void run() {System.out.println(8606);}}
    public static class Impl8607 implements DemoService {public int type() {return 8607;} public void run() {System.out.println(8607);}}
    public static class Impl8608 implements DemoService {public int type() {return 8608;} public void run() {System.out.println(8608);}}
    public static class Impl8609 implements DemoService {public int type() {return 8609;} public void run() {System.out.println(8609);}}
    public static class Impl8610 implements DemoService {public int type() {return 8610;} public void run() {System.out.println(8610);}}
    public static class Impl8611 implements DemoService {public int type() {return 8611;} public void run() {System.out.println(8611);}}
    public static class Impl8612 implements DemoService {public int type() {return 8612;} public void run() {System.out.println(8612);}}
    public static class Impl8613 implements DemoService {public int type() {return 8613;} public void run() {System.out.println(8613);}}
    public static class Impl8614 implements DemoService {public int type() {return 8614;} public void run() {System.out.println(8614);}}
    public static class Impl8615 implements DemoService {public int type() {return 8615;} public void run() {System.out.println(8615);}}
    public static class Impl8616 implements DemoService {public int type() {return 8616;} public void run() {System.out.println(8616);}}
    public static class Impl8617 implements DemoService {public int type() {return 8617;} public void run() {System.out.println(8617);}}
    public static class Impl8618 implements DemoService {public int type() {return 8618;} public void run() {System.out.println(8618);}}
    public static class Impl8619 implements DemoService {public int type() {return 8619;} public void run() {System.out.println(8619);}}
    public static class Impl8620 implements DemoService {public int type() {return 8620;} public void run() {System.out.println(8620);}}
    public static class Impl8621 implements DemoService {public int type() {return 8621;} public void run() {System.out.println(8621);}}
    public static class Impl8622 implements DemoService {public int type() {return 8622;} public void run() {System.out.println(8622);}}
    public static class Impl8623 implements DemoService {public int type() {return 8623;} public void run() {System.out.println(8623);}}
    public static class Impl8624 implements DemoService {public int type() {return 8624;} public void run() {System.out.println(8624);}}
    public static class Impl8625 implements DemoService {public int type() {return 8625;} public void run() {System.out.println(8625);}}
    public static class Impl8626 implements DemoService {public int type() {return 8626;} public void run() {System.out.println(8626);}}
    public static class Impl8627 implements DemoService {public int type() {return 8627;} public void run() {System.out.println(8627);}}
    public static class Impl8628 implements DemoService {public int type() {return 8628;} public void run() {System.out.println(8628);}}
    public static class Impl8629 implements DemoService {public int type() {return 8629;} public void run() {System.out.println(8629);}}
    public static class Impl8630 implements DemoService {public int type() {return 8630;} public void run() {System.out.println(8630);}}
    public static class Impl8631 implements DemoService {public int type() {return 8631;} public void run() {System.out.println(8631);}}
    public static class Impl8632 implements DemoService {public int type() {return 8632;} public void run() {System.out.println(8632);}}
    public static class Impl8633 implements DemoService {public int type() {return 8633;} public void run() {System.out.println(8633);}}
    public static class Impl8634 implements DemoService {public int type() {return 8634;} public void run() {System.out.println(8634);}}
    public static class Impl8635 implements DemoService {public int type() {return 8635;} public void run() {System.out.println(8635);}}
    public static class Impl8636 implements DemoService {public int type() {return 8636;} public void run() {System.out.println(8636);}}
    public static class Impl8637 implements DemoService {public int type() {return 8637;} public void run() {System.out.println(8637);}}
    public static class Impl8638 implements DemoService {public int type() {return 8638;} public void run() {System.out.println(8638);}}
    public static class Impl8639 implements DemoService {public int type() {return 8639;} public void run() {System.out.println(8639);}}
    public static class Impl8640 implements DemoService {public int type() {return 8640;} public void run() {System.out.println(8640);}}
    public static class Impl8641 implements DemoService {public int type() {return 8641;} public void run() {System.out.println(8641);}}
    public static class Impl8642 implements DemoService {public int type() {return 8642;} public void run() {System.out.println(8642);}}
    public static class Impl8643 implements DemoService {public int type() {return 8643;} public void run() {System.out.println(8643);}}
    public static class Impl8644 implements DemoService {public int type() {return 8644;} public void run() {System.out.println(8644);}}
    public static class Impl8645 implements DemoService {public int type() {return 8645;} public void run() {System.out.println(8645);}}
    public static class Impl8646 implements DemoService {public int type() {return 8646;} public void run() {System.out.println(8646);}}
    public static class Impl8647 implements DemoService {public int type() {return 8647;} public void run() {System.out.println(8647);}}
    public static class Impl8648 implements DemoService {public int type() {return 8648;} public void run() {System.out.println(8648);}}
    public static class Impl8649 implements DemoService {public int type() {return 8649;} public void run() {System.out.println(8649);}}
    public static class Impl8650 implements DemoService {public int type() {return 8650;} public void run() {System.out.println(8650);}}
    public static class Impl8651 implements DemoService {public int type() {return 8651;} public void run() {System.out.println(8651);}}
    public static class Impl8652 implements DemoService {public int type() {return 8652;} public void run() {System.out.println(8652);}}
    public static class Impl8653 implements DemoService {public int type() {return 8653;} public void run() {System.out.println(8653);}}
    public static class Impl8654 implements DemoService {public int type() {return 8654;} public void run() {System.out.println(8654);}}
    public static class Impl8655 implements DemoService {public int type() {return 8655;} public void run() {System.out.println(8655);}}
    public static class Impl8656 implements DemoService {public int type() {return 8656;} public void run() {System.out.println(8656);}}
    public static class Impl8657 implements DemoService {public int type() {return 8657;} public void run() {System.out.println(8657);}}
    public static class Impl8658 implements DemoService {public int type() {return 8658;} public void run() {System.out.println(8658);}}
    public static class Impl8659 implements DemoService {public int type() {return 8659;} public void run() {System.out.println(8659);}}
    public static class Impl8660 implements DemoService {public int type() {return 8660;} public void run() {System.out.println(8660);}}
    public static class Impl8661 implements DemoService {public int type() {return 8661;} public void run() {System.out.println(8661);}}
    public static class Impl8662 implements DemoService {public int type() {return 8662;} public void run() {System.out.println(8662);}}
    public static class Impl8663 implements DemoService {public int type() {return 8663;} public void run() {System.out.println(8663);}}
    public static class Impl8664 implements DemoService {public int type() {return 8664;} public void run() {System.out.println(8664);}}
    public static class Impl8665 implements DemoService {public int type() {return 8665;} public void run() {System.out.println(8665);}}
    public static class Impl8666 implements DemoService {public int type() {return 8666;} public void run() {System.out.println(8666);}}
    public static class Impl8667 implements DemoService {public int type() {return 8667;} public void run() {System.out.println(8667);}}
    public static class Impl8668 implements DemoService {public int type() {return 8668;} public void run() {System.out.println(8668);}}
    public static class Impl8669 implements DemoService {public int type() {return 8669;} public void run() {System.out.println(8669);}}
    public static class Impl8670 implements DemoService {public int type() {return 8670;} public void run() {System.out.println(8670);}}
    public static class Impl8671 implements DemoService {public int type() {return 8671;} public void run() {System.out.println(8671);}}
    public static class Impl8672 implements DemoService {public int type() {return 8672;} public void run() {System.out.println(8672);}}
    public static class Impl8673 implements DemoService {public int type() {return 8673;} public void run() {System.out.println(8673);}}
    public static class Impl8674 implements DemoService {public int type() {return 8674;} public void run() {System.out.println(8674);}}
    public static class Impl8675 implements DemoService {public int type() {return 8675;} public void run() {System.out.println(8675);}}
    public static class Impl8676 implements DemoService {public int type() {return 8676;} public void run() {System.out.println(8676);}}
    public static class Impl8677 implements DemoService {public int type() {return 8677;} public void run() {System.out.println(8677);}}
    public static class Impl8678 implements DemoService {public int type() {return 8678;} public void run() {System.out.println(8678);}}
    public static class Impl8679 implements DemoService {public int type() {return 8679;} public void run() {System.out.println(8679);}}
    public static class Impl8680 implements DemoService {public int type() {return 8680;} public void run() {System.out.println(8680);}}
    public static class Impl8681 implements DemoService {public int type() {return 8681;} public void run() {System.out.println(8681);}}
    public static class Impl8682 implements DemoService {public int type() {return 8682;} public void run() {System.out.println(8682);}}
    public static class Impl8683 implements DemoService {public int type() {return 8683;} public void run() {System.out.println(8683);}}
    public static class Impl8684 implements DemoService {public int type() {return 8684;} public void run() {System.out.println(8684);}}
    public static class Impl8685 implements DemoService {public int type() {return 8685;} public void run() {System.out.println(8685);}}
    public static class Impl8686 implements DemoService {public int type() {return 8686;} public void run() {System.out.println(8686);}}
    public static class Impl8687 implements DemoService {public int type() {return 8687;} public void run() {System.out.println(8687);}}
    public static class Impl8688 implements DemoService {public int type() {return 8688;} public void run() {System.out.println(8688);}}
    public static class Impl8689 implements DemoService {public int type() {return 8689;} public void run() {System.out.println(8689);}}
    public static class Impl8690 implements DemoService {public int type() {return 8690;} public void run() {System.out.println(8690);}}
    public static class Impl8691 implements DemoService {public int type() {return 8691;} public void run() {System.out.println(8691);}}
    public static class Impl8692 implements DemoService {public int type() {return 8692;} public void run() {System.out.println(8692);}}
    public static class Impl8693 implements DemoService {public int type() {return 8693;} public void run() {System.out.println(8693);}}
    public static class Impl8694 implements DemoService {public int type() {return 8694;} public void run() {System.out.println(8694);}}
    public static class Impl8695 implements DemoService {public int type() {return 8695;} public void run() {System.out.println(8695);}}
    public static class Impl8696 implements DemoService {public int type() {return 8696;} public void run() {System.out.println(8696);}}
    public static class Impl8697 implements DemoService {public int type() {return 8697;} public void run() {System.out.println(8697);}}
    public static class Impl8698 implements DemoService {public int type() {return 8698;} public void run() {System.out.println(8698);}}
    public static class Impl8699 implements DemoService {public int type() {return 8699;} public void run() {System.out.println(8699);}}
    public static class Impl8700 implements DemoService {public int type() {return 8700;} public void run() {System.out.println(8700);}}
    public static class Impl8701 implements DemoService {public int type() {return 8701;} public void run() {System.out.println(8701);}}
    public static class Impl8702 implements DemoService {public int type() {return 8702;} public void run() {System.out.println(8702);}}
    public static class Impl8703 implements DemoService {public int type() {return 8703;} public void run() {System.out.println(8703);}}
    public static class Impl8704 implements DemoService {public int type() {return 8704;} public void run() {System.out.println(8704);}}
    public static class Impl8705 implements DemoService {public int type() {return 8705;} public void run() {System.out.println(8705);}}
    public static class Impl8706 implements DemoService {public int type() {return 8706;} public void run() {System.out.println(8706);}}
    public static class Impl8707 implements DemoService {public int type() {return 8707;} public void run() {System.out.println(8707);}}
    public static class Impl8708 implements DemoService {public int type() {return 8708;} public void run() {System.out.println(8708);}}
    public static class Impl8709 implements DemoService {public int type() {return 8709;} public void run() {System.out.println(8709);}}
    public static class Impl8710 implements DemoService {public int type() {return 8710;} public void run() {System.out.println(8710);}}
    public static class Impl8711 implements DemoService {public int type() {return 8711;} public void run() {System.out.println(8711);}}
    public static class Impl8712 implements DemoService {public int type() {return 8712;} public void run() {System.out.println(8712);}}
    public static class Impl8713 implements DemoService {public int type() {return 8713;} public void run() {System.out.println(8713);}}
    public static class Impl8714 implements DemoService {public int type() {return 8714;} public void run() {System.out.println(8714);}}
    public static class Impl8715 implements DemoService {public int type() {return 8715;} public void run() {System.out.println(8715);}}
    public static class Impl8716 implements DemoService {public int type() {return 8716;} public void run() {System.out.println(8716);}}
    public static class Impl8717 implements DemoService {public int type() {return 8717;} public void run() {System.out.println(8717);}}
    public static class Impl8718 implements DemoService {public int type() {return 8718;} public void run() {System.out.println(8718);}}
    public static class Impl8719 implements DemoService {public int type() {return 8719;} public void run() {System.out.println(8719);}}
    public static class Impl8720 implements DemoService {public int type() {return 8720;} public void run() {System.out.println(8720);}}
    public static class Impl8721 implements DemoService {public int type() {return 8721;} public void run() {System.out.println(8721);}}
    public static class Impl8722 implements DemoService {public int type() {return 8722;} public void run() {System.out.println(8722);}}
    public static class Impl8723 implements DemoService {public int type() {return 8723;} public void run() {System.out.println(8723);}}
    public static class Impl8724 implements DemoService {public int type() {return 8724;} public void run() {System.out.println(8724);}}
    public static class Impl8725 implements DemoService {public int type() {return 8725;} public void run() {System.out.println(8725);}}
    public static class Impl8726 implements DemoService {public int type() {return 8726;} public void run() {System.out.println(8726);}}
    public static class Impl8727 implements DemoService {public int type() {return 8727;} public void run() {System.out.println(8727);}}
    public static class Impl8728 implements DemoService {public int type() {return 8728;} public void run() {System.out.println(8728);}}
    public static class Impl8729 implements DemoService {public int type() {return 8729;} public void run() {System.out.println(8729);}}
    public static class Impl8730 implements DemoService {public int type() {return 8730;} public void run() {System.out.println(8730);}}
    public static class Impl8731 implements DemoService {public int type() {return 8731;} public void run() {System.out.println(8731);}}
    public static class Impl8732 implements DemoService {public int type() {return 8732;} public void run() {System.out.println(8732);}}
    public static class Impl8733 implements DemoService {public int type() {return 8733;} public void run() {System.out.println(8733);}}
    public static class Impl8734 implements DemoService {public int type() {return 8734;} public void run() {System.out.println(8734);}}
    public static class Impl8735 implements DemoService {public int type() {return 8735;} public void run() {System.out.println(8735);}}
    public static class Impl8736 implements DemoService {public int type() {return 8736;} public void run() {System.out.println(8736);}}
    public static class Impl8737 implements DemoService {public int type() {return 8737;} public void run() {System.out.println(8737);}}
    public static class Impl8738 implements DemoService {public int type() {return 8738;} public void run() {System.out.println(8738);}}
    public static class Impl8739 implements DemoService {public int type() {return 8739;} public void run() {System.out.println(8739);}}
    public static class Impl8740 implements DemoService {public int type() {return 8740;} public void run() {System.out.println(8740);}}
    public static class Impl8741 implements DemoService {public int type() {return 8741;} public void run() {System.out.println(8741);}}
    public static class Impl8742 implements DemoService {public int type() {return 8742;} public void run() {System.out.println(8742);}}
    public static class Impl8743 implements DemoService {public int type() {return 8743;} public void run() {System.out.println(8743);}}
    public static class Impl8744 implements DemoService {public int type() {return 8744;} public void run() {System.out.println(8744);}}
    public static class Impl8745 implements DemoService {public int type() {return 8745;} public void run() {System.out.println(8745);}}
    public static class Impl8746 implements DemoService {public int type() {return 8746;} public void run() {System.out.println(8746);}}
    public static class Impl8747 implements DemoService {public int type() {return 8747;} public void run() {System.out.println(8747);}}
    public static class Impl8748 implements DemoService {public int type() {return 8748;} public void run() {System.out.println(8748);}}
    public static class Impl8749 implements DemoService {public int type() {return 8749;} public void run() {System.out.println(8749);}}
    public static class Impl8750 implements DemoService {public int type() {return 8750;} public void run() {System.out.println(8750);}}
    public static class Impl8751 implements DemoService {public int type() {return 8751;} public void run() {System.out.println(8751);}}
    public static class Impl8752 implements DemoService {public int type() {return 8752;} public void run() {System.out.println(8752);}}
    public static class Impl8753 implements DemoService {public int type() {return 8753;} public void run() {System.out.println(8753);}}
    public static class Impl8754 implements DemoService {public int type() {return 8754;} public void run() {System.out.println(8754);}}
    public static class Impl8755 implements DemoService {public int type() {return 8755;} public void run() {System.out.println(8755);}}
    public static class Impl8756 implements DemoService {public int type() {return 8756;} public void run() {System.out.println(8756);}}
    public static class Impl8757 implements DemoService {public int type() {return 8757;} public void run() {System.out.println(8757);}}
    public static class Impl8758 implements DemoService {public int type() {return 8758;} public void run() {System.out.println(8758);}}
    public static class Impl8759 implements DemoService {public int type() {return 8759;} public void run() {System.out.println(8759);}}
    public static class Impl8760 implements DemoService {public int type() {return 8760;} public void run() {System.out.println(8760);}}
    public static class Impl8761 implements DemoService {public int type() {return 8761;} public void run() {System.out.println(8761);}}
    public static class Impl8762 implements DemoService {public int type() {return 8762;} public void run() {System.out.println(8762);}}
    public static class Impl8763 implements DemoService {public int type() {return 8763;} public void run() {System.out.println(8763);}}
    public static class Impl8764 implements DemoService {public int type() {return 8764;} public void run() {System.out.println(8764);}}
    public static class Impl8765 implements DemoService {public int type() {return 8765;} public void run() {System.out.println(8765);}}
    public static class Impl8766 implements DemoService {public int type() {return 8766;} public void run() {System.out.println(8766);}}
    public static class Impl8767 implements DemoService {public int type() {return 8767;} public void run() {System.out.println(8767);}}
    public static class Impl8768 implements DemoService {public int type() {return 8768;} public void run() {System.out.println(8768);}}
    public static class Impl8769 implements DemoService {public int type() {return 8769;} public void run() {System.out.println(8769);}}
    public static class Impl8770 implements DemoService {public int type() {return 8770;} public void run() {System.out.println(8770);}}
    public static class Impl8771 implements DemoService {public int type() {return 8771;} public void run() {System.out.println(8771);}}
    public static class Impl8772 implements DemoService {public int type() {return 8772;} public void run() {System.out.println(8772);}}
    public static class Impl8773 implements DemoService {public int type() {return 8773;} public void run() {System.out.println(8773);}}
    public static class Impl8774 implements DemoService {public int type() {return 8774;} public void run() {System.out.println(8774);}}
    public static class Impl8775 implements DemoService {public int type() {return 8775;} public void run() {System.out.println(8775);}}
    public static class Impl8776 implements DemoService {public int type() {return 8776;} public void run() {System.out.println(8776);}}
    public static class Impl8777 implements DemoService {public int type() {return 8777;} public void run() {System.out.println(8777);}}
    public static class Impl8778 implements DemoService {public int type() {return 8778;} public void run() {System.out.println(8778);}}
    public static class Impl8779 implements DemoService {public int type() {return 8779;} public void run() {System.out.println(8779);}}
    public static class Impl8780 implements DemoService {public int type() {return 8780;} public void run() {System.out.println(8780);}}
    public static class Impl8781 implements DemoService {public int type() {return 8781;} public void run() {System.out.println(8781);}}
    public static class Impl8782 implements DemoService {public int type() {return 8782;} public void run() {System.out.println(8782);}}
    public static class Impl8783 implements DemoService {public int type() {return 8783;} public void run() {System.out.println(8783);}}
    public static class Impl8784 implements DemoService {public int type() {return 8784;} public void run() {System.out.println(8784);}}
    public static class Impl8785 implements DemoService {public int type() {return 8785;} public void run() {System.out.println(8785);}}
    public static class Impl8786 implements DemoService {public int type() {return 8786;} public void run() {System.out.println(8786);}}
    public static class Impl8787 implements DemoService {public int type() {return 8787;} public void run() {System.out.println(8787);}}
    public static class Impl8788 implements DemoService {public int type() {return 8788;} public void run() {System.out.println(8788);}}
    public static class Impl8789 implements DemoService {public int type() {return 8789;} public void run() {System.out.println(8789);}}
    public static class Impl8790 implements DemoService {public int type() {return 8790;} public void run() {System.out.println(8790);}}
    public static class Impl8791 implements DemoService {public int type() {return 8791;} public void run() {System.out.println(8791);}}
    public static class Impl8792 implements DemoService {public int type() {return 8792;} public void run() {System.out.println(8792);}}
    public static class Impl8793 implements DemoService {public int type() {return 8793;} public void run() {System.out.println(8793);}}
    public static class Impl8794 implements DemoService {public int type() {return 8794;} public void run() {System.out.println(8794);}}
    public static class Impl8795 implements DemoService {public int type() {return 8795;} public void run() {System.out.println(8795);}}
    public static class Impl8796 implements DemoService {public int type() {return 8796;} public void run() {System.out.println(8796);}}
    public static class Impl8797 implements DemoService {public int type() {return 8797;} public void run() {System.out.println(8797);}}
    public static class Impl8798 implements DemoService {public int type() {return 8798;} public void run() {System.out.println(8798);}}
    public static class Impl8799 implements DemoService {public int type() {return 8799;} public void run() {System.out.println(8799);}}
    public static class Impl8800 implements DemoService {public int type() {return 8800;} public void run() {System.out.println(8800);}}
    public static class Impl8801 implements DemoService {public int type() {return 8801;} public void run() {System.out.println(8801);}}
    public static class Impl8802 implements DemoService {public int type() {return 8802;} public void run() {System.out.println(8802);}}
    public static class Impl8803 implements DemoService {public int type() {return 8803;} public void run() {System.out.println(8803);}}
    public static class Impl8804 implements DemoService {public int type() {return 8804;} public void run() {System.out.println(8804);}}
    public static class Impl8805 implements DemoService {public int type() {return 8805;} public void run() {System.out.println(8805);}}
    public static class Impl8806 implements DemoService {public int type() {return 8806;} public void run() {System.out.println(8806);}}
    public static class Impl8807 implements DemoService {public int type() {return 8807;} public void run() {System.out.println(8807);}}
    public static class Impl8808 implements DemoService {public int type() {return 8808;} public void run() {System.out.println(8808);}}
    public static class Impl8809 implements DemoService {public int type() {return 8809;} public void run() {System.out.println(8809);}}
    public static class Impl8810 implements DemoService {public int type() {return 8810;} public void run() {System.out.println(8810);}}
    public static class Impl8811 implements DemoService {public int type() {return 8811;} public void run() {System.out.println(8811);}}
    public static class Impl8812 implements DemoService {public int type() {return 8812;} public void run() {System.out.println(8812);}}
    public static class Impl8813 implements DemoService {public int type() {return 8813;} public void run() {System.out.println(8813);}}
    public static class Impl8814 implements DemoService {public int type() {return 8814;} public void run() {System.out.println(8814);}}
    public static class Impl8815 implements DemoService {public int type() {return 8815;} public void run() {System.out.println(8815);}}
    public static class Impl8816 implements DemoService {public int type() {return 8816;} public void run() {System.out.println(8816);}}
    public static class Impl8817 implements DemoService {public int type() {return 8817;} public void run() {System.out.println(8817);}}
    public static class Impl8818 implements DemoService {public int type() {return 8818;} public void run() {System.out.println(8818);}}
    public static class Impl8819 implements DemoService {public int type() {return 8819;} public void run() {System.out.println(8819);}}
    public static class Impl8820 implements DemoService {public int type() {return 8820;} public void run() {System.out.println(8820);}}
    public static class Impl8821 implements DemoService {public int type() {return 8821;} public void run() {System.out.println(8821);}}
    public static class Impl8822 implements DemoService {public int type() {return 8822;} public void run() {System.out.println(8822);}}
    public static class Impl8823 implements DemoService {public int type() {return 8823;} public void run() {System.out.println(8823);}}
    public static class Impl8824 implements DemoService {public int type() {return 8824;} public void run() {System.out.println(8824);}}
    public static class Impl8825 implements DemoService {public int type() {return 8825;} public void run() {System.out.println(8825);}}
    public static class Impl8826 implements DemoService {public int type() {return 8826;} public void run() {System.out.println(8826);}}
    public static class Impl8827 implements DemoService {public int type() {return 8827;} public void run() {System.out.println(8827);}}
    public static class Impl8828 implements DemoService {public int type() {return 8828;} public void run() {System.out.println(8828);}}
    public static class Impl8829 implements DemoService {public int type() {return 8829;} public void run() {System.out.println(8829);}}
    public static class Impl8830 implements DemoService {public int type() {return 8830;} public void run() {System.out.println(8830);}}
    public static class Impl8831 implements DemoService {public int type() {return 8831;} public void run() {System.out.println(8831);}}
    public static class Impl8832 implements DemoService {public int type() {return 8832;} public void run() {System.out.println(8832);}}
    public static class Impl8833 implements DemoService {public int type() {return 8833;} public void run() {System.out.println(8833);}}
    public static class Impl8834 implements DemoService {public int type() {return 8834;} public void run() {System.out.println(8834);}}
    public static class Impl8835 implements DemoService {public int type() {return 8835;} public void run() {System.out.println(8835);}}
    public static class Impl8836 implements DemoService {public int type() {return 8836;} public void run() {System.out.println(8836);}}
    public static class Impl8837 implements DemoService {public int type() {return 8837;} public void run() {System.out.println(8837);}}
    public static class Impl8838 implements DemoService {public int type() {return 8838;} public void run() {System.out.println(8838);}}
    public static class Impl8839 implements DemoService {public int type() {return 8839;} public void run() {System.out.println(8839);}}
    public static class Impl8840 implements DemoService {public int type() {return 8840;} public void run() {System.out.println(8840);}}
    public static class Impl8841 implements DemoService {public int type() {return 8841;} public void run() {System.out.println(8841);}}
    public static class Impl8842 implements DemoService {public int type() {return 8842;} public void run() {System.out.println(8842);}}
    public static class Impl8843 implements DemoService {public int type() {return 8843;} public void run() {System.out.println(8843);}}
    public static class Impl8844 implements DemoService {public int type() {return 8844;} public void run() {System.out.println(8844);}}
    public static class Impl8845 implements DemoService {public int type() {return 8845;} public void run() {System.out.println(8845);}}
    public static class Impl8846 implements DemoService {public int type() {return 8846;} public void run() {System.out.println(8846);}}
    public static class Impl8847 implements DemoService {public int type() {return 8847;} public void run() {System.out.println(8847);}}
    public static class Impl8848 implements DemoService {public int type() {return 8848;} public void run() {System.out.println(8848);}}
    public static class Impl8849 implements DemoService {public int type() {return 8849;} public void run() {System.out.println(8849);}}
    public static class Impl8850 implements DemoService {public int type() {return 8850;} public void run() {System.out.println(8850);}}
    public static class Impl8851 implements DemoService {public int type() {return 8851;} public void run() {System.out.println(8851);}}
    public static class Impl8852 implements DemoService {public int type() {return 8852;} public void run() {System.out.println(8852);}}
    public static class Impl8853 implements DemoService {public int type() {return 8853;} public void run() {System.out.println(8853);}}
    public static class Impl8854 implements DemoService {public int type() {return 8854;} public void run() {System.out.println(8854);}}
    public static class Impl8855 implements DemoService {public int type() {return 8855;} public void run() {System.out.println(8855);}}
    public static class Impl8856 implements DemoService {public int type() {return 8856;} public void run() {System.out.println(8856);}}
    public static class Impl8857 implements DemoService {public int type() {return 8857;} public void run() {System.out.println(8857);}}
    public static class Impl8858 implements DemoService {public int type() {return 8858;} public void run() {System.out.println(8858);}}
    public static class Impl8859 implements DemoService {public int type() {return 8859;} public void run() {System.out.println(8859);}}
    public static class Impl8860 implements DemoService {public int type() {return 8860;} public void run() {System.out.println(8860);}}
    public static class Impl8861 implements DemoService {public int type() {return 8861;} public void run() {System.out.println(8861);}}
    public static class Impl8862 implements DemoService {public int type() {return 8862;} public void run() {System.out.println(8862);}}
    public static class Impl8863 implements DemoService {public int type() {return 8863;} public void run() {System.out.println(8863);}}
    public static class Impl8864 implements DemoService {public int type() {return 8864;} public void run() {System.out.println(8864);}}
    public static class Impl8865 implements DemoService {public int type() {return 8865;} public void run() {System.out.println(8865);}}
    public static class Impl8866 implements DemoService {public int type() {return 8866;} public void run() {System.out.println(8866);}}
    public static class Impl8867 implements DemoService {public int type() {return 8867;} public void run() {System.out.println(8867);}}
    public static class Impl8868 implements DemoService {public int type() {return 8868;} public void run() {System.out.println(8868);}}
    public static class Impl8869 implements DemoService {public int type() {return 8869;} public void run() {System.out.println(8869);}}
    public static class Impl8870 implements DemoService {public int type() {return 8870;} public void run() {System.out.println(8870);}}
    public static class Impl8871 implements DemoService {public int type() {return 8871;} public void run() {System.out.println(8871);}}
    public static class Impl8872 implements DemoService {public int type() {return 8872;} public void run() {System.out.println(8872);}}
    public static class Impl8873 implements DemoService {public int type() {return 8873;} public void run() {System.out.println(8873);}}
    public static class Impl8874 implements DemoService {public int type() {return 8874;} public void run() {System.out.println(8874);}}
    public static class Impl8875 implements DemoService {public int type() {return 8875;} public void run() {System.out.println(8875);}}
    public static class Impl8876 implements DemoService {public int type() {return 8876;} public void run() {System.out.println(8876);}}
    public static class Impl8877 implements DemoService {public int type() {return 8877;} public void run() {System.out.println(8877);}}
    public static class Impl8878 implements DemoService {public int type() {return 8878;} public void run() {System.out.println(8878);}}
    public static class Impl8879 implements DemoService {public int type() {return 8879;} public void run() {System.out.println(8879);}}
    public static class Impl8880 implements DemoService {public int type() {return 8880;} public void run() {System.out.println(8880);}}
    public static class Impl8881 implements DemoService {public int type() {return 8881;} public void run() {System.out.println(8881);}}
    public static class Impl8882 implements DemoService {public int type() {return 8882;} public void run() {System.out.println(8882);}}
    public static class Impl8883 implements DemoService {public int type() {return 8883;} public void run() {System.out.println(8883);}}
    public static class Impl8884 implements DemoService {public int type() {return 8884;} public void run() {System.out.println(8884);}}
    public static class Impl8885 implements DemoService {public int type() {return 8885;} public void run() {System.out.println(8885);}}
    public static class Impl8886 implements DemoService {public int type() {return 8886;} public void run() {System.out.println(8886);}}
    public static class Impl8887 implements DemoService {public int type() {return 8887;} public void run() {System.out.println(8887);}}
    public static class Impl8888 implements DemoService {public int type() {return 8888;} public void run() {System.out.println(8888);}}
    public static class Impl8889 implements DemoService {public int type() {return 8889;} public void run() {System.out.println(8889);}}
    public static class Impl8890 implements DemoService {public int type() {return 8890;} public void run() {System.out.println(8890);}}
    public static class Impl8891 implements DemoService {public int type() {return 8891;} public void run() {System.out.println(8891);}}
    public static class Impl8892 implements DemoService {public int type() {return 8892;} public void run() {System.out.println(8892);}}
    public static class Impl8893 implements DemoService {public int type() {return 8893;} public void run() {System.out.println(8893);}}
    public static class Impl8894 implements DemoService {public int type() {return 8894;} public void run() {System.out.println(8894);}}
    public static class Impl8895 implements DemoService {public int type() {return 8895;} public void run() {System.out.println(8895);}}
    public static class Impl8896 implements DemoService {public int type() {return 8896;} public void run() {System.out.println(8896);}}
    public static class Impl8897 implements DemoService {public int type() {return 8897;} public void run() {System.out.println(8897);}}
    public static class Impl8898 implements DemoService {public int type() {return 8898;} public void run() {System.out.println(8898);}}
    public static class Impl8899 implements DemoService {public int type() {return 8899;} public void run() {System.out.println(8899);}}
    public static class Impl8900 implements DemoService {public int type() {return 8900;} public void run() {System.out.println(8900);}}
    public static class Impl8901 implements DemoService {public int type() {return 8901;} public void run() {System.out.println(8901);}}
    public static class Impl8902 implements DemoService {public int type() {return 8902;} public void run() {System.out.println(8902);}}
    public static class Impl8903 implements DemoService {public int type() {return 8903;} public void run() {System.out.println(8903);}}
    public static class Impl8904 implements DemoService {public int type() {return 8904;} public void run() {System.out.println(8904);}}
    public static class Impl8905 implements DemoService {public int type() {return 8905;} public void run() {System.out.println(8905);}}
    public static class Impl8906 implements DemoService {public int type() {return 8906;} public void run() {System.out.println(8906);}}
    public static class Impl8907 implements DemoService {public int type() {return 8907;} public void run() {System.out.println(8907);}}
    public static class Impl8908 implements DemoService {public int type() {return 8908;} public void run() {System.out.println(8908);}}
    public static class Impl8909 implements DemoService {public int type() {return 8909;} public void run() {System.out.println(8909);}}
    public static class Impl8910 implements DemoService {public int type() {return 8910;} public void run() {System.out.println(8910);}}
    public static class Impl8911 implements DemoService {public int type() {return 8911;} public void run() {System.out.println(8911);}}
    public static class Impl8912 implements DemoService {public int type() {return 8912;} public void run() {System.out.println(8912);}}
    public static class Impl8913 implements DemoService {public int type() {return 8913;} public void run() {System.out.println(8913);}}
    public static class Impl8914 implements DemoService {public int type() {return 8914;} public void run() {System.out.println(8914);}}
    public static class Impl8915 implements DemoService {public int type() {return 8915;} public void run() {System.out.println(8915);}}
    public static class Impl8916 implements DemoService {public int type() {return 8916;} public void run() {System.out.println(8916);}}
    public static class Impl8917 implements DemoService {public int type() {return 8917;} public void run() {System.out.println(8917);}}
    public static class Impl8918 implements DemoService {public int type() {return 8918;} public void run() {System.out.println(8918);}}
    public static class Impl8919 implements DemoService {public int type() {return 8919;} public void run() {System.out.println(8919);}}
    public static class Impl8920 implements DemoService {public int type() {return 8920;} public void run() {System.out.println(8920);}}
    public static class Impl8921 implements DemoService {public int type() {return 8921;} public void run() {System.out.println(8921);}}
    public static class Impl8922 implements DemoService {public int type() {return 8922;} public void run() {System.out.println(8922);}}
    public static class Impl8923 implements DemoService {public int type() {return 8923;} public void run() {System.out.println(8923);}}
    public static class Impl8924 implements DemoService {public int type() {return 8924;} public void run() {System.out.println(8924);}}
    public static class Impl8925 implements DemoService {public int type() {return 8925;} public void run() {System.out.println(8925);}}
    public static class Impl8926 implements DemoService {public int type() {return 8926;} public void run() {System.out.println(8926);}}
    public static class Impl8927 implements DemoService {public int type() {return 8927;} public void run() {System.out.println(8927);}}
    public static class Impl8928 implements DemoService {public int type() {return 8928;} public void run() {System.out.println(8928);}}
    public static class Impl8929 implements DemoService {public int type() {return 8929;} public void run() {System.out.println(8929);}}
    public static class Impl8930 implements DemoService {public int type() {return 8930;} public void run() {System.out.println(8930);}}
    public static class Impl8931 implements DemoService {public int type() {return 8931;} public void run() {System.out.println(8931);}}
    public static class Impl8932 implements DemoService {public int type() {return 8932;} public void run() {System.out.println(8932);}}
    public static class Impl8933 implements DemoService {public int type() {return 8933;} public void run() {System.out.println(8933);}}
    public static class Impl8934 implements DemoService {public int type() {return 8934;} public void run() {System.out.println(8934);}}
    public static class Impl8935 implements DemoService {public int type() {return 8935;} public void run() {System.out.println(8935);}}
    public static class Impl8936 implements DemoService {public int type() {return 8936;} public void run() {System.out.println(8936);}}
    public static class Impl8937 implements DemoService {public int type() {return 8937;} public void run() {System.out.println(8937);}}
    public static class Impl8938 implements DemoService {public int type() {return 8938;} public void run() {System.out.println(8938);}}
    public static class Impl8939 implements DemoService {public int type() {return 8939;} public void run() {System.out.println(8939);}}
    public static class Impl8940 implements DemoService {public int type() {return 8940;} public void run() {System.out.println(8940);}}
    public static class Impl8941 implements DemoService {public int type() {return 8941;} public void run() {System.out.println(8941);}}
    public static class Impl8942 implements DemoService {public int type() {return 8942;} public void run() {System.out.println(8942);}}
    public static class Impl8943 implements DemoService {public int type() {return 8943;} public void run() {System.out.println(8943);}}
    public static class Impl8944 implements DemoService {public int type() {return 8944;} public void run() {System.out.println(8944);}}
    public static class Impl8945 implements DemoService {public int type() {return 8945;} public void run() {System.out.println(8945);}}
    public static class Impl8946 implements DemoService {public int type() {return 8946;} public void run() {System.out.println(8946);}}
    public static class Impl8947 implements DemoService {public int type() {return 8947;} public void run() {System.out.println(8947);}}
    public static class Impl8948 implements DemoService {public int type() {return 8948;} public void run() {System.out.println(8948);}}
    public static class Impl8949 implements DemoService {public int type() {return 8949;} public void run() {System.out.println(8949);}}
    public static class Impl8950 implements DemoService {public int type() {return 8950;} public void run() {System.out.println(8950);}}
    public static class Impl8951 implements DemoService {public int type() {return 8951;} public void run() {System.out.println(8951);}}
    public static class Impl8952 implements DemoService {public int type() {return 8952;} public void run() {System.out.println(8952);}}
    public static class Impl8953 implements DemoService {public int type() {return 8953;} public void run() {System.out.println(8953);}}
    public static class Impl8954 implements DemoService {public int type() {return 8954;} public void run() {System.out.println(8954);}}
    public static class Impl8955 implements DemoService {public int type() {return 8955;} public void run() {System.out.println(8955);}}
    public static class Impl8956 implements DemoService {public int type() {return 8956;} public void run() {System.out.println(8956);}}
    public static class Impl8957 implements DemoService {public int type() {return 8957;} public void run() {System.out.println(8957);}}
    public static class Impl8958 implements DemoService {public int type() {return 8958;} public void run() {System.out.println(8958);}}
    public static class Impl8959 implements DemoService {public int type() {return 8959;} public void run() {System.out.println(8959);}}
    public static class Impl8960 implements DemoService {public int type() {return 8960;} public void run() {System.out.println(8960);}}
    public static class Impl8961 implements DemoService {public int type() {return 8961;} public void run() {System.out.println(8961);}}
    public static class Impl8962 implements DemoService {public int type() {return 8962;} public void run() {System.out.println(8962);}}
    public static class Impl8963 implements DemoService {public int type() {return 8963;} public void run() {System.out.println(8963);}}
    public static class Impl8964 implements DemoService {public int type() {return 8964;} public void run() {System.out.println(8964);}}
    public static class Impl8965 implements DemoService {public int type() {return 8965;} public void run() {System.out.println(8965);}}
    public static class Impl8966 implements DemoService {public int type() {return 8966;} public void run() {System.out.println(8966);}}
    public static class Impl8967 implements DemoService {public int type() {return 8967;} public void run() {System.out.println(8967);}}
    public static class Impl8968 implements DemoService {public int type() {return 8968;} public void run() {System.out.println(8968);}}
    public static class Impl8969 implements DemoService {public int type() {return 8969;} public void run() {System.out.println(8969);}}
    public static class Impl8970 implements DemoService {public int type() {return 8970;} public void run() {System.out.println(8970);}}
    public static class Impl8971 implements DemoService {public int type() {return 8971;} public void run() {System.out.println(8971);}}
    public static class Impl8972 implements DemoService {public int type() {return 8972;} public void run() {System.out.println(8972);}}
    public static class Impl8973 implements DemoService {public int type() {return 8973;} public void run() {System.out.println(8973);}}
    public static class Impl8974 implements DemoService {public int type() {return 8974;} public void run() {System.out.println(8974);}}
    public static class Impl8975 implements DemoService {public int type() {return 8975;} public void run() {System.out.println(8975);}}
    public static class Impl8976 implements DemoService {public int type() {return 8976;} public void run() {System.out.println(8976);}}
    public static class Impl8977 implements DemoService {public int type() {return 8977;} public void run() {System.out.println(8977);}}
    public static class Impl8978 implements DemoService {public int type() {return 8978;} public void run() {System.out.println(8978);}}
    public static class Impl8979 implements DemoService {public int type() {return 8979;} public void run() {System.out.println(8979);}}
    public static class Impl8980 implements DemoService {public int type() {return 8980;} public void run() {System.out.println(8980);}}
    public static class Impl8981 implements DemoService {public int type() {return 8981;} public void run() {System.out.println(8981);}}
    public static class Impl8982 implements DemoService {public int type() {return 8982;} public void run() {System.out.println(8982);}}
    public static class Impl8983 implements DemoService {public int type() {return 8983;} public void run() {System.out.println(8983);}}
    public static class Impl8984 implements DemoService {public int type() {return 8984;} public void run() {System.out.println(8984);}}
    public static class Impl8985 implements DemoService {public int type() {return 8985;} public void run() {System.out.println(8985);}}
    public static class Impl8986 implements DemoService {public int type() {return 8986;} public void run() {System.out.println(8986);}}
    public static class Impl8987 implements DemoService {public int type() {return 8987;} public void run() {System.out.println(8987);}}
    public static class Impl8988 implements DemoService {public int type() {return 8988;} public void run() {System.out.println(8988);}}
    public static class Impl8989 implements DemoService {public int type() {return 8989;} public void run() {System.out.println(8989);}}
    public static class Impl8990 implements DemoService {public int type() {return 8990;} public void run() {System.out.println(8990);}}
    public static class Impl8991 implements DemoService {public int type() {return 8991;} public void run() {System.out.println(8991);}}
    public static class Impl8992 implements DemoService {public int type() {return 8992;} public void run() {System.out.println(8992);}}
    public static class Impl8993 implements DemoService {public int type() {return 8993;} public void run() {System.out.println(8993);}}
    public static class Impl8994 implements DemoService {public int type() {return 8994;} public void run() {System.out.println(8994);}}
    public static class Impl8995 implements DemoService {public int type() {return 8995;} public void run() {System.out.println(8995);}}
    public static class Impl8996 implements DemoService {public int type() {return 8996;} public void run() {System.out.println(8996);}}
    public static class Impl8997 implements DemoService {public int type() {return 8997;} public void run() {System.out.println(8997);}}
    public static class Impl8998 implements DemoService {public int type() {return 8998;} public void run() {System.out.println(8998);}}
    public static class Impl8999 implements DemoService {public int type() {return 8999;} public void run() {System.out.println(8999);}}
    public static class Impl9000 implements DemoService {public int type() {return 9000;} public void run() {System.out.println(9000);}}
    public static class Impl9001 implements DemoService {public int type() {return 9001;} public void run() {System.out.println(9001);}}
    public static class Impl9002 implements DemoService {public int type() {return 9002;} public void run() {System.out.println(9002);}}
    public static class Impl9003 implements DemoService {public int type() {return 9003;} public void run() {System.out.println(9003);}}
    public static class Impl9004 implements DemoService {public int type() {return 9004;} public void run() {System.out.println(9004);}}
    public static class Impl9005 implements DemoService {public int type() {return 9005;} public void run() {System.out.println(9005);}}
    public static class Impl9006 implements DemoService {public int type() {return 9006;} public void run() {System.out.println(9006);}}
    public static class Impl9007 implements DemoService {public int type() {return 9007;} public void run() {System.out.println(9007);}}
    public static class Impl9008 implements DemoService {public int type() {return 9008;} public void run() {System.out.println(9008);}}
    public static class Impl9009 implements DemoService {public int type() {return 9009;} public void run() {System.out.println(9009);}}
    public static class Impl9010 implements DemoService {public int type() {return 9010;} public void run() {System.out.println(9010);}}
    public static class Impl9011 implements DemoService {public int type() {return 9011;} public void run() {System.out.println(9011);}}
    public static class Impl9012 implements DemoService {public int type() {return 9012;} public void run() {System.out.println(9012);}}
    public static class Impl9013 implements DemoService {public int type() {return 9013;} public void run() {System.out.println(9013);}}
    public static class Impl9014 implements DemoService {public int type() {return 9014;} public void run() {System.out.println(9014);}}
    public static class Impl9015 implements DemoService {public int type() {return 9015;} public void run() {System.out.println(9015);}}
    public static class Impl9016 implements DemoService {public int type() {return 9016;} public void run() {System.out.println(9016);}}
    public static class Impl9017 implements DemoService {public int type() {return 9017;} public void run() {System.out.println(9017);}}
    public static class Impl9018 implements DemoService {public int type() {return 9018;} public void run() {System.out.println(9018);}}
    public static class Impl9019 implements DemoService {public int type() {return 9019;} public void run() {System.out.println(9019);}}
    public static class Impl9020 implements DemoService {public int type() {return 9020;} public void run() {System.out.println(9020);}}
    public static class Impl9021 implements DemoService {public int type() {return 9021;} public void run() {System.out.println(9021);}}
    public static class Impl9022 implements DemoService {public int type() {return 9022;} public void run() {System.out.println(9022);}}
    public static class Impl9023 implements DemoService {public int type() {return 9023;} public void run() {System.out.println(9023);}}
    public static class Impl9024 implements DemoService {public int type() {return 9024;} public void run() {System.out.println(9024);}}
    public static class Impl9025 implements DemoService {public int type() {return 9025;} public void run() {System.out.println(9025);}}
    public static class Impl9026 implements DemoService {public int type() {return 9026;} public void run() {System.out.println(9026);}}
    public static class Impl9027 implements DemoService {public int type() {return 9027;} public void run() {System.out.println(9027);}}
    public static class Impl9028 implements DemoService {public int type() {return 9028;} public void run() {System.out.println(9028);}}
    public static class Impl9029 implements DemoService {public int type() {return 9029;} public void run() {System.out.println(9029);}}
    public static class Impl9030 implements DemoService {public int type() {return 9030;} public void run() {System.out.println(9030);}}
    public static class Impl9031 implements DemoService {public int type() {return 9031;} public void run() {System.out.println(9031);}}
    public static class Impl9032 implements DemoService {public int type() {return 9032;} public void run() {System.out.println(9032);}}
    public static class Impl9033 implements DemoService {public int type() {return 9033;} public void run() {System.out.println(9033);}}
    public static class Impl9034 implements DemoService {public int type() {return 9034;} public void run() {System.out.println(9034);}}
    public static class Impl9035 implements DemoService {public int type() {return 9035;} public void run() {System.out.println(9035);}}
    public static class Impl9036 implements DemoService {public int type() {return 9036;} public void run() {System.out.println(9036);}}
    public static class Impl9037 implements DemoService {public int type() {return 9037;} public void run() {System.out.println(9037);}}
    public static class Impl9038 implements DemoService {public int type() {return 9038;} public void run() {System.out.println(9038);}}
    public static class Impl9039 implements DemoService {public int type() {return 9039;} public void run() {System.out.println(9039);}}
    public static class Impl9040 implements DemoService {public int type() {return 9040;} public void run() {System.out.println(9040);}}
    public static class Impl9041 implements DemoService {public int type() {return 9041;} public void run() {System.out.println(9041);}}
    public static class Impl9042 implements DemoService {public int type() {return 9042;} public void run() {System.out.println(9042);}}
    public static class Impl9043 implements DemoService {public int type() {return 9043;} public void run() {System.out.println(9043);}}
    public static class Impl9044 implements DemoService {public int type() {return 9044;} public void run() {System.out.println(9044);}}
    public static class Impl9045 implements DemoService {public int type() {return 9045;} public void run() {System.out.println(9045);}}
    public static class Impl9046 implements DemoService {public int type() {return 9046;} public void run() {System.out.println(9046);}}
    public static class Impl9047 implements DemoService {public int type() {return 9047;} public void run() {System.out.println(9047);}}
    public static class Impl9048 implements DemoService {public int type() {return 9048;} public void run() {System.out.println(9048);}}
    public static class Impl9049 implements DemoService {public int type() {return 9049;} public void run() {System.out.println(9049);}}
    public static class Impl9050 implements DemoService {public int type() {return 9050;} public void run() {System.out.println(9050);}}
    public static class Impl9051 implements DemoService {public int type() {return 9051;} public void run() {System.out.println(9051);}}
    public static class Impl9052 implements DemoService {public int type() {return 9052;} public void run() {System.out.println(9052);}}
    public static class Impl9053 implements DemoService {public int type() {return 9053;} public void run() {System.out.println(9053);}}
    public static class Impl9054 implements DemoService {public int type() {return 9054;} public void run() {System.out.println(9054);}}
    public static class Impl9055 implements DemoService {public int type() {return 9055;} public void run() {System.out.println(9055);}}
    public static class Impl9056 implements DemoService {public int type() {return 9056;} public void run() {System.out.println(9056);}}
    public static class Impl9057 implements DemoService {public int type() {return 9057;} public void run() {System.out.println(9057);}}
    public static class Impl9058 implements DemoService {public int type() {return 9058;} public void run() {System.out.println(9058);}}
    public static class Impl9059 implements DemoService {public int type() {return 9059;} public void run() {System.out.println(9059);}}
    public static class Impl9060 implements DemoService {public int type() {return 9060;} public void run() {System.out.println(9060);}}
    public static class Impl9061 implements DemoService {public int type() {return 9061;} public void run() {System.out.println(9061);}}
    public static class Impl9062 implements DemoService {public int type() {return 9062;} public void run() {System.out.println(9062);}}
    public static class Impl9063 implements DemoService {public int type() {return 9063;} public void run() {System.out.println(9063);}}
    public static class Impl9064 implements DemoService {public int type() {return 9064;} public void run() {System.out.println(9064);}}
    public static class Impl9065 implements DemoService {public int type() {return 9065;} public void run() {System.out.println(9065);}}
    public static class Impl9066 implements DemoService {public int type() {return 9066;} public void run() {System.out.println(9066);}}
    public static class Impl9067 implements DemoService {public int type() {return 9067;} public void run() {System.out.println(9067);}}
    public static class Impl9068 implements DemoService {public int type() {return 9068;} public void run() {System.out.println(9068);}}
    public static class Impl9069 implements DemoService {public int type() {return 9069;} public void run() {System.out.println(9069);}}
    public static class Impl9070 implements DemoService {public int type() {return 9070;} public void run() {System.out.println(9070);}}
    public static class Impl9071 implements DemoService {public int type() {return 9071;} public void run() {System.out.println(9071);}}
    public static class Impl9072 implements DemoService {public int type() {return 9072;} public void run() {System.out.println(9072);}}
    public static class Impl9073 implements DemoService {public int type() {return 9073;} public void run() {System.out.println(9073);}}
    public static class Impl9074 implements DemoService {public int type() {return 9074;} public void run() {System.out.println(9074);}}
    public static class Impl9075 implements DemoService {public int type() {return 9075;} public void run() {System.out.println(9075);}}
    public static class Impl9076 implements DemoService {public int type() {return 9076;} public void run() {System.out.println(9076);}}
    public static class Impl9077 implements DemoService {public int type() {return 9077;} public void run() {System.out.println(9077);}}
    public static class Impl9078 implements DemoService {public int type() {return 9078;} public void run() {System.out.println(9078);}}
    public static class Impl9079 implements DemoService {public int type() {return 9079;} public void run() {System.out.println(9079);}}
    public static class Impl9080 implements DemoService {public int type() {return 9080;} public void run() {System.out.println(9080);}}
    public static class Impl9081 implements DemoService {public int type() {return 9081;} public void run() {System.out.println(9081);}}
    public static class Impl9082 implements DemoService {public int type() {return 9082;} public void run() {System.out.println(9082);}}
    public static class Impl9083 implements DemoService {public int type() {return 9083;} public void run() {System.out.println(9083);}}
    public static class Impl9084 implements DemoService {public int type() {return 9084;} public void run() {System.out.println(9084);}}
    public static class Impl9085 implements DemoService {public int type() {return 9085;} public void run() {System.out.println(9085);}}
    public static class Impl9086 implements DemoService {public int type() {return 9086;} public void run() {System.out.println(9086);}}
    public static class Impl9087 implements DemoService {public int type() {return 9087;} public void run() {System.out.println(9087);}}
    public static class Impl9088 implements DemoService {public int type() {return 9088;} public void run() {System.out.println(9088);}}
    public static class Impl9089 implements DemoService {public int type() {return 9089;} public void run() {System.out.println(9089);}}
    public static class Impl9090 implements DemoService {public int type() {return 9090;} public void run() {System.out.println(9090);}}
    public static class Impl9091 implements DemoService {public int type() {return 9091;} public void run() {System.out.println(9091);}}
    public static class Impl9092 implements DemoService {public int type() {return 9092;} public void run() {System.out.println(9092);}}
    public static class Impl9093 implements DemoService {public int type() {return 9093;} public void run() {System.out.println(9093);}}
    public static class Impl9094 implements DemoService {public int type() {return 9094;} public void run() {System.out.println(9094);}}
    public static class Impl9095 implements DemoService {public int type() {return 9095;} public void run() {System.out.println(9095);}}
    public static class Impl9096 implements DemoService {public int type() {return 9096;} public void run() {System.out.println(9096);}}
    public static class Impl9097 implements DemoService {public int type() {return 9097;} public void run() {System.out.println(9097);}}
    public static class Impl9098 implements DemoService {public int type() {return 9098;} public void run() {System.out.println(9098);}}
    public static class Impl9099 implements DemoService {public int type() {return 9099;} public void run() {System.out.println(9099);}}
    public static class Impl9100 implements DemoService {public int type() {return 9100;} public void run() {System.out.println(9100);}}
    public static class Impl9101 implements DemoService {public int type() {return 9101;} public void run() {System.out.println(9101);}}
    public static class Impl9102 implements DemoService {public int type() {return 9102;} public void run() {System.out.println(9102);}}
    public static class Impl9103 implements DemoService {public int type() {return 9103;} public void run() {System.out.println(9103);}}
    public static class Impl9104 implements DemoService {public int type() {return 9104;} public void run() {System.out.println(9104);}}
    public static class Impl9105 implements DemoService {public int type() {return 9105;} public void run() {System.out.println(9105);}}
    public static class Impl9106 implements DemoService {public int type() {return 9106;} public void run() {System.out.println(9106);}}
    public static class Impl9107 implements DemoService {public int type() {return 9107;} public void run() {System.out.println(9107);}}
    public static class Impl9108 implements DemoService {public int type() {return 9108;} public void run() {System.out.println(9108);}}
    public static class Impl9109 implements DemoService {public int type() {return 9109;} public void run() {System.out.println(9109);}}
    public static class Impl9110 implements DemoService {public int type() {return 9110;} public void run() {System.out.println(9110);}}
    public static class Impl9111 implements DemoService {public int type() {return 9111;} public void run() {System.out.println(9111);}}
    public static class Impl9112 implements DemoService {public int type() {return 9112;} public void run() {System.out.println(9112);}}
    public static class Impl9113 implements DemoService {public int type() {return 9113;} public void run() {System.out.println(9113);}}
    public static class Impl9114 implements DemoService {public int type() {return 9114;} public void run() {System.out.println(9114);}}
    public static class Impl9115 implements DemoService {public int type() {return 9115;} public void run() {System.out.println(9115);}}
    public static class Impl9116 implements DemoService {public int type() {return 9116;} public void run() {System.out.println(9116);}}
    public static class Impl9117 implements DemoService {public int type() {return 9117;} public void run() {System.out.println(9117);}}
    public static class Impl9118 implements DemoService {public int type() {return 9118;} public void run() {System.out.println(9118);}}
    public static class Impl9119 implements DemoService {public int type() {return 9119;} public void run() {System.out.println(9119);}}
    public static class Impl9120 implements DemoService {public int type() {return 9120;} public void run() {System.out.println(9120);}}
    public static class Impl9121 implements DemoService {public int type() {return 9121;} public void run() {System.out.println(9121);}}
    public static class Impl9122 implements DemoService {public int type() {return 9122;} public void run() {System.out.println(9122);}}
    public static class Impl9123 implements DemoService {public int type() {return 9123;} public void run() {System.out.println(9123);}}
    public static class Impl9124 implements DemoService {public int type() {return 9124;} public void run() {System.out.println(9124);}}
    public static class Impl9125 implements DemoService {public int type() {return 9125;} public void run() {System.out.println(9125);}}
    public static class Impl9126 implements DemoService {public int type() {return 9126;} public void run() {System.out.println(9126);}}
    public static class Impl9127 implements DemoService {public int type() {return 9127;} public void run() {System.out.println(9127);}}
    public static class Impl9128 implements DemoService {public int type() {return 9128;} public void run() {System.out.println(9128);}}
    public static class Impl9129 implements DemoService {public int type() {return 9129;} public void run() {System.out.println(9129);}}
    public static class Impl9130 implements DemoService {public int type() {return 9130;} public void run() {System.out.println(9130);}}
    public static class Impl9131 implements DemoService {public int type() {return 9131;} public void run() {System.out.println(9131);}}
    public static class Impl9132 implements DemoService {public int type() {return 9132;} public void run() {System.out.println(9132);}}
    public static class Impl9133 implements DemoService {public int type() {return 9133;} public void run() {System.out.println(9133);}}
    public static class Impl9134 implements DemoService {public int type() {return 9134;} public void run() {System.out.println(9134);}}
    public static class Impl9135 implements DemoService {public int type() {return 9135;} public void run() {System.out.println(9135);}}
    public static class Impl9136 implements DemoService {public int type() {return 9136;} public void run() {System.out.println(9136);}}
    public static class Impl9137 implements DemoService {public int type() {return 9137;} public void run() {System.out.println(9137);}}
    public static class Impl9138 implements DemoService {public int type() {return 9138;} public void run() {System.out.println(9138);}}
    public static class Impl9139 implements DemoService {public int type() {return 9139;} public void run() {System.out.println(9139);}}
    public static class Impl9140 implements DemoService {public int type() {return 9140;} public void run() {System.out.println(9140);}}
    public static class Impl9141 implements DemoService {public int type() {return 9141;} public void run() {System.out.println(9141);}}
    public static class Impl9142 implements DemoService {public int type() {return 9142;} public void run() {System.out.println(9142);}}
    public static class Impl9143 implements DemoService {public int type() {return 9143;} public void run() {System.out.println(9143);}}
    public static class Impl9144 implements DemoService {public int type() {return 9144;} public void run() {System.out.println(9144);}}
    public static class Impl9145 implements DemoService {public int type() {return 9145;} public void run() {System.out.println(9145);}}
    public static class Impl9146 implements DemoService {public int type() {return 9146;} public void run() {System.out.println(9146);}}
    public static class Impl9147 implements DemoService {public int type() {return 9147;} public void run() {System.out.println(9147);}}
    public static class Impl9148 implements DemoService {public int type() {return 9148;} public void run() {System.out.println(9148);}}
    public static class Impl9149 implements DemoService {public int type() {return 9149;} public void run() {System.out.println(9149);}}
    public static class Impl9150 implements DemoService {public int type() {return 9150;} public void run() {System.out.println(9150);}}
    public static class Impl9151 implements DemoService {public int type() {return 9151;} public void run() {System.out.println(9151);}}
    public static class Impl9152 implements DemoService {public int type() {return 9152;} public void run() {System.out.println(9152);}}
    public static class Impl9153 implements DemoService {public int type() {return 9153;} public void run() {System.out.println(9153);}}
    public static class Impl9154 implements DemoService {public int type() {return 9154;} public void run() {System.out.println(9154);}}
    public static class Impl9155 implements DemoService {public int type() {return 9155;} public void run() {System.out.println(9155);}}
    public static class Impl9156 implements DemoService {public int type() {return 9156;} public void run() {System.out.println(9156);}}
    public static class Impl9157 implements DemoService {public int type() {return 9157;} public void run() {System.out.println(9157);}}
    public static class Impl9158 implements DemoService {public int type() {return 9158;} public void run() {System.out.println(9158);}}
    public static class Impl9159 implements DemoService {public int type() {return 9159;} public void run() {System.out.println(9159);}}
    public static class Impl9160 implements DemoService {public int type() {return 9160;} public void run() {System.out.println(9160);}}
    public static class Impl9161 implements DemoService {public int type() {return 9161;} public void run() {System.out.println(9161);}}
    public static class Impl9162 implements DemoService {public int type() {return 9162;} public void run() {System.out.println(9162);}}
    public static class Impl9163 implements DemoService {public int type() {return 9163;} public void run() {System.out.println(9163);}}
    public static class Impl9164 implements DemoService {public int type() {return 9164;} public void run() {System.out.println(9164);}}
    public static class Impl9165 implements DemoService {public int type() {return 9165;} public void run() {System.out.println(9165);}}
    public static class Impl9166 implements DemoService {public int type() {return 9166;} public void run() {System.out.println(9166);}}
    public static class Impl9167 implements DemoService {public int type() {return 9167;} public void run() {System.out.println(9167);}}
    public static class Impl9168 implements DemoService {public int type() {return 9168;} public void run() {System.out.println(9168);}}
    public static class Impl9169 implements DemoService {public int type() {return 9169;} public void run() {System.out.println(9169);}}
    public static class Impl9170 implements DemoService {public int type() {return 9170;} public void run() {System.out.println(9170);}}
    public static class Impl9171 implements DemoService {public int type() {return 9171;} public void run() {System.out.println(9171);}}
    public static class Impl9172 implements DemoService {public int type() {return 9172;} public void run() {System.out.println(9172);}}
    public static class Impl9173 implements DemoService {public int type() {return 9173;} public void run() {System.out.println(9173);}}
    public static class Impl9174 implements DemoService {public int type() {return 9174;} public void run() {System.out.println(9174);}}
    public static class Impl9175 implements DemoService {public int type() {return 9175;} public void run() {System.out.println(9175);}}
    public static class Impl9176 implements DemoService {public int type() {return 9176;} public void run() {System.out.println(9176);}}
    public static class Impl9177 implements DemoService {public int type() {return 9177;} public void run() {System.out.println(9177);}}
    public static class Impl9178 implements DemoService {public int type() {return 9178;} public void run() {System.out.println(9178);}}
    public static class Impl9179 implements DemoService {public int type() {return 9179;} public void run() {System.out.println(9179);}}
    public static class Impl9180 implements DemoService {public int type() {return 9180;} public void run() {System.out.println(9180);}}
    public static class Impl9181 implements DemoService {public int type() {return 9181;} public void run() {System.out.println(9181);}}
    public static class Impl9182 implements DemoService {public int type() {return 9182;} public void run() {System.out.println(9182);}}
    public static class Impl9183 implements DemoService {public int type() {return 9183;} public void run() {System.out.println(9183);}}
    public static class Impl9184 implements DemoService {public int type() {return 9184;} public void run() {System.out.println(9184);}}
    public static class Impl9185 implements DemoService {public int type() {return 9185;} public void run() {System.out.println(9185);}}
    public static class Impl9186 implements DemoService {public int type() {return 9186;} public void run() {System.out.println(9186);}}
    public static class Impl9187 implements DemoService {public int type() {return 9187;} public void run() {System.out.println(9187);}}
    public static class Impl9188 implements DemoService {public int type() {return 9188;} public void run() {System.out.println(9188);}}
    public static class Impl9189 implements DemoService {public int type() {return 9189;} public void run() {System.out.println(9189);}}
    public static class Impl9190 implements DemoService {public int type() {return 9190;} public void run() {System.out.println(9190);}}
    public static class Impl9191 implements DemoService {public int type() {return 9191;} public void run() {System.out.println(9191);}}
    public static class Impl9192 implements DemoService {public int type() {return 9192;} public void run() {System.out.println(9192);}}
    public static class Impl9193 implements DemoService {public int type() {return 9193;} public void run() {System.out.println(9193);}}
    public static class Impl9194 implements DemoService {public int type() {return 9194;} public void run() {System.out.println(9194);}}
    public static class Impl9195 implements DemoService {public int type() {return 9195;} public void run() {System.out.println(9195);}}
    public static class Impl9196 implements DemoService {public int type() {return 9196;} public void run() {System.out.println(9196);}}
    public static class Impl9197 implements DemoService {public int type() {return 9197;} public void run() {System.out.println(9197);}}
    public static class Impl9198 implements DemoService {public int type() {return 9198;} public void run() {System.out.println(9198);}}
    public static class Impl9199 implements DemoService {public int type() {return 9199;} public void run() {System.out.println(9199);}}
    public static class Impl9200 implements DemoService {public int type() {return 9200;} public void run() {System.out.println(9200);}}
    public static class Impl9201 implements DemoService {public int type() {return 9201;} public void run() {System.out.println(9201);}}
    public static class Impl9202 implements DemoService {public int type() {return 9202;} public void run() {System.out.println(9202);}}
    public static class Impl9203 implements DemoService {public int type() {return 9203;} public void run() {System.out.println(9203);}}
    public static class Impl9204 implements DemoService {public int type() {return 9204;} public void run() {System.out.println(9204);}}
    public static class Impl9205 implements DemoService {public int type() {return 9205;} public void run() {System.out.println(9205);}}
    public static class Impl9206 implements DemoService {public int type() {return 9206;} public void run() {System.out.println(9206);}}
    public static class Impl9207 implements DemoService {public int type() {return 9207;} public void run() {System.out.println(9207);}}
    public static class Impl9208 implements DemoService {public int type() {return 9208;} public void run() {System.out.println(9208);}}
    public static class Impl9209 implements DemoService {public int type() {return 9209;} public void run() {System.out.println(9209);}}
    public static class Impl9210 implements DemoService {public int type() {return 9210;} public void run() {System.out.println(9210);}}
    public static class Impl9211 implements DemoService {public int type() {return 9211;} public void run() {System.out.println(9211);}}
    public static class Impl9212 implements DemoService {public int type() {return 9212;} public void run() {System.out.println(9212);}}
    public static class Impl9213 implements DemoService {public int type() {return 9213;} public void run() {System.out.println(9213);}}
    public static class Impl9214 implements DemoService {public int type() {return 9214;} public void run() {System.out.println(9214);}}
    public static class Impl9215 implements DemoService {public int type() {return 9215;} public void run() {System.out.println(9215);}}
    public static class Impl9216 implements DemoService {public int type() {return 9216;} public void run() {System.out.println(9216);}}
    public static class Impl9217 implements DemoService {public int type() {return 9217;} public void run() {System.out.println(9217);}}
    public static class Impl9218 implements DemoService {public int type() {return 9218;} public void run() {System.out.println(9218);}}
    public static class Impl9219 implements DemoService {public int type() {return 9219;} public void run() {System.out.println(9219);}}
    public static class Impl9220 implements DemoService {public int type() {return 9220;} public void run() {System.out.println(9220);}}
    public static class Impl9221 implements DemoService {public int type() {return 9221;} public void run() {System.out.println(9221);}}
    public static class Impl9222 implements DemoService {public int type() {return 9222;} public void run() {System.out.println(9222);}}
    public static class Impl9223 implements DemoService {public int type() {return 9223;} public void run() {System.out.println(9223);}}
    public static class Impl9224 implements DemoService {public int type() {return 9224;} public void run() {System.out.println(9224);}}
    public static class Impl9225 implements DemoService {public int type() {return 9225;} public void run() {System.out.println(9225);}}
    public static class Impl9226 implements DemoService {public int type() {return 9226;} public void run() {System.out.println(9226);}}
    public static class Impl9227 implements DemoService {public int type() {return 9227;} public void run() {System.out.println(9227);}}
    public static class Impl9228 implements DemoService {public int type() {return 9228;} public void run() {System.out.println(9228);}}
    public static class Impl9229 implements DemoService {public int type() {return 9229;} public void run() {System.out.println(9229);}}
    public static class Impl9230 implements DemoService {public int type() {return 9230;} public void run() {System.out.println(9230);}}
    public static class Impl9231 implements DemoService {public int type() {return 9231;} public void run() {System.out.println(9231);}}
    public static class Impl9232 implements DemoService {public int type() {return 9232;} public void run() {System.out.println(9232);}}
    public static class Impl9233 implements DemoService {public int type() {return 9233;} public void run() {System.out.println(9233);}}
    public static class Impl9234 implements DemoService {public int type() {return 9234;} public void run() {System.out.println(9234);}}
    public static class Impl9235 implements DemoService {public int type() {return 9235;} public void run() {System.out.println(9235);}}
    public static class Impl9236 implements DemoService {public int type() {return 9236;} public void run() {System.out.println(9236);}}
    public static class Impl9237 implements DemoService {public int type() {return 9237;} public void run() {System.out.println(9237);}}
    public static class Impl9238 implements DemoService {public int type() {return 9238;} public void run() {System.out.println(9238);}}
    public static class Impl9239 implements DemoService {public int type() {return 9239;} public void run() {System.out.println(9239);}}
    public static class Impl9240 implements DemoService {public int type() {return 9240;} public void run() {System.out.println(9240);}}
    public static class Impl9241 implements DemoService {public int type() {return 9241;} public void run() {System.out.println(9241);}}
    public static class Impl9242 implements DemoService {public int type() {return 9242;} public void run() {System.out.println(9242);}}
    public static class Impl9243 implements DemoService {public int type() {return 9243;} public void run() {System.out.println(9243);}}
    public static class Impl9244 implements DemoService {public int type() {return 9244;} public void run() {System.out.println(9244);}}
    public static class Impl9245 implements DemoService {public int type() {return 9245;} public void run() {System.out.println(9245);}}
    public static class Impl9246 implements DemoService {public int type() {return 9246;} public void run() {System.out.println(9246);}}
    public static class Impl9247 implements DemoService {public int type() {return 9247;} public void run() {System.out.println(9247);}}
    public static class Impl9248 implements DemoService {public int type() {return 9248;} public void run() {System.out.println(9248);}}
    public static class Impl9249 implements DemoService {public int type() {return 9249;} public void run() {System.out.println(9249);}}
    public static class Impl9250 implements DemoService {public int type() {return 9250;} public void run() {System.out.println(9250);}}
    public static class Impl9251 implements DemoService {public int type() {return 9251;} public void run() {System.out.println(9251);}}
    public static class Impl9252 implements DemoService {public int type() {return 9252;} public void run() {System.out.println(9252);}}
    public static class Impl9253 implements DemoService {public int type() {return 9253;} public void run() {System.out.println(9253);}}
    public static class Impl9254 implements DemoService {public int type() {return 9254;} public void run() {System.out.println(9254);}}
    public static class Impl9255 implements DemoService {public int type() {return 9255;} public void run() {System.out.println(9255);}}
    public static class Impl9256 implements DemoService {public int type() {return 9256;} public void run() {System.out.println(9256);}}
    public static class Impl9257 implements DemoService {public int type() {return 9257;} public void run() {System.out.println(9257);}}
    public static class Impl9258 implements DemoService {public int type() {return 9258;} public void run() {System.out.println(9258);}}
    public static class Impl9259 implements DemoService {public int type() {return 9259;} public void run() {System.out.println(9259);}}
    public static class Impl9260 implements DemoService {public int type() {return 9260;} public void run() {System.out.println(9260);}}
    public static class Impl9261 implements DemoService {public int type() {return 9261;} public void run() {System.out.println(9261);}}
    public static class Impl9262 implements DemoService {public int type() {return 9262;} public void run() {System.out.println(9262);}}
    public static class Impl9263 implements DemoService {public int type() {return 9263;} public void run() {System.out.println(9263);}}
    public static class Impl9264 implements DemoService {public int type() {return 9264;} public void run() {System.out.println(9264);}}
    public static class Impl9265 implements DemoService {public int type() {return 9265;} public void run() {System.out.println(9265);}}
    public static class Impl9266 implements DemoService {public int type() {return 9266;} public void run() {System.out.println(9266);}}
    public static class Impl9267 implements DemoService {public int type() {return 9267;} public void run() {System.out.println(9267);}}
    public static class Impl9268 implements DemoService {public int type() {return 9268;} public void run() {System.out.println(9268);}}
    public static class Impl9269 implements DemoService {public int type() {return 9269;} public void run() {System.out.println(9269);}}
    public static class Impl9270 implements DemoService {public int type() {return 9270;} public void run() {System.out.println(9270);}}
    public static class Impl9271 implements DemoService {public int type() {return 9271;} public void run() {System.out.println(9271);}}
    public static class Impl9272 implements DemoService {public int type() {return 9272;} public void run() {System.out.println(9272);}}
    public static class Impl9273 implements DemoService {public int type() {return 9273;} public void run() {System.out.println(9273);}}
    public static class Impl9274 implements DemoService {public int type() {return 9274;} public void run() {System.out.println(9274);}}
    public static class Impl9275 implements DemoService {public int type() {return 9275;} public void run() {System.out.println(9275);}}
    public static class Impl9276 implements DemoService {public int type() {return 9276;} public void run() {System.out.println(9276);}}
    public static class Impl9277 implements DemoService {public int type() {return 9277;} public void run() {System.out.println(9277);}}
    public static class Impl9278 implements DemoService {public int type() {return 9278;} public void run() {System.out.println(9278);}}
    public static class Impl9279 implements DemoService {public int type() {return 9279;} public void run() {System.out.println(9279);}}
    public static class Impl9280 implements DemoService {public int type() {return 9280;} public void run() {System.out.println(9280);}}
    public static class Impl9281 implements DemoService {public int type() {return 9281;} public void run() {System.out.println(9281);}}
    public static class Impl9282 implements DemoService {public int type() {return 9282;} public void run() {System.out.println(9282);}}
    public static class Impl9283 implements DemoService {public int type() {return 9283;} public void run() {System.out.println(9283);}}
    public static class Impl9284 implements DemoService {public int type() {return 9284;} public void run() {System.out.println(9284);}}
    public static class Impl9285 implements DemoService {public int type() {return 9285;} public void run() {System.out.println(9285);}}
    public static class Impl9286 implements DemoService {public int type() {return 9286;} public void run() {System.out.println(9286);}}
    public static class Impl9287 implements DemoService {public int type() {return 9287;} public void run() {System.out.println(9287);}}
    public static class Impl9288 implements DemoService {public int type() {return 9288;} public void run() {System.out.println(9288);}}
    public static class Impl9289 implements DemoService {public int type() {return 9289;} public void run() {System.out.println(9289);}}
    public static class Impl9290 implements DemoService {public int type() {return 9290;} public void run() {System.out.println(9290);}}
    public static class Impl9291 implements DemoService {public int type() {return 9291;} public void run() {System.out.println(9291);}}
    public static class Impl9292 implements DemoService {public int type() {return 9292;} public void run() {System.out.println(9292);}}
    public static class Impl9293 implements DemoService {public int type() {return 9293;} public void run() {System.out.println(9293);}}
    public static class Impl9294 implements DemoService {public int type() {return 9294;} public void run() {System.out.println(9294);}}
    public static class Impl9295 implements DemoService {public int type() {return 9295;} public void run() {System.out.println(9295);}}
    public static class Impl9296 implements DemoService {public int type() {return 9296;} public void run() {System.out.println(9296);}}
    public static class Impl9297 implements DemoService {public int type() {return 9297;} public void run() {System.out.println(9297);}}
    public static class Impl9298 implements DemoService {public int type() {return 9298;} public void run() {System.out.println(9298);}}
    public static class Impl9299 implements DemoService {public int type() {return 9299;} public void run() {System.out.println(9299);}}
    public static class Impl9300 implements DemoService {public int type() {return 9300;} public void run() {System.out.println(9300);}}
    public static class Impl9301 implements DemoService {public int type() {return 9301;} public void run() {System.out.println(9301);}}
    public static class Impl9302 implements DemoService {public int type() {return 9302;} public void run() {System.out.println(9302);}}
    public static class Impl9303 implements DemoService {public int type() {return 9303;} public void run() {System.out.println(9303);}}
    public static class Impl9304 implements DemoService {public int type() {return 9304;} public void run() {System.out.println(9304);}}
    public static class Impl9305 implements DemoService {public int type() {return 9305;} public void run() {System.out.println(9305);}}
    public static class Impl9306 implements DemoService {public int type() {return 9306;} public void run() {System.out.println(9306);}}
    public static class Impl9307 implements DemoService {public int type() {return 9307;} public void run() {System.out.println(9307);}}
    public static class Impl9308 implements DemoService {public int type() {return 9308;} public void run() {System.out.println(9308);}}
    public static class Impl9309 implements DemoService {public int type() {return 9309;} public void run() {System.out.println(9309);}}
    public static class Impl9310 implements DemoService {public int type() {return 9310;} public void run() {System.out.println(9310);}}
    public static class Impl9311 implements DemoService {public int type() {return 9311;} public void run() {System.out.println(9311);}}
    public static class Impl9312 implements DemoService {public int type() {return 9312;} public void run() {System.out.println(9312);}}
    public static class Impl9313 implements DemoService {public int type() {return 9313;} public void run() {System.out.println(9313);}}
    public static class Impl9314 implements DemoService {public int type() {return 9314;} public void run() {System.out.println(9314);}}
    public static class Impl9315 implements DemoService {public int type() {return 9315;} public void run() {System.out.println(9315);}}
    public static class Impl9316 implements DemoService {public int type() {return 9316;} public void run() {System.out.println(9316);}}
    public static class Impl9317 implements DemoService {public int type() {return 9317;} public void run() {System.out.println(9317);}}
    public static class Impl9318 implements DemoService {public int type() {return 9318;} public void run() {System.out.println(9318);}}
    public static class Impl9319 implements DemoService {public int type() {return 9319;} public void run() {System.out.println(9319);}}
    public static class Impl9320 implements DemoService {public int type() {return 9320;} public void run() {System.out.println(9320);}}
    public static class Impl9321 implements DemoService {public int type() {return 9321;} public void run() {System.out.println(9321);}}
    public static class Impl9322 implements DemoService {public int type() {return 9322;} public void run() {System.out.println(9322);}}
    public static class Impl9323 implements DemoService {public int type() {return 9323;} public void run() {System.out.println(9323);}}
    public static class Impl9324 implements DemoService {public int type() {return 9324;} public void run() {System.out.println(9324);}}
    public static class Impl9325 implements DemoService {public int type() {return 9325;} public void run() {System.out.println(9325);}}
    public static class Impl9326 implements DemoService {public int type() {return 9326;} public void run() {System.out.println(9326);}}
    public static class Impl9327 implements DemoService {public int type() {return 9327;} public void run() {System.out.println(9327);}}
    public static class Impl9328 implements DemoService {public int type() {return 9328;} public void run() {System.out.println(9328);}}
    public static class Impl9329 implements DemoService {public int type() {return 9329;} public void run() {System.out.println(9329);}}
    public static class Impl9330 implements DemoService {public int type() {return 9330;} public void run() {System.out.println(9330);}}
    public static class Impl9331 implements DemoService {public int type() {return 9331;} public void run() {System.out.println(9331);}}
    public static class Impl9332 implements DemoService {public int type() {return 9332;} public void run() {System.out.println(9332);}}
    public static class Impl9333 implements DemoService {public int type() {return 9333;} public void run() {System.out.println(9333);}}
    public static class Impl9334 implements DemoService {public int type() {return 9334;} public void run() {System.out.println(9334);}}
    public static class Impl9335 implements DemoService {public int type() {return 9335;} public void run() {System.out.println(9335);}}
    public static class Impl9336 implements DemoService {public int type() {return 9336;} public void run() {System.out.println(9336);}}
    public static class Impl9337 implements DemoService {public int type() {return 9337;} public void run() {System.out.println(9337);}}
    public static class Impl9338 implements DemoService {public int type() {return 9338;} public void run() {System.out.println(9338);}}
    public static class Impl9339 implements DemoService {public int type() {return 9339;} public void run() {System.out.println(9339);}}
    public static class Impl9340 implements DemoService {public int type() {return 9340;} public void run() {System.out.println(9340);}}
    public static class Impl9341 implements DemoService {public int type() {return 9341;} public void run() {System.out.println(9341);}}
    public static class Impl9342 implements DemoService {public int type() {return 9342;} public void run() {System.out.println(9342);}}
    public static class Impl9343 implements DemoService {public int type() {return 9343;} public void run() {System.out.println(9343);}}
    public static class Impl9344 implements DemoService {public int type() {return 9344;} public void run() {System.out.println(9344);}}
    public static class Impl9345 implements DemoService {public int type() {return 9345;} public void run() {System.out.println(9345);}}
    public static class Impl9346 implements DemoService {public int type() {return 9346;} public void run() {System.out.println(9346);}}
    public static class Impl9347 implements DemoService {public int type() {return 9347;} public void run() {System.out.println(9347);}}
    public static class Impl9348 implements DemoService {public int type() {return 9348;} public void run() {System.out.println(9348);}}
    public static class Impl9349 implements DemoService {public int type() {return 9349;} public void run() {System.out.println(9349);}}
    public static class Impl9350 implements DemoService {public int type() {return 9350;} public void run() {System.out.println(9350);}}
    public static class Impl9351 implements DemoService {public int type() {return 9351;} public void run() {System.out.println(9351);}}
    public static class Impl9352 implements DemoService {public int type() {return 9352;} public void run() {System.out.println(9352);}}
    public static class Impl9353 implements DemoService {public int type() {return 9353;} public void run() {System.out.println(9353);}}
    public static class Impl9354 implements DemoService {public int type() {return 9354;} public void run() {System.out.println(9354);}}
    public static class Impl9355 implements DemoService {public int type() {return 9355;} public void run() {System.out.println(9355);}}
    public static class Impl9356 implements DemoService {public int type() {return 9356;} public void run() {System.out.println(9356);}}
    public static class Impl9357 implements DemoService {public int type() {return 9357;} public void run() {System.out.println(9357);}}
    public static class Impl9358 implements DemoService {public int type() {return 9358;} public void run() {System.out.println(9358);}}
    public static class Impl9359 implements DemoService {public int type() {return 9359;} public void run() {System.out.println(9359);}}
    public static class Impl9360 implements DemoService {public int type() {return 9360;} public void run() {System.out.println(9360);}}
    public static class Impl9361 implements DemoService {public int type() {return 9361;} public void run() {System.out.println(9361);}}
    public static class Impl9362 implements DemoService {public int type() {return 9362;} public void run() {System.out.println(9362);}}
    public static class Impl9363 implements DemoService {public int type() {return 9363;} public void run() {System.out.println(9363);}}
    public static class Impl9364 implements DemoService {public int type() {return 9364;} public void run() {System.out.println(9364);}}
    public static class Impl9365 implements DemoService {public int type() {return 9365;} public void run() {System.out.println(9365);}}
    public static class Impl9366 implements DemoService {public int type() {return 9366;} public void run() {System.out.println(9366);}}
    public static class Impl9367 implements DemoService {public int type() {return 9367;} public void run() {System.out.println(9367);}}
    public static class Impl9368 implements DemoService {public int type() {return 9368;} public void run() {System.out.println(9368);}}
    public static class Impl9369 implements DemoService {public int type() {return 9369;} public void run() {System.out.println(9369);}}
    public static class Impl9370 implements DemoService {public int type() {return 9370;} public void run() {System.out.println(9370);}}
    public static class Impl9371 implements DemoService {public int type() {return 9371;} public void run() {System.out.println(9371);}}
    public static class Impl9372 implements DemoService {public int type() {return 9372;} public void run() {System.out.println(9372);}}
    public static class Impl9373 implements DemoService {public int type() {return 9373;} public void run() {System.out.println(9373);}}
    public static class Impl9374 implements DemoService {public int type() {return 9374;} public void run() {System.out.println(9374);}}
    public static class Impl9375 implements DemoService {public int type() {return 9375;} public void run() {System.out.println(9375);}}
    public static class Impl9376 implements DemoService {public int type() {return 9376;} public void run() {System.out.println(9376);}}
    public static class Impl9377 implements DemoService {public int type() {return 9377;} public void run() {System.out.println(9377);}}
    public static class Impl9378 implements DemoService {public int type() {return 9378;} public void run() {System.out.println(9378);}}
    public static class Impl9379 implements DemoService {public int type() {return 9379;} public void run() {System.out.println(9379);}}
    public static class Impl9380 implements DemoService {public int type() {return 9380;} public void run() {System.out.println(9380);}}
    public static class Impl9381 implements DemoService {public int type() {return 9381;} public void run() {System.out.println(9381);}}
    public static class Impl9382 implements DemoService {public int type() {return 9382;} public void run() {System.out.println(9382);}}
    public static class Impl9383 implements DemoService {public int type() {return 9383;} public void run() {System.out.println(9383);}}
    public static class Impl9384 implements DemoService {public int type() {return 9384;} public void run() {System.out.println(9384);}}
    public static class Impl9385 implements DemoService {public int type() {return 9385;} public void run() {System.out.println(9385);}}
    public static class Impl9386 implements DemoService {public int type() {return 9386;} public void run() {System.out.println(9386);}}
    public static class Impl9387 implements DemoService {public int type() {return 9387;} public void run() {System.out.println(9387);}}
    public static class Impl9388 implements DemoService {public int type() {return 9388;} public void run() {System.out.println(9388);}}
    public static class Impl9389 implements DemoService {public int type() {return 9389;} public void run() {System.out.println(9389);}}
    public static class Impl9390 implements DemoService {public int type() {return 9390;} public void run() {System.out.println(9390);}}
    public static class Impl9391 implements DemoService {public int type() {return 9391;} public void run() {System.out.println(9391);}}
    public static class Impl9392 implements DemoService {public int type() {return 9392;} public void run() {System.out.println(9392);}}
    public static class Impl9393 implements DemoService {public int type() {return 9393;} public void run() {System.out.println(9393);}}
    public static class Impl9394 implements DemoService {public int type() {return 9394;} public void run() {System.out.println(9394);}}
    public static class Impl9395 implements DemoService {public int type() {return 9395;} public void run() {System.out.println(9395);}}
    public static class Impl9396 implements DemoService {public int type() {return 9396;} public void run() {System.out.println(9396);}}
    public static class Impl9397 implements DemoService {public int type() {return 9397;} public void run() {System.out.println(9397);}}
    public static class Impl9398 implements DemoService {public int type() {return 9398;} public void run() {System.out.println(9398);}}
    public static class Impl9399 implements DemoService {public int type() {return 9399;} public void run() {System.out.println(9399);}}
    public static class Impl9400 implements DemoService {public int type() {return 9400;} public void run() {System.out.println(9400);}}
    public static class Impl9401 implements DemoService {public int type() {return 9401;} public void run() {System.out.println(9401);}}
    public static class Impl9402 implements DemoService {public int type() {return 9402;} public void run() {System.out.println(9402);}}
    public static class Impl9403 implements DemoService {public int type() {return 9403;} public void run() {System.out.println(9403);}}
    public static class Impl9404 implements DemoService {public int type() {return 9404;} public void run() {System.out.println(9404);}}
    public static class Impl9405 implements DemoService {public int type() {return 9405;} public void run() {System.out.println(9405);}}
    public static class Impl9406 implements DemoService {public int type() {return 9406;} public void run() {System.out.println(9406);}}
    public static class Impl9407 implements DemoService {public int type() {return 9407;} public void run() {System.out.println(9407);}}
    public static class Impl9408 implements DemoService {public int type() {return 9408;} public void run() {System.out.println(9408);}}
    public static class Impl9409 implements DemoService {public int type() {return 9409;} public void run() {System.out.println(9409);}}
    public static class Impl9410 implements DemoService {public int type() {return 9410;} public void run() {System.out.println(9410);}}
    public static class Impl9411 implements DemoService {public int type() {return 9411;} public void run() {System.out.println(9411);}}
    public static class Impl9412 implements DemoService {public int type() {return 9412;} public void run() {System.out.println(9412);}}
    public static class Impl9413 implements DemoService {public int type() {return 9413;} public void run() {System.out.println(9413);}}
    public static class Impl9414 implements DemoService {public int type() {return 9414;} public void run() {System.out.println(9414);}}
    public static class Impl9415 implements DemoService {public int type() {return 9415;} public void run() {System.out.println(9415);}}
    public static class Impl9416 implements DemoService {public int type() {return 9416;} public void run() {System.out.println(9416);}}
    public static class Impl9417 implements DemoService {public int type() {return 9417;} public void run() {System.out.println(9417);}}
    public static class Impl9418 implements DemoService {public int type() {return 9418;} public void run() {System.out.println(9418);}}
    public static class Impl9419 implements DemoService {public int type() {return 9419;} public void run() {System.out.println(9419);}}
    public static class Impl9420 implements DemoService {public int type() {return 9420;} public void run() {System.out.println(9420);}}
    public static class Impl9421 implements DemoService {public int type() {return 9421;} public void run() {System.out.println(9421);}}
    public static class Impl9422 implements DemoService {public int type() {return 9422;} public void run() {System.out.println(9422);}}
    public static class Impl9423 implements DemoService {public int type() {return 9423;} public void run() {System.out.println(9423);}}
    public static class Impl9424 implements DemoService {public int type() {return 9424;} public void run() {System.out.println(9424);}}
    public static class Impl9425 implements DemoService {public int type() {return 9425;} public void run() {System.out.println(9425);}}
    public static class Impl9426 implements DemoService {public int type() {return 9426;} public void run() {System.out.println(9426);}}
    public static class Impl9427 implements DemoService {public int type() {return 9427;} public void run() {System.out.println(9427);}}
    public static class Impl9428 implements DemoService {public int type() {return 9428;} public void run() {System.out.println(9428);}}
    public static class Impl9429 implements DemoService {public int type() {return 9429;} public void run() {System.out.println(9429);}}
    public static class Impl9430 implements DemoService {public int type() {return 9430;} public void run() {System.out.println(9430);}}
    public static class Impl9431 implements DemoService {public int type() {return 9431;} public void run() {System.out.println(9431);}}
    public static class Impl9432 implements DemoService {public int type() {return 9432;} public void run() {System.out.println(9432);}}
    public static class Impl9433 implements DemoService {public int type() {return 9433;} public void run() {System.out.println(9433);}}
    public static class Impl9434 implements DemoService {public int type() {return 9434;} public void run() {System.out.println(9434);}}
    public static class Impl9435 implements DemoService {public int type() {return 9435;} public void run() {System.out.println(9435);}}
    public static class Impl9436 implements DemoService {public int type() {return 9436;} public void run() {System.out.println(9436);}}
    public static class Impl9437 implements DemoService {public int type() {return 9437;} public void run() {System.out.println(9437);}}
    public static class Impl9438 implements DemoService {public int type() {return 9438;} public void run() {System.out.println(9438);}}
    public static class Impl9439 implements DemoService {public int type() {return 9439;} public void run() {System.out.println(9439);}}
    public static class Impl9440 implements DemoService {public int type() {return 9440;} public void run() {System.out.println(9440);}}
    public static class Impl9441 implements DemoService {public int type() {return 9441;} public void run() {System.out.println(9441);}}
    public static class Impl9442 implements DemoService {public int type() {return 9442;} public void run() {System.out.println(9442);}}
    public static class Impl9443 implements DemoService {public int type() {return 9443;} public void run() {System.out.println(9443);}}
    public static class Impl9444 implements DemoService {public int type() {return 9444;} public void run() {System.out.println(9444);}}
    public static class Impl9445 implements DemoService {public int type() {return 9445;} public void run() {System.out.println(9445);}}
    public static class Impl9446 implements DemoService {public int type() {return 9446;} public void run() {System.out.println(9446);}}
    public static class Impl9447 implements DemoService {public int type() {return 9447;} public void run() {System.out.println(9447);}}
    public static class Impl9448 implements DemoService {public int type() {return 9448;} public void run() {System.out.println(9448);}}
    public static class Impl9449 implements DemoService {public int type() {return 9449;} public void run() {System.out.println(9449);}}
    public static class Impl9450 implements DemoService {public int type() {return 9450;} public void run() {System.out.println(9450);}}
    public static class Impl9451 implements DemoService {public int type() {return 9451;} public void run() {System.out.println(9451);}}
    public static class Impl9452 implements DemoService {public int type() {return 9452;} public void run() {System.out.println(9452);}}
    public static class Impl9453 implements DemoService {public int type() {return 9453;} public void run() {System.out.println(9453);}}
    public static class Impl9454 implements DemoService {public int type() {return 9454;} public void run() {System.out.println(9454);}}
    public static class Impl9455 implements DemoService {public int type() {return 9455;} public void run() {System.out.println(9455);}}
    public static class Impl9456 implements DemoService {public int type() {return 9456;} public void run() {System.out.println(9456);}}
    public static class Impl9457 implements DemoService {public int type() {return 9457;} public void run() {System.out.println(9457);}}
    public static class Impl9458 implements DemoService {public int type() {return 9458;} public void run() {System.out.println(9458);}}
    public static class Impl9459 implements DemoService {public int type() {return 9459;} public void run() {System.out.println(9459);}}
    public static class Impl9460 implements DemoService {public int type() {return 9460;} public void run() {System.out.println(9460);}}
    public static class Impl9461 implements DemoService {public int type() {return 9461;} public void run() {System.out.println(9461);}}
    public static class Impl9462 implements DemoService {public int type() {return 9462;} public void run() {System.out.println(9462);}}
    public static class Impl9463 implements DemoService {public int type() {return 9463;} public void run() {System.out.println(9463);}}
    public static class Impl9464 implements DemoService {public int type() {return 9464;} public void run() {System.out.println(9464);}}
    public static class Impl9465 implements DemoService {public int type() {return 9465;} public void run() {System.out.println(9465);}}
    public static class Impl9466 implements DemoService {public int type() {return 9466;} public void run() {System.out.println(9466);}}
    public static class Impl9467 implements DemoService {public int type() {return 9467;} public void run() {System.out.println(9467);}}
    public static class Impl9468 implements DemoService {public int type() {return 9468;} public void run() {System.out.println(9468);}}
    public static class Impl9469 implements DemoService {public int type() {return 9469;} public void run() {System.out.println(9469);}}
    public static class Impl9470 implements DemoService {public int type() {return 9470;} public void run() {System.out.println(9470);}}
    public static class Impl9471 implements DemoService {public int type() {return 9471;} public void run() {System.out.println(9471);}}
    public static class Impl9472 implements DemoService {public int type() {return 9472;} public void run() {System.out.println(9472);}}
    public static class Impl9473 implements DemoService {public int type() {return 9473;} public void run() {System.out.println(9473);}}
    public static class Impl9474 implements DemoService {public int type() {return 9474;} public void run() {System.out.println(9474);}}
    public static class Impl9475 implements DemoService {public int type() {return 9475;} public void run() {System.out.println(9475);}}
    public static class Impl9476 implements DemoService {public int type() {return 9476;} public void run() {System.out.println(9476);}}
    public static class Impl9477 implements DemoService {public int type() {return 9477;} public void run() {System.out.println(9477);}}
    public static class Impl9478 implements DemoService {public int type() {return 9478;} public void run() {System.out.println(9478);}}
    public static class Impl9479 implements DemoService {public int type() {return 9479;} public void run() {System.out.println(9479);}}
    public static class Impl9480 implements DemoService {public int type() {return 9480;} public void run() {System.out.println(9480);}}
    public static class Impl9481 implements DemoService {public int type() {return 9481;} public void run() {System.out.println(9481);}}
    public static class Impl9482 implements DemoService {public int type() {return 9482;} public void run() {System.out.println(9482);}}
    public static class Impl9483 implements DemoService {public int type() {return 9483;} public void run() {System.out.println(9483);}}
    public static class Impl9484 implements DemoService {public int type() {return 9484;} public void run() {System.out.println(9484);}}
    public static class Impl9485 implements DemoService {public int type() {return 9485;} public void run() {System.out.println(9485);}}
    public static class Impl9486 implements DemoService {public int type() {return 9486;} public void run() {System.out.println(9486);}}
    public static class Impl9487 implements DemoService {public int type() {return 9487;} public void run() {System.out.println(9487);}}
    public static class Impl9488 implements DemoService {public int type() {return 9488;} public void run() {System.out.println(9488);}}
    public static class Impl9489 implements DemoService {public int type() {return 9489;} public void run() {System.out.println(9489);}}
    public static class Impl9490 implements DemoService {public int type() {return 9490;} public void run() {System.out.println(9490);}}
    public static class Impl9491 implements DemoService {public int type() {return 9491;} public void run() {System.out.println(9491);}}
    public static class Impl9492 implements DemoService {public int type() {return 9492;} public void run() {System.out.println(9492);}}
    public static class Impl9493 implements DemoService {public int type() {return 9493;} public void run() {System.out.println(9493);}}
    public static class Impl9494 implements DemoService {public int type() {return 9494;} public void run() {System.out.println(9494);}}
    public static class Impl9495 implements DemoService {public int type() {return 9495;} public void run() {System.out.println(9495);}}
    public static class Impl9496 implements DemoService {public int type() {return 9496;} public void run() {System.out.println(9496);}}
    public static class Impl9497 implements DemoService {public int type() {return 9497;} public void run() {System.out.println(9497);}}
    public static class Impl9498 implements DemoService {public int type() {return 9498;} public void run() {System.out.println(9498);}}
    public static class Impl9499 implements DemoService {public int type() {return 9499;} public void run() {System.out.println(9499);}}
    public static class Impl9500 implements DemoService {public int type() {return 9500;} public void run() {System.out.println(9500);}}
    public static class Impl9501 implements DemoService {public int type() {return 9501;} public void run() {System.out.println(9501);}}
    public static class Impl9502 implements DemoService {public int type() {return 9502;} public void run() {System.out.println(9502);}}
    public static class Impl9503 implements DemoService {public int type() {return 9503;} public void run() {System.out.println(9503);}}
    public static class Impl9504 implements DemoService {public int type() {return 9504;} public void run() {System.out.println(9504);}}
    public static class Impl9505 implements DemoService {public int type() {return 9505;} public void run() {System.out.println(9505);}}
    public static class Impl9506 implements DemoService {public int type() {return 9506;} public void run() {System.out.println(9506);}}
    public static class Impl9507 implements DemoService {public int type() {return 9507;} public void run() {System.out.println(9507);}}
    public static class Impl9508 implements DemoService {public int type() {return 9508;} public void run() {System.out.println(9508);}}
    public static class Impl9509 implements DemoService {public int type() {return 9509;} public void run() {System.out.println(9509);}}
    public static class Impl9510 implements DemoService {public int type() {return 9510;} public void run() {System.out.println(9510);}}
    public static class Impl9511 implements DemoService {public int type() {return 9511;} public void run() {System.out.println(9511);}}
    public static class Impl9512 implements DemoService {public int type() {return 9512;} public void run() {System.out.println(9512);}}
    public static class Impl9513 implements DemoService {public int type() {return 9513;} public void run() {System.out.println(9513);}}
    public static class Impl9514 implements DemoService {public int type() {return 9514;} public void run() {System.out.println(9514);}}
    public static class Impl9515 implements DemoService {public int type() {return 9515;} public void run() {System.out.println(9515);}}
    public static class Impl9516 implements DemoService {public int type() {return 9516;} public void run() {System.out.println(9516);}}
    public static class Impl9517 implements DemoService {public int type() {return 9517;} public void run() {System.out.println(9517);}}
    public static class Impl9518 implements DemoService {public int type() {return 9518;} public void run() {System.out.println(9518);}}
    public static class Impl9519 implements DemoService {public int type() {return 9519;} public void run() {System.out.println(9519);}}
    public static class Impl9520 implements DemoService {public int type() {return 9520;} public void run() {System.out.println(9520);}}
    public static class Impl9521 implements DemoService {public int type() {return 9521;} public void run() {System.out.println(9521);}}
    public static class Impl9522 implements DemoService {public int type() {return 9522;} public void run() {System.out.println(9522);}}
    public static class Impl9523 implements DemoService {public int type() {return 9523;} public void run() {System.out.println(9523);}}
    public static class Impl9524 implements DemoService {public int type() {return 9524;} public void run() {System.out.println(9524);}}
    public static class Impl9525 implements DemoService {public int type() {return 9525;} public void run() {System.out.println(9525);}}
    public static class Impl9526 implements DemoService {public int type() {return 9526;} public void run() {System.out.println(9526);}}
    public static class Impl9527 implements DemoService {public int type() {return 9527;} public void run() {System.out.println(9527);}}
    public static class Impl9528 implements DemoService {public int type() {return 9528;} public void run() {System.out.println(9528);}}
    public static class Impl9529 implements DemoService {public int type() {return 9529;} public void run() {System.out.println(9529);}}
    public static class Impl9530 implements DemoService {public int type() {return 9530;} public void run() {System.out.println(9530);}}
    public static class Impl9531 implements DemoService {public int type() {return 9531;} public void run() {System.out.println(9531);}}
    public static class Impl9532 implements DemoService {public int type() {return 9532;} public void run() {System.out.println(9532);}}
    public static class Impl9533 implements DemoService {public int type() {return 9533;} public void run() {System.out.println(9533);}}
    public static class Impl9534 implements DemoService {public int type() {return 9534;} public void run() {System.out.println(9534);}}
    public static class Impl9535 implements DemoService {public int type() {return 9535;} public void run() {System.out.println(9535);}}
    public static class Impl9536 implements DemoService {public int type() {return 9536;} public void run() {System.out.println(9536);}}
    public static class Impl9537 implements DemoService {public int type() {return 9537;} public void run() {System.out.println(9537);}}
    public static class Impl9538 implements DemoService {public int type() {return 9538;} public void run() {System.out.println(9538);}}
    public static class Impl9539 implements DemoService {public int type() {return 9539;} public void run() {System.out.println(9539);}}
    public static class Impl9540 implements DemoService {public int type() {return 9540;} public void run() {System.out.println(9540);}}
    public static class Impl9541 implements DemoService {public int type() {return 9541;} public void run() {System.out.println(9541);}}
    public static class Impl9542 implements DemoService {public int type() {return 9542;} public void run() {System.out.println(9542);}}
    public static class Impl9543 implements DemoService {public int type() {return 9543;} public void run() {System.out.println(9543);}}
    public static class Impl9544 implements DemoService {public int type() {return 9544;} public void run() {System.out.println(9544);}}
    public static class Impl9545 implements DemoService {public int type() {return 9545;} public void run() {System.out.println(9545);}}
    public static class Impl9546 implements DemoService {public int type() {return 9546;} public void run() {System.out.println(9546);}}
    public static class Impl9547 implements DemoService {public int type() {return 9547;} public void run() {System.out.println(9547);}}
    public static class Impl9548 implements DemoService {public int type() {return 9548;} public void run() {System.out.println(9548);}}
    public static class Impl9549 implements DemoService {public int type() {return 9549;} public void run() {System.out.println(9549);}}
    public static class Impl9550 implements DemoService {public int type() {return 9550;} public void run() {System.out.println(9550);}}
    public static class Impl9551 implements DemoService {public int type() {return 9551;} public void run() {System.out.println(9551);}}
    public static class Impl9552 implements DemoService {public int type() {return 9552;} public void run() {System.out.println(9552);}}
    public static class Impl9553 implements DemoService {public int type() {return 9553;} public void run() {System.out.println(9553);}}
    public static class Impl9554 implements DemoService {public int type() {return 9554;} public void run() {System.out.println(9554);}}
    public static class Impl9555 implements DemoService {public int type() {return 9555;} public void run() {System.out.println(9555);}}
    public static class Impl9556 implements DemoService {public int type() {return 9556;} public void run() {System.out.println(9556);}}
    public static class Impl9557 implements DemoService {public int type() {return 9557;} public void run() {System.out.println(9557);}}
    public static class Impl9558 implements DemoService {public int type() {return 9558;} public void run() {System.out.println(9558);}}
    public static class Impl9559 implements DemoService {public int type() {return 9559;} public void run() {System.out.println(9559);}}
    public static class Impl9560 implements DemoService {public int type() {return 9560;} public void run() {System.out.println(9560);}}
    public static class Impl9561 implements DemoService {public int type() {return 9561;} public void run() {System.out.println(9561);}}
    public static class Impl9562 implements DemoService {public int type() {return 9562;} public void run() {System.out.println(9562);}}
    public static class Impl9563 implements DemoService {public int type() {return 9563;} public void run() {System.out.println(9563);}}
    public static class Impl9564 implements DemoService {public int type() {return 9564;} public void run() {System.out.println(9564);}}
    public static class Impl9565 implements DemoService {public int type() {return 9565;} public void run() {System.out.println(9565);}}
    public static class Impl9566 implements DemoService {public int type() {return 9566;} public void run() {System.out.println(9566);}}
    public static class Impl9567 implements DemoService {public int type() {return 9567;} public void run() {System.out.println(9567);}}
    public static class Impl9568 implements DemoService {public int type() {return 9568;} public void run() {System.out.println(9568);}}
    public static class Impl9569 implements DemoService {public int type() {return 9569;} public void run() {System.out.println(9569);}}
    public static class Impl9570 implements DemoService {public int type() {return 9570;} public void run() {System.out.println(9570);}}
    public static class Impl9571 implements DemoService {public int type() {return 9571;} public void run() {System.out.println(9571);}}
    public static class Impl9572 implements DemoService {public int type() {return 9572;} public void run() {System.out.println(9572);}}
    public static class Impl9573 implements DemoService {public int type() {return 9573;} public void run() {System.out.println(9573);}}
    public static class Impl9574 implements DemoService {public int type() {return 9574;} public void run() {System.out.println(9574);}}
    public static class Impl9575 implements DemoService {public int type() {return 9575;} public void run() {System.out.println(9575);}}
    public static class Impl9576 implements DemoService {public int type() {return 9576;} public void run() {System.out.println(9576);}}
    public static class Impl9577 implements DemoService {public int type() {return 9577;} public void run() {System.out.println(9577);}}
    public static class Impl9578 implements DemoService {public int type() {return 9578;} public void run() {System.out.println(9578);}}
    public static class Impl9579 implements DemoService {public int type() {return 9579;} public void run() {System.out.println(9579);}}
    public static class Impl9580 implements DemoService {public int type() {return 9580;} public void run() {System.out.println(9580);}}
    public static class Impl9581 implements DemoService {public int type() {return 9581;} public void run() {System.out.println(9581);}}
    public static class Impl9582 implements DemoService {public int type() {return 9582;} public void run() {System.out.println(9582);}}
    public static class Impl9583 implements DemoService {public int type() {return 9583;} public void run() {System.out.println(9583);}}
    public static class Impl9584 implements DemoService {public int type() {return 9584;} public void run() {System.out.println(9584);}}
    public static class Impl9585 implements DemoService {public int type() {return 9585;} public void run() {System.out.println(9585);}}
    public static class Impl9586 implements DemoService {public int type() {return 9586;} public void run() {System.out.println(9586);}}
    public static class Impl9587 implements DemoService {public int type() {return 9587;} public void run() {System.out.println(9587);}}
    public static class Impl9588 implements DemoService {public int type() {return 9588;} public void run() {System.out.println(9588);}}
    public static class Impl9589 implements DemoService {public int type() {return 9589;} public void run() {System.out.println(9589);}}
    public static class Impl9590 implements DemoService {public int type() {return 9590;} public void run() {System.out.println(9590);}}
    public static class Impl9591 implements DemoService {public int type() {return 9591;} public void run() {System.out.println(9591);}}
    public static class Impl9592 implements DemoService {public int type() {return 9592;} public void run() {System.out.println(9592);}}
    public static class Impl9593 implements DemoService {public int type() {return 9593;} public void run() {System.out.println(9593);}}
    public static class Impl9594 implements DemoService {public int type() {return 9594;} public void run() {System.out.println(9594);}}
    public static class Impl9595 implements DemoService {public int type() {return 9595;} public void run() {System.out.println(9595);}}
    public static class Impl9596 implements DemoService {public int type() {return 9596;} public void run() {System.out.println(9596);}}
    public static class Impl9597 implements DemoService {public int type() {return 9597;} public void run() {System.out.println(9597);}}
    public static class Impl9598 implements DemoService {public int type() {return 9598;} public void run() {System.out.println(9598);}}
    public static class Impl9599 implements DemoService {public int type() {return 9599;} public void run() {System.out.println(9599);}}
    public static class Impl9600 implements DemoService {public int type() {return 9600;} public void run() {System.out.println(9600);}}
    public static class Impl9601 implements DemoService {public int type() {return 9601;} public void run() {System.out.println(9601);}}
    public static class Impl9602 implements DemoService {public int type() {return 9602;} public void run() {System.out.println(9602);}}
    public static class Impl9603 implements DemoService {public int type() {return 9603;} public void run() {System.out.println(9603);}}
    public static class Impl9604 implements DemoService {public int type() {return 9604;} public void run() {System.out.println(9604);}}
    public static class Impl9605 implements DemoService {public int type() {return 9605;} public void run() {System.out.println(9605);}}
    public static class Impl9606 implements DemoService {public int type() {return 9606;} public void run() {System.out.println(9606);}}
    public static class Impl9607 implements DemoService {public int type() {return 9607;} public void run() {System.out.println(9607);}}
    public static class Impl9608 implements DemoService {public int type() {return 9608;} public void run() {System.out.println(9608);}}
    public static class Impl9609 implements DemoService {public int type() {return 9609;} public void run() {System.out.println(9609);}}
    public static class Impl9610 implements DemoService {public int type() {return 9610;} public void run() {System.out.println(9610);}}
    public static class Impl9611 implements DemoService {public int type() {return 9611;} public void run() {System.out.println(9611);}}
    public static class Impl9612 implements DemoService {public int type() {return 9612;} public void run() {System.out.println(9612);}}
    public static class Impl9613 implements DemoService {public int type() {return 9613;} public void run() {System.out.println(9613);}}
    public static class Impl9614 implements DemoService {public int type() {return 9614;} public void run() {System.out.println(9614);}}
    public static class Impl9615 implements DemoService {public int type() {return 9615;} public void run() {System.out.println(9615);}}
    public static class Impl9616 implements DemoService {public int type() {return 9616;} public void run() {System.out.println(9616);}}
    public static class Impl9617 implements DemoService {public int type() {return 9617;} public void run() {System.out.println(9617);}}
    public static class Impl9618 implements DemoService {public int type() {return 9618;} public void run() {System.out.println(9618);}}
    public static class Impl9619 implements DemoService {public int type() {return 9619;} public void run() {System.out.println(9619);}}
    public static class Impl9620 implements DemoService {public int type() {return 9620;} public void run() {System.out.println(9620);}}
    public static class Impl9621 implements DemoService {public int type() {return 9621;} public void run() {System.out.println(9621);}}
    public static class Impl9622 implements DemoService {public int type() {return 9622;} public void run() {System.out.println(9622);}}
    public static class Impl9623 implements DemoService {public int type() {return 9623;} public void run() {System.out.println(9623);}}
    public static class Impl9624 implements DemoService {public int type() {return 9624;} public void run() {System.out.println(9624);}}
    public static class Impl9625 implements DemoService {public int type() {return 9625;} public void run() {System.out.println(9625);}}
    public static class Impl9626 implements DemoService {public int type() {return 9626;} public void run() {System.out.println(9626);}}
    public static class Impl9627 implements DemoService {public int type() {return 9627;} public void run() {System.out.println(9627);}}
    public static class Impl9628 implements DemoService {public int type() {return 9628;} public void run() {System.out.println(9628);}}
    public static class Impl9629 implements DemoService {public int type() {return 9629;} public void run() {System.out.println(9629);}}
    public static class Impl9630 implements DemoService {public int type() {return 9630;} public void run() {System.out.println(9630);}}
    public static class Impl9631 implements DemoService {public int type() {return 9631;} public void run() {System.out.println(9631);}}
    public static class Impl9632 implements DemoService {public int type() {return 9632;} public void run() {System.out.println(9632);}}
    public static class Impl9633 implements DemoService {public int type() {return 9633;} public void run() {System.out.println(9633);}}
    public static class Impl9634 implements DemoService {public int type() {return 9634;} public void run() {System.out.println(9634);}}
    public static class Impl9635 implements DemoService {public int type() {return 9635;} public void run() {System.out.println(9635);}}
    public static class Impl9636 implements DemoService {public int type() {return 9636;} public void run() {System.out.println(9636);}}
    public static class Impl9637 implements DemoService {public int type() {return 9637;} public void run() {System.out.println(9637);}}
    public static class Impl9638 implements DemoService {public int type() {return 9638;} public void run() {System.out.println(9638);}}
    public static class Impl9639 implements DemoService {public int type() {return 9639;} public void run() {System.out.println(9639);}}
    public static class Impl9640 implements DemoService {public int type() {return 9640;} public void run() {System.out.println(9640);}}
    public static class Impl9641 implements DemoService {public int type() {return 9641;} public void run() {System.out.println(9641);}}
    public static class Impl9642 implements DemoService {public int type() {return 9642;} public void run() {System.out.println(9642);}}
    public static class Impl9643 implements DemoService {public int type() {return 9643;} public void run() {System.out.println(9643);}}
    public static class Impl9644 implements DemoService {public int type() {return 9644;} public void run() {System.out.println(9644);}}
    public static class Impl9645 implements DemoService {public int type() {return 9645;} public void run() {System.out.println(9645);}}
    public static class Impl9646 implements DemoService {public int type() {return 9646;} public void run() {System.out.println(9646);}}
    public static class Impl9647 implements DemoService {public int type() {return 9647;} public void run() {System.out.println(9647);}}
    public static class Impl9648 implements DemoService {public int type() {return 9648;} public void run() {System.out.println(9648);}}
    public static class Impl9649 implements DemoService {public int type() {return 9649;} public void run() {System.out.println(9649);}}
    public static class Impl9650 implements DemoService {public int type() {return 9650;} public void run() {System.out.println(9650);}}
    public static class Impl9651 implements DemoService {public int type() {return 9651;} public void run() {System.out.println(9651);}}
    public static class Impl9652 implements DemoService {public int type() {return 9652;} public void run() {System.out.println(9652);}}
    public static class Impl9653 implements DemoService {public int type() {return 9653;} public void run() {System.out.println(9653);}}
    public static class Impl9654 implements DemoService {public int type() {return 9654;} public void run() {System.out.println(9654);}}
    public static class Impl9655 implements DemoService {public int type() {return 9655;} public void run() {System.out.println(9655);}}
    public static class Impl9656 implements DemoService {public int type() {return 9656;} public void run() {System.out.println(9656);}}
    public static class Impl9657 implements DemoService {public int type() {return 9657;} public void run() {System.out.println(9657);}}
    public static class Impl9658 implements DemoService {public int type() {return 9658;} public void run() {System.out.println(9658);}}
    public static class Impl9659 implements DemoService {public int type() {return 9659;} public void run() {System.out.println(9659);}}
    public static class Impl9660 implements DemoService {public int type() {return 9660;} public void run() {System.out.println(9660);}}
    public static class Impl9661 implements DemoService {public int type() {return 9661;} public void run() {System.out.println(9661);}}
    public static class Impl9662 implements DemoService {public int type() {return 9662;} public void run() {System.out.println(9662);}}
    public static class Impl9663 implements DemoService {public int type() {return 9663;} public void run() {System.out.println(9663);}}
    public static class Impl9664 implements DemoService {public int type() {return 9664;} public void run() {System.out.println(9664);}}
    public static class Impl9665 implements DemoService {public int type() {return 9665;} public void run() {System.out.println(9665);}}
    public static class Impl9666 implements DemoService {public int type() {return 9666;} public void run() {System.out.println(9666);}}
    public static class Impl9667 implements DemoService {public int type() {return 9667;} public void run() {System.out.println(9667);}}
    public static class Impl9668 implements DemoService {public int type() {return 9668;} public void run() {System.out.println(9668);}}
    public static class Impl9669 implements DemoService {public int type() {return 9669;} public void run() {System.out.println(9669);}}
    public static class Impl9670 implements DemoService {public int type() {return 9670;} public void run() {System.out.println(9670);}}
    public static class Impl9671 implements DemoService {public int type() {return 9671;} public void run() {System.out.println(9671);}}
    public static class Impl9672 implements DemoService {public int type() {return 9672;} public void run() {System.out.println(9672);}}
    public static class Impl9673 implements DemoService {public int type() {return 9673;} public void run() {System.out.println(9673);}}
    public static class Impl9674 implements DemoService {public int type() {return 9674;} public void run() {System.out.println(9674);}}
    public static class Impl9675 implements DemoService {public int type() {return 9675;} public void run() {System.out.println(9675);}}
    public static class Impl9676 implements DemoService {public int type() {return 9676;} public void run() {System.out.println(9676);}}
    public static class Impl9677 implements DemoService {public int type() {return 9677;} public void run() {System.out.println(9677);}}
    public static class Impl9678 implements DemoService {public int type() {return 9678;} public void run() {System.out.println(9678);}}
    public static class Impl9679 implements DemoService {public int type() {return 9679;} public void run() {System.out.println(9679);}}
    public static class Impl9680 implements DemoService {public int type() {return 9680;} public void run() {System.out.println(9680);}}
    public static class Impl9681 implements DemoService {public int type() {return 9681;} public void run() {System.out.println(9681);}}
    public static class Impl9682 implements DemoService {public int type() {return 9682;} public void run() {System.out.println(9682);}}
    public static class Impl9683 implements DemoService {public int type() {return 9683;} public void run() {System.out.println(9683);}}
    public static class Impl9684 implements DemoService {public int type() {return 9684;} public void run() {System.out.println(9684);}}
    public static class Impl9685 implements DemoService {public int type() {return 9685;} public void run() {System.out.println(9685);}}
    public static class Impl9686 implements DemoService {public int type() {return 9686;} public void run() {System.out.println(9686);}}
    public static class Impl9687 implements DemoService {public int type() {return 9687;} public void run() {System.out.println(9687);}}
    public static class Impl9688 implements DemoService {public int type() {return 9688;} public void run() {System.out.println(9688);}}
    public static class Impl9689 implements DemoService {public int type() {return 9689;} public void run() {System.out.println(9689);}}
    public static class Impl9690 implements DemoService {public int type() {return 9690;} public void run() {System.out.println(9690);}}
    public static class Impl9691 implements DemoService {public int type() {return 9691;} public void run() {System.out.println(9691);}}
    public static class Impl9692 implements DemoService {public int type() {return 9692;} public void run() {System.out.println(9692);}}
    public static class Impl9693 implements DemoService {public int type() {return 9693;} public void run() {System.out.println(9693);}}
    public static class Impl9694 implements DemoService {public int type() {return 9694;} public void run() {System.out.println(9694);}}
    public static class Impl9695 implements DemoService {public int type() {return 9695;} public void run() {System.out.println(9695);}}
    public static class Impl9696 implements DemoService {public int type() {return 9696;} public void run() {System.out.println(9696);}}
    public static class Impl9697 implements DemoService {public int type() {return 9697;} public void run() {System.out.println(9697);}}
    public static class Impl9698 implements DemoService {public int type() {return 9698;} public void run() {System.out.println(9698);}}
    public static class Impl9699 implements DemoService {public int type() {return 9699;} public void run() {System.out.println(9699);}}
    public static class Impl9700 implements DemoService {public int type() {return 9700;} public void run() {System.out.println(9700);}}
    public static class Impl9701 implements DemoService {public int type() {return 9701;} public void run() {System.out.println(9701);}}
    public static class Impl9702 implements DemoService {public int type() {return 9702;} public void run() {System.out.println(9702);}}
    public static class Impl9703 implements DemoService {public int type() {return 9703;} public void run() {System.out.println(9703);}}
    public static class Impl9704 implements DemoService {public int type() {return 9704;} public void run() {System.out.println(9704);}}
    public static class Impl9705 implements DemoService {public int type() {return 9705;} public void run() {System.out.println(9705);}}
    public static class Impl9706 implements DemoService {public int type() {return 9706;} public void run() {System.out.println(9706);}}
    public static class Impl9707 implements DemoService {public int type() {return 9707;} public void run() {System.out.println(9707);}}
    public static class Impl9708 implements DemoService {public int type() {return 9708;} public void run() {System.out.println(9708);}}
    public static class Impl9709 implements DemoService {public int type() {return 9709;} public void run() {System.out.println(9709);}}
    public static class Impl9710 implements DemoService {public int type() {return 9710;} public void run() {System.out.println(9710);}}
    public static class Impl9711 implements DemoService {public int type() {return 9711;} public void run() {System.out.println(9711);}}
    public static class Impl9712 implements DemoService {public int type() {return 9712;} public void run() {System.out.println(9712);}}
    public static class Impl9713 implements DemoService {public int type() {return 9713;} public void run() {System.out.println(9713);}}
    public static class Impl9714 implements DemoService {public int type() {return 9714;} public void run() {System.out.println(9714);}}
    public static class Impl9715 implements DemoService {public int type() {return 9715;} public void run() {System.out.println(9715);}}
    public static class Impl9716 implements DemoService {public int type() {return 9716;} public void run() {System.out.println(9716);}}
    public static class Impl9717 implements DemoService {public int type() {return 9717;} public void run() {System.out.println(9717);}}
    public static class Impl9718 implements DemoService {public int type() {return 9718;} public void run() {System.out.println(9718);}}
    public static class Impl9719 implements DemoService {public int type() {return 9719;} public void run() {System.out.println(9719);}}
    public static class Impl9720 implements DemoService {public int type() {return 9720;} public void run() {System.out.println(9720);}}
    public static class Impl9721 implements DemoService {public int type() {return 9721;} public void run() {System.out.println(9721);}}
    public static class Impl9722 implements DemoService {public int type() {return 9722;} public void run() {System.out.println(9722);}}
    public static class Impl9723 implements DemoService {public int type() {return 9723;} public void run() {System.out.println(9723);}}
    public static class Impl9724 implements DemoService {public int type() {return 9724;} public void run() {System.out.println(9724);}}
    public static class Impl9725 implements DemoService {public int type() {return 9725;} public void run() {System.out.println(9725);}}
    public static class Impl9726 implements DemoService {public int type() {return 9726;} public void run() {System.out.println(9726);}}
    public static class Impl9727 implements DemoService {public int type() {return 9727;} public void run() {System.out.println(9727);}}
    public static class Impl9728 implements DemoService {public int type() {return 9728;} public void run() {System.out.println(9728);}}
    public static class Impl9729 implements DemoService {public int type() {return 9729;} public void run() {System.out.println(9729);}}
    public static class Impl9730 implements DemoService {public int type() {return 9730;} public void run() {System.out.println(9730);}}
    public static class Impl9731 implements DemoService {public int type() {return 9731;} public void run() {System.out.println(9731);}}
    public static class Impl9732 implements DemoService {public int type() {return 9732;} public void run() {System.out.println(9732);}}
    public static class Impl9733 implements DemoService {public int type() {return 9733;} public void run() {System.out.println(9733);}}
    public static class Impl9734 implements DemoService {public int type() {return 9734;} public void run() {System.out.println(9734);}}
    public static class Impl9735 implements DemoService {public int type() {return 9735;} public void run() {System.out.println(9735);}}
    public static class Impl9736 implements DemoService {public int type() {return 9736;} public void run() {System.out.println(9736);}}
    public static class Impl9737 implements DemoService {public int type() {return 9737;} public void run() {System.out.println(9737);}}
    public static class Impl9738 implements DemoService {public int type() {return 9738;} public void run() {System.out.println(9738);}}
    public static class Impl9739 implements DemoService {public int type() {return 9739;} public void run() {System.out.println(9739);}}
    public static class Impl9740 implements DemoService {public int type() {return 9740;} public void run() {System.out.println(9740);}}
    public static class Impl9741 implements DemoService {public int type() {return 9741;} public void run() {System.out.println(9741);}}
    public static class Impl9742 implements DemoService {public int type() {return 9742;} public void run() {System.out.println(9742);}}
    public static class Impl9743 implements DemoService {public int type() {return 9743;} public void run() {System.out.println(9743);}}
    public static class Impl9744 implements DemoService {public int type() {return 9744;} public void run() {System.out.println(9744);}}
    public static class Impl9745 implements DemoService {public int type() {return 9745;} public void run() {System.out.println(9745);}}
    public static class Impl9746 implements DemoService {public int type() {return 9746;} public void run() {System.out.println(9746);}}
    public static class Impl9747 implements DemoService {public int type() {return 9747;} public void run() {System.out.println(9747);}}
    public static class Impl9748 implements DemoService {public int type() {return 9748;} public void run() {System.out.println(9748);}}
    public static class Impl9749 implements DemoService {public int type() {return 9749;} public void run() {System.out.println(9749);}}
    public static class Impl9750 implements DemoService {public int type() {return 9750;} public void run() {System.out.println(9750);}}
    public static class Impl9751 implements DemoService {public int type() {return 9751;} public void run() {System.out.println(9751);}}
    public static class Impl9752 implements DemoService {public int type() {return 9752;} public void run() {System.out.println(9752);}}
    public static class Impl9753 implements DemoService {public int type() {return 9753;} public void run() {System.out.println(9753);}}
    public static class Impl9754 implements DemoService {public int type() {return 9754;} public void run() {System.out.println(9754);}}
    public static class Impl9755 implements DemoService {public int type() {return 9755;} public void run() {System.out.println(9755);}}
    public static class Impl9756 implements DemoService {public int type() {return 9756;} public void run() {System.out.println(9756);}}
    public static class Impl9757 implements DemoService {public int type() {return 9757;} public void run() {System.out.println(9757);}}
    public static class Impl9758 implements DemoService {public int type() {return 9758;} public void run() {System.out.println(9758);}}
    public static class Impl9759 implements DemoService {public int type() {return 9759;} public void run() {System.out.println(9759);}}
    public static class Impl9760 implements DemoService {public int type() {return 9760;} public void run() {System.out.println(9760);}}
    public static class Impl9761 implements DemoService {public int type() {return 9761;} public void run() {System.out.println(9761);}}
    public static class Impl9762 implements DemoService {public int type() {return 9762;} public void run() {System.out.println(9762);}}
    public static class Impl9763 implements DemoService {public int type() {return 9763;} public void run() {System.out.println(9763);}}
    public static class Impl9764 implements DemoService {public int type() {return 9764;} public void run() {System.out.println(9764);}}
    public static class Impl9765 implements DemoService {public int type() {return 9765;} public void run() {System.out.println(9765);}}
    public static class Impl9766 implements DemoService {public int type() {return 9766;} public void run() {System.out.println(9766);}}
    public static class Impl9767 implements DemoService {public int type() {return 9767;} public void run() {System.out.println(9767);}}
    public static class Impl9768 implements DemoService {public int type() {return 9768;} public void run() {System.out.println(9768);}}
    public static class Impl9769 implements DemoService {public int type() {return 9769;} public void run() {System.out.println(9769);}}
    public static class Impl9770 implements DemoService {public int type() {return 9770;} public void run() {System.out.println(9770);}}
    public static class Impl9771 implements DemoService {public int type() {return 9771;} public void run() {System.out.println(9771);}}
    public static class Impl9772 implements DemoService {public int type() {return 9772;} public void run() {System.out.println(9772);}}
    public static class Impl9773 implements DemoService {public int type() {return 9773;} public void run() {System.out.println(9773);}}
    public static class Impl9774 implements DemoService {public int type() {return 9774;} public void run() {System.out.println(9774);}}
    public static class Impl9775 implements DemoService {public int type() {return 9775;} public void run() {System.out.println(9775);}}
    public static class Impl9776 implements DemoService {public int type() {return 9776;} public void run() {System.out.println(9776);}}
    public static class Impl9777 implements DemoService {public int type() {return 9777;} public void run() {System.out.println(9777);}}
    public static class Impl9778 implements DemoService {public int type() {return 9778;} public void run() {System.out.println(9778);}}
    public static class Impl9779 implements DemoService {public int type() {return 9779;} public void run() {System.out.println(9779);}}
    public static class Impl9780 implements DemoService {public int type() {return 9780;} public void run() {System.out.println(9780);}}
    public static class Impl9781 implements DemoService {public int type() {return 9781;} public void run() {System.out.println(9781);}}
    public static class Impl9782 implements DemoService {public int type() {return 9782;} public void run() {System.out.println(9782);}}
    public static class Impl9783 implements DemoService {public int type() {return 9783;} public void run() {System.out.println(9783);}}
    public static class Impl9784 implements DemoService {public int type() {return 9784;} public void run() {System.out.println(9784);}}
    public static class Impl9785 implements DemoService {public int type() {return 9785;} public void run() {System.out.println(9785);}}
    public static class Impl9786 implements DemoService {public int type() {return 9786;} public void run() {System.out.println(9786);}}
    public static class Impl9787 implements DemoService {public int type() {return 9787;} public void run() {System.out.println(9787);}}
    public static class Impl9788 implements DemoService {public int type() {return 9788;} public void run() {System.out.println(9788);}}
    public static class Impl9789 implements DemoService {public int type() {return 9789;} public void run() {System.out.println(9789);}}
    public static class Impl9790 implements DemoService {public int type() {return 9790;} public void run() {System.out.println(9790);}}
    public static class Impl9791 implements DemoService {public int type() {return 9791;} public void run() {System.out.println(9791);}}
    public static class Impl9792 implements DemoService {public int type() {return 9792;} public void run() {System.out.println(9792);}}
    public static class Impl9793 implements DemoService {public int type() {return 9793;} public void run() {System.out.println(9793);}}
    public static class Impl9794 implements DemoService {public int type() {return 9794;} public void run() {System.out.println(9794);}}
    public static class Impl9795 implements DemoService {public int type() {return 9795;} public void run() {System.out.println(9795);}}
    public static class Impl9796 implements DemoService {public int type() {return 9796;} public void run() {System.out.println(9796);}}
    public static class Impl9797 implements DemoService {public int type() {return 9797;} public void run() {System.out.println(9797);}}
    public static class Impl9798 implements DemoService {public int type() {return 9798;} public void run() {System.out.println(9798);}}
    public static class Impl9799 implements DemoService {public int type() {return 9799;} public void run() {System.out.println(9799);}}
    public static class Impl9800 implements DemoService {public int type() {return 9800;} public void run() {System.out.println(9800);}}
    public static class Impl9801 implements DemoService {public int type() {return 9801;} public void run() {System.out.println(9801);}}
    public static class Impl9802 implements DemoService {public int type() {return 9802;} public void run() {System.out.println(9802);}}
    public static class Impl9803 implements DemoService {public int type() {return 9803;} public void run() {System.out.println(9803);}}
    public static class Impl9804 implements DemoService {public int type() {return 9804;} public void run() {System.out.println(9804);}}
    public static class Impl9805 implements DemoService {public int type() {return 9805;} public void run() {System.out.println(9805);}}
    public static class Impl9806 implements DemoService {public int type() {return 9806;} public void run() {System.out.println(9806);}}
    public static class Impl9807 implements DemoService {public int type() {return 9807;} public void run() {System.out.println(9807);}}
    public static class Impl9808 implements DemoService {public int type() {return 9808;} public void run() {System.out.println(9808);}}
    public static class Impl9809 implements DemoService {public int type() {return 9809;} public void run() {System.out.println(9809);}}
    public static class Impl9810 implements DemoService {public int type() {return 9810;} public void run() {System.out.println(9810);}}
    public static class Impl9811 implements DemoService {public int type() {return 9811;} public void run() {System.out.println(9811);}}
    public static class Impl9812 implements DemoService {public int type() {return 9812;} public void run() {System.out.println(9812);}}
    public static class Impl9813 implements DemoService {public int type() {return 9813;} public void run() {System.out.println(9813);}}
    public static class Impl9814 implements DemoService {public int type() {return 9814;} public void run() {System.out.println(9814);}}
    public static class Impl9815 implements DemoService {public int type() {return 9815;} public void run() {System.out.println(9815);}}
    public static class Impl9816 implements DemoService {public int type() {return 9816;} public void run() {System.out.println(9816);}}
    public static class Impl9817 implements DemoService {public int type() {return 9817;} public void run() {System.out.println(9817);}}
    public static class Impl9818 implements DemoService {public int type() {return 9818;} public void run() {System.out.println(9818);}}
    public static class Impl9819 implements DemoService {public int type() {return 9819;} public void run() {System.out.println(9819);}}
    public static class Impl9820 implements DemoService {public int type() {return 9820;} public void run() {System.out.println(9820);}}
    public static class Impl9821 implements DemoService {public int type() {return 9821;} public void run() {System.out.println(9821);}}
    public static class Impl9822 implements DemoService {public int type() {return 9822;} public void run() {System.out.println(9822);}}
    public static class Impl9823 implements DemoService {public int type() {return 9823;} public void run() {System.out.println(9823);}}
    public static class Impl9824 implements DemoService {public int type() {return 9824;} public void run() {System.out.println(9824);}}
    public static class Impl9825 implements DemoService {public int type() {return 9825;} public void run() {System.out.println(9825);}}
    public static class Impl9826 implements DemoService {public int type() {return 9826;} public void run() {System.out.println(9826);}}
    public static class Impl9827 implements DemoService {public int type() {return 9827;} public void run() {System.out.println(9827);}}
    public static class Impl9828 implements DemoService {public int type() {return 9828;} public void run() {System.out.println(9828);}}
    public static class Impl9829 implements DemoService {public int type() {return 9829;} public void run() {System.out.println(9829);}}
    public static class Impl9830 implements DemoService {public int type() {return 9830;} public void run() {System.out.println(9830);}}
    public static class Impl9831 implements DemoService {public int type() {return 9831;} public void run() {System.out.println(9831);}}
    public static class Impl9832 implements DemoService {public int type() {return 9832;} public void run() {System.out.println(9832);}}
    public static class Impl9833 implements DemoService {public int type() {return 9833;} public void run() {System.out.println(9833);}}
    public static class Impl9834 implements DemoService {public int type() {return 9834;} public void run() {System.out.println(9834);}}
    public static class Impl9835 implements DemoService {public int type() {return 9835;} public void run() {System.out.println(9835);}}
    public static class Impl9836 implements DemoService {public int type() {return 9836;} public void run() {System.out.println(9836);}}
    public static class Impl9837 implements DemoService {public int type() {return 9837;} public void run() {System.out.println(9837);}}
    public static class Impl9838 implements DemoService {public int type() {return 9838;} public void run() {System.out.println(9838);}}
    public static class Impl9839 implements DemoService {public int type() {return 9839;} public void run() {System.out.println(9839);}}
    public static class Impl9840 implements DemoService {public int type() {return 9840;} public void run() {System.out.println(9840);}}
    public static class Impl9841 implements DemoService {public int type() {return 9841;} public void run() {System.out.println(9841);}}
    public static class Impl9842 implements DemoService {public int type() {return 9842;} public void run() {System.out.println(9842);}}
    public static class Impl9843 implements DemoService {public int type() {return 9843;} public void run() {System.out.println(9843);}}
    public static class Impl9844 implements DemoService {public int type() {return 9844;} public void run() {System.out.println(9844);}}
    public static class Impl9845 implements DemoService {public int type() {return 9845;} public void run() {System.out.println(9845);}}
    public static class Impl9846 implements DemoService {public int type() {return 9846;} public void run() {System.out.println(9846);}}
    public static class Impl9847 implements DemoService {public int type() {return 9847;} public void run() {System.out.println(9847);}}
    public static class Impl9848 implements DemoService {public int type() {return 9848;} public void run() {System.out.println(9848);}}
    public static class Impl9849 implements DemoService {public int type() {return 9849;} public void run() {System.out.println(9849);}}
    public static class Impl9850 implements DemoService {public int type() {return 9850;} public void run() {System.out.println(9850);}}
    public static class Impl9851 implements DemoService {public int type() {return 9851;} public void run() {System.out.println(9851);}}
    public static class Impl9852 implements DemoService {public int type() {return 9852;} public void run() {System.out.println(9852);}}
    public static class Impl9853 implements DemoService {public int type() {return 9853;} public void run() {System.out.println(9853);}}
    public static class Impl9854 implements DemoService {public int type() {return 9854;} public void run() {System.out.println(9854);}}
    public static class Impl9855 implements DemoService {public int type() {return 9855;} public void run() {System.out.println(9855);}}
    public static class Impl9856 implements DemoService {public int type() {return 9856;} public void run() {System.out.println(9856);}}
    public static class Impl9857 implements DemoService {public int type() {return 9857;} public void run() {System.out.println(9857);}}
    public static class Impl9858 implements DemoService {public int type() {return 9858;} public void run() {System.out.println(9858);}}
    public static class Impl9859 implements DemoService {public int type() {return 9859;} public void run() {System.out.println(9859);}}
    public static class Impl9860 implements DemoService {public int type() {return 9860;} public void run() {System.out.println(9860);}}
    public static class Impl9861 implements DemoService {public int type() {return 9861;} public void run() {System.out.println(9861);}}
    public static class Impl9862 implements DemoService {public int type() {return 9862;} public void run() {System.out.println(9862);}}
    public static class Impl9863 implements DemoService {public int type() {return 9863;} public void run() {System.out.println(9863);}}
    public static class Impl9864 implements DemoService {public int type() {return 9864;} public void run() {System.out.println(9864);}}
    public static class Impl9865 implements DemoService {public int type() {return 9865;} public void run() {System.out.println(9865);}}
    public static class Impl9866 implements DemoService {public int type() {return 9866;} public void run() {System.out.println(9866);}}
    public static class Impl9867 implements DemoService {public int type() {return 9867;} public void run() {System.out.println(9867);}}
    public static class Impl9868 implements DemoService {public int type() {return 9868;} public void run() {System.out.println(9868);}}
    public static class Impl9869 implements DemoService {public int type() {return 9869;} public void run() {System.out.println(9869);}}
    public static class Impl9870 implements DemoService {public int type() {return 9870;} public void run() {System.out.println(9870);}}
    public static class Impl9871 implements DemoService {public int type() {return 9871;} public void run() {System.out.println(9871);}}
    public static class Impl9872 implements DemoService {public int type() {return 9872;} public void run() {System.out.println(9872);}}
    public static class Impl9873 implements DemoService {public int type() {return 9873;} public void run() {System.out.println(9873);}}
    public static class Impl9874 implements DemoService {public int type() {return 9874;} public void run() {System.out.println(9874);}}
    public static class Impl9875 implements DemoService {public int type() {return 9875;} public void run() {System.out.println(9875);}}
    public static class Impl9876 implements DemoService {public int type() {return 9876;} public void run() {System.out.println(9876);}}
    public static class Impl9877 implements DemoService {public int type() {return 9877;} public void run() {System.out.println(9877);}}
    public static class Impl9878 implements DemoService {public int type() {return 9878;} public void run() {System.out.println(9878);}}
    public static class Impl9879 implements DemoService {public int type() {return 9879;} public void run() {System.out.println(9879);}}
    public static class Impl9880 implements DemoService {public int type() {return 9880;} public void run() {System.out.println(9880);}}
    public static class Impl9881 implements DemoService {public int type() {return 9881;} public void run() {System.out.println(9881);}}
    public static class Impl9882 implements DemoService {public int type() {return 9882;} public void run() {System.out.println(9882);}}
    public static class Impl9883 implements DemoService {public int type() {return 9883;} public void run() {System.out.println(9883);}}
    public static class Impl9884 implements DemoService {public int type() {return 9884;} public void run() {System.out.println(9884);}}
    public static class Impl9885 implements DemoService {public int type() {return 9885;} public void run() {System.out.println(9885);}}
    public static class Impl9886 implements DemoService {public int type() {return 9886;} public void run() {System.out.println(9886);}}
    public static class Impl9887 implements DemoService {public int type() {return 9887;} public void run() {System.out.println(9887);}}
    public static class Impl9888 implements DemoService {public int type() {return 9888;} public void run() {System.out.println(9888);}}
    public static class Impl9889 implements DemoService {public int type() {return 9889;} public void run() {System.out.println(9889);}}
    public static class Impl9890 implements DemoService {public int type() {return 9890;} public void run() {System.out.println(9890);}}
    public static class Impl9891 implements DemoService {public int type() {return 9891;} public void run() {System.out.println(9891);}}
    public static class Impl9892 implements DemoService {public int type() {return 9892;} public void run() {System.out.println(9892);}}
    public static class Impl9893 implements DemoService {public int type() {return 9893;} public void run() {System.out.println(9893);}}
    public static class Impl9894 implements DemoService {public int type() {return 9894;} public void run() {System.out.println(9894);}}
    public static class Impl9895 implements DemoService {public int type() {return 9895;} public void run() {System.out.println(9895);}}
    public static class Impl9896 implements DemoService {public int type() {return 9896;} public void run() {System.out.println(9896);}}
    public static class Impl9897 implements DemoService {public int type() {return 9897;} public void run() {System.out.println(9897);}}
    public static class Impl9898 implements DemoService {public int type() {return 9898;} public void run() {System.out.println(9898);}}
    public static class Impl9899 implements DemoService {public int type() {return 9899;} public void run() {System.out.println(9899);}}
    public static class Impl9900 implements DemoService {public int type() {return 9900;} public void run() {System.out.println(9900);}}
    public static class Impl9901 implements DemoService {public int type() {return 9901;} public void run() {System.out.println(9901);}}
    public static class Impl9902 implements DemoService {public int type() {return 9902;} public void run() {System.out.println(9902);}}
    public static class Impl9903 implements DemoService {public int type() {return 9903;} public void run() {System.out.println(9903);}}
    public static class Impl9904 implements DemoService {public int type() {return 9904;} public void run() {System.out.println(9904);}}
    public static class Impl9905 implements DemoService {public int type() {return 9905;} public void run() {System.out.println(9905);}}
    public static class Impl9906 implements DemoService {public int type() {return 9906;} public void run() {System.out.println(9906);}}
    public static class Impl9907 implements DemoService {public int type() {return 9907;} public void run() {System.out.println(9907);}}
    public static class Impl9908 implements DemoService {public int type() {return 9908;} public void run() {System.out.println(9908);}}
    public static class Impl9909 implements DemoService {public int type() {return 9909;} public void run() {System.out.println(9909);}}
    public static class Impl9910 implements DemoService {public int type() {return 9910;} public void run() {System.out.println(9910);}}
    public static class Impl9911 implements DemoService {public int type() {return 9911;} public void run() {System.out.println(9911);}}
    public static class Impl9912 implements DemoService {public int type() {return 9912;} public void run() {System.out.println(9912);}}
    public static class Impl9913 implements DemoService {public int type() {return 9913;} public void run() {System.out.println(9913);}}
    public static class Impl9914 implements DemoService {public int type() {return 9914;} public void run() {System.out.println(9914);}}
    public static class Impl9915 implements DemoService {public int type() {return 9915;} public void run() {System.out.println(9915);}}
    public static class Impl9916 implements DemoService {public int type() {return 9916;} public void run() {System.out.println(9916);}}
    public static class Impl9917 implements DemoService {public int type() {return 9917;} public void run() {System.out.println(9917);}}
    public static class Impl9918 implements DemoService {public int type() {return 9918;} public void run() {System.out.println(9918);}}
    public static class Impl9919 implements DemoService {public int type() {return 9919;} public void run() {System.out.println(9919);}}
    public static class Impl9920 implements DemoService {public int type() {return 9920;} public void run() {System.out.println(9920);}}
    public static class Impl9921 implements DemoService {public int type() {return 9921;} public void run() {System.out.println(9921);}}
    public static class Impl9922 implements DemoService {public int type() {return 9922;} public void run() {System.out.println(9922);}}
    public static class Impl9923 implements DemoService {public int type() {return 9923;} public void run() {System.out.println(9923);}}
    public static class Impl9924 implements DemoService {public int type() {return 9924;} public void run() {System.out.println(9924);}}
    public static class Impl9925 implements DemoService {public int type() {return 9925;} public void run() {System.out.println(9925);}}
    public static class Impl9926 implements DemoService {public int type() {return 9926;} public void run() {System.out.println(9926);}}
    public static class Impl9927 implements DemoService {public int type() {return 9927;} public void run() {System.out.println(9927);}}
    public static class Impl9928 implements DemoService {public int type() {return 9928;} public void run() {System.out.println(9928);}}
    public static class Impl9929 implements DemoService {public int type() {return 9929;} public void run() {System.out.println(9929);}}
    public static class Impl9930 implements DemoService {public int type() {return 9930;} public void run() {System.out.println(9930);}}
    public static class Impl9931 implements DemoService {public int type() {return 9931;} public void run() {System.out.println(9931);}}
    public static class Impl9932 implements DemoService {public int type() {return 9932;} public void run() {System.out.println(9932);}}
    public static class Impl9933 implements DemoService {public int type() {return 9933;} public void run() {System.out.println(9933);}}
    public static class Impl9934 implements DemoService {public int type() {return 9934;} public void run() {System.out.println(9934);}}
    public static class Impl9935 implements DemoService {public int type() {return 9935;} public void run() {System.out.println(9935);}}
    public static class Impl9936 implements DemoService {public int type() {return 9936;} public void run() {System.out.println(9936);}}
    public static class Impl9937 implements DemoService {public int type() {return 9937;} public void run() {System.out.println(9937);}}
    public static class Impl9938 implements DemoService {public int type() {return 9938;} public void run() {System.out.println(9938);}}
    public static class Impl9939 implements DemoService {public int type() {return 9939;} public void run() {System.out.println(9939);}}
    public static class Impl9940 implements DemoService {public int type() {return 9940;} public void run() {System.out.println(9940);}}
    public static class Impl9941 implements DemoService {public int type() {return 9941;} public void run() {System.out.println(9941);}}
    public static class Impl9942 implements DemoService {public int type() {return 9942;} public void run() {System.out.println(9942);}}
    public static class Impl9943 implements DemoService {public int type() {return 9943;} public void run() {System.out.println(9943);}}
    public static class Impl9944 implements DemoService {public int type() {return 9944;} public void run() {System.out.println(9944);}}
    public static class Impl9945 implements DemoService {public int type() {return 9945;} public void run() {System.out.println(9945);}}
    public static class Impl9946 implements DemoService {public int type() {return 9946;} public void run() {System.out.println(9946);}}
    public static class Impl9947 implements DemoService {public int type() {return 9947;} public void run() {System.out.println(9947);}}
    public static class Impl9948 implements DemoService {public int type() {return 9948;} public void run() {System.out.println(9948);}}
    public static class Impl9949 implements DemoService {public int type() {return 9949;} public void run() {System.out.println(9949);}}
    public static class Impl9950 implements DemoService {public int type() {return 9950;} public void run() {System.out.println(9950);}}
    public static class Impl9951 implements DemoService {public int type() {return 9951;} public void run() {System.out.println(9951);}}
    public static class Impl9952 implements DemoService {public int type() {return 9952;} public void run() {System.out.println(9952);}}
    public static class Impl9953 implements DemoService {public int type() {return 9953;} public void run() {System.out.println(9953);}}
    public static class Impl9954 implements DemoService {public int type() {return 9954;} public void run() {System.out.println(9954);}}
    public static class Impl9955 implements DemoService {public int type() {return 9955;} public void run() {System.out.println(9955);}}
    public static class Impl9956 implements DemoService {public int type() {return 9956;} public void run() {System.out.println(9956);}}
    public static class Impl9957 implements DemoService {public int type() {return 9957;} public void run() {System.out.println(9957);}}
    public static class Impl9958 implements DemoService {public int type() {return 9958;} public void run() {System.out.println(9958);}}
    public static class Impl9959 implements DemoService {public int type() {return 9959;} public void run() {System.out.println(9959);}}
    public static class Impl9960 implements DemoService {public int type() {return 9960;} public void run() {System.out.println(9960);}}
    public static class Impl9961 implements DemoService {public int type() {return 9961;} public void run() {System.out.println(9961);}}
    public static class Impl9962 implements DemoService {public int type() {return 9962;} public void run() {System.out.println(9962);}}
    public static class Impl9963 implements DemoService {public int type() {return 9963;} public void run() {System.out.println(9963);}}
    public static class Impl9964 implements DemoService {public int type() {return 9964;} public void run() {System.out.println(9964);}}
    public static class Impl9965 implements DemoService {public int type() {return 9965;} public void run() {System.out.println(9965);}}
    public static class Impl9966 implements DemoService {public int type() {return 9966;} public void run() {System.out.println(9966);}}
    public static class Impl9967 implements DemoService {public int type() {return 9967;} public void run() {System.out.println(9967);}}
    public static class Impl9968 implements DemoService {public int type() {return 9968;} public void run() {System.out.println(9968);}}
    public static class Impl9969 implements DemoService {public int type() {return 9969;} public void run() {System.out.println(9969);}}
    public static class Impl9970 implements DemoService {public int type() {return 9970;} public void run() {System.out.println(9970);}}
    public static class Impl9971 implements DemoService {public int type() {return 9971;} public void run() {System.out.println(9971);}}
    public static class Impl9972 implements DemoService {public int type() {return 9972;} public void run() {System.out.println(9972);}}
    public static class Impl9973 implements DemoService {public int type() {return 9973;} public void run() {System.out.println(9973);}}
    public static class Impl9974 implements DemoService {public int type() {return 9974;} public void run() {System.out.println(9974);}}
    public static class Impl9975 implements DemoService {public int type() {return 9975;} public void run() {System.out.println(9975);}}
    public static class Impl9976 implements DemoService {public int type() {return 9976;} public void run() {System.out.println(9976);}}
    public static class Impl9977 implements DemoService {public int type() {return 9977;} public void run() {System.out.println(9977);}}
    public static class Impl9978 implements DemoService {public int type() {return 9978;} public void run() {System.out.println(9978);}}
    public static class Impl9979 implements DemoService {public int type() {return 9979;} public void run() {System.out.println(9979);}}
    public static class Impl9980 implements DemoService {public int type() {return 9980;} public void run() {System.out.println(9980);}}
    public static class Impl9981 implements DemoService {public int type() {return 9981;} public void run() {System.out.println(9981);}}
    public static class Impl9982 implements DemoService {public int type() {return 9982;} public void run() {System.out.println(9982);}}
    public static class Impl9983 implements DemoService {public int type() {return 9983;} public void run() {System.out.println(9983);}}
    public static class Impl9984 implements DemoService {public int type() {return 9984;} public void run() {System.out.println(9984);}}
    public static class Impl9985 implements DemoService {public int type() {return 9985;} public void run() {System.out.println(9985);}}
    public static class Impl9986 implements DemoService {public int type() {return 9986;} public void run() {System.out.println(9986);}}
    public static class Impl9987 implements DemoService {public int type() {return 9987;} public void run() {System.out.println(9987);}}
    public static class Impl9988 implements DemoService {public int type() {return 9988;} public void run() {System.out.println(9988);}}
    public static class Impl9989 implements DemoService {public int type() {return 9989;} public void run() {System.out.println(9989);}}
    public static class Impl9990 implements DemoService {public int type() {return 9990;} public void run() {System.out.println(9990);}}
    public static class Impl9991 implements DemoService {public int type() {return 9991;} public void run() {System.out.println(9991);}}
    public static class Impl9992 implements DemoService {public int type() {return 9992;} public void run() {System.out.println(9992);}}
    public static class Impl9993 implements DemoService {public int type() {return 9993;} public void run() {System.out.println(9993);}}
    public static class Impl9994 implements DemoService {public int type() {return 9994;} public void run() {System.out.println(9994);}}
    public static class Impl9995 implements DemoService {public int type() {return 9995;} public void run() {System.out.println(9995);}}
    public static class Impl9996 implements DemoService {public int type() {return 9996;} public void run() {System.out.println(9996);}}
    public static class Impl9997 implements DemoService {public int type() {return 9997;} public void run() {System.out.println(9997);}}
    public static class Impl9998 implements DemoService {public int type() {return 9998;} public void run() {System.out.println(9998);}}
    public static class Impl9999 implements DemoService {public int type() {return 9999;} public void run() {System.out.println(9999);}}
}
